#ifndef _CPLANE_H_
#define _CPLANE_H_

#include <qimage.h>
#include <qtextstream.h>

#include <string>
#include <vector>
#include <sstream>
#include <iostream>
#include <fstream>
#include <cassert>

using namespace std;

static const int ERS7_FRAME_WIDTH = 208;
static const int ERS7_FRAME_HEIGHT = 160;

static const QColor colorRed("red");
static const QColor colorBlue("blue");
static const QColor colorOrange("orange");
static const QColor colorGreen("green");
static const QColor colorYellow("yellow");
static const QColor colorPink("pink");
static const QColor colorGrey("grey");
static const QColor colorWhite("white");
static const QColor colorBlack("black");

static const QColor colorTable[] = {
    colorOrange,
    colorBlue,
    colorGreen,
    colorYellow,
    colorPink,
    colorBlue,
    colorRed,
    colorGreen,
    colorGrey,
    colorWhite,
    colorBlack,
    colorBlack
};



class CPlane{
    public:
        //WARNING: heuristic , in fact 160*2 < minimumCPlaneSize < 160*204*2
        static const int minimumCPlaneSize = 160;
        static const int maximumCPlaneSize = 60000;

        CPlane(int width = ERS7_FRAME_WIDTH, int height = ERS7_FRAME_HEIGHT){
            this->width = width;
            this->height = height;                
            cplaneBlobArray.clear();
            image = new QImage(width, height,32);
        }
        ~CPlane(){
            delete(image);
        }

        bool parseCPlaneFromStream(istream &ss){
            cplaneBlobArray.clear();

            int u = -1 , v = -1;
            int size = -1;
            ss >> size;

            int readSize = 0;
            while (!ss.eof() && readSize < size ){
                u = -1; v = -1;
                ss >> u >> v;
                if (u == -1 || v == -1) break;
                readSize += 2;
                cplaneBlobArray.push_back(make_pair(u,v));
            }
            ss.ignore(100000,'\n');
            return readSize == size;
        }

        bool parseCPlaneFromLine(string line){
            istringstream ss(line);
            return parseCPlaneFromStream(ss);
        }

        bool parseCPlaneFromCompressedPChar(char *s, int size){
            cplaneBlobArray.clear();

            int compressedLength = *(int *)s;
            /*
            for (int k = 0;k<4;k++)
                cout << (int)s[k] << ", ";
            cout << endl;
            */
            cout << "Compressed Length = " << compressedLength << endl;
            //assert(compressedLength % 2 == 0);
            int i = 4, j = 0;
            if (compressedLength > minimumCPlaneSize && compressedLength < maximumCPlaneSize && compressedLength % 2 == 0){
                for (j =0;j<compressedLength && i < size; j += 2, i+= 2){
                    unsigned char count = (unsigned char)s[i];
                    unsigned char color = (unsigned char)s[i+1];
                    cplaneBlobArray.push_back(make_pair(count,color));
                }
                cout << " -- " << size - compressedLength << " bytes left " << endl;
            }
            else 
                cout << "Discard , wrong size " << compressedLength << endl;
            return (j==compressedLength);
        }

        QImage *toImage(){
            unsigned int i = 0 ;       
            int color = 0 ;
            int count = 0 ;
            for (int y = 0; y < height;y++)
                for (int x = 0; x < width ; x++){
                    if (count == 0){
                        if (i >= cplaneBlobArray.size()){
                            count = 1000;
                            count = 0;
                        }
                        else{
                            count = cplaneBlobArray[i].first;
                            color = cplaneBlobArray[i].second;                       
                            i += 1;
                        }
                    }
                    image->setPixel(x,y, colorTable[color].rgb());
                    count -= 1;
                }
            return image;
        }

        QImage *getImage(){
            image = toImage();
            return image;
        }

        void saveToOpenedFile(QTextStream &outStream){
            //cout << "Saving " << cplaneBlobArray.size() << endl;
            outStream << 2 * cplaneBlobArray.size() << " "  ;
            for (unsigned int i=0;i < cplaneBlobArray.size() ; i++){
                outStream << cplaneBlobArray[i].first << " " << cplaneBlobArray[i].second << " " ;
            }
            //cout << " -- done saving " << endl;
            outStream << endl;
        }

    private:
        int width, height;
        vector<pair<int,int> > cplaneBlobArray;
        QImage *image;


};


#endif // _CPLANE_H_
