#include "CPlaneDisplay.h"
#include "CPlane.h"
#include <string>
#include <vector>
#include <sstream>
#include <fstream>
#include <iostream>
#include <cassert>

#include <qdatetime.h>
#include <qtextstream.h>
#include <qcheckbox.h>
#include <qcombobox.h>
#include "../../../../robot7/share/VisionDef.h"

using namespace std;

static bool bDebugParsing = false;
int numObject; //HACK: should put into class variable instead of global
bool bUpdateCPlane = true;//HACK: should put into class variable instead of global
char *lastObjectData; //to display when not updating

vector<OPlane> seenObject;

enum ReadingStateEnum{
    rsNoReading,
    rsCPlaneSize,
    rsCPlane,
    rsSensors,
    rsSanity,
    rsNumObject,
    rsObjects,
    rsAllState
};

int stateSize[rsAllState] = {
    0,
    4,
    -1,//mean not detemined yet, depends on other state
    12,
    28,
    4,
    -1
};

int readingState;

/* 
 *  Constructs a CPlaneDisplay which is a child of 'parent', with the 
 *  name 'name' and widget flags set to 'f' 
 */
CPlaneDisplay::CPlaneDisplay(const char *logFileName, QWidget* parent,  const char* name, WFlags fl )
    : FormCPlaneDisplay( parent, name, fl ), fLogFile(logFileName){
        //loadFromFile(logFileName);
        this->logFileName = logFileName;
        bLogToFile = false;
        cout << "Got " << frames.size() << " frames " << endl;
        pixmap = NULL;
        currentFrame = 0;
        timer = new QTimer();
        clientSocket = new QSocket();
        serverSocket = NULL;	
        readingState = rsNoReading;
        slotListen();
        objectCheckBoxes.push_back(chbBall);
        objectCheckBoxes.push_back(chbBlueGoal);
        objectCheckBoxes.push_back(chbYellowGoal);
        objectCheckBoxes.push_back(chbBluePink);
        objectCheckBoxes.push_back(chbPinkBlue);
        objectCheckBoxes.push_back(NULL);
        objectCheckBoxes.push_back(NULL);
        objectCheckBoxes.push_back(chbYellowPink);
        objectCheckBoxes.push_back(chbPinkYellow);
        objectCheckBoxes.push_back(chbRed1);
        objectCheckBoxes.push_back(chbRed2);
        objectCheckBoxes.push_back(chbRed3);
        objectCheckBoxes.push_back(chbRed4);
        objectCheckBoxes.push_back(chbBlue1);
        objectCheckBoxes.push_back(chbBlue2);
        objectCheckBoxes.push_back(chbBlue3);
        objectCheckBoxes.push_back(chbBlue4);

        cbxObject->clear();
        for (unsigned int i=0;i<objectCheckBoxes.size();i++)
            if (objectCheckBoxes[i] != NULL){
                cbxObject->insertItem(objectCheckBoxes[i]->text());
            }
            else 
                cbxObject->insertItem("NULL");

        currentObject = cbxObject->currentItem();
        //playNext();
    }

/*
   void CPlaneDisplay::paintEvent(QPaintEvent *paintEvent){
   paintEvent = paintEvent;
   if (pixmap != NULL) 
   bitBlt((QWidget*)lblCPlane,0,0,pixmap);
   }
 */

/*  
 *  Destroys the object and frees any allocated resources
 */
CPlaneDisplay::~CPlaneDisplay(){
    cout << " Destroying CPlaneDisplay" << endl;
    delete (timer);
    for (unsigned int i=0;i<frames.size();i++){
        delete(frames[i]);
    }
    delete (pixmap);
    delete (clientSocket);
    delete (serverSocket);
}

bool CPlaneDisplay::loadFromFile(const string &fileName){
    QTime clock;
    int readingTime = 0;
    int imagingTime = 0;
    
    ifstream f(fileName.c_str());
    if (! f.is_open()){
        cout << "Error opening file " << endl;
        return false;
    }

    f >> version;
    CPlane *cplane = NULL;
    while (1){
        clock.start();
        cplane = new CPlane();
        if (cplane->parseCPlaneFromStream(f))
            frames.push_back(cplane);
        else
            break;

        readingTime += clock.elapsed();
        clock.start();
        cplane->getImage();
        imagingTime += clock.elapsed();

    }
    cout << "Reading Time = " << readingTime << endl;
    cout << "Imaging Time = " << imagingTime << endl;

    return true;
}

int CPlaneDisplay::playBackLength(){
    return frames.size();
}


/*
 * public slot
 */
void CPlaneDisplay::slotBack() {
    qWarning( "CPlaneDisplay::slotBack() not yet implemented!" );
}

/*
 * public slot
 */
void CPlaneDisplay::slotForward() {
    qWarning( "CPlaneDisplay::slotForward() not yet implemented!" );
}

/*
 * public slot
 */
void CPlaneDisplay::slotAnimate() {
    qWarning( "CPlaneDisplay::slotAnimate() ");
    if (timer)
        delete(timer);
    timer = new QTimer();
    if (currentFrame == playBackLength())
        currentFrame = 0;
    connect(timer,SIGNAL(timeout()),this,SLOT(playNext()));
    timer->start(40, false);
}

void CPlaneDisplay::playNext(){
    if ( (unsigned int) currentFrame >= frames.size() ){
        slotStop();
        return;
    }
    cout << "Current = " << currentFrame << endl;
    int w = lblCPlane->width();
    int h = lblCPlane->height();        
    pixmap = new QPixmap(frames[currentFrame]->getImage()->scale(w,h) );
    //TODO try repaint()
    bitBlt((QWidget*)lblCPlane,0,0,pixmap);
    delete(pixmap);
    currentFrame += 1;
}

void CPlaneDisplay::slotStop(){
    timer->stop() ;     
}

void CPlaneDisplay::slotConnect(){
    
    cout << "slotConnect" << endl;
    QString ip = ledIP->text();
    cout << ip << endl;
    int port = ledPort->text().toInt();
    connect( clientSocket, SIGNAL(readyRead()),
                this, SLOT(readRemoteCPlane()) );
    clientSocket->connectToHost(ip,port);
    cout << "state = " << clientSocket->state() << endl;
    cout << "peerName = " << clientSocket->peerName() << endl;
    cout << "Connection established , local port = " << clientSocket->port() << endl;
}
    
void CPlaneDisplay::slotListen(){
    cout << "CPlaneDisplay::slotListen" << endl;
    int port = ledPort->text().toInt();
    cout << "port " << port << endl;
    if (serverSocket)
        delete(serverSocket);
    serverSocket = new SimpleServer(port);
    connect( serverSocket, SIGNAL(newConnect(ClientSocket*)),
                this, SLOT(newConnection(ClientSocket*)) );
    cout << "I'm listening on " << port << endl;

}

void CPlaneDisplay::newConnection(ClientSocket *socket){
    cout << "New connection " << endl;
    clientSocket = socket;
    connect( socket, SIGNAL(readyRead()),
                this, SLOT(readRemoteCPlane()) );

}

static const int QUEUE_SIZE = 2 * CPlane::maximumCPlaneSize;
static const int DUPLICATE_CPLANE = CPlane::maximumCPlaneSize;
static char buffer[QUEUE_SIZE + DUPLICATE_CPLANE];
static int currentPos = 0, firstPos =0 , lastPos = 0;

int getQueueSize(){
    if (lastPos >= firstPos ){
        return lastPos - firstPos;
    }
    else if (lastPos < firstPos ){
        return lastPos + QUEUE_SIZE - firstPos;
    }
    return 0;
}

bool enoughByteInQueue(QSocket *clientSocket, int size){
    int queueSize = getQueueSize();

    if (queueSize >= size)
        return true;
    
    if ( (queueSize + (int) clientSocket->bytesAvailable()) >= size ){
        int needed = size - queueSize;
        QDataStream ts(clientSocket);
        if (lastPos + needed >= QUEUE_SIZE){ //read in 2 pieces
            ts.readRawBytes(buffer + lastPos, QUEUE_SIZE - lastPos);
            ts.readRawBytes(buffer , needed - (QUEUE_SIZE - lastPos));
            lastPos += needed - QUEUE_SIZE;
        }
        else{ //reading just one piece
            ts.readRawBytes(buffer + lastPos, needed);
            lastPos += needed;
        }

        if (bDebugParsing){
            cout << "After reading " << clientSocket->bytesAvailable() << endl;
            cout << "Reading in " << needed << " last pos = " << lastPos << endl;
        }
        return true;
    }

    return false;
}

void clearByteInQueue(int size){
    if (bDebugParsing)
        cout << "Clearing " << size << endl;
    if (getQueueSize()  < size ){
        cout << "WARNING: clear shorted buffer" << currentPos << "< " << size << endl;
        return;
    }
    firstPos += size;
    if (firstPos >= QUEUE_SIZE)
        firstPos -= QUEUE_SIZE;

}


void checkMissingEndThenDuplicate(int endPoint){
    if (endPoint >= QUEUE_SIZE){
        cout << "Duplicate " << endPoint - QUEUE_SIZE << endl;
        for (int i=QUEUE_SIZE; i < endPoint;i++)
            buffer[i] = buffer[i-QUEUE_SIZE];
    }
}

void CPlaneDisplay::parseCPlaneInState(){
    if (bDebugParsing)
        cout << "Parsing state " << readingState << " firstPos = " << firstPos << endl;
    char *p = buffer + firstPos;
    switch (readingState){
        case rsCPlaneSize:
        {
            int cpSize = *(int *) p;
            if (bDebugParsing)
                cout << " Got CPlane size = " << cpSize << endl;
            stateSize[rsCPlane] = cpSize;
            break;
        }
        case rsCPlane:
        {
            if (bDebugParsing)
                cout << "Display cplane size = " << stateSize[rsCPlane] << endl;

            if (bUpdateCPlane){
                //kim:tricky here
                int endPoint = lastPos + stateSize[rsCPlane];
                checkMissingEndThenDuplicate(endPoint);
                //JUST HACKED, should check buffer + lastPos - 4
                displayCPlane(buffer + firstPos - 4, stateSize[rsCPlane] + 4);
            }
            break;
        }
        case rsSensors:
        {
            if (bDebugParsing)
                cout << "Reading sensor " << stateSize[rsSensors] << endl;
            break;
        }
        case rsSanity:
        {
            if (bDebugParsing)
                cout << "Reading sanity " << stateSize[rsSensors] << endl;
            break;
        }
        case rsNumObject:
        {
            numObject = *(int *) p;
            if (bDebugParsing)
                cout << " Got Object size = " << numObject << endl;
            stateSize[rsObjects] = numObject * ONE_OBJ_SIZE + NUM_ROBOT_DATA * INT_BYTES;
            break;
        }
        case rsObjects:
        {
            if (bDebugParsing)
                cout << "Reading objects " << stateSize[rsObjects] << endl;
            int endPoint = lastPos + stateSize[rsObjects];
            if (bUpdateCPlane){
                checkMissingEndThenDuplicate(endPoint);
                displayObjects(buffer + firstPos, stateSize[rsObjects]);
            }
            break;
        }
        default:
        if (bDebugParsing)
            cout << " Not a valid reading state" << readingState << endl;
    }
    clearByteInQueue(stateSize[readingState]);
    if (bDebugParsing)
        cout << "Done parsing bytes left=" << firstPos << " " << lastPos << " " << lastPos - firstPos << " " << lastPos - firstPos + QUEUE_SIZE << endl;
}

void CPlaneDisplay::readRemoteCPlane(){
    QDataStream ts(clientSocket);
    /*
    */
    //static bool continuePacket = false;
    cout << "Available bytes = " << clientSocket->bytesAvailable() << endl;
    if (readingState == rsNoReading){
        readingState = rsCPlaneSize;
    }

    while ( enoughByteInQueue(clientSocket, stateSize[readingState])  ) {
        parseCPlaneInState();
        readingState = ReadingStateEnum ( (int) readingState + 1 );
        if (readingState == rsAllState) readingState = rsCPlaneSize;
    }
}

/*
void CPlaneDisplay::parseSensorsSanityObjects(char *data, int size){
    if  (size < (NUM_SENSOR_COORDS + NUM_SANITY_COORDS + 2) * INT_BYTES)
        return ; //Not enough data
    cout << "Parsing " << size << endl;
    int t;
    for (int i = 0; i < NUM_SENSOR_COORDS ; i++){
        t = *(int *)data ;
        cout << "Got sensor " << i << " = " << t << endl;
        data += INT_BYTES; // = 4
    }
    
    for (int i = 0; i < NUM_SANITY_COORDS ; i++){
        t = *(int *)data ;
        //cout << "Got sanity " << i << " = " << t << endl;
        data += INT_BYTES; // = 4
    }
    t = *(int *)data ;
    cout << "panx = " << t << endl;
    data += INT_BYTES; // = 4
    
    t = *(int *)data ;
    cout << "tilty = " << t << endl;
    data += INT_BYTES; // = 4

    int bytesLeft = size - (NUM_SENSOR_COORDS + NUM_SANITY_COORDS + 2 ) * INT_BYTES;
    int num_obj = 0;
    cout << "Bytes Left " << bytesLeft << endl;
    if (bytesLeft % INT_BYTES == 0)
        num_obj = bytesLeft / INT_BYTES;

    cout << "Getting " << num_obj << " objects " << endl;

    for (int i = 0; i < num_obj ; i++){
        //cout << "Got object " << i << " : " << endl;
        //cout << "  (obj_index, x, y, width, height, d) = " ;
        for (int j = 0 ; j < NUM_INT_COORDS; j++){
            t = *(int *)data ;
            //cout << t << "  " ;
            data += INT_BYTES; // = 4
        }
        //cout << endl;
    }
    
    
}

*/

void CPlaneDisplay::displayCPlane(char *cplaneCompressed, int size){
    static CPlane cplane;
    static int cc = 0;
    static const int maxEle = 50;
    static int timeStamp[maxEle];
    static QTime qt;

    cout << "Displaying " << endl;
    bool ok = cplane.parseCPlaneFromCompressedPChar(cplaneCompressed,size);
    cout << "Done " << endl;
    if (ok){
        if (cc == 0){
            qt.start();
        }
        else{
            timeStamp[cc%maxEle] = qt.elapsed();
            if (cc > maxEle){
                int timePerMaxEle = timeStamp[cc%maxEle] - timeStamp[(cc+1)%maxEle];
                lcdFrameRate->display(double( maxEle * 1000.0) / timePerMaxEle);
            }
        }

        //TODO try repaint() 
        cout << "Sucessful " << cc ++ << endl;
        int w = lblCPlane->width();
        int h = lblCPlane->height();        
        if (pixmap)
            delete(pixmap);
        pixmap = new QPixmap(cplane.getImage()->scale(w,h) );
        //pixmap = new QPixmap(*(cplane.getImage()) );
        bitBlt((QWidget*)lblCPlane,0,0,pixmap);

        if (bLogToFile){
            QTextStream stream(&fLogFile);
            cplane.saveToOpenedFile(stream);
        }
    }
    else{
        cout << "Not valid cplane" << endl;
    }

}

string intToString(int x){
    ostringstream ss;
    ss << x;
    return ss.str();
}

//HACK: use global numObject instead of size
void CPlaneDisplay::displayObjects(char *objectData, int size){
    size = size;
    lastObjectData = objectData;
    int *p = (int *) objectData;
    double pan = *p++;
    pan = pan;
    double tilt = *p++;
    tilt = tilt;
    cout << "got " << numObject << " " << size << " bytes " ;
    seenObject.clear();
    for (int i=0;i<numObject;i++){
        OPlane oplane;
        oplane.objNum = *p++;
        oplane.x = *p++;
        oplane.y = *p++;
        oplane.w = *p++;
        oplane.h = *p++;
        oplane.d = *p++;
        seenObject.push_back(oplane);
    }
    displaySeenObjectOnGUI(seenObject);

}

void CPlaneDisplay::displaySeenObjectOnGUI(vector<OPlane> &seenObject){
    for (unsigned int i=0;i<objectCheckBoxes.size();i++)
        if (objectCheckBoxes[i] != NULL)
            objectCheckBoxes[i]->setChecked(false);
    bool currentObjectSeen = false;
    for (unsigned int i = 0;i< seenObject.size();i++){
        int objNum = seenObject[i].objNum;
        if (objNum == currentObject){
            currentObjectSeen = true;
            ledValueX->setText(intToString(seenObject[i].x).c_str());
            ledValueY->setText(intToString(seenObject[i].y).c_str());
            ledValueW->setText(intToString(seenObject[i].w).c_str());
            ledValueH->setText(intToString(seenObject[i].h).c_str());
            ledValueD->setText(intToString(seenObject[i].d).c_str());
        }
        //cout << " Got object " << objNum << " y=" << x << " y=" << y << " w=" << w << " h=" << h << " d=" << d << endl;
        if (objNum >= 0 && objNum < (int)objectCheckBoxes.size() && objectCheckBoxes[objNum] != NULL)
            objectCheckBoxes[objNum]->setChecked(true);
        else
            cout << "Invalid object number " << objNum << endl;
    }
    if ( !currentObjectSeen){
        ledValueX->setText("");
        ledValueY->setText("");
        ledValueW->setText("");
        ledValueH->setText("");
        ledValueD->setText("");
    }
}


bool CPlaneDisplay::close(bool alsoDelete){
    cout <<" Blah" << endl;
    return FormCPlaneDisplay::close(alsoDelete);
}

void CPlaneDisplay::slotToggleLogging(){
    if (bLogToFile) { //should save file now
        bLogToFile = false;
        fLogFile.close();
        cout << "Log file saved " << endl;
    }
    else{
        if (fLogFile.open(IO_WriteOnly)){
            cout << " file opened OK " << endl;
            QTextStream fs(&fLogFile);
            fs << "v1234" << endl;
            bLogToFile = true;
        }
        else{
            cout << "Error openning file " << endl;
            bLogToFile = false;
        }
    }
}

void CPlaneDisplay::slotChangeCurrentObject(int item)
{
    cout << "CPlaneDisplay::slotChangeCurrentObject : " << item << endl;
    currentObject = item;
    if (! bUpdateCPlane){
        displaySeenObjectOnGUI(seenObject);
    }
}


void CPlaneDisplay::slotChangeUpdatingCPlane( int state )
{	
    bUpdateCPlane = not ((bool) state);
    cout << "Stop Updating CPlane ? " << bUpdateCPlane << endl;
    lblCPlane->setPixmap(*pixmap);

}
//#include "CPlaneDisplay.moc"
//#include "SimpleServer.moc"
