from powell import *
from threading import Event
import thread #low-level threads
import sys

dummyID = 0
bDebugThread = False
def debug(*arg):
    if bDebugThread:
        print arg
def checkLogFile(logFileName):
    try:
        f = open(logFileName,"r")
        f.readline()
        line = f.readline()
        while True:
            if line[:4] != MARKING:
                return False
            line = f.readline() #base
            line = f.readline() #dirVec
            line = f.readline() #x,fx,fcalls
            line = f.readline() 
            if line == "": break
        f.close()
        return True
    except:
        return False
        
#todo : multiple runs for each params?
class PowellDescent:
    def __init__(self, startFromScratch = True, logFileName = "powell.log"):
##~         sys.setcheckinterval(1)
        if not startFromScratch and checkLogFile(logFileName):
            self.startFromScratch = False
            print "Resuming from previous run..."
        else:
            self.startFromScratch = True
            print "Starting from scratch..."
        self.logFileName = logFileName
        self.running = False
        self.eventDog = Event()
        self.eventPowell = Event()
        self.nextParameters = None
        self.evaluation = None
        self.bestParams,self.bestValue = [],-10000000
        self.isStopped = False
        self.cachingEvaluation = {}
        
        #2 variables to ensure only one param is evaluated and powell 
        #thread is resumed only when the evaluation is done
        self.paramsIsDone = False
        self.paramsInUse = False 

    def startOptimization(self):
        #need to stop the dog thread
        self.running = True
        self.eventDog.clear()
        thread.start_new_thread(fminPowellNew,(self.walkingFunction,self.parameters, self.startFromScratch))
        debug( "Sleeping..")
        self.eventDog.wait()
        debug( "..done" )
    
    #note:this is run in a separate thread - powell thread
    def walkingFunction(self,params):
        
        self.nextParameters = params
        if params is None: #powell tells to stop
            print "Powell is stoped"
            self.isStopped = True
            self.eventDog.set()
            return None
        
        if self.cachingEvaluation.get(params,None) is not None:            
            self.evaluation = self.cachingEvaluation[params]
            print "Got already evaluated policy : ",self.evaluation
            return self.evaluation
        else:
            self.evaluation = None
            self.paramsIsDone = False
            self.eventPowell.clear()
            self.eventDog.set()
            debug( "Powell Waiting...", self.nextParameters)
            self.eventPowell.wait()
            debug( "Powell executing...")
            if self.evaluation is None:
                raise "Invalid evaluation"
            return self.evaluation #minimization
##~             return -self.evaluation #maximization
        
    def writeToLog(self, msg):
        self.logFile = open(self.logFileName,"a")   
        self.logFile.write(msg + "\n")
        self.logFile.close()
        
    def setParameters(self,parameters):
        self.parameters = parameters
        
    def setDeltaParameters(self,deltaParams):
        pass

    def getCurrentParameters(self):
        return self.parameters
    
    def getNextEvaluation(self):
        if self.isStopped :
            print "Already stop"
            return (None,None)
        debug( "getNextEvaluation")
        if not self.running:
            debug( "Starting up optimization")
            self.startOptimization()
           
        if self.paramsIsDone:
            self.paramsIsDone = False
            debug( "Evaluated, resume powell")
            self.eventDog.clear()
            self.eventPowell.set()
            debug( "Dog Waiting...")
            self.eventDog.wait()
            if self.isStopped:
                print "Already stoped"
                return (None,None)
            debug( "Dog executing...")
        elif self.paramsInUse:
            print "current Params is being used, got nothing more to do"
            return (None,None) 
        self.paramsInUse = True
        return (dummyID,self.nextParameters)

    def setEvaluation(self,_,value):
        debug( "Setting ", value)
        self.evaluation = -value #since we want to maximize walkingFunction
        if value > self.bestValue:
            self.bestValue = value
            self.bestParams = self.nextParameters #actually is current 
        self.paramsInUse = False
        self.paramsIsDone = True
        
        params = self.nextParameters
        if self.cachingEvaluation.setdefault(params,value) != value:
            print "WARNING: cached value is different from this value"
            print "Param =", params, "  old :", self.cachingEvaluation[params], "  new:",value
        
##~         self.eventPowell.set()
        
                
    def cancelEvaluation(self,_):
        self.paramsInUse = False
        
    def getBestParameter(self):
        return (self.bestParams,self.bestValue)
        
    ##################### private function #####################    
    
    def loadFromFile(self,fileName):
        pass
        
