import sys
from qt import *
from WdgRobotInfo import WdgRobotInfo
import WalkLearner

def learnParamToStr(para):
    return str(int(round(para)))

VERY_SLOW_SPEED = 0 #0cm/sec
LOG_FILE_NAME = "walking_robot.log"

class RobotInfo(WdgRobotInfo):
    def __init__(self,robotID = 0, parent = None,name = None,fl = 0):
        WdgRobotInfo.__init__(self,parent,name,fl)
        self.robotID = robotID
        self.parent = parent
        self.learningState = None
        self.forthTime = -1
        self.backTime = -1
        
        self.lcdRobotID.display( self.robotID )
        self.bStarted = False
        self.learningParameters = None  # this will be set in slotGetNewCommand()
        self.policyID = -1
        self.slotGetNewCommand()
        
        
    def setStatus(self, learningState, forth, back):
        self.learningState = learningState
        self.forthTime = forth
        self.backTime = back
        self.ledLearningState.setText( str(learningState) )
        self.ledForth.setText( "%.2f"%forth )
        self.ledBack.setText( "%.2f"%back )
        if self.parent.bStartLearning:
            self.continueLearning()
            
    def saveToLogFile(self,policy, value):
        f = open(LOG_FILE_NAME,"a")
        f.write("%s %f\n" % (policy,value))
        f.close()
        
    def continueLearning(self):
        if self.learningState == "lsDONE" :
            if self.forthTime + self.backTime > 0:
                self.saveToLogFile(self.learningParameters, self.forthTime + self.backTime)
            self.slotGetNewCommand()
            #if there's no policy for this iteration, just wait
            if self.policyID != -1:
                self.slotBeReady()
            
        elif self.learningState == "lsREADY":
            if self.learningParameters is None:
                self.slotGetNewCommand()
            if self.policyID != -1:
                self.ledParams.setText( " ".join( map(learnParamToStr,self.learningParameters)  ) ) # [1,2,3] => "1 2 3"
                self.slotLearnThis()
                self.slotStart()
            
    def slotGetNewCommand(self):
        print "RobotInfo.slotGetNewCommand"
        if not self.btSkipThisRun.isEnabled():
            self.btSkipThisRun.setEnabled(True)
            print "Skipped the result" 
            return #skipped, don't ask for new policy, just use old one
            
        nextPolicy = self.parent.requestNewCommand(self.policyID, self.forthTime, self.backTime)
        if nextPolicy is not None and nextPolicy != (None,None):
            (self.policyID, self.learningParameters) = nextPolicy
        else:
            print "++++++++++++++++++++++++++++++++++++++++ Got None", 
            self.policyID = -1

    def isDone(self):
        return self.learningState == "lsDone"
     
    def slotLearnThis(self):
        print "RobotInfo.slotLearnThis(): "
        learningParams = str(self.ledParams.text())
        if len(learningParams) < 94:
            msg = WalkLearner.WrlCommand("slp",self.robotID,learningParams)
        else:
            raise "learningParams too long"        
##~         print self.parentWidget() , type(self.parentWidget())
##~         print self.parentWidget().parentWidget() , type(self.parentWidget().parentWidget())
        print "learn this ", learningParams
        self.parent.parent.slotServerSend(msg)
        
    def slotSkipThisRun(self):
        self.slotStop()
##~         self.btSkipThisRun.setEnabled(False)
##~         print self.btSkipThisRun.isEnabled()
##~         if self.policyID != -1:
##~             self.parent.cancelThisRun(self.policyID)
##~             self.slotStop()
##~             self.policyID = -1
##~         self.btSkipThisRun.setEnabled(True)
##~         print self.btSkipThisRun.isEnabled()
        
    def slotStop(self):
        msg = WalkLearner.WrlCommand("stop",self.robotID) 
        self.btStartStop.setText("&GO !")
        self.parent.parent.slotServerSend(msg)
        
    def slotStart(self):
        msg = WalkLearner.WrlCommand("start" , self.robotID) 
        self.btStartStop.setText("&Stop !")
        self.parent.parent.slotServerSend(msg)
        
    def slotStartStop(self):
        print "RobotInfo.slotStartStop():"
        if self.bStarted:
            self.slotStop()
        else:
            self.slotStart()
        self.bStarted = not self.bStarted
        
    def slotBeReady(self):
        print "RobotInfo.slotBeReady(): "
        msg = WalkLearner.WrlCommand("ready",self.robotID)
        self.parent.parent.slotServerSend(msg)
        
    def slotSetDefault(self):
        print "RobotInfo.slotSetDefault():"
        self.ledParams.setText(" ".join( map(learnParamToStr,WalkLearner.STARTING_PARAMETERS) ) )
            
    def slotSetDone(self):
        print "RobotInfo.slotSetDone():"
        msg = WalkLearner.WrlCommand("set",self.robotID,"11") #11 = lsDONE
        self.parent.parent.slotServerSend(msg)
        
    def slotRunTooSlow(self):
        self.forthTime = self.backTime = VERY_SLOW_SPEED
        self.slotGetNewCommand()
        self.slotStop()
    
    def slotDogCrash(self):
        self.slotStop()
        self.parent.cancelThisRun(self.policyID)
        
if __name__ == "__main__":
    a = QApplication(sys.argv)
    QObject.connect(a,SIGNAL("lastWindowClosed()"),a,SLOT("quit()"))
    w = RobotInfo()
    a.setMainWidget(w)
    w.show()
    a.exec_loop()
