from GradientDescent import *
from PowellDescent import *
from LearnError import LearnError
from sys import *

class ErrorFunctionBase:
    def __init__(self):
        self.callCount = 0
    def getCallCount(self):
        return self.callCount
        
class HumanErrorFunction(ErrorFunctionBase):
    def __init__(self):
        ErrorFunctionBase.__init__(self)
        
    def getFirstParams(self):
        return [80, 90, 130,  45,  15,   -65,    5, 40,  20, 45, 40]
        
    def getDeltaParams(self):
        return [ 4,  2,  2,   1 ,  1 ,    1 ,  1,   2,  2,  2,  2]
        
    def evaluate(self,params):
        self.callCount += 1
        print " Cost " + `params` + " ? " ,
        cost = float(stdin.readline())
        print "Ok, got " , cost
        return cost

class SimpleErrorFunction(ErrorFunctionBase):
    def __init__(self):
        ErrorFunctionBase.__init__(self)
        
    def getFirstParams(self):
        return [0,0,0]
        
    def getDeltaParams(self):
        return [ 1,1,1 ]
        
    def evaluate(self,params):
        self.callCount += 1
        [x,y,z] = params
        cost = - x * y * z + - 100 * x * y + 5 * x * z + 9 * y * z
        return cost
        
class RosenbrockFunction(ErrorFunctionBase):
    def __init__(self):
        ErrorFunctionBase.__init__(self)

    def getFirstParams(self):
        return [0.8,1.2,0.7]
        
    def getDeltaParams(self):
        return [ 0.001,0.001,0.001 ]
        
    def evaluate(self,params):
        self.callCount += 1
        [x,y,z] = map(float,params)
        cost = 100 * (y - x ** 2) ** 2 + 100 * (z - y ** 2 ) ** 2 + ( 1 - x ) ** 2 + ( 1 - y ) ** 2
        return -cost
        
class TestLearning:
    def __init__(self, learningAlgo, errFun):
        self.learningAlgo = learningAlgo
        self.errFun = errFun
    
    def run(self):
        params = self.errFun.getFirstParams()
        self.learningAlgo.setParameters(params)
        deltaParams = self.errFun.getDeltaParams()
        self.learningAlgo.setDeltaParameters(deltaParams)
        while True:
            (rid, params) = self.learningAlgo.getNextEvaluation()
            if rid is None: break
            
            cost = self.errFun.evaluate( params )
            if cost == -1 : break
            if cost == -2: 
                self.learningAlgo.cancelEvaluation(rid)
            else:
                self.learningAlgo.setEvaluation(rid, cost)
            print "Param, cost = ", params, cost
        

if __name__ == "__main__":
    la = PowellDescent(True)
##~     la = GradientDescent()
##~     la = LearnError()
##~     errFun = HumanErrorFunction()
##~     errFun = SimpleErrorFunction()
    errFun = RosenbrockFunction()
    test = TestLearning(la,errFun)
    try:
        test.run()
    except KeyboardInterrupt:
        pass
        
    print la.getBestParameter()
    print "Number of fucntion calls: ",errFun.getCallCount()
