import numarray as Num
import optimize


##################### Caching and loading functions ######
LOG_FILE_NAME = "powell.log"
MARKING = "----"
cachingLineSearch = []

def parseFloatList(s):
    s = s.strip("\t\n []")
    l = s.split(",")
    return map(float,l)
    
def loadPowellFromFile(logFileName):
    global cachingLineSearch
    f = open(logFileName,"r")
    print "First line : ",f.readline()
    line = f.readline()
    while line[:4] == MARKING:
        base = parseFloatList(f.readline())
        dirVec = parseFloatList(f.readline())
        x,fx,fcalls = f.readline().split()
        x = float(x)
        fx = float(fx)
        fcalls = int(fcalls)
        cachingLineSearch.append( (base,dirVec,x,fx,fcalls))
        line = f.readline()
    f.close()

def storeCachedLineSearch(base,dirVec, x, fx, fcalls):
    f = open(LOG_FILE_NAME,"a")
    f.write("%s\n" % MARKING)
    f.write("\t%s\n" % base.tolist())
    f.write("\t%s\n" % dirVec.tolist())
    f.write("\t%f %f %d\n" % (x,fx,fcalls))
    f.close()
    
FLOAT_ERROR = 0.00001
def equalFloat(a,b):
    return abs(a-b) < FLOAT_ERROR
    
def equalFloatList(listA,listB):
    if len(listA) != len(listB):
        return False
    for (a,b) in zip(listA,listB):
        if not equalFloat(a,b):
            return False
    return True
    
def loadCachedSearch(base,dirVec):
    for (_base,_dirVec,x,fx,fcalls) in cachingLineSearch:
        if equalFloatList(base,_base) and \
            equalFloatList(dirVec,_dirVec):
            return x,fx,fcalls
    return None
    
################################################

def SQR(x):
    return x*x

def line_searchA(f, xk, pk, args=(), xtol=1e-4, ftol=1e-4):
    print "minimizing from: "
    print xk
    print " Direction: "
    print pk
    
    gfk = apply(optimize.approx_fprime,(xk,f)+args)
    alpha, fc, gc = optimize.line_search_BFGS(f, xk, pk, gfk, args)
    p = xk + alpha*pk
    ev = apply(f,(p,)+args)
    print "Found alpha: %f" % alpha
    return alpha, ev, fc

def SIGN(a, b):
    if (b > 0):
        return abs(a)
    else:
        return -abs(a)

def max(a, b):
    if (a >= b):
        return a
    else:
        return b

def mnbrak(f, base, dirVec, args, ax = 0, bx = 1, maxiter = 100):
    GOLD = 1.618034

    if (ax == bx):
        bx += 1
    
    fcalls = 0
    
    fa = apply(f,((base + ax*dirVec),)+args)
    fcalls += 1
    fb = apply(f,((base + bx*dirVec),)+args)
    fcalls += 1
    
    if (fb > fa):
        dum = ax
        ax = bx
        bx = dum
        dum = fb
        fb = fa
        fa = dum
    
    cx = bx+GOLD*(bx-ax)    # Guess first point using the golden ratio
    fc = apply(f,((base + cx*dirVec),)+args)
    fcalls += 1

    iter = 0

    while (fb > fc):                            # Keep returning here until we bracket.
        iter += 1
        if (iter > maxiter):
            print "Too many iterations in mnbrak"
            return ax, fa, bx, fb, cx, fc, fcalls
        r = (bx-ax)*(fb-fc)                     # Compute u by parabolic extrapolation from a,b,c.
        q = (bx-cx)*(fb-fa)
        u = bx-((bx-cx)*q-(bx-ax)*r)/(2.0*SIGN(max(abs(q-r),TINY),q-r)) #TINY is used to prevent any possible division by zero
        ulim = bx+GLIMIT*(cx-bx) 

        if ((bx-u)*(u-cx) > 0.0):
            fu = apply(f,((base + u*dirVec),)+args)
            fcalls += 1
            if (fu < fc):
                ax = bx
                bx = u
                fa = fb
                fb = fu
                return ax, fa, bx, fb, cx, fc, fcalls
            elif (fu > fb):
                cx = u
                fc = fu
                return ax, fa, bx, fb, cx, fc, fcalls
            u=cx+GOLD*(cx-bx)
            fu = apply(f,((base + u*dirVec),)+args)
            fcalls += 1
        elif ((cx-u)*(u-ulim) > 0.0):
            fu = apply(f,((base + u*dirVec),)+args)
            fcalls += 1
            if (fu < fc):
                bx = cx
                cx = u
                u = cx+GOLD*(cx-bx)
                fb = fc
                fc = fu
                fu = apply(f,((base + u*dirVec),)+args)
                fcalls += 1
        elif ((u-ulim)*(ulim-cx) >= 0.0):
            u = ulim
            fu = apply(f,((base + u*dirVec),)+args)
            fcalls += 1
        else:
            u=cx+GOLD*(cx-bx);
            fu = apply(f,((base + u*dirVec),)+args)
            fcalls += 1
        ax = bx
        bx = cx
        cx = u
        fa = fb
        fb = fc
        fc = u
    return ax, fa, bx, fb, cx, fc, fcalls


TINY = 1.0e-20
GLIMIT = 100.0
    
def line_search(f, base, dirVec, args=(), xtol=1e-4, ftol=1e-4, maxiter=None, ax = 0, bx = 1):
    res = loadCachedSearch(base,dirVec)
    if res is not None:
        return res
    
    CGOLD = 0.3819660
    ZEPS = 1.0e-10
    
    if maxiter is None:
        maxiter = 100
    iter = 0
    
    fa = 0
    fb = 0
    fc = 0
    fbx = 0
    xmin = 0
    fx = 0
    e = 0

    ax, fa, bx, fb, cx, fc, fcalls = mnbrak(f, base, dirVec, args, ax, bx, maxiter)
    
    if (ax < cx):
        a = ax
        b = cx
    else:
        a = cx
        b = ax
    
    x = w = v = bx
    fbx = fw = fv = fx = fb
    while (iter < maxiter):
        iter += 1
        xm = (a+b)/2.0
        tol1 = ftol*abs(x)+ZEPS
        tol2 = 2.0*tol1
        if (abs(x-xm) <= (tol2-(b-a)/2)):
            # print "alpha: %f f(alpha*dirVec+base): %f nCalls: %d" % (x, fx, fcalls)
            storeCachedLineSearch( base,dirVec,x,fx,fcalls )
            return x, fx, fcalls    # Exit with best values.
        if (abs(e) > tol1): # Construct a trial parabolic fit.
            r = (x-w)*(fx-fv)
            q = (x-v)*(fx-fw)
            p = (x-v)*q-(x-w)*r
            q = 2.0*(q-r)
            if (q > 0):
                p = -p
            q = abs(q)
            etemp = e
            e = d
            if (abs(p) >= abs(q*etemp/2)) or (p <= q*(a-x)) or (p >= q*(b-x)):
                if (x >= xm):
                    e = a-x
                else:
                    e = b-x
                d = CGOLD*e
            else:
                d = p/q
                u = x+d
                if ((u-a < tol2) or (b-u < tol2)):
                    d = SIGN(tol1, xm-x)
        else:
            if (x >= xm):
                e = a-x
            else:
                e = b-x
            d = CGOLD*e
        if (abs(d) >= tol1):
            u = x+d
        else:
            u = x+SIGN(tol1,d)
        fu = apply(f,((base + u*dirVec),)+args)
        fcalls += 1
        if (fu <= fx):
            if (u >= x):
                a = x
            else:
                b = x
            v = w
            w = x
            x = u
            fv = fw
            fw = fx
            fx = fu
        else:
            if (u < x):
                a = u
            else:
                b = u
            if ((fu <= fw) or (x == x)):
                v = w
                w = u
                fv = fw
                fw = fu
            elif ((fu <= fv) or (v == x) or (v == w)):
                v = u
                fv = fu
    print "Too many iterations"
    storeCachedLineSearch( base,dirVec,x,fx,fcalls )
    return x, fx, fcalls


        
def fminPowellNew(func, x0, startFromScratch = True,args=(), xtol=1e-4, ftol=1e-4, xi=None, maxiter=None, maxfun=None, fulloutput=0, printmessg=1):
    if startFromScratch:
        print "Starting Powell from scratch..."
        f = open(LOG_FILE_NAME,"w")
        f.write("Powell log file\n")
        f.close()
    else:
        print "Loading Powell from file..."
        loadPowellFromFile(LOG_FILE_NAME)
    fminPowell(func, x0, args, xtol, ftol, xi, maxiter, maxfun, fulloutput, printmessg)
    
def fminPowell(func, x0, args=(), xtol=10e-4, ftol=.2, xi=None, maxiter=None, maxfun=None, fulloutput=0, printmessg=1):
    """xopt,{fval,warnflag} = fmin(function, x0, args=(), xtol=1e-4, ftol=1e-4,
    maxiter=200*len(x0), maxfun=200*len(x0), fulloutput=0, printmessg=0)

    Uses a Powell's algorithm to find the minimum of function
    of one or more variables.  See Section 10.5 of Numerical Recipes in C.
    """
    
    p = Num.asarray(x0)
    assert (len(p.shape)==1)
    n = len(p)
    if maxiter is None:
        maxiter = 200
    iter = 0
    if maxfun is None:
        maxfun = n * 500
    func_calls = 0
    if xi is None:
        xi = Num.identity(n, p.typecode())

    fret = apply(func,(p,)+args)
    func_calls += 1
    pt = p.copy()
    ptt = p.copy()
    xit = xi[0].copy()
    
    while True:
        ibig = -1
        delta = -1e20
        iter += 1
        if (iter > maxiter):
            if printmessg:
                print "Warning: Maximum number of iterations has been exceeded"
            warnflag = True
            break
        if (func_calls > maxfun):
            if printmessg:
                print "Warning: Maximum number of function evaluations has been exceeded"
            warnflag = True
            break
        # perform a line search in each direction
    
        fptt = fp = fret
        for i in range(0,n):
            fptt = fret
            alpha, fret, fc = line_search(func, p, xi[i], args, xtol, 1e-3)
            #print "add %f times " % alpha
            #print xi[i]
            p += alpha * xi[i]
            func_calls += fc
            if ((fptt - fret) > delta):
                delta = fptt - fret
                ibig = i
        if (delta < 0):
            if printmessg:
                print "Warning: All directions head up hill"
            warnflag = True
            break
        if (2.0*abs(fp - fret) <= ftol*(abs(fp)+abs(fret))):
            if printmessg:
                print "Optimisation successful"
            warnflag = False
            break
        
        # find the total movement - and extrapolate in that direction
        
        #ptt = 2*p-pt
        xit = p-pt
        pt = p
        
##~         print "prev xi: (ibig %d)" % ibig
##~         print xi
        
        xi[ibig]=xi[0]
        xi[0] = xit
    

    if printmessg:
        print "         Current function value: %f" % fret
        print "         Iterations: %d" % iter
        print "         Function evaluations: %d" % func_calls
    try:
        apply(func, (None,)) #tell the function thread to stop
    except TypeError:
        print "Function not support None argument"
        
    if fulloutput:
        return p, fret, func_calls, warnflag
    else:        
        return p
    

if __name__ == "__main__":
    import string
    import time

    
    times = []
    algor = []
    x0 = [0.8,1.2,0.7]
    
    start = time.time()
    x = optimize.fmin(optimize.rosen,x0)
    print x
    times.append(time.time() - start)
    algor.append('Nelder-Mead Simplex\t')
    
    start = time.time()
    x = fminPowellNew(optimize.rosen,x0,False)
    print x
    times.append(time.time() - start)
    algor.append('Powells Method\t')

    start = time.time()
    x = optimize.fminBFGS(optimize.rosen, x0, avegtol=1e-4, maxiter=100)
    print x
    times.append(time.time() - start)
    algor.append('BFGS without gradient\t')

    print "\nMinimizing the Rosenbrock function of order 3\n"
    print " Algorithm \t\t\t       Seconds"
    print "===========\t\t\t      ========="
    for k in range(len(algor)):
        print algor[k], "\t -- ", times[k]
