#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: ActiveLocalise.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# ActiveLocalise.py
#
# Implements smart active localising based on self covariance
#
#===============================================================================

import math
import Global
import VisionLink
import HelpShort
import HelpLong
import HelpTrack
import Constant

#===============================================================================
# Global variables of this module.
#===============================================================================

##~ # debugging..
##~ head1 = 0
##~ head2 = 0
##~ head3 = 0
##~ head4 = 0



CHECK_BEACONS = [2, 3, 6, 7] # These are the GPS coordinate array indices of
                             # the beacons to check for active localising.
                             # Note: they're not vob indices.

activeLocaliseCounter = 0    # Use to switch between beaconBestReduceVar and beaconAlternate.

localiseBeacon  = 0     # Which beacon active localise should look at.
                        # Either set this yourself or use SmartSetBeacon.

beaconAlternate = 0     # The beacon (except the beaconBestReduceVar) observable in 
                        # least pan, as an alternate choice to beaconBestReduceVar.

isFocusing = False

myaGoingTo = Constant.dCLOCKWISE

#===============================================================================
# Functions that belong to this module
#===============================================================================



#--------------------------------------
# Makes the dog look at the beacon designated by localiseBeacon
def DecideNextAction(bShowLED = True):

    global activeLocaliseCounter, localiseBeacon, isFocusing

##~     print "Active-localising!!"

    # Increase counter each time we want to active localise.
##~     activeLocaliseCounter += 1
    
    # Set debugging information.
    if bShowLED:
        HelpLong.showRedEye(True)

    # Get shared variables from elsewhere.
    beaconPos   = VisionLink.getGPSCoordArray()
    selfLoc     = Global.selfLoc

    if localiseBeacon == 0:
        print "ERROR (ActiveLocalise.DecideNextAction): localiseBeacon = 0"
        return

    whichBeacon = localiseBeacon
##~     print "ActiveLocalise at beacon: ", whichBeacon

    relPos      = map(lambda b,s:b-s, beaconPos[whichBeacon], selfLoc.getPos())
    relPos      = HelpShort.rotate(relPos, 90 - selfLoc.getHeading())
    targetX     = -relPos[0]
    targetY     = relPos[1]

    HelpLong.setHeadParams(targetX, Constant.BEACON_HEIGHT, targetY, Constant.HTAbs_xyz)
    
    # Check if the head is in where it should be already. FWHead will check it.
    # -ive to make it has the same sign as the dog's local pan.
##~     targetPan = -math.atan2(targetX, targetY)
##~ 
##~     if (abs(targetPan) > 90 and abs(Global.degPan > 80)):
##~         isFocusing = True
##~    
##~     elif (abs(targetPan-Global.degPan) < 20):
##~         isFocusing = True
##~     else:
##~         isFocusing = False

#--------------------------------------
# Sets which beacon the dog will look at based on self covariance. panLimit
# and distLimit specify which beacons should be considered. If you want to
# consider all beacons the dog could see from its current field position, use
# the defaults

def SmartSetBeacon(panLimit = 90, distLimit = 550):

    global localiseBeacon

    # Get shared variables from elsewhere.
    selfLoc = Global.selfLoc
    cov     = VisionLink.getGPSSelfCovariance()
    
    # Calculate the direction of the larger axis of the self covariance
    b = -(cov[0][0] + cov[1][1])
    c = cov[0][0] * cov[1][1] - cov[1][0] * cov[1][0]
    det = b * b - 4 * c
    
    if (det < 0.00001):
        sol = -b / 2
    else:
        sol = (-b + math.sqrt(det)) / 2
    
    if abs(cov[0][0] - sol) < 0.01 and cov[0][1] < 0.01:
    
        # If self position variance is circular, all beacons are equally good
        # to localise off, and so choose the one closest to current heading.
        if abs(cov[1][1] - sol) < 0.01:
            head = selfLoc.getHeading()
        else:
            head = HelpShort.RAD2DEG(math.atan2(-cov[1][0], cov[1][1] - sol))
            
    else:
        head = HelpShort.RAD2DEG(math.atan2(sol - cov[0][0], cov[0][1]))
            
    # Find the beacon that is closest to the variance direction.
    beaconPos = VisionLink.getGPSCoordArray()
    curCompHead = 360
    localiseBeacon = 0    
    
    for beacon in CHECK_BEACONS:
        relPos = map(lambda b,s:b-s, beaconPos[beacon], selfLoc.getPos())
        beaconHead = HelpShort.RAD2DEG(math.atan2(relPos[1], relPos[0]))
        beaconDist = HelpShort.getLength(relPos)
        localHead  = HelpShort.normalizeAngle_180(beaconHead - selfLoc.getHeading())
    
        if abs(localHead) < panLimit and beaconDist < distLimit:
            compHead = abs(HelpShort.normalizeAngle_180(head - beaconHead))
            if compHead > 90:
                compHead = 180 - compHead
            
            if compHead < curCompHead:
                curCompHead  = compHead
                localiseBeacon = beacon

#--------------------------------------
# Sets the beacon the dog should look at to the one with smallest pan from
# current heading

def AlternateSetBeacon(panLimit = 90, distLimit = 550):
    global localiseBeacon

    selfLoc = Global.selfLoc
    beaconPos = VisionLink.getGPSCoordArray()
    curCompHead = panLimit
    localiseBeacon = 0    
    
    for beacon in CHECK_BEACONS:
        relPos = map(lambda b,s:b-s, beaconPos[beacon], selfLoc.getPos())
        beaconHead = HelpShort.RAD2DEG(math.atan2(relPos[1], relPos[0]))
        beaconDist = HelpShort.getLength(relPos)
        localHead  = HelpShort.normalizeAngle_180(beaconHead - selfLoc.getHeading())
    
        if abs(localHead) < curCompHead and beaconDist < distLimit:
            curCompHead  = localHead
            localiseBeacon = beacon

#--------------------------------------

def beaconLocalise():
    global myaGoingTo,alcounter
    
    # If current GPS is bad, then my will be wrong.
    if Global.selfLoc.getHeadingVar() > HelpShort.get95cf(80) or\
        Global.selfLoc.getPosVar() > HelpShort.get95cf(60):
        HelpTrack.stationaryLocalise()
    else:
        beaconPos = [(0,0),
            (Constant.FIELD_WIDTH,0),
            (Constant.FIELD_WIDTH,Constant.FIELD_LENGTH),
            (0,Constant.FIELD_LENGTH)]
        panAngle = []
        myX,myY = Global.selfLoc.getPos()
        myH = HelpShort.normalizeAngle_180( Global.selfLoc.getHeading() )
        for (x,y) in beaconPos:
            deg1 = HelpShort.getHeadingBetween( myX,myY,x,y) 
            panAngle.append( HelpShort.normalizeAngle_180(deg1 - myH ) )
        panAngle.sort(lambda x,y : cmp(abs(x) , abs(y) ) )
        panA,panB = panAngle[:2]
        if panA > panB:
            t = panA
            panA = panB
            panB = t
    
        if myaGoingTo == Constant.dCLOCKWISE:
            panx = panA
            if abs(Global.degPan - panx) <= 5:
                myaGoingTo = Constant.dANTICLOCKWISE
        else:
            panx = panB
            if abs(Global.degPan - panx) <= 5:
                myaGoingTo = Constant.dCLOCKWISE
        HelpLong.setHeadParams(panx,0,0,Constant.HTAbs_h)
