#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: BallChaser.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# CoolDemo.py
#
# Demonstrate UPennKick
#
#===============================================================================
 

"""
OpenChallenger Ball Chaser
"""

import Global
import HelpLong
import HelpShort
import HelpTrack
import VisionLink
import OpenChallenger
import Constant
from StateBasedAction import *
import math
##~ import ActiveLocalise

FP_COOL = [0,1,2,1,0]
FP_KICKING = [1,0,0,0,1]

myPos = (0.0,0.0)
myHeading = 0
myBallDistance = 0
localBallX = 0
ballPos = (0.0,0.0)
goalPos = (Constant.FIELD_WIDTH/2,Constant.FIELD_LENGTH)

def resetEachFrame():
    global myBallDistance,myPos,ballPos,\
        myHeading,\
        localBallX
    myBallDistance = Global.vBall.getVobDistance()
    
    myPos = Global.selfLoc.getPos()
    myHeading = Global.selfLoc.getHeading()
    ballPos = (Global.ballX,Global.ballY)
    
    localBallX = Global.vBall.getVobDistance() * \
            math.sin(HelpShort.DEG2RAD(Global.vBall.getHeading()))

def RandomUPennKick():
    if Global.frame % 2 == 0:
        return Constant.UpennLeftWT
    else:
        return Constant.UpennRightWT

def RandomKick():
    if Global.frame % 2 == 0:
        return Constant.HandKickLeftWT
    else:
        return Constant.HandKickRightWT

class MyActiveLocalise(StateBasedAction):
    ACTIVE_LOCALISE_INTERVAL = 60
    DOING_ACTIVE_LOCALISE = 10
    CLOSE_DISTANCE = 60
    CLOSE_DISTANCE_SLOW_BALL = 20
    
    def __init__(self):
        StateBasedAction.__init__(self)
        self.activeLocaliseCounter = self.ACTIVE_LOCALISE_INTERVAL
        
    
        
    def finishAction(self): #do reseting counter
        self.activeLocaliseCounter = self.ACTIVE_LOCALISE_INTERVAL
        
    def DoAction(self):
        if (Global.lostBall < 5):
            HelpTrack.trackVisualBall()
        elif Global.lostBall <= 10:
            HelpLong.setHeadParams(0,-20,-20,Constant.HTAbs_h)
        elif (Global.lostBall <= 30
                and abs(Global.gpsLocalBall.getHeading()) < 90): 
            HelpTrack.trackGpsBall()
        else:
            HelpTrack.stationaryLocalise()
                
##~         #decreasing counter
##~         self.activeLocaliseCounter -= 1
##~         if self.activeLocaliseCounter < - self.DOING_ACTIVE_LOCALISE:
##~             return None #stop action, finishAction called()
##~         elif self.activeLocaliseCounter < 0:
##~            
##~             #ball too close, should not active localise
##~             if Global.lostBall == 0:
##~                 if Global.ballD < MyActiveLocalise.CLOSE_DISTANCE:
##~                     return self 
##~ 
##~             ActiveLocalise.SmartSetBeacon(90,500)
##~             ActiveLocalise.DecideNextAction(False)
##~             
        return self
        
class TurnAndKick(StateBasedAction):
    
    TURNING_LEFT_720 = 11.8
    TURNING_RIGHT_720 = 11.0
    UPENN_ANGLE = 10
    goodRanges1 = [
        [(0,5),(42,60),(79,80)], #UpennRightWT
        [(0,18),(44,65)]  #UpennLeftWT
    ]
    
    goodRanges2 = [
        [(6,41),(61,78)], #UpennRightWT
        [(19,43),(66,80)]  #UpennLeftWT
    ]
    
##~     goodRanges = [
##~         [(0,5),(42,60),(79,80)], #UpennRightWT
##~         [(0,18),(44,65),(76,80)]  #UpennLeftWT
##~     ]
    def __init__(self):
        StateBasedAction.__init__(self)
        self.turningTime1 = 0
        self.turningTime2 = 0
        self.angle1 = 0
        self.angle2 = 0
        self.theKick = Constant.UpennLeftWT
        self.turnDir = Constant.dCLOCKWISE
        self.turnNeeded = 0
        self.isKicked = False
        
    def calculateDesiredHeading(self):
        attackHeading = HelpShort.getHeadingBetween(*(myPos + goalPos))
        self.turnNeeded = HelpShort.normalizeAngle_180(attackHeading - myHeading)
        print "Turn needed = ",self.turnNeeded," my H = ", myHeading, attackHeading
        
    def getTurningTime(self,turnAngle):
        if abs(turnAngle) < 1:
            return 0
        elif turnAngle > 0:
            return int(abs(self.TURNING_LEFT_720 * 30 / (720.0/turnAngle)))
        else:
            return int(abs(self.TURNING_RIGHT_720 * 30 / (720.0/turnAngle)))
    
    def calculateTurningTime(self):
##~         if self.turnNeeded > 0 :
##~             if self.turnDir == Constant.dCLOCKWISE :
##~                 if self.theKick == Constant.UpennLeftWT:
##~                     self.angle1 = - (360 - self.turnNeeded + TurnAndKick.UPENN_ANGLE)
##~                     self.angle2 = 2 * TurnAndKick.UPENN_ANGLE
##~                 else:
##~                     self.angle1 = 0
##~                     self.angle2 = - ( 360 - self.turnNeeded + TurnAndKick.UPENN_ANGLE)
##~             else:
##~                 if self.theKick == Constant.UpennLeftWT:
##~                     self.angle1 = 0
##~                     self.angle2 = self.turnNeeded + TurnAndKick.UPENN_ANGLE
##~                 else:
##~                     self.angle1 = self.turnNeeded + TurnAndKick.UPENN_ANGLE
##~                     self.angle2 = -2 * TurnAndKick.UPENN_ANGLE
##~         else:
##~             if self.turnDir == Constant.dCLOCKWISE :
##~                 if self.theKick == Constant.UpennLeftWT:
##~                     self.angle1 = self.turnNeeded - TurnAndKick.UPENN_ANGLE
##~                     self.angle2 = 2 * TurnAndKick.UPENN_ANGLE
##~                 else:
##~                     self.angle1 = 0
##~                     self.angle2 = self.turnNeeded - TurnAndKick.UPENN_ANGLE
##~             else:
##~                 if self.theKick == Constant.UpennLeftWT:
##~                     self.angle1 = 0 
##~                     self.angle2 = 360 + self.turnNeeded + TurnAndKick.UPENN_ANGLE
##~                 else:
##~                     self.angle1 = 360 + self.turnNeeded + TurnAndKick.UPENN_ANGLE
##~                     self.angle2 = -2 * TurnAndKick.UPENN_ANGLE
##~                 
##~         
        self.theKick = RandomUPennKick()
        if self.turnNeeded > 0:
            self.turnDir = Constant.dANTICLOCKWISE            
        else:
            self.turnDir = Constant.dCLOCKWISE
        if self.theKick == Constant.UpennRightWT:
            self.angle2 = self.turnNeeded - TurnAndKick.UPENN_ANGLE
        else:
            self.angle2 = self.turnNeeded + TurnAndKick.UPENN_ANGLE
        
        if self.turnNeeded > 0:
            self.theKick = Constant.UpennRightWT
        else:
            self.theKick = Constant.UpennLeftWT
        self.angle1 = 0
            
        self.turningTime1 = self.getTurningTime(self.angle1)
        self.turningTime2 = self.getTurningTime(self.angle2)
        print "Angle1 = %d , turningTime = %d",(self.angle1,self.turningTime1)
        print "Angle2 = %d , turningTime = %d",(self.angle2,self.turningTime2)
        
        
    def beginAction(self):
##~         if Global.frame % 4 == 0:
##~             self.turnDir = Constant.dCLOCKWISE
##~             self.theKick = Constant.UpennLeftWT
##~         elif Global.frame % 4 == 1:
##~             self.turnDir = Constant.dCLOCKWISE
##~             self.theKick = Constant.UpennRightWT
##~         elif Global.frame % 4 == 2:
##~             self.turnDir = Constant.dANTICLOCKWISE
##~             self.theKick = Constant.UpennLeftWT
##~         else:
##~             self.turnDir = Constant.dCLOCKWISE
##~             self.theKick = Constant.UpennRightWT
        self.isKicked = False
        self.calculateDesiredHeading()
        self.calculateTurningTime()
        HelpLong.openMouth(-600000)
            
    def goodRanges(self,theKick, turningAngle):
    
        if theKick == Constant.UpennLeftWT :
            oddDir = turningAngle > 0
        else:
            oddDir = turningAngle < 0
            
        
        if oddDir:
            return self.goodRanges1[theKick - Constant.UpennRightWT]
        else:
            return self.goodRanges2[theKick - Constant.UpennRightWT]
        
    def isGoodStep(self,step,kick):
        for (a,b) in self.goodRanges(kick,self.angle2):
            if step >= a and step <= b:
                return True
        return False
        
    def doTheKick(self):
        if self.isGoodStep(Global.pWalkInfo.getCurrentStep(),self.theKick):            
            Global.finalAction[Constant.AAWalktype] = self.theKick
            HelpLong.setForceStepComplete()
            
            self.isKicked = True
            print "Kick %d , Current step %d" % (self.theKick, Global.pWalkInfo.getCurrentStep() )
        
    def finishAction(self):
        HelpLong.openMouth(-10000)
        
    def doTurning(self,angle):
        if angle > 0:
            Global.finalAction = list(Constant.TurnHoldingCounterClockWise)
        else:
            Global.finalAction = list(Constant.TurnHoldingClockWise)
        
    def DoAction(self):
        HelpLong.showFacePattern(FP_KICKING)
        if self.counter > 10:
            if VisionLink.getAnySensor(0) - (-600000) < 10000:
                return None
##~             print "Mouth",VisionLink.getAnySensor(0) + 600000
        if self.counter < self.turningTime1:
            self.doTurning(self.angle1)
        elif self.counter < self.turningTime1 + self.turningTime2:
            self.doTurning(self.angle2)
        elif self.isKicked:
            return None
        else:
            self.doTurning(self.angle2)
            self.doTheKick()
            
        return self
            
class KickTurning(StateBasedAction):
    TURNING_LEFT_720 = 7.2
    TURNING_RIGHT_720 = 7.2
    def __init__(self):
        StateBasedAction.__init__(self)
        self.desiredHeading = 0
        self.turningTime = 0
        self.walkParams = None
        
    def calculateDesiredHeading(self):
        self.desiredHeading = 90
##~         print "My Pos = ",myPos
##~         print "desiredHeading=",self.desiredHeading

        
    def calculateTurningTime(self):
        turnNeeded = HelpShort.normalizeAngle_180(self.desiredHeading - myHeading)
        print "Turn needed = ",turnNeeded," my H = ", myHeading

        if turnNeeded > 0:
            calibrate = self.TURNING_LEFT_720
            self.walkParams = list(Constant.TurnHoldingCounterClockWise)
        else:
            calibrate = self.TURNING_RIGHT_720
            self.walkParams = list(Constant.TurnHoldingClockWise)
            
        if turnNeeded == 0:
            self.turningTime = 0
        else:
            self.turningTime = calibrate * 30 / (720.0/abs(turnNeeded))
        print "Turning time = ",self.turningTime
        
    def beginAction(self):
        self.calculateDesiredHeading()
        self.calculateTurningTime()
        HelpLong.openMouth(-350000)
        
    def finishAction(self):
        HelpLong.openMouth(Constant.IND_MOUTH_CLOSED)
        
        
    def DoAction(self):
        Global.finalAction = self.walkParams[:]
##~         print "Turning "
        if self.counter >= self.turningTime:
            return self.outState
        else:
            return self

class UPennKicker(StateBasedAction):
    APPROACHING_DISTANCE = 25
    def __init__(self):
        StateBasedAction.__init__(self)
        self.myHoverToBall = OpenChallenger.MyHoverToBall()
        self.doApproaching = OpenChallenger.ApproachingBall(10,self.APPROACHING_DISTANCE)
        self.doGrabBall = OpenChallenger.MyGrabBall()
        self.turnAndKick = TurnAndKick()
        
        self.doLocateBall = OpenChallenger.MyLocateBall()
        self.doActiveLocalise = MyActiveLocalise()
        
        self.doLocateBall.setOutState(self.doApproaching)
        self.doApproaching.setOutState(self.doGrabBall)
        self.doGrabBall.setOutState(self.turnAndKick)
        
        self.debug = False
        
        
    def DoAction(self):
        HelpLong.showFacePattern(FP_COOL)
        self.doActiveLocalise.DecideNextAction()
        if self.subState is None:     
            
            if Global.lostBall > 0:
                self.doSubState(self.doLocateBall)
            else:
                self.doSubState(self.myHoverToBall)
        else:
            if self.subState == self.myHoverToBall and \
                Global.ballD < self.APPROACHING_DISTANCE:
                self.doSubState(self.doApproaching)
            elif self.subState == self.doApproaching and \
                Global.lostBall == 0 and \
                self.doGrabBall.shouldGrabTheBall():
                if self.debug: print "GrabBall"
                self.doSubState(self.doGrabBall)
            else:
                self.doSubState()
        return self
     
            
class CoolKick(KickTurning):
    def __init__(self):
        KickTurning.__init__(self)
        
    def calculateDesiredHeading(self):
        self.desiredHeading = HelpShort.getHeadingBetween(*(myPos + goalPos))
        self.theKick = RandomKick()
    
    def DoAction(self):
        print "Turning ", self.counter
        if self.counter >= self.turningTime:
            Global.finalAction[Constant.AAWalktype] = self.theKick
            HelpLong.setForceStepComplete()
            print "Kicked"
            return self.outState
        else:
            KickTurning.DoAction(self)
            return self
        
class HandKicker(StateBasedAction):
    APPROACHING_DISTANCE = 25
    def __init__(self):
        StateBasedAction.__init__(self)
        self.myHoverToBall = OpenChallenger.MyHoverToBall()
        self.doApproaching = OpenChallenger.ApproachingBall(10,self.APPROACHING_DISTANCE)
        self.doGrabBall = OpenChallenger.MyGrabBall()
        self.turnAndKick = CoolKick()
        
        self.doLocateBall = OpenChallenger.MyLocateBall()
        self.doActiveLocalise = MyActiveLocalise()
        
        self.doLocateBall.setOutState(self.doApproaching)
        self.doApproaching.setOutState(self.doGrabBall)
        self.doGrabBall.setOutState(self.turnAndKick)
        
        self.debug = False
        
        
    def DoAction(self):
        HelpLong.showFacePattern(FP_COOL)
        self.doActiveLocalise.DecideNextAction()
        if self.subState is None:     
            
            if Global.lostBall > 0:
                self.doSubState(self.doLocateBall)
            else:
                self.doSubState(self.myHoverToBall)
        else:
            if self.subState == self.myHoverToBall and \
                Global.ballD < self.APPROACHING_DISTANCE:
                self.doSubState(self.doApproaching)
            elif self.subState == self.doApproaching and \
                Global.lostBall == 0 and \
                self.doGrabBall.shouldGrabTheBall():
                if self.debug: print "GrabBall"
                self.doSubState(self.doGrabBall)
            else:
                self.doSubState()
        return self
        
THE_LIMIT = 16
class DiveKicker(StateBasedAction):
    APPROACHING_DISTANCE = 25
    def __init__(self):
        StateBasedAction.__init__(self)
        self.myHoverToBall = OpenChallenger.MyHoverToBall()
        self.doApproaching = OpenChallenger.ApproachingBall(10,self.APPROACHING_DISTANCE)
        
        
        self.doLocateBall = OpenChallenger.MyLocateBall()
        self.doActiveLocalise = MyActiveLocalise()
    
        
        self.debug = False
        
        
    def DoAction(self):
        HelpLong.showFacePattern(FP_COOL)
        self.doActiveLocalise.DecideNextAction()
        if self.subState is None:     
            
            if Global.lostBall > 0:
                self.doSubState(self.doLocateBall)
            else:
                self.doSubState(self.myHoverToBall)
        else:
            if self.subState == self.myHoverToBall and \
                Global.ballD < self.APPROACHING_DISTANCE:
                self.doSubState(self.doApproaching)
            else:
                self.doSubState()
        return self
        
upennKicker = UPennKicker()
diveKicker = DiveKicker()

def initPlayer():
    pass
    
def DecideNextAction():

    Global.frameReset()
    HelpLong.superDebug()
    OpenChallenger.resetEachFrame()
    resetEachFrame()
    diveKicker.DecideNextAction()
    
initPlayer()
