#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: Behaviou.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# Behaviou.py
#
# This is the top level module in our python behaviour program.
# processFrame() got called first and it will call the "DecideNextAction()" of
# relevant player.
#
#===============================================================================

# In python, each code file is called a module. Python programs work by having 
# one and only one top-level module, which executes the "main function" and 
# import other modules to carry out more complex tasks.

import Player
import Global
import Reload
import Constant
import VisionLink
import HelpLong
import Debug #everything you need for debugging
import profile
import pstats

evalResult = "None has been evaluated"

#These are 4 levels of pausing.
#level 1: pause the leg only. --> The dog stands still and thinks.
#level 2: pause both leg and head --> The dog is thinking from a particular cplane.
#level 3: stop DecideNextAction()  --> The dog is really pausing, everything in its mind in the previous state is remained
#level 4: (mode 0) Stop everything, cannot communicate at all, except for sending cplane and world model

#the commands for them are (pairs of pausing/resuming)
# level 1 : (pyc) n pleg            /   (pyc) n rleg
# level 2 : (pyc) n paus[e]       /   (pyc) n resu[me]
# level 3 : (pyc) n nodna         /   (pyc) n nodna
# level 4 : (mode) 0                  /   (mode) 1
# n is the dog's number.

bStopMotion = False
bStopLegOnly = False
bStopDecideNextAction = False


Debug.bDebugOn = True # just to stop pychecker warning

profilerInstance = None

#===============================================================================
# Functions belong to this module.
#===============================================================================

def dumpProfileStats():
    global profilerInstance
    
    if profilerInstance is not None:
    
        mystats = pstats.Stats(profilerInstance)
        profilerInstance = None
        mystats.dump_stats("/MS/profstat")
        print "profiler stats dumped"
    else:
        print "profiler was not switched on"

def startProfiler():
    global profilerInstance
    
    print "starting profiler"
    
    profilerInstance = profile.Profile()

#--------------------------------------
# Called by processFrame() in robot7/vision/pyEmbed.cc,
# is used for dynamic evaluation
def processFrame():

    global profilerInstance,bStopMotion

    # increment the frame counter
    # Make sure frame is initialised in Behaviou.py initBehavModule()
    Global.frame += 1
    
    if not bStopDecideNextAction:
        if profilerInstance is None:
            Player.player.DecideNextAction()
        else:
            profilerInstance.runcall(Player.player.DecideNextAction)

    debugOnDemand()
    dynamic_gain()

    Global.lastForward = Global.finalAction[Constant.AAForward]
    Global.lastLeft    = Global.finalAction[Constant.AALeft]
    Global.lastTurnCCW = Global.finalAction[Constant.AATurnCCW]
        
    if bStopLegOnly:
        HelpLong.setNormalWalk(0,0,0)
    if not bStopMotion:
        VisionLink.sendAtomicAction(*tuple(Global.finalAction) )
        if VisionLink.getPressSensorCount(2) > 60:
            bStopMotion = True
    else:
        if VisionLink.getPressSensorCount(0) > 5:
            bStopMotion = False
    if Global.finalIndicators != Global.lastIndicators:
        VisionLink.sendIndicators(*tuple(Global.finalIndicators) )
        Global.lastIndicators = Global.finalIndicators[:] 

#--------------------------------------
# Called when received a wireless command.
def processCommand(cmd):
    print "Python:I just got this command: ",cmd
    global evalResult, bDebugForwardLeftTurn
    global bStopMotion,bStopLegOnly,bStopDecideNextAction
    cmd = cmd.strip()
    if cmd[:4] == "eval":
        evalResult = eval(cmd[4:])
        VisionLink.sendMessageToBase("****"+str(evalResult))
    elif cmd[:4] == "exec":
        print "Executing (%s)" % cmd[4:].strip()
        exec(cmd[4:].strip())
##~         print bDebugForwardLeftTurn
    elif cmd[:4] == "pleg": #put the dog stand still, head still moves
        bStopLegOnly = True
    elif cmd[:4] == "rleg": #resume legs
        bStopLegOnly = False    
    elif cmd[:4] == "paus": #allow to pause the dog(both head and legs)
        bStopMotion = True
    elif cmd[:4] == "resu": #allow to resume the dog
        bStopMotion = False
    elif cmd[:4] == "colo": #allow to set team color
        color = int(cmd.split()[-1])
        VisionLink.setGPSGoal(color)
    elif cmd[:5] == "nodna":
        bStopDecideNextAction = not bStopDecideNextAction
    else:
        try:
            Player.player.processCommand(cmd)
        except:
            pass #that's fine, some player doesn't handle message


#--------------------------------------
#Print debug information only when some flag is turned on ( preferable via Dynamic Execution)
def debugOnDemand():
    if Global.bDebugForwardLeftTurn:
        print "F = %.2f, L = %.2f, T = %.2F" % \
            (Global.finalAction[Constant.AAForward],Global.finalAction[Constant.AALeft],Global.finalAction[Constant.AATurnCCW])

def setLowGain():

    if Global.isHighGain:
        Global.isHighGain = False
        Global.finalAction[Constant.AACmd] = Constant.aaSetLowGain

def setHighGain():
    
    if not Global.isHighGain:
        Global.isHighGain = True
        Global.finalAction[Constant.AACmd] = Constant.aaSetHighGain
    
# -------------------------------
# set the motor join gain by battery current
def dynamic_gain ():

    current = VisionLink.getBatteryCurrent()
    
    wt = Global.finalAction[Constant.AAWalktype]
    if wt in [Constant.ChestPushWT,Constant.FwdKickWT,
        Constant.FastKickWT       ,Constant.UpennRightWT,
        Constant.UpennLeftWT      ,Constant.DiveKickWT, 
        Constant.HandKickRightWT  ,Constant.HandKickLeftWT]:
        Global.forceHighGain = True
##~         print "Kicking, force high gain"
    
    if Global.forceHighGain is None: #do dynamic_gain
        if current < 2300 and not Global.isHighGain:
            setHighGain()
        elif current > 3000 and Global.isHighGain:
            setLowGain()
    else:
        if Global.forceHighGain :
            setHighGain()
        else:
            setLowGain()
            
#--------------------------------------
# Initialisations before any frame is processed.
def initBehaviourModule():
    print "Starting initialization of Behaviour.py"

    VisionLink.setCallbacks(processFrame, processCommand)
    
    print "Finish initialization of Behaviour.py"


#===============================================================================
# These following statments are sort of like main function in C.
#===============================================================================

# Only the top-level module (ie. this file) should have this (non-function-scope
# statements). If other modules have statements not in function scope, like this, 
# they will be executed whenever they are imported, which is not what we want.


# Reload every dependent modules as described in Reload.py. 
reload(Reload)

initBehaviourModule()


