#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: Constant.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# Constant.py
#
# Constants used in behaviours codes. Most player modules will need to import this.
#
#===============================================================================

import VisionLink
import math

# Directions used in FWHead.py
dCLOCKWISE     = 0
dANTICLOCKWISE = 1
dAUTO = 1

NUM_OF_EFFECTORS = VisionLink.NUM_OF_EFFECTORS # = 15
NUM_SENSORS      = VisionLink.NUM_SENSORS # = 34

#===============================================================================
# Constants used by finalAction - it will be sent to actuatorControl.

# Indices in AtomicAction , depends on PyEmbed.cc::sendAtomicAction()
# Indices for finalAction.
AACmd       = 0
AAWalktype  = 1
AAForward   = 2
AALeft      = 3
AATurnCCW   = 4
AAPg        = 5
AAHf        = 6
AAHb        = 7
AAHdf       = 8
AAHdb       = 9
AAFfo       = 10
AAFso       = 11
AABfo       = 12
AABso       = 13
AAHeadType  = 14
AAPanx      = 15
AATilty     = 16
AACranez    = 17

# Indices for  finalIndicators.
iMouth      = 0
iTailV      = 1
iTailH      = 2
iEarL       = 3
iEarR       = 4
iLed1       = 5
iLed2       = 6
iLed3       = 7
iLed4       = 8
iLed5       = 9
iLed6       = 10
iLed7       = 11
iLed8       = 12
iLed9       = 13
iLed10      = 14
iLed11      = 15
iLed12      = 16
iLed13      = 17
iLed14      = 18
iHeadColor  = 19
iHeadWhite  = 20
iModeRed    = 21
iModeGreen  = 22
iModeBlue   = 23
iWireless   = 24
iBackRearColor      = 25
iBackRearWhite      = 26
iBackMiddleColor    = 27
iBackMiddleWhite    = 28
iBackFrontColor     = 29
iBackFrontWhite     = 30

# Role constants.
NOROLE           = 0
FACEBACKOFF      = 1
ATTACKER         = 2
STRIKER          = 3
ATTACKBACKOFF    = 4
GETBEHINDBACKOFF = 5
SUPPORT          = 6
SIDEBACKOFF      = 7
SUPPORTBACKOFF   = 8
DEFENDER         = 9
DEFENDERGOALBOX  = 10
STEALTHDOG       = 11
STUCKER          = 12
GOALIEATTACK     = 13
GOALIEDEFENDSEEB = 14
GOALIEDEFENDNOTS = 15

# Indicator intensity constant.
INTENSITY_DIVIDE = 10

# Can only store up to 8 (0-7) color, please use LEDs (showFacePattern) if you
# have more things to show in head (and use OTHER for myRole).

# RGB Color Information for the light on head, used in MakeHeadColor in HelpLong.py
RGB_NO_COLOR    = (False, False, False)
RGB_BLUE        = (False, False, True)
RGB_GREEN       = (False, True, False)
RGB_LIGHT_BLUE  = (False, True, True)
RGB_RED         = (True, False, False)
RGB_PURPLE      = (True, False, True)
RGB_YELLOW      = (True, True, False)
RGB_WHITE       = (True, True, True)

# Face and back lights constants. 
IND_LED3_INTENSITY_MIN = 0x00000000
IND_LED3_INTENSITY_MAX = 0x000000FF
IND_LED3_RED_EYE_VAL   = 0x000000FE
IND_LED3_MODE_A        = 0x00000000
IND_LED3_MODE_B        = 0x00000F00
IND_LED2_ON            = 1
IND_LED2_OFF           = 2
IND_LED_UNCHANGED      = -1
IND_LED_ON             = 1
IND_LED_OFF            = 0
IND_MOUTH_OPEN         = -950000
IND_MOUTH_CLOSED       = -10000

# Mouth and Tail constants
IND_MOUTH_CLOSED = -10000
IND_MOUTH_OPEN = -950000
IND_TAIL_RIGHT = 1000000
IND_TAIL_LEFT = -1000000
IND_TAIL_UP = 1000000
IND_TAIL_DOWN = 17000
IND_TAIL_H_CENTRED = (IND_TAIL_LEFT + IND_TAIL_RIGHT) / 2
IND_TAIL_V_CENTRED = (IND_TAIL_UP + IND_TAIL_DOWN) / 2

# Face patterns constants.
# For easy lookup, please put in ascending order.
FP_BACKOFF_RUN_BEHIND   = [1,0,1,0,1]
FP_BACKOFF_SUPPORT      = [1,0,1,0,2]
FP_BACKOFF_ATTACKER     = [1,0,1,0,3]
FP_BACKOFF_IMMEDIATE    = [1,0,1,1,3]

FP_FIRE_PAW_KICK        = [1,1,1,1,1]
FP_HOVER_TO_BALL        = [1,2,1,2,1]
FP_BIRD_OF_PREY         = [2,1,2,1,2]
FP_DIR_KICK             = [3,0,2,0,3]

FP_GET_BEHIND_LE        = [3,0,0,0,0]
FP_GET_BEHIND_RE        = [0,0,0,0,3]
FP_GET_BEHIND_BE        = [0,0,1,0,0]
FP_GET_BEHIND_TLE       = [0,2,0,0,0]
FP_GET_BEHIND_TRE       = [0,0,0,2,0]
FP_GET_BEHIND_TLC       = [1,0,0,0,0]
FP_GET_BEHIND_TRC       = [0,0,0,0,1]

FP_DKD_LEFT             = [0,0,2,2,2]
FP_DKD_RIGHT            = [2,2,2,0,0]
FP_DKD_BOTTOM_LEFT      = [2,0,0,2,2]
FP_DKD_BOTTOM_RIGHT     = [2,2,0,0,2]

# Headtype, read actuatorControl/HeadMotion.cc for details.
HTNull_h    = 0
HTRel_h     = 1
HTAbs_h     = 2
HTAbs_pt    = 3
HTAbs_xyz   = 4
HTRel_z     = 5

# Values for AACmd, as defined in robot7/share/ActioProDef.h
# For the first element of finalAction to use.
aaStop                  = 0
aaStill                 = 1
aaTailUp                = 2
aaParaWalk              = 3
aaRelaxed               = 4
aaTurnOnJointDebug      = 5
aaSetJointSpeedLimit    = 6
aaSetAJointGain         = 7
aaSetHighGain           = 8
aaSetLowGain            = 9
aaReloadSpecialAction   = 10
aaForceStepComplete   = 11

# JointEnum enumeration in PWalkDef.h
jNothing                = 0
jHeadTilt               = 1
jHeadPan                = 2
jHeadCrane              = 3
jRightFrontJoint        = 4
jRightFrontShoulder     = 5
jRightFrontKnee         = 6
jLeftFrontJoint         = 7
jLeftFrontShoulder      = 8
jLeftFrontKnee          = 9
jRightRearJoint         = 10
jRightRearShoulder      = 11
jRightRearKnee          = 12
jLeftRearJoint          = 13
jLeftRearShoulder       = 14
jLeftRearKnee           = 15

# Stand constants (mm)
SHOULDER_HEIGHT         = 81
HIP_HEIGHT              = 110
DOG_LENGTH              = 130
DOG_WIDTH               = 100

# Walktype enum , from share/PWalkDef.h
NormalWalkWT            = 0
CanterWalkWT            = 1
ZoidalWalkWT            = 2
OffsetWalkWT            = 3 
EllipseWalkWT           = 4 
AlanWalkWT              = 5
ChestPushWT             = 6
FwdKickWT               = 7
FastKickWT              = 8
GetupFrontWT            = 9
GetupSideWT             = 10
GetupRearWT             = 11
BackKickWT              = 12
HoldBallWT              = 13
CheerWT                 = 14
DanceWT                 = 15
UpennRightWT            = 16
UpennLeftWT             = 17
BackWalkWT              = 18
SkiWalkWT               = 19
SkellipticalWalkWT      = 20
DiveKickWT              = 21
HandKickRightWT         = 22
HandKickLeftWT          = 23
HeadForwardKick         = 24
HeadLeftKick            = 25
HeadRightKick           = 26

HANDKICKANGLEOFFSET     = 0

# A few default values for finalAction to be used.
# It is really handled in PyEmbed.cc::sendAtomicAction()
NormalWalkStandStill    = (aaParaWalk,NormalWalkWT,0,0,0,5,90,110,0,0,55,15,-55,5,HTAbs_h,0,0,0)
EllipticalWalkFast      = (aaParaWalk,EllipseWalkWT,0,0,0,80,90,130,10,25,45,15,-65,5,HTAbs_h,0,0,0)
SkellipticalWalkFast    = (aaParaWalk,SkellipticalWalkWT,0,0,0,80,90,130,10,25,45,15,-65,5,HTAbs_h,0,0,0)
NormalWalkSideWayFast   = (aaParaWalk,NormalWalkWT,0,0,0,32,95,110,15,25,60,5,-55,5,HTAbs_h,0,0,0)
NormalWalkFast          = (aaParaWalk,NormalWalkWT,0,0,0,40,90,110,20,20,59,10,-50,5,HTAbs_h,0,0,0)
ChestPushKick           = (aaParaWalk,ChestPushWT,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
UPleftKick              = (aaParaWalk,UpennLeftWT,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
UPrightKick             = (aaParaWalk,UpennRightWT,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
HandKickRight           = (aaParaWalk,HandKickRightWT,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
HandKickLeft            = (aaParaWalk,HandKickLeftWT,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
DiveKick                = (aaParaWalk,DiveKickWT,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
ForwardKick             = (aaParaWalk,FwdKickWT,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
FastKick                = (aaParaWalk,FastKickWT,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
NormalWalkSlow          = (aaParaWalk,NormalWalkWT,0,0,0,40,70,100,15,20,60,10,-55,5,HTAbs_h,0,0,0)
##~ NormalWalkSlow          = NormalWalkFast
BackWalkClockWise           = (aaParaWalk,BackWalkWT,0,3,-10,30,70,135,10,45,95,-8,-55,20,HTAbs_h,0,-40,40)
BackWalkCounterClockWise    = (aaParaWalk,BackWalkWT,0,-3,10,30,70,135,10,45, 95, -8, -55,20,HTAbs_h,0,-40,40)
BackWalkSitDownHoldingBall  = (aaParaWalk,BackWalkWT,0, 0,  0,5,20, 40,10,45,145,-15, -70,20,HTAbs_h,0, 0 ,0)
TurnKickClockWise           = (aaParaWalk,NormalWalkWT,3, -2,  -20,40,70, 100,10,20,95,15, -55,5,HTAbs_h,0, 0 ,0)
TurnKickCounterClockWise    = (aaParaWalk,NormalWalkWT,3, 2,  20,40,70, 100,10,20,95,15, -55,5,HTAbs_h,0, 0 ,0)

# 13/6/04 The following is not calibrated for gps, takes 36 seconds for 10 revolutions
TurnHoldingClockWise           = (aaParaWalk,NormalWalkWT,0, -3,  -40,40,80, 120,10,25,75,0, -55,10,HTAbs_h,30, -55 ,40)
TurnHoldingCounterClockWise    = (aaParaWalk,NormalWalkWT,0, 3,  40,40,80, 120,10,25,75,0, -55,10,HTAbs_h,-30, -55 ,40)
TURN_LIMIT_EllipticalWalkFast  = 15

# Default value for finalIndicators to use.
INDICATORS_DEFAULT = (IND_MOUTH_CLOSED, # iMouth \
                IND_TAIL_UP,       # iTailV
                IND_TAIL_H_CENTRED,# iTailH      = 2
                0,                 # iEarL       = 3
                0,                 # iEarR       = 4
                IND_LED_OFF,       # iLed1       = 5
                IND_LED_OFF,       # iLed2       = 6
                IND_LED_OFF,       # iLed3       = 7
                IND_LED_OFF,       # iLed4       = 8
                IND_LED_OFF,       # iLed5       = 9
                IND_LED_OFF,       # iLed6       = 10
                IND_LED_OFF,       # iLed7       = 11
                IND_LED_OFF,       # iLed8       = 12
                IND_LED_OFF,       # iLed9       = 13
                IND_LED_OFF,       # iLed10      = 14
                IND_LED_OFF,       # iLed11      = 15
                IND_LED_OFF,       # iLed12      = 16
                IND_LED_OFF,       # iLed13      = 17
                IND_LED_OFF,       # iLed14      = 18
                IND_LED2_OFF,      # iHeadColor  = 19
                IND_LED2_OFF,      # iHeadWhite  = 20
                IND_LED2_OFF,      # iModeRed    = 21
                IND_LED2_OFF,      # iModeGreen  = 22
                IND_LED2_OFF,      # iModeBlue   = 23
                IND_LED2_OFF,      # iWireless   = 24
                IND_LED3_INTENSITY_MIN,  # iBackRearColor      = 25
                IND_LED3_INTENSITY_MIN,  # iBackRearWhite      = 26
                IND_LED3_INTENSITY_MIN,  # iBackMiddleColor    = 27
                IND_LED3_INTENSITY_MIN,  # iBackMiddleWhite    = 28
                IND_LED3_INTENSITY_MIN,  # iBackFrontColor     = 29
                IND_LED3_INTENSITY_MIN,  # iBackFrontWhite     = 30
                )

#===============================================================================
# Other constants

LARGE_VAL           = 10000000
BLUE_TEAM           = 0 #compatible with VisionLink.getTeamColor()
RED_TEAM            = 1

# Physical measurements  , from share/BallDef.h
BallRadius            = VisionLink.BALL_RADIUS
BallDiameter          = 2 * BallRadius

# Physical dimension of the field
FIELD_WIDTH            = 290
FIELD_LENGTH           = 440
FIELD_DIAGONAL         = math.sqrt(FIELD_WIDTH ** 2 + FIELD_LENGTH ** 2)
BEACON_HEIGHT          = 30
WALL_THICKNESS         = 10
GOAL_WIDTH             = 60
GOAL_LENGTH            = 30
CORNER_WIDTH           = 50
GOALBOX_WIDTH          = 100
GOALBOX_DEPTH          = 50
CIRCLE_DIAMETER        = 30
TARGET_GOAL_X          = FIELD_WIDTH / 2.0
TARGET_GOAL_Y          = FIELD_LENGTH
OWN_GOAL_X             = FIELD_WIDTH / 2.0
OWN_GOAL_Y             = 0
RIGHT_GOALBOX_EDGE_X   = TARGET_GOAL_X + GOALBOX_WIDTH / 2.0
LEFT_GOALBOX_EDGE_X    = TARGET_GOAL_X - GOALBOX_WIDTH / 2.0
LEFT_GOALBOX_EDGE_Y    = TARGET_GOAL_Y - WALL_THICKNESS - GOALBOX_DEPTH
RIGHT_GOALBOX_EDGE_Y   = LEFT_GOALBOX_EDGE_Y
LEFT_GOAL_POST         = TARGET_GOAL_X - GOAL_WIDTH/2.0
RIGHT_GOAL_POST        = TARGET_GOAL_X + GOAL_WIDTH/2.0
OWN_GOALBOX_EDGE_Y     = OWN_GOAL_Y + WALL_THICKNESS + GOALBOX_DEPTH 
IMAGE_WIDTH            = 208
IMAGE_HEIGHT           = 160

# Context  , from Common.h - for gps ball info
CTGlobal                = 0
CTLocal                 = 1

#Visual Object indices , from vision/VisualCortex.h
vobBall = 0
vobBlueGoal = 1
vobYellowGoal = 2
vobBlueLeftBeacon = 3
vobBlueRightBeacon = 4
vobGreenLeftBeacon = 5
vobGreenRightBeacon = 6
vobYellowLeftBeacon = 7
vobYellowRightBeacon = 8    
vobRedDog = 9
vobRedDog2 = 10
vobRedDog3 = 11
vobRedDog4 = 12
vobBlueDog = 13
vobBlueDog2 = 14
vobBlueDog3 = 15
vobBlueDog4 = 16

# Ball usage constants
VISION_BALL    = 0
GPS_BALL       = 1
WIRELESS_BALL  = 2
VBALL_LOST     = 12
VBALL_ANYWHERE = 40
LOSTBALL_DIST  = 1000

USE_WIRELESS_BALL_VAR = 100

# Wireless constants
NUM_TEAM_MEMBER = VisionLink.NUM_TEAM_MEMBER

# Basic behaviour constants, also used as an attack mode constants
FIRE_PAWKICK_AUTO      = 0
GET_BEHIND_BALL        = 1
FIELD_EDGE_SPIN        = 2
SPIN_DRIBBLE           = 3
WIRELESS_TEAMMATE_INFO = 4
HOVER_TO_BALL          = 5        # This is like the default case 
FIRE_PAWKICK_LEFT      = 6
FIRE_PAWKICK_RIGHT     = 7

# Indicator indices, taken from share/IndicatorsDef.h
IND_MOUTH_START     = 0;
IND_TAIL_START      = 1;
IND_EAR_START       = 3;
IND_FACE_LED3_START = 5;
IND_LED2_START      = 19;
IND_BACK_LED3_START = 25;

# Supporter constants.
ON_EDGE_OFFSET = WALL_THICKNESS + 20

# DKD Region constants.
DKDBlank       = 0
DKDRight       = 1
DKDLeft        = 2
DKDBottomRight = 3
DKDBottomLeft  = 4

# GameController related constants.
READYSTATE               = 7  # Copy from Behaviour.h. Correspond to mReadyMatch.
SETSTATE                 = 5  # Copy from Behaviour.h. Correspond to mStartMatch.
PLAYINGSTATE             = 0  # Copy from Behaviour.h. Correspond to mNormal.
FINISHED                 = 8  # Copy from Behaviour.h. Correspond to mPythonDemobilized.
ROBOCUP_KICKOFF_INVALID  = 0x00
ROBOCUP_KICKOFF_OWN      = 0x01
ROBOCUP_KICKOFF_OPPONENT = 0x02

# Striker constants.
FURTHEST_OFFSET = 30

# Stuck detection constants.
STUCK_DIST_TO_BALL_THRESHOLD               = 10
STUCK_NUM_OF_FRAMES_WITHIN_THRESHOLD_LIMIT = 60
STUCK_SOLUTION_ONE_STANDSTILL              = 0
STUCK_SOLUTION_TWO_READYSTATE              = 1
STUCK_THRESHOLD                            = 1000

#Stuck detection related, which leg gets stuck>
legFRONT_RIGHT = 1
legFRONT_LEFT  = 2
legBACK_RIGHT  = 3
legBACK_LEFT   = 4

# Ready state signal constants.
NOTDECIDED          = 0
DECIDEPOSITION_A    = 1
DECIDEPOSITION_B    = 2
DECIDEPOSITION_C    = 3
DECIDEPOSITION_D    = 4
EDGE_OFFSET         = 5
POS_OFFSET_IN_KICK  = 15
POS_OFFSET_OUT_KICK = 30
POSITION_OFFSET     = 20
HEADING_OFFSET      = 20

# Stealth dog constans.
STEALTHDOG_OFFSET_ANGLE = 80

# Goalie constants.
WIDTH_MOVE_THRESHOLD  = 203
HEIGHT_MOVE_THRESHOLD = 155





CIRCLE_BOP    = 1
NORMAL_BOP    = 0
AVOIDBOX_BOP  = 2
POSTAVOID_BOP = 3

#Sensor values
ssMOUTH                 = 0
ssCHIN                  = 1
ssHEAD_TILT2            = 2
ssHEAD                  = 3
ssINFRARED_NEAR         = 4 
ssINFRARED_FAR          = 5
ssHEAD_PAN              = 6
ssHEAD_TILT1            = 7
ssFL_PALM               = 8
ssFL_KNEE               = 9
ssFL_ABDUCTOR           = 10
ssFL_ROTATOR            = 11
ssRL_PALM               = 12
ssRL_KNEE               = 13
ssRL_ABDUCTOR           = 14
ssRL_ROTATOR            = 15
ssFR_PALM               = 16
ssFR_KNEE               = 17
ssFR_ABDUCTOR           = 18
ssFR_ROTATOR            = 19
ssRR_PALM               = 20
ssRR_KNEE               = 21
ssRR_ABDUCTOR           = 22
ssRR_ROTATOR            = 23
ssTAIL_H                = 24
ssTAIL_V                = 25
ssACCEL_FOR             = 26
ssACCEL_SIDE            = 27
ssACCEL_Z               = 28
ssCHEST_INFRARED        = 29
ssWLAN                  = 30
ssBACK_REAR             = 31
ssBACK_MIDDLE           = 32
ssBACK_FRONT            = 33
