#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: Dribble.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




"""
Dirbbler when close to goal.
"""

import Global
import HelpShort
import HelpLong
import Constant
import VisionLink
import Kick
import HoverToBall
import FWHead

SpinTimer       = 0
GrabTimer       = 0
Diff            = 0
TurnDirection   = Constant.dCLOCKWISE
KickDirection   = Constant.dCLOCKWISE

AVOID_KNOCK_BALL_TIME   = 5
OPEN_MOUTH_TIME         = 15
TOTAL_GRAB_TIME         = 23
HOLD_BALL_TIME          = 30
LOCALISE_TIME           = 50
SPIN_PREPARE            = 65
SPIN_LIMIT              = 95
CHARGING                = 125
STUCK_THRESHOLD         = 15000
GRABBING_MOUTH_OPEN     = -600000
HOLD_BALL               = (Constant.aaParaWalk,Constant.NormalWalkWT,0,0,0,5,90,120,0,0,110,-15,-55,15,Constant.HTAbs_h,0,0,0)
SLOW_WALK               = (Constant.aaParaWalk,Constant.NormalWalkWT,0,0,0,40,70,100,15,20,50,10,-55,5,Constant.HTAbs_h,0,0,0)

#=============================
# Convenient helpers
def resetSpinDribble():
    
    global GrabTimer, SpinTimer, Diff

    Diff      = 0
    GrabTimer = 0
    SpinTimer = 0
    
    HelpLong.openMouth(Constant.IND_MOUTH_CLOSED)

def ShouldIGrabBall():

# u can see the ball and looking really down, for ball grabbing
##~     print "p %.2f   t %.2f   c %.2f" % (Global.degPan,Global.degTilt,Global.degCrane)
##~     print "ballD %.2f" % (Global.ballD)
    
##~     ballx = math.sin(HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
##~     
##~     print "ballx %.2f   ballH %.2f"  % (ballx,Global.ballH)
##~     print "ballD %.2f" % (Global.ballD)
##~     
##~     if abs(ballx) < 4 and Global.ballD < 14 and Global.lostBall == 0:
##~         return True
    if (Global.ballD < 14 and Global.vBall.getConfidence() > 0 and
        Global.degTilt < -10 and abs(Global.degPan) < 18 and Global.degCrane < -10):
        return True
        
    return False

def setGrabBallPose():

    Global.finalAction[Constant.AACmd:Constant.AAHeadType] = list(HOLD_BALL[Constant.AACmd:Constant.AAHeadType])

def ballGrabbedUnderChin():

    return (VisionLink.getAnySensor(Constant.ssCHIN) > 0 or
            VisionLink.getAnySensor(Constant.ssMOUTH) - GRABBING_MOUTH_OPEN > STUCK_THRESHOLD)
    
    
#---------------------------
# Note this slow walk has a Different fs0 form the one in Constant

def setNormalSlowWalk(forward,left,turn):

    Global.finalAction[Constant.AACmd:Constant.AAHeadType] = list(Constant.NormalWalkSlow[Constant.AACmd:Constant.AAHeadType])
    Global.finalAction[Constant.AAHf] = 70
    Global.finalAction[Constant.AAHb] = 100
    Global.finalAction[Constant.AAFfo] = 95
    Global.finalAction[Constant.AAFso] = 5
    HelpLong.setForwardLeftTurn(forward,left,turn)

#==============================
# grabBall and SpinDribble is for the spin dribble action only
def grabBall():

    global GrabTimer, SpinTimer, Diff
    
    GrabTimer += 1
    
##~     print "Grab ball timer " , GrabTimer , " mouth " , (VisionLink.getAnySensor(Constant.ssMOUTH)) , \
##~         " mouth " , (VisionLink.getAnySensor(Constant.ssMOUTH) - GRABBING_MOUTH_OPEN > STUCK_THRESHOLD)

    if GrabTimer < AVOID_KNOCK_BALL_TIME:
##~         HelpLong.setHeadParams(0,-5,15,Constant.HTAbs_h)
        setNormalSlowWalk(4.5,0,0)
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,-5,15,Constant.HTAbs_h)

    elif GrabTimer < OPEN_MOUTH_TIME:
##~         HelpLong.setHeadParams(0,-70,50,Constant.HTAbs_h)
        setNormalSlowWalk (4.0,0,0)
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,-70,50,Constant.HTAbs_h)
        
    elif GrabTimer < TOTAL_GRAB_TIME:
##~         HelpLong.setHeadParams(0,-70,50,Constant.HTAbs_h)
        HelpLong.openMouth (GRABBING_MOUTH_OPEN)
        setNormalSlowWalk(3.0,0,0)
       
        if Global.lostBall == 0:
            resetSpinDribble()
##~             print "I see ball at counter = ", GrabTimer
            return False
            
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,-70,50,Constant.HTAbs_h)        

    elif GrabTimer == TOTAL_GRAB_TIME:
    
        if Global.lostBall == 0:
##~             print "I see ball at counter = ", GrabTimer
            resetSpinDribble()
            return False
        
        if ballGrabbedUnderChin():
            setGrabBallPose()
            HelpLong.openMouth(GRABBING_MOUTH_OPEN)
##~             HelpLong.setHeadParams(0,-70,50,Constant.HTAbs_h)
            FWHead.compulsoryAction = FWHead.fixHead
            FWHead.fixHeadAt        = (0,-70,50,Constant.HTAbs_h)        

##~             print "ball is grabbed"
            
        else:
##~             print "cant sense a ball"
            resetSpinDribble()
            return False

    elif GrabTimer < HOLD_BALL_TIME:

##~         print "grabbing ball"
        setGrabBallPose()
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,-70,50,Constant.HTAbs_h)
##~         HelpLong.setHeadParams(0,-70,50,Constant.HTAbs_h)
        HelpLong.openMouth (GRABBING_MOUTH_OPEN)
        
    elif GrabTimer == HOLD_BALL_TIME:
    
        if Global.lostBall == 0:
##~             print "I see ball at counter = ", GrabTimer
            resetSpinDribble()
            return False
        
        if ballGrabbedUnderChin():
##~             print "ball is grabbed"
            setGrabBallPose()
            HelpLong.openMouth(Constant.IND_MOUTH_CLOSED)
            FWHead.compulsoryAction = FWHead.mustLocalise
##~             ActiveLocalise.SmartSetBeacon(90,500)
##~             ActiveLocalise.DecideNextAction()
        else:
##~             print "cant sense a ball"
            resetSpinDribble()
            return False
            
    elif GrabTimer < LOCALISE_TIME:
##~         print "localising"
        setGrabBallPose()
        FWHead.compulsoryAction = FWHead.mustLocalise
##~         ActiveLocalise.SmartSetBeacon(90,500)
##~         ActiveLocalise.DecideNextAction()
        
    elif GrabTimer < SPIN_PREPARE:
        Global.finalAction[Constant.AACmd:Constant.AAHeadType] = Constant.TurnHoldingCounterClockWise [Constant.AACmd:Constant.AAHeadType]
        HelpLong.setForwardLeftTurn(0,0,0)
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,-70,50,Constant.HTAbs_h)
##~         HelpLong.setHeadParams (0,-70,50,Constant.HTAbs_h)
    
    elif GrabTimer < SPIN_LIMIT:
        print "setting spin timer"
        myx , myy   = Global.selfLoc.getPos()
        myh         = Global.selfLoc.getHeading()
        minGoalHead = HelpShort.getHeadingBetween (myx,myy,Constant.RIGHT_GOAL_POST-10,Constant.FIELD_LENGTH)
        maxGoalHead = HelpShort.getHeadingBetween (myx,myy,Constant.LEFT_GOAL_POST+10 ,Constant.FIELD_LENGTH)
        goalHead    = (minGoalHead + maxGoalHead) / 2.0
        Diff        = HelpShort.normalizeAngle_180 (goalHead - myh)
        SpinTimer   = int (abs(Diff) * 11.0 / 24.0) + SPIN_PREPARE
        
        print "minGoalHead %.2f   maxGoalHead %.2f" % (minGoalHead,maxGoalHead)
        print "goalHead    %.2f   myh         %.2f" % (goalHead, myh)
        
        if abs(Diff) > 8:
            print "Spinning to goal"
            if Diff > 0:
                Global.finalAction = list(Constant.TurnHoldingCounterClockWise)
            else:
                Global.finalAction = list(Constant.TurnHoldingClockWise)
            FWHead.compulsoryAction = FWHead.fixHead
            FWHead.fixHeadAt        = (0,-70,50,Constant.HTAbs_h)
##~         HelpLong.setHeadParams(0,-70,50,Constant.HTAbs_h)    
##~             HelpLong.setNormalWalk(0,0,HelpShort.CLIP (HelpShort.normalizeAngle_180 (dkd - myh) , 30.0 ))
##~             HelpLong.setHeadParams(0,-70,50,Constant.HTAbs_h)
        else:
            GrabTimer = SPIN_LIMIT

    elif GrabTimer < CHARGING:
    
        print "LOOKING UP and PUSHING"
        
        if (Global.vTGoal.getConfidence() > 0):
            HoverToBall.DecideNextAction(Global.vTGoal.getDistance(),Global.vTGoal.getHeading())
        else:
            HoverToBall.DecideNextAction(40,0)
            
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,0,5,Constant.HTAbs_h)
        
##~         Global.finalAction = list(Constant.TurnHoldingCounterClockWise)
##~         HelpLong.setHeadParams(0,0,5,Constant.HTAbs_h)
##~         FWHead.compulsoryAction = FWHead.fixHead
##~         FWHead.fixHeadAt        = (0,0,5,Constant.HTAbs_h)
##~         HelpLong.setForwardLeftTurn(0,0,0)
        
    elif GrabTimer == CHARGING:
        resetSpinDribble()
        return True
    
##~     elif GrabTimer == 300:
##~         print "break out, timer expired"
##~         resetSpinDribble()
##~         return False
        
    return None
    

def SpinDribble():

    if (Global.vTGoal.getConfidence() > 0):
        HoverToBall.DecideNextAction(Global.vTGoal.getDistance(),Global.vTGoal.getHeading())
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,0,5,Constant.HTAbs_h)

    


#==============================
# do a turn + upenn kicks for the below
def grabBall_forUpenn():

    global GrabTimer, SpinTimer, TurnDirection, KickDirection
    
    GrabTimer += 1
    
##~     print "Grab ball timer " , GrabTimer , " mouth " , (VisionLink.getAnySensor(Constant.ssMOUTH)) , \
##~         " mouth " , (VisionLink.getAnySensor(Constant.ssMOUTH) - GRABBING_MOUTH_OPEN > STUCK_THRESHOLD)
    
    # Assume that the upenn is in 45 degrees
    if GrabTimer == 1:
        myh   = Global.selfLoc.getHeading()
        angle = HelpShort.normalizeAngle_180 (myh - Global.DKD[0])
        
        #------------------------ less opponent area
##~         closeOpp,closeOppDist = HelpLong.closestGpsOpponentWithinVar(80)
##~         if closeOpp is not None and closeOppDist is not None:
##~             print "d %.2f  , h %.2f" % (closeOppDist, closeOpp.getHeading())
##~         
##~         if (closeOppDist is not None and
##~             closeOpp is not None and
##~             closeOppDist < 80 and 
##~             closeOpp.getHeading() <= 180 and 
##~             closeOpp.getHeading() >= 0):
##~             KickDirection = Constant.dCLOCKWISE
##~             TurnDirection = Constant.dCLOCKWISE
##~             angle         = HelpShort.normalizeAngle_180 (Global.DKD[0] + 45 - myh)
##~             
##~         elif (closeOpp is not None and  
##~               closeOppDist is not None and 
##~               closeOppDist < 80 and 
##~               closeOpp.getHeading() > -180 and 
##~               closeOpp.getHeading() < 0):
##~             KickDirection = Constant.dANTICLOCKWISE
##~             TurnDirection = Constant.dANTICLOCKWISE
##~             angle         = HelpShort.normalizeAngle_180 (Global.DKD[0] - 45 - myh)
            
        if angle >= 0 and angle <= 180:
            KickDirection = Constant.dCLOCKWISE
            
            if angle <= 45 and angle >= 0:
                TurnDirection = Constant.dANTICLOCKWISE
            else:
                TurnDirection = Constant.dCLOCKWISE
            angle         = HelpShort.normalizeAngle_180 (Global.DKD[0] + 45 - myh)
            
        else:
            KickDirection = Constant.dANTICLOCKWISE
            
            if angle <= 0 and angle >= -45:
                TurnDirection = Constant.dCLOCKWISE
            else:
                TurnDirection = Constant.dANTICLOCKWISE
            angle         = HelpShort.normalizeAngle_180 (Global.DKD[0] - 45 - myh)
        
##~         if angle >= 0 and angle <= 180:
##~             angle     = HelpShort.normalizeAngle_180 (Global.DKD[0] + 45 - myh)
##~             TurnDirection = Constant.dCLOCKWISE
##~         else:
##~             angle     = HelpShort.normalizeAngle_180 (Global.DKD[0] - 45 - myh)
##~             TurnDirection = Constant.dANTICLOCKWISE
            
        SpinTimer     = int (abs(angle) * 0.3) + HOLD_BALL_TIME + 1
##~         print ""
##~         print "DKD is %.2f, myh %.2f" % (Global.DKD[0],myh) 
##~         print "KickDirection %d , TurnDirection %d , SpinTimer %d" % (KickDirection,TurnDirection,SpinTimer)
            
##~         print "setting up turn + upenn"
##~         print "myh %.2f , Diff %.2f, SpinTimer %.2f TurnDirection %d" % (myh,Diff,SpinTimer,TurnDirection)
        
    if GrabTimer < AVOID_KNOCK_BALL_TIME:
##~         print "avoid knock ball out"
        setNormalSlowWalk(0,0,0)
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,-5,15,Constant.HTAbs_h)

    elif GrabTimer < OPEN_MOUTH_TIME:
        setNormalSlowWalk (0,0,0)
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,-70,50,Constant.HTAbs_h)
        
    elif GrabTimer < TOTAL_GRAB_TIME:
        HelpLong.openMouth (GRABBING_MOUTH_OPEN)
        setNormalSlowWalk(0,0,0)
       
        if Global.lostBall == 0:
##~             print "See the ball , return false, at frame " , GrabTimer
            resetSpinDribble()
            return False
            
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,-70,50,Constant.HTAbs_h)        

    elif GrabTimer == TOTAL_GRAB_TIME:
        
        # If ball is seen, abort
        if Global.lostBall == 0:
##~             print "See the ball , return false, at frame " , GrabTimer
            resetSpinDribble()
            return False
            
        # If a ball is sensed, good
        if ballGrabbedUnderChin():
            Global.finalAction[Constant.AACmd:Constant.AAHeadType] = Constant.TurnHoldingClockWise [Constant.AACmd:Constant.AAHeadType]
            HelpLong.setForwardLeftTurn(0,0,0)
            HelpLong.openMouth(GRABBING_MOUTH_OPEN)
            FWHead.compulsoryAction = FWHead.fixHead
            FWHead.fixHeadAt        = (0,-70,50,Constant.HTAbs_h)        
            
        # If a ball is not sensed, abort
        else:
##~             print "cant sense the ball, return false, at frame " , GrabTimer
            resetSpinDribble()
            return False
            
    elif GrabTimer < HOLD_BALL_TIME:
        
        Global.finalAction[Constant.AACmd:Constant.AAHeadType] = Constant.TurnHoldingClockWise [Constant.AACmd:Constant.AAHeadType]
        HelpLong.setForwardLeftTurn(0,0,0)
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,-70,50,Constant.HTAbs_h)
        HelpLong.openMouth(GRABBING_MOUTH_OPEN)
        
    elif GrabTimer == HOLD_BALL_TIME:
        
        if Global.lostBall == 0:
##~             print "See the ball , return false, at frame " , GrabTimer
            resetSpinDribble()
            return False
         
        if ballGrabbedUnderChin():
            
            if TurnDirection == Constant.dCLOCKWISE:
                Global.finalAction = list(Constant.TurnHoldingClockWise)
            else:   
                Global.finalAction = list(Constant.TurnHoldingCounterClockWise)
            
            FWHead.compulsoryAction = FWHead.fixHead
            p  = Global.finalAction[Constant.AAPanx]
            t  = Global.finalAction[Constant.AATilty]
            c  = Global.finalAction[Constant.AACranez]
            ht = Global.finalAction[Constant.AAHeadType]
            FWHead.fixHeadAt = (p,t,c,ht)
            HelpLong.openMouth(-650000)
        
        else:
##~             print "cant sense the ball, return false, at frame " , GrabTimer
            resetSpinDribble()
            return False
            
    elif GrabTimer < SpinTimer:
        
        if TurnDirection == Constant.dCLOCKWISE:
            Global.finalAction = list(Constant.TurnHoldingClockWise)
        else:   
            Global.finalAction = list(Constant.TurnHoldingCounterClockWise)
        
        FWHead.compulsoryAction = FWHead.fixHead
        p  = Global.finalAction[Constant.AAPanx]
        t  = Global.finalAction[Constant.AATilty]
        c  = Global.finalAction[Constant.AACranez]
        ht = Global.finalAction[Constant.AAHeadType]
        FWHead.fixHeadAt = (p,t,c,ht)
        HelpLong.openMouth(-650000)
        
    elif GrabTimer == SpinTimer:
    
        resetSpinDribble()
        if KickDirection == Constant.dCLOCKWISE:
            Kick.UPleft()
        else:
            Kick.UPright()
        
        return True

    return None
    

# =============================
# grabBall_forHandKick    
def grabBall_forHandKick():

    global GrabTimer, SpinTimer, TurnDirection, KickDirection
    
    GrabTimer += 1
    
##~     print "Grab ball timer " , GrabTimer , " mouth " , (VisionLink.getAnySensor(Constant.ssMOUTH)) , \
##~         " mouth " , (VisionLink.getAnySensor(Constant.ssMOUTH) - GRABBING_MOUTH_OPEN > STUCK_THRESHOLD)
    
    # Assume that the upenn is in 45 degrees
    if GrabTimer == 1:
        myh   = Global.selfLoc.getHeading()
        angle = HelpShort.normalizeAngle_180 (myh - Global.DKD[0])
        
        #------------------------ less opponent area
##~         closeOpp,closeOppDist = HelpLong.closestGpsOpponentWithinVar(80)
##~         if closeOpp is not None and closeOppDist is not None:
##~             print "d %.2f  , h %.2f" % (closeOppDist, closeOpp.getHeading())
##~         
##~         if (closeOppDist is not None and
##~             closeOpp is not None and
##~             closeOppDist < 80 and 
##~             closeOpp.getHeading() <= 180 and 
##~             closeOpp.getHeading() >= 0):
##~             KickDirection = Constant.dCLOCKWISE
##~             TurnDirection = Constant.dCLOCKWISE
##~             angle         = HelpShort.normalizeAngle_180 (Global.DKD[0] + 45 - myh)
##~             
##~         elif (closeOpp is not None and  
##~               closeOppDist is not None and 
##~               closeOppDist < 80 and 
##~               closeOpp.getHeading() > -180 and 
##~               closeOpp.getHeading() < 0):
##~             KickDirection = Constant.dANTICLOCKWISE
##~             TurnDirection = Constant.dANTICLOCKWISE
##~             angle         = HelpShort.normalizeAngle_180 (Global.DKD[0] - 45 - myh)
            
        if angle >= 0 and angle <= 180:
            KickDirection = Constant.dCLOCKWISE
            TurnDirection = Constant.dCLOCKWISE
            angle         = HelpShort.normalizeAngle_180 (Global.DKD[0] - myh)
            
        else:
            KickDirection = Constant.dANTICLOCKWISE
            TurnDirection = Constant.dANTICLOCKWISE
            angle         = HelpShort.normalizeAngle_180 (Global.DKD[0] - myh)
        
##~         if angle >= 0 and angle <= 180:
##~             angle     = HelpShort.normalizeAngle_180 (Global.DKD[0] + 45 - myh)
##~             TurnDirection = Constant.dCLOCKWISE
##~         else:
##~             angle     = HelpShort.normalizeAngle_180 (Global.DKD[0] - 45 - myh)
##~             TurnDirection = Constant.dANTICLOCKWISE
            
        SpinTimer     = int (abs(angle) * 0.3) + HOLD_BALL_TIME - 5
##~         print ""
##~         print "DKD is %.2f, myh %.2f" % (Global.DKD[0],myh) 
##~         print "KickDirection %d , TurnDirection %d , SpinTimer %d" % (KickDirection,TurnDirection,SpinTimer)
            
##~         print "setting up turn + upenn"
##~         print "myh %.2f , Diff %.2f, SpinTimer %.2f TurnDirection %d" % (myh,Diff,SpinTimer,TurnDirection)
        
    if GrabTimer < AVOID_KNOCK_BALL_TIME:
##~         print "avoid knock ball out"
        setNormalSlowWalk(0,0,0)
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,-5,15,Constant.HTAbs_h)

    elif GrabTimer < OPEN_MOUTH_TIME:
        setNormalSlowWalk (0,0,0)
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,-70,50,Constant.HTAbs_h)
        
    elif GrabTimer < TOTAL_GRAB_TIME:
        HelpLong.openMouth (GRABBING_MOUTH_OPEN)
        setNormalSlowWalk(0,0,0)
       
        if Global.lostBall == 0:
##~             print "See the ball , return false, at frame " , GrabTimer
            resetSpinDribble()
            return False
            
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,-70,50,Constant.HTAbs_h)        

    elif GrabTimer == TOTAL_GRAB_TIME:
        
        # If ball is seen, abort
        if Global.lostBall == 0:
##~             print "See the ball , return false, at frame " , GrabTimer
            resetSpinDribble()
            return False
            
        # If a ball is sensed, good
        if ballGrabbedUnderChin():
            Global.finalAction[Constant.AACmd:Constant.AAHeadType] = Constant.TurnHoldingClockWise [Constant.AACmd:Constant.AAHeadType]
            HelpLong.setForwardLeftTurn(0,0,0)
            HelpLong.openMouth(GRABBING_MOUTH_OPEN)
            FWHead.compulsoryAction = FWHead.fixHead
            FWHead.fixHeadAt        = (0,-70,50,Constant.HTAbs_h)        
            
        # If a ball is not sensed, abort
        else:
##~             print "cant sense the ball, return false, at frame " , GrabTimer
            resetSpinDribble()
            return False
            
    elif GrabTimer < HOLD_BALL_TIME:
        
        Global.finalAction[Constant.AACmd:Constant.AAHeadType] = Constant.TurnHoldingClockWise [Constant.AACmd:Constant.AAHeadType]
        HelpLong.setForwardLeftTurn(0,0,0)
        FWHead.compulsoryAction = FWHead.fixHead
        FWHead.fixHeadAt        = (0,-70,50,Constant.HTAbs_h)
        HelpLong.openMouth(GRABBING_MOUTH_OPEN)
        
    elif GrabTimer == HOLD_BALL_TIME:
        
        if Global.lostBall == 0:
##~             print "See the ball , return false, at frame " , GrabTimer
            resetSpinDribble()
            return False
         
        if ballGrabbedUnderChin():
            
            if TurnDirection == Constant.dCLOCKWISE:
                Global.finalAction = list(Constant.TurnHoldingClockWise)
            else:   
                Global.finalAction = list(Constant.TurnHoldingCounterClockWise)
            
            FWHead.compulsoryAction = FWHead.fixHead
            p  = Global.finalAction[Constant.AAPanx]
            t  = Global.finalAction[Constant.AATilty]
            c  = Global.finalAction[Constant.AACranez]
            ht = Global.finalAction[Constant.AAHeadType]
            FWHead.fixHeadAt = (p,t,c,ht)
            HelpLong.openMouth(-650000)
        
        else:
##~             print "cant sense the ball, return false, at frame " , GrabTimer
            resetSpinDribble()
            return False
            
    elif GrabTimer < SpinTimer:
        
        if TurnDirection == Constant.dCLOCKWISE:
            Global.finalAction = list(Constant.TurnHoldingClockWise)
        else:   
            Global.finalAction = list(Constant.TurnHoldingCounterClockWise)
        
        FWHead.compulsoryAction = FWHead.fixHead
        p  = Global.finalAction[Constant.AAPanx]
        t  = Global.finalAction[Constant.AATilty]
        c  = Global.finalAction[Constant.AACranez]
        ht = Global.finalAction[Constant.AAHeadType]
        FWHead.fixHeadAt = (p,t,c,ht)
        HelpLong.openMouth(-650000)
        
    elif GrabTimer == SpinTimer:
    
        resetSpinDribble()
        
        if KickDirection == Constant.dCLOCKWISE:
            Kick.handKickRight()
        else:
            Kick.handKickLeft()
        HelpLong.setForceStepComplete()
        return True

    return None
    
