#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: EdgeGoalie.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# Goalie.py
#
# Our rUNSWift 2004 Goal keeper
#
#===============================================================================

"""
Goal keeper - hangs further back - more conservative
"""

import math
import Kick
import Debug
import Global
import PawKick
import Constant
import HelpLong
import HelpShort
import HelpWhere
import HelpTrack
import SetPlayer
import GoalieHead
import VisionLink
import ReadyPlayer
import HoverToBall
import AttackerCruyff
import Ellipse

#===============================================================================
# Global variables of this module.
#===============================================================================

attackApproach = None
amAttacking = False
WIRELESS_BALL_VAR = 50

# Coordinates of where goalie should stand.
defendPos = [Constant.FIELD_WIDTH / 2, Constant.WALL_THICKNESS + 10, 90]

defendRadiusA = (Constant.GOAL_WIDTH / 2.0 + 5.0, 15.0)
defendRadiusB = (Constant.GOAL_WIDTH / 2.0 + 5.0, 0.5)
defendRadius = defendRadiusB
pivotPt = (Constant.FIELD_WIDTH / 2.0, Constant.WALL_THICKNESS + 0.0)

# Variables used when turning toward the field.
getOutOfGoalCounter   = 0
getOutOfGoalDirection = None

defendPosnAngle = 0

#===============================================================================
# Functions belong to this module.
#===============================================================================
def DecideNextAction():
    Global.frameReset()

    # Finished state - the robot can't move at all.
    if Global.state == Constant.FINISHED:
        HelpLong.showBatteryStatus()
        HelpShort.resetEverything()
    
    # Ready state - let me move to my correct position.
    elif Global.state == Constant.READYSTATE or Debug.mustReady:
        ReadyPlayer.readyMatchPlaying()

    # Set state - I can't move my legs, but I can move heads to look for ball.
    elif Global.state == Constant.SETSTATE:
        SetPlayer.DecideNextAction()

    # Playing state - lets play a game.
    else:
        playGameMode()

    # Show debug information using head lights.
    HelpLong.superDebug()
    sendGoalieWirelessInfo()

#-------------------------------------------------------------------------------
# Main flow of the goalie
def playGameMode():
    global attackApproach, defendRadius, defendRadiusA, amAttacking

    HelpLong.setDefaultAction()
    GoalieHead.DecideNextAction()

    shouldIKick, kick  = shouldIKickBall()    
    shouldIAttack      = shouldIAttackBall()
    shouldITurnToField = shouldITurnTowardTheField()
    
    amAttacking = False
    
    # Need to move toward the field ?
    if shouldITurnToField:
        if Debug.goalieDebug:
            print "Turning toward the field!"
        HelpLong.setNormalWalk(0, 0, getOutOfGoalDirection)
        moveAngleToMiddle()
        moveAngleToMiddle()
        return
    
    # Is the ball under my chin?
    if shouldIKick:
        defendRadius = defendRadiusA
        findDefendAngle(Global.ballX, Global.ballY, defendRadius)
    
        chooseKick(kick)
        
    # Want to approach the ball?
    elif shouldIAttack:
        defendRadius = defendRadiusA
        findDefendAngle(Global.ballX, Global.ballY, defendRadius)
        amAttacking = True
        attackBall()

    # Want to move back and defend.
    else:
        # Reset the planned attack approach if not intending to attack.
        attackApproach = None
        
        defendGoal()
    
    if Kick.shouldIContinueKick():
        Kick.continueKick()

def moveAngleToMiddle():
    global defendPosnAngle
    
    if defendPosnAngle < 0:
        defendPosnAngle += 0.5
    else:
        defendPosnAngle -= 0.5

# Defending when not seeing the ball.
def defendGoal():
    global defendRadius, defendRadiusA, defendRadiusB, defendPosnAngle, WIRELESS_BALL_VAR
    # Defending when not see vision ball = green.
    Global.myRole = Constant.GOALIEDEFENDNOTS
    
    # Last known vision position.
    if Global.lostBall < 5:
        if Debug.goalieDefendDebug:
            print "lostBall < 5 case!"
        if Global.lostBall == 0:
            Global.myRole = Constant.GOALIEDEFENDSEEB
        
        defendRadius = defendRadiusA
        findDefendAngle(Global.ballX, Global.ballY, defendRadius)
    
    # Wireless case.
    elif Global.sharedBallVar < HelpShort.get95cf(WIRELESS_BALL_VAR):
        if Debug.goalieDefendDebug:
            print "Wireless ball case!"
        defendRadius = defendRadiusB
        findDefendAngle(Global.sharedBallX, Global.sharedBallX, defendRadius)
        
    # GPS ball now. 
    elif (Global.lostBall <= Constant.VBALL_LOST
            and abs(Global.gpsLocalBall.getHeading()) < 90):
        if Debug.goalieDefendDebug:
            print "GPS ball case!"
        defendRadius = defendRadiusB
        findDefendAngle(Global.ballX, Global.ballY, defendRadius)
        
    # Default position.
    else:
        if Debug.goalieDefendDebug:
            print "I am confused, moving back to the original position!"
        # use previous angles, but decay back to looking straight out over time
        moveAngleToMiddle()
    
    findDefendPosn(defendRadius)
    moveToDefendPos()
    
# Moves the robot to the defend position. Stop the robot if it is in position.
def moveToDefendPos():
    global defendPos
    
    diffX = defendPos[0] - Global.selfLoc.getPos()[0]
    diffY = defendPos[1] - Global.selfLoc.getPos()[1]
    diffH = defendPos[2] - Global.selfLoc.getHeading()
    
    if (abs(diffX) <= 10 and abs(diffY) <= 10 and abs(diffH) <= 15):
        return
    else:
        HelpTrack.saGoToTargetFacingHeading(defendPos[0], defendPos[1], defendPos[2])

# Return true if I want to kick the ball.    
def shouldIKickBall():
    selfH      = HelpShort.normalizeAngle_90(Global.selfLoc.getHeading())
    ifUPenKick = AttackerCruyff.BallUdnerChinForUpenn()
    
    # Don't kick if you are facing the own goal.
    if selfH >= 225 or selfH <= -45:
        return [False, None]
    
    # At this region only left UPenn would be allowed.
    elif selfH < 225 and selfH >= 135 and ifUPenKick:
        if Debug.goalieDebug:
            print "Only left UPenn region!"
        return [True, Constant.UpennLeftWT]
 
    # At this region only right UPenn would be allowed.
    elif (selfH > -45 and selfH <= 45) and ifUPenKick:
        if Debug.goalieDebug:
            print "Right UPenn success! Heading - %f" %selfH
        return [True, Constant.UpennRightWT]
  
    # At this point, the self heading should be: 45 < self-heading < 135.
    if not (selfH > 45 and selfH < 135):
        # will get in here if we can't UPennKick
        return [False, None]
            
    # Now decided upon the enemy position.
    # Unsure region.
    if ifUPenKick:
        return [True, Constant.DiveKickWT]

    return [False, None]

# Return a list of opponent index sorting from smallest heading to largest heading.
def findGapsForKicking():
    gaps = []
    for i in range(Constant.NUM_TEAM_MEMBER):
        if Global.vOpponents[i].getConfidence() > 0 and \
           Global.vOpponents[i].getDistance() > 50:
           continue
           
        gaps.append(Global.vOpponents[i].getHeading())
        
    if gaps != None:
        gaps.sort(lambda x, y: cmp(x , y))

    return gaps
    
def chooseKick(kick):
    if kick == Constant.UpennRightWT:
        Kick.UPright()
    elif kick == Constant.UpennLeftWT:
        Kick.UPleft()
    elif kick == Constant.DiveKickWT:
        Kick.diveKick()
    else:
        Debug.goAndMakeTraceBack()
    
def isGoalieUPrightOk():
    if HelpShort.canSeeBall():
        ballx = math.sin (HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
        bally = math.cos (HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
                
        if (ballx < 3.5 and ballx > -5.5 and
            bally < 4.4 * Constant.BallRadius and
            Global.ballD < 18 and
            Global.degTilt < -9 and abs(Global.degPan) < 30):
            return True            
            
    return False
    
def isGoalieUPleftOk():
    if HelpShort.canSeeBall():
        ballx = math.sin (HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
        bally = math.cos (HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
        
        if (ballx < 3.0 and ballx > -5.0 and
            bally < 4.4 * Constant.BallRadius and
            Global.ballD < 18 and
            Global.degTilt < -9 and abs(Global.degPan) < 30):
            return True
            
    return False

def goalieBallIsGoingToArriveAtYouSoon():
    # Returns a tuple of (vel x[cm/frame], vel y[cm/frame], speed[cm/frame],
    # vel angle [degrees], vel angle - 90[degrees], normalized angle [doesn't work])
    vel = VisionLink.getGPSBallVInfo(Constant.CTLocal)
    
    # If heading away then ignore.
    if vel[1] >= 0:
        return False
    
    # Stop dividing zero.
    if vel[0] == 0:
        vel[0] = 0.01
        
    topEquation    = vel[1]
    bottomEquation = vel[0]
        
    m = topEquation / bottomEquation
    b = Global.ballY - Global.ballX * m
    x = -b / m
    
    if x < -30 or x > 30:
        return False

    if Global.ballD <= vel[2] * 75:
        return True
    else:
        return False
    
# Returns true if the goalie should attack the ball.
def shouldIAttackBall():
    global amAttacking
    
    attackHysteresis = 0
    
    if amAttacking:
        attackHysteresis = 20
    
    ## Immediate danger cases.
    ##
    if not HelpShort.canSeeBall():
        if Debug.goalieDebug:
            print "No attack because can't see the ball!"
        return False

    if HelpLong.teammateHasBall():
        if Debug.goalieDebug:
            print "Not attacking because my teammate is doing it!"
        return False
        
    # If you have gone too far, don't attack.
    if not HelpWhere.isInOwnGoalBox(10+attackHysteresis):
        if Debug.goalieDebug:
            print "No attack because I am not in the goal box (too far)!"
        return False
          
    if not HelpWhere.ballInOwnGoalBox(10+attackHysteresis):
        if Debug.goalieDebug:
            print "Ball not in the goal box!"
        return False

    # Ignore the middle scrum because the above case should cover that indirectly.
    #if (not HelpWhere.ballInOwnGoalBox(10)) and isBallDistanceTooFarInversly():
    #    if Debug.goalieDebug:
    #        print "I am too far away, no attack!"
    #    return False
       
    # If opponent is much closer than me, don't do attack.
    oppLoc, oppDistToBall = HelpLong.closestToBallGpsOpponentWithinVar(60)

    if oppLoc != None:
        # Come back later, is 25 a good distance? Maybe check the 2003 code.
        if oppDistToBall <= (Global.ballD - 25 - attackHysteresis):
            if Debug.goalieDebug:
                print "Opp ball distance: ", oppDistToBall , " My Ball dist: " , Global.ballD
                print "Opponent is much closer than me for the ball, hence no attack!", oppDistToBall
            return False
        else:
            if Debug.goalieDebug:
                print "Opp ball distance: ", oppDistToBall , " My Ball dist: " , Global.ballD

    if Debug.goalieDebug:
        print "Attacking the ball!"
        
    return True

def isBallDistanceTooFarInversly():
    goalieDistanceFromHome = HelpShort.getDistanceBetween(Global.selfLoc.getPos()[0], \
            Global.selfLoc.getPos()[1], Constant.FIELD_WIDTH / 2, 0)

    if not HelpShort.canSeeBall() or (HelpShort.canSeeBall() and \
        Global.ballD >= 75):
        return True
            
    requiredBallDist = 600.0 / goalieDistanceFromHome

    if Global.ballD >= requiredBallDist:
        if Debug.goalieDebug:
            print "requiredBallDist: " , requiredBallDist
        return True
    else:
        return False

#-------------------------------------------------------------------------------
# Makes the dog approach the ball, and kick if it is very close.
def attackBall():
    global attackApproach

    # Goalie attack = red.
    Global.myRole = Constant.GOALIEATTACK
    
    if attackApproach != None:
        if attackApproach == PawKick.FIRE_PAWKICK_AUTO:
            PawKick.firePawKick(PawKick.FIRE_PAWKICK_AUTO)
        elif attackApproach == PawKick.FIRE_PAWKICK_LEFT:
            PawKick.firePawKick(PawKick.FIRE_PAWKICK_LEFT)
        elif attackApproach == PawKick.FIRE_PAWKICK_RIGHT:
            PawKick.firePawKick(PawKick.FIRE_PAWKICK_RIGHT)
        elif attackApproach == Constant.HOVER_TO_BALL:
            HoverToBall.DecideNextAction()
        else:
            print "The attacking approach: %f" %attackApproach             
            Debug.goAndMakeTraceBack()        
        return
        
    if Global.ballD <= 30 or abs(Global.ballH) >= 30:
        if Debug.goalieDebug:
            print "Choose hovertoBall!"
        attackApproach = Constant.HOVER_TO_BALL
        HoverToBall.DecideNextAction()
        
    elif Global.selfLoc.getHeading() < 45:
        if Debug.goalieDebug:
            print "Choose Right-PawKick!"
        attackApproach = PawKick.FIRE_PAWKICK_RIGHT
        PawKick.firePawKick(PawKick.FIRE_PAWKICK_RIGHT)
        
    elif Global.selfLoc.getHeading() > 135:
        if Debug.goalieDebug:
            print "Choose Left-PawKick!"
        attackApproach = PawKick.FIRE_PAWKICK_LEFT
        PawKick.firePawKick(PawKick.FIRE_PAWKICK_LEFT)
    else:
        if Debug.goalieDebug:
            print "Choose Auto-Pawkick!"
        attackApproach = PawKick.FIRE_PAWKICK_AUTO
        PawKick.firePawKick(PawKick.FIRE_PAWKICK_AUTO)
    
# Send the data to other forwards.
def sendGoalieWirelessInfo():
    sendBallDistOrKickOff = Constant.LOSTBALL_DIST
    isStriker, isGoalie, haveBall, kickedOff, isDefending, isBackOff = \
                        [False, True, False, False, False, False]

    # Does goalie need to send haveBall signal?
                        
    # If I can see the ball, send the ball distance I calculated. If I can't
    # see the ball, do nothing as sendBallDist is initalised as LOSTBALL_DIST.
    if HelpShort.canSeeBall():
        sendBallDistOrKickOff = Global.ballD

    if Global.state == Constant.READYSTATE or Debug.mustReady:
        sendBallDistOrKickOff = Global.kickOffState

    VisionLink.sendMyBehaviourInfo(sendBallDistOrKickOff, isStriker, isGoalie, haveBall, \
                                   kickedOff, isDefending, isBackOff)

##############################################                                   
## Turning toward the soccer field.
##############################################                                   

# This is a direct port of the hacking made in 2003 goalie code. Although this was
# a hack, but it is very effective.
def shouldITurnTowardTheField():  
    global getOutOfGoalCounter, getOutOfGoalDirection

    ownGoal = Global.vOGoal
    turnCCW = -1
    
    # If we can see our own goal real big or we've got our face in either goal or we're locked
    # into a get-out-of-goal routine.
    if ((ownGoal.getConfidence() > 0 and (ownGoal.getWidth() >  \
        Constant.WIDTH_MOVE_THRESHOLD or (ownGoal.getHeight() > \
        Constant.HEIGHT_MOVE_THRESHOLD and ownGoal.getWidth() > \
        Constant.IMAGE_WIDTH / 2))) or getOutOfGoalCounter > 0):
        if getOutOfGoalDirection == None:
            if ownGoal.getHeading() >= 0:
                turnCCW = -30
            else:
                turnCCW = 30
            getOutOfGoalDirection = turnCCW
        else:
            turnCCW = getOutOfGoalDirection
                        
        HelpLong.setNormalWalk(0, 0, turnCCW)
        HelpLong.setHeadParams(0, 40, 0, Constant.HTAbs_h)
        getOutOfGoalCounter = getOutOfGoalCounter + 1
        
        MAX_INFRA_RED = 90.0
        irFar = VisionLink.getAnySensor(Constant.ssINFRARED_FAR)
        
        if getOutOfGoalCounter > 60 or Global.vTGoal.getConfidence() > 0 or \
            (irFar / 10000.0) >= MAX_INFRA_RED or Global.seenBeacon:
                getOutOfGoalCounter = 0
    
        return True
    else:
        getOutOfGoalDirection = None
        return False

##############################################                                   
## Intercept calculating functions
##############################################

def findDefendAngle(ballX, ballY, pivotRadius):
    global defendPosnAngle, pivotPt
    
    pivotX = pivotPt[0]
    pivotY = pivotPt[1]

    #print
    #print "pivot: " , pivotX , pivotY , " ball: " , ballX , ballY , " radii: " , pivotRadius[0], pivotRadius[1]

    if ballY < pivotY:
        ballY = pivotY

    # Find intercept between line through ball and pivot point and ellipse
    # centred at pivot point and of the pivot axis.
    # Perform translation to make pivot point (0, 0)
    relX = ballX - pivotX
    relY = ballY - pivotY
    
    if relX == 0 and relY == 0:
        relY = 1
    
    #print "relXY: " , relX , relY
    
    (angle, dist) = Ellipse.distToEllipse((relX, relY), pivotRadius)
    
    angleB = abs(angle-90)
    shrinkAngle = 45.0
    if angleB < shrinkAngle:
        sgn = ((angle - 90) < 0)
        angleB = shrinkAngle*math.sqrt(angleB/shrinkAngle)
        if sgn:
            angle = -angleB + 90
        else:
            angle = angleB + 90

    angle = Ellipse.getReducedAngle(angle, pivotRadius)
    
    defendPosnAngle = angle - 90
    
    return defendPosnAngle

def findDefendPosn(pivotRadius):
    global defendPosnAngle, pivotPt
    
    #print "Angle:", defendPosnAngle
    
    redAngle = defendPosnAngle + 90
    
    normAngle = Ellipse.getNormalAngle(redAngle, pivotRadius)
    
    #print "redAngle:", redAngle
    
    defendPos[0] = pivotPt[0] + pivotRadius[0] * math.cos(HelpShort.DEG2RAD(redAngle))
    defendPos[1] = pivotPt[1] + pivotRadius[1] * math.sin(HelpShort.DEG2RAD(redAngle))
    defendPos[2] = normAngle
    
    #print " dp: " , defendPos[0] , defendPos[1] , defendPos[2]
    
    return defendPos
