#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: FWHead.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# FWHead.py
#
# This module controls the forward's head movement.
#
#===============================================================================

import Debug
import Global
import LocateBall
import Constant
import HelpLong
import HelpShort
import HelpTrack
import ActiveLocalise

#===============================================================================
# Some globals that are needed
#===============================================================================

# Don't set these constants outside of this module.
defaultHeadAction   = 0
fixHead             = 1
mustSeeBall         = 2   # In some lockmode, the head should look at the ball, no matter what.
mustLocalise        = 3
mustSeeGpsBall      = 4
mustStatLocalise    = 5
localiseOnlyBallFar = 6

FOCUS_ON_BEACON_DEFAULT     = 10 # frames
MIN_POS_VAR_DEFAULT         = 55
MIN_HEAD_VAR_DEFAULT        = 25
DISALLOWLOCALISE_PERIOD     = 30

# The following can be set by outside modules, 
# but they'll be (and should be) reset to default in resetEachFrame() for every frame.
fixHeadAt               = (0,0,0,Constant.HTAbs_h)
minPosVariance          = MIN_POS_VAR_DEFAULT
minHeadVariance         = MIN_HEAD_VAR_DEFAULT
focusTotalTime          = FOCUS_ON_BEACON_DEFAULT
disAllowLocalisePeriod  = DISALLOWLOCALISE_PERIOD

# Assign compulsoryAction to fixHead, mustLocalise, mustLocalise, or defaultHeadAction
compulsoryAction        = defaultHeadAction

# Local timers. Other modules should not touch these.
focusDuration           = 0
timerSinceLastLocalise  = 0
isLocalising            = False

#--------------------------------------
# These variables need to be reset in each frame.
def resetEachFrame():
    global mustSeeBall, mustLocalise, fixHead, fixHeadAt
    global defaultHeadAction, compulsoryAction
    global minPosVariance, minHeadVariance, MIN_POS_VAR_DEFAULT, MIN_HEAD_VAR_DEFAULT
    global focusTotalTime, FOCUS_ON_BEACON_DEFAULT
    global disAllowLocalisePeriod, DISALLOWLOCALISE_PERIOD
    
    minPosVariance   = MIN_POS_VAR_DEFAULT
    minHeadVariance  = MIN_HEAD_VAR_DEFAULT
    focusTotalTime   = FOCUS_ON_BEACON_DEFAULT
    compulsoryAction = defaultHeadAction
    fixHeadAt        = (0,0,0,Constant.HTAbs_h)

#--------------------------------------
# Set the head parameters.
#
# The algorithm flow:
#
#   1. Special cases
#   2. Active localise
#   3. Vision ball
#   4. Look at previous location
#   5. Wireless
#   6. GPS
#   7. Locateball
def DecideNextAction():
    global mustSeeBall, mustLocalise, fixHead, fixHeadAt, compulsoryAction
    global focusTotalTime, focusDuration, timerSinceLastLocalise, isLocalising
    global minHeadVariance, minPosVariance
    global disAllowLocalisePeriod, DISALLOWLOCALISE_PERIOD

    #---------------------------------------------------------------------------
    # Special cases.
    #---------------------------------------------------------------------------
    
    #---------------------------------------------------------------------------
    # Special case 1: Sometimes players want to have their own heading.
    if (compulsoryAction == fixHead):
        HelpLong.setHeadParams (*tuple(fixHeadAt))
        return

    # Special case 2: Sometimes players want to see the ball no matter what.
    if (compulsoryAction == mustSeeBall):
        HelpTrack.trackVisualBall()
        return
        
    # Spcical case 3: Force a localisation - caller will set its own timers.
    if (compulsoryAction == mustLocalise):
        isLocalising    = True
        focusDuration   = 0
        ActiveLocalise.SmartSetBeacon(90,500)
        ActiveLocalise.DecideNextAction()
        return

    # Spcical case 4:  Force to track the gps ball.
    if (compulsoryAction == mustSeeGpsBall):
        HelpTrack.trackGpsBall()
        return
        
    # Spcical case 5:  Force to stationary localise
    if (compulsoryAction == mustStatLocalise):
        HelpTrack.stationaryLocalise()
        return
        

    #---------------------------------------------------------------------------
    # Default behaviour of the head.
    #---------------------------------------------------------------------------
    #
    # Notice that striker, supporter etc are using the default behaviour.
    
    #---------------------------------------------------------------------------
    # Already localising.
    if (isLocalising and focusDuration < focusTotalTime):
        # Because of beaconAlternate, we do not want to do smartSet so many times, 
        # otherwise the head will flutuate.
        # ActiveLocalise.SmartSetBeacon(90,500)

        ActiveLocalise.DecideNextAction()
        focusDuration += 1

        # Counter-act the effect of increasing lostBall counter while localising,
        # if not have this, locateBall will be (inappropriate) quickly triggereds.
        Global.lostBall = HelpShort.DECREMENT(Global.lostBall)

    #---------------------------------------------------------------------------
    # Reset variables if finished focusing on beacon.
    elif (isLocalising and focusDuration >= focusTotalTime):
        isLocalising            = False
        timerSinceLastLocalise  = 0
        disAllowLocalisePeriod  = DISALLOWLOCALISE_PERIOD

    #---------------------------------------------------------------------------
    # Trigger the active localise only if
    # - Have been quite a while since localising last time. AND
    # - a) Dog not sure where it is, or where it is heading, OR
    #   b) Haven't done it for a while (var is small doesn't mean the position is defiinitely correct)
    #
    # TODO: Active localise only if you think no opponent nearby the ball.
    elif ((timerSinceLastLocalise > disAllowLocalisePeriod)  and
            not (compulsoryAction == localiseOnlyBallFar and HelpShort.canSeeBall()
                and Global.ballD <= 40)
         and (#timerSinceLastLocalise > 150
                Global.selfLoc.getPosVar()     > HelpShort.get95cf(minPosVariance)
                or Global.selfLoc.getHeadingVar() > HelpShort.get95cf(minHeadVariance)
             )
        ):

        focusDuration   = 0     # Increase from this moment on.
        isLocalising    = True
        ActiveLocalise.SmartSetBeacon(90,500)
        ActiveLocalise.DecideNextAction()

    # If we want always localise after the disAllowLocalisePeriod, then un-comment below out.
##~     elif ((timerSinceLastLocalise > disAllowLocalisePeriod)  
##~         focusDuration   = 0     # Increase from this moment on.
##~         isLocalising    = True
##~         ActiveLocalise.SmartSetBeacon(90,500)
##~         ActiveLocalise.DecideNextAction()
    
    #---------------------------------------------------------------------------
    # The following codes can be compared with the beckhamGeneralHead().
    else:
        timerSinceLastLocalise += 1
        
        # If there is a vision ball, then track it. If there is no vision ball (gps or wireless)
        # but only lost it for few frames then track at the last vision ball position.
        if (Global.lostBall < 5): 
            HelpTrack.trackVisualBall()
            if Debug.headDebug:
                print "FWHead: lostBall < 5!"
            
        # If the vision ball is still lost not for very long (however the ballSource is either
        # wireless or gps), then look at the gps ball if the heading is appropriate.
        elif (Global.lostBall <= Constant.VBALL_LOST
                and abs(Global.gpsLocalBall.getHeading()) < 90): 
            HelpTrack.trackGpsBall()
            if Debug.headDebug:
                print "FWHead: look at gps ball < 5!"

        elif (Global.lostBall <= Constant.VBALL_LOST):
            HelpTrack.trackVisualBall()
            
        else:
            LocateBall.headFindBall(-10, 0)
            
##~         elif (Global.lostBall <= Constant.VBALL_LOST):
##~             LocateBall.headFindBall(-50, -5)
##~             if Debug.headDebug:
##~                 print "track visual ball >= 5"
##~                 
##~         elif (Global.sharedBallVar < Constant.LARGE_VAL and abs(sh) < 90):
##~             HelpLong.setHeadParams(-sx,Constant.BallDiameter,sy,Constant.HTAbs_xyz)
##~             if Debug.headDebug:
##~                 print "track wireless ball"
##~                 
##~         else:
##~             HelpTrack.trackVisualBall()

