#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: Forward.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# Forward.py
#
# Our rUNSWift 2004 Forward Player!!
#
#===============================================================================

"""
Forward, switching between Cruyff, Beckenbaurer and Zidane
"""

import Kick
import Debug
import FWHead
import Global
import Constant
import HelpLong
import HelpShort
import HelpTrack
import SetPlayer
import ReadyPlayer
import BackOffAnand
import DefenderHomer
import SupportZidane
import StrikerBecken
import AttackerCruyff
import LandmarkTester

# Why Cruyff, Zidane and Beckenbaurer? :)
#
#       >> Because according to my quick-research, they are the top 3 "Best European 
#           Players" in the past 50 years. The poll is from 7 million votes and conducted 
#           by UEFA as they celebrate their 50th anniversary since 1954 - daniel

#==================================================================
# Functions belong to this module.
#===============================================================================

# Main diffeence in behaviour codes in 2004 compare to 2003:
#
#       >> Instead of having lockmodes to lock into a particular skill, try to ensure
#           that the same decision will be yield again. We may be able to achieve 
#           this by considering "what actions were made before?", plus careful 
#           condition.
#

#--------------------------------------
# New designed 2004 decision tree. This decision tree can do everything the old tree can do, but it is
# more flexible. Hence less hack should be introduced.
def playGameMode():
    resetEachFrame()    

    shouldIStartDefend = DefenderHomer.shouldIStartDefend()
    shouldIEndDefend   = DefenderHomer.shouldIEndDefend()
    shouldIStriker     = StrikerBecken.shouldIBecomeStriker()
    shouldIBackOff     = BackOffAnand.shouldIVisualBackOff()
    shouldISupport     = SupportZidane.ShouldIBecomeSupporter()
    
    if ((shouldIStartDefend or DefenderHomer.lockDefender) and \
         not shouldIEndDefend) or Debug.mustDefender:
        Global.myRole = Constant.DEFENDER
        DefenderHomer.DecideNextAction()

    elif shouldIStriker or Debug.hackStriker:
        StrikerBecken.DecideNextAction()
        
    elif shouldIBackOff:
        BackOffAnand.DecideNextAction()

    elif shouldISupport:
        SupportZidane.DecideNextAction()
        
    else:    
        AttackerCruyff.DecideNextAction()
        
        if BackOffAnand.shouldISideBackOff():
            Global.myRole = Constant.SIDEBACKOFF
            Global.finalAction[Constant.AAForward] = \
                Global.finalAction[Constant.AAForward] * 0.6        
                
    FWHead.DecideNextAction()

    if Kick.shouldIContinueKick():
        Kick.continueKick()
        
#--------------------------------------
# Act according to different states.

def DecideNextAction():
    if (Global.debugBehaviour_line):
        print "..... new frame: ", Global.frame

    Global.frameReset()

    # Ted: In python only three states are supported. Other states are in still
    #          in C++. Why? If I port all the states to python then none of the C++
    #          behaviour would work. I think Kim still need his kimForward.cc
    #          and walkingLearner.cc.
    
    # Finished state - the robot can't move at all.
    if Global.state == Constant.FINISHED:
        HelpLong.showBatteryStatus()
        HelpShort.resetEverything()

    # Ready state - let me move to my correct position.
    elif Global.state == Constant.READYSTATE or Debug.mustReady:
        ReadyPlayer.readyMatchPlaying()
        DefenderHomer.lockDefender = False # shouldn't keep birding on new kickoff

    # Set state - I can't move my legs, but I can move heads to look for ball.
    elif Global.state == Constant.SETSTATE:
        SetPlayer.DecideNextAction()
        DefenderHomer.lockDefender = False # shouldn't keep birding on new kickoff

    # Playing state - lets play a game.
    else:
        playGameMode()

    # Show debug information using head lights.
    # If landmarkDebugging is on, no more face pattern for skills will be shown.
    # superDebug will be run inside it also, so don't need to call twice.
    landmarkDebugging = False
    if (landmarkDebugging):
        LandmarkTester.DecideNextAction()
    else:
        HelpLong.superDebug()

    # Send wireless info to the teammates.
    HelpLong.sendWirelessInfo()
    
#--------------------------------------
# General Initialisation
def resetEachFrame():
    Global.DKD = HelpTrack.getDKDRange()
    
    if Global.state != Constant.FINISHED:
        Global.myRole  = Constant.NOROLE
        
    HelpLong.setDefaultAction()
    HelpLong.setDefaultIndicators()
    FWHead.resetEachFrame()
