#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: Global.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# Global.py
#
# Contains visual and gps information, and also commands to be sent to
# actuators. Every player modules import this to use these info.
#
#===============================================================================

import math
import Kick
import Constant
import HelpLong
import HelpShort
import HelpTrack
import VisionLink
import ReadyPlayer

#===============================================================================
# Global variables of this module.
#===============================================================================

# These variables are for any behaviour module to read and write.
# Other modules (any player codes) can access this variable by something like
# Global.finalAction

#-----------------------------------------------------------------------------------------------------------------------------------------
lockStuck   = 0
lastForward = None
lastLeft    = None
lastTurnCCW = None

# Alpha.
alphaInvereseSquare = 80.0

# Ready state kickoff signal.
kickOffState = Constant.NOTDECIDED

# A list of forward/left/turn/etc for behaviour codes to write to. Converted to 
# list so that it is mutable (the original is a tuple which is immutable).
finalAction = list(Constant.NormalWalkStandStill)
finalIndicators = list(Constant.INDICATORS_DEFAULT)
lastIndicators = None

# The current frame number, set incremented by Behaviou.processFrame()
frame    = 0
lastFrameReset = 0 #to prevent frameReset() more than twice in the same frame 
##lockmode = None

# Robot state.
state     = None

# My team color
teamColor = None

# Head motion
pan         = None  # These are in radians.
tilt        = None
crane       = None
degPan      = None  # These are in degrees.
degTilt     = None
degCrane    = None

lastPan         = 0     # Used by HelpTrack.trackVisualBall
lastTilt        = 0     # The last pan, tilt, crane you see a ball.
lastCrane       = 0
lastHeadType    = Constant.HTAbs_h
desiredHead     = (0,0,0) # (desiredPan, desiredTilt, desiredCrane)
desiredPan,desiredTilt,desiredCrane = desiredHead

# GPS (self) information.
gpsGlobalBall = None  # Vector class below
gpsLocalBall  = None  # Vector 
selfLoc       = None  # WMObj
teammatesLoc  = None  # Tuple of WMObj.

# GPS (shared - wireless) information
teamPlayers                 = None # Contain info beside location.
framesSince3ForwardsPresent = 0
framesSince2ForwardsPresent = 0

# Wireless ball from teammate with ok-low-variance
# It does not have a class, but simply a tuple of (x,y, variance, source robot)
# from getSharedBallInfo.
sharedBall      = None
sharedBallX     = None      # = sharedBall[0], for convenience
sharedBallY     = None      # = sharedBall[1], for convenience
sharedBallVar   = None      # = sharedBall[2], for convenience
sharedBallSourceRobot     = None      # = sharedBall[3], for convenience

# Vision information
vBall       = None      # VisualObject
vTeammates  = None      # List of VisualObject
vOpponents  = None
vTGoal      = None
vOGoal      = None

# For convenience, got set by HelpTrack.determineBallSource()
#--------------------------------------
# NOTE: ballX and ballY are global coordinate
# BUT:    ballH and ballD are relative to the robot neck base
# Want a local one? Go to HelpShort.getLocalCoordinate

ballX       = None
ballY       = None
ballD       = None
ballH       = None
ballSource  = None

#PWalkInfo
pWalkInfo   = None      # PWalkInfo

seenBeacon  = None

# How many frames vision ball lost
lostBall    = 0

# How many frames vision has ball 
haveBall    = 0

# Debug variables.
myRole  = Constant.NOROLE
debugBehaviour_line = False
bDebugForwardLeftTurn = False

# DKDRange (dkd, narrowMin, narrowMax, broadMin, broadMax)
DKD     = None

# The number of forwards
forwardAlive = 1

# The player number I am assigned, default is 1
myPlayerNum  = VisionLink.getMyPlayerNum()

# Motor gain - is it high gain
isHighGain      = False
forceHighGain   = None #True if forced to be High Gain, False if forced to be low gain

# Stuck detection. 1 second = 30 frames.
STUCK_NUMBER_OF_FRAMES = 60
lastMaxPWMValues = [(0,0)] * STUCK_NUMBER_OF_FRAMES

#===============================================================================
# Functions belong to this module
#===============================================================================

#--------------------------------------
# Called by processFrame() in Behaviou.py, update ball/self location to latest
# information for each frame.
def frameReset():
    global lastFrameReset, forceHighGain
    
    if frame == lastFrameReset:
        return #already reset in this frame, quit
    else:
        lastFrameReset = frame
    
    global myRole
    
    myRole = Constant.NOROLE
    
    # Need to be reset, otherwise it over written dynamic_gain, which is very dangerous
    forceHighGain = None
    
    if not Kick.shouldIContinueKick() and Kick.framesSinceLastKicking <= 500:
        Kick.framesSinceLastKicking += 1
    
    #-----------------------------------
    # Get the head motion info
    global pan, tilt, crane
    global degPan, degTilt, degCrane
    
    degPan   = VisionLink.getJointSensor(Constant.jHeadPan)
    degTilt  = VisionLink.getJointSensor(Constant.jHeadTilt)
    degCrane = VisionLink.getJointSensor(Constant.jHeadCrane)

    pan   = HelpShort.DEG2RAD(degPan)
    tilt  = HelpShort.DEG2RAD(degTilt)
    crane = HelpShort.DEG2RAD(degCrane)
    
    #-----------------------------------
    # Get the GPS info    
    global selfLoc, teammatesLoc, gpsLocalBall, gpsGlobalBall
    
    gpsGlobalBall = Vector(*VisionLink.getGPSBallInfo(Constant.CTGlobal))
    gpsLocalBall  = Vector(*VisionLink.getGPSBallInfo(Constant.CTLocal))
    selfLoc       = WMObj(*VisionLink.getGPSSelfInfo())

    teammatesLoc = []
    for i in range(Constant.NUM_TEAM_MEMBER):
        teammatesLoc.append(
                    WMObj(*VisionLink.getGPSTeammateInfo(
                                    Constant.CTGlobal, i+1)))

    #--------------------------------------
    # Get the Shared ball
    global sharedBall, sharedBallX, sharedBallY, sharedBallVar, sharedBallSourceRobot

    sharedBall = VisionLink.getSharedBallInfo();
    sharedBallX, sharedBallY, sharedBallVar, sharedBallSourceRobot = sharedBall

    #-----------------------------------
    # Get pWalkInfo
    
    global pWalkInfo,desiredHead,desiredPan,desiredTilt,desiredCrane,isHighGain
    
    pWalkInfo     = PWalkInfo(*VisionLink.getPWalkInfo())
    desiredHead   = pWalkInfo.getDesiredHead()
    desiredPan,desiredTilt,desiredCrane = desiredHead
    isHighGain = pWalkInfo.isUsingHighGain()
    
    #-----------------------------------
    # Get the VisualObjects
    
    global teamColor
    
    teamColor = VisionLink.getTeamColor()
    if teamColor == Constant.RED_TEAM: # RED
        teammateRange = range (Constant.vobRedDog,Constant.vobBlueDog)
        opponentRange = range (Constant.vobBlueDog,Constant.vobBlueDog4+1)
        ownGoal       = Constant.vobYellowGoal
        targetGoal    = Constant.vobBlueGoal
    elif teamColor == Constant.BLUE_TEAM: # BLUE
        teammateRange = range (Constant.vobBlueDog,Constant.vobBlueDog4+1)
        opponentRange = range (Constant.vobRedDog,Constant.vobBlueDog)
        ownGoal       = Constant.vobBlueGoal
        targetGoal    = Constant.vobYellowGoal
    else:
        print "Warning: teamColor is None, the dog's coordinate is wrong"

    global vBall, vTeammates, vOpponents, vTGoal, vOGoal, seenBeacon

    blueLeft    = VisualObject(*VisionLink.getVisualObject(Constant.vobBlueLeftBeacon))
    blueRight   = VisualObject(*VisionLink.getVisualObject(Constant.vobBlueRightBeacon))
    yellowLeft  = VisualObject(*VisionLink.getVisualObject(Constant.vobYellowLeftBeacon)) 
    yellowRight = VisualObject(*VisionLink.getVisualObject(Constant.vobYellowRightBeacon))

    if blueLeft.getConfidence() > 0 or blueRight.getConfidence() > 0 or \
        yellowLeft.getConfidence() > 0 or yellowRight.getConfidence() > 0:
        seenBeacon = True
    else:
        seenBeacon = False
    
    vBall   = VisualObject(*VisionLink.getVisualObject(Constant.vobBall))
    
    vTeammates = []
    for i in teammateRange:
        vTeammates.append(VisualObject(*VisionLink.getVisualObject(i)))

    vOpponents = []
    for i in opponentRange:
        vOpponents.append(VisualObject(*VisionLink.getVisualObject(i)))
    
    vOGoal = VisualObject(*VisionLink.getVisualObject(ownGoal))
    vTGoal = VisualObject(*VisionLink.getVisualObject(targetGoal))
    
    #-----------------------------------
    # Goals doesnt change thier global positions
    vOGoal.setPos(Constant.OWN_GOAL_X   ,Constant.OWN_GOAL_Y   )
    vTGoal.setPos(Constant.TARGET_GOAL_X,Constant.TARGET_GOAL_Y)

    #-----------------------------------
    # Set default final actions/indicators.
    global finalIndicators, finalAction
    
##~     finalIndicators = list(Constant.INDICATORS_DEFAULT)
    finalAction = list(Constant.NormalWalkStandStill)
    
    #-----------------------------------
    # Other variable to be reset
    global lostBall, haveBall

    # Increase the lostBall counter if can't see a ball.
    if HelpShort.canSeeBall():
        lostBall = 0
        haveBall += 1
    else:
        lostBall += 1
        haveBall = 0

    HelpTrack.determineBallSource()
        
    #-----------------------------------
    # Set the wireless teammate communication variables.
    global teamPlayers, forwardAlive, framesSince3ForwardsPresent, framesSince2ForwardsPresent
    wirelessInfo = VisionLink.getWirelessTeammateInfo()

    teamPlayers = []
    for i in range(Constant.NUM_TEAM_MEMBER):
        teamPlayers.append( \
            WirelessTeammateInfo( \
                [wirelessInfo[i], \
                 wirelessInfo[i+4], \
                 wirelessInfo[i+8],  \
                 wirelessInfo[i+12], \
                 wirelessInfo[i+16], \
                 wirelessInfo[i+20], \
                 wirelessInfo[i+24], \
                 wirelessInfo[i+28]]))
    
    
    #------------------------------------------------
    # This is the RoboCup Game state, see the rule book
    
    global state
    
    # If first call to ready state, reset the readyplayer global
    if (VisionLink.getTheCurrentMode() == Constant.READYSTATE and
        state != Constant.READYSTATE):
        ReadyPlayer.initReadyMatchPlaying()

    state = VisionLink.getTheCurrentMode()
    
    if HelpLong.forwardAlive() == 3:
        framesSince3ForwardsPresent = 0
    else:
        framesSince3ForwardsPresent = framesSince3ForwardsPresent + 1

    if HelpLong.forwardAlive() >= 2:
        framesSince2ForwardsPresent = 0
    else:
        framesSince2ForwardsPresent = framesSince2ForwardsPresent + 1

    global lastMaxPWMValues
    lastMaxPWMValues.pop(0)
    lastMaxPWMValues.append(HelpLong.getMaxPWMDuty())
    
#===============================================================================
# Sub-classes of this module
#===============================================================================

class WirelessTeammateInfo:
    # Wireless information from teammates.
    
    # counter: Whenever wireless info arrive, counter is set to 12 (set by gps.cc)
    #                 When nothing is received, it'll be decreased by 1 unless already 0.
    #                 Tell you how recent the information is.
    # isDefender: Set if is a defender.
    # isStriker:      Set if is a striker.
    # ballDist:     Distance to ball.
    # isGoalie:    Set if is a goalie, forward otherwise.
    # hasBall:   Set if have ball and want support.
    # isBackOff:  Set if is backing off.
    # kickedOff:  Set if have just kickedOff

    def __init__(self, arg):
        if len(arg) == 8:
            self.__counter, self.__isDefender, self.__isStriker, self.__ballDist, self.__isGoalie, self.__hasBall, self.__isBackOff, self.__kickedOff = arg
        else:
            print "arg length = 8", len(arg)
            raise Exception("Number of argument to Vector() must be 8")
    def getCounter(self):
        return (self.__counter)
    def isDefender(self):
        return (self.__isDefender)
    def isStriker(self):
        return (self.__isStriker)
    def getBallDist(self):
        return (self.__ballDist)
    def isGoalie(self):
        return (self.__isGoalie)
    def hasBall(self):
        return (self.__hasBall)
    def isBackOff(self):
        return (self.__isBackOff)
        
    #Ted's hacked value, why define a function, why not use getBallDist() --> $#@%
    def getKickedOff(self):
        return (self.__ballDist)
        
    def isKickedOff(self):
        return self.__kickedOff
    def getTuple(self):
        return (self.__counter, self.__isDefender, self.__isStriker, self.__ballDist, self.__isGoalie, self.__hasBall, self.__isBackOff, self.__kickedOff )
        
#--------------------------------------
# Self information. 
class Vector:
    # x,y: Cartesian coordinates of the vector, usually mean coordinates on field.
    # theta, d: Polar format.
    # head: theta - 90deg (I don't know what for).
    # angleNorm: Angle before wrapping round to 0 again. Not used yet, may del later.
    
    def __init__(self,*arg):
        if len(arg) == 6:
            self.__x , self.__y, self.__d , self.__theta , self.__head , self.__angleNorm = arg
        else:
            print "arg length = ", len(arg)
            raise Exception("Number of argument to Vector() must be 6")
    def getX(self):
        return (self.__x)
    def getY(self):
        return (self.__y)
    def getPos(self):
        return (self.__x, self.__y)        
    def getDistance(self):
        return self.__d        
    def getTheta(self):
        return self.__theta        
    def getAngleNorm(self):
        return self.__angleNorm
    def getHeading(self):
        return self.__head
    def sub(self,otherVec):
        self.__x -= otherVec.__x
        self.__y -= otherVec.__y
        return self
    def calcTheta(self):
        if (self.__x == 0 or self.__y == 0):
            self.__theta = 0
        else:
            self.__theta = HelpShort.RAD2DEG(math.atan2(self.__y, self.__x));
    def calcHead(self):
        self.calcTheta()

        # principle() not ported, I think the one below is equivalent - dan.
        # self.__head = principle(self.__theta - HALF_CIRCLE / 2);
        self.__head = HelpShort.normalizeAngle_0_360(self.__theta - 90);
    def getTuple(self):
        return (self.__x , self.__y, self.__d , self.__theta , self.__head , self.__angleNorm)
        
#--------------------------------------
# World model objects information.
class WMObj:
    def __init__(self,*arg):
        if len(arg) != 5:
            raise Exception("Number of argument to WMObj() must be 5")
        self.__x, self.__y, self.__posVar, self.__h, self.__hVar = arg
    def getX(self):
        return (self.__x)
    def getY(self):
        return (self.__y)
    def getPos(self):
        return (self.__x, self.__y)        
    def getPosVar(self):
        return self.__posVar
    def getHeading(self):
        return self.__h
    def getHeadingVar(self):
        return self.__hVar
    def getTuple(self):
        return (self.__x, self.__y, self.__posVar, self.__h, self.__hVar)
        
#--------------------------------------
# Visual objects information.

class VisualObject:
    def __init__(self,*arg):
        if len(arg) != 14:
            raise Exception("Number of argument to VisualObject() must be 14, while it's %d" % len(arg))
        self.__cf, self.__head, self.__d, self.__var, \
            self.__x, self.__y,  self.__cx, self.__cy, \
            self.__height, self.__width, \
            self.__imgHead, self.__imgElev, self.__vobDist,\
            self.__elev = arg
    def getConfidence(self):
        return self.__cf
    def getVariance(self):
        return self.__var
    def getCentroid(self):
        return (self.__cx, self.__cy)
    def getX(self):
        return self.__x
    def getY(self):
        return self.__y
    def getPos(self):
        return (self.__x, self.__y) # note it is the global x,y position on field
    def getHeight(self):
        return self.__height
    def getWidth(self):
        return self.__width
    def getImgHeading(self):
        return self.__imgHead
    def getImgElevation(self):
        return self.__imgElev
    def getHeading(self):         # the heading of the object relative tot the robot's neckbase
        return self.__head
    def getDistance(self):        # the flat distance from the robot's neck-base to the vob
        return self.__d
    def getVobDistance(self):
        return self.__vobDist
    def setPos(self,x,y):
        self.__x = x
        self.__y = y
    def getElev(self):
        return self.__elev
    def getTuple(self):
        return (self.__cf, self.__head, self.__d, self.__var, \
            self.__x, self.__y,  self.__cx, self.__cy, \
            self.__height, self.__width, \
            self.__imgHead, self.__imgElev,self.__vobDist,self.__elev)
    
    
#--------------------------------------
# PWalk information. 
class PWalkInfo:
    def __init__(self,*arg):
        if len(arg) != 10:
            raise Exception("Number of argument to PWalkInfo() must be 10, but it was %d" % len(arg))
    
        self.__walkType, self.__step, self.__PG, \
            self.__leadingLeg, self.__frontShoulderHeight, self.__bodyTilt,\
            self.__desiredPan, self.__desiredTilt, self.__desiredCrane, \
            self.__isHighGain = arg

    def getCurrentWalkType(self):
        return self.__walkType

    def getCurrentStep(self):
        return self.__step

    def getCurrentPG(self):
        return self.__PG        

    def getLeadingLeg(self):
        return self.__leadingLeg

    def getFrontShoulderHeight(self):
        return self.__frontShoulderHeight
        
    def getBodyTilt(self):
        return self.__bodyTilt
        
    def getDesiredHead(self):
        return (self.__desiredPan, self.__desiredTilt, self.__desiredCrane)

    def isUsingHighGain(self):
        return self.__isHighGain
        
    def getTuple(self):
        return (self.__walkType, self.__step, self.__PG, \
            self.__leadingLeg, self.__frontShoulderHeight, self.__bodyTilt,\
            self.__desiredPan, self.__desiredTilt, self.__desiredCrane, \
            self.__isHighGain)
