#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: Goalie.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# Goalie.py
#
# Our rUNSWift 2004 Goal keeper
#
#===============================================================================

"""
Goal keeper
"""

import math
import Kick
import Debug
import Global
import PawKick
import Constant
import HelpLong
import HelpShort
import HelpWhere
import HelpTrack
import SetPlayer
import GoalieHead
import LocateBall
import VisionLink
import ReadyPlayer
import HoverToBall
import AttackerCruyff

#===============================================================================
# Global variables of this module.
#===============================================================================

LAZINESS = 8            # How reluctant goalie is to move from current spot
ACCURACY = 3            # Accuracy goalie should aim for (less than LAZINESS)
MAX_VARIANCE = 500      # Max variance allowed for gps or wireless ball use
ATTACK_CONE = 40        # Maximum ball heading allowed to attack
ATTACK_DIST = 25        # Ba##~     if not grabbingBall:
MAX_ATTACK_DIST = 50    # Once goalie attacking, this is ball distance breakoff
BALL_LOST_LIMIT = 90    # Frames before goalie resorts to central position
DEFEND_SCALER = (35, 25)# Defines the defence ellipse size
DEFEND_PIVOT = (Constant.FIELD_WIDTH / 2, 10) # Pivot is the ellipse center



shouldMove = False      # Indicates whether goalie should be moving
attacking = False       # Indicates if goalie is attacking or not
ballLostTimer = 0       # Number of frames since ball was last seen
attackApproach = None



# Coordinates of where goalie should stand.
defendPos = [Constant.FIELD_WIDTH / 2, Constant.WALL_THICKNESS + 10, 90]

# Variables used when turning toward the field.
getOutOfGoalCounter   = 0
getOutOfGoalDirection = None

#===============================================================================
# Functions belong to this module.
#===============================================================================
def DecideNextAction():
    Global.frameReset()

    # Finished state - the robot can't move at all.
    if Global.state == Constant.FINISHED:
        HelpLong.showBatteryStatus()
        HelpShort.resetEverything()
    
    # Ready state - let me move to my correct position.
    elif Global.state == Constant.READYSTATE or Debug.mustReady:
        ReadyPlayer.readyMatchPlaying()

    # Set state - I can't move my legs, but I can move heads to look for ball.
    elif Global.state == Constant.SETSTATE:
        SetPlayer.DecideNextAction()

    # Playing state - lets play a game.
    else:
        playGameMode()

    # Show debug information using head lights.
    HelpLong.superDebug()
    sendGoalieWirelessInfo()

#-------------------------------------------------------------------------------
# Main flow of the goalie
def playGameMode():
    global attackApproach

    HelpLong.setDefaultAction()
    GoalieHead.DecideNextAction()

    shouldIKick, kick  = shouldIKickBall()    
    shouldIAttack      = shouldIAttackBall()
    shouldITurnToField = shouldITurnTowardTheField()
        
    # Need to move toward the field ?
    if shouldITurnToField:
        if Debug.goalieDebug:
            print "Turning toward the field!"
        return
    
    # Is the ball under my chin?
    if shouldIKick:
        chooseKick(kick)
        
    # Want to approach the ball?
    elif shouldIAttack:
        attackBall()

    # Want to move back and defend.
    else:
        # Reset the planned attack approach if not intending to attack.
        attackApproach = None
        
        if HelpShort.canSeeBall():
            defendWhenBallSeen()
        else:
            defendWhenBallNotSeen()
    
        moveToDefendPos()

    if Kick.shouldIContinueKick():
        Kick.continueKick()

#  Defending when seeing the ball.
def defendWhenBallSeen():
    # Defending when see vision ball = yellow.
    Global.myRole = Constant.GOALIEDEFENDSEEB
    calculateDefendPositionHeading()
    
# Defending when not seeing the ball.
def defendWhenBallNotSeen():
    # Defending when not see vision ball = green.
    Global.myRole = Constant.GOALIEDEFENDNOTS
    
    # Last known vision position.
    if Global.lostBall < 5:
        if Debug.goalieDefendDebug:
            print "lostBall < 5 case!"
        calculateDefendPositionHeading()
    
    # Wireless case.
    elif Global.sharedBallVar < HelpShort.get95cf(70):
        if Debug.goalieDefendDebug:
            print "Wireless ball case!"
        LocateBall.headFindBall(-50, 5)
        calculateSearchDefendPositionHeading()
        
    # GPS ball now. 
    elif (Global.lostBall <= Constant.VBALL_LOST
            and abs(Global.gpsLocalBall.getHeading()) < 90):
        if Debug.goalieDefendDebug:
            print "GPS ball case!"
        calculateSearchDefendPositionHeading()
        HelpTrack.trackGpsBall()
        
    # Default position.
    else:
        if Debug.goalieDefendDebug:
            print "I am confused, moving back to the original position!"
        LocateBall.headFindBall(-50, 5)
        calculateDefaultPositionHeading()
    
def calculateDefaultPositionHeading():
    global defendPos
    defendPos[0] = Constant.FIELD_WIDTH / 2.0
    defendPos[1] = Constant.WALL_THICKNESS
    defendPos[2] = 90
    
# Moves the robot to the defend position. Stop the robot if it is in position.
def moveToDefendPos():
    global defendPos
    
    diffX = defendPos[0] - Global.selfLoc.getPos()[0]
    diffY = defendPos[1] - Global.selfLoc.getPos()[1]
    diffH = defendPos[2] - Global.selfLoc.getHeading()
    
    if (abs(diffX) <= 10 and abs(diffY) <= 10 and abs(diffH) <= 15):
        return
    else:
        HelpTrack.saGoToTargetFacingHeading(defendPos[0], defendPos[1], defendPos[2])

# Return true if I want to kick the ball.    
def shouldIKickBall():
    selfH      = Global.selfLoc.getHeading()
    ifUPenKick = AttackerCruyff.BallUdnerChinForUpenn()
    
    # Don't kick if you are facing the own goal.
    if selfH >= 225 and selfH <= 315:
        return [False, None]
    
    # At this region only left UPenn would be allowed.
    elif selfH < 225 and selfH >= 135 and ifUPenKick:
        if Debug.goalieDebug:
            print "Only left UPenn region!"
        return [True, Constant.UpennLeftWT]
 
    # At this region only right UPenn would be allowed.
    elif (selfH > 315 or selfH <= 45) and ifUPenKick:
        if Debug.goalieDebug:
            print "Right UPenn success! Heading - %f" %selfH
        return [True, Constant.UpennRightWT]
 
    if not (selfH > 45 and selfH < 135):
        return [False, None]
 
    # At this point, the self heading should be: 135 < self-heading < 45.
    if not (selfH > 45 and selfH < 135):
        print selfH
        Debug.goAndMakeTraceBack()
        return
            
    # Now decided upon the enemy position.
    # Unsure region.
    if ifUPenKick:
       return [True, Constant.DiveKickWT]

    return [False, None]
     
def chooseKick(kick):
    if kick == Constant.UpennRightWT:
        Kick.UPright()
    elif kick == Constant.UpennLeftWT:
        Kick.UPleft()     
    elif kick == Constant.DiveKickWT:
        Kick.diveKick()
    else:
        Debug.goAndMakeTraceBack()
    
def isGoalieUPrightOk():
    if HelpShort.canSeeBall():
        ballx = math.sin (HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
        bally = math.cos (HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
                
        if (ballx < 3.5 and ballx > -5.5 and
            bally < 4.4 * Constant.BallRadius and
            Global.ballD < 18 and
            Global.degTilt < -9 and abs(Global.degPan) < 30):
            return True            
            
    return False
    
def isGoalieUPleftOk():
    if HelpShort.canSeeBall():
        ballx = math.sin (HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
        bally = math.cos (HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
        
        if (ballx < 3.0 and ballx > -5.0 and
            bally < 4.4 * Constant.BallRadius and
            Global.ballD < 18 and
            Global.degTilt < -9 and abs(Global.degPan) < 30):
            return True
            
    return False

def goalieBallIsGoingToArriveAtYouSoon():
    # Returns a tuple of (vel x[cm/frame], vel y[cm/frame], speed[cm/frame],
    # vel angle [degrees], vel angle - 90[degrees], normalized angle [doesn't work])
    vel = VisionLink.getGPSBallVInfo(Constant.CTLocal)
    
    # If heading away then ignore.
    if vel[1] >= 0:
        return False
    
    # Stop dividing zero.
    if vel[0] == 0:
        vel[0] = 0.01
        
    topEquation    = vel[1]
    bottomEquation = vel[0]
        
    m = topEquation / bottomEquation
    b = Global.ballY - Global.ballX * m
    x = -b / m
    
    if x < -30 or x > 30:
        return False

    if Global.ballD <= vel[2] * 75:
        return True
    else:
        return False
    
# Returns true if the goalie should attack the ball.
def shouldIAttackBall():
    ## Immediate danger cases.
    ##
    if not HelpShort.canSeeBall():
        if Debug.goalieDebug:
            print "No attack because can't see the ball!"
        return False

    if HelpLong.teammateHasBall():
        if Debug.goalieDebug:
            print "Not attacking because my teammate is doing it!"
        return False
        
    # If you have gone too far, don't attack.
    if not HelpWhere.isInOwnGoalBox(10):
        if Debug.goalieDebug:
            print "No attack because I am not in the goal box (too far)!"
        return False
        
    # Ignore the middle scrum because the above case should cover that indirectly.
    if isBallDistanceTooFarInversly():
        if Debug.goalieDebug:
            print "I am too far away, no attack!"
        return False
      
    if not HelpWhere.ballInOwnGoalBox(10):
        if Debug.goalieDebug:
            print "Ball not in the goal box!"
        return False
       
    # If opponent is much closer than me, don't do attack.
    _, oppDist = HelpLong.closestGpsOpponentWithinVar(60, Constant.CTLocal)
    oppInfo, _ = HelpLong.closestGpsOpponentWithinVar(60, Constant.CTGlobal)

    if oppInfo != None and oppDist != None:
        oppX = oppInfo.getX()
        oppY = oppInfo.getY()
        oppToBallDist = HelpShort.getDistanceBetween(oppX, oppY, Global.ballX, \
                        Global.ballY)
    
        # Come back later, is 25 a good distance? Maybe check the 2003 code.
        if oppToBallDist <= (Global.ballD - 25):
            if Debug.goalieDebug:
                print "Opponent is much closer than me for the ball, hence no attack!", oppToBallDist
            return False
        else:
            print "Ball distance: ", oppToBallDist

    if Debug.goalieDebug:
        print "Attacking the ball!"
        
    return True

def isBallDistanceTooFarInversly():
    goalieDistanceFromHome = HelpShort.getDistanceBetween(Global.selfLoc.getPos()[0], \
            Global.selfLoc.getPos()[1], Constant.FIELD_WIDTH / 2, 0)

    if not HelpShort.canSeeBall() or (HelpShort.canSeeBall() and \
        Global.ballD >= 75):
        return True
            
    requiredBallDist = 600.0 / goalieDistanceFromHome

    if Global.ballD >= requiredBallDist:
        if Debug.goalieDebug:
            print requiredBallDist
        return True
    else:
        return False

#-------------------------------------------------------------------------------
# Makes the dog approach the ball, and kick if it is very close.
def attackBall():
    global attackApproach

    # Goalie attack = red.
    Global.myRole = Constant.GOALIEATTACK
    
    if attackApproach != None:
        if attackApproach == PawKick.FIRE_PAWKICK_AUTO:
            PawKick.firePawKick(PawKick.FIRE_PAWKICK_AUTO)
        elif attackApproach == PawKick.FIRE_PAWKICK_LEFT:
            PawKick.firePawKick(PawKick.FIRE_PAWKICK_LEFT)
        elif attackApproach == PawKick.FIRE_PAWKICK_RIGHT:
            PawKick.firePawKick(PawKick.FIRE_PAWKICK_RIGHT)
        elif attackApproach == Constant.HOVER_TO_BALL:
            HoverToBall.DecideNextAction()
        else:
            print "The attacking approach: %f" %attackApproach             
            Debug.goAndMakeTraceBack()        
        return
        
    if Global.ballD <= 30 or abs(Global.ballH) >= 30:
        if Debug.goalieDebug:
            print "Choose hovertoBall!"
        attackApproach = Constant.HOVER_TO_BALL
        HoverToBall.DecideNextAction()
        
    elif Global.selfLoc.getHeading() < 45:
        if Debug.goalieDebug:
            print "Choose Right-PawKick!"
        attackApproach = PawKick.FIRE_PAWKICK_RIGHT
        PawKick.firePawKick(PawKick.FIRE_PAWKICK_RIGHT)
        
    elif Global.selfLoc.getHeading() > 135:
        if Debug.goalieDebug:
            print "Choose Left-PawKick!"
        attackApproach = PawKick.FIRE_PAWKICK_LEFT
        PawKick.firePawKick(PawKick.FIRE_PAWKICK_LEFT)
    else:
        if Debug.goalieDebug:
            print "Choose Auto-Pawkick!"
        attackApproach = PawKick.FIRE_PAWKICK_AUTO
        PawKick.firePawKick(PawKick.FIRE_PAWKICK_AUTO)
    
##~ #-------------------------------------------------------------------------------
##~ # Calculates the optimum field position for the goalie to defend against goals
##~ def calculateDefendPos():
##~     global ballLostTimer, defendPos
##~ 
##~     ballLostTimer = ballLostTimer + 1
##~ 
##~     # Use either gps or wireless ball position to calculate defend position.
##~     if VisionLink.getGPSBallMaxVar() < MAX_VARIANCE:
##~         ballLostTimer = 0
##~         defendPos = calculateIntercept(Global.gpsGlobalBall.getPos())
##~ 
##~     elif Global.sharedBallVar < MAX_VARIANCE:
##~         ballLostTimer = 0;
##~         defendPos = calculateIntercept((Global.sharedBallX, Global.sharedBallY))
##~ 
##~     # If neither gps or wireless are good enough, eventually move back to
##~     # central goal position.        
##~     elif ballLostTimer > BALL_LOST_LIMIT:
##~         defendPos = (145, 40, math.pi / 2)
##~ 
##~ #-------------------------------------------------------------------------------
##~ # Returns how far (in cm) the gps ball is outside our own goalbox
##~ def calculateBallBoxDist(ballPos):
##~     # Calculate the ball x & y distances from the goal mouth's center
##~     goalcentre = (Constant.FIELD_WIDTH / 2, 0)
##~     relBall = map(lambda b,g:abs(b-g), ballPos, goalcentre)
##~     
##~     # Now calculate x & y distances to closest point in goalbox
##~     relBall[0] = relBall[0] - Constant.GOALBOX_WIDTH / 2
##~     relBall[1] = relBall[1] - Constant.GOALBOX_DEPTH
##~ 
##~     if relBall[0] < 0:
##~         relBall[0] = 0
##~     if relBall[1] < 0:
##~         relBall[1] = 0
##~  
##~     # Return distance to closest point in goalbox
##~     return HelpShort.getLength(relBall)

# Send the data to other forwards.
def sendGoalieWirelessInfo():
    sendBallDistOrKickOff = Constant.LOSTBALL_DIST
    isStriker, isGoalie, haveBall, kickedOff, isDefending, isBackOff = \
                        [False, True, False, False, False, False]

    # Does goalie need to send haveBall signal?
                        
    # If I can see the ball, send the ball distance I calculated. If I can't
    # see the ball, do nothing as sendBallDist is initalised as LOSTBALL_DIST.
    if HelpShort.canSeeBall():
        sendBallDistOrKickOff = Global.ballD

    if Global.state == Constant.READYSTATE or Debug.mustReady:
        sendBallDistOrKickOff = Global.kickOffState

    VisionLink.sendMyBehaviourInfo(sendBallDistOrKickOff, isStriker, isGoalie, haveBall, \
                                   kickedOff, isDefending, isBackOff)

##############################################                                   
## Turning toward the soccer field.
##############################################                                   

# This is a direct port of the hacking made in 2003 goalie code. Although this was
# a hack, but it is very effective.
def shouldITurnTowardTheField():  
    global getOutOfGoalCounter, getOutOfGoalDirection

    ownGoal = Global.vOGoal
    turnCCW = -1
    
    # If we can see our own goal real big or we've got our face in either goal or we're locked
    # into a get-out-of-goal routine.
    if ((ownGoal.getConfidence() > 0 and (ownGoal.getWidth() >  \
        Constant.WIDTH_MOVE_THRESHOLD or (ownGoal.getHeight() > \
        Constant.HEIGHT_MOVE_THRESHOLD and ownGoal.getWidth() > \
        Constant.IMAGE_WIDTH / 2))) or getOutOfGoalCounter > 0):
        if getOutOfGoalDirection == None:
            if ownGoal.getHeading() >= 0:
                turnCCW = -30
            else:
                turnCCW = 30
            getOutOfGoalDirection = turnCCW
        else:
            turnCCW = getOutOfGoalDirection
                        
        HelpLong.setNormalWalk(0, 0, turnCCW)
        HelpLong.setHeadParams(0, 40, 0, Constant.HTAbs_h)
        getOutOfGoalCounter = getOutOfGoalCounter + 1
        
        MAX_INFRA_RED = 90.0
        irFar = VisionLink.getAnySensor(Constant.ssINFRARED_FAR)
        
        if getOutOfGoalCounter > 60 or Global.vTGoal.getConfidence() > 0 or \
            (irFar / 10000.0) >= MAX_INFRA_RED or Global.seenBeacon:
                getOutOfGoalCounter = 0
    
        return True
    else:
        return False

##############################################                                   
## Intercept calculating functions
##############################################

# Return a tuple of (x, y, h) for search defend case.
def calculateSearchDefendPositionHeading():
    global defendPos

    pivotX = Constant.FIELD_WIDTH / 2.0
    pivotY = Constant.WALL_THICKNESS
    pivotMajorAxis = 35.0
    pivotMinorAxis = 5.0
    
    ballX, ballY = [None, None]
    
    # Expect the source of the ball to be wireless or gps.
    if Global.ballSource == Constant.VISION_BALL and False:
        Debug.goAndMakeTraceBack()
        
    elif (Global.ballSource == Constant.WIRELESS_BALL or True):
        if Debug.goalieDefendDebug:
            print "In search(), the ball source is wireless!"
        ballX = Global.sharedBallX
        ballY = Global.sharedBallY
        
##~         print "------"
##~         print ballY
##~         print ballX
##~         print Global.selfLoc.getPos()
##~         print Global.selfLoc.getHeading()
##~         print "============="
        
        ballH = HelpShort.normalizeAngle_180(HelpShort.RAD2DEG(math.atan2 \
                (ballY - Global.selfLoc.getPos()[1], ballX - \
                 Global.selfLoc.getPos()[0])))
        
    else:
        ballX = Global.gpsGlobalBall.getPos()[0]
        ballY = Global.gpsGlobalBall.getPos()[1]
        ballH = HelpShort.normalizeAngle_180(HelpShort.RAD2DEG(math.atan2 \
                (ballY - Global.selfLoc.getPos()[1], ballX - \
                 Global.selfLoc.getPos()[0])))
        
    # Find intercept between line through ball and pivot point and ellipse
    # centred at pivot point and of the pivot axis.
    # Perform translation to make pivot point (0, 0)
    relX = ballX - pivotX
    relY = ballY - pivotY
    
    # Perform scaling to convert ellipse into unit circle.
    scaleX = relX / pivotMajorAxis
    scaleY = relY / pivotMinorAxis
    
    # Find intersection of line (0,0) -> (x,y) and unit circle.
    scaleLength = math.sqrt(HelpShort.SQUARE(scaleX) + HelpShort.SQUARE(scaleY))
    intX        = scaleX / scaleLength
    intY        = scaleY / scaleLength
    
    # Here we assume bally >= pivoty.
    if intY < 0:
        print scaleY, scaleLength, relY, pivotMinorAxis, ballY, pivotY
##~         Debug.goAndMakeTraceBack()
        intY = 0
    
    intH = HelpShort.RAD2DEG(math.atan2(intY, intX))
    tanAngle = None
    
    if intX > 0:
        tanAngle = 90 - intH
    else:
        tanAngle = intH - 90

    # Reverse scaling.
    unscaleX = intX * pivotMajorAxis
    unscaleY = intY * pivotMinorAxis
    opposite = math.atan(HelpShort.DEG2RAD(tanAngle))
    opposite = opposite * (pivotMinorAxis / pivotMajorAxis)
    unscaleTanAngle = HelpShort.RAD2DEG(math.atan(opposite))
    
    if intX > 0:
        defendPos[2] = 90 - unscaleTanAngle
    else:
        defendPos[2] = 90 + unscaleTanAngle
        
    defendPos[2] = ballH
    print defendPos[2]
        
    # Reverse translation.
    defendPos[0] = unscaleX + pivotX
    defendPos[1] = unscaleY + pivotY
        
# Return a tuple of (x, y, h). I simply don't really understand Derrick's new implementation.
# So revert back to the 2003 code.
def calculateDefendPositionHeading():
    pivotX = Constant.FIELD_WIDTH / 2.0
    pivotY = Constant.WALL_THICKNESS
    pivotR = 35 # Radius of the circle.
    ballX  = Global.gpsGlobalBall.getX()
    ballY  = Global.gpsGlobalBall.getY()
    
    # Find intercept between line through ball and pivot point
    # and circle around pivot point.
    #
    # Now been changed to weight toward the edges.    
    # Perform translation to make pivot point (0,0)
    relX = ballX - pivotX
    relY = ballY - pivotY
    
    if relX == 0 and relY == 0:
        relX = 0.01
        relY = 0.01
    
    # Determine angle compared to straight up the field.
    angle = HelpShort.normalizeAngle_180(90.0 - \
                        HelpShort.RAD2DEG(math.atan2(relY, relX)))
    weightLimit = 30.0
    
    if abs(angle) > weightLimit:
        sign = None
        if angle < 0:
            sign = -1
        else:
            sign = 1
            
        val = abs(angle) - weightLimit
        weightConjugate = 90.0 - weightLimit
        newval = weightConjugate * math.sqrt(val / weightConjugate)
        angle = sign * (newval + weightLimit)
        
    defendPos[0] = pivotX + pivotR * math.sin(HelpShort.DEG2RAD(angle))
    defendPos[1] = pivotY + pivotR * math.cos(HelpShort.DEG2RAD(angle))
    newLocX      = ballX - defendPos[0]
    newLocY      = ballY - defendPos[1]
    defendPos[2] = HelpShort.RAD2DEG(math.atan2(newLocY, newLocX))
    
##~ # Returns the intercept (gps coordinates) of a line from ball position to
##~ #   goal center, and an ellipse surrounding the goal
##~ def calculateIntercept(ballPos):
##~     # To calculate intercept, scale the relative ball position (from ellipse
##~     # center) so that the ellipse becomes a unit circle.
##~     relBall = map(lambda b,p,s:(b-p)/s, ballPos, DEFEND_PIVOT, DEFEND_SCALER)
##~     
##~     # Find the intersection with a unit circle.
##~     relBall = HelpShort.normalise(relBall)
##~     
##~     # Convert the intersection point back into normal global coordinates.
##~     pos = map(lambda r,s,p:r*s+p, relBall, DEFEND_SCALER, DEFEND_PIVOT)
##~ 
##~     # Calculate heading from intersection point to ball position. 
##~     relBall = map(lambda b,p:b-p, ballPos, pos)
##~     pos = pos + [math.pi / 2 - math.atan2(relBall[0], relBall[1])]
##~ 
##~     return tuple(pos) 
