#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: GoalieHead.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# GoalieHead.py
#
# This module controls the goal keeper's head movement.
#
#===============================================================================

import Debug
import Global 
import Constant
import HelpLong
import HelpShort
import HelpWhere
import HelpTrack
import LocateBall
import ActiveLocalise

#===============================================================================
# Global variables of this module.
#===============================================================================

ORIENT_TIME = 120           # How long goalie stationary localises for at start
BEACON_LOOK_TIME = 20       # Amount of time spent active localising
BEACON_LOOK_INTERVAL = 80   # Min. time between active localises
MIN_SELF_VARIANCE = 500     # Min. self variance needed to localise
MIN_BALL_DISTANCE = 30      # Min. ball distance needed to localise
GLANCE_TIME = 15            # Approx. time taken to move head to look at a point
HEAD_SEARCH_PERIOD = 70     # Number of frames for one head search cycle

headMode = None             # Can be either Start, Localise or BallTrack
localiseTimer = 0           # Determines when to switch modes
ballSearchTimer = 0         # Timer for ball search function
wideLocalise = False        # If goalie should localise off beacons outside
                            #   the predicted pan range
ballSeenCounter = 0

gbLocalisePt = 0            # which of the 3 directions to look down

#===============================================================================
# Functions belonging to this module.
#===============================================================================
# Sets the goalie's head angles
def DecideNextAction():
    global headMode

    if headMode == None:
        headMode = start

    # Execute current head movement code.
    headMode()

#-------------------------------------------------------------------------------
# Makes the goalie do a stationary localise
def start():
    global headMode, localiseTimer
    HelpTrack.stationaryLocalise()
    
    localiseTimer = localiseTimer + 1
    if localiseTimer > ORIENT_TIME:
        headMode = ballTrack
        localiseTimer = 0
        
#-------------------------------------------------------------------------------
# Makes the goalie look towards a beacon for a short amount of time
def localise():
    global headMode, localiseTimer
    ActiveLocalise.DecideNextAction()
    
    # If we have spent long enough localising, switch back to ball tracking
    localiseTimer = localiseTimer + 1
    if localiseTimer > BEACON_LOOK_TIME:
        headMode = ballTrack
        localiseTimer = 0
    
#-------------------------------------------------------------------------------
# Makes the goalie look directly at the ball.
def ballTrack():
    global headMode, localiseTimer, wideLocalise
    global ballSeenCounter, ballSearchTimer

    ballSeenCounter = ballSeenCounter + 1
    if not HelpShort.canSeeBall():
        ballSeenCounter = 0

    if ballSeenCounter < 2:
        ballSearchTimer = ballSearchTimer + 1
    else:
        ballSearchTimer = 0
        
    # If we can see the ball, track it normally, or quickly check last position.
    if HelpShort.canSeeBall() or ballSearchTimer < GLANCE_TIME:
        if Debug.goalieHeadDebug:
            print "Vision or last known ball position case!"
        HelpTrack.trackVisualBall()
    
    # If this doesn't work try GPS ball position.
    elif ballSearchTimer < (2 * GLANCE_TIME) and \
         abs(Global.gpsLocalBall.getHeading()) < 90:
        if Debug.goalieHeadDebug:
            print "GPS ball case by the goalie head!"
        HelpTrack.trackGpsBall()

    # Otherwise, do a full head search.
    else:
        ballSearch()

    # Occasionally localise
    if localiseTimer <= 500:
        localiseTimer = localiseTimer + 1
    
    if shouldLocalise():
        # WideLocalise is needed because the goalie gps sometimes thinks it's
        #   further forward than it really is, and hence won't localise off
        #   close beacons since (it thinks) they're outside head pan range.
        if wideLocalise:
            ActiveLocalise.SmartSetBeacon(150, 300)
            wideLocalise = False
        else:
            ActiveLocalise.SmartSetBeacon(95, 500)
            wideLocalise = True
        
        headMode = goalBoxLocalise
        # headMode = localise
        localiseTimer = 0

#-------------------------------------------------------------------------------
# Determines if the goalie needs to look for a beacon or not
def shouldLocalise():
    global localiseTimer

    # Localise at intervals, and only if unsure of self position.
    if (localiseTimer < BEACON_LOOK_INTERVAL):
        return False;

##~     # Localise at intervals, and only if unsure of self position.
##~     if (localiseTimer < BEACON_LOOK_INTERVAL 
##~         Global.selfLoc.getPosVar() < MIN_SELF_VARIANCE):
##~         return False;

    # We still need to localise while doing a ball search, since we probably
    # won't see the close beacons while searching.
    if not HelpShort.canSeeBall():
        return True;
    
    # Don't localise if the ball is closeby.
    return (Global.ballD > MIN_BALL_DISTANCE)

#-------------------------------------------------------------------------------
# Searches for the ball by scanning and panning
def ballSearch():
    global ballSearchTimer
    modCounter = ballSearchTimer % 40
    
    # A new ball search for the goalie only. Should be useful for edge detection.
    if (Global.ballD >= 70 and Global.selfLoc.getPosVar() <= HelpShort.
        get95cf(60) and HelpWhere.isInOwnGoalBox() and modCounter >
        20 and modCounter <= 40 and Global.ballSource != Constant.WIRELESS_BALL):
        if Debug.goalieHeadDebug:
            print "New ball search for the goalie!"
        LocateBall.headFindBall(-50, -5)
        return
    
    if ballSearchTimer > (2 * GLANCE_TIME + HEAD_SEARCH_PERIOD):
        ballSearchTimer = 2 * GLANCE_TIME
    
    if Debug.goalieHeadDebug:
        print "Head find ball for the goalie!"
    LocateBall.headFindBall(-50, -5)

# Look at the corners of the goal box to try and localise
def goalBoxLocalise():
    global headMode, localiseTimer, gbLocalisePt
    # If we have spent long enough localising, switch back to ball tracking for NEXT frame
    localiseTimer = localiseTimer + 1
    if localiseTimer > BEACON_LOOK_TIME:
        headMode = ballTrack
        localiseTimer = 0
        gbLocalisePt = (gbLocalisePt + 1) % 3
    
    if gbLocalisePt == 0:
        HelpLong.setHeadParams(0, 30, -30, Constant.HTAbs_h)
    elif gbLocalisePt == 1:
        HelpLong.setHeadParams(-60, 30, -30, Constant.HTAbs_h)
    elif gbLocalisePt == 2:
        HelpLong.setHeadParams(60, 30, -30, Constant.HTAbs_h)
    