#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: HandKick.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#===============================================================================

"""
Vladmir Kramnik.
"""

import Kick
import Debug
import FWHead
import Global
import HelpLong
import Constant
import HelpShort
import VisionLink
import OpenChallenger

# This player is written by state machine. Below are the states.
RUNNING_TO_THE_BALL        = 0
TURNING_WITH_THE_BALL      = 1
PREPARING_TURNING_THE_BALL = 2
DECIDE_KICK                = 3
HOLD_THE_BALL              = 4
HEAD_LOOK_UP               = 5
MUST_CHOOSE_LEFT_KICK      = 6
MUST_CHOOSE_RIGHT_KICK     = 7
DECIDE_IF_TURNING_TO_GAPS  = 8
TURNING_AGAIN              = 9
TURNING_AGAIN_HEAD_DOWN    = 10

# Dummy State. Don't know what this really mean.
ds = OpenChallenger.SuccessfulState()

# A wrapped grab ball implementation from OpenChallenger.
doRunToGrabBall = OpenChallenger.RunToGrabBall()

# My current state.
myState = RUNNING_TO_THE_BALL

# The grab ball timer.
grabBallTimer = 0

# The turning (spinning) timer.
spinTimer     = 0

turningDirection        = None
seeTargetGoalWhenHeadUp = False
HOLD_BALL               = (Constant.aaParaWalk,Constant.NormalWalkWT,0,0,0,5,90,120,0,0,110,-15,-55,15,Constant.HTAbs_h,0,0,0)

# The target shooting heading in local and global coordinatives.
shootHead       = None
shootHeadGlobal = None

# Should call this function to reset everything before calling doGrabBallHandKick().
def resetEverything():
    global myState, grabBallTimer, turningDirection, spinTimer, shootHead
    global RUNNING_TO_THE_BALL, seeTargetGoalWhenHeadUp, shootHeadGlobal

    myState       = RUNNING_TO_THE_BALL
    
    if Debug.handKickDebug:
        print "Before resetting, grabBallTimer = ", grabBallTimer
    
    grabBallTimer    = 0
    turningDirection = None
    spinTimer   = 0
    seeTargetGoalWhenHeadUp = False
    shootHead = None
    HelpLong.openMouth(0)

def runningToTheBall():
    global myState, doRunToGrabBall
    FWHead.resetEachFrame()
    FWHead.DecideNextAction()    
    grabSuccess = doRunToGrabBall.DecideNextAction()
    doRunToGrabBall.setOutState(ds)

    # Grab sucessfully?
    if grabSuccess == ds:
        myState = PREPARING_TURNING_THE_BALL
        return True
        
    # Grab failed?
    elif grabSuccess == None:
        resetEverything()
        return False
        
    return True
        
def prepareToGrabAndTurnTheBall():
    global myState, spinTimer, turningDirection, HOLD_BALL, HOLD_THE_BALL

    # Now decide my turning angle. The robot won't be able to decide until it pulls its head up.
    headingNeedToFaceTheGoal = HelpShort.getHeadingToFaceAt(Constant.FIELD_WIDTH / 2, \
                               Constant.FIELD_LENGTH)
                                   
    # Just fix the kicking in this case.
    if abs(headingNeedToFaceTheGoal) <= 10:
        print "Just choose left kick!"
        myState = MUST_CHOOSE_LEFT_KICK
    else:
        if headingNeedToFaceTheGoal > 0:
            turningDirection = Constant.dANTICLOCKWISE
        else:
            turningDirection = Constant.dCLOCKWISE
        
        myState   = HOLD_THE_BALL
        spinTimer = int(abs(headingNeedToFaceTheGoal) * 0.3) + 30 + 1
        Global.finalAction[Constant.AACmd:Constant.AAHeadType] = list(HOLD_BALL[Constant.AACmd:Constant.AAHeadType])
        HelpLong.setHeadParams(0, -70, 50, Constant.HTAbs_h)
        
    return True

def holdTheBall():
    global myState, grabBallTimer
    if grabBallTimer >= 10:
        myState = TURNING_WITH_THE_BALL
    Global.finalAction[Constant.AACmd:Constant.AAHeadType] = list(HOLD_BALL[Constant.AACmd:Constant.AAHeadType])
    HelpLong.setHeadParams(0, -70, 50, Constant.HTAbs_h)
    print "Hold the ball!"
    return True

def turnWithTheBall():
    global myState, grabBallTimer, spinTimer, HEAD_LOOK_UP, turningDirection
    if grabBallTimer >= spinTimer:
        myState = HEAD_LOOK_UP
        HelpLong.setHeadParams(0, -70, 50, Constant.HTAbs_h)        
    else:
        if turningDirection == Constant.dCLOCKWISE:
            Global.finalAction = list(Constant.TurnHoldingClockWise)
        else:
            Global.finalAction = list(Constant.TurnHoldingCounterClockWise)
        HelpLong.setHeadParams(0, -70, 50, Constant.HTAbs_h)        
    return True
        
def headLookUp():
    global myState, spinTimer, seeTargetGoalWhenHeadUp, shootHead, \
            shootHeadGlobal, grabBallTimer

    print "Head up!"
        
    # If see ball quit immediately.
    if HelpShort.canSeeBall():        
        print "See ball when head look up, drop!"
        resetEverything()
        return False

    Global.finalAction[Constant.AACmd:Constant.AAHeadType] = \
        list(Constant.BackWalkSitDownHoldingBall[Constant.AACmd:Constant.AAHeadType])
        
    if grabBallTimer <= spinTimer + 3:
        HelpLong.setHeadParams(0, -70, 50, Constant.HTAbs_h)        
        return True
    elif grabBallTimer <= spinTimer + 6:
        HelpLong.setHeadParams(0, 20, 10, Constant.HTAbs_pt)
        return True

    # Give this state a few counters for looking at the enemy goal. Janice.
    if grabBallTimer >= spinTimer + 15:
        myState = DECIDE_IF_TURNING_TO_GAPS

    if Global.vTGoal.getConfidence() > 0:
        seeTargetGoalWhenHeadUp = True
        tdkmin, tdkmax  = VisionLink.getHeadingToBestGap()
        shootHead       = (tdkmin + tdkmax) / 2.0
        myState         = DECIDE_IF_TURNING_TO_GAPS
        shootHeadGlobal = HelpShort.normalizeAngle_0_360(Global.selfLoc.getHeading() + shootHead)
        print "I can see the target goal Hacking! ShootHead = , shootHeadGlobal ", shootHead, shootHeadGlobal
        HelpLong.setHeadParams(0, -70, 50, Constant.HTAbs_h)
        return True
        
    else:
        print "Didn't see the goal when head up, look at it from gps!"
        headToFaceGoal = HelpShort.getHeadingToFaceAt(Constant.FIELD_WIDTH / 2.0, \
                         Constant.FIELD_LENGTH)
                             
        # If the heading make sense.
        if headToFaceGoal <= abs(90):
            HelpLong.setHeadParams(headToFaceGoal, 10, 10, Constant.HTAbs_pt)
        else:
            print "Heading doesn't make sense!"
            resetEverything()
            return False

    return True

def decideIfTurningToGaps():
    global myState, MUST_CHOOSE_RIGHT_KICK, MUST_CHOOSE_LEFT_KICK
    global TURNING_AGAIN_HEAD_DOWN

    # If see ball drop immediately.
    if HelpShort.canSeeBall():
        print "See ball when head decide gaps, drop!"
        resetEverything()        
        return False
        
    # If can't see the target goal.
    print "myState = DECIDE_IF_TURNING_TO_GPS!"
    
    if not seeTargetGoalWhenHeadUp:
        print "seeTargetGoalWhenHeadUp is False!"
        resetEverything()
        return False        
    else:

##~         Global.finalAction[Constant.AACmd:Constant.AAHeadType] = \
##~             list(Constant.BackWalkSitDownHoldingBall[Constant.AACmd:Constant.AAHeadType])
        HelpLong.setHeadParams(0, -70, 50, Constant.HTAbs_h)
        Global.finalAction[Constant.AACmd:Constant.AAHeadType] = list(HOLD_BALL[Constant.AACmd:Constant.AAHeadType])

        print "See the target goal in decide_if_turning!"
        ballGoHeadingIfLeftKick  = Global.selfLoc.getHeading() - Constant.HANDKICKANGLEOFFSET
        ballGoHeadingIfRightKick = Global.selfLoc.getHeading() + Constant.HANDKICKANGLEOFFSET

        if abs(ballGoHeadingIfLeftKick - shootHeadGlobal) <= 4:
            print "Choose left kick in DECIDE_IF_TURNING_TO_GAPS!"
            myState = MUST_CHOOSE_LEFT_KICK
        elif abs(ballGoHeadingIfRightKick - shootHeadGlobal) <= 4:
            print "Choose right kick in DECIDE_IF_TURNING_TO_GAPS!"
            myState = MUST_CHOOSE_RIGHT_KICK
        else:
            print "Keep turning!"
            myState = TURNING_AGAIN_HEAD_DOWN
        return True

def putHeadDownForNextTurn():
    global myState, grabBallTimer, spinTimer, TURNING_AGAIN
    
    Global.myRole = Constant.SUPPORT
    print "Head push again!"

    HelpLong.setHeadParams(0, -70, 50, Constant.HTAbs_h)

##~     Global.finalAction[Constant.AACmd:Constant.AAHeadType] = \
##~             list(Constant.BackWalkSitDownHoldingBall[Constant.AACmd:Constant.AAHeadType])
    
    Global.finalAction[Constant.AACmd:Constant.AAHeadType] = list(HOLD_BALL[Constant.AACmd:Constant.AAHeadType])

    HelpLong.openMouth(-650000)
    
    if grabBallTimer >= spinTimer + 30:    
        if VisionLink.getAnySensor(0) < -500000:
            print "Go from turning again head down to turning again!"
            myState = TURNING_AGAIN
            return True
        else:
            print "No ball under, quit!", VisionLink.getAnySensor(0)
            resetEverything()
            return False

    return True

# If this function returns True, then the skill is not finished. The skill is finished when False is
# returned.
def doGrabBallHandKick():
    global myState, grabBallTimer, doRunToGrabBall, turningDirection, RUNNING_TO_THE_BALL, TURNING_WITH_THE_BALL, turnAgainCounter, PREPARING_TURNING_THE_BALL, DECIDE_KICK, HOLD_THE_BALL, HEAD_LOOK_UP, turningCounter, seeTargetGoalWhenHeadUp
    global shootHead, MUST_CHOOSE_LEFT_KICK, MUST_CHOOSE_RIGHT_KICK, shootHeadGlobal
    global DECIDE_IF_TURNING_TO_GAPS, TURINING_AGAIN, TURNING_AGAIN_HEAD_DOWN

     # This is the counter starting from when the robot grabs the ball.
    if grabBallTimer >= 75:
         print "3 seconds limit!"
         # resetEverything()
         # return False
        
    # When running to the ball.
    if myState == RUNNING_TO_THE_BALL:
        return runningToTheBall()

    # The actual grabbing and turning is start here. This state takes 1 frame.
    elif myState == PREPARING_TURNING_THE_BALL:
        grabBallTimer = 0
        return prepareToGrabAndTurnTheBall()

    # Holding the ball. This state starts to increment the grabbing timer.
    elif myState == HOLD_THE_BALL:
        grabBallTimer += 1
        return holdTheBall()

    # Turning with the ball.
    elif myState == TURNING_WITH_THE_BALL:
        grabBallTimer += 1
        return turnWithTheBall()
    
    elif myState == HEAD_LOOK_UP:
        grabBallTimer += 1
        return headLookUp()
    
    elif myState == DECIDE_IF_TURNING_TO_GAPS:
        grabBallTimer += 1
        return decideIfTurningToGaps()
    
    elif myState == TURNING_AGAIN_HEAD_DOWN:
        grabBallTimer += 1
        return putHeadDownForNextTurn()

    elif myState == TURNING_AGAIN:
        grabBallTimer += 1
        return turningWithTheBallAgain()
    
    elif myState == MUST_CHOOSE_LEFT_KICK or myState == MUST_CHOOSE_RIGHT_KICK:
        if myState == MUST_CHOOSE_LEFT_KICK:
            Kick.handKickLeft()
        else:
            Kick.handKickRight()
        
        HelpLong.setForceStepComplete()
        resetEverything()
        return False

    else:
        Debug.goAndMakeTraceBack()
        return True

def turningWithTheBallAgain():
    global myState, shootHead, shootHeadGlobal

    Global.myRole = Constant.SUPPORT    
    print "Turning Again*********************************************"
    ballGoHeadingIfLeftKick  = Global.selfLoc.getHeading() - Constant.HANDKICKANGLEOFFSET
    ballGoHeadingIfRightKick = Global.selfLoc.getHeading() + Constant.HANDKICKANGLEOFFSET
            
    if abs(ballGoHeadingIfLeftKick - shootHeadGlobal) <= 4:    
        print "123 Choose left kick in DECIDE_IF_TURNING_TO_GAPS!"
        myState = MUST_CHOOSE_LEFT_KICK
        # Need this?

        HelpLong.setHeadParams(0, -70, 50,Constant.HTAbs_h)                    
        Global.finalAction[Constant.AACmd:Constant.AAHeadType] = list(HOLD_BALL[Constant.AACmd:Constant.AAHeadType])
##~         Global.finalAction[Constant.AACmd:Constant.AAHeadType] = \
##~             list(Constant.BackWalkSitDownHoldingBall[Constant.AACmd:Constant.AAHeadType])

        Kick.handKickLeft()        
        HelpLong.setForceStepComplete()
        resetEverything()
        return False

    elif abs(ballGoHeadingIfRightKick - shootHeadGlobal) <= 4:
        print "123 Choose right kick in DECIDE_IF_TURNING_TO_GAPS!"
        myState = MUST_CHOOSE_RIGHT_KICK
        
        # Need this?
        HelpLong.setHeadParams(0, -70, 50,Constant.HTAbs_h)                    
        Global.finalAction[Constant.AACmd:Constant.AAHeadType] = list(HOLD_BALL[Constant.AACmd:Constant.AAHeadType])        
##~         Global.finalAction[Constant.AACmd:Constant.AAHeadType] = \
##~             list(Constant.BackWalkSitDownHoldingBall[Constant.AACmd:Constant.AAHeadType])

        Kick.handKickRight()        
        HelpLong.setForceStepComplete()
        resetEverything()
        return False

    else:
        print "Keep turning hehe!", ballGoHeadingIfLeftKick, shootHeadGlobal
        if shootHead <= 0:
        
##~             Global.finalAction = list(Constant.BackWalkClockWise)
            Global.finalAction = list(Constant.TurnHoldingClockWise)
            
        else:
        
##~             Global.finalAction = list(Constant.BackWalkCounterClockWise)        
            Global.finalAction = list(Constant.TurnHoldingCounterClockWise)
            
        HelpLong.setHeadParams(0, -70, 50,Constant.HTAbs_h)            
    
    return True
    
def DecideNextAction():
    Global.frameReset()
    OpenChallenger.resetEachFrame()
    doGrabBallHandKick()
    HelpLong.superDebug()
