#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: HelpLong.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# HelpLong.py
#
# Different simple behaviours that might be used in any player codes. Compare to
# HelpShort.py, functions in this file is more complex.
#
#===============================================================================
#TODO:
#   setHeadParams for headType ABS_Z should use standard coordinates system
#

import math
import Debug
import Global
import Constant
import HelpShort
import HelpWhere
import HelpTrack
import VisionLink

# Used for frame debugging.
tail_down_count = 0

#===============================================================================
# Functions belong to this module.
#===============================================================================

#--------------------------------------
# Return numeric value.
# Get the maximum PWM values of current joints. When motor is jamed (something
# blocks the dog's way), it should be much higher.
def getMaxPWMDuty():
    maxVal = 0
    whichLeg = 0 #see Constant.legFRONT_RIGHT, ....
    for i in range(3,Constant.NUM_OF_EFFECTORS): #ignore head pwm
        thisJointVal = abs(VisionLink.getJointPWMDuty(i+1))

        if (maxVal < thisJointVal):
            maxVal = thisJointVal
            whichLeg = i / 3

    return (maxVal,whichLeg)
        
def getStuckInfo():
    return max(Global.lastMaxPWMValues) #return (maxPWM,whichLeg)
    
#================================================
# Helper to avoid typing in lots of names and avoid careless bugs
# To set finalAction
#================================================

def setHeadParams (panx,tilty,cranez,type):
    Global.finalAction[Constant.AAHeadType] = type
    Global.finalAction[Constant.AAPanx]     = panx
    Global.finalAction[Constant.AATilty]    = tilty
    Global.finalAction[Constant.AACranez]   = cranez

#if only set when the value is not None:
#for ex: if want to just set forward: setForwardLeftTurn(7,None,None)
def setForwardLeftTurn(forward,left,turnCCW):
    if forward is not None:
        Global.finalAction[Constant.AAForward] = forward
    if left is not None:
        Global.finalAction[Constant.AALeft]    = left
    if turnCCW is not None:
        Global.finalAction[Constant.AATurnCCW] = turnCCW
    
def setNormalWalk(forward, left, turnCCW):
    Global.finalAction[Constant.AAWalktype:Constant.AAHeadType] = list (Constant.NormalWalkFast[Constant.AAWalktype:Constant.AAHeadType])
    setForwardLeftTurn(forward,left,turnCCW)
    
def setSkellipticalWalk(forward,left,turnCCW):
    Global.finalAction[Constant.AAWalktype] = Constant.SkellipticalWalkWT
    setForwardLeftTurn(forward,left,turnCCW)
    
def setEllipticalWalk (forward, left, turnCCW):
    Global.finalAction[Constant.AAWalktype:Constant.AAHeadType] = list (Constant.EllipticalWalkFast[Constant.AAWalktype:Constant.AAHeadType])
    setForwardLeftTurn(forward,left,turnCCW)
    
def setNormalSideWalk (forward, left, turnCCW):
    Global.finalAction[Constant.AAWalktype:Constant.AAHeadType] = list (Constant.NormalWalkSideWayFast[Constant.AAWalktype:Constant.AAHeadType])
    setForwardLeftTurn(forward,left,turnCCW)
    
def setForceStepComplete():
    Global.finalAction[Constant.AACmd] = Constant.aaForceStepComplete

#--------------------------------------
# Set the default action parameters of the dog, stand still.
def setDefaultAction():
    Global.finalAction = list(Constant.NormalWalkStandStill)


#--------------------------------------
# Set the default action parameters of the dog, stand still.
def setDefaultIndicators():
    Global.finalIndicators = list(Constant.INDICATORS_DEFAULT)

#--------------------------------------
# Show the appropriate indicators.
def superDebug():
    showBallIndicator()
    showTeamIndicator()
    showRoleIndicator()
    showDropFrameIndicator()
    showHighLowGainIndicator()
    if Global.state == None or Global.state == Constant.READYSTATE or \
       Global.state == Constant.SETSTATE:
        showKickOffIndicator()
    else:
        showVisualTeammateIndicator()

#--------------------------------------
def showVisualTeammateIndicator():
    if canSeeTeammateWithinDist(Constant.LARGE_VAL):
        Global.finalIndicators[Constant.iBackMiddleColor] = Constant.IND_LED3_INTENSITY_MAX / Constant.INTENSITY_DIVIDE
    else:
        Global.finalIndicators[Constant.iBackMiddleColor] = Constant.IND_LED3_INTENSITY_MIN / Constant.INTENSITY_DIVIDE

#--------------------------------------
def showHighLowGainIndicator():
    pass
##~     if Global.pWalkInfo.isUsingHighGain():
##~         Global.finalIndicators[Constant.iHeadWhite] = Constant.IND_LED2_ON
##~     else:
##~         Global.finalIndicators[Constant.iHeadWhite] = Constant.IND_LED2_OFF

#--------------------------------------
# Show whether a ball is in the current frame.
def showBallIndicator():

##~     if HelpShort.canSeeBall():
##~         Global.finalIndicators[Constant.iHeadColor] = Constant.IND_LED2_ON
##~     else:
##~         Global.finalIndicators[Constant.iHeadColor] = Constant.IND_LED2_OFF


    if HelpShort.canSeeBall():
        Global.finalIndicators[Constant.iHeadWhite] = Constant.IND_LED2_ON
    else:
        Global.finalIndicators[Constant.iHeadWhite] = Constant.IND_LED2_OFF

#--------------------------------------
# Show which team I am in.
def showTeamIndicator():
    if Global.teamColor == Constant.BLUE_TEAM:
        changeTeamIndicator(True, False)
    else:
        changeTeamIndicator(False, True)

def changeTeamIndicator(isBlueTeam, isRedTeam):
    if (isBlueTeam and isRedTeam) or (not isBlueTeam and not isRedTeam):
        Debug.goAndMakeTraceBack()
        return

    if isRedTeam:
        Global.finalIndicators[Constant.iBackRearColor]  = Constant.IND_LED3_INTENSITY_MIN / Constant.INTENSITY_DIVIDE
        Global.finalIndicators[Constant.iBackRearWhite]  = Constant.IND_LED3_INTENSITY_MIN / Constant.INTENSITY_DIVIDE
        Global.finalIndicators[Constant.iBackFrontColor] = Constant.IND_LED3_INTENSITY_MAX / Constant.INTENSITY_DIVIDE
        Global.finalIndicators[Constant.iBackFrontWhite] = Constant.IND_LED3_INTENSITY_MIN / Constant.INTENSITY_DIVIDE
    else:
        Global.finalIndicators[Constant.iBackRearColor]  = Constant.IND_LED3_INTENSITY_MAX / Constant.INTENSITY_DIVIDE
        Global.finalIndicators[Constant.iBackRearWhite]  = Constant.IND_LED3_INTENSITY_MIN / Constant.INTENSITY_DIVIDE
        Global.finalIndicators[Constant.iBackFrontColor] = Constant.IND_LED3_INTENSITY_MIN / Constant.INTENSITY_DIVIDE
        Global.finalIndicators[Constant.iBackFrontWhite] = Constant.IND_LED3_INTENSITY_MIN / Constant.INTENSITY_DIVIDE
    
#--------------------------------------
# Show the red "eyes" on face. If this function is not called (and with on=True),
# no red eyes on face.
def showRedEye(on = True):

    if (on):
        Global.finalIndicators[Constant.iLed14] =  Constant.IND_LED3_RED_EYE_VAL
    else :
        Global.finalIndicators[Constant.iLed14] =  Constant.IND_LED3_INTENSITY_MIN


#--------------------------------------
# Show pattern in the face.
def showFacePattern(pattern):

    # Input is a tuple of 5 ints. Each represent a coloumn of lights in face.
    # 0 means no light, 3 means 3 blocks of lights.
    # For the 1st and 5th column there're 3 blocks, for the middle 3 there can 
    # only be 2 blocks.
    
    # The first column.
    if (pattern[0] == Constant.IND_LED_OFF):
        Global.finalIndicators[Constant.iLed1] = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed3] = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed7] = Constant.IND_LED3_INTENSITY_MIN
    elif (pattern[0] == 1):
        Global.finalIndicators[Constant.iLed1] = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed3] = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed7] = Constant.IND_LED3_INTENSITY_MAX
    elif (pattern[0] == 2):
        Global.finalIndicators[Constant.iLed1] = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed3] = Constant.IND_LED3_INTENSITY_MAX
        Global.finalIndicators[Constant.iLed7] = Constant.IND_LED3_INTENSITY_MAX
    elif (pattern[0] == 3):
        Global.finalIndicators[Constant.iLed1] = Constant.IND_LED3_INTENSITY_MAX
        Global.finalIndicators[Constant.iLed3] = Constant.IND_LED3_INTENSITY_MAX
        Global.finalIndicators[Constant.iLed7] = Constant.IND_LED3_INTENSITY_MAX

    # The second column.
    if (pattern[1] == Constant.IND_LED_OFF):
        Global.finalIndicators[Constant.iLed5] = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed9] = Constant.IND_LED3_INTENSITY_MIN
    elif (pattern[1] == 1):
        Global.finalIndicators[Constant.iLed5] = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed9] = Constant.IND_LED3_INTENSITY_MAX
    elif (pattern[1] == 2):
        Global.finalIndicators[Constant.iLed5] = Constant.IND_LED3_INTENSITY_MAX
        Global.finalIndicators[Constant.iLed9] = Constant.IND_LED3_INTENSITY_MAX

    # The third column.
    if (pattern[2] == Constant.IND_LED_OFF):
        Global.finalIndicators[Constant.iLed11] = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed12] = Constant.IND_LED3_INTENSITY_MIN
    elif (pattern[2] == 1):
        Global.finalIndicators[Constant.iLed11] = Constant.IND_LED3_INTENSITY_MAX
        Global.finalIndicators[Constant.iLed12] = Constant.IND_LED3_INTENSITY_MIN
    elif (pattern[2] == 2):
        Global.finalIndicators[Constant.iLed11] = Constant.IND_LED3_INTENSITY_MAX
        Global.finalIndicators[Constant.iLed12] = Constant.IND_LED3_INTENSITY_MAX

    # The forth column.
    if (pattern[3] == Constant.IND_LED_OFF):
        Global.finalIndicators[Constant.iLed6]  = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed10] = Constant.IND_LED3_INTENSITY_MIN
    elif (pattern[3] == 1):
        Global.finalIndicators[Constant.iLed6]  = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed10] = Constant.IND_LED3_INTENSITY_MAX
    elif (pattern[3] == 2):
        Global.finalIndicators[Constant.iLed6]  = Constant.IND_LED3_INTENSITY_MAX
        Global.finalIndicators[Constant.iLed10] = Constant.IND_LED3_INTENSITY_MAX

    # The fifth column.
    if (pattern[4] == Constant.IND_LED_OFF):
        Global.finalIndicators[Constant.iLed2] = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed4] = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed8] = Constant.IND_LED3_INTENSITY_MIN
    elif (pattern[4] == 1):
        Global.finalIndicators[Constant.iLed2] = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed4] = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed8] = Constant.IND_LED3_INTENSITY_MAX
    elif (pattern[4] == 2):
        Global.finalIndicators[Constant.iLed2] = Constant.IND_LED3_INTENSITY_MIN
        Global.finalIndicators[Constant.iLed4] = Constant.IND_LED3_INTENSITY_MAX
        Global.finalIndicators[Constant.iLed8] = Constant.IND_LED3_INTENSITY_MAX
    elif (pattern[4] == 3):
        Global.finalIndicators[Constant.iLed2] = Constant.IND_LED3_INTENSITY_MAX
        Global.finalIndicators[Constant.iLed4] = Constant.IND_LED3_INTENSITY_MAX
        Global.finalIndicators[Constant.iLed8] = Constant.IND_LED3_INTENSITY_MAX
    
    
#--------------------------------------
# Make up the color in the head.
# Input color is a tuple (bool, bool, bool) where each boolean represents if the 
# respective color should be on or not in the head.
def showHeadColor(color):

    red   = Constant.IND_LED2_OFF
    green = Constant.IND_LED2_OFF
    blue  = Constant.IND_LED2_OFF

    if (color[0]):
        red     = Constant.IND_LED2_ON

    if (color[1]):
        green   = Constant.IND_LED2_ON

    if (color[2]):
        blue    = Constant.IND_LED2_ON

    Global.finalIndicators[Constant.iModeRed]   =  red
    Global.finalIndicators[Constant.iModeGreen] =  green
    Global.finalIndicators[Constant.iModeBlue]  =  blue

#--------------------------------------
# Show which role this player is in.
def showRoleIndicator():

    if Global.myRole == Constant.SUPPORT or \
        Global.myRole == Constant.SUPPORTBACKOFF or \
        Global.myRole == Constant.GOALIEDEFENDSEEB:
        if Debug.roleIndicator:
            print "Yellow case!"    
        showHeadColor(Constant.RGB_YELLOW)

    elif Global.myRole == Constant.STEALTHDOG:
        showHeadColor(Constant.RGB_LIGHT_BLUE)

    elif Global.myRole == Constant.ATTACKER or \
         Global.myRole == Constant.ATTACKBACKOFF or \
         Global.myRole == Constant.GOALIEATTACK:
        if Debug.roleIndicator:
            print "Red case!"
##~         showHeadColor(Constant.RGB_RED)
        showHeadColor(Constant.RGB_NO_COLOR)

    elif Global.myRole == Constant.STRIKER:
        showHeadColor(Constant.RGB_BLUE)

    elif Global.myRole == Constant.DEFENDER or \
         Global.myRole == Constant.STUCKER or \
         Global.myRole == Constant.GOALIEDEFENDNOTS:
        showHeadColor(Constant.RGB_GREEN) 

    elif Global.myRole == Constant.GETBEHINDBACKOFF:
        showHeadColor(Constant.RGB_WHITE)

    elif Global.myRole == Constant.SIDEBACKOFF or \
         Global.myRole == Constant.DEFENDERGOALBOX:
        showHeadColor(Constant.RGB_PURPLE)

    else:
        showHeadColor(Constant.RGB_NO_COLOR)
    
#--------------------------------------
# Show whether drop frames occur due to computation-too-heavy.
def showDropFrameIndicator():

    # Frames come at a rate of 30 frames/sec.
    # If cpu can't finish processing the frame in 1/30 sec, frame drops
    # occur.
    
    # Max time before the frame become dropped.
    # CPU_OVER = (1.0 / 30.0) * 1000000.0
    # CPU_80          = CPU_OVER * 0.8
    CAMERA_FRAME_MS = 1000.0 / 30.0
    JOINT_FRAME_MS  = 8     
    FRAME_SKIP      = (CAMERA_FRAME_MS * 1.2 / JOINT_FRAME_MS) + 2

    # Each frame-drop occur will at least incur how many frames of tail down.
    # Ted: Should the tail_down_duration be 30 rather than 25?
    TAIL_DOWN_DURATION = 30;

    # Tail indicator for CPU time.
    global tail_down_count
    
    # OPENR frame ID.
    # cpuUsage = VisionLink.getCPUUsage()
    currentFrameID, lastFrameID = VisionLink.getCurrentPreviousFrameID()
    
    if tail_down_count <= 0:
        if currentFrameID - lastFrameID > FRAME_SKIP and lastFrameID > 0:
            Global.finalIndicators[Constant.iTailV] = Constant.IND_TAIL_DOWN
            tail_down_count = TAIL_DOWN_DURATION            
    else: 
        Global.finalIndicators[Constant.iTailV] = Constant.IND_TAIL_DOWN
        tail_down_count = tail_down_count - 1
    
#--------------------------------------
# Return number of robots alive including yourself.
def robotAlive():
    # I am alive otherwise this function won't be called.
    alive = 1
    
    for i in range(Constant.NUM_TEAM_MEMBER):
        if Global.teamPlayers[i].getCounter() > 0 and i != \
            Global.myPlayerNum - 1:            
            alive = alive + 1
            
    return alive

#--------------------------------------
def showKickOffIndicator():
    if VisionLink.getKickOffState() == Constant.ROBOCUP_KICKOFF_OWN:
        Global.finalIndicators[Constant.iBackMiddleColor] =\
            Constant.IND_LED3_INTENSITY_MAX / Constant.INTENSITY_DIVIDE
    else:
        Global.finalIndicators[Constant.iBackMiddleColor] = \
            Constant.IND_LED3_INTENSITY_MIN / Constant.INTENSITY_DIVIDE

#--------------------------------------
# Returns true iff 'have ball' signal received from another teammate.
def teammateHasBall():
    for i in range(Constant.NUM_TEAM_MEMBER):
        if (Global.teamPlayers[i].getCounter() > 0 and 
            i != Global.myPlayerNum - 1 and
            Global.teamPlayers[i].hasBall()):
            return True
    return False

#--------------------------------------
def canSeeTeammateWithinDist(dist):
    for i in range(Constant.NUM_TEAM_MEMBER):
        # Should this be changed to check variance?
        if (Global.vTeammates[i].getConfidence() > 0 
                and Global.vTeammates[i].getDistance() < dist):
            return True
    return False

#--------------------------------------    
def canSeeOpponentWithinDist(dist):
    for i in range(Constant.NUM_TEAM_MEMBER):
        if (Global.vOpponents[i].getConfidence() > 0
                and Global.vOpponents[i].getDistance() < dist):
            return True
    return False

#-----------------------------
# return a list of their player num (0 to 3)
def whichGpsOpponentWithinDistVar (dist, var, coordinate = Constant.CTLocal):
    res = []

    for i in range (Constant.NUM_TEAM_MEMBER):
        gpsOppVar = VisionLink.getGPSOpponentCovMax(i)
        gpsOppPos = Global.Vector(*VisionLink.getGPSOpponentInfo(i, coordinate))
        
        if gpsOppVar < HelpShort.get95cf(var) and gpsOppPos.getDistance() < dist:
            
            res.append (gpsOppPos)
        
    return res

def closestGpsOpponentWithinVar (var, coordinate = Constant.CTLocal):
    closestOpp = None
    minDist    = Constant.LARGE_VAL
    
    for i in range (Constant.NUM_TEAM_MEMBER):
        gpsOppVar = VisionLink.getGPSOpponentCovMax(i)        
        gpsOppPos = Global.Vector(*VisionLink.getGPSOpponentInfo(i, coordinate))
        
        if gpsOppVar < HelpShort.get95cf(var) and gpsOppPos.getDistance() < minDist:            
            closestOpp = gpsOppPos
            minDist    = gpsOppPos.getDistance()
        
    return [closestOpp, minDist]
    
def closestToBallGpsOpponentWithinVar (var):
    closestOpp = None
    minDist    = Constant.LARGE_VAL
    
    for i in range (Constant.NUM_TEAM_MEMBER):
        gpsOppVar = VisionLink.getGPSOpponentCovMax(i)
        if gpsOppVar < HelpShort.get95cf(var):
            gpsOppPos = Global.Vector(*VisionLink.getGPSOpponentInfo(i, Constant.CTGlobal))
            oppDistToBall = HelpShort.getDistanceBetween(gpsOppPos.getX(), gpsOppPos.getY(), \
                        Global.ballX, Global.ballY)
            if oppDistToBall < minDist:            
                closestOpp = gpsOppPos
                minDist    = oppDistToBall
        
    return [closestOpp, minDist]
    
def closestVisualRobot (TeammateRobot = True):

    closeRobot = None
    minDist    = Constant.LARGE_VAL
    
    if TeammateRobot:
        for i in range (Constant.NUM_TEAM_MEMBER):
            tm = Global.vTeammates[i]
            if (tm.getConfidence() > 0 and
                tm.getVobDistance() < minDist):
                closeRobot = tm
                
    else:
        for i in range (Constant.NUM_TEAM_MEMBER):
            opp = Global.vOpponents[i]
            if (opp.getConfidence() > 0 and 
                opp.getVobDistance() < minDist):
                closeRobot = opp
        
    if closeRobot is None:
        return None
    else:
        return closeRobot
    
#--------------------------------------
def forwardAlive():
    alive = 1
    
    for i in range(Constant.NUM_TEAM_MEMBER):
        if (Global.teamPlayers[i].getCounter() > 0
            and not Global.teamPlayers[i].isGoalie()
            and i != Global.myPlayerNum-1):
            alive = alive + 1
    
    return alive

#--------------------------------------
# Send wireless info to teammates. The 2003 team didn't group the sending together, and
# the code became so messy.
#
# However by grouping them, the assigning may have some problems. But two or more
# could be set at the same time. Come back later.
def sendWirelessInfo():
    sendBallDistOrKickOff = Constant.LOSTBALL_DIST
    isStriker, isGoalie, haveBall, kickedOff, isDefending, isBackOff = \
                        [False, False, False, False, False, False]

    # If I can see the ball, send the ball distance I calculated. I I can't
    # see the ball, do nothing as sendBallDist is initalised as LOSTBALL_DIST.
    if VisionLink.getBallConfidence() > 0:
        sendBallDistOrKickOff = Global.ballD
    
    # Should I tell my teammates that I am a striker?
    if Global.myRole == Constant.STRIKER:
        isStriker = True
        
    # Should I tell my teammates that I am a goalie?
    isGoalie = False
    
    # Should I tell my teammates that I am kicking off? Similarly for the goalie.
    kickedOff = False
    
    # If in ready state, then send the kickoff signal.
    # Kim and Daniel both agreed, so at the moment ballDist in the readyState is
    # really not a ball distance, it is the kickoff signal. Reason is the kickOff signal
    # in interpBehav is a boolean with two values, but kickoff signal need 5
    # values. However any other state (eg: set), the ballDist is really the ball
    # distance. Notice that in the readyState robots don't need to send
    # ball distance to other robots as they are not even tracking the ball.
    if Global.state == Constant.READYSTATE or Debug.mustReady:
        sendBallDistOrKickOff = Global.kickOffState
    
    # Should I tell my teammates that I am defending?
    if Global.myRole == Constant.DEFENDER:
        isDefending = True
    
    # Should I tell my teammates that I am backing-off?
    # Need to check this carefully from the C++ code.
    # Ted: I think in 2003, this signal is only called when getBehindBallBackOff
    #          or supportBackOff. You can grep setBeckhamBackingOff(). Not called
    #          when attackBackOff.
    if Global.myRole == Constant.GETBEHINDBACKOFF or \
       Global.myRole == Constant.SUPPORTBACKOFF:
        isBackOff = True
        
    # Should I tell my teammates that I have the ball?
    if HelpTrack.doIHaveTheBall() and not isStriker and not isBackOff and \
        Global.myRole == Constant.ATTACKER:
        haveBall = True

    sendMyBehaviourInfo(sendBallDistOrKickOff, isStriker, isGoalie, haveBall, \
        kickedOff, isDefending, isBackOff)

def sendMyBehaviourInfo(sendBallDist=0, isStriker=False, isGoalie=False, haveBall=False,\
        kickedOff = False, isDefending = False, isBackOff = False):
    VisionLink.sendMyBehaviourInfo(sendBallDist, isStriker, isGoalie, haveBall, \
                                   kickedOff, isDefending, isBackOff)

def showBatteryStatus():
    current = VisionLink.getRemainingCapacity()
    
    if current >= 100:
        showFacePattern([1,1,1,1,1])
    elif current >= 80:
        showFacePattern([0,1,1,1,1])
    elif current >= 60:
        showFacePattern([0,1,1,1,0])
    elif current >= 40:
        showFacePattern([0,0,1,1,0])
    elif current >= 20:
        showFacePattern([0,0,1,0,0])
    else:
        showFacePattern([0,0,0,0,0])

def openMouth(angle = Constant.IND_MOUTH_OPEN):
    Global.finalIndicators[Constant.iMouth] = angle

#return number of milliseconds of the current time
def getCurrentSimpleTime():
    return HelpShort.getSimpleTime(VisionLink.getCurrentTime())
