#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: HelpShort.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# HelpShort.py
#
# This file store helper functions for the behaviour codes to use.
#
#===============================================================================

##################################################
##  Helper functions 
##
import math # built-in python maths library
import Global
import Constant
import VisionLink

#===============================================================================
# Functions belong to this module.
#===============================================================================
def resetMovement():
    Global.finalAction[Constant.AAForward] = 0
    Global.finalAction[Constant.AATurnCCW] = 0
    Global.finalAction[Constant.AALeft]    = 0

def resetEverything():
    Global.finalAction[Constant.AAHeadType] = Constant.HTAbs_xyz
    Global.finalAction[Constant.AAPanx]     = 10
    Global.finalAction[Constant.AATilty]    = 10
    Global.finalAction[Constant.AACranez]   = 10
    Global.finalAction[Constant.AAForward]  = 0
    Global.finalAction[Constant.AATurnCCW]  = 0
    Global.finalAction[Constant.AALeft]     = 0

def get95cf(stddev):
    return SQUARE(stddev) / 4.0

def get95sd(var):
    return math.sqrt(var * 4.0)

def DECREMENT(x):
    if x > 0:
        x -= 1
    else:
        x = 0
    return x
    
#---------------------------------------
# Convert micro-"something" to "something"
def MICRO2NORMAL(x):
    return (x / 1000000)

#---------------------------------------
# Convert "something" to micro-"something"
def NORMAL2MICRO(x):
    return (x * 1000000)

#---------------------------------------
# Convert coordinates from top-left corner oriented to center oriented.
def TOPLEFT2CENTER_X(x):
    return ((x) - 104)   # 104 = CPLANE_WIDTH/2 

def TOPLEFT2CENTER_Y(y):
    return (80 - (y))    # 80 = CPLANE_HEIGHT/2


#---------------------------------------
# Return True if visual ball is observed in cplane.
def canSeeBall():
    return VisionLink.getBallConfidence() > 0

#---------------------------------------
# Return angle satisfying: 0 <= angle < 360, given degree.
def normalizeAngle_0_360(angle):
    while angle < 0:
        angle += 360
    while angle >= 360:
        angle -= 360
    return angle

#---------------------------------------
# Return angle satisfying: -180 < angle <= 180, given degree.
def normalizeAngle_180(angle):
    while angle <= -180:
        angle += 360
    while angle > 180:
        angle -= 360
    return angle

#---------------------------------------
# Return angle satisfying: -90 < angle <= 270, given degree.
def normalizeAngle_90(angle):
    while angle <= -90:
        angle += 360
    while angle > 270:
        angle -= 360
    return angle

#---------------------------------------
# Returns the cartesian length of a vector of any dimension
def getLength(vec):
    length = 0
    for dimension in vec:
        length = length + dimension * dimension
    return math.sqrt(length)

#---------------------------------------
# Scales the given n-vector to unit length
def normalise(vec):
    length = getLength(vec)
    if (length > 0):
        return map(lambda x:x/length, vec)
    else:
        return vec

#---------------------------------------
# Rotates a 2D vector counter-clockwise by the given angle in degrees
def rotate(vec, angle, rad = False):
    if rad:
        sinA = math.sin(angle)
        cosA = math.cos(angle)
    else:
        sinA = math.sin(DEG2RAD(angle))
        cosA = math.cos(DEG2RAD(angle))
    x = cosA * vec[0] - sinA * vec[1]
    y = sinA * vec[0] + cosA * vec[1]
    return (x, y)

#---------------------------------------
def getHeadingToFaceAt(objX, objY):
    myX = Global.selfLoc.getX()
    myY = Global.selfLoc.getY()
    myH = Global.selfLoc.getHeading()
    return getHeadingToMe(myX,myY,myH, objX,objY)

#---------------------------------------
def getHeadingToBall(x,y):
    """Convert coordinate (x,y) relatively to the robot
    to heading (-pi/2 : left , pi/2 : right )
    """
    if x == 0: return 0
    if x > 0: return math.atan( y / x) - math.pi / 2 
    else: return math.atan( y / x) + math.pi / 2 

#---------------------------------------
def getDistanceBetween(fromX, fromY, toX, toY):
    relx = toX - fromX
    rely = toY - fromY
    return math.sqrt((relx * relx) + (rely * rely)) 
    
#---------------------------------------
def getHeadingBetween(fromX, fromY, toX, toY):
    relx = toX - fromX
    rely = toY - fromY
    if relx == 0 and rely == 0:
        return 0
    return RAD2DEG(math.atan2(rely, relx))

#---------------------------------------
def getPointRelative(baseX, baseY, offsetHead, offsetDist):
    rx = baseX + offsetDist * math.cos(DEG2RAD(offsetHead))
    ry = baseY + offsetDist * math.sin(DEG2RAD(offsetHead))
    return (rx,ry)
    
#---------------------------------------
def RAD2DEG(x):
    return x * 180.0 / math.pi 
    
#---------------------------------------
def DEG2RAD(x):
    return x * math.pi / 180.0
    
#---------------------------------------
def CLIP(x, LIM):
    if x > LIM : return LIM
    elif x < -LIM : return -LIM
    else: return x

#---------------------------------------
def CLIPTO(x, min, max):
    if x > max : return max
    elif x < min : return min
    else: return x

#opposed to CLIP(), make sure abs(x) >= LIM
def extendToLimit(x,LIM):
    if x >= 0 and x < LIM : return LIM
    elif x < 0 and x > -LIM : return -LIM
    else: return x
    
    
#clip a coordinate to a point in the field    
def clipPosition((x,y)):
    return ( CLIPTO(x, 0,Constant.FIELD_WIDTH) , CLIPTO(y, 0,Constant.FIELD_LENGTH) )
    
    
#---------------------------------------
# Return Cartesian coordinates (x,y) given polar coordinates (dist, degree).
# (Polar coords: right = 0, front = 90, so on)
def convertToCartFromPolar(r,alpha):
    t = DEG2RAD(alpha)
    return (r * math.sin(t) , r * math.cos(t) )

#---------------------------------------
# Return polar coordinates (dist, degree) given Cartesian coordinates (x,y) relative to origin.
# Result will be >= 0 and < 360.
def convertToPolarFromCart(dx,dy):
    return normalizeAngle_0_360(RAD2DEG(math.atan2(dy, dx)))

#---------------------------------------
def array2D(nrow,ncol):
    l = []
    for _ in range(nrow):
        sl = []
        for _ in range(ncol):
            sl.append(0)
        l.append(sl)
    return l

#---------------------------------------
def SQUARE(x):
    return x * x
    
#---------------------------------------
# Given global coordinates x y, calculate my distance and local heading to it.
# Local heading: How much turning will make me face at it, +ive left, -ive right.
def findAngleDist(x, y):
    dx = x - Global.selfLoc.getX()
    dy = y - Global.selfLoc.getY()
    dist = math.sqrt((dx * dx) + (dy * dy))
    
    head = normalizeAngle_180(RAD2DEG(math.atan2(dy, dx)) - Global.selfLoc.getHeading())

    return [dist, head]
    
#---------------------------------------
# Return the abs angle between 2 points, taking this pivot point as the origin
# eg of use, a is robot ,b is goal, pivot is ball
# ie. the angle between the robot and the goal via the ball ie. angle a-pivot-b.
def absAngleBetweenTwoPointsFromPivotPoint(pointax, pointay, pointbx, pointby, pivotx, pivoty):
    asquared = SQUARE(pointbx - pointax) + SQUARE(pointby - pointay)
    bsquared = SQUARE(pointbx - pivotx)  + SQUARE(pointby - pivoty)
    csquared = SQUARE(pivotx - pointax)  + SQUARE(pivoty - pointay)

    squaredSum  = bsquared + csquared - asquared
    twobc       = 2 * math.sqrt(bsquared) * math.sqrt(csquared)
    
    if abs(twobc) <= 0.0001:
        twobc += 0.01

    result = abs(RAD2DEG(math.acos(squaredSum / twobc)))
    return result

#angle a-pivot-b.
#TODO: make the above use this one for coherence
def angleBetweenTwoPointsFromPivotPoint(pointax, pointay, pointbx, pointby, pivotx, pivoty):
    h1 = getHeadingBetween(pivotx,pivoty,pointax,pointay)
    h2 = getHeadingBetween(pivotx,pivoty,pointbx,pointby)
    return normalizeAngle_180(h2-h1)
    
# return a xor b , since python doesn't have one????
def xor(a,b):
    return ( a and not b ) or (b and not a)

#return the number of milli second from within an interval of 1000 seconds
# sothat range = 0...10e6
def getSimpleTime((seconds,microSeconds)):
    return (seconds % 1000 ) * 1000 + microSeconds / 1000
    
#get the two intersect point from two circles
def intersectCircle(x1,y1,r1,x2,y2,r2):
    d = getDistanceBetween(x1,y1,x2,y2)
    res = []
    if d > r1+r2: return res
    
    semi_p = (d + r1 + r2)/2.0
    area = math.sqrt(semi_p * (semi_p - d) * (semi_p - r1) * (semi_p - r2))
    h = 2*area / d;
    alpha = math.asin(h/r1)
    if r2 > d and r2 > r1: # alpha > 90
        alpha = math.pi - alpha
    
    heading = getHeadingBetween(x1,y1,x2,y2)
    (x,y) = getPointRelative( x1,y1, heading, r1)
    dx,dy = x-x1,y-y1
    dx1,dy1 = rotate( (dx,dy) , alpha, rad=True)
    dx2,dy2 = rotate( (dx,dy) , -alpha, rad=True)
    res.append( (x1 + dx1, y1 + dy1 ) )
    res.append( (x1 + dx2, y1 + dy2 ) )
    return res
    
#convert global to local coordinate
#for eg: myX,myY,myH = self position, destX,destY = global ball
#return local ball.
def getLocalCoordinate(myX,myY,myH,x,y):
    x -= myX
    y -= myY
    return rotate( (x,y), 90-myH)

#convert local to global coordinate
def getGlobalCoordinate(myX,myY,myH,x,y):
    dx,dy = rotate((x,y), myH - 90)    
    return (dx + myX, dy + myY)

#get Heading from global coordinate
def getHeadingToMe(myX,myY,myH,x,y):
    return normalizeAngle_180( getHeadingBetween(myX,myY,x,y) - myH)

def distanceToLeftEdge(x,h):
    return x / math.cos(DEG2RAD(180 - h))
    
def distanceToRightEdge(x,h):
    return (Constant.FIELD_WIDTH - x) / math.cos(DEG2RAD(h))
    
def distanceToBottomEdge(y,h):
    return y / math.cos(DEG2RAD(270 - h))
    
def distanceToTopEdge(y,h):
    return (Constant.FIELD_LENGTH - y) / math.cos(DEG2RAD(90 - h))
    
#how far I am from the edge, facing that heading
def distanceToEdge(x,y,h):
    minD = 1000000
    if h > 90 and h < 270:
        minD = min(minD, distanceToLeftEdge(x,h))
    elif h < 90 or h > 270:   
        minD = min(minD, distanceToRightEdge(x,h))
        
    if h > 0 and h < 180:
        minD = min(minD, distanceToTopEdge(y,h))
    elif h > 180:
        minD = min(minD, distanceToBottomEdge(y,h))
    
    return minD

#
def getDistanceToClosestEdge(x,y):
    minD = 1000000
    if x < minD: minD = x
    if Constant.FIELD_WIDTH - x < minD : minD = Constant.FIELD_WIDTH - x
    if y < minD: minD = y
    if Constant.FIELD_LENGTH - y < minD : minD = Constant.FIELD_LENGTH - y
    return minD
