#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: HelpTrack.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# HelpTrack.py
#
# Different simple behaviours that might be used in any player codes. Compare to
# HelpShort.py, functions in this file is more complex.
#
#===============================================================================

import math
import Debug
import Global
import Constant
import HelpShort
import HelpLong
import HelpWhere
import VisionLink

# A global variable used by stationaryLocalise. This variable indicate what direction the robot
# want to pan to when localising.
panDirection = Constant.dCLOCKWISE

#--------------------------------------
# Adjust the pan, tilt and crane to look at the ball, if there is one on scene.
def trackVisualBall():

    ballPos = VisionLink.getProjectedBall()

    if Global.vBall.getConfidence() > 0:
            
        y = VisionLink.getGPSBallVInfo(Constant.CTLocal)[1]
        
        # The ball is coming to u
        adjustmentZ = y * 2.0
        adjustmentY = -2.5
        
        
        if Global.ballD < 25:
            adjustmentY = -1
            adjustmentZ = -2.5
        
##~         print "Y velocity %.2f , adjustment %.2f" % (y,adjustment)
##~         print "x %.2f   y %.2f    z %.2f" % (ballPos[0],-2.5,ballPos[1] + adjustment)
        
        # Note that xyz in getPointProjection is different from xyz to headMotion
        Global.lastHeadType = Global.finalAction[Constant.AAHeadType] = Constant.HTAbs_xyz
        Global.lastTilt     = Global.finalAction[Constant.AATilty]    = adjustmentY # Adj for new dog.
        Global.lastPan      = Global.finalAction[Constant.AAPanx]     = ballPos[0]
        Global.lastCrane    = Global.finalAction[Constant.AACranez]   = ballPos[1] + adjustmentZ

    else:
        Global.finalAction[Constant.AAHeadType] = Global.lastHeadType
        Global.finalAction[Constant.AAPanx]     = Global.lastPan
        Global.finalAction[Constant.AATilty]    = Global.lastTilt
        Global.finalAction[Constant.AACranez]   = Global.lastCrane

    # Make sure the head don't swing all the time.
    clipFireballPan()

    
#--------------------------------------
# Set the head not to have any pan movement.
# This assume the headType is relative (no point to do clip on absolute anyway).
def clipFireballPan():

    if Global.vBall.getConfidence() > 0:
        
        ballx = math.sin(HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
        bally = math.cos(HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
        
##~         print "local x %.2f,  y%.2f   ballH %.2f" % (ballx, bally,Global.ballH)
        if abs(ballx)<=3.0 and bally<=3.5*Constant.BallRadius:
            Global.finalAction[Constant.AAPanx] = 0
##~             print "clip clip"
    
    
#--------------------------------------
# Given a gps ball location, direct your head to it
def trackGpsBall():
    if abs(Global.gpsLocalBall.getHeading()) < 90:
        bx, by = Global.gpsLocalBall.getPos()
        HelpLong.setHeadParams(-bx,Constant.BallDiameter,by,Constant.HTAbs_xyz)
        
#--------------------------------------
# Move the point under the dog's neck to a location on the field
# Note that turnCCW is relative to the centre of the dog's body, not under its neck
def saGoToTargetFacingHeading(targetX, targetY, targetH, maxSpeed = 10, maxTurn = 30):

    # The speed constants from 2003 in Behaviour.h is 5.5, 15 for speed/turn.
    # Ted: I try to change this to faster speed whenever I have time.

    # Variables relative to self localisation.
    selfPos  = Global.selfLoc.getPos()
    selfH    = Global.selfLoc.getHeading()
    relX     = targetX - selfPos[0]
    relY     = targetY - selfPos[1]
    relH     = HelpShort.normalizeAngle_180(targetH - selfH)
    # now take into account that as we turn, the position of our head will change
    relX     += Constant.DOG_LENGTH/2.0/10.0*(math.cos(math.radians(selfH)) - math.cos(math.radians(targetH)))
    relY     += Constant.DOG_LENGTH/2.0/10.0*(math.sin(math.radians(selfH)) - math.sin(math.radians(targetH)))
    relD     = HelpShort.getLength((relX, relY))
    
    if relX == 0 and relY == 0:
        #on the dog, math.atan2(0,0) give "Value Error: math domain error"
        relTheta = 0
    else:
        relTheta = HelpShort.normalizeAngle_180(HelpShort.RAD2DEG(math.atan2(relY, relX)) - selfH)
    
    forward  = HelpShort.CLIP(relD, maxSpeed) * math.cos(HelpShort.DEG2RAD(relTheta))
    left     = HelpShort.CLIP(relD, maxSpeed) * math.sin(HelpShort.DEG2RAD(relTheta))
    turnCCW  = HelpShort.CLIP(relH, maxTurn)
        
    # in future if a better walktype, replace this
    HelpLong.setNormalWalk(forward, left, turnCCW)

#--------------------------------------
def saGoToTarget(targetX, targetY, maxSpeed = 5.5, maxTurn = 15):
    selfPos  = Global.selfLoc.getPos()
    selfH    = Global.selfLoc.getHeading()
    relX     = targetX - selfPos[0]
    relY     = targetY - selfPos[1]
    relD     = HelpShort.getLength((relX, relY))
    
    if relX == 0 and relY == 0:
        relTheta = 0
    else:
        relTheta = HelpShort.normalizeAngle_180(HelpShort.RAD2DEG(math.atan2(relY, relX)) - selfH)

    forward  = HelpShort.CLIP(relD, maxSpeed) * math.cos(HelpShort.DEG2RAD(relTheta))
    left     = HelpShort.CLIP(relD, maxSpeed) * math.sin(HelpShort.DEG2RAD(relTheta))
    turnCCW  = HelpShort.CLIP(relTheta, maxTurn)

    HelpLong.setNormalWalk(forward, left, turnCCW)

#--------------------------------------
# There are two dkd functions for some reason. This one correspond to the version in
# beckhamForward.cc.
# Note: this returns a global heading between 0 and 360 (like the gps self heading)
# returns a tuple of 5 numbers (avgDir, narrowMin, narrowMax, wideMin, wideMax)
def getDKDRange():

    nearEdgeDist = Constant.WALL_THICKNESS + 20
    debugDKD     = False
    length       = Constant.FIELD_LENGTH
    width        = Constant.FIELD_WIDTH

    # Make sure determine ballSource is called
    if Global.ballSource is None:
        if debugDKD:
            print "Warning: in getDKDRange() havent determined the ball source before getDKD()"
        return None
        
    X_ = Global.ballX
    Y_ = Global.ballY
    if debugDKD:
        print "Ball Source: ", Global.ballSource, "   X_ ", X_, "  Y_ " , Y_ 
    
    # If you don't know where the ball is determine dkd based on where you are.
    if (Global.ballSource == Constant.GPS_BALL and
        VisionLink.getGPSBallMaxVar() >= HelpShort.get95cf(35)):
        
        X_, Y_ = Global.selfLoc.getPos()
        if debugDKD:
            print "Warning: in getDKDRange(): I dont trust GPS ball right now: use my own pos is X_ ", X_, "  Y_ " , Y_ 

    Y_ = HelpShort.CLIPTO(Y_,0,length)
    X_ = HelpShort.CLIPTO(X_,0,width)
    
    # If the ball is near one of the side edges and not in corners then the direction becomes up the field.
    if Y_ > nearEdgeDist:
        if X_ < nearEdgeDist: # on the left edge
            if debugDKD:
                print "On the left edge"
            return (90, 70, 100, 10, 110)
        elif X_ > (width - nearEdgeDist): #on the right edge
            if debugDKD:
                print "On the right edge"
            return (90, 80, 110, 70, 170)

    # upper 60%, can I use voak?

    if Y_ > 0.6*length:
    
        # If target goal seen and is close and not many obstacles, it has a good heading, get the best heading like VOAK
        # But first find out how many blocking robots
        numOfBlockingRobots = 0
        i = 0
        while i < Constant.NUM_TEAM_MEMBER:
            if Global.vOpponents[i].getConfidence() > 0:
                numOfBlockingRobots = numOfBlockingRobots + 1
            if Global.vTeammates[i].getConfidence() > 0:
                numOfBlockingRobots = numOfBlockingRobots + 1
            i = i + 1
            
        if (Global.vTGoal.getConfidence() > 3 and 
            Global.ballSource == Constant.VISION_BALL and
            numOfBlockingRobots < 4):
            
            toY       = length - Y_
            toX       = width/2.0 - X_
            xRange    = Constant.GOAL_WIDTH/2.0 + (toY * math.tan(HelpShort.DEG2RAD(30)))
            inVOAKcal = abs(toX) < xRange
            
            if inVOAKcal:
            
                tdkdmin, tdkdmax = VisionLink.getHeadingToBestGap()
                tdkdmin = HelpShort.normalizeAngle_0_360 (tdkdmin+Global.selfLoc.getHeading())
                tdkdmax = HelpShort.normalizeAngle_0_360 (tdkdmax+Global.selfLoc.getHeading())
                dkd = (tdkdmin+ tdkdmax) / 2.0
                if debugDKD:
                    print "VOAK calculation, numOfBlockingRobots ", numOfBlockingRobots
                
                if (dkd < 150 and dkd > 30):
                    return (dkd,tdkdmin,tdkdmax,10,170)
    
    # Find the dkd fron the center of the target and oen goals
    dkdup     = HelpShort.getHeadingBetween(X_, Y_, (width / 2.0), length)
    dkdbottom = HelpShort.getHeadingBetween((width/2.0),0,X_,Y_)
    
    # if i cant use voak, then the headhing to the center of target goal is used as dkd
    if Y_ > 0.6*length:
        
        tdkdmin = HelpShort.getHeadingBetween (X_,Y_,Constant.RIGHT_GOAL_POST,length)
        tdkdmax = HelpShort.getHeadingBetween (X_,Y_,Constant.LEFT_GOAL_POST,length)
        if debugDKD:
            print "close to target goal, dkd directed to the target goal"
        return (dkdup,tdkdmin,tdkdmax,10,170)
         
    # if close to own goal, direted out of own goal
    elif Y_ < 0.4*length:
    
        if debugDKD:
            print "close to own goal, dkd directed out of the own goal"
        return (dkdbottom, 20, 160, 10 , 170)
        
    # Now, the middle 20% of the field should have a conutinuous chnage of direction.   
    else:
        myy   = Y_ - 0.4*length
        ratio = myy/(0.2*length)
        dkd   = dkdup * ratio + dkdbottom * (1-ratio) 
        
        if debugDKD:
            print "middle 20%, linear interpolation"
        return (dkd, 20, 160, 10, 170 )
        
#--------------------------------------
# NOTE: ballX and ballY are global coordinate
# BUT:    ballH and ballD are relative to the robot neck base
# Want a local one? Go to HelpShort.getLocalCoordinate
def determineBallSource():
    selfX = Global.selfLoc.getPos()[0]
    selfY = Global.selfLoc.getPos()[1]
    selfH = Global.selfLoc.getHeading()
    
    # If vision ball is available, use it since it is the most accurate.
    if Global.vBall.getConfidence() > 0:
        Global.ballSource          = Constant.VISION_BALL
        Global.ballX, Global.ballY = Global.vBall.getPos()
        Global.ballH               = Global.vBall.getHeading()
        Global.ballD               = Global.vBall.getVobDistance()

    # Check if wireless ball is accurate.
    # CRM: Maybe teammate got lost, then their "reliable" ball's gps coord will be wrong.
    elif (Global.sharedBallVar < (VisionLink.getGPSBallMaxVar() + \
            HelpShort.get95cf(Constant.USE_WIRELESS_BALL_VAR))
            and Global.lostBall > Constant.VBALL_LOST):
        Global.ballSource = Constant.WIRELESS_BALL
        Global.ballX      = Global.sharedBallX
        Global.ballY      = Global.sharedBallY
        Global.ballH      = HelpShort.normalizeAngle_180(HelpShort.getHeadingBetween\
        (selfX, selfY, Global.ballX, Global.ballY) - selfH)
        Global.ballD      = HelpShort.getDistanceBetween(Global.ballX, Global.ballY, selfX, selfY)

    # Use gps ball if nothing better.
    else:
        Global.ballSource = Constant.GPS_BALL
        Global.ballX, Global.ballY = Global.gpsGlobalBall.getPos()
        Global.ballH      = Global.gpsLocalBall.getHeading()
        Global.ballD      = Global.gpsLocalBall.getDistance()

#--------------------------------------
def stationaryLocalise(speed = 5, leftLimit = 80, rightLimit = -80):
    global panDirection
    
    if Global.degPan >= leftLimit and panDirection == Constant.dANTICLOCKWISE:
        panDirection = Constant.dCLOCKWISE
    elif Global.degPan <= rightLimit and panDirection == Constant.dCLOCKWISE:
        panDirection = Constant.dANTICLOCKWISE

    if panDirection == Constant.dCLOCKWISE:
        panx = -speed
    else:
        panx = speed

    if abs(Global.degCrane) > 5:
        cranez = (0 - Global.degCrane) / 3
    else:
        cranez = 0
    
    if abs(Global.degTilt - 10) > 5:
        tilty = (10 - Global.degTilt) / 3
    else:
        tilty = 0
            
    Global.finalAction[Constant.AAHeadType] = Constant.HTRel_h
    Global.finalAction[Constant.AAPanx]     = panx
    Global.finalAction[Constant.AATilty]    = tilty
    Global.finalAction[Constant.AACranez]   = cranez


#--------------------------------------
currentPanx = 0
panningCounter = 0
PANNING_TOO_LONG = 60
def spinningLocalise(clockwise = True,  lowCrane = -20, highCrane = 20, tilty = -20,speed = 5, minPan = -80, maxPan = 80):
    global panDirection,currentPanx, panningCounter
    
    if (panningCounter > PANNING_TOO_LONG or Global.degPan >= maxPan) and panDirection == Constant.dANTICLOCKWISE:
        panDirection = Constant.dCLOCKWISE
        panningCounter = 0
    elif (panningCounter > PANNING_TOO_LONG or Global.degPan <= minPan) and panDirection == Constant.dCLOCKWISE:
        panDirection = Constant.dANTICLOCKWISE
        panningCounter = 0
    
    panningCounter += 1
    
    if HelpShort.xor(clockwise,panDirection == Constant.dCLOCKWISE):
        cranez = lowCrane
    else:
        cranez = highCrane
        
    if panDirection == Constant.dCLOCKWISE:
        currentPanx += -speed
    else:
        currentPanx += speed
        
    HelpLong.setHeadParams(currentPanx,tilty,cranez,Constant.HTAbs_h)

# Correspond to beckhamForward.isBeckhamHaveBall().
# Of course this function should be called after determineBallSource.
def doIHaveTheBall():
    # Warning: Assume if HelpShort.canSeeBall() is true then the ballD in Global is the vision ball.

    # Close to ball and can't see teammate. Maybe 60 should change to 45, 
    # ask Daniel about robot distance.
    if (HelpShort.canSeeBall() and Global.ballD < 60 and 
        not HelpLong.canSeeTeammateWithinDist(60)):
        return True
    
    # Close to ball and heading roughly correct.
    # Ted and Kevin: utilIsUp should be using dkd, mean if the robot is close to the dkd.
    #                             At the moment, it used with the global heading. Come back later.
    
    if (Global.DKD is not None and 
        HelpShort.canSeeBall() and 
        Global.ballD < 80 and 
        HelpWhere.utilIsUp(Global.DKD)
       ):
        return True
        
    return False
