#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: HelpWhere.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




import Global
import Constant
import HelpShort

CloseEdgeDist     = 40
BallCloseEdgeDist = 30

# =====================================================
# Cases about my own positions
# =====================================================
def isInOwnGoalBox(offset = 0):
    (posx, posy) = Global.selfLoc.getPos()
    
    return (posy < (Constant.GOALBOX_DEPTH + Constant.WALL_THICKNESS + offset) and \
        abs(posx - Constant.OWN_GOAL_X) < (Constant.GOALBOX_WIDTH / 2 + offset))

def isNotAlongEdge(offset = CloseEdgeDist):
    return (not isOnBEdge(offset) and 
            not isOnLEdge(offset) and
            not isOnREdge(offset) and
            not isOnTLEdge(offset) and
            not isOnTREdge(offset) )
            
    
def isOnLEdge (offset = CloseEdgeDist):
    return ( Global.selfLoc.getX() < (Constant.WALL_THICKNESS + offset))

def isOnREdge(offset = CloseEdgeDist):
    return (Global.selfLoc.getX() > (Constant.FIELD_WIDTH - Constant.WALL_THICKNESS - offset))

def isOnTEdge(offset = CloseEdgeDist):
    return (Global.selfLoc.getY()> (Constant.FIELD_LENGTH - Constant.WALL_THICKNESS - offset) )
    
def isOnBEdge (offset = CloseEdgeDist):
    return (Global.selfLoc.getY() < (Constant.WALL_THICKNESS + offset))

def isOnTLEdge(offset = CloseEdgeDist):
    return ( (Global.selfLoc.getX() < Constant.LEFT_GOAL_POST) and
             (Global.selfLoc.getY() > Constant.FIELD_LENGTH - Constant.WALL_THICKNESS - offset) )
    
def isOnTREdge(offset = CloseEdgeDist):
    return ( (Global.selfLoc.getX() > Constant.RIGHT_GOAL_POST) and
             (Global.selfLoc.getY() > Constant.FIELD_LENGTH - Constant.WALL_THICKNESS - offset) )
    
def isOnTheField(x,y,offset = 0):
    return x >= offset and x <= Constant.FIELD_WIDTH - offset and \
        y >=offset and y <= Constant.FIELD_WIDTH - offset
    
# =====================================================
# Cases about the ball positions
# =====================================================
def ballInOwnGoalBox(offset = 0):
    # Use the global ball position regardless of the actual source.
    if (Global.ballX >= Constant.LEFT_GOALBOX_EDGE_X - offset and
        Global.ballX <= Constant.RIGHT_GOALBOX_EDGE_X + offset and
        Global.ballY <= Constant.OWN_GOALBOX_EDGE_Y + offset):
        return True
    return False

def ballOnLREdge ():
    return (ballOnLEdge() or ballOnREdge())

def ballOnLEdge(offset = BallCloseEdgeDist):
    if Global.ballSource != Constant.VISION_BALL:
        return False
    return (Global.ballX < (Constant.WALL_THICKNESS + offset))

def ballOnREdge (offset = BallCloseEdgeDist):
    if Global.ballSource != Constant.VISION_BALL:
        return False
    return (Global.ballX > (Constant.FIELD_WIDTH - Constant.WALL_THICKNESS - offset))

def ballOnTEdge(offset = BallCloseEdgeDist):
    if Global.ballSource != Constant.VISION_BALL:
        return False
    return (Global.ballY > (Constant.FIELD_LENGTH - Constant.WALL_THICKNESS - offset))
    
def ballInTLCorner ():
    return ( (Global.ballY > (Constant.FIELD_LENGTH - Constant.WALL_THICKNESS - 40)) and
        (Global.ballX < (Constant.WALL_THICKNESS + 40)) )

def ballInTRCorner():    
    return ( (Global.ballY > (Constant.FIELD_LENGTH - Constant.WALL_THICKNESS - 40)) and
        (Global.ballX > (Constant.FIELD_WIDTH - Constant.WALL_THICKNESS - 40)) )

def ballInBLCorner():
    return (Global.ballX <= (Constant.WALL_THICKNESS + 40)) and \
           (Global.ballY <= (Constant.WALL_THICKNESS + 40))
    
def ballInBRCorner():
    return (Global.ballX >= (Constant.FIELD_WIDTH - Constant.WALL_THICKNESS - 40)) and\
           (Global.ballY <= (Constant.WALL_THICKNESS + 40))
        
def ballOnBEdge(offset = BallCloseEdgeDist):
    if Global.ballSource != Constant.VISION_BALL:
        return False
    
    # At this point Global.ballx and Global.bally should be vision ballX should be vision ball from determineBallSource().
    return (Global.ballY < Constant.WALL_THICKNESS + offset)

def ballOnTLEdge(offset = BallCloseEdgeDist):
    if Global.ballSource != Constant.VISION_BALL:
        return False
        
    return ( (Global.ballX < Constant.LEFT_GOAL_POST) and
             (Global.ballY > (Constant.FIELD_LENGTH - Constant.WALL_THICKNESS - offset)) )
             

def ballOnTREdge(offset = BallCloseEdgeDist):
    if Global.ballSource != Constant.VISION_BALL:
        return False
        
    return ( (Global.ballX > Constant.RIGHT_GOAL_POST) and
             (Global.ballY > (Constant.FIELD_LENGTH - Constant.WALL_THICKNESS - offset)) )

def ballNotAlongEdge():
    if Global.ballSource != Constant.VISION_BALL:
        return False
    
    return (not ballOnBEdge() and 
            not ballOnLEdge() and
            not ballOnREdge() and
            not ballOnTLEdge() and
            not ballOnTREdge() )

##~     return ( (Global.ballX > (Constant.WALL_THICKNESS + BallCloseEdgeDist)) and
##~         (Global.ballX < (Constant.FIELD_WIDTH - Constant.WALL_THICKNESS - BallCloseEdgeDist)) and
##~         (Global.ballY > (Constant.WALL_THICKNESS + BallCloseEdgeDist )) and
##~         (Global.ballY < (Constant.FIELD_LENGTH - Constant.WALL_THICKNESS - BallCloseEdgeDist)) )
    
def isInFrontTargetGoal ():
    posx = Global.selfLoc.getX()
    return ( (abs(posx - Constant.FIELD_WIDTH / 2)) < Constant.GOAL_WIDTH/2 + 10 )

def ballInOwnHalf ():
    bally = Global.gpsGlobalBall.getPos()[1]
    return (bally < Constant.FIELD_LENGTH/ 2)

def isInRange (min, max):
    realmin = HelpShort.normalizeAngle_0_360(min)
    realmax = HelpShort.normalizeAngle_0_360(max)
    if realmax < realmin:
        realmax = realmax + 360
    
    head = Global.selfLoc.getHeading()
    
    if head < realmin:
        realmin = realmin - 360
        realmax = realmax - 360
        
    if head > realmax:
        realmin = realmin + 360
        realmax = realmax + 360
            
    if head >= min and head <= max:
        return True
    else:
        return False

# When this function is called, make sure that the DKD is not None, 
# putting in DKD as argument helps this
def utilIsUp(DKD):
    (_, tdkdmin, tdkdmax, wdkdmin, wdkdmax) = DKD
    return isInRange ( (tdkdmin+wdkdmin)/2.0 , (tdkdmax+wdkdmax)/2.0 )

def inTargetGoalBox(x,y):
    return (y >= Constant.LEFT_GOALBOX_EDGE_Y and 
            x >= Constant.LEFT_GOALBOX_EDGE_X  and
            x <= Constant.RIGHT_GOALBOX_EDGE_X)

def isHeadingForward(h):
    return h > 0 and h <= 180
    
def isHeadingLeftward(h):
    return h > 90 and h <= 270
    
def edgeIsOnMyRight(x,y,h):
    #if near left edge heading backward or near right edge heading forward
    return ( x < Constant.FIELD_WIDTH/2 and not isHeadingForward(h) ) \
        or ( x > Constant.FIELD_WIDTH/2 and isHeadingForward(h) ) \
        or ( y < Constant.FIELD_LENGTH/2 and not isHeadingLeftward(h) ) \
        or ( x > Constant.FIELD_LENGTH/2 and isHeadingLeftward(h) )

def isInTopLeftCorner(x,y,offset = 20):
    return x < offset and y > Constant.FIELD_LENGTH - offset
    
def isInTopRightCorner(x,y,offset = 20):
    return x > Constant.FIELD_WIDTH - offset and y > Constant.FIELD_LENGTH - offset
    
def isInBottomLeftCorner(x,y,offset = 20):
    return x < offset and y < offset

def isInBottomRightCorner(x,y,offset = 20):
    return x > Constant.FIELD_WIDTH - offset and y < offset
    
def isInTopEdgeCorner(_,y,offset = 20):
    return y > Constant.FIELD_LENGTH - offset

def isInBottomEdgeCorner(_,y,offset = 20):
    return y < offset
    
def isInRightEdgeCorner(x,_,offset = 20):
    return x > Constant.FIELD_WIDTH - offset 
    
def isInLeftEdgeCorner(x,_,offset = 20):
    return x < offset
