#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: JoystickPlayer.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#   JoystickPlayer.py
#
#  Tool to read joint values, mostly when the dog is relaxed
#
#===============================================================================


"""
Response to command sent by joystick controller
"""
import Global
import HeadKick
import Constant
import HelpLong
import OpenChallenger
from StateBasedAction import StateBasedAction
import SetPlayer
import HelpShort
import ReadyPlayer
import HelpTrack
import LocateBall

headKickRightTurn = True
headKickLeft      = False
headKickRight     = False
bInitGlobal       = True
bPrintJointSensor = False

class ActuatorAction(StateBasedAction):
    def __init__(self):
        StateBasedAction.__init__(self)
        self.walkAction = Global.finalAction[Constant.AAWalktype:Constant.AAHeadType]
        
    def storeFinalAction(self):
        self.walkAction = Global.finalAction[Constant.AAWalktype:Constant.AAHeadType]
        
    def DecideNextAction(self):
        Global.finalAction[Constant.AAWalktype:Constant.AAHeadType] = self.walkAction[:]
        return self
    
class DoItOnce(StateBasedAction):
    def __init__(self):
        StateBasedAction.__init__(self)
        self.walkAction = Global.finalAction[Constant.AAWalktype:Constant.AAHeadType]
        
    def storeFinalAction(self):
        self.walkAction = Global.finalAction[Constant.AAWalktype:Constant.AAHeadType]
        
    def DecideNextAction(self):
        Global.finalAction[Constant.AAWalktype:Constant.AAHeadType] = self.walkAction[:]
        HelpLong.setForceStepComplete()
        print "DoItOnce", Global.finalAction
        return None
        
class AutoHeadControl(StateBasedAction):
    def __init__(self):
        StateBasedAction.__init__(self)
                
        
    def DecideNextAction(self):
        if Global.lostBall == 0:
            HelpTrack.trackVisualBall()
        else:
            LocateBall.headFindBall(-10,0)
        
        return self
        
        
actuatorAction = ActuatorAction()
grabBallAction = OpenChallenger.RunToGrabBall()
autoHeadControl= AutoHeadControl()
doItOnce       = DoItOnce()

theAction      = None
theHead        = None

#persistent command: set head, set joint ( finalAction is set every frame)
#non-persistent command : set indicators (finalIndicators only set on new command)
            
        
def doJoyStickPlayer():
    global theAction, headKickLeft, headKickRight
    
    OpenChallenger.resetEachFrame()
    
    if headKickLeft:
        headKickLeft = HeadKick.headKickLeft()
        return
        
    if headKickRight:
        headKickRight = HeadKick.headKickRight()
        return
    
    if theHead is not None:
        theHead.DecideNextAction()
    if theAction is not None:
        theAction = theAction.DecideNextAction()
    
        
def DecideNextAction():   
    Global.frameReset()
    HelpLong.superDebug()
 
    # Finished state - the robot can't move at all.
    if Global.state == Constant.FINISHED:
        HelpLong.showBatteryStatus()
        HelpShort.resetEverything()

    # Ready state - let me move to my correct position.
    elif Global.state == Constant.READYSTATE:
        ReadyPlayer.readyMatchPlaying()

    # Set state - I can't move my legs, but I can move heads to look for ball.
    elif Global.state == Constant.SETSTATE:
        SetPlayer.DecideNextAction()

    # Playing state - lets play a game.
    else:
        doJoyStickPlayer()

pan,tilt,crane = 0,0,0

def setHeadParams(p,t,c):
    global pan,tilt,crane
    pan,tilt,crane = p,t,c
    HelpLong.setHeadParams(pan,tilt,crane,Constant.HTAbs_h)
    
def processCommand(cmd):
    global theAction, theHead, headKickLeft, headKickRight, headKickRightTurn
    print "JoystickPlayer got |%s|"%cmd
    cmd = cmd.strip()
    if cmd[:3] == "ind":
        print "Indicator command",cmd[3:]
        indicator,value = cmd[3:].split()
        indicator = int(indicator)
        value = int(value)
        Global.finalIndicators[indicator] = value
    elif cmd[:4] == "head":
        print "Head command ", cmd[4:]
        p,t,c = map(float,cmd[4:].split())
        setHeadParams(p,t,c)
        actuatorAction.storeFinalAction()
        theAction = actuatorAction    
    elif cmd[:8] == "joy move":        
        f,l,t,wt = map(float,cmd[8:].split())
        wt = int(wt)
        print "Walk command ", cmd[8:], " wt = ",wt
        if wt == Constant.EllipseWalkWT:
            HelpLong.setEllipticalWalk(f,l,t)
        elif wt == Constant.NormalWalkWT:
            HelpLong.setNormalWalk(f,l,t)
        else:
            Global.finalAction[Constant.AAWalktype] = wt
            HelpLong.setForwardLeftTurn(f,l,t)
        
        actuatorAction.storeFinalAction()
        theAction = actuatorAction
    elif cmd[:8] == "joy butt":
        command = cmd[8:].strip()  
        if command == "grab":
            print "Do auto grab"
            theAction = grabBallAction
        elif command == "autohead":
            print "Do autohead"
            if theHead is None:
                theHead = autoHeadControl
            else:
                theHead = None
        else:
            if command == "divekick":
                wt = Constant.DiveKickWT
            elif command == "sideR":
                wt = Constant.UpennRightWT
            elif command == "sideL":
                wt = Constant.UpennLeftWT
            elif command == "HKleft":
                wt = Constant.HandKickLeftWT
            elif command == "HKright":
                wt = Constant.HandKickRightWT
            elif command == "headKick":
                wt = Constant.NormalWalkWT
                if headKickRightTurn:
                    headKickRightTurn = False
                    headKickRight = True
                else:
                    headKickRightTurn = True
                    headKickLeft = True
            else:
                return
            theAction = doItOnce
            print "Joybutton command ", command, " wt = ",wt        
            Global.finalAction[Constant.AAWalktype] = wt
            doItOnce.storeFinalAction()
            
