#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: Kick.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




import math
import Global
import Constant
import HelpLong
import HelpShort

# NormalWalkWT means no kick is going on, ie. currentKick = True.
currentKick    = Constant.NormalWalkWT

#CRM - Need to get rid of Global.lockmode, lockmodes = EVIL

framesSinceLastKicking = 0

#=================================
# Upleft DOES NOT mean the ball travels left (in fact it travels right)
# It means which hand

def UPleft ():
    global currentKick, framesSinceLastKicking
    
    if Global.pWalkInfo.getCurrentWalkType()==Constant.UpennLeftWT:
        currentKick = Constant.NormalWalkWT
        HelpLong.setDefaultAction()
    else:
        currentKick          = Constant.UpennLeftWT
        Global.finalAction = list (Constant.UPleftKick)

    framesSinceLastKicking = 0
        
#=================================
# condition for upenn suitability

def isUPleftOk ():

    if Global.vBall.getConfidence() > 0:
    
        ballx = math.sin (HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
        bally = math.cos (HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
        
        if (ballx < 3.2 and ballx > -2.0 and
            bally < 3.4*Constant.BallRadius and
            Global.ballD < 12 and
            Global.degTilt < -7 and abs(Global.degPan) < 30):
            return True
            
    return False

#=================================
def UPright ():

    global currentKick, framesSinceLastKicking
    
    if Global.pWalkInfo.getCurrentWalkType()==Constant.UpennRightWT:
        currentKick = Constant.NormalWalkWT
        HelpLong.setDefaultAction()
    else:
        currentKick          = Constant.UpennRightWT
        Global.finalAction = list (Constant.UPrightKick)

    framesSinceLastKicking = 0

        
def isUPrightOk():

    if Global.vBall.getConfidence() > 0:
    
        ballx = math.sin (HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
        bally = math.cos (HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
        
        if (ballx < 2.0 and ballx > -3.2 and
            bally < 3.4*Constant.BallRadius and
            Global.ballD < 12 and
            Global.degTilt < -7 and abs(Global.degPan) < 30):
            return True
            
    return False


#==================================
# Also called goalie kick in histroy

def ForwardKick ():

    global currentKick, framesSinceLastKicking
    
    if Global.pWalkInfo.getCurrentWalkType()==Constant.FwdKickWT:
        currentKick = Constant.NormalWalkWT
        HelpLong.setDefaultAction()
    else:
        currentKick           = Constant.FwdKickWT
        Global.finalAction  = list (Constant.ForwardKick)

    framesSinceLastKicking = 0

#==================================
# LightningKick have less setup time, just one step less

def LightningKick ():

    global currentKick, framesSinceLastKicking
    
    if Global.pWalkInfo.getCurrentWalkType()==Constant.FastKickWT:
        currentKick = Constant.NormalWalkWT
        HelpLong.setDefaultAction()
    else:
        currentKick           = Constant.FastKickWT
        Global.finalAction  = list (Constant.FastKick)

    framesSinceLastKicking = 0

def diveKick():
    global currentKick, framesSinceLastKicking
    
    if Global.pWalkInfo.getCurrentWalkType() == Constant.DiveKickWT:
        currentKick = Constant.NormalWalkWT
        HelpLong.setDefaultAction()
    else:
        currentKick        = Constant.DiveKickWT
        Global.finalAction = list(Constant.DiveKick)

    framesSinceLastKicking = 0
    
def handKickRight():
    global currentKick, framesSinceLastKicking
    
    if Global.pWalkInfo.getCurrentWalkType() == Constant.HandKickRightWT:
        currentKick = Constant.NormalWalkWT
        HelpLong.setDefaultAction()
    else:
        currentKick        = Constant.HandKickRightWT
        Global.finalAction = list(Constant.HandKickRight)
    
    framesSinceLastKicking = 0
    
def handKickLeft():
    global currentKick, framesSinceLastKicking
    
    if Global.pWalkInfo.getCurrentWalkType() == Constant.HandKickLeftWT:
        currentKick = Constant.NormalWalkWT
        HelpLong.setDefaultAction()
    else:
        currentKick        = Constant.HandKickLeftWT
        Global.finalAction = list(Constant.HandKickLeft)
    
    framesSinceLastKicking = 0
    
#==================================
# Have i done the Kicks I wanted, if not, continue and return True.
def shouldIContinueKick():

    global currentKick

    if currentKick == Constant.NormalWalkWT:
        return False

    return True
    
#==================================
# Continue to do the desired kick.
def continueKick():
    global currentKick

    debug       = False
    whatKick    = None

    if currentKick == Constant.UpennLeftWT:
        whatKick = "Doing UPLeft kick."
        UPleft()
    
    elif currentKick == Constant.UpennRightWT:
        whatKick = "Doing UPRight kick."
        UPright()

    elif currentKick == Constant.FwdKickWT:
        whatKick = "Doing Forward kick."
        ForwardKick()

    elif currentKick == Constant.FastKickWT:
        whatKick = "Doing Lightning kick."
        LightningKick()

    if debug and whatKick != None:
        print "currentKick: ", currentKick, " - Kick? ", whatKick

