#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: KimPlayer.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# KimPlayer.py
#
# Kim's player for everyone's reference. 
#
# (This file may keep changing from time to time as Kim wants to try his own
# stuff. Comments as in 22/4/2004 written by Daniel, pls correct if you pick up
# anything)
#
#===============================================================================


# Some principles that I think of and believe we should follow so that everyone 
# can work happily! Comments welcome~~
#
# >> Do not modify finalAction values at the end of DecideNextAction base on 
#       lockmodes or whatever, ie. NO HACK!
#
# >> Keep EVERY line no longer than 80, if not, should never exceeds 100 char.
#
# >> Break things into modules when it grows too large, like too many functions 
#       in a file, a single task became too complex, etc (furthur explained below).
#
# >> If a function is returning something rather than "None" ("None" is like 
#       "void" in C++), specify that in the function header comment.
#
# >> More items here welcome.
#

"""
KimPlayer
"""

import Global 
import VisionLink
import HelpShort
import Constant
import math



#===============================================================================
# Global variables of this module.
#===============================================================================

# Note that there is no const declaration in python. So we will use Capital 
# letters to indicate that we want this to be a constant.
GET_BEHIND_DISTANCE = 50
NUMBER_OF_FRAMES    = 50
PawKickRadius         = 7.5
dutyRange           = HelpShort.array2D(Constant.NUM_OF_EFFECTORS,NUMBER_OF_FRAMES)
counter             = 0
#--------------------------------------
# Basic Action
BABallTracking          = 0
BAPawKick               = 1
BAActiveLocalise        = 2
BAGetBehindBall         = 3
BARunBehindBall         = 4
BALocateBall            = 5

finalAction     = list(Constant.NormalWalkStandStill) #converting to list to be mutable
ballLoc         = None
selfLoc         = None
useBasicAction  = None
lockMode        = None



#===============================================================================
# Functions belong to this module.
#===============================================================================

# Notes on functions: Any function can return any type of data! If the calling 
# function expects something different from what get returned, error can be found 
# using pychecker.
#
# So whatever function you write, it is a good habit to state what you are going
# to return in the  beginning.    
    
#------------------------------------------------------------------------------
# Main Behaviours
#------------------------------------------------------------------------------


#--------------------------------------
# Called by processFrame() in Behaviou.py.
# Return None
#
# Every player will have this file and it should always be the first function 
# for readability.
def DecideNextAction():

    # How to make our codes neat and manageable?
    # Write codes in a new module in appropriate times. What does that mean?
    #
    # Generally speaking, deviding complex task into smaller modules are good, 
    # but there are cases when you can't. Let say Module A import Module B. 
    # Module A can call functions of and use variables in Module B, but not the 
    # other way round. So you can't just cut things into modules all the time.

    global lockMode
    initAction()
    lockMode = testingMode
    processLockMode()
    Global.finalAction = finalAction

    # Notes on boolean: you can use constants True / False. If it is an integer,
    # zero is False and non-zero is True.
    if useBasicAction is not None:
        VisionLink.doBasicBehaviour(useBasicAction)


#--------------------------------------
def initAction():
    global ballLoc, selfLoc, useBasicAction
    
    useBasicAction = None
    ballLoc = VisionLink.getBallLocation()
    selfLoc = VisionLink.getSelfLocation()
    return None


#--------------------------------------
# Note that we will not be using lockMode terminlogy from now on. When Kim wrote
# this things were not well decided yet. 
def chooseLockMode():
    if canSeeBall():
        return clearBallAction
    else:
        return lostBallAction


#--------------------------------------
def getLockModeName(lm):
    return str(lm).split()[0][11:]
    
##~     if isinstance(lm,ActiveLocalise):
##~         return "ActiveLocalise"
##~     elif isinstance(lm,ClearBallAction):
##~         return "ClearBallAction"
##~     elif isinstance(lm,FirePawKickAction):
##~         return "FirePawKickAction"
##~     elif isinstance(lm,GetBehindBall):
##~         return "GetBehindBall"
##~     elif isinstance(lm,HoverToBall):
##~         return "HoverToBall"
##~     elif isinstance(lm,LostBallAction):
##~         return "LostBallAction"
##~     elif isinstance(lm,PawKickAction):
##~         return "PawKickAction"
##~     elif isinstance(lm,RunBehindBall):
##~         return "RunBehindBall"
##~     elif isinstance(lm,RunTowardBall):
##~         return "RunTowardBall"
##~     elif isinstance(lm,StandStillAction):
##~         return "StandStillAction"
##~     elif isinstance(lm,WalkAroundBall):
##~         return "WalkAroundBall"
##~     else:
##~         return "Unknown Lockmode",  type(lm)


#--------------------------------------
def processLockMode():
    global lockMode
    if lockMode is None:
        lockMode = chooseLockMode()
    print getLockModeName(lockMode)
    #todo: add a loop
    lockMode = lockMode.DecideNextAction()

    
#--------------------------------------
def printSensorValues():
    print "Sensor "
    for i in range(Constant.NUM_OF_EFFECTORS):
        print "%7.2f" % VisionLink.getJointPWMDuty(i),
    print ""


#--------------------------------------
def printDutyRange():
    global dutyRange
    print "Duty ",
    for i in range(Constant.NUM_OF_EFFECTORS):
        print "%7.2f" % max(dutyRange[i]),
    print ""
    

#--------------------------------------
def updatePWMDuty():
    global counter, dutyRange
    counter = (counter + 1) % NUMBER_OF_FRAMES
    for i in range(Constant.NUM_OF_EFFECTORS):
        t = VisionLink.getJointPWMDuty(i)
        dutyRange[i][counter] = abs(t)


#--------------------------------------
def dutyMaxList():
    global dutyRange
    l = []
    for i in range(Constant.NUM_OF_EFFECTORS):
        t = 0
        for j in range(NUMBER_OF_FRAMES):
            if abs(dutyRange[i][j]) > t:
                t = abs(dutyRange[i][j])
        l.append(t)
    return l


#--------------------------------------
def isStuckDetected():
    global dutyRange
    for i in range(Constant.NUM_OF_EFFECTORS):
        t = max(dutyRange[i])
        if t > 300.0:
           return True
    return False


#--------------------------------------
def testStuckDection():
    global dutyRange
    print "sendingAA"
    updatePWMDuty()
    printSensorValues()
    printDutyRange()
    if (isStuckDetected()):
        printDutyRange()
        print "Stuck detected -------------------------------------------"


#--------------------------------------
def doTesting():
#    testStuckDection()
#    print VisionLink.getAnySensor(29)
#    print VisionLink.getBallLocationGlobal()   
    pass


#------------------------------------------------------------------------------
# HelpShort Behaviours
#------------------------------------------------------------------------------

#--------------------------------------
def canSeeBall():
    return VisionLink.getBallConfidence() > 0



#===============================================================================
# Sub-classes of this module
#===============================================================================

# Because this file is just for a basic testing, lengthy actions are just sub-
# classes below. However in the future, when they become lengthy, they should
# be many modules (files) instead. That's why each of them have the 
# DecideNextAction function as if they are player modules.

#------------------------------------------------------------------------------
#  Module initialization
#------------------------------------------------------------------------------

#--------------------------------------
class FirePawKickAction:
    def DecideNextAction(self):
        global useBasicAction
        useBasicAction = BAPawKick
        return self
        
#--------------------------------------
class PawKickAction:
    LEFT_PAW = 1
    RIGHT_PAW = 2
    def __init__(self):
        self.lostBallPeriod = 0
        self.lastCommand = list(Constant.EllipticalWalkFast)
        
    def DecideNextAction(self):
        if not canSeeBall():
            self.lostBallPeriod += 1
        else:
            self.lostBallPeriod = 0
            
        if self.lostBallPeriod > 50:
            return lostBallAction
            
        global finalAction,useBasicAction
            
        if canSeeBall():
            h = VisionLink.getBallHeading()
            d = VisionLink.getBallDistance()
            #choose paw
            if h > 0:  paw = PawKickAction.LEFT_PAW
            else: paw = PawKickAction.RIGHT_PAW
            
            offHeading = HelpShort.RAD2DEG( math.asin(PawKickRadius/d) )
            print "h = %3.1f , off = %3.1f " % (h,offHeading),
            if (paw == PawKickAction.RIGHT_PAW ):
                h += offHeading
            else:
                h -= offHeading
            
            if d < 30 and h > 40:
                self.lastCommand = list(Constant.NormalWalkFast)
            else:
                self.lastCommand = list(Constant.EllipticalWalkFast)
            self.lastCommand[Constant.AATurnCCW] = HelpShort.CLIP(h, 40)
            print " turn = %3.1f" % self.lastCommand[Constant.AATurnCCW]
            
        finalAction = self.lastCommand
        useBasicAction = BABallTracking
        return self
        
#--------------------------------------
class ActiveLocalise:
    def DecideNextAction(self):
        global useBasicAction
        useBasicAction = BAActiveLocalise
        return self
        
#--------------------------------------
class GetBehindBall:
    def DecideNextAction(self):
        global useBasicAction
        useBasicAction = BAGetBehindBall
        h = VisionLink.getBallHeading()
        if canSeeBall():
            if h > 0 and h < 180:
                return pawKickAction
            return self
        else:
            return None

#--------------------------------------
class RunBehindBall:
    def DecideNextAction(self):
        global useBasicAction
        useBasicAction = BARunBehindBall
        return self


#--------------------------------------
class RunTowardBall:
    def __init__(self):
        self.lastCommand = list(Constant.EllipticalWalkFast)
        
    def DecideNextAction(self):
        global finalAction,useBasicAction
        finalAction = list(Constant.EllipticalWalkFast)
        h = VisionLink.getBallHeading()
        finalAction[Constant.AATurnCCW] = h/2.0
        useBasicAction = BABallTracking
        if VisionLink.getBallDistance() < GET_BEHIND_DISTANCE:
            return getBehindBall
        return self

#--------------------------------------
class HoverToBall:
    def DecideNextAction(self):
        global finalAction,useBasicAction
        finalAction = list(Constant.EllipticalWalkFast)
        h = VisionLink.getBallHeading()
        finalAction[Constant.AATurnCCW] = h/4.0
        useBasicAction = BABallTracking
        return self
         
#--------------------------------------
class StandStillAction:
    def DecideNextAction(self):
        global finalAction
        finalAction = list(Constant.NormalWalkStandStill)
        print VisionLink.getSelfLocation()
        return self
        
#--------------------------------------
class WalkAroundBall:
    def DecideNextAction(self):
        global finalAction
        heading = VisionLink.getSelfLocation()[3]
        print "heading %3.1f" % heading
        if heading > 60 and heading < 120:
            return pawKickAction
        elif heading >= 120 and heading <= 270:
            finalAction = list(Constant.NormalWalkSideWayFast)
            finalAction[Constant.AATurnCCW] = -10
            finalAction[Constant.AALeft] = 6
        else:
            finalAction = list(Constant.NormalWalkSideWayFast)
            finalAction[Constant.AATurnCCW] = 10
            finalAction[Constant.AALeft] = -6
        return self
            
        
#--------------------------------------
class ClearBallAction:
    def DecideNextAction(self):
        global finalAction, useBasicAction
        d = VisionLink.getBallDistance()
        if not canSeeBall():
            return lostBallAction
        else:
            robotHeading = VisionLink.getSelfLocation()[3]
            if (robotHeading > 0 and robotHeading < 180) or (d > GET_BEHIND_DISTANCE):
                return runTowardBall
            else:
                return getBehindBall
      
#--------------------------------------
class LostBallAction:
    def DecideNextAction(self): 
        global useBasicAction
        useBasicAction = BALocateBall
        if canSeeBall():
            return None
        return self

#--------------------------------------
class TestingMode:
    def  DecideNextAction(self):
        global finalAction
        finalAction = list(Constant.NormalWalkFast)
        finalAction[Constant.AAForward] = 7
        finalAction[Constant.AATurnCCW] = 0
        finalAction[Constant.AAHeadType] = Constant.HTAbs_h
        
        finalAction[Constant.AATilty] = -45
        return self
       



# The following is defined not on top, but here because classes are defined in 
# middle. In python, there's no other way around this order.

# The following are like getting instances from a class in java.
# Note that in the future each action will be in its own module, they won't be
# in a class so there will be no need to instantiate them like this.

#  List of lockmodes
clearBallAction = ClearBallAction()
lostBallAction  = LostBallAction()
standStillAction = StandStillAction()
walkAroundBall  = WalkAroundBall()
pawKickAction   = PawKickAction()
runTowardBall   = RunTowardBall()
hoverToBall     = HoverToBall()
activeLocalise  = ActiveLocalise()
getBehindBall   = GetBehindBall()
runBehindBall   = RunBehindBall()
testingMode     = TestingMode()

