#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: LocateBall.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# LocateBall.py
#
# This is a player that try to find the ball and look at it. Note that right now the Global.finalAction[..]
# are commented out, so it cannot control its headMotion unless some module called this module
# on top, and called FWHead.DecideNextAction separately (just like Forward).
#===============================================================================

"""
LocateBall
"""

import math
import Global
import Constant
import HelpLong
import HelpShort

import FWHead



#===============================================================================
# Some globals that are needed
#===============================================================================


# num frames in 1/2 cycle of head finding ball
HEAD_PERIOD = 22

# BL stands for "Ball Lost", number of frames after not seeing a ball.
# All these constants assume a frame rate of 30

BL_LOOK_GPS     = Constant.VBALL_LOST + 1   # Start looking at gps ball.
BL_FIND_ON_SPOT = BL_LOOK_GPS     + HEAD_PERIOD*2 + 10  # Start looking around.
BL_SLOW_FIND    = BL_FIND_ON_SPOT + HEAD_PERIOD*2 + 50  # Start walking on the spot slowly.
BL_FAST_FIND    = BL_SLOW_FIND + 120                    # Start spining fast with head aside.
                        # After all these, spin and move to a spot whilst doing head find ball

TILT_HIGH_NORMAL = -5

spinDir = Constant.dANTICLOCKWISE;
lookDir = Constant.dANTICLOCKWISE;

#--------------------------------------
# 3 stages to locate the ball, depending on how long the ball has been lost. 
def DecideNextAction(baseFrame = Constant.VBALL_LOST + 1):

    global lookDir, spinDir
    global BL_LOOK_GPS, BL_SLOW_FIND, BL_FAST_FIND, BL_FIND_ON_SPOT

    if baseFrame != Constant.VBALL_LOST + 1:
        BL_LOOK_GPS         = baseFrame
        BL_FIND_ON_SPOT     = BL_LOOK_GPS     + HEAD_PERIOD*2 + 10
        BL_SLOW_FIND        = BL_FIND_ON_SPOT + HEAD_PERIOD*2 + 50
        BL_FAST_FIND        = BL_SLOW_FIND + 120
    
    lball = Global.lostBall
    
    # WARNING: set yr localise timer less then BL_LOOK_GPS is ok, but not larger
    
    if lball<BL_LOOK_GPS:
        return

    # Set parameter if this is the first time get called.
    if (lball == BL_LOOK_GPS):
        setLocateBallParams()

    # 1st stage: Use only the head to find ball.
    if (lball <= BL_FIND_ON_SPOT):
        HelpLong.setDefaultAction()
        headFindBall(-60, TILT_HIGH_NORMAL)

    # 2nd stage: spin on the spot with head moving to find the ball. 
    elif (lball <= BL_SLOW_FIND):

        # Head motion is set in spinFindBall also, 
        # but will be overwritten in headFindBall following it.
        spinFindBall(25) 
        lookDir = spinDir 
        headFindBall(-50, TILT_HIGH_NORMAL) 

    # 3rd stage: spin on the spot quickly to find the ball. */
    elif (lball <= BL_FAST_FIND):
        spinFindBall(50) # spinSpeed = 50

    # 4th stage: spin and move to a particular spot.
    else:
        spinToFieldCenter()

    # In case this skill is called by player, it should not active localise.
    FWHead.compulsoryAction = FWHead.fixHead


#--------------------------------------
# First time you call locateBall you should call this, otherwise the dog will always spin anticlockwise
def setLocateBallParams():

    global spinDir

    (x,_) = Global.selfLoc.getPos()

    if (x > Constant.FIELD_WIDTH/2) :
        spinDir = Constant.dANTICLOCKWISE
    else:
        spinDir = Constant.dCLOCKWISE
    


#--------------------------------------
# Spin the the center of the field.
def spinToFieldCenter():
    spinToSpot(Constant.FIELD_WIDTH/2, Constant.FIELD_LENGTH/2)


#--------------------------------------
# The 4th stage of locateBall. Spin to the given spot in hope to find ball.
def spinToSpot(destX, destY):

    global lookDir, spinDir

    selfH = Global.selfLoc.getHeading()

    target = Global.Vector(destX + 8.0 * math.cos(HelpShort.DEG2RAD(selfH)),\
                    destY + 8.0 * math.sin(HelpShort.DEG2RAD(selfH)),\
                    0,0,0,0) # Only interested in x,y now.

    current = Global.Vector(Global.selfLoc.getX(), \
                     Global.selfLoc.getY(), \
                     0,0,0,0)

    # Vector to destination point
    diffVec     = current.sub(target)
    diffVec.calcHead()
    angleInBetw = HelpShort.normalizeAngle_180(diffVec.getHeading() - selfH)

    lookDir = spinDir
    headFindBall(-50, TILT_HIGH_NORMAL)

    if (spinDir == Constant.dANTICLOCKWISE):
        turnCCW = 20
    else:
        turnCCW = -20

    forward = -5.0 * math.cos(HelpShort.DEG2RAD(angleInBetw))
    left    = -5.0 * math.sin(HelpShort.DEG2RAD(angleInBetw))

    if (diffVec.getDistance() < 20):
        forward /= 4.0;
        left /= 4.0;

    # Set the calculated values.
    HelpLong.setNormalWalk(forward,left,turnCCW)

#--------------------------------------
# Spin the body on the spot to find ball. 
# Use the input variable to determine direction. 
def spinFindBall(spinSpeed):

    global spinDir

    if spinDir == Constant.dCLOCKWISE:
        panx    = -50
        turnCCW = -spinSpeed
    else:
        panx    = 50
        turnCCW = spinSpeed

    HelpLong.setNormalWalk(0,0,turnCCW)
    FWHead.fixHeadAt = (panx,0,0,Constant.HTAbs_xyz)

#--------------------------------------
# The head moves like a rectangle either clockwise or anti-clockwise in this motion.
# The tilt angle in upper and lower locus is specified by lowtilt/hightilt.
#
# tiltLow is an input because the amount of tilt in different stages are different.
def headFindBall(tiltLow, tiltHigh):

    global lookDir, lostBall
    global HEAD_PERIOD, BL_LOOK_GPS

    lowfirst = False

    # If first time in calling this function, set lookDir.
    if (Global.lostBall <= (BL_LOOK_GPS + 1)):

        lowfirst = True

        h = Global.gpsLocalBall.getHeading()
        if  ((h >= 0 and lowfirst) or \
             (h < 0  and not lowfirst)):
            lookDir = Constant.dCLOCKWISE
        else:
            lookDir = Constant.dANTICLOCKWISE

    headStep = Global.lostBall % (HEAD_PERIOD)
            
    # Every head period alternate between low first and high first
    # 1 == True, 0 == False.
    if (((Global.lostBall / HEAD_PERIOD) % 2) == lowfirst):

        # look down first
        panx = -90 + (180 / (HEAD_PERIOD)) * headStep
        tilty = tiltLow + 30
        cranez = -15

    else:
        # look up first
        panx = 90 - (180 / (HEAD_PERIOD)) * headStep
        tilty = tiltHigh
        cranez = 5

    # Set pan direction.
    if (lookDir == Constant.dCLOCKWISE):
        panx *= 1
    else:
        panx *= -1

    FWHead.fixHeadAt = (panx,tilty,cranez,Constant.HTAbs_h)

    # Usually only setting the FWHead is enough. However special case is that 
    # FWHead.DecideNextAction also uses headFindBall. So we should set finalAction here as well.
    HelpLong.setHeadParams(panx,tilty,cranez,Constant.HTAbs_h)

