#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: OdoLearner.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#
# TODO: 
#   fix StopAndLocalise timeout 30s
"""
Odometry learner. 
Step 1: Sample walking data: localise, walk for a while, then 
localise, get the difference..
Step 2: ???
"""

import HelpLong
import HelpTrack
import Global
import HelpShort
import Constant
import math
import WalkingSkills
import LabArchive
from StateBasedAction import StateBasedAction
import Debug


LOG_FILE_NAME = "/MS/odo.log" 

VERY_SMALL = 1e-6
counter = 1000000 #not doing until being reset()
testingTime = 10 #seconds

FP_FINDING_POSITION         = [0, 1, 0, 1, 0]
FP_FINDING_POSITION_LOST    = [1, 0, 0, 0, 1]
FP_DO_WALKING               = [2, 0, 2, 0, 2]
FP_COLLECTING_DATA          = [0, 3, 0, 3, 0]
FP_IDLE                     = [0, 0, 0, 0, 0]

theState = None
wirelessCommand = None
samplingForward = 0
samplingLeft = 0
samplingTurn = 0



def solve2LinesCrossEachOther( equation1, equation2 ):
    (a1,b1,c1) = equation1
    (a2,b2,c2) = equation2
    try:
        y = - float(a2*c1 - a1*c2) / (a2*b1 - a1*b2)
        x = float(c1 - b1*y) / a1
        return (x,y)
    except : #division by zero
        return (None,None)
        
#handle all cases
def getInterpolatedDeltaFLT(startX,startY,startH, endX,endY,endH,numberOfStep):
    startX,startY,startH, endX,endY,endH = map(float,(startX,startY,startH, endX,endY,endH))
    if abs(endH - startH) >= 360:
        return (None,None,None)
    elif abs(endH - startH) < VERY_SMALL:
        dt = 0
        dX,dY = (endX - startX) / numberOfStep,(endX - startX) / numberOfStep
        df,dl = HelpShort.rotate( (dX,dY), -startH)
        return (df,dl,dt)
    elif abs(endH - startH) > 180:
        diffH = HelpShort.normalizeAngle_180(endH - startH)
        complementStep = float(numberOfStep) * abs(diffH) / abs(endH - startH)
        df,dl,dt = getInterpolatedDeltaFLT_180(startX,startY,startH, endX,endY,startH + diffH,complementStep)
        return (-df,-dl,-dt)
    else:
        return getInterpolatedDeltaFLT_180(startX,startY,startH, endX,endY,endH,numberOfStep)

#handle all cases that 0 < |endH - startH|    < 180
def getInterpolatedDeltaFLT_180(startX,startY,startH, endX,endY,endH,numberOfStep):
    startX,startY,startH, endX,endY,endH = map(float,(startX,startY,startH, endX,endY,endH))
    ha = startH
    alpha = HelpShort.getHeadingBetween(startX,startY,endX,endY)
    hb = endH
    beta = 180 + alpha
    bigA = - (ha - alpha)
    bigB = - (hb - beta)

    t = HelpShort.normalizeAngle_180(bigA - bigB) / 2
    ta = alpha - t
    tb = alpha + t
    ta_r = HelpShort.DEG2RAD(ta)
    tb_r = HelpShort.DEG2RAD(tb)
    if abs(ta-tb) < VERY_SMALL:
        Xo, Yo = (startX + endX) / 2.0 , (startY + endY) / 2.0
    else:
        Xo, Yo = solve2LinesCrossEachOther( 
            ( -math.sin(ta_r), math.cos(ta_r), math.sin(ta_r) * startX - math.cos(ta_r) * startY),
            ( -math.sin(tb_r), math.cos(tb_r), math.sin(tb_r) * endX - math.cos(tb_r) * endY ) )

    radius = HelpShort.getDistanceBetween(startX,startY, Xo,Yo)
    totalArc = HelpShort.normalizeAngle_180(180 + 2*t)  # = 360 - (180 - 2t)
    eachStepArc = totalArc / numberOfStep
    deltaHeading = HelpShort.getHeadingBetween(startX,startY, Xo,Yo) + \
        eachStepArc + 180
    deltaX,deltaY = HelpShort.getPointRelative(Xo,Yo,deltaHeading, radius)
    deltaX -= startX
    deltaY -= startY
    deltaF,deltaL = HelpShort.rotate( (deltaX,deltaY), -startH)
    deltaT = (endH - startH) / numberOfStep
    #WARNING: endH - startH should be in (-360,360)
    #Test: more than one 360 circle
    return (deltaF,deltaL,deltaT)
    
def moveOneStep(startX,startY,startH,deltaF,deltaL,deltaT):
    x,y,h = startX,startY,startH
    x,y = HelpShort.getPointRelative(x,y,h,deltaF) #forward component
    x,y = HelpShort.getPointRelative(x,y,h+90,deltaL) #left component
    h += deltaT
    return (x,y,h)
    
def getFinalPosition(startX,startY,startH,deltaF,deltaL,deltaT, numberOfStep):
    x,y,h = startX,startY,startH
    for _ in range(numberOfStep):
        x,y,h = moveOneStep(x,y,h,deltaF,deltaL,deltaT)
    return (x,y,h)
     
class OdoLearnerStateBase(StateBasedAction):
    def __init__(self): #constructor
        StateBasedAction.__init__(self)
        self.counter = 0
        self.forward,self.left,self.turn = 0,0,0
    def setMovement(self,forward,left,turn):
        self.forward,self.left,self.turn = map(float,[forward,left,turn])
    def getMovement(self):
        return (self.forward,self.left,self.turn )
        
    def copyMovement(self,other):
        self.setMovement(other.forward,other.left,other.turn)

    
class FindingStartPosition(OdoLearnerStateBase):
    TIME_OUT = 1800
    
    def __init__(self): #constructor
        OdoLearnerStateBase.__init__(self)
        self.startX, self.startY, self.startH = None,None,None
        self.doGoToPosition = WalkingSkills.GoToPosition()
        
    def checkInRectangularBound(self,x,y,minX,minY, maxX,maxY):
        return x >= minX and x <= maxX and y >= minY and y <= maxY
        
    def getBoundaryPosition(self,df,dl,dt,totalFrames):
        x = 0
        y = 0
        h = 90
        Xmin = 0
        Ymin = 0
        Xmax = 0
        Ymax = 0
        for _ in range(totalFrames):
            x,y,h = moveOneStep(x,y,h,df,dl,dt)
            Xmin = min(Xmin,x)
            Xmax = max(Xmax,x)
            Ymin = min(Ymin,y)
            Ymax = max(Ymax,y)
        return (Xmin,Ymin,Xmax,Ymax)
            
    def calculateStartPosition(self):
        PG = 40.0
        df = self.forward / (PG * 2)
        dl = self.left / (PG * 2)
        dt = self.turn / (PG * 2)
        totalFrames = testingTime * 125        
        Xmin,Ymin,Xmax,Ymax = self.getBoundaryPosition(df,dl,dt,totalFrames)
        
        self.startX = Constant.FIELD_WIDTH / 2.0 - (Xmin + Xmax) / 2.0
        self.startY = Constant.FIELD_LENGTH / 2.0 - (Ymin + Ymax) / 2.0
        self.startH = 90
        print "Starting at %.2f %.2f %.2f" %(self.startX,self.startY,self.startH)
        
    def DoAction(self): 
        HelpTrack.stationaryLocalise()
        if self.startX is None:
            self.calculateStartPosition()
        if self.counter == 0:
            self.doGoToPosition.setTarget(self.startX,self.startY,self.startH)
        
        self.doGoToPosition.DecideNextAction()
        if self.doGoToPosition.isFinished() or self.counter > self.TIME_OUT:
            return firstLocalisation
        else:
            return self

WALKING_PERIOD = 30 * 5

class WalkingState(OdoLearnerStateBase):
    def __init__(self):
        OdoLearnerStateBase.__init__(self)
        self.headingSequence = []
        self.startX, self.startY, self.startH = None,None,None
        
    def beginAction(self): #should I get current location from FindingStartPosition or figure out myself?
        self.copyMovement(findingStartPosition)
        self.headingSequence = []
        if self.startX is None:
            self.startX, self.startY = Global.selfLoc.getPos()
            self.startH = Global.selfLoc.getHeading()
            
    def DoAction(self):
        HelpLong.showFacePattern(FP_DO_WALKING)
        HelpTrack.stationaryLocalise()
        self.headingSequence.append( Global.selfLoc.getHeading())
        #Change the walk here for different learning
        HelpLong.setNormalWalk(self.forward,self.left,self.turn)        
        if self.counter >= WALKING_PERIOD:
            HelpLong.setNormalWalk(0,0,0)
            HelpLong.setForceStepComplete()
            return secondLocalisation
        return self
        
    def calculateHeading(self):
        f = open("/MS/turning.log","w")
        for h in self.headingSequence:
            f.write("%f\n" % h)
        f.close()
        
        turn360 = 0
        for i in range(1,len(self.headingSequence) ):
            if self.headingSequence[i] - self.headingSequence[i-1] > 300:
                turn360 -= 360.0
            elif self.headingSequence[i] - self.headingSequence[i-1] < -300:
                turn360 += 360.0
                
        return turn360

class StopAndLocalise(OdoLearnerStateBase):
    TIME_OUT = 900
    def __init__(self,timeout = TIME_OUT):
        OdoLearnerStateBase.__init__(self)
        self.setTimeOut(timeout)
        self.endX, self.endY, self.endH = None,None,None
        
    def setTimeOut(self,timeout):
        self.timeout = timeout
        
    def DoAction(self):
        HelpLong.setNormalWalk(0,0,0)
        LabArchive.doMyActiveLocalise()
        if self.counter >= self.timeout:
            self.endX, self.endY = Global.selfLoc.getPos()
            self.endH = Global.selfLoc.getHeading()
            return self.outState
        else:
            return self
            
    def getLocation(self):
        return (self.endX, self.endY, self.endH)
        
class WriteResult(OdoLearnerStateBase):
    TIME_OUT = 300
    def __init__(self):
        OdoLearnerStateBase.__init__(self)

    def DoAction(self):
        HelpLong.showFacePattern(FP_COLLECTING_DATA)
        self.collectData()
        self.writeResultToLogFile()
        return self
        
    def collectData(self):
        self.forward, self.left, self.turn = walkingState.getMovement()
        self.startX, self.startY, self.startH = firstLocalisation.getLocation()
        self.endX, self.endY, self.endH = secondLocalisation.getLocation()
        self.roundTurningMade = walkingState.calculateHeading()
        
    def writeResultToLogFile(self):
        try:
            f = open(LOG_FILE_NAME,"a")
        except IOError:
            f = open(LOG_FILE_NAME,"w")
        
        logResult = ("F,L,T = %d %d %d  Start Pos = %d %d %d , End Pos = %d %d %d\
            turning made = %d\n" % \
            (self.forward, self.left, self.turn, 
            self.startX, self.startY, self.startH, 
            self.endX, self.endY, self.endH, 
            self.roundTurningMade + self.endH - self.startH)
            )
        f.write(logResult)
        f.close()
        
        Debug.showDebug(logResult)
        res = getInterpolatedDeltaFLT(self.startX, 
            self.startY, self.startH, 
            self.endX, self.endY, 
            self.endH,
            int(WALKING_PERIOD * 125/30.0) )
        Debug.showDebug(str(res),True)
        
        
findingStartPosition = FindingStartPosition()
firstLocalisation = StopAndLocalise(150)
walkingState = WalkingState()
secondLocalisation = StopAndLocalise(150)
writeResult = WriteResult()

firstLocalisation.setOutState(walkingState)
secondLocalisation.setOutState(writeResult)

def randomFloat(fromFloat,toFloat):
    pass
    
def generateRandomWalk():
    forward = randomFloat(-7,7)
    if abs(forward) < 3:
        left = randomFloat(-6,6)
    else:
        left = randomFloat(-5,5)
    turn = randomFloat(-30,30)
    return (forward,left,turn)
    
def processCommand(cmd):
    global theState, samplingForward,samplingLeft,samplingTurn
    global wirelessCommand 
    wirelessCommand = None
    cmd = cmd.strip()
    print "Odometry got ", cmd
    if cmd[:5] == "start":
        samplingForward, samplingLeft, samplingTurn = map(int,cmd[5:].split())
        theState = findingStartPosition
        findingStartPosition.setMovement(samplingForward,samplingLeft,samplingTurn)
    if cmd[:5] == "random":
        samplingForward, samplingLeft, samplingTurn = generateRandomWalk()
        theState = findingStartPosition
    elif cmd == "sloc":
        wirelessCommand = HelpTrack.stationaryLocalise


#
# wireless command -> findingStartPosition -> walkingState -> localiseAndResult 
# -> findingStartPosition...
#

def DecideNextAction():
    global wirelessCommand,theState
    Global.frameReset()
    
    if theState is not None:
        theState = theState.DecideNextAction()
        
    if wirelessCommand is not None:
        wirelessCommand()
