#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: PawKick.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
#   Non-directional paw kick. Edited by Ted Wong.
#===============================================================================

import math
import Global
import Constant
import HelpLong
import HelpShort
import HelpWhere

#=========================
# Constants for paw kick module.
#=========================

FIRE_PAWKICK_LEFT    = 0
FIRE_PAWKICK_RIGHT   = 1
FIRE_PAWKICK_AUTO    = 2
firePaw              = FIRE_PAWKICK_LEFT

#=======================================================================    
# This function returns a list with two elements. The first element is True or False, representing
# whether it is ok for the robot to do the paw kick. The second element is which paw the robot
# should kick. Notice that the second element only makes sense if the first element is True.
# In another words, don't process the second element if the first element is False.

# Paw kick should be used on edges.
# Paw kick shouldn't be used when the ball is very close and specially it is under the chin
# (isUNSW2004BallUnderChin() in 2003) even the ball and the robot are both near the edge.

def IsPawKickOk():

    global FIRE_PAWKICK_LEFT, FIRE_PAWKICK_RIGHT, FIRE_PAWKICK_AUTO

    debug = False
    
    # Don't do paw kick if you can't physically see the ball.
    if Global.ballSource != Constant.VISION_BALL:
        if debug:
            print "Come on, paw kick is only used for vision ball, at lease for the moment!\n"
        return [False, FIRE_PAWKICK_AUTO]

    # Ted: Position variance in the following cases have been disabled. Come back later.

    # Paw kick left edge case.
    # 1.) If ball is near the left edge.
    # 2.) If the robot is near the left edge.
    # 3.) If the robot heading is toward the target goal.
    # 4.) If the ball heading relative to the robot is not much (fewer turning).
    if (HelpWhere.ballOnLEdge() 
            and HelpWhere.isOnLEdge() 
            and HelpWhere.isInRange(60,120) 
            and ( (abs(Global.ballH) < 30 and Global.ballD > 30) or 
                  (abs(Global.ballH) < 40 and Global.ballD <= 30) )
        ):
        if debug:
            print "Hey, I am doing the left edge paw kick now!\n"    
        return [True, FIRE_PAWKICK_LEFT]

    # Paw kick right edge case.
    # 1.) If ball is near the right edge.
    # 2.) If the robot is near the right edge.
    # 3.) If the robot heading is toward the target goal.
    # 4.) If the ball heading relative to the robot is not much (fewer turning).
    if (HelpWhere.ballOnREdge() 
            and HelpWhere.isOnREdge() 
            and HelpWhere.isInRange(60,120) 
            and ( (abs(Global.ballH) < 30 and Global.ballD > 30) or 
                  (abs(Global.ballH) < 40 and Global.ballD <= 30) )
        ):
        if debug:
            print "Hey, I am doing the right edge paw kick now!\n"
        return [True, FIRE_PAWKICK_RIGHT]
    
    # Paw kick top edge right half.
    # 1.) If ball is on the top right edge.
    # 2.) If the robot is on the top right edge.
    # 3.) If the robot heading is toward the target goal.
    # 4.) If the ball heading relative to the robot is not much (fewer turning).
    # 5.) If the robot is right of the goal box.
    # 6.) If the ball is right of the goal box.    
    # 7.) If the ball heading is not large.
    if (HelpWhere.ballOnTREdge()
        and HelpWhere.isOnTREdge() 
        and HelpWhere.isInRange(150,190)
        and ( (abs(Global.ballH) < 30 and Global.ballD > 30) or 
              (abs(Global.ballH) < 40 and Global.ballD <= 30) )
        ):
        if debug:
            print "Paw kick on the top edge right half!\n"
        return [True, FIRE_PAWKICK_AUTO]

    # Paw kick top edge left half.
    if (HelpWhere.ballOnTLEdge() 
            and HelpWhere.isOnTLEdge()
            and not HelpWhere.isInRange(30,350) 
            and ( (abs(Global.ballH) < 30 and Global.ballD > 30) or 
                  (abs(Global.ballH) < 40 and Global.ballD <= 30) )
        ):
        if debug:
            print "Hey, I am doing the top edge left paw kick now!\n"
        return [True, FIRE_PAWKICK_AUTO]
                
    # Paw kick bottom left half
    if (HelpWhere.ballOnBEdge() 
            and HelpWhere.isOnBEdge()
            and HelpWhere.isInRange(150,190)
            and Global.selfLoc.getPos()[0] < Constant.FIELD_WIDTH / 2
            and Global.ballX < Constant.LEFT_GOALBOX_EDGE_X
            and ( (abs(Global.ballH) < 30 and Global.ballD > 30) or 
                  (abs(Global.ballH) < 40 and Global.ballD <= 30) )
        ):
        if debug:
            print "Hey, I am doing the bottom right edge paw kick now!\n"
        return [True, FIRE_PAWKICK_AUTO]
        
    # If Paw kick bottom right right.
    if (HelpWhere.ballOnBEdge()
            and HelpWhere.isOnBEdge()
            and not HelpWhere.isInRange(30,350)
            and Global.selfLoc.getPos()[0] > Constant.FIELD_LENGTH / 2
            and Global.ballX > Constant.RIGHT_GOALBOX_EDGE_X
            and ( (abs(Global.ballH) < 30 and Global.ballD > 30) or 
                  (abs(Global.ballH) < 40 and Global.ballD <= 30) )
        ):
        if debug:
            print "Hey, I am doing the bottom right edge paw kick now!\n"
        return [True, FIRE_PAWKICK_AUTO]
        
    return [False, FIRE_PAWKICK_AUTO]


#=======================================================================    
def firePawKick(paw = FIRE_PAWKICK_AUTO):

    global firePaw, FIRE_PAWKICK_LEFT, FIRE_PAWKICK_RIGHT, FIRE_PAWKICK_AUTO

    if (Global.ballSource != Constant.VISION_BALL):
        return None
    
    HelpLong.showFacePattern(Constant.FP_FIRE_PAW_KICK)
    
    FIRE_PAWKICK_YOFFSET        = Constant.BallRadius
    FIRE_PAWKICK_OFFSET         = 7.0

    ballx = math.sin (HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
    bally = math.cos (HelpShort.DEG2RAD(Global.ballH)) * Global.ballD
    
    # The 2003 offset is 6 and it is not too bad.
    
    # Decide on a paw. Once decided stick with it until ball rolls past the other paw
    # Try to avoid problen with switching between left and right paws.
    
    if paw == FIRE_PAWKICK_AUTO:
        if firePaw == FIRE_PAWKICK_LEFT and ballx < -FIRE_PAWKICK_OFFSET / 2:
            firePaw = FIRE_PAWKICK_RIGHT
        elif firePaw == FIRE_PAWKICK_RIGHT and ballx > FIRE_PAWKICK_OFFSET / 2:
            firePaw = FIRE_PAWKICK_LEFT
    else:
        firePaw = paw

    if firePaw == FIRE_PAWKICK_RIGHT:
        target_x = ballx + FIRE_PAWKICK_OFFSET
    else:
        target_x = ballx - FIRE_PAWKICK_OFFSET

    target_y = bally - FIRE_PAWKICK_YOFFSET
    target_d = math.sqrt(HelpShort.SQUARE(target_x) + HelpShort.SQUARE(target_y))
    target_h = HelpShort.RAD2DEG(math.asin(target_x / target_d))
    
    if abs(target_h) > 40 and target_x > 10:#need large turning, use NormalWalk
        
        HelpLong.setNormalWalk(0,0,HelpShort.CLIP(target_h,60))
    else:
        
        # Run at ball
        turnCCW = HelpShort.CLIP(target_h / 3, 30.0)
    
        # Once the ball is lined up and well within range for the forward, the robot should
        # commit to charge at the ball at full speed.
        
        if abs(target_x) <= 1.0 and target_d < Constant.BallRadius * 1.5:
            turnCCW = HelpShort.CLIP(turnCCW, 10.0)
        
        HelpLong.setEllipticalWalk(0,0, turnCCW)
