#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: ReadJointPlayer.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#   ReadJointPlayer.py
#
#  Tool to read joint values, mostly when the dog is relaxed
#
#===============================================================================


"""
Tool to read joint values, mostly when the dog is relaxed.
To read joint values, simply set bPrintJointSensor to True 
using SRC.

Can also be used to experiment with indicators, get/set joint values...
"""
import Global
import VisionLink
import Constant
import HelpLong

bInitGlobal = True
bPrintJointSensor = False
bForceStepComplete = False
bSendRobotBlobArea = False
bEnablePersistant  = True
bDoItOnce          = False
myFinalAction = list(Constant.NormalWalkStandStill)

#persistent command: set head, set joint ( finalAction is set every frame)
#non-persistent command : set indicators (finalIndicators only set on new command)
def doPersistentCommand():
##~     HelpLong.setHeadParams(pan,tilt,crane,Constant.HTAbs_h)
    Global.finalAction = myFinalAction[:]
    if bForceStepComplete:
            HelpLong.setForceStepComplete()
            
def doSendingRobotBlobInfo():
    if bSendRobotBlobArea:
##~         for vOpp in Global.vOpponents:
##~             if vOpp.getConfidence() > 0:
##~                 VisionLink.sendMessageToBase("----"+vOpp.getHeight())
##~                 break
        VisionLink.sendMessageToBase("----"+str(VisionLink.getTestingInfo()))
        
def DecideNextAction():
    global bEnablePersistant,bDoItOnce
    if bInitGlobal:
        Global.frameReset()
    if bPrintJointSensor:
        values = ""
        for i in range(15):
            values += "%6.2f  " % VisionLink.getJointSensor(i+1) 
        print values
    if bDoItOnce:
        doPersistentCommand()
        HelpLong.setForceStepComplete()
        bDoItOnce         = False
        bEnablePersistant = False
    elif bEnablePersistant:
        doPersistentCommand()
        
    doSendingRobotBlobInfo()
    
pan,tilt,crane = 0,0,0

def setHeadParams(p,t,c):
    global pan,tilt,crane
    pan,tilt,crane = p,t,c
    HelpLong.setHeadParams(pan,tilt,crane,Constant.HTAbs_h)
    
def processCommand(cmd):
    global bForceStepComplete,myFinalAction, \
        bSendRobotBlobArea,bDoItOnce, bEnablePersistant
    print "ReadJointPlayer got |%s|"%cmd
    cmd = cmd.strip()
    if cmd[:3] == "ind":
        print "Indicator command",cmd[3:]
        indicator,value = cmd[3:].split()
        indicator = int(indicator)
        value = int(value)
        Global.finalIndicators[indicator] = value
    elif cmd[:3] == "sen":
        print "Sensor command ", cmd[3:]
        indicator = int(cmd[3:].split()[0])
        print VisionLink.getAnySensor(indicator)
    elif cmd[:4] == "head":
        bEnablePersistant = True
        print "Head command ", cmd[4:]
        p,t,c = map(float,cmd[4:].split())
        setHeadParams(p,t,c)
        myFinalAction = Global.finalAction[:]
    elif cmd[:4] == "walk":
        bEnablePersistant = True
        f,l,t,wt = map(float,cmd[4:].split())
        wt = int(wt)
        print "Walk command ", cmd[4:], " wt = ",wt
        if wt == Constant.EllipseWalkWT:
            HelpLong.setEllipticalWalk(f,l,t)
        elif wt == Constant.NormalWalkWT:
            HelpLong.setNormalWalk(f,l,t)
        else:
            Global.finalAction[Constant.AAWalktype] = wt
            HelpLong.setForwardLeftTurn(f,l,t)
        
        myFinalAction = Global.finalAction[:]
    elif cmd[:4] == "once":
        bDoItOnce = True
        f,l,t,wt = map(float,cmd[4:].split())
        wt = int(wt)
        print "Walk command[once] ", cmd[4:], " wt = ",wt
        if wt == Constant.EllipseWalkWT:
            HelpLong.setEllipticalWalk(f,l,t)
        elif wt == Constant.NormalWalkWT:
            HelpLong.setNormalWalk(f,l,t)
        else:
            Global.finalAction[Constant.AAWalktype] = wt
            HelpLong.setForwardLeftTurn(f,l,t)
        
        myFinalAction = Global.finalAction[:]
    elif cmd[:3] == "cpl":
        print "Force Step Complete command"
        bForceStepComplete = True
    elif cmd[:4] == "area": 
        bSendRobotBlobArea = not int(cmd[4:])
