#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: StealthDog.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# StealthDog - edited by Ted Wong.
#===============================================================================

"""
Stealth Dog.
"""

import Debug
import Global
import Constant
import HelpLong
import HelpShort
import HelpWhere

# Fire stealth backoff left component.
leftComp = None    
    
# Stealth dog. Return False if the stealthing is not trigger, else return True.
def stealthDog(mustSeeBall):
    if mustSeeBall and not HelpShort.canSeeBall():
        if Debug.stealthTrigger:
            print "False because I can't see the ball!"
        return False

    if mustSeeBall and Global.ballD < 30:
        if Debug.stealthTrigger:
            print "False - I can see the ball, but the ball is too far away!"
        return False
            
    knowWhereIam = (Global.selfLoc.getPosVar() < HelpShort.get95cf(75))    
    onLeft       = HelpWhere.isOnLEdge() and knowWhereIam and HelpWhere.isInRange (60,300)
    onRight      = HelpWhere.isOnREdge() and knowWhereIam and not HelpWhere.isInRange (120,240)
    onTop        = HelpWhere.isOnTEdge() and knowWhereIam and not HelpWhere.isInRange (210,330)
    onBottom     = HelpWhere.isOnBEdge() and knowWhereIam and not HelpWhere.isInRange (30,150)

    if onLeft or onRight or onTop or onBottom:
        if Debug.stealthTrigger:
            print "False because of edge problems!"
        return False
        
    index1, index2  = [-1, -1]
    
    for i in range(Constant.NUM_TEAM_MEMBER):
        opponent = Global.vOpponents[i]
    
        if opponent.getConfidence() == 0:
            continue
        
        # If I am closer to the ball why not go straight?
        if mustSeeBall and opponent.getDistance() >= Global.ballD:
            if Debug.stealthTrigger:
                print "I can see the ball and I am closer to the ball. Why not go straight?"
            continue
        
##~         # Opponent too close to the ball. Check if vOpponents is also global.
##~         if math.sqrt(HelpShort.SQUARE(Global.vOpponents[i].getX() - Global.ballX) + \
##~                      HelpShort.SQUARE(Global.vOpponents[i].getY() - Global.ballY) < 20):
##~             if Debug.stealthTrigger:
##~                 print "Opponent too close to the ball!"
##~             continue
            
        # Wrong distance to start avoiding. 90.
        if opponent.getDistance() >= 110:
            if Debug.stealthTrigger:
                print "Wrong distance to start avoiding!"
            continue
            
        # Shoulder charge.
        if opponent.getDistance() < 40 and \
           abs(opponent.getHeading()) > 30:
            if Debug.stealthTrigger:
                print "Opponent at the side!"
            continue
            
        if index1 == -1 or Global.vOpponents[i].getHeading() > \
           Global.vOpponents[index1].getHeading():
            index1 = i
        
        if index2 == -1 or Global.vOpponents[i].getHeading() < \
           Global.vOpponents[index2].getHeading():
            index2 = i
            
    if index1 == -1:
        if Debug.stealthTrigger:
            print "Index1 == -1!"
        return False
        
    Opp_left   = Global.vOpponents[index1]
    Opp_right  = Global.vOpponents[index2]
    leftAngle  = 0
    rightAngle = 0
        
    if HelpShort.getDistanceBetween (Opp_right.getX(),Opp_right.getY(),Opp_left.getX(),Opp_left.getY()) < 20:
        leftAngle  = Global.vOpponents[index1].getHeading()
        rightAngle = Global.vOpponents[index2].getHeading()
    else:        
        if Opp_left.getDistance() < Opp_right.getDistance():
            leftAngle  = Opp_left.getHeading()
            rightAngle = Opp_left.getHeading()
        else:
            leftAngle  = Opp_right.getHeading()
            rightAngle = Opp_right.getHeading()
            
    if Debug.stealthTrigger:
        print "Yeah, stealth dog is on!"

    stealthAction(mustSeeBall, leftAngle, rightAngle)        
    return True

def stealthAction(mustSeeBall, leftAngle, rightAngle):
    turnCCW       = None        
    stealthTurn   = None
    Global.myRole = Constant.STEALTHDOG
    
    # The average of the two headings.
    averageHead = (leftAngle + rightAngle) / 2

    if Debug.stealthDebug:
        print "Ball heading: %f"  %Global.ballH
        print "Robot heading: %f" %averageHead

    if mustSeeBall:
        # If the ball heading is smaller than the robot heading, we stealth the right (anti-clockwise) direction.
        if Global.ballH < averageHead:
            stealthTurn = rightAngle - Constant.STEALTHDOG_OFFSET_ANGLE            
            if stealthTurn < Global.ballH:
                if Debug.stealthDebug:
                    print "Anti-clockwise but smaller than ball heading case!"
                turnCCW = stealthTurn
            else:
                if Debug.stealthDebug:
                    print "Anti-clockwise but greater than ball heading case!"
                turnCCW = Global.ballH
                    
        # Here the direction is left (clockwise) direction.
        else:
            stealthTurn = leftAngle + Constant.STEALTHDOG_OFFSET_ANGLE
            if stealthTurn > Global.ballH:
                if Debug.stealthDebug:
                    print "Clockwise but smaller than ball heading case!"
                turnCCW = stealthTurn
            else:
                if Debug.stealthDebug:
                    print "Clockwise but greater than ball heading case!"
                turnCCW = Global.ballH
         
    # Try this one.
    else:    
        if leftAngle == rightAngle:
            if leftAngle <= 0:
                turnCCW = leftAngle + Constant.STEALTHDOG_OFFSET_ANGLE
                HelpLong.setHeadParams(-45, 0, 0, Constant.HTAbs_pt)                
            else:
                turnCCW = rightAngle - Constant.STEALTHDOG_OFFSET_ANGLE
                HelpLong.setHeadParams(45, 0, 0, Constant.HTAbs_pt)
                
        elif leftAngle <= 0 and rightAngle <= 0:
            turnCCW = leftAngle + Constant.STEALTHDOG_OFFSET_ANGLE
            HelpLong.setHeadParams(-45, 0, 0, Constant.HTAbs_pt)                
        elif leftAngle > 0 and rightAngle > 0:
            turnCCW = rightAngle - Constant.STEALTHDOG_OFFSET_ANGLE
            HelpLong.setHeadParams(45, 0, 0, Constant.HTAbs_pt)
         
        # leftAngle and rightAngle are different signs.
        else:
            # Want to go straight into the middle?
            if abs(leftAngle) <= abs(rightAngle):
                turnCCW = leftAngle + Constant.STEALTHDOG_OFFSET_ANGLE
                HelpLong.setHeadParams(-45, 0, 0, Constant.HTAbs_pt)                                
            else:
                turnCCW = rightAngle - Constant.STEALTHDOG_OFFSET_ANGLE
                HelpLong.setHeadParams(45, 0, 0, Constant.HTAbs_pt)

##~     # New stealth dog added in 2004.
##~     else:
##~         # Opponent is left side, so should stealth right side.
##~         if averageHead >= 0:
##~             if Debug.stealthDebug:
##~                 print ">= 0 case!"
##~             turnCCW = rightAngle - Constant.STEALTHDOG_OFFSET_ANGLE
##~             HelpLong.setHeadParams(45, 0, 0, Constant.HTAbs_pt)
##~             
##~         # Opponent is right side, so should stealth left side.
##~         else:
##~             if Debug.stealthDebug:
##~                 print "< 0 case!"
##~             turnCCW = leftAngle + Constant.STEALTHDOG_OFFSET_ANGLE
##~             HelpLong.setHeadParams(-45, 0, 0, Constant.HTAbs_pt)
                
    if Debug.stealthDebug:
        print "TurnCCW: %f" %turnCCW 

    HelpLong.setEllipticalWalk(7, 0, HelpShort.CLIP(turnCCW / 2, 30))

################################################## 
##  Fire stealth dog.
##################################################
def fireStealthDog():
    knowWhereIam = (Global.selfLoc.getPosVar() < HelpShort.get95cf(75))    
    onLeft       = HelpWhere.isOnLEdge() and knowWhereIam and HelpWhere.isInRange (60,300)
    onRight      = HelpWhere.isOnREdge() and knowWhereIam and not HelpWhere.isInRange (120,240)
    onTop        = HelpWhere.isOnTEdge() and knowWhereIam and not HelpWhere.isInRange (210,330)
    onBottom     = HelpWhere.isOnBEdge() and knowWhereIam and not HelpWhere.isInRange (30,150)

    if onLeft or onRight or onTop or onBottom:
        if Debug.stealthTrigger:
            print "False because of edge problems!"
        return False
        
    index1, index2  = [-1, -1]
    
    for i in range(Constant.NUM_TEAM_MEMBER):
        opponent = Global.vOpponents[i]
    
        if opponent.getConfidence() == 0:
            continue
        
        if opponent.getDistance() > 30:
            if Debug.stealthTrigger:
                print "Wrong distance to start avoiding!"
            continue
            
            # Shoulder charge.
            if abs(opponent.getHeading()) > 70:
                if Debug.stealthTrigger:
                    print "Opponent at the side!"
                continue
                
        if index1 == -1 or Global.vOpponents[i].getHeading() > \
           Global.vOpponents[index1].getHeading():
            index1 = i
        
        if index2 == -1 or Global.vOpponents[i].getHeading() < \
           Global.vOpponents[index2].getHeading():
            index2 = i
            
    if index1 == -1:
        if Debug.stealthTrigger:
            print "Index1 == -1!"
        return False
        
    Opp_left   = Global.vOpponents[index1]
    Opp_right  = Global.vOpponents[index2]
    leftAngle  = 0
    rightAngle = 0
        
    if HelpShort.getDistanceBetween (Opp_right.getX(),Opp_right.getY(),Opp_left.getX(),Opp_left.getY()) < 20:
        leftAngle  = Global.vOpponents[index1].getHeading()
        rightAngle = Global.vOpponents[index2].getHeading()
    else:        
        if Opp_left.getDistance() < Opp_right.getDistance():
            leftAngle  = Opp_left.getHeading()
            rightAngle = Opp_left.getHeading()
        else:
            leftAngle  = Opp_right.getHeading()
            rightAngle = Opp_right.getHeading()
            
    if Debug.stealthTrigger:
        print "Yeah, stealth dog is on!"

    fireStealthAction(leftAngle, rightAngle)        
    return True

def fireStealthAction(leftAngle, rightAngle):
    global leftComp
    left          = None        
    Global.myRole = Constant.STEALTHDOG
        
    if leftAngle == rightAngle:
        if leftAngle <= 0:
            left = 15
            HelpLong.setHeadParams(-45, 0, 0, Constant.HTAbs_pt)                
        else:
            left = -15
            HelpLong.setHeadParams(45, 0, 0, Constant.HTAbs_pt)
                
    elif leftAngle <= 0 and rightAngle <= 0:
        left = 15
        HelpLong.setHeadParams(-45, 0, 0, Constant.HTAbs_pt)                
    elif leftAngle > 0 and rightAngle > 0:
        left = -15
        HelpLong.setHeadParams(45, 0, 0, Constant.HTAbs_pt)
         
    # leftAngle and rightAngle are different signs.
    else:
        # Want to go straight into the middle?
        if abs(leftAngle) <= abs(rightAngle):
            left = 15
            HelpLong.setHeadParams(-45, 0, 0, Constant.HTAbs_pt)                                
        else:
            left = -15
            HelpLong.setHeadParams(45, 0, 0, Constant.HTAbs_pt)

    leftComp = left
    HelpLong.setNormalWalk(0, left, 0)
