#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: VOAK.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# CoolDemo.py
#
# Demonstrate UPennKick
#
#===============================================================================
 

"""
VOAK
"""

import Global
import HelpLong
import HelpShort
import HelpTrack
##~ import VisionLink
import OpenChallenger
import Constant
from StateBasedAction import *
import math
import ActiveLocalise


FP_TURNING_TO_VOAK  = [1,0,0,0,0]
FP_SIT_DOWN         = [1,2,0,0,0]
FP_LOOKING_UP       = [1,2,1,0,0]
FP_TURN_KICK        = [2,2,0,0,0]
FP_HANDKICKER       = [1,0,1,0,1]

myPos = (0.0,0.0)
myHeading = 0
myBallDistance = 0
localBallX = 0
ballPos = (0.0,0.0)
goalPos = (Constant.FIELD_WIDTH/2,Constant.FIELD_LENGTH)
##~ headingToGapCenter = 0
lastTurnDir = None
headingToGoal = 0 #hack global
seenTheGoal = False

def resetEachFrame():
    global myBallDistance,myPos,ballPos,\
        myHeading,\
        localBallX,headingToGapCenter
    myBallDistance = Global.vBall.getVobDistance()
    
    myPos = Global.selfLoc.getPos()
    myHeading = Global.selfLoc.getHeading()
    ballPos = (Global.ballX,Global.ballY)
    
    localBallX = Global.vBall.getVobDistance() * \
        math.sin(HelpShort.DEG2RAD(Global.vBall.getHeading()))
        
##~     minH,maxH = VisionLink.getHeadingToBestGap()
##~     headingToGapCenter = (minH + maxH) / 2
##~     
def RandomKick():
    if Global.frame % 2 == 0:
        return Constant.HandKickLeftWT
    else:
        return Constant.HandKickRightWT

class MyActiveLocalise(StateBasedAction):
    ACTIVE_LOCALISE_INTERVAL = 60
    DOING_ACTIVE_LOCALISE = 10
    CLOSE_DISTANCE = 60
    CLOSE_DISTANCE_SLOW_BALL = 20
    
    def __init__(self):
        StateBasedAction.__init__(self)
        self.activeLocaliseCounter = self.ACTIVE_LOCALISE_INTERVAL
        
    
        
    def finishAction(self): #do reseting counter
        self.activeLocaliseCounter = self.ACTIVE_LOCALISE_INTERVAL
        
    def DoAction(self):
        if (Global.lostBall < 5):
            HelpTrack.trackVisualBall()
        elif Global.lostBall <= 10:
            HelpLong.setHeadParams(0,-20,-20,Constant.HTAbs_h)
        elif (Global.lostBall <= 30
                and abs(Global.gpsLocalBall.getHeading()) < 90): 
            HelpTrack.trackGpsBall()
        else:
            HelpTrack.stationaryLocalise()
                
        #decreasing counter
        self.activeLocaliseCounter -= 1
        if self.activeLocaliseCounter < - self.DOING_ACTIVE_LOCALISE:
            return None #stop action, finishAction called()
        elif self.activeLocaliseCounter < 0:
           
            #ball too close, should not active localise
            if Global.lostBall == 0:
                if Global.ballD < MyActiveLocalise.CLOSE_DISTANCE:
                    return self 

            ActiveLocalise.SmartSetBeacon(90,500)
            ActiveLocalise.DecideNextAction(False)
            
        return self
          
STUCK_THRESHOLD = 10000
GRABBING_MOUTH_OPEN = OpenChallenger.GRABBING_MOUTH_OPEN
class KickTurning(StateBasedAction):
    TURNING_LEFT_720 = 7.2
    TURNING_RIGHT_720 = 7.2
    CAN_CHECK_CHIN_TIME = 18
    
    def __init__(self):
        StateBasedAction.__init__(self)
        self.desiredHeading = 0
        self.turningTime = 0
        self.walkParams = None
        self.turnDir = None
        
    def calculateDesiredHeading(self):
        self.desiredHeading = 90
##~         print "My Pos = ",myPos
##~         print "desiredHeading=",self.desiredHeading

        
    def calculateTurningTime(self):
        turnNeeded = HelpShort.normalizeAngle_180(self.desiredHeading - myHeading)
        print "Turn needed = ",turnNeeded," my H = ", myHeading

        if turnNeeded > 0:
            calibrate = self.TURNING_LEFT_720
            self.walkParams = list(Constant.TurnHoldingCounterClockWise)
            self.turnDir = Constant.dANTICLOCKWISE
        else:
            calibrate = self.TURNING_RIGHT_720
            self.walkParams = list(Constant.TurnHoldingClockWise)
            self.turnDir = Constant.dCLOCKWISE
            
        if turnNeeded == 0:
            self.turningTime = 0
        else:
            self.turningTime = calibrate * 30 / (720.0/abs(turnNeeded))
        print "Turning time = ",self.turningTime
        
    def beginAction(self):
        self.calculateDesiredHeading()
        self.calculateTurningTime()
        HelpLong.openMouth(-350000)
        
    def finishAction(self):
        HelpLong.openMouth(Constant.IND_MOUTH_CLOSED)
        
        
    def DoAction(self):
        HelpLong.showFacePattern(FP_TURNING_TO_VOAK)
        Global.finalAction = self.walkParams[:]
##~         if self.counter >= self.CAN_CHECK_CHIN_TIME and \
##~             VisionLink.getAnySensor(Constant.ssMOUTH) - GRABBING_MOUTH_OPEN < STUCK_THRESHOLD:
##~                 return None
                
##~         print "Turning "
        if self.counter >= self.turningTime:
            return self.outState
        else:
            return self
 
class TurnToVOAK(KickTurning):
    def __init__(self):
        KickTurning.__init__(self)
        
    def calculateDesiredHeading(self):
        self.desiredHeading = HelpShort.getHeadingBetween(*(myPos + goalPos ))
##~         Global.selfLoc.getHeading() + headingToGapCenter
##~         print "headingToGapCenter = ", headingToGapCenter
##~         print "desiredHeading = " , self.desiredHeading
        self.theKick = RandomKick()
    
TURNING_KICK_OFFSET = 20
class HoldTurning(KickTurning):
    def __init__(self):
        KickTurning.__init__(self)
    
    def calculateDesiredHeading(self):
        self.desiredHeading = HelpShort.getHeadingBetween(*(myPos + goalPos ))
        turnNeeded = HelpShort.normalizeAngle_180(myHeading - self.desiredHeading)
        if abs(turnNeeded) < TURNING_KICK_OFFSET:
            self.desiredHeading = myHeading
        else:
            if turnNeeded > 0:
                self.desiredHeading += TURNING_KICK_OFFSET
            else:
                self.desiredHeading -= TURNING_KICK_OFFSET
        
SIT_DOWN_TIMER = 5
class SitDown(StateBasedAction):
    
    def __init__(self):
        StateBasedAction.__init__(self)
        
    def DoAction(self):
        HelpLong.showFacePattern( FP_SIT_DOWN )
        Global.finalAction[Constant.AAWalktype:] = Constant.BackWalkSitDownHoldingBall[Constant.AAWalktype:]
        HelpLong.setHeadParams(0,-60,40,Constant.HTAbs_h)
        if self.counter >= SIT_DOWN_TIMER:
            return self.outState
        else:
            return self
       
TIMER_TO_PAN = 10

class LookingAtGoal(StateBasedAction):
    def __init__(self):
        StateBasedAction.__init__(self)
        self.pan = 0
        
        
##~     def beginAction(self):
##~         self.pan = HelpShort.extendToLimit(headingToGapCenter, 90)
    def setPan(self,pan = 90):
        self.pan = pan
        
    def beginAction(self):
        global seenTheGoal
        seenTheGoal = False
        
    def DoAction(self): 
        global headingToGoal,seenTheGoal
        HelpLong.showFacePattern(FP_LOOKING_UP)
        Global.finalAction[Constant.AAWalktype:] = Constant.BackWalkSitDownHoldingBall[Constant.AAWalktype:]
        if Global.vTGoal.getConfidence() > 0:
            print "Yeah, see goal", Global.vTGoal.getHeading(), Global.vTGoal.getDistance()
            headingToGoal = Global.vTGoal.getHeading()
            HelpLong.showRedEye(True)
            seenTheGoal = True
            
        if self.counter < TIMER_TO_PAN:
            HelpLong.setHeadParams(-45, 0,10,Constant.HTAbs_h)
            return self
        elif self.counter < 2 * TIMER_TO_PAN:
            HelpLong.setHeadParams(45, 0,10,Constant.HTAbs_h)
            return self
        else:
            return self.outState

class SmartHandKick(StateBasedAction):
    def __init__(self):
        StateBasedAction.__init__(self)
        self.headingToGoal = 0
        
    def finishAction(self):
        HelpLong.showRedEye(False)
        
    def DoAction(self):
        if seenTheGoal:
            if abs(headingToGoal) > 40:
                Global.finalAction = list(Constant.TurnHoldingCounterClockWise)
                HelpLong.setHeadParams(0,-60,50,Constant.HTAbs_h)
                if self.counter >= 15:
                    if headingToGoal < -40:
                        Global.finalAction[Constant.AAWalktype] = Constant.UpennLeftWT
                    elif headingToGoal > 40:
                        Global.finalAction[Constant.AAWalktype] = Constant.UpennRightWT
                    HelpLong.setForceStepComplete()
                    return self.outState
                else:
                    return self
            else:
                if headingToGoal <= 10:
                    Global.finalAction[Constant.AAWalktype] = Constant.HandKickLeftWT
                else:
                    Global.finalAction[Constant.AAWalktype] = Constant.HandKickRightWT
                HelpLong.setForceStepComplete()
                return self.outState
        else:
            if Global.lostBall == 0:
                HelpTrack.trackVisualBall()
            else:
                HelpLong.setHeadParams(0,-20,-20,Constant.HTAbs_h)
            return self.outState
            
class HandKicker(StateBasedAction):
    APPROACHING_DISTANCE = 25
    def __init__(self):
        StateBasedAction.__init__(self)
        self.runToAndGrabBall = OpenChallenger.RunToGrabBall()
        self.turnToVoak = TurnToVOAK()
        self.sitDown = SitDown()
        self.lookingAtGoal  = LookingAtGoal()
        self.kickTheBall = SmartHandKick()

        
        self.doLocateBall = OpenChallenger.MyLocateBall()
        self.doActiveLocalise = MyActiveLocalise()
        
        self.doLocateBall.setOutState(self.runToAndGrabBall)
        
        self.runToAndGrabBall.setOutState(self.turnToVoak)
        self.turnToVoak.setOutState(self.sitDown)
        self.sitDown.setOutState(self.lookingAtGoal)
        self.lookingAtGoal.setOutState(self.kickTheBall)

        
        self.debug = False
        
    def beginAction(self):
        self.doSubState(runToAndGrabBall,immediate = False)
        
    def DoAction(self):
        HelpLong.showFacePattern(FP_HANDKICKER)
        self.doActiveLocalise.DecideNextAction()
        if self.subState is None:     
            if Global.lostBall > 0:
                self.doSubState(self.doLocateBall)
            else:
                self.doSubState(self.runToAndGrabBall)
        else:
            self.doSubState()
            #hack
            if self.subState == self.sitDown:
                if self.turnToVoak.turnDir == Constant.dCLOCKWISE:
                    self.lookingAtGoal.setPan(-90)
                else:
                    self.lookingAtGoal.setPan(90)
        
        if self.subState is None:
            return None
        else:
            return self
 
class TurnKick(StateBasedAction):
    TURN_KICK_TIME = 30
    def __init__(self):
        StateBasedAction.__init__(self)
        self.turnDir = 0
        self.lastTurnSet = -1
        
    def beginAction(self):
        self.lastTurnSet = Global.frame
        
    def setTurnDir(self,dir = Constant.dCLOCKWISE):
        if Global.frame - self.lastTurnSet > 5:
            self.turnDir = dir
            self.lastTurnSet = Global.frame
        
    def DoAction(self):
        
        HelpLong.showFacePattern(FP_TURN_KICK)
        if Global.lostBall == 0 or self.counter < 25:
            if Global.lostBall == 0 and Global.ballD > 20:
                return None
            if self.counter < self.TURN_KICK_TIME:
                OpenChallenger.setNormalWalkTurnKick(self.turnDir)
                if self.counter == 0:
                    HelpLong.setForceStepComplete()
                return self
            else:
                return None
        else:
            return None
            
class TurnHoldingKick(StateBasedAction):
    def __init__(self):
        StateBasedAction.__init__(self)
        self.runToAndGrabBall = OpenChallenger.RunToGrabBall()
        self.holdTurning = HoldTurning()
        self.turnKick = TurnKick()
        
        self.doLocateBall = OpenChallenger.MyLocateBall()
        self.doActiveLocalise = MyActiveLocalise()
        
        self.doLocateBall.setOutState(self.runToAndGrabBall)
        
        self.runToAndGrabBall.setOutState(self.holdTurning)
        self.holdTurning.setOutState(self.holdTurning)
        self.holdTurning.setOutState(self.turnKick)
        
    def DoAction(self):
        self.doActiveLocalise.DecideNextAction()
        if self.subState is None:     
            if Global.lostBall > 0:
                self.doSubState(self.doLocateBall)
            else:
                self.doSubState(self.runToAndGrabBall)
        else:
            self.doSubState()
            if self.subState == self.holdTurning:   
                self.turnKick.setTurnDir(self.holdTurning.turnDir)
        if self.subState is None:
            return None
        else:
            return self
        
handKicker = HandKicker()
turnKick = TurnHoldingKick()

def initPlayer():
    pass
    
def DecideNextAction():

    Global.frameReset()
    HelpLong.superDebug()
    OpenChallenger.resetEachFrame()
    resetEachFrame()
##~     handKicker.DecideNextAction()
    turnKick.DecideNextAction()
##~     if Global.vTGoal.getConfidence() > 0:
##~         HelpLong.showFacePattern([1,1,1,1,1])
##~     else:
##~         HelpLong.showFacePattern([0,0,0,0,0])
initPlayer()
