#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: WalkingSkills.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




"""
Walking behaviours, gotoTarget, go this way, go that way...
"""
from StateBasedAction import StateBasedAction    
import HelpLong
import HelpShort
import HelpTrack
import Global
import Constant
##~ import LabArchive

############ Smarter Active goToTarger facing heading ######
### best used for "offline" dog (laboratory dog)


def setWalkFromPlaceToPlace(fromX,fromY,fromH, toX,toY,toH):
    dx,dy = toX - fromX , toY - fromY
    forward,left = HelpShort.rotate( (dx,dy), -fromH )
    forward = HelpShort.CLIP(forward,6)
    left = HelpShort.CLIP(left,6)
    turnCCW = HelpShort.CLIP( HelpShort.normalizeAngle_180(toH - fromH) , 30)
    HelpLong.setNormalWalk(forward,left,turnCCW)
 
def getSmartHeading(currentY):
    if currentY <= Constant.FIELD_LENGTH / 2:
        return 270.0
    else:
        return 90.0
        
def shouldFlipHeading(currentHeading):
    y = Global.selfLoc.getY()
    
    if getSmartHeading(y) != currentHeading:
        if currentHeading == 90.0:
            return y - Constant.FIELD_LENGTH/2.0 < -10
        else:
            return y - Constant.FIELD_LENGTH/2.0 > 10
    return False
    
class GoToPosition(StateBasedAction):
    FIRST_TURNING_STEP = "FIRST_TURNING_STEP"
    HORIZONTAL_STEP = "HORIZONTAL_STEP"
    VERTICAL_STEP = "VERTICAL_STEP"
    LAST_TURNING_STEP = "LAST_TURNING_STEP"
    FINISHED_STEP = "FINISHED_STEP"
    
    STOPPING_ENOUGH = 5
    STOPPING_ENOUGH_LAST_POSITION = 5
    STOPPING_ENOUGH_LAST = 5
    
    STATLOC_TIME = 30
    
    POSITION_ERROR = 10
    HEADING_ERROR = 10
    TIME_OUT = 900
    def __init__(self,targetX=0,targetY=0,targetH=0):
        StateBasedAction.__init__(self)
        self.setTarget(targetX,targetY,targetH)        
        self.stage = None   
        self.firstTurningStepCounter = 0
        self.horizontalStepCounter = 0
        self.verticalStepCounter = 0
        self.lastTurningStepCounter = 0
        self.desiredHeading = 0
        self.stoppingCounter = 0

    def setTarget(self,targetX=0,targetY=0,targetH=0):
        self.targetX,self.targetY,self.targetH = targetX,targetY,targetH
        
    def beginAction(self): 
        self.firstTurningStepCounter = 0
        self.horizontalStepCounter = 0
        self.verticalStepCounter = 0
        self.lastTurningStepCounter = 0
        self.stage = GoToPosition.FIRST_TURNING_STEP
                
    
    def setStage(self,stageNo):
        #hacked, should fix this bug (share stoppingCounter) properly
        if self.stage != stageNo:
            self.stage = stageNo
            self.stoppingCounter = 0
        
    def doFirstTurningStep(self):
        if self.firstTurningStepCounter < GoToPosition.STATLOC_TIME: #find current Heading
            HelpTrack.stationaryLocalise()
        else:
            if self.firstTurningStepCounter == GoToPosition.STATLOC_TIME: #find desired Heading
                self.desiredHeading = getSmartHeading(Global.selfLoc.getY())
            myHeading = Global.selfLoc.getHeading()
            if abs(myHeading - self.desiredHeading) < 5:
                self.setStage(GoToPosition.HORIZONTAL_STEP)
            turnCCW = HelpShort.normalizeAngle_180(self.desiredHeading-myHeading) / 2
            turnCCW = HelpShort.CLIP(turnCCW,30)
            HelpLong.setNormalWalk(0,0,turnCCW )
        self.firstTurningStepCounter += 1
    
    def doHorizontalStep(self,targetX):
        (x,y) = Global.selfLoc.getPos()
        var = Global.selfLoc.getPosVar()
        if HelpShort.get95sd(var) <= 40 and abs(x - targetX) <= GoToPosition.POSITION_ERROR:
##~             print "H",HelpShort.get95sd(var), abs(x - targetX)
            self.stoppingCounter += 1
            if self.stoppingCounter == GoToPosition.STOPPING_ENOUGH:
                self.setStage(GoToPosition.VERTICAL_STEP)
                return
        else:
##~             print "-H",HelpShort.get95sd(var), abs(x - targetX)
            self.stoppingCounter = 0
            
        if shouldFlipHeading(self.desiredHeading): 
            self.desiredHeading = 360 - self.desiredHeading #flip
        else:
            h = Global.selfLoc.getHeading()
            if abs(self.desiredHeading - h) < 20:
                setWalkFromPlaceToPlace(x,y,h, targetX,y,self.desiredHeading)
            else:
                setWalkFromPlaceToPlace(x,y,h, x,y,self.desiredHeading)            
            #hacked, set forward to 0
            HelpLong.setForwardLeftTurn(0,None,None)

    def doVerticalStep(self,_,targetY):
        (x,y) = Global.selfLoc.getPos()
        var = Global.selfLoc.getPosVar()
        if HelpShort.get95sd(var) <= 40 and abs(y - targetY) <= GoToPosition.POSITION_ERROR:
##~             print "V",HelpShort.get95sd(var), abs(y - targetY),self.stoppingCounter
            self.stoppingCounter += 1
            if self.stoppingCounter == GoToPosition.STOPPING_ENOUGH_LAST_POSITION:
                self.setStage(GoToPosition.LAST_TURNING_STEP)
                return
        else:
##~             print "-V",HelpShort.get95sd(var), abs(y - targetY)
            self.stoppingCounter = 0
            
        if shouldFlipHeading(self.desiredHeading):
            self.desiredHeading = 360 - self.desiredHeading #flip
        else:
            h = Global.selfLoc.getHeading()
            if abs(self.desiredHeading - h) < 20:
                setWalkFromPlaceToPlace(x,y,h, x,targetY,self.desiredHeading)
            else:
                setWalkFromPlaceToPlace(x,y,h, x,y,self.desiredHeading)
            #hacked, set left to 0
            HelpLong.setForwardLeftTurn(None,0,None)
        
    def doLastTurningStep(self,targetH):
        global stoppingCounter
        if self.lastTurningStepCounter > GoToPosition.TIME_OUT:
            print "turningStep : timeout"
            self.setStage(GoToPosition.FINISHED_STEP)
            return 
        self.lastTurningStepCounter += 1
        h = Global.selfLoc.getHeading()
        x,y = Global.selfLoc.getPos()
        var = Global.selfLoc.getPosVar()
        if HelpShort.get95sd(var) <= 50 and abs(HelpShort.normalizeAngle_180(h - targetH)) <= GoToPosition.HEADING_ERROR:
            self.stoppingCounter += 1
            if self.stoppingCounter == GoToPosition.STOPPING_ENOUGH_LAST:                
                self.setStage(GoToPosition.FINISHED_STEP)
                return
        else:
##~             print "-T",HelpShort.get95sd(var),abs(h - targetH)
            self.stoppingCounter = 0
        setWalkFromPlaceToPlace(x,y,h, x,y,targetH)
            
    #currently handle of all sub-states, 
    #the proper solution would leave the framework to handle it
    def DoAction(self): 
##~         LabArchive.doMyActiveLocalise()
##~         print self.stage
        if self.stage == GoToPosition.FIRST_TURNING_STEP: 
            self.doFirstTurningStep()
        elif self.stage == GoToPosition.HORIZONTAL_STEP:
            self.doHorizontalStep(self.targetX)
        elif self.stage == GoToPosition.VERTICAL_STEP:
            self.doVerticalStep(self.targetX,self.targetY)
        elif self.stage == GoToPosition.LAST_TURNING_STEP:
            self.doLastTurningStep(self.targetH)
        else:
            pass 
        return self
    
    def isFinished(self):
        return self.stage == GoToPosition.FINISHED_STEP
