/*

   Copyright 2004 The University of New South Wales(UNSW) and National  
   ICT Australia(NICTA).

   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
   redistribute it and/or modify it under the terms of the GNU General  
   Public License as published by the Free Software Foundation; either  
   version 2 of the License, or(at your option) any later version as  
   modified below.  As the original licensors, we add the following  
   conditions to that license:

   In paragraph 2.b), the phrase "distribute or publish" should be  
   interpreted to include entry into a competition, and hence the source  
   of any derived work entered into a competition must be made available  
   to all parties involved in that competition under the terms of this  
   license.

   In addition, if the authors of a derived work publish any conference  
   proceedings, journal articles or other academic papers describing that  
   derived work, then appropriate academic citations to the original work  
   must be included in that publication.

   This rUNSWift source is distributed in the hope that it will be useful,  
   but WITHOUT ANY WARRANTY; without even the implied warranty of  
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
   General Public License for more details.

   You should have received a copy of the GNU General Public License along  
   with this source code; if not, write to the Free Software Foundation,  
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: PWalk.cc 4364 2004-09-22 07:04:12Z kcph007 $
 *
 * Copyright(c) 2004 UNSW
 * All Rights Reserved.
 * 
 **/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <fstream>
#include <string>
#include <math.h>
#include "PWalk.h"
#include "../share/OdometerDef.h"
#include "legsGeometry.h"
#include "../share/sensors.h"

using namespace std;


static const bool bDebugWalkCommand = false;
static const bool bDebugPSTEP = false;
static const bool bDebugAll = false;
static const bool useSensorsForHeightTilt = true;

#ifdef OFFLINE
static string POS_FILE_PATH = "../../robot7/pos";
#else
static string POS_FILE_PATH = "/MS/pos";
#endif


static const string CHESH_PUSH_FILE = POS_FILE_PATH + "/cpush.pos";
static const string FWDKICK_FILE = POS_FILE_PATH + "/fwdkick.pos";
static const string FASTKICK_FILE = POS_FILE_PATH + "/fastkick.pos";
static const string GETUP_F_FILE = POS_FILE_PATH + "/getupf.pos";
static const string GETUP_S_FILE = POS_FILE_PATH + "/getups.pos";
static const string GETUP_R_FILE = POS_FILE_PATH + "/getupr.pos";
static const string BACKKICK = POS_FILE_PATH + "/bckkick.pos";
static const string CHEER_FILE = POS_FILE_PATH + "/cheer.pos";
static const string DANCE_FILE = POS_FILE_PATH + "/dance.pos";
static const string UPENNRIGH_FILE = POS_FILE_PATH + "/UPright.pos";
static const string UPENNLEFT_FILE = POS_FILE_PATH + "/UPleft.pos";

static const string HANK_KICK_RIGHT_FILE = POS_FILE_PATH + "/HKright.pos";
static const string HANK_KICK_LEFT_FILE = POS_FILE_PATH + "/HKleft.pos";
static const string DIVE_KICK_FILE = POS_FILE_PATH + "/diveKick.pos";

static const string STARTUP_FILE = POS_FILE_PATH + "/startup.pos";

/*
 * PWalk
 */
PWalk::PWalk(const long *sensorValTemp, const long *dutyCycleValTemp) : jointCommand() {
	cout << "This is PWalk constructor" << endl;

    //Odometer should not be done in JointAgent
	//sOdometer = om;

	sp = NULL;
	idle = false;
    forceComplete = false;

	walktype = -1;
	bufferedCommand.walkType = -1;

	delta_forward = delta_left = delta_turn = 0;

	sensorVal = sensorValTemp;
	dutyCycleVal = dutyCycleValTemp;

	/* Special actions. */
	chestpush = new SpecialAction();
	chestpush->load(CHESH_PUSH_FILE);
	fwdkick = new SpecialAction();
	fwdkick->load(FWDKICK_FILE);
	fastkick = new SpecialAction();
	fastkick->load(FASTKICK_FILE);

	getupf = new SpecialAction();
	getupf->load(GETUP_F_FILE);
	getups = new SpecialAction();
	getups->load(GETUP_S_FILE);
	getupr = new SpecialAction();
	getupr->load(GETUP_R_FILE);
	backKick = new SpecialAction();
	backKick->load(BACKKICK);

	cheer = new SpecialAction();
	cheer->load(CHEER_FILE);
	dance = new SpecialAction();
	dance->load(DANCE_FILE);

	startup = new SpecialAction();
	startup->load(STARTUP_FILE);

	upennRight = new SpecialAction();
	upennRight->load(UPENNRIGH_FILE);
	upennLeft = new SpecialAction();
	upennLeft->load(UPENNLEFT_FILE);

	hkLeft = new SpecialAction();
	hkLeft->load(HANK_KICK_LEFT_FILE);

	hkRight = new SpecialAction();
	hkRight->load(HANK_KICK_RIGHT_FILE);

	diveKick = new SpecialAction();
	diveKick->load(DIVE_KICK_FILE);



	normalWalk = new NormalWalk();
	backWalk = new BackWalk();
	offsetWalk = new OffsetWalk();
	ellipticalWalk = new EllipticalWalk();
	alanWalk = new AlanWalk();
	skiWalk = new SkiWalk(this);
	skellipticalWalk = new SkellipticalWalk();

	/* Pointer to shared variables walktype, step_ and PG, provide read-only access outside PWalk. */
	//How does it actually work?
	ptr_walktype = &walktype;
	ptr_step = &step;
	ptr_PG = &PG;
	ptr_frontShoulderHeight = &frontShoulderHeight;
	ptr_bodyTilt = &bodyTilt;


	sp = startup;
	sp->setPointNum( 0 );

	readLocusOffets("/MS/offsets.txt");
}

//read parameters that's set by vision
void PWalk::updateParameters(){
    if (bDebugPSTEP)
        cout << "wt = " << walktype << " completed = " << isCompleted() << " " << bufferedCommand.walkType << endl;
    if ( bufferedCommand.walkType != UpennLeftWT  && bufferedCommand.walkType != UpennRightWT && ! isCompleted() ){ // isCompleted tells whether or not the step can be changed
        return ;
    }
    
    if (bDebugWalkCommand)
        cout << "walk " << bufferedCommand.walkType << " " << bufferedCommand.Forward << " " << bufferedCommand.Left << " " << bufferedCommand.turnCCW << " " << bufferedCommand.speed << endl;

    walktype = bufferedCommand.walkType;

    BaseAction *prevStep = sp;
	switch (walktype){
		case -1:
		    // haven't receieved a command yet - do nothing
		    sp = NULL;
                    break;
		case NormalWalkWT:
		case CanterWalkWT:
			normalWalk->setParameters(bufferedCommand);
			sp = normalWalk;
			break;
        case BackWalkWT:
			backWalk->setParameters(bufferedCommand);
			sp = backWalk;
			break;
		case OffsetWalkWT:
			offsetWalk->setParameters(bufferedCommand);
			sp = offsetWalk;
			break;
		case EllipseWalkWT:
			ellipticalWalk->setParameters(bufferedCommand);
			sp = ellipticalWalk;
			break;
        case AlanWalkWT:
			alanWalk->setParameters(bufferedCommand);
            sp = static_cast<SpecialAction*>(alanWalk);
            break;
        case SkiWalkWT:
			skiWalk->setParameters(bufferedCommand);
            sp = skiWalk;
            break;
        case SkellipticalWalkWT:
			skellipticalWalk->setParameters(bufferedCommand);
            sp = skellipticalWalk;
            break;
		case ChestPushWT:
			sp = chestpush;
			break;
		case FwdKickWT:
			sp = fwdkick;
			break;
		case FastKickWT:
			sp = fastkick;
			break;
		case GetupFrontWT:
			sp = getupf;
			break;
		case GetupSideWT:
			sp = getups;
			break;
		case GetupRearWT:
			sp = getupr;
			break;
		case BackKickWT:
			sp = backKick;
			break;
		case CheerWT:
			sp = cheer;
			break;
		case DanceWT:
            sp = dance;
            break;
        case UpennRightWT:
            sp = upennRight;
            break;
        case UpennLeftWT:
            sp = upennLeft;
            break;
        case DiveKickWT:
            sp = diveKick;
            break;
        case HandKickRightWT:
            sp = hkRight;
            break;
        case HandKickLeftWT:
            sp = hkLeft;
            break;
        default:
            cout << "Unknown walktype " << walktype << endl;
			sp = NULL;
			break;
	}
	if (sp != NULL) {
		PG = sp->getNumPoints();
        if (prevStep != NULL && prevStep != sp){ //if the walk is switched, set the appropiate step cycle ( smooth switching, 0 percentage = right front hand moving forward)
            if (!sp->isSpecialAction())
                if (prevStep->isSpecialAction() ){
                    sp->setPointNum( 0 );
                    //cout << "set cycle 0" << endl;
                }
                else {
                    sp->setPointNum( int(stepCycle * PG) );
                    //cout << "set cycle " << int(stepCycle * PG) <<  endl;
                }
            //cout << "prev step " << prevStep->getPointNum() << endl;
        }
	}
}


//continuing steps, is called when OPENR is ready to receive command
void PWalk::pStep(bool highGain){
	idle = false;
//	if (bDebugPSTEP)
//        cout << "pStep() " << walktype << " PG = " << PG << endl;

	updateParameters();
	jointCommand.reset();

    if (sp == NULL) {
        if (bDebugAll) 
            cout << "sp is NULL in pStep where it shouldn't be" << endl;
        return;
    }

    if (sp->isCompleted()){
        sp->reset();
    }

    sp->setCurrentJointCommand(jointCommand);
	sp->goNextFrame();
    sp->getOdometry(delta_forward, delta_left, delta_turn, highGain);
    step = sp->getPointNum();
    if (!sp->isSpecialAction() ) //if it is a walk, memorize where it is in the cycle (percentage)
    {
        stepCycle = step * 1.0 / sp->getNumPoints();
    }

    if (useSensorsForHeightTilt) {
    	// N.B. Sensors will be set from previous joint angles when offline
		JOINT_Values flTheta = {MICRO2RAD(sensorVal[ssFL_ROTATOR]),
								MICRO2RAD(sensorVal[ssFL_ABDUCTOR]),
								MICRO2RAD(sensorVal[ssFL_KNEE])};
		JOINT_Values blTheta = {MICRO2RAD(sensorVal[ssRL_ROTATOR]),
								MICRO2RAD(sensorVal[ssRL_ABDUCTOR]),
								MICRO2RAD(sensorVal[ssRL_KNEE])};
		JOINT_Values frTheta = {MICRO2RAD(sensorVal[ssFR_ROTATOR]),
								MICRO2RAD(sensorVal[ssFR_ABDUCTOR]),
								MICRO2RAD(sensorVal[ssFR_KNEE])};
		JOINT_Values brTheta = {MICRO2RAD(sensorVal[ssRR_ROTATOR]),
								MICRO2RAD(sensorVal[ssRR_ABDUCTOR]),
								MICRO2RAD(sensorVal[ssRR_KNEE])};
		
		calcHeightTilt(flTheta, blTheta, frTheta, brTheta);
	} else {
		JOINT_Values flTheta = {jointCommand.getJointValueInRads(leftFrontJoint-1),
								jointCommand.getJointValueInRads(leftFrontShoulder-1),
								jointCommand.getJointValueInRads(leftFrontKnee-1)};
		JOINT_Values blTheta = {jointCommand.getJointValueInRads(leftRearJoint-1),
								jointCommand.getJointValueInRads(leftRearShoulder-1),
								jointCommand.getJointValueInRads(leftRearKnee-1)};
		JOINT_Values frTheta = {jointCommand.getJointValueInRads(rightFrontJoint-1),
								jointCommand.getJointValueInRads(rightFrontShoulder-1),
								jointCommand.getJointValueInRads(rightFrontKnee-1)};
		JOINT_Values brTheta = {jointCommand.getJointValueInRads(rightRearJoint-1),
								jointCommand.getJointValueInRads(rightRearShoulder-1),
								jointCommand.getJointValueInRads(rightRearKnee-1)};
		
		calcHeightTilt(flTheta, blTheta, frTheta, brTheta);
	}
	
	// cout << jointCommand << endl;
}

void PWalk::pStep(AtomicAction &newCommand){

    if (bDebugPSTEP){
        cout << "buffer is updated old WT=" << bufferedCommand.walkType << " new WT=" << newCommand.walkType << endl;
    }
    bufferedCommand = newCommand;
    //if the dog is standing, PG should be 5 sothat the next command would bear no delay
    if (bufferedCommand.Forward == 0 && bufferedCommand.Left == 0 && bufferedCommand.turnCCW == 0){
        bufferedCommand.speed = 5;
        bufferedCommand.hdF = 0;
        bufferedCommand.hdB = 0;
    }
    if (bufferedCommand.cmd == aaForceStepComplete)
        forceComplete = true;
    else
        forceComplete = false;
    if (idle){
        cout << "PWalk is Idle, weird, please tell Kim " ;
        pStep(true);
    }

    idle = true;

}  /*(finally) end function pStep */

bool PWalk::isCompleted(){
	if (sp == NULL)
		return true;
        /*
    if (walktype == NormalWalkWT || walktype == EllipseWalkWT)
        return true;
        */
    if (!sp->isSpecialAction() && forceComplete) return true; // we dont want to force special action
	return sp->isCompleted();
}

void PWalk::readLocusOffets(const char *filename)
{
	offsetWalk->readLocusOffets(filename);
}

bool PWalk::NotUsingHead(){
    if (sp == NULL)
        return false;
    else
        return !sp->usingHead();
}

void PWalk::reloadAllSpecialAction()
{
    chestpush->reload();
    fwdkick->reload();
    fastkick->reload();
    getupf->reload();
    getups->reload();
    getupr->reload();
    backKick->reload();
    cheer->reload();        
    dance->reload();        
    upennLeft->reload();
    upennRight->reload();
    hkRight->reload();
    hkLeft->reload();
    diveKick->reload();
}

void PWalk::getAngleHeight(double &angle, double &shoulderHeight,
		double bX, double bY, double fX, double fY) {
	// given two X,Y coordinates (front and back)
	// Both with origin the front shoulder, but with opposite signs for X
	// And with the X axis being horozontal through the dogs body
	// find the resting body tilt angle of the dog
	// and the height of the front shoulder from the ground
	angle = atan2(bY-fY, bX+fX);
	double alpha = atan2(fX, fY);
	double gamma = alpha - angle;
	double fdist = sqrt(SQUARE(fX)+SQUARE(fY));
	shoulderHeight = fdist*cos(gamma);
	// cout << "bX: " << bX << " bY: " << bY << " fX: " << fX << " fY: " << fY << " ang: " << RAD2DEG(angle) << " h: " << height << endl;
}

void PWalk::calcHeightTilt(const JOINT_Values &flTheta, const JOINT_Values &blTheta, const JOINT_Values &frTheta, const JOINT_Values &brTheta) {
/*
	cout << "fl: " << flTheta << endl;
	cout << "fr: " << frTheta << endl;
	cout << "bl: " << blTheta << endl;
	cout << "br: " << brTheta << endl;
*/
	XYZ_Coord flToe;
	kinematics(flTheta, tLeg, fbLegPadEdge, flToe);
	
	XYZ_Coord frToe;
	kinematics(frTheta, tLeg, fbLegPadEdge, frToe);
	
	XYZ_Coord flBase;
	kinematics(flTheta, tLeg, fbLegBaseEdge, flBase);
	
	XYZ_Coord frBase;
	kinematics(frTheta, tLeg, fbLegBaseEdge, frBase);
	
	XYZ_Coord blToe;
	kinematics(blTheta, tLeg, bbLegToes, blToe);
	
	XYZ_Coord brToe;
	kinematics(brTheta, tLeg, bbLegToes, brToe);
	
	double bX, bY;
	
	if (blToe.y < brToe.y) {
		bX = brToe.x;
		bY = brToe.y;
	} else {
		bX = blToe.x;
		bY = blToe.y;
	}
	
	if (bY < SHOULDER_TO_BASE) {
		bY = SHOULDER_TO_BASE;
		bX = SHOULDER_TO_REAR;
	}
	
	bX += BODY_LENGTH;
	
	double tempBodyTilt = PI/2;
	double tempFrontShoulderHeight = -100;

	double bestBodyTilt = tempBodyTilt;
	double bestFrontShoulderHeight = tempFrontShoulderHeight;

	// now check all the possible front contact points in turn
	
	// front left toes
	getAngleHeight(tempBodyTilt, tempFrontShoulderHeight, bX, bY, flToe.x, flToe.y);
	if (tempBodyTilt < bestBodyTilt) {
		bestBodyTilt = tempBodyTilt;
		bestFrontShoulderHeight = tempFrontShoulderHeight;
	}
	
	// front right toes
	getAngleHeight(tempBodyTilt, tempFrontShoulderHeight, bX, bY, frToe.x, frToe.y);
	if (tempBodyTilt < bestBodyTilt) {
		bestBodyTilt = tempBodyTilt;
		bestFrontShoulderHeight = tempFrontShoulderHeight;
	}
	
	// front left base
	getAngleHeight(tempBodyTilt, tempFrontShoulderHeight, bX, bY, flBase.x, flBase.y);
	if (tempBodyTilt < bestBodyTilt) {
		bestBodyTilt = tempBodyTilt;
		bestFrontShoulderHeight = tempFrontShoulderHeight;
	}
	
	// front right base
	getAngleHeight(tempBodyTilt, tempFrontShoulderHeight, bX, bY, frBase.x, frBase.y);
	if (tempBodyTilt < bestBodyTilt) {
		bestBodyTilt = tempBodyTilt;
		bestFrontShoulderHeight = tempFrontShoulderHeight;
	}
	
	// front of chest
	getAngleHeight(tempBodyTilt, tempFrontShoulderHeight, bX, bY, SHOULDER_TO_FRONT, SHOULDER_TO_BASE);
	if (tempBodyTilt < bestBodyTilt) {
		bestBodyTilt = tempBodyTilt;
		bestFrontShoulderHeight = tempFrontShoulderHeight;
	}
	
	frontShoulderHeight = bestFrontShoulderHeight;
	bodyTilt = bestBodyTilt;
	backHipHeight = frontShoulderHeight + BODY_LENGTH*sin(bodyTilt);
	// cout << "Sh: " << frontShoulderHeight << " bt: " << RAD2DEG(bodyTilt) << endl;
}


/*
 * end PWalk
 */

