/*

   Copyright 2004 The University of New South Wales (UNSW) and National  
   ICT Australia (NICTA).

   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
   redistribute it and/or modify it under the terms of the GNU General  
   Public License as published by the Free Software Foundation; either  
   version 2 of the License, or (at your option) any later version as  
   modified below.  As the original licensors, we add the following  
   conditions to that license:

   In paragraph 2.b), the phrase "distribute or publish" should be  
   interpreted to include entry into a competition, and hence the source  
   of any derived work entered into a competition must be made available  
   to all parties involved in that competition under the terms of this  
   license.

   In addition, if the authors of a derived work publish any conference  
   proceedings, journal articles or other academic papers describing that  
   derived work, then appropriate academic citations to the original work  
   must be included in that publication.

   This rUNSWift source is distributed in the hope that it will be useful,  
   but WITHOUT ANY WARRANTY; without even the implied warranty of  
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
   General Public License for more details.

   You should have received a copy of the GNU General Public License along  
   with this source code; if not, write to the Free Software Foundation,  
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

/*
 * Last modification background information
 * $Id: SkellipticalWalk.cc 4364 2004-09-22 07:04:12Z kcph007 $
 *
 * Copyright (c) 2004 UNSW
 * All Rights Reserved.
 *
**/



#include "legsGeometry.h"
#include "SkellipticalWalk.h"
#include "../share/Common.h"
#include "JointCommand.h"

static const bool ellipticalLeadIn = false;
static const bool debugFSH = false;
/*
 * Command for walk calibration
 * __05wtcal__99 20 8.5 0 0 40 15 15 30 25 0.5 0.55 0.01 0.01 0.01 0.01 0.1 0.2 0.2 0.4 55 15 80 -55 5 115 0 0 3 0 0
 */

void SkellipticalWalk::setDefaultParameters() {
	struct skelWalkParms tempWalkParams;
	struct skelStepParms tempStepParams;
	
	tempStepParams.cmdFwd = 0;
	tempStepParams.cmdLeft = 0;
	tempStepParams.cmdTurnCCW = 0;
	tempStepParams.shortStep = 1;
	
	tempWalkParams.init();

	step = 0;
	time = 1;

	setWalkParams(tempWalkParams);
	setStepParams(tempStepParams);
}

void SkellipticalWalk::setParameters(const AtomicAction &bufferedCommand) {

	if (false) {
		skelWalkParms tempWalkParams;

		tempWalkParams.init();

		tempWalkParams.halfStepTime = bufferedCommand.speed;
		tempWalkParams.frontFwdHeight = bufferedCommand.hdF;
		tempWalkParams.frontTrnLftHeight = bufferedCommand.hdF;
		tempWalkParams.backFwdHeight = bufferedCommand.hdB;
		tempWalkParams.backTrnLftHeight = bufferedCommand.hdB;
		tempWalkParams.frontF = bufferedCommand.ffO;
		tempWalkParams.frontS = bufferedCommand.fsO;
		tempWalkParams.frontH = bufferedCommand.hF;
		tempWalkParams.backF = bufferedCommand.bfO;
		tempWalkParams.backS = bufferedCommand.bsO;
		tempWalkParams.backH = bufferedCommand.hB;

		setWalkParams(tempWalkParams);
	}

	skelStepParms tempStepParams;
	
	tempStepParams.cmdFwd = bufferedCommand.Forward;
	tempStepParams.cmdLeft = bufferedCommand.Left;
	tempStepParams.cmdTurnCCW = bufferedCommand.turnCCW;
	tempStepParams.shortStep = 1;

	setStepParams(tempStepParams);
}

void SkellipticalWalk::setWalkParams(const struct skelWalkParms &newParms) {
	walkParms = newParms;

	cout << "Set SkellipticalWalk parameters: ";
	newParms.print(cout);
	cout << endl;

	if (walkParms.halfStepTime < 1) {
		// stop a whole lot of problems with div by 0
		// You can't do anything in zero or -ve time
		walkParms.halfStepTime = 1;
		standing = true;
	} else {
		standing = false;
	}

	// add some limits to the parameters so that the timing makes sense

	if (walkParms.frontDutyCycle < 0) {
		walkParms.frontDutyCycle = 0;
	} else if (walkParms.frontDutyCycle > 1) {
		walkParms.frontDutyCycle = 1;
	}
	if (walkParms.backDutyCycle < 0) {
		walkParms.backDutyCycle = 0;
	} else if (walkParms.backDutyCycle > 1) {
		walkParms.backDutyCycle = 1;
	}

	if (walkParms.frontLeadOutFrac < 0) {
		walkParms.frontLeadOutFrac = 0;
	} else if (walkParms.frontLeadOutFrac > 1) {
		walkParms.frontLeadOutFrac = 1;
	}
	if (walkParms.backLeadOutFrac < 0) {
		walkParms.backLeadOutFrac = 0;
	} else if (walkParms.backLeadOutFrac > 1) {
		walkParms.backLeadOutFrac = 1;
	}
	if (walkParms.frontLeadInFrac < 0) {
		walkParms.frontLeadInFrac = 0;
	} else if (walkParms.frontLeadInFrac > 1) {
		walkParms.frontLeadInFrac = 1;
	}
	if (walkParms.backLeadInFrac < 0) {
		walkParms.backLeadInFrac = 0;
	} else if (walkParms.backLeadInFrac > 1) {
		walkParms.backLeadInFrac = 1;
	}

	
	bodyTilt = asin((walkParms.backH - walkParms.frontH)/lsh);
	shoulderDist = lsh*cos(bodyTilt);
	
	double dF = shoulderDist/2;
	double dS = BODY_WIDTH/2;
	
	flFOffset = dF + walkParms.frontF - walkParms.turnCenterF;
	flLOffset = -dS - walkParms.frontS + walkParms.turnCenterL + walkParms.sideOffset;
	flTheta0 = atan2(flFOffset, flLOffset);
	flRadius = sqrt(flFOffset*flFOffset + flLOffset*flLOffset);

	frFOffset = dF + walkParms.frontF - walkParms.turnCenterF;
	frLOffset = dS + walkParms.frontS + walkParms.turnCenterL + walkParms.sideOffset;
	frTheta0 = atan2(frFOffset, frLOffset);
	frRadius = sqrt(frFOffset*frFOffset + frLOffset*frLOffset);

	blFOffset = -dF + walkParms.backF - walkParms.turnCenterF;
	blLOffset = -dS - walkParms.backS + walkParms.turnCenterL + walkParms.sideOffset;
	blTheta0 = atan2(blFOffset, blLOffset);
	blRadius = sqrt(blFOffset*blFOffset + blLOffset*blLOffset);

	brFOffset = -dF + walkParms.backF - walkParms.turnCenterF;
	brLOffset = dS + walkParms.backS + walkParms.turnCenterL + walkParms.sideOffset;
	brTheta0 = atan2(brFOffset, brLOffset);
	brRadius = sqrt(brFOffset*brFOffset + brLOffset*brLOffset);

	step = (int)step;	// round this when we set walk parameters
}

void SkellipticalWalk::setStepParams(const struct skelStepParms &newParms) {
	stepParms = newParms;
	
	// cout << "Set SkellipticalWalk step parameters: " << newParms.cmdFwd << ", " << newParms.cmdLeft << ", " << newParms.cmdTurnCCW << ", " << newParms.shortStep << endl;

	double oldTime = time;
	double oldStep = step;
	
	time = 2*walkParms.halfStepTime*stepParms.shortStep;
	
	if (oldTime == 0) {
		step = 0;
	} else {
		step = oldStep/oldTime*time;	// move to the correct location in the new step
	}
	
	calibrateWalk(stepParms.cmdFwd, stepParms.cmdLeft, stepParms.cmdTurnCCW, stepParms.shortStep,
				fwdLength, leftLength, turnLength);
	
	forwardSpeed = stepParms.cmdFwd/walkParms.halfStepTime;
	leftSpeed = stepParms.cmdLeft/walkParms.halfStepTime;
	turnSpeed = stepParms.cmdTurnCCW/walkParms.halfStepTime;

	double heightMix;
	if ((fwdLength != 0) || (leftLength != 0) || (turnLength != 0)) {
		heightMix = fabs(fwdLength)/(fabs(fwdLength) + fabs(leftLength) + fabs(turnLength)*flRadius);
	} else {
		heightMix = 1;
	}
	
	frontHeight = heightMix*walkParms.frontFwdHeight + (1-heightMix)*walkParms.frontTrnLftHeight;
	backHeight = heightMix*walkParms.backFwdHeight + (1-heightMix)*walkParms.backTrnLftHeight;

	frontStartLeadOutTime = time/2*walkParms.frontDutyCycle * (1-walkParms.frontLeadOutFrac);
	backStartLeadOutTime = time/2*walkParms.backDutyCycle * (1-walkParms.backLeadOutFrac);
	frontEndLeadInTime = time - time/2*walkParms.frontDutyCycle * (1-walkParms.frontLeadInFrac);
	backEndLeadInTime = time - time/2*walkParms.backDutyCycle * (1-walkParms.backLeadInFrac);
	frontEndLeadOutTime = time/2 * walkParms.frontDutyCycle;
	backEndLeadOutTime = time/2 * walkParms.backDutyCycle;
	frontStartLeadInTime = time - time/2 * walkParms.frontDutyCycle;
	backStartLeadInTime = time - time/2 * walkParms.backDutyCycle;
}

void SkellipticalWalk::calibrateWalk(double &cmdFwd, double &cmdLeft, double &cmdTurn, double shortStep,
				double &fwdLen, double &leftLen, double &turnLen) {
	// first clip things to the valid range
	// this changes the commands themselves
	cmdFwd = CLIP(cmdFwd, MAX_SKEL_FORWARD);
	cmdLeft = CLIP(cmdLeft, MAX_SKEL_LEFT);
	cmdTurn = CLIP(cmdTurn, MAX_SKEL_TURN);
	
	// now calibrate the lengths to get the correct walking speed
	// If you're doing calibration, this is the bit you want to change
	// If the walk theory was perfect, these would be simple assignments:
	// fwdLen = cmdFwd, etc.
	
	fwdLen = cmdFwd;
	leftLen = cmdLeft;
	turnLen = cmdTurn;
	
	// now shrink the lengths to take account of the short step
	
	fwdLen *= shortStep;
	leftLen *= shortStep;
	turnLen *= shortStep;
	
	// finally, convert into the correct units
	
	fwdLen *= 10;	// cm into mm
	leftLen *= 10;	// cm into mm
	turnLen = DEG2RAD(turnLen);	// degrees into rads
}

void SkellipticalWalk::getOdometry(double &delta_forward, double &delta_left, double &delta_turn, bool highGain) {
	delta_forward = forwardSpeed;
	delta_left = leftSpeed;
	delta_turn = turnSpeed;
}

void SkellipticalWalk::handleTilt(JOINT_Values &joints, bool front) {
	if (front) {
		joints.rotator += bodyTilt;	// handle the tilt
	} else {
		joints.rotator -= bodyTilt;	// handle the tilt
	}
}

void SkellipticalWalk::getLocusX(const double step,
					double &X,
					const bool front,
					const double length) {

	const double endLeadOutTime = front?frontEndLeadOutTime:backEndLeadOutTime;
	const double startLeadInTime = front?frontStartLeadInTime:backStartLeadInTime;

	double wrappedStep = step;
	if (wrappedStep > time/2)
		wrappedStep -= time;

	const double mySpeed = length*2.0/time;

	if ((step <= endLeadOutTime) || (step >= startLeadInTime)) {
		// bottom of locus
		X = -wrappedStep*mySpeed;
	} else {
		// recovery phase
		double theta;
		double sideLength;
		if (step < time/2) {
			theta = (step-endLeadOutTime)/(time/2-endLeadOutTime)*PI/2;
			sideLength = mySpeed*endLeadOutTime;
		} else {
			theta = PI - (startLeadInTime - step)/(startLeadInTime - time/2)*PI/2;
			sideLength = mySpeed*(time-startLeadInTime);
		}
		X = -sideLength * cos(theta);
		// cout << "step: " << step << " theta: " << theta << " length: " << length << " len: " << sideLength << " X: " << X << endl;
	}
}

void SkellipticalWalk::getLocusY(const double step,
					double &Y,
					const bool front,
					const double height) {

	const double startLeadInTime = front?frontStartLeadInTime:backStartLeadInTime;
	const double endLeadInTime = front?frontEndLeadInTime:backEndLeadInTime;
	const double startLeadOutTime = front?frontStartLeadOutTime:backStartLeadOutTime;
	const double endLeadOutTime = front?frontEndLeadOutTime:backEndLeadOutTime;

	double wrappedStep = step;
	if (wrappedStep > time/2)
		wrappedStep -= time;
	
	double frac;
	
	if (wrappedStep > 0) {
		frac = front?walkParms.frontLeadUpFrac:walkParms.backLeadUpFrac;
	} else {
		frac = front?walkParms.frontLeadDownFrac:walkParms.backLeadDownFrac;
	}

	if ((step <= startLeadOutTime) || (step >= endLeadInTime)) {
		// flat section along bottom of locus
		Y = 0;
	} else if ((step < endLeadOutTime) || (step > startLeadInTime)) {
		// handle lead in and out
		double eX;
		// first calculate the X position within the ellipse
		if (wrappedStep > 0) {
			eX =  (step - startLeadOutTime)/(endLeadOutTime - startLeadOutTime);
		} else {
			eX =  (endLeadInTime - step)/(endLeadInTime - startLeadInTime);
		}
		
		// then figure out the Y value
		if (ellipticalLeadIn) {
			// elliptical lead in
			Y = height*frac * (1 - sqrt(1 - eX*eX));
		} else {
			// linear lead in
			Y = eX * height * frac;
		}
		
	} else {
		// return stroke
		double theta;
		if (step < time/2) {
			theta = (step-endLeadOutTime)/(time/2-endLeadOutTime)*PI/2;
		} else {
			theta = PI - (startLeadInTime - step)/(startLeadInTime - time/2)*PI/2;
		}
		Y = height*frac + height*(1-frac)*sin(theta);
	}
}

void SkellipticalWalk::makeFSFwd(double X, double &f, double &s) {
	f = X;
	s = 0;
}

void SkellipticalWalk::makeFSLft(double X, double &f, double &s, bool left) {
	f = 0;
	s = left?X:-X;
}

void SkellipticalWalk::makeFSTrn(double X, double &f, double &s, bool left, bool front) {
	if (front) {
		if (left) {
			f = sin(X + flTheta0)*flRadius - flFOffset;
			s = -(cos(X + flTheta0)*flRadius - flLOffset);
		} else {
			f = sin(X + frTheta0)*frRadius - frFOffset;
			s = cos(X + frTheta0)*frRadius - frLOffset;
		}
	} else {
		if (left) {
			f = sin(X + blTheta0)*blRadius - blFOffset;
			s = -(cos(X + blTheta0)*blRadius - blLOffset);
		} else {
			f = sin(X + brTheta0)*brRadius - brFOffset;
			s = cos(X + brTheta0)*brRadius - brLOffset;
		}
	}
}

void SkellipticalWalk::getLocusJoints(double step, bool front, bool left, JOINT_Values &joints) {
	double X, Y, f, s, h, fTemp, sTemp;
	XYZ_Coord xyz;

	getLocusY(step, Y, front, front?frontHeight:backHeight);
	h = Y;

	// add the thrust
	{
		double startLeadOutTime = front?frontStartLeadOutTime:backStartLeadOutTime;
		double endLeadOutTime = front?frontEndLeadOutTime:backEndLeadOutTime;
		if (step < startLeadOutTime)
			h -= step / startLeadOutTime * (walkParms.thrustHeight);
		else if (step < endLeadOutTime) {
			double diff = endLeadOutTime - startLeadOutTime;
			h -= (endLeadOutTime - step) / diff * (walkParms.thrustHeight);
		}
	}
	// and the canter
	double stepCycle = step / time * 4.0 * pi;
	h -= walkParms.canterHeight * cos(stepCycle);
	
	getLocusX(step, X, front, fwdLength);	// fwd
	makeFSFwd(X, f, s);
	
	getLocusX(step, X, front, leftLength);	// left
	makeFSLft(X, fTemp, sTemp, left);
	f += fTemp;
	s += sTemp;
	
	getLocusX(step, X, front, turnLength);	// turn
	makeFSTrn(X, fTemp, sTemp, left, front);
	f += fTemp;
	s += sTemp;
	
	// rotate2DCoordinate(f, h, walkParms.locRotate);

	if (debugFSH)
		cout << "fl: " << front << left << " fsh: " << f << " " << s << " " << h << endl;
	
	// convert to xyz coords
	if (front) {
		xyz.x = walkParms.frontF + f;
		xyz.z = walkParms.frontS + s + (left?-walkParms.sideOffset:walkParms.sideOffset);
		xyz.y = walkParms.frontH - h; 
	} else {
		xyz.x = walkParms.backF + f;
		xyz.z = walkParms.backS + s + (left?-walkParms.sideOffset:walkParms.sideOffset);
		xyz.y = walkParms.backH - h; 
	}
	
	if (front) {
		invKinematics(xyz, tLeg, fbLegPadCenter, joints);
	} else {
		// inverseKinematicsBack(xyz, joints);
		xyz.x *= -1;
		invKinematics(xyz, tLeg, bbLegPad, joints);
	}
	handleTilt(joints, front);
}

void SkellipticalWalk::setCurrentJointCommand(JointCommand &jointCommand) {
	JOINT_Values leg;
	
	double step_ = step + time/2;
	if (step_ > time)
		step_ -= time;
	
	if (standing) {	// handle standing still
		step = step_ = 0;
	}
	
	getLocusJoints(step, true, true, leg);
    jointCommand.setJointEnumValue(leftFrontJoint, RAD2MICRO(leg.rotator) ); 
    jointCommand.setJointEnumValue(leftFrontShoulder, RAD2MICRO(leg.shoulder) ); 
    jointCommand.setJointEnumValue(leftFrontKnee, RAD2MICRO(leg.knee) ); 
	
	getLocusJoints(step_, true, false, leg);
    jointCommand.setJointEnumValue(rightFrontJoint, RAD2MICRO(leg.rotator) );
    jointCommand.setJointEnumValue(rightFrontShoulder, RAD2MICRO(leg.shoulder) ); 
    jointCommand.setJointEnumValue(rightFrontKnee, RAD2MICRO(leg.knee) ); 
	
	getLocusJoints(step_, false, true, leg);
    jointCommand.setJointEnumValue(leftRearJoint, RAD2MICRO(leg.rotator) ); 
    jointCommand.setJointEnumValue(leftRearShoulder, RAD2MICRO(leg.shoulder) ); 
    jointCommand.setJointEnumValue(leftRearKnee, RAD2MICRO(leg.knee) ); 

	getLocusJoints(step, false, false, leg);
    jointCommand.setJointEnumValue(rightRearJoint, RAD2MICRO(leg.rotator) ); 
    jointCommand.setJointEnumValue(rightRearShoulder, RAD2MICRO(leg.shoulder) ); 
    jointCommand.setJointEnumValue(rightRearKnee, RAD2MICRO(leg.knee) ); 
}
