/*

   Copyright 2004 The University of New South Wales (UNSW) and National  
   ICT Australia (NICTA).

   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
   redistribute it and/or modify it under the terms of the GNU General  
   Public License as published by the Free Software Foundation; either  
   version 2 of the License, or (at your option) any later version as  
   modified below.  As the original licensors, we add the following  
   conditions to that license:

   In paragraph 2.b), the phrase "distribute or publish" should be  
   interpreted to include entry into a competition, and hence the source  
   of any derived work entered into a competition must be made available  
   to all parties involved in that competition under the terms of this  
   license.

   In addition, if the authors of a derived work publish any conference  
   proceedings, journal articles or other academic papers describing that  
   derived work, then appropriate academic citations to the original work  
   must be included in that publication.

   This rUNSWift source is distributed in the hope that it will be useful,  
   but WITHOUT ANY WARRANTY; without even the implied warranty of  
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
   General Public License for more details.

   You should have received a copy of the GNU General Public License along  
   with this source code; if not, write to the Free Software Foundation,  
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: SpecialAction.h 4364 2004-09-22 07:04:12Z kcph007 $
 *
 * Copyright(c) 2003 UNSW
 * All Rights Reserved.
 *
 **/

#ifndef _SpecialAction_h_DEFINED
#define _SpecialAction_h_DEFINED

#include <fstream>
#include <iostream>
#include "../share/PWalkDef.h"
#include "JointCommand.h"
#include "BaseAction.h"

using namespace std;

/* posa = position array, format (double):
   duration,
   head tilt, pan, roll,
   right front joint, shoulder, knee,
   left front joint, shoulder, knee,
   right hind joint, shoulder, knee,
   left hind joint, shoulder, knee
   */


class SpecialAction : public BaseAction {
	protected:
		int numOfSteps;
		int duration;
		int step;
		int currIndex;

		bool actionCompleted;

		double *posa;
		const char *filename;

	public:
		SpecialAction() {
			numOfSteps = 0;
			duration = 0;
			posa = NULL;
			reset();
		}

		virtual ~SpecialAction() {
			delete(posa);
		}

		virtual void reset(){
			actionCompleted = false;
			currIndex = 0;
			step = 0;
		}

        virtual void load(const string &str){
            load(str.c_str());
        }

		virtual void load(const char *filename) {
			this->filename = filename;

			ifstream in(filename);
			numOfSteps = 0;
			duration = 0;
            if (posa)
                delete(posa);

			if (in) {
				in >> numOfSteps;

				if (numOfSteps > 0) {
					posa = (double *)(new double[numOfSteps*posaLength]);

					for (int i = 0; i < numOfSteps; ++i) {
						for (int j = 0; j < posaLength; ++j) {
							in >> posa[i*posaLength+j];
						}
						//WARNING: 
						//posa[i*posaLength] *= 5;
						duration += (int) posa[i*posaLength];
						posa[i*posaLength] = duration;
					}
				}
				cout << "Step duration " << duration << endl;
				in.close();
			}
		}

		void reload(){
			load(this->filename);
		}

		virtual double getCurrentJointValue(int jointIndex){
			return posa[currIndex * posaLength + jointIndex];
		}

		virtual void setCurrentJointCommand(JointCommand &j) {
			for (int i=1; i<posaLength; ++i) {
				j.setJointValue(i-1,DEG2MICRO(getCurrentJointValue(i)) );
			}
		}

		virtual void getOdometry(double &delta_forward, double &delta_left, double &delta_turn, bool highGain) {
			delta_forward = delta_left = delta_turn = 0;
            highGain = false;
		}

		virtual void goNextFrame(){
			step ++;
			if (step >= duration) {
				step = 0;
				currIndex = 0;
				actionCompleted = true;
			}
			else{
				if (step >= posa[currIndex * posaLength] ){
					currIndex ++;
				}
			}
			//cout << "now in step " << step << " ind = " << currIndex << endl;
		}

		virtual bool isCompleted(){
			return actionCompleted;
		}

		virtual int getNumPoints() {
			return (int)duration;
		}
		
		virtual void setPointNum(int number) {
			step = number;
            currIndex = 0;
            while (step >= posa[currIndex * posaLength] ){
                currIndex ++;
            }
            //cout << "got set number " << step << " currIndex " << endl;
		}

		virtual int getPointNum() {
			return step;
		}

		virtual bool usingHead() {
			return true;
		}

        virtual bool isSpecialAction(){
            return true;
        }

};


#endif //_SpecialAction_h_DEFINED

