/*

   Copyright 2004 The University of New South Wales(UNSW) and National  
   ICT Australia(NICTA).

   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
   redistribute it and/or modify it under the terms of the GNU General  
   Public License as published by the Free Software Foundation; either  
   version 2 of the License, or(at your option) any later version as  
   modified below.  As the original licensors, we add the following  
   conditions to that license:

   In paragraph 2.b), the phrase "distribute or publish" should be  
   interpreted to include entry into a competition, and hence the source  
   of any derived work entered into a competition must be made available  
   to all parties involved in that competition under the terms of this  
   license.

   In addition, if the authors of a derived work publish any conference  
   proceedings, journal articles or other academic papers describing that  
   derived work, then appropriate academic citations to the original work  
   must be included in that publication.

   This rUNSWift source is distributed in the hope that it will be useful,  
   but WITHOUT ANY WARRANTY; without even the implied warranty of  
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
   General Public License for more details.

   You should have received a copy of the GNU General Public License along  
   with this source code; if not, write to the Free Software Foundation,  
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: UNSWSkills.cc 4363 2004-09-22 06:58:24Z kcph007 $
 *
 * Copyright(c) 2004 UNSW
 * All Rights Reserved.
 *
 **/

#include "Behaviours.h"
#include "UNSW2004.h"
#include <sys/time.h>
#ifndef OFFLINE
#include <MCOOP.h>
#endif

#include "../vision/CommonSense.h"
#ifndef OFFLINE
#include "../vision/Vision.h"
#endif // OFFLINE
#include "../share/WirelessTypeDef.h"
#include <sstream>

static const bool delayDebugMsg = false;
static const bool bDebugGain = false;
static const bool bDebugDelayBug = false;

//TODO : figure out how to send burst command properly
static const int BURST_NUMBER            = 5;
static const int MOBILIZE_COUNT_HEAD     = 4;
static const int MOBILIZE_COUNT_BACK     = LARGE_VAL; //16; //ERIC CHANGED THIS TO DEACTIVATE BACK BUTTON
static const int DEMOBILIZE_COUNT_HEAD   = 24;
static const int DEMOBILIZE_COUNT_BACK   = LARGE_VAL; //96; //ERIC CHANGED THIS TO DEACTIVATE BACK BUTTON
static const int PRESSURE_RECOVERY_VALUE = -100;
static const int TAIL_KICKOFF            = 300000;

static const double DISTANCE_ACCURACY = 5;
static const double ANGLE_ACCURACY = 3;
static const double POSITION_VAR_ACCURACY = get95CF(20);

static const int DISPLAY_WORLD_MODEL_ALL = 9;
static const int NEWOBS_CHALLENGE        = 23;

static const int NUM_FRAMES_TO_PAN = 148;

static int yuvSaveNumber = -1;

static AtomicAction nextAA;
static int mouthCounter = 0;
static const int BALL_LOST = 5;

bool Behaviours::jointRelaxed = false;

#ifndef OFFLINE
// added on 4-04, since power is main concern in 2004
static OPowerStatus currentStatus;
#endif // OFFLINEn


#ifdef LOG_YUV
static const int logYUVFrequency = 25;
static int logYUVTimer = 0;
#endif

void saveYUVPhoto();

void Behaviours::UpdateMode() {
	// Detect and reset fallen down state
	if (sensors->hasFallen != NotFallenFS && mode_ != mGettingUp && *stepComplete) {
		modeBeforeFallen = mode_;
		mode_ = mGettingUp;
	} else if (mode_ == mGettingUp) {
		// Detect gotten up from fallen down
		if (sensors->hasFallen == NotFallenFS) {
			sensors->resetFallenState();
			switch(modeBeforeFallen) {
				case mDemobilized:
				case mStationaryLocalize:
				case mStartMatch:
					mode_ = mDemobilized;
					break;
				default:
					mode_ = mNormal;
					break;
			}
			// Reset();
		}
	}

	// Detect head or back press to mobilize and demobilize
	if (mode_ == mDemobilized) {
		if (sensors->press_sensor[PRESS_HEAD].count > MOBILIZE_COUNT_HEAD) {

            if (PLAYER == 1) { //HACK: magic number = Behaviours.cc:PHOTO_TAKER
                saveYUVPhoto();
			} 
            else if (minimalModes) {
                mode_ = mNormal;
                //Reset();
            } else {
                mode_ = mStationaryLocalize;
                actionTimer = NUM_FRAMES_TO_PAN;
            }
        }

	} else if (mode_ == mStationaryLocalize) {
		if (actionTimer == 0) {
			mode_ = mStartMatch;
			// Produce_Sound(0);
		}
	} else if (mode_ == mStartMatch) {
		if (sensors->press_sensor[PRESS_HEAD].count > MOBILIZE_COUNT_HEAD) {
			cerr << "games begun" << endl;
			mode_ = mNormal;
			//Reset();
			//sensors->press_sensor[PRESS_HEAD].count = PRESSURE_RECOVERY_VALUE;
		}
	} 

	else if (sensors->press_sensor[PRESS_HEAD].count > DEMOBILIZE_COUNT_HEAD) {
        if (!minimalModes) {
    		mode_ = mDemobilized;
	    	//sensors->press_sensor[PRESS_HEAD].count = PRESSURE_RECOVERY_VALUE;
		    Reset();
        }
	} 

	else if (mode_ == mKickoffPos) {

		double distDiff = sqrt(SQUARE(gps->self().pos.x-kickInix)
				+ SQUARE(gps->self().pos.y-kickIniy));
		double angDiff = ABS(gps->self().h - kickInih);

		if ((distDiff < DISTANCE_ACCURACY && angDiff < ANGLE_ACCURACY &&
					gps->self().posVar < POSITION_VAR_ACCURACY) ||
				(sensors->press_sensor[PRESS_BACK_FRONT].count > MOBILIZE_COUNT_HEAD)) {
			setStandParams();
			makeParaWalk(walkType, forward, left, turnCCW, PG, 
					hF, hB, hdF, hdB, ffO, fsO, bfO, bsO);
			//actionTimer = NUM_FRAMES_TO_PAN;
			//mode_ = mStationaryLocalize;
			mode_ = mStartMatch;
		}

	} else if (mode_ == mReceivePos) {

		double distDiff = sqrt(SQUARE(gps->self().pos.x-recvInix)
				+ SQUARE(gps->self().pos.y-recvIniy));
		double angDiff = ABS(gps->self().h - recvInih);

		if ((distDiff < DISTANCE_ACCURACY && angDiff < ANGLE_ACCURACY &&
					gps->self().posVar < POSITION_VAR_ACCURACY) ||
				(sensors->press_sensor[PRESS_BACK_FRONT].count > MOBILIZE_COUNT_HEAD)) {
			setStandParams();
			makeParaWalk(walkType, forward, left, turnCCW, PG, 
					hF, hB, hdF, hdB, ffO, fsO, bfO, bsO);
			//actionTimer = NUM_FRAMES_TO_PAN;
			//mode_ = mStationaryLocalize;
			mode_ = mStartMatch;
		}

	}
}
//Kim:06/03/04 This is the last calibrate, since the walk is changed, every thing need to be re-calibrate. This is kind of a hack to push my calibration. Remove it when we have stable walks
void ultimateCalibrate(AtomicAction &nextAA){

    if (nextAA.walkType == CanterWalkWT || nextAA.walkType == ZoidalWalkWT || 
            nextAA.walkType == OffsetWalkWT )
        nextAA.walkType = NormalWalkWT;

}

static bool isHighGain = false;

// void Behaviours::dynamic_gain (AtomicAction &next) {
//   /*
//     int hesteresisCounter = 0; 
//     if (hesteresisCounter > 0) hesteresisCounter --;

//     //only switch to high gain if the dog stays below current for at least 90 frames
//     if ( currentStatus.current < -3000.0){
//         hesteresisCounter = 90; //staty the same for at least 30 frames
//         if ( isHighGain)
//             setLowGain();
//     }
//     else if (next.walkType == EllipseWalkWT && !isHighGain && hesteresisCounter == 0){
//         setHighGain();
//         hesteresisCounter = 10; //just to avoid alternating too fast
//     }
//     else if (next.walkType != EllipseWalkWT && isHighGain && hesteresisCounter == 0){
//         setLowGain();
//     }
//     */

// #ifndef OFFLINE
//     if ( currentStatus.current > -2000.0 && !isHighGain ) {
//         setHighGain ();
//     }
  
//     if ( currentStatus.current < -3000.0 && isHighGain ) {
//         setLowGain();
//     }
// #endif // OFFLINE

//     //KEV_DEBUG (cout << "The current battery current is: " 
//     //      << currentStatus.current << "   am i using HI GAIN NOW " 
//     //      << isHighGain << endl;);

//     /*
//     if (! isSpecialAction(nextAA.walkType) && nextAA.Forward != 0 && nextAA.turnCCW != 0 && nextAA.Left == 0){
//         nextAA.walkType = EllipseWalkWT;
//     }
//     */
// }

//TODO:remove
// for testing: set default gain either low or high here
void Behaviours::setDefaultGain(){
    static bool first = true;
    if (first)
        setLowGain();

    first = false;
}

static bool modeDebug = false;

/* A higher level decision-making to determine which stage of a game the dog
** is in. Just started? Penalty? Getting up? etc. Next level depends on
** player. */ 
void Behaviours::NextAction() 
{
    if (modeDebug)
        cout << "The current mode is: " << mode_ << endl;

    #ifndef OFFLINE
	if ( (OPENR::GetPowerStatus(&currentStatus)) != oSUCCESS) {
	  cout << "ERROR in getting the battery status" << endl;
	}
    #endif // OFFLINE

	frame++;
	UpdateMode();
	initialize();

	// Keep track of how long we have lost the ball for.
	if (gps->canSee(vobBall)) {        
		lostBall = 0;        
	} 
	else {
		lostBall++;            
	}

	int inds[NUM_OF_INDICATORS];

    // What to do for each mode_?
	switch(mode_) {

        // The very beginning when the dog just booted up.
		case mDemobilized:
			// Show battery status when the robot is not moving.
			batteryLEDS();

			/// The first position we want the dog to be after boot up.
			makeParaWalk(NormalWalkWT, 0, 0, 0, 60, 73, 107, 0, 0, 50, 15, -50, 15, ABS_H, 0, 0, 0);
			// makeParaWalk(0,2,0,0,60,73,107,0,0,50,15,-50,15,ABS_H,0,0);

			// Send now because the sending is disable in the code later because of conflict between python behaviour.
			makeIndicators(inds);

            #ifndef OFFLINE
			transmission->sendIndicators(inds, sizeof(inds));
            #endif // OFFLINE

			break;

		case mGettingUp:
			if (sensors->hasFallen == FallenFrontFS)
				makeParaWalk(GetupFrontWT,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);
			else if (sensors->hasFallen == FallenRearFS)
				makeParaWalk(GetupRearWT,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);
			else if (sensors->hasFallen == FallenSideFS)
				makeParaWalk(GetupSideWT,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);
			break;

		case mStationaryLocalize:
			StationaryLocalize();

			// Show battery status when the robot is not moving.
			batteryLEDS();

			/// The first position we want the dog to be after boot up.
			makeParaWalk(NormalWalkWT, 0, 0, 0, 60, 73, 107, 0, 0, 50, 15, -50, 15, ABS_H, 0, 0, 0);
			// makeParaWalk(0,2,0,0,60,73,107,0,0,50,15,-50,15,ABS_H,0,0);

			// Send now because the sending is disable in the code later because of conflict between python behaviour.
			makeIndicators(inds);
            #ifndef OFFLINE
			transmission->sendIndicators(inds, sizeof(inds));
            #endif // OFFLINE
			break;

		case mKickoffPos:
		
		/* Ted removed because these variables are removed. Come back later.

			// show battery status
			BatteryLEDS();
			kickingOff =(sqrt(SQUARE(kickInix-FIELD_WIDTH/2) + SQUARE(kickIniy-FIELD_LENGTH/2)) < 20);
			smartSetBeacon();
			smartActiveGpsLocalise();
			//saGoToTargetFacingHeading(kickInix, kickIniy, kickInih);
			UNSW2004::doFastGoToPointHeading(kickInix, kickIniy, kickIniy);
			setWalkParams();
			makeParaWalk(walkType, forward, left, turnCCW, PG, 
					hF, hB, hdF, hdB, ffO, fsO, bfO, bsO);
			makeHead(headtype, panx, tilty, cranez);
			
		*/
			break;

		case mReceivePos:

		/* See the comment above.
		
			// show battery status
			BatteryLEDS();

			kickingOff = false;
			smartSetBeacon();
			smartActiveGpsLocalise();
			//saGoToTargetFacingHeading(recvInix, recvIniy, recvInih);
			UNSW2004::doFastGoToPointHeading(recvInix, recvIniy, recvInih);
			setWalkParams();
			makeParaWalk(walkType, forward, left, turnCCW, PG, 
					hF, hB, hdF, hdB, ffO, fsO, bfO, bsO);
			makeHead(headtype, panx, tilty, cranez);
	
		*/
			break;

		// Ted: I removed the case statement for mStartMatch. Revision 3213.

		// mNormal, mStartMatch, mReadyMatch, mPythonDemobilized.
		default:

			// Not sure the following linux time evaluation functions work in OffVision.
            bool debugProcessTime = false;
            double t1, t2;
            
            if (flag_OFFLINE && debugProcessTime) {
                timeval start;
                gettimeofday(&start, NULL);
                t1 = start.tv_sec + (start.tv_usec);
            }

			DecideNextAction();

            if (mode_ == mStartMatch) {
                if (standTall)
                    hF = 85;			

                makeParaWalk(walkType, forward, left, turnCCW, PG, hF, hB,
                        hdF, hdB, ffO, fsO, bfO, bsO, headtype, panx, tilty, cranez);			
            }

            if (flag_OFFLINE && debugProcessTime) {
                timeval end;
                gettimeofday(&end, NULL);
                t2 = end.tv_sec + (end.tv_sec);
                cout << "Time took by behaviour: " << t2 - t1 << " and the player number is: " << PLAYER_NUMBER << endl;
            }

	}

    // Set all the neccessary indicators.
    // superDebug();
    // Don't want to overwrite debug indicators set in python ones.

    makeHead(headtype, panx, tilty, cranez);
        
	// send information for other dogs
#ifndef OFFLINE
	if (mode_ != mDemobilized && gameState.state != ROBOCUP_STATE_PENALIZED) {
		transmission->shareEnvironment();
	}
#endif // OFFLINE

	// send any debug information
	// send cplane data
	if (cmdCPlane == PLAYER_NUMBER) {
#ifndef OFFLINE
		transmission->sendCompressedCPlane();
#endif // OFFLINE
	}
	// send burst data
	if (cmdBurst == PLAYER_NUMBER && Behaviours::burstCounter > 0 ) {
#ifndef OFFLINE
		transmission->sendCompressedCPlane();
		burstCounter--;
#endif // OFFLINE
	}

	// send oplane data
	if (cmdOPlane == PLAYER_NUMBER) {
#ifndef OFFLINE
		transmission->sendOPlane();
#endif // OFFLINE
	}

	// send world model stuff
	if (cmdWDisplay == PLAYER_NUMBER || cmdWDisplay == DISPLAY_WORLD_MODEL_ALL) {
#ifndef OFFLINE
		transmission->displayEnvironment();
#endif // OFFLINE
	}

#ifdef LOG_YUV
	if (logYUVTimer == logYUVFrequency) {
		transmission->sendYUVPlane();
		logYUVTimer = 0;            
	} else {
		logYUVTimer++;            
	}            
#endif

#ifndef OFFLINE
	if (transmission->gotTMWM) {
		// turn bar on
		wireless = IND_LED2_ON;
		transmission->gotTMWM = 0;
	} else {
		// turn bar off
		wireless = IND_LED2_OFF;
	}

	// if there are three or more teammates with valid world models turn bar on
	int tm = 0;
	//cout << "friends = {";
	for (int i=0 ; i<NUM_TEAM_MEMBER ; i++) {
		if (gps->tmObj(i).counter > 0) {
			//cout << i+1 << " ";
			tm++;
		}
	}
	//cout << "}" << endl;
	if (tm >= 3) {
		// Please do not use this anymore, it does not work on the new dog.
		wireless = IND_LED2_ON;
		tail_h = IND_TAIL_H_CENTRED;
	} 
	else if (tm > 0) {
		tail_h = IND_TAIL_LEFT;
	} 
	else {
		tail_h = IND_TAIL_RIGHT;
	}

    /* Want to show indicators for remote player. */ 
    if (PLAYER == 0) { // hack: 0 is remote player.
        //int inds[NUM_OF_INDICATORS];
        makeIndicators(inds);
        transmission->sendIndicators(inds, sizeof(inds));
    }

	ultimateCalibrate(nextAA);
	//dynamic_gain (nextAA);
    //setDefaultGain();
	lastHPan = hPan;
	lastHTilt = hTilt;
	lastHCrane = hCrane;
	
    if (bDebugDelayBug){
        cout << "sending " << nextAA.walkType << endl;
    }
	transmission->sendAA(nextAA);
#endif // OFFLINE
}

void Behaviours::showTeamIndicator() {
    // If I am a blue dog.
	if (gps->targetGoal == vobYellowGoal) {
	    back_rear_color  = IND_LED3_INTENSITY_MAX;
	    back_rear_white  = IND_LED3_INTENSITY_MIN;
        back_front_color = IND_LED3_INTENSITY_MIN;
        back_front_white = IND_LED3_INTENSITY_MIN;        
    }
    // If we are red dog.   
    else {
	    back_rear_color  = IND_LED3_INTENSITY_MIN;
	    back_rear_white  = IND_LED3_INTENSITY_MIN;
        back_front_color = IND_LED3_INTENSITY_MAX;
        back_front_white = IND_LED3_INTENSITY_MIN;        
    }
}

void Behaviours::batteryLEDS() {
#ifndef OFFLINE
	OStatus result;
	OPowerStatus currentStatus;
	result = OPENR::GetPowerStatus(&currentStatus);
	
	if (result == oSUCCESS) {
		word status = currentStatus.remainingCapacity;		
		if (status >= 100)
			showEyes(1, 1, 1, 1, 1);
		else if (status >= 80)
			showEyes(0, 1, 1, 1, 1);
		else if (status >= 60)
			showEyes(0, 1, 1, 1, 0);
		else if (status >= 40)
			showEyes(0, 0, 1, 1, 0);
		else if (status >= 20)
			showEyes(0, 0, 1, 0, 0);
		else
			showEyes(0, 0, 0, 0, 0);	
	}
	else {
		cout << "I can't display leds for battery in UNSWSkils.cc::BatteryLEDS()!" << endl;	
	}
#endif
}

void Behaviours::StationaryLocalize() {
	if (actionTimer > 0) {
		if (ABS(hPan - destPan) <= 10) {
			destPan =(destPan < 0) ? 85 : -85;
		}
		headtype = REL_H;
		panx =(destPan > 0) ? 5 : -5;
		tilty =(10 - hTilt) / 5.0;
		cranez = -hCrane / 4.0;
		makeHead(headtype, panx, tilty, 0);
		actionTimer--;
	}
}

/* Note:

** 1. whichJoint is according to JointEnum in ../share/PWalkDef.h
** 2. How come a makeParaWalk is different to makeRelaxJoint,
**    makeEnableJoint, setDogRelaxed?
**
**    Actually, they are the same. But at the end of the behav loop,
**    makeParaWalk is called, which sets the nextAA.cmd to walk,
**    so one way to have our own cmd working is to create an ActomicAction
**    and send it all ourselves, in the exact way that the walk does 
*/

void Behaviours::setLowGain ()
{
        isHighGain = false;
        AtomicAction aa ;
	aa.cmd      = aaSetLowGain;
	//cout << "eWHICH JOINTS: "<< aa.walkType << endl;
#ifndef OFFLINE
	transmission->sendAA(aa);
#endif //OFFLINE
}

void Behaviours::setHighGain ()
{
        isHighGain = true;
        AtomicAction aa ;
	aa.cmd      = aaSetHighGain;
	//cout << "eWHICH JOINTS: "<< aa.walkType << endl;
#ifndef OFFLINE
	transmission->sendAA(aa);
#endif //OFFLINE
}

void Behaviours::makeRelaxAJoint (int whichJoint)
{
        makeSetJointGain (whichJoint,0,0,0);
}

void Behaviours::makeEnableAJoint (int whichJoint)
{
        makeSetJointGain(whichJoint,1000,1000,1000);
}

void Behaviours::makeSetJointGain (int whichJoint, int P , int I , int D)
{
  	AtomicAction aa ;
	aa.cmd      = aaSetAJointGain;
	aa.walkType = whichJoint;
	aa.panx     = P;
	aa.tilty    = I;
	aa.cranez   = D;
	//cout << "eWHICH JOINTS: "<< aa.walkType << endl;
#ifndef OFFLINE
	transmission->sendAA(aa);
#endif //OFFLINE
}


void Behaviours::makeParaWalk(int walkType, double Forward, double Left,
		double turnCCW, int speed, double hF,
		double hB, double hdF, double hdB,
		double ffO, double fsO, double bfO, double bsO) {
	nextAA.cmd = acCmd;
	nextAA.walkType = walkType;
	nextAA.Forward = Forward;
	nextAA.Left = Left;
	nextAA.turnCCW = turnCCW;
	nextAA.speed = speed;
	nextAA.hF = hF;
	nextAA.hB = hB;
	nextAA.hdF = hdF;
	nextAA.hdB = hdB;
	nextAA.ffO = ffO;
	nextAA.fsO = fsO;
	nextAA.bfO = bfO;
	nextAA.bsO = bsO;
	nextAA.headType = NULL_H;
	defaultInit();
}

void Behaviours::makeParaWalk(int walkType, double Forward, double Left,
		double turnCCW, int speed, double hF,
		double hB, double hdF, double hdB,
		double ffO, double fsO, double bfO,
		double bsO, int headType, double panx, double tilty, double cranez) {
	nextAA.cmd = acCmd;
	nextAA.walkType = walkType;
	nextAA.Forward = Forward;
	nextAA.Left = Left;
	nextAA.turnCCW = turnCCW;
	nextAA.speed = speed;
	nextAA.hF = hF;
	nextAA.hB = hB;
	nextAA.hdF = hdF;
	nextAA.hdB = hdB;
	nextAA.ffO = ffO;
	nextAA.fsO = fsO;
	nextAA.bfO = bfO;
	nextAA.bsO = bsO;
	nextAA.headType = headType;
	nextAA.tilty = tilty;
	nextAA.panx = panx;
	nextAA.cranez = cranez;
	//  nextAA.mouth = mouth;
	defaultInit();
}

void Behaviours::makeHead(int headType, double panx, double tilty, double cranez) {

    const string debugHead = "Behaviours::makeHead| ";

    //cout << debugHead << "headType: " << headType << " panx: " << panx << " tilty: " 
    //    << tilty << "cranez: " << cranez << endl;

	nextAA.cmd = acCmd;
	//  nextAA.walkType =1000;
	nextAA.headType = headType;
	nextAA.panx = panx;
	nextAA.tilty = tilty;
	nextAA.cranez = cranez;
	//  nextAA.mouth = 0;
	defaultInit();
}

void Behaviours::defaultInit(void) {
	// common initialization for all commands
	nextAA.step = 0;
	nextAA.finished = false;
}

double Behaviours::microRad2Deg(double rad) {
	return rad * 180.0 /(1000000.0 * pi);
}

// given an object at(x,y), find its heading,distance
// relative to the robot(rx,ry,rh)
void Behaviours::findAngleDist(double x, double y, double *head, double *dist) {

	const WMObj& r = gps->self();
	*dist = sqrt((r.pos.x - x) *(r.pos.x - x) +(r.pos.y - y) *(r.pos.y - y));
	*head = asin(abs(r.pos.y - y) / *dist) * 180 / pi;
	if (y >= r.pos.y) {
		if (x <= r.pos.x)
			*head = 180 - *head;
	} else {
		if (x >= r.pos.x)
			*head *= -1;
		else
			*head -= 180;
	}
	*head -= r.h;
	while(*head > 180) {
		*head -= 360;
	}
	while(*head <= -180) {
		*head += 360;
	}
}

// cleaned up code by using operator= and several
// object operations to make it simple
void Behaviours::initialize(void) 
{
	// Setting player(team and goalie) indicators
	if (playerIndicatorNotSet) {

		// Orange indicator for Red robots
		mode_blue = (gps->targetGoal == vobYellowGoal) ? IND_LED2_ON : IND_LED2_OFF;

		// Blue indidcator for Blue robots
		mode_red = (gps->targetGoal == vobYellowGoal) ? IND_LED2_OFF : IND_LED2_ON;

		// Indicator for Goalie
		wireless = IND_LED2_OFF;
		playerIndicatorNotSet = false;
	}
	// ball from current image
	if (vision->vob[vobBall].cf == 0) {
		vBall.vob = &(vision->vob[vobBall]);
		vBall.reset();
	} else {
		vBall = vision->vob[vobBall];

		const WMObj& r = gps->self();
		vBall.calcXY(r.pos.x, r.pos.y, r.h);
	}

	//!!ERIC MAKE THIS COME BACK LATER
	//!!ERIC MAKE THIS COME BACK LATER
	//!!ERIC MAKE THIS COME BACK LATER
	/*
	// ball from world model
	// can't use x and y in wmobj, because they are not reversed for the blue team
	wmBall = wmobj[vobBall];
	wmBall.calcXY(rx, ry, rh);

	//wout << "gps best ball x, y, h, d " << wlBestBall.x << "," << wlBestBall.y << "," << wlBestBall.h << "," << wlBestBall.d << "\n";
	wlBall = wlBestBall;
	wlBall.calcXY(rx, ry, rh);

	// tGoal -targetGoal- from the world model
	wmtGoal.x = FIELD_WIDTH / 2;
	wmtGoal.y = FIELD_LENGTH + GOAL_LENGTH / 2;    // these are fixed
	//wmtGoal.calcHD(rx,ry);
	findAngleDist(wmtGoal.x, wmtGoal.y, &wmtGoal.h, &wmtGoal.d);
	wmtGoal.cf = rcf;
	wmtGoal.var = rvar;

	// oGoal -ownGoal- from the world model
	wmoGoal.x = FIELD_WIDTH / 2;
	wmoGoal.y = -GOAL_LENGTH / 2;
	//wmoGoal.calcHD(rx,ry);
	findAngleDist(wmoGoal.x, wmoGoal.y, &wmoGoal.h, &wmoGoal.d);
	wmoGoal.cf = rcf;
	wmoGoal.var = rvar;

*/

	// tGoal from current image
	if (vision->vob[gps->targetGoal].cf == 0) {
		vtGoal.reset();
	} else {
		// tGoal from current image
		vtGoal = vision->vob[gps->targetGoal];
		vtGoal.x = FIELD_WIDTH / 2;
		vtGoal.y = FIELD_LENGTH + GOAL_LENGTH / 2;
	}

	// oGoal from current image
	if (vision->vob[gps->ownGoal].cf == 0) {
		voGoal.reset();
	} else {
		voGoal = vision->vob[gps->ownGoal];
		voGoal.x = FIELD_WIDTH / 2;
		voGoal.y = -GOAL_LENGTH / 2;
	}

	// convert head sensors to degrees
	hTilt  = microRad2Deg(sensors->sensorVal[ssHEAD_TILT1]);
	hPan   = microRad2Deg(sensors->sensorVal[ssHEAD_PAN]);
	//hRoll  = microRad2Deg(sensors->sensorVal[ssMOUTH]);
	hCrane = microRad2Deg(sensors->sensorVal[ssHEAD_TILT2]);

	//cout << "hTilt: " << hTilt  <<"    hPan: " << hPan << "    hCrane: " << hCrane << endl;

	// other robots
	int wmOpponentBase = vobBlueDog, wmTeamBase = vobRedDog;
	int wlOpponentBase = NUM_TEAM_MEMBER, wlTeamBase = 0;
	if (gps->targetGoal == vobYellowGoal) {
		wmOpponentBase = vobRedDog, wmTeamBase = vobBlueDog;
		wlOpponentBase = 0, wlTeamBase = NUM_TEAM_MEMBER;
	}
	int vOpponentBase = wmOpponentBase, vTeamBase = wmTeamBase;

	// this is needed for visual backoff!
	for (int i = 0; i < NUM_TEAM_MEMBER; i++) {
		// team members from vision
		vTeammate[i] = vision->vob[vTeamBase++];
		vTeammate[i].calcXY(gps->self().pos.x, gps->self().pos.y, gps->self().getHackedH());
		// opponent members from vision
		vOpponent[i] = vision->vob[vOpponentBase++];
		vOpponent[i].calcXY(gps->self().pos.x, gps->self().pos.y, gps->self().getHackedH());
	}            

	//!!ERIC MAKE THIS COME BACK LATER
	//!!ERIC MAKE THIS COME BACK LATER
	//!!ERIC MAKE THIS COME BACK LATER
	/*
	   for (int i = 0; i < NUM_TEAM_MEMBER; i++) {
	// team members as told by them
	teammate[i] = wlTeam[i];
	teammate[i].calcXY(rx, ry, rh);

	// team members from vision
	vTeammate[i] = vision->vob[vTeamBase++];
	vTeammate[i].calcXY(rx, ry, rh);

	// teammates from world model
	wmTeammate[i] = wmobj[wmTeamBase++];
	wmTeammate[i].calcXY(rx, ry, rh);

	// teammates from other ppl and your own world model
	wlTeammate[i] = wlRobots[wlTeamBase++];
	wlTeammate[i].calcXY(rx, ry, rh);


	// opponent from world model
	wmOpponent[i] = wmobj[wmOpponentBase++];
	wmOpponent[i].calcXY(rx, ry, rh);

	// opponent from other ppl and your own world model
	wlOpponent[i] = wlRobots[wlOpponentBase++];
	wlOpponent[i].calcXY(rx, ry, rh);
	}
	*/

	standTall = false;
}

// Kevin: I removeed the makeNormalWalk , makeEdgeWalk , makeKickWalk here,
// I belive they are ancient stuff get it back at r2648

void Behaviours::makeIndicators(int inds[])
{
	inds[0]  = mouth;
	inds[1]  = tail_v;
	inds[2]  = tail_h;
	inds[3]  = ear_l;
	inds[4]  = ear_r;   
	inds[5]  = led1  | mode_led1;
	inds[6]  = led2  | mode_led2;
	inds[7]  = led3  | mode_led3;
	inds[8]  = led4  | mode_led4; 
	inds[9]  = led5  | mode_led5;
	inds[10] = led6  | mode_led6;
	inds[11] = led7  | mode_led7;
	inds[12] = led8  | mode_led8;
	inds[13] = led9  | mode_led9;
	inds[14] = led10 | mode_led10;
	inds[15] = led11 | mode_led11;
	inds[16] = led12 | mode_led12;
	inds[17] = led13 | mode_led13;
	inds[18] = led14 | mode_led14;
	inds[19] = head_color;
	inds[20] = head_white;
	inds[21] = mode_red;
	inds[22] = mode_green;
	inds[23] = mode_blue;
	inds[24] = wireless;
	inds[25] = back_rear_color;
	inds[26] = back_rear_white;
	inds[27] = back_middle_color;
	inds[28] = back_middle_white;
	inds[29] = back_front_color;
	inds[30] = back_front_white;
}

void Behaviours::wreceive(int from, int type, int size, const void *data) {
	switch(type) {
		case WL_SHARE_OBJ:
			//received a worldmodel packet from teammate
			//gps->processShareInfo(from,(const WMShareObj *) data);
			break;

		case WL_NNMC:
			vision->wlLoadNNMC((const unsigned char *) data, size);
			break;

		case WL_COMMAND:
			/*strcpy(cmdName,(const char *) data);

			  if (delayDebugMsg) {
			  cout << "UNSWSkills.cc name = " << cmdName << endl;
			  }*/

			break;

		case WL_VALUE:
			/*strcpy(cmdValue,(const char *) data);
			// check for mode modification
			if (strcmp(cmdName, "ip") == 0) {
			PLAYER_NUMBER = atoi(cmdValue);
			cmdName[0] = '\0';
			cmdValue[0] = '\0';
			cout << "UNSWSkills.cc My Player Number = " << PLAYER_NUMBER << endl;
			transmission->sendPlayerNumber(PLAYER_NUMBER);

			} else if (strcmp(cmdName, "ko") == 0) {
			int value = atoi(cmdValue);                            
			if (value == 1) {
			utherKickingOff = true;
			} else {
			utherKickingOff = false;
			} 
			} else if (strcmp(cmdName, "mo")==0 || strcmp(cmdName, "mode")==0) {
			mode_       = atoi(cmdValue);
			cmdName[0]  = '\0';
			cmdValue[0] = '\0';

			if (mode_==mStationaryLocalize) {
			actionTimer = NUM_FRAMES_TO_PAN;
			}
			}

			if (delayDebugMsg) {
			cout << "UNSWSkills.cc value = " << cmdValue << endl;
			}*/

			break;
	}
}

static bool debugProcessCommand = true;

void Behaviours::processCommand(const CommandData& cmd) {

    if (debugProcessCommand) {
        cout << "processCommand() " << cmd.RC_COMMAND << " " << cmd.rawValue << endl;
    }

    // const string debugHead = "Behaviours::processCommand| ";    
	// cout << debugHead << "Got wireless command of size " << sizeof(cmd) << " bytes." << endl;
    /* cmd.type: CommandType. */
    /* cmd.RC_COMMAND: Instruction. */ 
    /* Both in share/CommandData.h. */
	// cout << debugHead << "cmd.type = " << cmd.type << endl;
	// cout << debugHead << "cmd.RC_COMMAND = " << cmd.RC_COMMAND << endl;

	if (cmd.type == ROBOCOMMANDER_DATA) {
		if (cmd.RC_COMMAND == UTHER_KICKING_OFF) {
			/*            
				      int kickOff =(int)(cmd.value);
				      if (kickOff) {
				      utherKickingOff = true;
				      } else {
				      utherKickingOff = false;
				      }
		    */        
		} else if (cmd.RC_COMMAND == MODE_CHANGE) {
			mode_ = (StateModes)(cmd.value);

			if (mode_==mStationaryLocalize) {
				actionTimer = NUM_FRAMES_TO_PAN;
			}
			command.clear();

		} 
		// setting the sending of debug, moved from Behaviours.cc
		else if (cmd.RC_COMMAND == CPLANE_SEND) {
			cmdCPlane =(int)(cmd.value);
			cout << "Got CPLANE for " << cmdCPlane << endl;
		}
		// setting the sending of debug
		else if (cmd.RC_COMMAND == BURST_MODE) {
			cmdBurst =(int)(cmd.value);
			burstCounter = BURST_NUMBER;
		}

		else if (cmd.RC_COMMAND == WM_SEND) {
			cmdWDisplay =(int)(cmd.value);
		}

		else if (cmd.RC_COMMAND == YUVPLANE_SEND) {
#ifndef OFFLINE            
			transmission->sendYUVPlane();
#endif //OFFLINE            
		}

		else if (cmd.RC_COMMAND == OPLANE_SEND) {
			cmdOPlane =(int)(cmd.value);
		}
		else
			command = cmd;
	} else { // cmd.type == GAMECONTROLLER_DATA

		if ((cmd.GC_DATA.state == ROBOCUP_STATE_FINISHED) &&
			(cmd.GC_DATA.ownScore > gameState.ownScore)) {

			lockMode = CelebrationLock;

		}

        // This function added by Ted Wong.
        processControllerData(cmd.GC_DATA);

    }

}

// Process data received from RoboCupGameController. Work for both new
// and old controller.
//
// Ted: The gamecontroll will be in C++ at least at the moment. The problem porting to python is that unless we
//      change the behaviour switch statements, it makes no sense to send gamecontroller data to the python
//      directly. For example, if gameState.State is ROBOCUP_STATE_INITIAL. Python received this command, but what
//      to do? The robot stopping code is handled in C++, python need to send some command back to C++ in order to
//      stop the robot.
void Behaviours::processControllerData(RoboCupGameControlData gameState) {

	// Receive signal from the gamecontroller whether my team is kicking off.
	kickOff = gameState.kickoff;

    // Look at the game state.
    switch(gameState.state) {
            
        case ROBOCUP_STATE_INITIAL:
            if (debugProcessCommand)
                cout << "Receive initial state!" << endl;
			 gps->setPenalized(false);
			 changeMode(mDemobilized);
			 actionTimer = NUM_FRAMES_TO_PAN;
             break;
  
        case ROBOCUP_STATE_READY:
            if (debugProcessCommand)
                cout << "Receive ready state!" << endl;
            gps->setPenalized(false);
            changeMode(mReadyMatch);
            break;

        case ROBOCUP_STATE_SET:
            if (debugProcessCommand)
                cout << "Receive set state!" << endl;
		    gps->setPenalized(false);
			changeMode(mStartMatch);                
			break;                            

        case ROBOCUP_STATE_PLAYING:
            if (debugProcessCommand)
                cout << "Receive playing state!" << endl;
			gps->setPenalized(false);
		    changeMode(mNormal);
			break;

        case ROBOCUP_STATE_PENALIZED:
            if (debugProcessCommand)
                cout << "Receive penalized state!" << endl;
			setPenaltyPos();
			gps->setPenalized(true);
			changeMode(mDemobilized);
			break;


		// Switch mode to python's mDemobilized. This mode is really same as C++ mDemobilized, except done in python. Useful
		// when debugging strategy - indicatos remain the same when the robot pause.
        case ROBOCUP_STATE_FINISHED:
            if (debugProcessCommand)
                cout << "Receive finished state!" << endl;
			changeMode(mPythonDemobilized);
		    break;

        cout << "The game state is not RState!" << endl;

    }

    // Look at the color.
    switch(gameState.teamColor) {            
        // Initial state. Not allowed to move.
        case ROBOCUP_TEAMCOLOR_RED:
            if (debugProcessCommand)
                cout << "Receive Red Color!" << endl;
			gps->SetGoals(true);

#ifndef OFFLINE
		    transmission->sendPlayerNumber(PLAYER_NUMBER, true);
#endif

            break;
        
        case ROBOCUP_TEAMCOLOR_BLUE:
            if (debugProcessCommand)
                cout << "Receive Blue color!" << endl;
            gps->SetGoals(false);

#ifndef OFFLINE
		    transmission->sendPlayerNumber(PLAYER_NUMBER, false);
#endif
            break;
    }
    
    // Look at the kickoff.
    switch(gameState.kickoff) {
        case ROBOCUP_KICKOFF_INVALID:
            if (debugProcessCommand)
                cout << "ROBOCUP_KICKOFF_INVALID received!" << endl;
            break;
            
        case ROBOCUP_KICKOFF_OWN:
            if (debugProcessCommand)
                cout << "ROBOCUP_KICKOFF_OWN received!" << endl;
            break;
        
        case ROBOCUP_KICKOFF_OPPONENT:
            if (debugProcessCommand)
                cout << "ROBOCUP_KICKOFF_OPPONENT received!" << endl;
            break;               
    }

    // Look at the penalty.
    switch(gameState.penalty) {
        case PENALTY_NONE:
            if (debugProcessCommand)
                cout << "PENALTY_NONE received!" << endl;
            break;
        case PENALTY_BALL_HOLDING:
            if (debugProcessCommand)
                cout << "PENALTY_BALL_HOLDING received!" << endl;
            break;
        case PENALTY_KEEPER_CHARGE:    
            if (debugProcessCommand)
                cout << "PENALTY_KEEPER_CHARGE received!" << endl;
            break;
        case PENALTY_FIELD_PLAYER_CHARGE:
            if (debugProcessCommand)
                cout << "PENALTY_FIELD_PLAYER_CHARGE received!" << endl;
            break;
        case PENALTY_ILLEGAL_DEFENDER:
            if (debugProcessCommand)
                cout << "PENALTY_ILLEGAL_DEFENDER received!" << endl;
            break;
        case PENALTY_ILLEGAL_DEFENSE:
            if (debugProcessCommand)
                cout << "PENALTY_ILLEGAL_DEFENSE received!" << endl;
            break;
        case PENALTY_OBSTRUCTION:
            if (debugProcessCommand)
                cout << "PENALTY_OBSTRUCTION received!" << endl;
            break;
        case PENALTY_REQ_FOR_PICKUP:
            if (debugProcessCommand)
                cout << "PENALTY_REQ_FOR_PICKUP received!" << endl;
            break;    
    }


    // ownScore = gameState.ownScore
    // opponentScore = gameState.opponentScore.

}

void Behaviours::setPenaltyPos() {

	double angle = atan(180.0/(double)FIELD_WIDTH);
	double sinA = sin(DEG2RAD(angle));
	double cosA = cos(DEG2RAD(angle));
	double ev1 = get95CF(FIELD_WIDTH/2);
	double ev2 = get95CF(10);
	double ev3 = get95CF(10);

	// the eigenvectors for the first and third eigenvalues, in terms
	// of x and h
	double x1 = cosA;
	double h1 = sinA;
	double x2 = -sinA;
	double h2 = cosA;

	MMatrix3 newC(0);
	newC(0,0) = ev1*x1*x1 + ev3*x2*x2;
	newC(2,0) = ev1*x1*h1 + ev3*x2*h2;
	newC(0,2) = newC(1,0);
	newC(2,2) = ev1*h1*h1 + ev3*h2*h2;
	newC(1,1) = ev2;

	gps->GPSSetVals(FIELD_WIDTH/2,FIELD_LENGTH/2,QUARTER_CIRCLE,newC);
}

void saveYUVPhoto(){
    using Behaviours::vision;
    using Behaviours::leds;
    //save YUV plane to the stick.
    char * bflFile;
    bflFile =(char*)malloc(512);
    yuvSaveNumber ++;
    cout << "Saving YUV photo " << yuvSaveNumber << endl;
    int inds[NUM_OF_INDICATORS];

    sprintf(bflFile, "/MS/yuv%.2d.bfl",(int)yuvSaveNumber);

    //                sprintf(bflFile, "%s", "/MS/yuvPlaneXX.bfl");
    //                const char *bflFile = "/MS/exx.bfl";
    //                bflFile[13] =(char)((yuvSaveNumber%10)+48);
    //                bflFile[12] =(char)((yuvSaveNumber/10)+48);
    //                const char *bflFile = "/MS/blah.bfl";
    ofstream yuvFile(bflFile);
    if (yuvFile)
    {

        memset(inds,0,sizeof(inds));
        inds[5] = 255;
        inds[6] = 255;
        inds[7] = 255;
        inds[8] = 255;
        inds[9] = 255;

#ifndef OFFLINE
        Behaviours::transmission->sendIndicators(inds, sizeof(inds));
#endif // OFFLINE

        //for (int i = 0; i < YUVPLANE_SIZE; i ++)
        //{
        //    yuvFile <<(char)(((char*)(vision->yplane))[i]);
        //}

        for (int h = 0; h < HEIGHT; h ++)
        {
            /* This format is the bfl format. Has to be same as it
             * get read in mc/colour calibration tools. */
            //write Y
            for (int w = 0; w < WIDTH; w ++)
            {
                yuvFile <<(char)(((char*)(vision->yplane))[h*6*WIDTH+w]);
            }
            //write U
            for (int w = 0; w < WIDTH; w ++)
            {
                yuvFile <<(char)(((char*)(vision->yplane))[(h*6+1)*WIDTH+w]);
            }
            //write V
            for (int w = 0; w < WIDTH; w ++)
            {
                yuvFile <<(char)(((char*)(vision->yplane))[(h*6+2)*WIDTH+w]);
            }

            // write the 3 unused Y values.  
            for (int w = 0; w < WIDTH; w ++)
            {
                yuvFile <<(char)(((char*)(vision->yplane))[(h*6+3)*WIDTH+w]);
            }
            for (int w = 0; w < WIDTH; w ++)
            {
                yuvFile <<(char)(((char*)(vision->yplane))[(h*6+4)*WIDTH+w]);
            }
            for (int w = 0; w < WIDTH; w ++)
            {
                yuvFile <<(char)(((char*)(vision->yplane))[(h*6+5)*WIDTH+w]);
            }

            //write C
            for (int w = 0; w < WIDTH; w ++)
            {
                yuvFile <<(char)(((char*)(vision->cplane))[h*WIDTH+w]);
            }
        }
        if (yuvSaveNumber%2)
        {
            leds(1,1,1,2,2,2);
        }
        else
        {
            leds(2,2,2,1,1,1);
        }
        yuvFile.close();

        memset(inds,0,sizeof(inds));
#ifndef OFFLINE
        Behaviours::transmission->sendIndicators(inds, sizeof(inds));
#endif // OFFLINE
    }
    else
    {
        leds(1,2,1,1,2,1);
    }
}
