/*

   Copyright 2003 The University of New South Wales (UNSW) and National  
   ICT Australia (NICTA).

   This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
   redistribute it and/or modify it under the terms of the GNU General  
   Public License as published by the Free Software Foundation; either  
   version 2 of the License, or (at your option) any later version as  
   modified below.  As the original licensors, we add the following  
   conditions to that license:

   In paragraph 2.b), the phrase "distribute or publish" should be  
   interpreted to include entry into a competition, and hence the source  
   of any derived work entered into a competition must be made available  
   to all parties involved in that competition under the terms of this  
   license.

   In addition, if the authors of a derived work publish any conference  
   proceedings, journal articles or other academic papers describing that  
   derived work, then appropriate academic citations to the original work  
   must be included in that publication.

   This rUNSWift source is distributed in the hope that it will be useful,  
   but WITHOUT ANY WARRANTY; without even the implied warranty of  
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
   General Public License for more details.

   You should have received a copy of the GNU General Public License along  
   with this source code; if not, write to the Free Software Foundation,  
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
 * Beckham Forward
 *
 **/ 


#include "kim.h"
#include "Behaviours.h"

#ifdef COMPILE_ALL_CPP
#include "UNSW2004.h"
#endif //COMPILE_ALL_CPP

#include "../share/Common.h"
#include "../share/WalkLearningParams.h"

#ifndef OFFLINE
#include <OPENR/OPENRAPI.h>
#include "../vision/Vision.h"
#endif //OFFLINE

#include <iostream>
#include <sstream>
#include <iomanip>
#include "../vision/InfoManager.h"

#define OUT(x) if (Behaviours::bWRITE_TO_CONSOLE) cout << x 
const bool bDebug = false;
const bool bDebugCommand = false;

using namespace Behaviours;
using namespace std;

bool greaterThan(const double &x, const double &y);
bool lessThan(const double &x, const double &y);
static const double epsilon = 0.001;

ostream & operator<<(ostream &out, const Vector &v) {
	out << "[" << v.x << "," << v.y << "]";
	return out;
}

namespace Kim {
	int actionCounter = 0;



	bool wanderingAround();
	bool goToTheCenter();
	bool goGetBall();
	bool wannaDance();
	bool ballSearch(double lastImgHead);
	bool scanBall(int direction); 
	bool turnAroundLow();
	bool teammateHasBall();
	bool gotoTeammateBall();

	void initKimForward() {
		cout << "Initializing Kim's Forward..." << endl;
	}

	static const int ON = 2;
	static const int OFF = 1;
	static const int HETERESIS = 25;

	static const int ALPHA = 0;
	static const int BETA = -30;
	static const int CLOCKWISE = 1;
	static const int COUNTERCW = -1;
	static const int OUTMOST = 85;

	static const int MAX_SCANNING = 4; //# times passing the y axis
	static const int MAX_TURNING = 80; //# vision frames 50 ~ 2 seconds
	static const int MAX_WANDERING = 5000; //# of vision frames



	void doKimForward() {
		actionCounter ++;
		//cout << "Doing Kim's Forward" << endl;
		static int count = 0;

		static double lastImgHead = 0;

		if (vision->vob[vobBall].cf > 0) {
			//cout << "ball found : tracking" << endl ;
#ifdef COMPILE_ALL_CPP
			UNSW2004::doUNSW2004TrackVisualBall();;
#endif //COMPILE_ALL_CPP
			goGetBall();
			if (wannaDance()) {
				cout << "Dancing" << endl;
			}
		}
		else {
			//cout << "ball lost : searching" << endl;
			ballSearch(lastImgHead);
			setWalkParams();
			// Kim, do this yourself, thank you!
		}

		lastImgHead = vBall.imgHead;
	}

	static  double ELEV_THRESHOLD = -5;
	static  double OSCILATOR = 10;
	//static  double SMALL_SHAKE = 5;
	static  double HEAD_THRESHOLD = 0;

	void trackTheBall() {
		static double alpha = 0.7;
		static double shift = 0;
		VisualObject &ball = vision->vob[vobBall];
		OUT("ball.imgElev = " << ball.imgElev << endl);
		OUT("ball.imgHead = " << ball.imgHead << endl);

		/*
		   cout << "GPS ball .x = " << gps->getVBall(LOCAL).x << endl;
		   cout << "GPS ball .y = " << gps->getVBall(LOCAL).y << endl;
		   cout << "GPS ball .head = " << gps->getVBall(LOCAL).head << endl;
		   */

		double elev = ball.imgElev;
		double head = ball.imgHead;
		double distance = ball.d;
		//
		/*
		   if (distance < 20){
		   ELEV_THRESHOLD = -10;
		   }
		   else{
		   ELEV_THRESHOLD = 0;
		   }
		   */

		panx = 0;
		tilty = 0;
		double lastShift = shift;

		double elevUrg = abs(elev - ELEV_THRESHOLD);
		double headUrg = abs(head - HEAD_THRESHOLD);
		static const double something = 0.3;

		if (elevUrg > headUrg) {
			shift = (elev - ELEV_THRESHOLD) * something;
		}
		else {
			shift = (head - HEAD_THRESHOLD) * something;
		}
		shift = alpha * lastShift + (1 - alpha) * shift;
		if (elevUrg > headUrg) {
			tilty = shift;
		}
		else {
			panx = shift;
		}

		headtype = REL_H;
		OUT("old = " << lastShift << " , new = " << shift << endl);
		OUT("elev = " << elev << " , tilty = " << tilty << endl);
		OUT("panx = " << head << " , panx = " << panx << endl);
		OUT("distance = " << distance << endl);
	}

	void chooseKimForwardStrategy() {
		OUT("choosing Kim's Forward Strategy" << endl);
	}

	void chooseKimShootStrategy() {
		OUT("choosing Kim's Shoot Strategy" << endl);
	}

	bool wannaDance() {
		VisualObject &ball = vision->vob[vobBall];
		cout << "Ball Dist " << ball.d << endl;
		if (ball.d < 15 && abs(panx) < 30) {
			walkType = UpennLeftWT;
			return true;
		}
		return false;
	}

	bool goGetBall() {
        
		VisualObject &ball = vision->vob[vobBall];
		turnCCW = ball.head;
		setWalkParams();

		if (abs(turnCCW) < 5) {
			//avoid oscilating
			turnCCW = 0;
		}

		turnCCW = CLIP(turnCCW, 20.0);

        walkType = EllipseWalkWT;
        forward = 8.5;
        left = 0;
        turnCCW /= 3;

		return true;
	}

	int scanningDirWhenBallLost(double lastImgHead) {
		static int lastDirection = COUNTERCW;
		//cout << " lth " << lastImgHead;
		if (lastImgHead > epsilon) {
			lastDirection = COUNTERCW;
		}
		else if (lastImgHead < -epsilon) {
			lastDirection = CLOCKWISE;
		}
		return lastDirection;
	}

	/* true if got action, false if not */
	bool ballSearch(double lastImgHead) {
		static bool reset = false;
		if (reset) {
			reset = false;
			return true;
		}
		else {
			setDefaultParams();
			setWalkParams();
			if (teammateHasBall()) {
				gotoTeammateBall();
				return true;
			}
			if (scanBall(scanningDirWhenBallLost(lastImgHead)))
				return true;
			if (turnAroundLow())
				return true;
			if (goToTheCenter())
				return true;
			if (wanderingAround())
				return true;
			reset = true;
			return false;
		}
	}

	static const int UP = 0;
	static const int LEFT = 1;
	static const int DOWN = 2;
	static const int RIGHT = 3;
	static const int HEAD_SPEED = 10;

	static const int countercwNext[4][2] = {
		{1,0}, {0, -1}, { - 1,0}, {0,1}
	}; //4 direction corresponding to 4 sides of the rectangle
	static const int clockwiseNext[4][2] = {
		{ - 1,0}, {0,1}, {1,0}, {0, -1}
	};

	//moving around a box bounded by ( higherTily, lowerTily, leftPanx, rightPanx ) !!! leftPanx > 0 > rightPanx
	void moveHeadAround(bool reset,
			int direction,
			int higherTily,
			int lowerTily,
			int leftPanx,
			int rightPanx) {
		static int curPanx, curTily;
		static int side = 0;

		if (reset) {
			curPanx = 0;
			curTily = higherTily;
			side = UP;
		}

		//change dir if goes over boudaries
		if (curPanx > leftPanx) {
			side = LEFT;
			curPanx = leftPanx;
		}
		else if (curPanx < rightPanx) {
			side = RIGHT;
			curPanx = rightPanx;
		}
		else if (curTily < lowerTily) {
			side = DOWN;
			curTily = lowerTily;
		}
		else if (curTily > higherTily) {
			side = UP;
			curTily = higherTily;
		}

		//cout << "side = " << side << " " << curPanx << " " << curTily ;

		int offPanx, offTily;
		if (direction == CLOCKWISE) {
			offPanx = clockwiseNext[side][0] * HEAD_SPEED;
			offTily = clockwiseNext[side][1] * HEAD_SPEED;
		}
		else {
			offPanx = countercwNext[side][0] * HEAD_SPEED;
			offTily = countercwNext[side][1] * HEAD_SPEED;
		}

		curPanx = curPanx + offPanx;
		curTily = curTily + offTily;

		headtype = ABS_H;
		panx = curPanx;
		tilty = curTily;
		/*
		   if (greaterThan(panx ,leftPanx) ) panx = leftPanx + 1;
		   else if (lessThan(panx,rightPanx) ) panx = rightPanx - 1;
		   else if (greaterThan(tilty, higherTily) ) tilty = higherTily+ 1;
		   else if (lessThan(tilty , lowerTily) ) tilty = lowerTily - 1;
		   */

		//cout << " -> " << panx << " " << tilty << endl;
	}

	bool scanBall(int direction) {
		static int lastCounter = 0;
		static int passed = 0;
		static double lastPanx = 0;
		bool reset = false;

		//cout << "scanning " << direction << endl;
		if (actionCounter != lastCounter + 1) {
			//we have missed an action, restart
			cout << "RESSSSSSSSSSSSSSET" << endl;
			reset = true;
			passed = 0;
		}
		else {
			reset = false;
		}

		lastCounter = actionCounter;

		if (passed > MAX_SCANNING) {
			panx = 0;
			return false; //number of allowed loop exceeded ,stop scanning
		}

		double thisPanx = MICRO2DEG(sensors->sensorVal[ssHEAD_PAN]);
		//double thisTily = MICRO2DEG(sensors->sensorVal[ssHEAD_TILT1]); 
		if (lastPanx * thisPanx < 0) {
			//passed the center
			cout << "passsssssssssssssssed" << endl;
			passed += 1;
		}

		moveHeadAround(reset, direction, ALPHA, BETA, OUTMOST, -OUTMOST);

		lastPanx = thisPanx;
		return true;
	}

	bool turnAroundLow() {
		static int duration = 0;
		static int lastCounter = 0;

		if (actionCounter != lastCounter + 1) {
			cout << "Turning low " << endl;
			duration = 0;
		}
		duration += 1;
		lastCounter = actionCounter;

		setWalkParams();
		if (duration > MAX_TURNING) {
			return false;
		}

		headtype = ABS_H;
		panx = 80;
		tilty = -90;

		walkType = NormalWalkWT;
		turnCCW = 40;
		forward = 0;
		left = 0;
		return true;
	}

	bool goToTheCenter() {
		cout << "Going to the center" << endl;
		double findPositionx = FIELD_WIDTH / 2.0,
		findPositiony = FIELD_LENGTH / 2.0;

		Vector target(vCART,
				findPositionx
				+ 8.0 * cos(DEG2RAD(gps->self().getHackedH())),
				findPositiony + 8.0 * sin(DEG2RAD(gps->self().getHackedH())));
		Vector robot(vCART, (double) gps->self().pos.x, (double) gps->self().pos.y);

		//vector to destination point
		Vector vx(robot);
		vx.sub(target);

		cout << "+" << vx.x << " " << vx.y << " " << vx.d ;
		if (vx.d < 50) {
			return false;
		}

		double pointAngle2 = utilNormaliseAngle((vx.getHackedTheta()
					- gps->self().getHackedH()));
		cout << " " << pointAngle2 << endl;

		moveHeadAround(false, COUNTERCW, ALPHA, BETA, OUTMOST, -OUTMOST);
		walkType = CanterWalkWT;
		turnCCW = 20;
		forward = -5.0 * cos(DEG2RAD(pointAngle2));
		left = -5.0 * sin(DEG2RAD(pointAngle2));

		if (vx.d < 20) {
			forward /= 4.0;
			left /= 4.0;
		}
		return true;
	}

	bool wanderingAround() {
		static int duration = 0;
		static int lastCounter = 0;
		static int direction = 1;

		if (actionCounter != lastCounter + 1) {
			cout << "Wandering" << endl;
			direction = rand() % 2 == 0 ? 1 : -1;
			duration = 0;
		}
		/*
		   if (duration % 500 == 0){
		   cout << "Change dir " << duration << endl;
		   }
		   */
		duration += 1;
		lastCounter = actionCounter;

		if (duration > MAX_WANDERING) {
			cout << "Stop wanderingAround" << endl;
			return false;
		}

		walkType = NormalWalkWT;
		forward = 6;
		turnCCW = 10 * direction;
		moveHeadAround(false, direction, ALPHA, BETA, OUTMOST, -OUTMOST);

		return true;
	}

	bool teammateHasBall() {
		return false;
	}

	bool gotoTeammateBall() {
		return true;
	}

	void doPawKickTesting() {
		//not work
		cout << "paw  " ;
		VisualObject &ball = vision->vob[vobBall] ;
		if (ball.cf > 0) {
			// saPawKick(ball.d);
		}
		else {
			headtype = ABS_H;
			panx = 0;
			tilty = 0;
		}
	}

	void doTeammateTesting() {
		static int count = 0;
		count ++;
		if (count % 20 != 0) {
			return;
		}
		for (int i = 0; i < NUM_TEAM_MEMBER; i++) {
			WMObj tmBall = gps->tmBallObj(i);
			WMObj tm = gps->tmObj(i);

			interpBehav info = (interpBehav) tm.behavioursVal;
			cout << "Team member " << i << " x=" << tm.pos.x << " y=" << tm.pos.y
				<< " ballDist= " << info.ballDist << " x = " << tmBall.pos.x
				<< " y = " << tmBall.pos.y << endl;
		}
	}

	void doIRTesting() {
		cout << "IR sensed: " << sensors->sensorVal[ssINFRARED_FAR] << endl;
	}

	void doReadJointTesting() {
		int HT = (int) MICRO2DEG(sensors-> sensorVal[ssHEAD_TILT1]) ,
		HP = (int) MICRO2DEG(sensors->sensorVal[ssHEAD_PAN]) ,
		HC = (int) MICRO2DEG(sensors->sensorVal[ssHEAD_TILT2]) ;

		int FLJ = (int) MICRO2DEG(sensors-> sensorVal[ssFL_ROTATOR]) ,
		FLS = (int) MICRO2DEG(sensors->sensorVal[ssFL_ABDUCTOR]) ,
		FLK = (int) MICRO2DEG(sensors->sensorVal[ssFL_KNEE]) ,

		FRJ = (int) MICRO2DEG(sensors->sensorVal[ssFR_ROTATOR]) ,
		FRS = (int) MICRO2DEG(sensors->sensorVal[ssFR_ABDUCTOR]) ,
		FRK = (int) MICRO2DEG(sensors->sensorVal[ssFR_KNEE]) ,

		RLJ = (int) MICRO2DEG(sensors->sensorVal[ssRL_ROTATOR]) ,
		RLS = (int) MICRO2DEG(sensors->sensorVal[ssRL_ABDUCTOR]) ,
		RLK = (int) MICRO2DEG(sensors->sensorVal[ssRL_KNEE]) ,

		RRJ = (int) MICRO2DEG(sensors->sensorVal[ssRR_ROTATOR]) ,
		RRS = (int) MICRO2DEG(sensors->sensorVal[ssRR_ABDUCTOR]) ,
		RRK = (int) MICRO2DEG(sensors->sensorVal[ssRR_KNEE]) ;

		cout << setw(5) << HT << setw(5) << HP << setw(5) << HC << setw(5) << FLJ << setw(5) << FLS << setw(5) << FLK << setw(5)
			<< FRJ << setw(5) << FRS << setw(5) << FRK << setw(5) << RLJ
			<< setw(5) << RLS << setw(5) << RLK << setw(5) << RRJ << setw(5)
			<< RRS << setw(5) << RRK << endl;
	}

	void doLegTesting() {
		static int count = 0;
		static int wt = 4;
		if (count % 100 == 0) {
			cout << "New WT" << wt << endl;
			walkType = wt;
			wt += 1;
		}
		count ++;
	}

    void doSimpleTesting(){
        forward = 8;
        Behaviours::left = 0;
        turnCCW = 0;
        walkType = NormalWalkWT;
    }

    void doTestStepComplete(){
        static int count =0 ;
        if (count ++ %2 == 0){
            walkType = LightningKick;
            cout << "Kick  " ;
        }
        else{
            walkType = ChestPushWT;
            cout << "Push  " ;
        }
    }

    AtomicAction action = {aaParaWalk,0,0,0,0 , 40, 90,120,0,0,55,10,-55,10,0,0,0,0, 0,0};
    void calibrateWalk(const char *command){
        if (bDebugCommand)
            cout << "got walk calibrating paramters " << endl;
        istringstream ss(command);
        action.headType = ABS_H;
        action.panx = 0;
        action.tilty = 0;
        action.cranez = 0;
        ss >> action.walkType >> action.Forward >> action.Left >> action.turnCCW ;
        switch (action.walkType){
            case SkellipticalWalkWT:
                {
                    WalkLearningParams wlParams;
                    wlParams.walkType = SkellipticalWalkWT;
                    wlParams.sp.read(ss);
#ifndef OFFLINE
                    Behaviours::transmission->sendLearningParams(wlParams);
#endif //OFFLINE
                    break;
                }
            case SkiWalkWT:
                {
                    WalkLearningParams wlParams;
                    wlParams.walkType = SkiWalkWT;
                    wlParams.skp.read(ss);
#ifndef OFFLINE
                    Behaviours::transmission->sendLearningParams(wlParams);
#endif //OFFLINE
                    break;
                }
            case EllipseWalkWT:
            default:
                ss >> action.speed >> action.hF >> action.hB >> action.hdF >> action.hdB >> action.ffO >> action.fsO >> action.bfO >> action.bsO >> action.headType >> action.panx >> action.tilty >> action.cranez;
                break;
        }

    }

    //play around with PWalk parameters
    //TODO: move to WALK LEARNING player
    void doWalkCalibrate(){
        if (bDebug){
            cout << "Do walk calibrate " << endl;
            action.printFull();
        }
        setDefaultParams();
        Behaviours::walkType = action.walkType;
        Behaviours::forward = action.Forward;
        Behaviours::left = action.Left;
        Behaviours::turnCCW = action.turnCCW;
        Behaviours::PG = action.speed;
        Behaviours::hF = action.hF;
        Behaviours::hB = action.hB;
        Behaviours::hdB = action.hdB;
        Behaviours::hdF = action.hdF;
        Behaviours::ffO = action.ffO;
        Behaviours::fsO = action.fsO;
        Behaviours::bfO = action.bfO;
        Behaviours::bsO = action.bsO;
        Behaviours::headtype = action.headType;
        Behaviours::panx = action.panx;
        Behaviours::tilty = action.tilty;
        Behaviours::cranez = action.cranez;
        //action.printFull();
        //cout << "Battery current: " << batteryCurrent << "mA " << gInfoManager.getAnySensor(29) << " current WT = " << *currentWalkType << endl;
    }

    void doPointProjectTesting(){
        Behaviours::headtype = ABS_H;
        Behaviours::panx = 0;
        Behaviours::tilty = -10;
        Behaviours::cranez = -10;
		VisualObject &ball = vision->vob[vobBall];

    }

    static const double CRAZY_HEAD = 50;

    void doTestHead(){
        static double lastPanx = 0;
        static double lastTilty = 0;
        static double lastCranez = 0;
        static int counter = 0;
        if (counter ++ % 2 == 0){
            lastPanx +=  ( (rand()%2)*2 -1 ) * CRAZY_HEAD  ;
            lastPanx = CLIP( lastPanx , 90.0) ;
            lastTilty += ( (rand()%2)*2 -1 ) * CRAZY_HEAD;
            lastTilty = CLIP(lastTilty + 45, 45.0) - 45;
            lastCranez += ( (rand()%2)*2 -1 ) * CRAZY_HEAD;
            lastCranez = CLIP( lastCranez - 25, 45.0) + 25;
        }
        headtype = ABS_H;
        panx = lastPanx;
        tilty = 0;
        cranez = lastCranez;
        cout << "Battery current: " << batteryCurrent << "mA " << panx << " " << tilty << " " << cranez << endl;

    }

	void doKimTesting() {
		//doPawKickTesting();
		//cout << "DoNothing" << endl;
		//doReadJointTesting();
		//doTeammateTesting();
		//doLegTesting();
        //doSimpleTesting();
        //doTestStepComplete();
        //doWalkCalibrate();
        //doTestHead();
        //doPointProjectTesting();
        //doIRTesting();
        //doGpsTesting();
	}
}

bool greaterThan(const double &x, const double &y) {
	return x > y + epsilon;
}

bool lessThan(const double &x, const double &y) {
	return greaterThan(y, x);
}


