/*

   Copyright 2004 The University of New South Wales (UNSW) and National  
   ICT Australia (NICTA).

   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
   redistribute it and/or modify it under the terms of the GNU General  
   Public License as published by the Free Software Foundation; either  
   version 2 of the License, or (at your option) any later version as  
   modified below.  As the original licensors, we add the following  
   conditions to that license:

   In paragraph 2.b), the phrase "distribute or publish" should be  
   interpreted to include entry into a competition, and hence the source  
   of any derived work entered into a competition must be made available  
   to all parties involved in that competition under the terms of this  
   license.

   In addition, if the authors of a derived work publish any conference  
   proceedings, journal articles or other academic papers describing that  
   derived work, then appropriate academic citations to the original work  
   must be included in that publication.

   This rUNSWift source is distributed in the hope that it will be useful,  
   but WITHOUT ANY WARRANTY; without even the implied warranty of  
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
   General Public License for more details.

   You should have received a copy of the GNU General Public License along  
   with this source code; if not, write to the Free Software Foundation,  
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

/*
 * Last modification background information
 * $Id: RoboCupGameControlData.h 4364 2004-09-22 07:04:12Z kcph007 $
 *
 * Copyright (c) 2004 UNSW
 * All Rights Reserved.
 *
**/



/*
 * Copyright 2002,2003 Sony Corporation
 *
 * Permission to use, copy, modify, and redistribute this software is
 * hereby granted.
 *
 * This software is provided "as is" without warranty of any kind,
 * either expressed or implied, including but not limited to the
 * implied warranties of fitness for a particular purpose.
 *
 *
 * Modified at University Bremen
 * for RoboCup 2004 - Martin Fritsche, March 2004
 */

#ifndef _RoboCupGameControlData_h_DEFINED
#define _RoboCupGameControlData_h_DEFINED


#ifdef __cplusplus
extern "C" {
#endif


/* Game state */

typedef char RState;

const RState ROBOCUP_STATE_INITIAL   = 0x00; /* initial state, never sent */
const RState ROBOCUP_STATE_READY     = 0x01; /* position for kickoff      */
const RState ROBOCUP_STATE_SET       = 0x02; /* stop and wait for kickoff */
const RState ROBOCUP_STATE_PLAYING   = 0x03; /* play game                 */
const RState ROBOCUP_STATE_PENALIZED = 0x04; /* penalized, do not move    */
const RState ROBOCUP_STATE_FINISHED  = 0x05; /* game set                  */



/* Kickoff */

typedef char RKickOff;

const RKickOff ROBOCUP_KICKOFF_INVALID  = 0x00;  /* sent in all states but "ready" and "set" */
const RKickOff ROBOCUP_KICKOFF_OWN      = 0x01;
const RKickOff ROBOCUP_KICKOFF_OPPONENT = 0x02;



/* Color */

typedef char RTeamColor;

const RTeamColor ROBOCUP_TEAMCOLOR_RED  = 0x00; 
const RTeamColor ROBOCUP_TEAMCOLOR_BLUE = 0x01;



/* Penalty */

typedef char RPenalty;
const int NUM_PENALTY = 8;                      /* used by GameController */

const RPenalty PENALTY_NONE                = 0; /* sent in all states but "penalized" */
const RPenalty PENALTY_BALL_HOLDING        = 1;
const RPenalty PENALTY_KEEPER_CHARGE       = 2;
const RPenalty PENALTY_FIELD_PLAYER_CHARGE = 3;
const RPenalty PENALTY_ILLEGAL_DEFENDER    = 4;
const RPenalty PENALTY_ILLEGAL_DEFENSE     = 5;
const RPenalty PENALTY_OBSTRUCTION         = 6;
const RPenalty PENALTY_REQ_FOR_PICKUP      = 7;



/* Data type definition */

typedef struct {
    RState      state;                  /* current state of player */
    RTeamColor  teamColor;              /* team color              */
    RKickOff    kickoff;                /* kickoff information     */
    short       ownScore;               /* score of own team       */
    short       opponentScore;          /* score of opponent team  */
    RPenalty    penalty;                /* current penalty, only valid in "penalized" */
} RoboCupGameControlData;


#ifdef __cplusplus
}
#endif


#endif /* _RoboCupGameControlData_h_DEFINED */

