/*

   Copyright 2003 The University of New South Wales (UNSW) and National  
   ICT Australia (NICTA).

   This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
   redistribute it and/or modify it under the terms of the GNU General  
   Public License as published by the Free Software Foundation; either  
   version 2 of the License, or (at your option) any later version as  
   modified below.  As the original licensors, we add the following  
   conditions to that license:

   In paragraph 2.b), the phrase "distribute or publish" should be  
   interpreted to include entry into a competition, and hence the source  
   of any derived work entered into a competition must be made available  
   to all parties involved in that competition under the terms of this  
   license.

   In addition, if the authors of a derived work publish any conference  
   proceedings, journal articles or other academic papers describing that  
   derived work, then appropriate academic citations to the original work  
   must be included in that publication.

   This rUNSWift source is distributed in the hope that it will be useful,  
   but WITHOUT ANY WARRANTY; without even the implied warranty of  
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
   General Public License for more details.

   You should have received a copy of the GNU General Public License along  
   with this source code; if not, write to the Free Software Foundation,  
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * UNSW 2003 Robocup (Nicodemus Sutanto)
 *
 * Last modification background information
 * $Id: SharedMemoryDef.h 4364 2004-09-22 07:04:12Z kcph007 $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
 **/

#ifndef SharedMemoryDef_h_DEFINED
#define SharedMemoryDef_h_DEFINED

#ifndef OFFLINE
#include <OPENR/OObject.h>
#endif // OFFLINE

#include <ostream>

#include "SwitchBoard.h"
#include "VisionDef.h"
#include "PackageDef.h"
#include "Common.h"
#include "WirelessSwitchboard.h"
#include "hton.h"


// For inter-robot latency testing
static const int NORMAL_WM = 0;
static const int INIT_LAT_WM = -1;

// used to store information form kalman filter updates for sending to teammates
// uses floats to save space, floats are 4 bytes, doubles are 8 bytes
struct IFKUpdateData {
	float iVecx;
	float iVecy;

	float iMat11;
	float iMat12;
	float iMat21;
	float iMat22;
};

// The information to be passed when sharing world model
// with other robots
struct WMShareObj {
	//WMObj obj;
	float posx, posy;
	float posh;
	float posVar, hVar;

	unsigned int behavioursVal;
	//team maverick stuff
	/*
	   unsigned int ballDist:29;
	   unsigned int amThirdPlayer:1;
	   unsigned int playerType:2;
	   */

	double ballx;
	double bally;
	double ballvar;

	// int teamCForwardRole;

	// information for information form kalman filter tracking opponents
	IFKUpdateData opponentsUpdate[NUM_OPPONENTS];

#ifdef IR_LATENCY_TEST
	// will be NORMAL_WM, INIT_LAT_WM, or the player number to the robot
	// we are responding to for latency testing
	int latMode;
	SystemTime waited;
#endif // IR_LATENCY_TEST
};

struct WMShareInfo {
	char header[4];
	int playerNum;
	int team;
	WMShareObj obj;
};

#ifdef DELAY_TESTING
static const bool isDelayTesting = true;
#else
static const bool isDelayTesting = false;
#endif

//Define Type of the message that is being sent to Wireless
//shmdt = Share Mem Data Type
enum {
	shmdtMessages,
	shmdtCPlane = 8,  //just to be competible with legacy CPLANE_DATA_TYPE ..., if things work out, remove it
	shmdtOPlane = 9,
	shmdtYUVPlaneCPlane = 10,
	shmdtYUVPlane = 11,
	shmdtActuatorMessage = 15,
	shmdtEdgeDetectDebug = 20,
	shmdtWorldModel = 100
};


// Debug information to be sent to Robocommander
struct RC_DEBUG_DATA {
	char message[100];

	char world_model_debug_data[560];

#ifdef DELAY_TESTING
	char cplane_data[INT_BYTES + COMPRESSED_CPLANE_SIZE + SENSOR_SIZE + SANITY_SIZE + OBJ_SIZE + GPS_SIZE + 4]; // for counter
#else
	char cplane_data[INT_BYTES + COMPRESSED_CPLANE_SIZE + SENSOR_SIZE + SANITY_SIZE + OBJ_SIZE + GPS_SIZE];
#endif // DELAY_TESTING

	char oplane_data[OBJ_SIZE];

	char yuvplane_data[YUVPLANE_SIZE];
	char yuvCplane_data[CPLANE_SIZE];

	char line_debug_data[CPLANE_SIZE *DEBUG_PLANES];
};

struct actSensorData {
	long frameID;

	long accelFor, accelSide, accelZ;

    long ssHeadTilt, ssHeadPan, ssHeadCrane;
    long ssFrontLeftJoint, ssFrontLeftShoulder, ssFrontLeftKnee;
    long ssFrontRightJoint, ssFrontRightShoulder, ssFrontRightKnee;
    long ssRearLeftJoint, ssRearLeftShoulder, ssRearLeftKnee;
    long ssRearRightJoint, ssRearRightShoulder, ssRearRightKnee;

    long PWMHeadTilt, PWMHeadPan, PWMHeadCrane;
    long PWMFrontLeftJoint, PWMFrontLeftShoulder, PWMFrontLeftKnee;
    long PWMFrontRightJoint, PWMFrontRightShoulder, PWMFrontRightKnee;
    long PWMRearLeftJoint, PWMRearLeftShoulder, PWMRearLeftKnee;
    long PWMRearRightJoint, PWMRearRightShoulder, PWMRearRightKnee;

	void dtoh() {
		frameID = dtohl(frameID);
	
		accelFor = dtohl(accelFor);
		accelSide = dtohl(accelSide);
		accelZ = dtohl(accelZ);

    	ssHeadTilt = dtohl(ssHeadTilt);
		ssHeadPan = dtohl(ssHeadPan);
		ssHeadCrane = dtohl(ssHeadCrane);
    	ssFrontLeftJoint = dtohl(ssFrontLeftJoint);
		ssFrontLeftShoulder = dtohl(ssFrontLeftShoulder);
		ssFrontLeftKnee = dtohl(ssFrontLeftKnee);
    	ssFrontRightJoint = dtohl(ssFrontRightJoint);
		ssFrontRightShoulder = dtohl(ssFrontRightShoulder);
		ssFrontRightKnee = dtohl(ssFrontRightKnee);
    	ssRearLeftJoint = dtohl(ssRearLeftJoint);
		ssRearLeftShoulder = dtohl(ssRearLeftShoulder);
		ssRearLeftKnee = dtohl(ssRearLeftKnee);
    	ssRearRightJoint = dtohl(ssRearRightJoint);
		ssRearRightShoulder = dtohl(ssRearRightShoulder);
		ssRearRightKnee = dtohl(ssRearRightKnee);

    	PWMHeadTilt = dtohl(PWMHeadTilt);
		PWMHeadPan = dtohl(PWMHeadPan);
		PWMHeadCrane = dtohl(PWMHeadCrane);
    	PWMFrontLeftJoint = dtohl(PWMFrontLeftJoint);
		PWMFrontLeftShoulder = dtohl(PWMFrontLeftShoulder);
		PWMFrontLeftKnee = dtohl(PWMFrontLeftKnee);
    	PWMFrontRightJoint = dtohl(PWMFrontRightJoint);
		PWMFrontRightShoulder = dtohl(PWMFrontRightShoulder);
		PWMFrontRightKnee = dtohl(PWMFrontRightKnee);
    	PWMRearLeftJoint = dtohl(PWMRearLeftJoint);
		PWMRearLeftShoulder = dtohl(PWMRearLeftShoulder);
		PWMRearLeftKnee = dtohl(PWMRearLeftKnee);
    	PWMRearRightJoint = dtohl(PWMRearRightJoint);
		PWMRearRightShoulder = dtohl(PWMRearRightShoulder);
		PWMRearRightKnee = dtohl(PWMRearRightKnee);
	}

	void htod() {
		frameID = htodl(frameID);
	
		accelFor = htodl(accelFor);
		accelSide = htodl(accelSide);
		accelZ = htodl(accelZ);

    	ssHeadTilt = htodl(ssHeadTilt);
		ssHeadPan = htodl(ssHeadPan);
		ssHeadCrane = htodl(ssHeadCrane);
    	ssFrontLeftJoint = htodl(ssFrontLeftJoint);
		ssFrontLeftShoulder = htodl(ssFrontLeftShoulder);
		ssFrontLeftKnee = htodl(ssFrontLeftKnee);
    	ssFrontRightJoint = htodl(ssFrontRightJoint);
		ssFrontRightShoulder = htodl(ssFrontRightShoulder);
		ssFrontRightKnee = htodl(ssFrontRightKnee);
    	ssRearLeftJoint = htodl(ssRearLeftJoint);
		ssRearLeftShoulder = htodl(ssRearLeftShoulder);
		ssRearLeftKnee = htodl(ssRearLeftKnee);
    	ssRearRightJoint = htodl(ssRearRightJoint);
		ssRearRightShoulder = htodl(ssRearRightShoulder);
		ssRearRightKnee = htodl(ssRearRightKnee);

    	PWMHeadTilt = htodl(PWMHeadTilt);
		PWMHeadPan = htodl(PWMHeadPan);
		PWMHeadCrane = htodl(PWMHeadCrane);
    	PWMFrontLeftJoint = htodl(PWMFrontLeftJoint);
		PWMFrontLeftShoulder = htodl(PWMFrontLeftShoulder);
		PWMFrontLeftKnee = htodl(PWMFrontLeftKnee);
    	PWMFrontRightJoint = htodl(PWMFrontRightJoint);
		PWMFrontRightShoulder = htodl(PWMFrontRightShoulder);
		PWMFrontRightKnee = htodl(PWMFrontRightKnee);
    	PWMRearLeftJoint = htodl(PWMRearLeftJoint);
		PWMRearLeftShoulder = htodl(PWMRearLeftShoulder);
		PWMRearLeftKnee = htodl(PWMRearLeftKnee);
    	PWMRearRightJoint = htodl(PWMRearRightJoint);
		PWMRearRightShoulder = htodl(PWMRearRightShoulder);
		PWMRearRightKnee = htodl(PWMRearRightKnee);
	}
};

struct actCommandData {
	long lastSensedFrameID;
    long cmdHeadTilt, cmdHeadPan, cmdHeadCrane;
    long cmdFrontLeftJoint, cmdFrontLeftShoulder, cmdFrontLeftKnee;
    long cmdFrontRightJoint, cmdFrontRightShoulder, cmdFrontRightKnee;
    long cmdRearLeftJoint, cmdRearLeftShoulder, cmdRearLeftKnee;
    long cmdRearRightJoint, cmdRearRightShoulder, cmdRearRightKnee;

	void dtoh() {
		lastSensedFrameID = dtohl(lastSensedFrameID);
    	cmdHeadTilt = dtohl(cmdHeadTilt);
		cmdHeadPan = dtohl(cmdHeadPan);
		cmdHeadCrane = dtohl(cmdHeadCrane);
    	cmdFrontLeftJoint = dtohl(cmdFrontLeftJoint);
		cmdFrontLeftShoulder = dtohl(cmdFrontLeftShoulder);
		cmdFrontLeftKnee = dtohl(cmdFrontLeftKnee);
    	cmdFrontRightJoint = dtohl(cmdFrontRightJoint);
		cmdFrontRightShoulder = dtohl(cmdFrontRightShoulder);
		cmdFrontRightKnee = dtohl(cmdFrontRightKnee);
    	cmdRearLeftJoint = dtohl(cmdRearLeftJoint);
		cmdRearLeftShoulder = dtohl(cmdRearLeftShoulder);
		cmdRearLeftKnee = dtohl(cmdRearLeftKnee);
    	cmdRearRightJoint = dtohl(cmdRearRightJoint);
		cmdRearRightShoulder = dtohl(cmdRearRightShoulder);
		cmdRearRightKnee = dtohl(cmdRearRightKnee);
	}

	void htod() {
		lastSensedFrameID = htodl(lastSensedFrameID);
    	cmdHeadTilt = htodl(cmdHeadTilt);
		cmdHeadPan = htodl(cmdHeadPan);
		cmdHeadCrane = htodl(cmdHeadCrane);
    	cmdFrontLeftJoint = htodl(cmdFrontLeftJoint);
		cmdFrontLeftShoulder = htodl(cmdFrontLeftShoulder);
		cmdFrontLeftKnee = htodl(cmdFrontLeftKnee);
    	cmdFrontRightJoint = htodl(cmdFrontRightJoint);
		cmdFrontRightShoulder = htodl(cmdFrontRightShoulder);
		cmdFrontRightKnee = htodl(cmdFrontRightKnee);
    	cmdRearLeftJoint = htodl(cmdRearLeftJoint);
		cmdRearLeftShoulder = htodl(cmdRearLeftShoulder);
		cmdRearLeftKnee = htodl(cmdRearLeftKnee);
    	cmdRearRightJoint = htodl(cmdRearRightJoint);
		cmdRearRightShoulder = htodl(cmdRearRightShoulder);
		cmdRearRightKnee = htodl(cmdRearRightKnee);
	}
};

enum ActuatorWirelessDataEntryType{
		actSensorDataT,
		actCommandDataT
};

struct ActuatorWirelessDataEntry {
	int type;
	
	union {
		actSensorData s;
		actCommandData c;
	};
	
	void dtoh() {
		type = dtohl(type);
		switch (type) {
			case actSensorDataT:
				s.dtoh();
			break;
			case actCommandDataT:
				c.dtoh();
			break;
			default:
			break;
		}
	}

	void htod() {
		switch (type) {
			case actSensorDataT:
				s.htod();
			break;
			case actCommandDataT:
				c.htod();
			break;
			default:
			break;
		}
		type = htodl(type);
	}

};

static inline std::ostream &operator<<(std::ostream &out , const actSensorData &a){
	out << '[';

	out << a.frameID << endl;

	out << '[' << a.accelFor << ", " << a.accelSide << ", " << a.accelZ << ']' << endl;

	out << '[';	
    out << '[' << a.ssHeadTilt << ", " << a.ssHeadPan << ", " << a.ssHeadCrane << ']' << endl;
    out << '[' << a.ssFrontLeftJoint << ", " << a.ssFrontLeftShoulder << ", " << a.ssFrontLeftKnee << ']' << endl;
    out << '[' << a.ssFrontRightJoint << ", " << a.ssFrontRightShoulder << ", " << a.ssFrontRightKnee << ']' << endl;
    out << '[' << a.ssRearLeftJoint << ", " << a.ssRearLeftShoulder << ", " << a.ssRearLeftKnee << ']' << endl;
    out << '[' << a.ssRearRightJoint << ", " << a.ssRearRightShoulder << ", " << a.ssRearRightKnee << ']' << endl;
	out << ']' << endl;	

	out << '[';	
    out << '[' << a.PWMHeadTilt << ", " << a.PWMHeadPan << ", " << a.PWMHeadCrane << ']' << endl;
    out << '[' << a.PWMFrontLeftJoint << ", " << a.PWMFrontLeftShoulder << ", " << a.PWMFrontLeftKnee << ']' << endl;
    out << '[' << a.PWMFrontRightJoint << ", " << a.PWMFrontRightShoulder << ", " << a.PWMFrontRightKnee << ']' << endl;
    out << '[' << a.PWMRearLeftJoint << ", " << a.PWMRearLeftShoulder << ", " << a.PWMRearLeftKnee << ']' << endl;
    out << '[' << a.PWMRearRightJoint << ", " << a.PWMRearRightShoulder << ", " << a.PWMRearRightKnee << ']' << endl;
	out << "]]" << endl;
	
	return out;
}

static inline std::ostream &operator<<(std::ostream &out , const actCommandData &a){
	out << '[';	
	out << a.lastSensedFrameID << endl;
    out << '[' << a.cmdHeadTilt << ", " << a.cmdHeadPan << ", " << a.cmdHeadCrane << ']' << endl;
    out << '[' << a.cmdFrontLeftJoint << ", " << a.cmdFrontLeftShoulder << ", " << a.cmdFrontLeftKnee << ']' << endl;
    out << '[' << a.cmdFrontRightJoint << ", " << a.cmdFrontRightShoulder << ", " << a.cmdFrontRightKnee << ']' << endl;
    out << '[' << a.cmdRearLeftJoint << ", " << a.cmdRearLeftShoulder << ", " << a.cmdRearLeftKnee << ']' << endl;
    out << '[' << a.cmdRearRightJoint << ", " << a.cmdRearRightShoulder << ", " << a.cmdRearRightKnee << ']' << endl;
	out << ']' << endl;

	return out;
}

static inline std::ostream &operator<<(std::ostream &out , const ActuatorWirelessDataEntry &a){
	switch (a.type) {
		case actSensorDataT:
			out << "[ sensor " << a.s << ']';
		break;
		case actCommandDataT:
			out << "[ command " << a.c << ']';
		break;
		default:
			out << "[ Unknown ActuatorWirelessData type ]";
		break;
	}
	return out;
}

#endif // SharedMemoryDef_h_DEFINED
