/*

   Copyright 2003 The University of New South Wales (UNSW) and National  
   ICT Australia (NICTA).

   This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
   redistribute it and/or modify it under the terms of the GNU General  
   Public License as published by the Free Software Foundation; either  
   version 2 of the License, or (at your option) any later version as  
   modified below.  As the original licensors, we add the following  
   conditions to that license:

   In paragraph 2.b), the phrase "distribute or publish" should be  
   interpreted to include entry into a competition, and hence the source  
   of any derived work entered into a competition must be made available  
   to all parties involved in that competition under the terms of this  
   license.

   In addition, if the authors of a derived work publish any conference  
   proceedings, journal articles or other academic papers describing that  
   derived work, then appropriate academic citations to the original work  
   must be included in that publication.

   This rUNSWift source is distributed in the hope that it will be useful,  
   but WITHOUT ANY WARRANTY; without even the implied warranty of  
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
   General Public License for more details.

   You should have received a copy of the GNU General Public License along  
   with this source code; if not, write to the Free Software Foundation,  
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: GeneralisedColourDetect.h 4362 2004-09-22 06:54:12Z kcph007 $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
 * classify pixels of the images according to the 9 distinct colours to form the c-plane (Colour plane)
 * Ball  : orange
 * Beacon: light blue, green, yellow, pink
 * Robots: dark blue, red
 * Field : green
 * Background colour
 * Unclassified colour
 *
 **/

#ifndef _GCD_h_DEFINED
#define _GCD_h_DEFINED

#include "../share/Common.h"
#include "../share/VisionDef.h"
#include "../share/SwitchBoard.h"

//#define VISION_DEBUG

static const int GCD = 0;
static const int DECISION_TREE = 1;
static const int NN_MULTI_COLOUR = 2;

/*
static const int BALL = 0;
static const int BLUE = 1;
static const int GREEN = 2;
static const int YELLOW = 3;
static const int PINK = 4;
static const int BLUE_DOG = 5;
static const int RED_DOG = 6;
static const int GREEN_FIELD = 7;
static const int ROBOT_GREY = 8;
static const int FIELDWHITE = 9;
static const int BLACK = 10;
static const int FIELD_LINE = 11;
static const int FIELD_BORDER = 12;

static const int NONE = 127;
*/

typedef enum {
	cBALL = 0,
	cBEACON_BLUE = 1,
	cBEACON_GREEN = 2,
	cBEACON_YELLOW = 3,
	cBEACON_PINK = 4,
	cROBOT_BLUE = 5,
	cROBOT_RED = 6,
	cFIELD_GREEN = 7,
	cROBOT_GREY = 8,
	cWHITE = 9,
	cBLACK = 10,
	cFIELD_LINE = 11,
	cFIELD_BORDER = 12,
	cNUM_COLOR = 13,
	cNONE = 127
}
Color;

static const Color BACKGROUND = cROBOT_GREY;
static const Color MAX_COLOR = cBLACK;
static const Color LAST_USABLE_COLOR = cROBOT_RED;
static const int NUM_USABLE_COLORS = LAST_USABLE_COLOR+1;

static const int COLOR_MASK = 0x0F;
static const int MAYBE_BIT = 0x10;

static const char *ColorName[cNUM_COLOR] = {
	"Ball Orange", 
    "Beacon Blue",
    "Beacon Green", 
    "Beacon Yellow", 
    "Beacon Pink", 
    "Dog Blue", 
    "Dog Red", 
    "Field Green", 
    "Robot Grey", 
    "Field White", 
    "Black", 
    "Line", 
    "Edge"
};

struct DogColourDistortion {
	double My;	// coefficient
	double Cy;	// constant
	double Mu;
	double Cu;
	double Mv;
	double Cv;
};

static inline std::ostream &operator<<(std::ostream &out , const DogColourDistortion &dcd) {
	out << "[\nY = " << dcd.My << "*y + " << dcd.Cy << ",\nU = " << dcd.Mu << "*u + " << dcd.Cu << ",\nV = " << dcd.Mv << "*v + " << dcd.Cv << ']';
	return out;
}

/*
   static const int BACKGROUND = 8;
#ifdef EDGE_DETECT
static const int FIELDWHITE = 9;
#else //EDGE_DETECT
static const int UNCLASSIFIED = 9;
#endif //EDGE_DETECT
*/

static const int NUM_PLANES = 8;

static const int RES = 0;
static const int WIDTH = CPLANE_WIDTH;
static const int HEIGHT = CPLANE_HEIGHT;
const double DIST_CONST = (50.0 * 65.0) / 20.0;

#ifdef OFFLINE
static const int SKIP = 2 * WIDTH;
#else
static const int SKIP = 5 * WIDTH;
#endif //OFFLINE

static const int PIXELS = HEIGHT *WIDTH;
static const int USABLE_HEIGHT = (HEIGHT - 1);
static const int USABLE_PIXELS = USABLE_HEIGHT *WIDTH;

static const int MAXY = 128;
static const int MAXU = 128;
static const int MAXV = 128;

static const int CUBE_VOLUME = MAXY *MAXU *MAXV;
static const int FACTOR = 2;

class GeneralisedColourDetect {
	public:

		GeneralisedColourDetect();
		~GeneralisedColourDetect() {
			delete[]nn_multi_colour;
		}
		// void replaceCplane(uchar *yp, uchar *up, uchar *vp, long skip);
		uchar cplane[PIXELS];

#ifdef VISION_DEBUG

		/*
		 * Change the calibration method on the fly.
		 **/
		void changeCalibration(int type);

		int getCalibration();
		void switchFile(int suff);
		void updateCalibration(int col, int row, int newcolour, uchar *yplane);
#endif

		void wlLoadNNMC(const uchar *data, int size);

		// uchar nn_multi_colour[CUBE_VOLUME];
		unsigned char *nn_multi_colour;
	private:
		int method;

		// used when we clip the cube
		int min_y, min_u, min_v, max_y, max_u, max_v;
		int u_size, v_size;

		// GCD methods
		int determinePlane(int val);
		int plane[256];
		uchar *seachart[NUM_PLANES];
};

#endif


