/*

   Copyright 2004 The University of New South Wales (UNSW) and National  
   ICT Australia (NICTA).

   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
   redistribute it and/or modify it under the terms of the GNU General  
   Public License as published by the Free Software Foundation; either  
   version 2 of the License, or (at your option) any later version as  
   modified below.  As the original licensors, we add the following  
   conditions to that license:

   In paragraph 2.b), the phrase "distribute or publish" should be  
   interpreted to include entry into a competition, and hence the source  
   of any derived work entered into a competition must be made available  
   to all parties involved in that competition under the terms of this  
   license.

   In addition, if the authors of a derived work publish any conference  
   proceedings, journal articles or other academic papers describing that  
   derived work, then appropriate academic citations to the original work  
   must be included in that publication.

   This rUNSWift source is distributed in the hope that it will be useful,  
   but WITHOUT ANY WARRANTY; without even the implied warranty of  
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
   General Public License for more details.

   You should have received a copy of the GNU General Public License along  
   with this source code; if not, write to the Free Software Foundation,  
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Copyright (c) 2004 UNSW
 * All Rights Reserved.
 *
 **/

#define OPENR_STUBGEN

#include <OPENR/core_macro.h>

#include <MCOOP.h>
#include <OPENR/OPENR.h>
#include <OPENR/OPENRMessages.h>
#include <OPENR/core_macro.h>
#include <ERA201D1.h>

#include "Vision.h"
#include "VisualCortex-priv.h"
#include "../share/Common.h"
#include "../share/VisionDef.h"
#include "../share/SharedMemoryDef.h"
#include "../share/WirelessTypeDef.h"
#include "Odometer.h"
#include "../share/OdometerDef.h"
#include "../behaviours/Behaviours.h"
#include "PyEmbed.h"
#include "InfoManager.h"

static const bool delayDebugMsg = false;
static const bool yuvDebugMsg = false;
static const bool newOplaneDebugMsg = false;
static const bool sensorValueDebugMsg = false;
static const bool sanityValuesDebugMsg = false;
static const bool compressedCPlaneDebugMsg = false;

// Must be one of send modes defined in Vision.h.
int sendCPlaneMode  = SEND_CPLANE_NORMAL; 
int CPLANE_INTERVAL = 30;

/*
 * The roll is multiplied by this factor so that only 4 bytes is used in the transmission, instead of 8.
 * has to agree among the following files: robot/vision/Vision.cc, client.h and RoboConstant.java
 **/
static const int SEND_SCALE = 1000;

// the threshold for determining if the dog has booted up as red or blue
static const int YELLOW_GOAL_MOUTH_THRESHOLD = -400000;

// this is sent when the robot recognise any object in the colour plane
static const unsigned char zeroBytes[INT_BYTES] = {
	0, 0, 0, 0
};

static const int objectOrder[NUM_OBJ] = {
	vobBlueGoal,
	vobYellowGoal,
	vobBlueLeftBeacon,
	vobBlueRightBeacon,
	vobGreenLeftBeacon,
	vobGreenRightBeacon,
	vobYellowLeftBeacon,
	vobYellowRightBeacon,
	vobRedDog,
	vobRedDog2,
	vobRedDog3,
	vobRedDog4,
	vobBlueDog,
	vobBlueDog2,
	vobBlueDog3,
	vobBlueDog4,
	vobBall
};

Vision::Vision() {
	OSYSDEBUG(("Vision::Vision()\n"));

	checkPrimitiveSize();

    Behaviours::InitBehaviours();

	powerIsOn = false;
	sharedMemIsSetup = false;
	camNotSet = true;
	wait4settle = 30;
	sharedrc = NULL;
	sharedDebugMemory = NULL;

	GetSystemTime(&lastTime);
	GetSystemTime(&curTime);

	frameCounter = 0;
	LastframePt = 0;

	Last_WM_Pt = 10 * 125;
    lastWalkType = -1;

    // if (pythonBehaviours) {
    PyBehaviours::sensors = &sensors;
    PyBehaviours::vision = &vision;
    PyBehaviours::gps = &gps;
    PyBehaviours::transmission = this;
    PyBehaviours::PWalkInfo = &PWalkInfo;
    //}
    //else {
    Behaviours::sensors = &sensors;
    Behaviours::vision = &vision;
    Behaviours::gps = &gps;
    Behaviours::transmission = this;
    //}
    gInfoManager.setSensors(&sensors);
    gInfoManager.setGPS(&gps);
    gInfoManager.setVisualCortex(&vision);

	cplaneCounter  = 0;
	oplaneCounter  = 0;

	displayEnvironmentCounter = 0;
	shareEnvironmentCounter = 0;

	delayTestingCounter = 0;

	gotTMWM = 0;
}

OStatus Vision::DoInit(const OSystemEvent &event) {
	OSYSDEBUG(("Vision::DoInit()\n"));

	// NewComData  /* generates the necessary number of subjects and observers */
	// SetComData  /* registers the communication service with other objects */
	NEW_ALL_SUBJECT_AND_OBSERVER;
	REGISTER_ALL_ENTRY;
	SET_ALL_READY_AND_NOTIFY_ENTRY;

	// wireless
	objectOut = subject[sbjSPEAKER];
	dogCommOut = subject[sbjDogComm];
	dogPhotoOut = subject[sbjDogPhoto];
	playerNumberOut = subject[sbjPlayerNumber];

	sSetIndication = subject[sbjSetIndication];
	sExecuteAA = subject[sbjExecuteAA];
	sGaitOptimiser = subject[sbjGaitOptimiser];
	sWalkingLearner = subject[sbjWalkingLearner];

	sensors.oSensor = observer[obsSensor];

	//wireless
	objectIn = observer[obsExternalCommand];
	dogCommIn = observer[obsDogComm];
	//dogCommLTIn = observer[obsDogCommLT];

	sharedrc = new RCRegion(sizeof(RC_DEBUG_DATA));
	sharedDebugMemory = (RC_DEBUG_DATA *) sharedrc->Base();

	EtherDriverGetMACAddressMsg myMsg;
	EtherStatus result;
	if ((result = ERA201D1_GetMACAddress(&myMsg)) != ETHER_OK) {
		cout << "Vision::DoInit: Bad ERA201D1_GetMACAddress return val: " << etherStatusStr(result) << endl;
	} else {
		ethernetMACaddress = ((myMsg.address.octet[2] << 24) | (myMsg.address.octet[3] << 16) |
								(myMsg.address.octet[4] << 8) | (myMsg.address.octet[5]));
		cout << "Got MAC address for dog: " << myMsg.address << " with low bytes: " << ethernetMACaddress << endl;
	}

	vision.SetDogParams(ethernetMACaddress);

	return oSUCCESS;
}

void Vision::sendWirelessMessage(int msgType,
		int dataType,
		int dataLength,
		void *sharedMem,
		const void *msg) {
	if (sharedMem == NULL) {
		switch (dataType) {
			case shmdtMessages: sharedMem = sharedDebugMemory->message;
					    break;
			case shmdtWorldModel: sharedMem = sharedDebugMemory->world_model_debug_data;
					      break;
			case shmdtCPlane: sharedMem = sharedDebugMemory->cplane_data;
					  break;
			case shmdtOPlane: sharedMem = sharedDebugMemory->oplane_data;
					  break;
			case shmdtYUVPlaneCPlane: sharedMem = sharedDebugMemory->yuvCplane_data;
						  break;
			case shmdtYUVPlane: sharedMem = sharedDebugMemory->yuvplane_data;
					    break;
			case shmdtEdgeDetectDebug: sharedMem = sharedDebugMemory->line_debug_data;
						   break;
			default: cout
				 << "WARNING Vision::sendWirelessMessage : setting unknown dataType or NULL shared memory"
					 << endl; 
				 return;
		}
	}

	memcpy(sharedMem, msg, dataLength);

	objectOut->ClearBuffer();
	objectOut->SetData(&msgType, sizeof(int));
	objectOut->SetData(&dataType, sizeof(int));
	objectOut->SetData(&dataLength, sizeof(int));
	objectOut->NotifyObservers();
}

void Vision::sendWirelessMessage(const string &str){
    sendWirelessMessage(msgTypeNORMAL,
            shmdtMessages,
            str.length(),
            NULL,
            str.c_str());
}

OStatus Vision::DoStart(const OSystemEvent &event) {
	OSYSDEBUG(("Vision::DoStart()\n"));

	ENABLE_ALL_SUBJECT;
	ASSERT_READY_TO_ALL_OBSERVER;
	/*
	   for (int i = 0; i < numOfObserver; ++i)
	   if (observer[i]->AssertReady () != oSUCCESS)
	   OSYSDEBUG(("AssertReady() failed. observer[%d]\n",i));
	   */
	// send player number to wireless
	sendPlayerNumber(Behaviours::PLAYER_NUMBER, -1);

	int msgType = msgTypeESTABLISHING;
	MemoryRegionID mem = sharedrc->MemID();
	size_t offset = sharedrc->Offset();
	size_t size = sharedrc->Size();

	objectOut->ClearBuffer();
	objectOut->SetData(&msgType, sizeof(int));
	objectOut->SetData(&mem, sizeof(mem));
	objectOut->SetData(&offset, sizeof(offset));
	objectOut->SetData(&sharedDebugMemory, sizeof(sharedDebugMemory));
	objectOut->SetData(&size, sizeof(size));
	objectOut->NotifyObservers();

	return oSUCCESS;
}

OStatus Vision::DoStop(const OSystemEvent &event) {
	//eCommander.stop();

	DISABLE_ALL_SUBJECT;
	DEASSERT_READY_TO_ALL_OBSERVER;

	//for (int i = 0; i < numOfObserver; ++i)
	//  observer[i]->DeassertReady();

	return oSUCCESS;
}

OStatus Vision::DoDestroy(const OSystemEvent &event) {
	DELETE_ALL_SUBJECT_AND_OBSERVER;

	/* deletes all observers and subjects spawned in DoInit() */
	DeleteComData         

		return oSUCCESS;
}


void Vision::ResultCamera(const ONotifyEvent &event) {
	// Setting the camera
	// not sure if this is mandatory
	// camNotSet = false;
	if (camNotSet) {
		SetCamera();
		observer[obsYUV]->AssertReady();
		return;
	}

	if (!sharedMemIsSetup) {
		observer[obsYUV]->AssertReady();
		return;
	}

	// Wait for everything to settle
	if (wait4settle > 0) {
		wait4settle--;
		observer[obsYUV]->AssertReady(event.SenderID());
		return;
	}
	// Setting goal
	if (gps.goalNotSet) {
		bool goalDir;
		int mouthPos = sensors.sensorVal[ssMOUTH];
	    if (mouthPos < YELLOW_GOAL_MOUTH_THRESHOLD) {
		// jaws open
		goalDir = false;
		cout << "Go to Yellow goal" << endl;
	    } else {
		// jaws closed
		goalDir = true;
		cout << "Go to Blue goal" << endl;
	    }
#ifdef ONLY_ONE_DOG
        if (Behaviours::PLAYER_NUMBER <= 2){
            goalDir = false;
        }
        else{
            goalDir = true;
        }
#endif//ONLY_ONE_DOG
                
		gps.SetGoals(goalDir);
		sendPlayerNumber(Behaviours::PLAYER_NUMBER, goalDir);
		//      if (targetGoal == vobYellowGoal)
		//          Produce_Sound(5);

        vision.invalidateEdgeMap();
	}

    if ((*PWalkInfo.currentWalkType) != lastWalkType) {
        vision.invalidateEdgeMap();
        lastWalkType = (*PWalkInfo.currentWalkType);
    }
    
    gps.sendCPlane = false;
    vision.bodyMoving = (Behaviours::left != 0 ||
                         Behaviours::forward != 0 ||
                         Behaviours::turnCCW != 0);
    vision.headMovement = sensors.getHeadSpeed();

#ifdef FRAMECOUNT
	SystemTime icur;
	GetSystemTime(&icur);
#endif //FRAMECOUNT

    long s0, s1, s2;
    s0 = sensors.sensorVal[ssHEAD_TILT1];
    s1 = sensors.sensorVal[ssHEAD_PAN];
    s2 = sensors.sensorVal[ssHEAD_TILT2];
	vision.ProcessImage(event, s0, s1, s2, sensors.range2obstacle, &PWalkInfo);

	/*if (vision.VISUALCORTEX_TIME_PRINT == 1) {
		//if the blobber has taken more than 30000us send a YUV plane of the incident. 
		if (vision.peakBlobTime > 30000) {
			sendYUVPlane();
		}
	}*/

#ifdef FRAMECOUNT
	SystemTime icur2;
	GetSystemTime(&icur2);
	SystemTime idiff = icur2 - icur;
	imgProcessTime = imgProcessTime + idiff;
#endif //FRAMECOUNT

    if (flag_LOC_CHALLENGE) {
        if (vision.lookForBeaconsAndGoals) {
        	gps.GPSVisionUpdate(vision.vob, vision.ballDistMethod);
            gps.MapLocalisePink(vision.localisePink, vision.numLocalisePink);
        } else {
            gps.LocalisePinkUpdate(vision.localisePink, vision.numLocalisePink);
        }

    } else gps.GPSVisionUpdate(vision.vob, vision.ballDistMethod);
        

    gps.GPSEdgeUpdate(&vision, &sensors);

#ifdef FRAMECOUNT
	SystemTime gcur;
	GetSystemTime(&gcur);
	SystemTime gdiff = gcur - icur2;
	gpsVisTime = gpsVisTime + gdiff;

	// Counting the number of frames
	frameCounter++;
    
#ifndef OFFLINE
        OPowerStatus currentStatus;
        OPENR::GetPowerStatus(&currentStatus);
        Behaviours::batteryCurrent = currentStatus.current ;
#endif

	if (vision.cameraFrame > LastframePt + 5 * 125) {
		LastframePt += 5 * 125;
		//cout << "Last 5 seconds averaged " << frameCounter/5.0 << " camera frames per second." << endl;
		//wout << "Last 5 seconds averaged " << frameCounter/5.0 << " camera frames per second.\n";
		GetSystemTime(&curTime);
		SystemTime diff = curTime - lastTime;
		double frames = frameCounter
			/ (diff.seconds + (diff.useconds / 1000000.0));
		int piSpeed = (int)
			((imgProcessTime.seconds*1000000.0 + imgProcessTime.useconds)
			 / frameCounter);
		int gpsSpeed = (int)
			((gpsVisTime.seconds*1000000.0 + gpsVisTime.useconds)
			 / frameCounter);
		int bSpeed = (int)
			((behavioursTime.seconds*1000000.0 + behavioursTime.useconds)
			 / frameCounter);

		sprintf(dataResultCamera,
				"(%d) - %d.%.6d s - %f cps\n   Aves : PI - %dus, gps - %dus, b - %dus current=%d\n",
				 Behaviours::PLAYER_NUMBER,
				(int)
				diff.seconds,
				(int)
				diff.useconds,
				frames,
				piSpeed,
				gpsSpeed,
				bSpeed,
                Behaviours::batteryCurrent);
		sendWirelessMessage(msgTypeNORMAL,
				shmdtMessages,
				strlen(dataResultCamera),
				sharedDebugMemory->message,
				dataResultCamera);
		frameCounter = 0;
		lastTime = curTime;
		imgProcessTime.seconds = 0;
		imgProcessTime.useconds = 0;
		gpsVisTime.seconds = 0;
		gpsVisTime.useconds = 0;
		behavioursTime.seconds = 0;
		behavioursTime.useconds = 0;
	}

#endif // FRAMECOUNT

	// Debug world model

	if (vision.cameraFrame > Last_WM_Pt + 1 * 125) {
		Last_WM_Pt += 1 * 125;

		/*
		   cout << "Start World Model here ";
		   cout << (long) rx << " " << (long) ry << " " << (long) rh << " " <<
		   (long) rcf;
		   cout << " " << (long) wmobj[0].d << " " << (long) wmobj[0].
		   h << " " << (long) wmobj[0].cf;
		   cout << " " << (long) wmobj[vobRedDog].
		   d << " " << (long) wmobj[vobRedDog].
		   h << " " << (long) wmobj[vobRedDog].cf;
		   cout << " " << (long) wmobj[vobBlueDog].
		   d << " " << (long) wmobj[vobBlueDog].
		   h << " " << (long) wmobj[vobBlueDog].cf;
		   cout << " End World Model here";
		   cout << endl;
		   */
	}

#ifdef FRAMECOUNT  
	SystemTime bcur;
	GetSystemTime(&bcur);
#endif // FRAMECOUNT
    Behaviours::NextAction();
#ifdef FRAMECOUNT  
	SystemTime bcur2;
	GetSystemTime(&bcur2);
	SystemTime bdiff = bcur2 - bcur;
	behavioursTime = behavioursTime + bdiff;

	SystemTime tdiff = bcur2 - icur;
    Behaviours::currentFrameID = vision.cameraFrame;

	lastFrameID = currentFrameID;
	currentFrameID = vision.cameraFrame;
	Behaviours::cpuUsage = (tdiff.seconds * 1000000.0 + tdiff.useconds);
#endif // FRAMECOUNT

	observer[obsYUV]->AssertReady(event.SenderID());

	// send across visual data to all other robots
}

//set default camera settings
void Vision::SetCamera() {

	//Parameter:
	//  White Balance:
	//      const longword ocamparamWB_INDOOR_MODE  = 1;
	//      const longword ocamparamWB_OUTDOOR_MODE = 2;
	//      const longword ocamparamWB_FL_MODE      = 3; // Fluorescent Lamp
	//
	//  Gain:
	//      const longword ocamparamGAIN_LOW        = 1;
	//      const longword ocamparamGAIN_MID        = 2;
	//      const longword ocamparamGAIN_HIGH       = 3;
	//
	//  Shutter Speed:
	//      const longword ocamparamSHUTTER_SLOW    = 1;
	//      const longword ocamparamSHUTTER_MID     = 2;
	//      const longword ocamparamSHUTTER_FAST    = 3;

	longword wb_val = ocamparamWB_INDOOR_MODE;
	//longword wb_val = ocamparamWB_OUTDOOR_MODE;
	//longword wb_val = ocamparamWB_FL_MODE;

	//longword gain_val = ocamparamGAIN_LOW;
	//longword gain_val = ocamparamGAIN_MID;
	longword gain_val = ocamparamGAIN_HIGH;

	//longword shutter_val = ocamparamSHUTTER_SLOW;
	//longword shutter_val = ocamparamSHUTTER_MID;
	longword shutter_val = ocamparamSHUTTER_FAST;

	SetCamera(wb_val, gain_val, shutter_val);

}

void Vision::SetCamera(longword wb_val, longword gain_val, longword shutter_val) {

	OPrimitiveID fbkID;
	OPENR::OpenPrimitive("PRM:/r1/c1/c2/c3/i1-FbkImageSensor:F1", &fbkID);

	// White Balance 
	OPrimitiveControl_CameraParam wb(wb_val);
	if (OPENR::ControlPrimitive(fbkID,
				oprmreqCAM_SET_WHITE_BALANCE,
				&wb,
				sizeof(wb),
				0,
				0) == oSUCCESS) {
		OSYSDEBUG(("CAM_SET_WHITE_BALANCE : Success\n"));
	}
	else {
		OSYSDEBUG(("CAM_SET_WHITE_BALANCE : Failed\n"));
	}

	// Gain 
	OPrimitiveControl_CameraParam gain(gain_val);
	if (OPENR::ControlPrimitive(fbkID,
				oprmreqCAM_SET_GAIN,
				&gain,
				sizeof(gain),
				0,
				0) == oSUCCESS) {
		OSYSDEBUG(("CAM_SET_GAIN : Success\n"));
	}
	else {
		OSYSDEBUG(("CAM_SET_GAIN : Failed\n"));
	}

	// Shutter Speed
	OPrimitiveControl_CameraParam shutter(shutter_val);
	if (OPENR::ControlPrimitive(fbkID,
				oprmreqCAM_SET_SHUTTER_SPEED,
				&shutter,
				sizeof(shutter),
				0,
				0) == oSUCCESS) {
		OSYSDEBUG(("CAM_SET_SHUTTER_SPEED : Success\n"));
	}
	else {
		OSYSDEBUG(("CAM_SET_SHUTTER_SPEED : Failed\n"));
	}

	camNotSet = false;
}

//the purpose of this fuction is to call GPSMotionUpdate
//and then assert ready.
//needs to be done here because gps doesn't have access to the observer array
void Vision::GPSMotionUpdateAssert(const ONotifyEvent &event) {
#ifdef FRAMECOUNT  
	SystemTime gcur;
	GetSystemTime(&gcur);
#endif // FRAMECOUNT   
	const double delta_forward = * (const double *) (event.Data(0));
	const double delta_left = * (const double *) (event.Data(1));
	const double delta_turn = * (const double *) (event.Data(2));
    long PWM = sensors.getMaxPWMValue();

	gps.GPSMotionUpdate(delta_forward, delta_left, delta_turn, PWM);
	observer[event.ObsIndex()]->AssertReady();
#ifdef FRAMECOUNT  
	SystemTime gcur2;
	GetSystemTime(&gcur2);
	SystemTime gdiff = gcur2 - gcur;
	gpsVisTime = gpsVisTime + gdiff;
#endif // FRAMECOUNT
}

void Vision::OdometerAssert(const ONotifyEvent &event) {
	/* Temporary, if it works, use stepComplete! */
	const OdometerData *data = (const OdometerData *) (event.Data(0));
	Odometer::inverse(data, (sensors.press_sensor[PRESS_BACK_REAR].count > 4));
	observer[event.ObsIndex()]->AssertReady();
}

/*void Vision::ResultWireless(const ONotifyEvent& event) {

//const char* text = (const char *)event.Data(0);
//OSYSPRINT(("Vision::ResultWireless() %s\n", text));
//++msgReceiveCounter;

processWireless(event);

objectIn->AssertReady();
}*/

void Vision::ExternalCommand(const ONotifyEvent &event) {
	const CommandData &cmd = *(const CommandData *) event.Data(0);
    Behaviours::processCommand(cmd);

	objectIn->AssertReady();
}

void Vision::TeammateWM(const ONotifyEvent &event) {
	gotTMWM = 1;
	const WMShareInfo *wm = (const WMShareInfo *) event.Data(0);

	/*KEV_DEBUG (cout << "In Vision TeammateWM\n" ;);
	KEV_DEBUG (cout << " ---- RECEIVED PLAYER: " << wm->playerNum << "    my pos: ( " << wm->obj.posx 
		      << " , " << wm->obj.posy << " )" << "       ball pos: ( " << wm->obj.ballx << " , " <<
		      wm->obj.posy << " )" << endl;);*/

	gps.processShareInfo(wm->playerNum, wm->obj);

	dogCommIn->AssertReady();
}

// Taken out by Ross due to change in shared memory use
/*void Vision::LatencyTesting(const ONotifyEvent& event) {

  int msgType = *(const int *)event.Data(0);
  int offset = 0;

//Wireless Latency
offset = REGION_H;
WDataInfo info = *(WDataInfo *) (sharedBase+offset);
int arrayLength = info.size;

// Test for array of 'a'
char world[arrayLength];
memcpy(world, sharedBase+offset+sizeof(WDataInfo), arrayLength);


//byte *world = (byte *) sharedBase+offset+sizeof(WDataInfo);
//Test World model Data
for (int a = 0; a < arrayLength; a++){
world [a] = 'b';
//cout << a << (char) world[a] << "  ";
}
//cout << "Received correctly.. " << endl;

//Behaviours::wreceive(info.robot, info.type, info.size, obj);


int dataType = 500;
//cout << "ArrayLength : " << arrayLength << endl;
memcpy(sharedBase, &dataType, sizeof(int));
memcpy(sharedBase+sizeof(int), &arrayLength, sizeof(int));
memcpy(sharedBase+REGION_F, world, arrayLength);
int msgType2 = 1;
objectOut->ClearBuffer();
objectOut->SetData(&msgType2, sizeof(int));
objectOut->NotifyObservers();

dogCommLTIn->AssertReady();
}*/

void Vision::ResultSensor(const ONotifyEvent &event) {
	sensors.ResultSensor(event);
	/* Store new set of joint angles. */
	Odometer::storeData(sensors.sensorVal[ssFL_ROTATOR],
			sensors.sensorVal[ssFL_ABDUCTOR],
			sensors.sensorVal[ssFL_KNEE],
			sensors.sensorVal[ssFR_ROTATOR],
			sensors.sensorVal[ssFR_ABDUCTOR],
			sensors.sensorVal[ssFR_KNEE],
			sensors.sensorVal[ssRL_ROTATOR],
			sensors.sensorVal[ssRL_ABDUCTOR],
			sensors.sensorVal[ssRL_KNEE],
			sensors.sensorVal[ssRR_ROTATOR],
			sensors.sensorVal[ssRR_ABDUCTOR],
			sensors.sensorVal[ssRR_KNEE]);
}

void Vision::SetPAgent(const ONotifyEvent &event) {
	const MemoryRegionID mem = *(const MemoryRegionID *) event.Data(0);
	const size_t offset = *(const size_t *) event.Data(1);
	const voidPtr base = *(const voidPtr *) event.Data(2);
	const size_t size = *(const size_t *) event.Data(3);

	if (size != sizeof(ActuatorSharedMemory)) {
		cout << "ERROR: Actuator Shared Memory definition mismatch!" << endl;
	}

	shmPWalk = new RCRegion(mem, offset, base, size);

	ActuatorSharedMemory *ptr = (ActuatorSharedMemory *) shmPWalk->Base();

	if (ptr == NULL) {
		cout << "Bad Actuator Control Shared Memory Base Pointer" << endl;
	}
	else {
		cout << "Vision has received ActuatorControl shared memory: " << (int) ptr
			<< endl;
        Behaviours::leading_leg = &(ptr->leading_leg);
        Behaviours::currentWalkType = &(ptr->currentWalkType);
        Behaviours::shareStep_ = &(ptr->currentStep);
        Behaviours::currentPG = &(ptr->currentPG);
        Behaviours::desiredPan = &(ptr->desiredPan);
        Behaviours::desiredTilt = &(ptr->desiredTilt);
        Behaviours::desiredCrane = &(ptr->desiredCrane);
        //Behaviours::stepComplete = &(ptr->stepComplete);

        PWalkInfo.leading_leg = &(ptr->leading_leg);
        PWalkInfo.currentWalkType = &(ptr->currentWalkType);
        PWalkInfo.currentStep = &(ptr->currentStep);
        PWalkInfo.currentPG = &(ptr->currentPG);
        PWalkInfo.frontShoulderHeight = &(ptr->frontShoulderHeight);
        PWalkInfo.bodyTilt = &(ptr->bodyTilt);
        PWalkInfo.desiredPan = &(ptr->desiredPan);
        PWalkInfo.desiredTilt = &(ptr->desiredTilt);
        PWalkInfo.desiredCrane = &(ptr->desiredCrane);
		PWalkInfo.isHighGain = &(ptr->isHighGain);

		vision.currentWalkType = &(ptr->currentWalkType);

		sharedMemIsSetup = true;
	}
	//I dont assert ready at the end
	//because i dont expect to be getting a message to set up the shared mem region
	//thus activating this function more than once.

}




void Vision::shareEnvironment() {
	if (shareEnvironmentCounter == 0) {
	  
		WMShareObj obj = gps.getShareInfo();

		dogCommOut->ClearBuffer();
		dogCommOut->SetData(&obj, sizeof(obj));
		dogCommOut->NotifyObservers();
	}
	shareEnvironmentCounter++;
	shareEnvironmentCounter %= SHARE_ENVIRONMENT_INTERVAL;
}

void Vision::sendPlayerNumber(int playerNumber, bool isRedRobot) {
	int team = isRedRobot?0:1;
	sendPlayerNumber(playerNumber, team);
}

void Vision::sendPlayerNumber(int playerNumber, int team) {
	playerNumberOut->SetData(&playerNumber, sizeof(int));
	playerNumberOut->SetData(&team, sizeof(int));
	playerNumberOut->NotifyObservers();
}

void Vision::displayEnvironment() {
	if (displayEnvironmentCounter == 0) {
		static const int TEAMMEMBERS_BEING_SENT = (NUM_TEAM_MEMBER - 1);
		static const int NUM_ATTR_SENT_ROB = 9;
		static const int NUM_ATTR_VIS_PULL = 3;
		static const int NUM_ATTR_SENT_TEAMMATE = 5;
		static const int NUM_ATTR_SENT_BALL = 9;    //for now x y vx vy varx vary varvx varvy
		static const int NUM_ATTR_SENT_WBALL = 3; // source robot num, x & y of wireless ball position
		static const int NUM_ATTR_SENT_OPPONENTS = 5;

		const int arraySize = 2
			+                      // for the player number and the team number

			1 +                      // for the view variable describing things we can see

			3
			+                      // for the last direct update values

			1 +                      // for the size of the remaining vision model data
			NUM_ATTR_SENT_ROB
			+      // the new world model data (x y h varx vary varh varxy varxh varyh)
			NUM_ATTR_VIS_PULL +   // the exponentially decaying visual pull
			NUM_ATTR_SENT_BALL
			+ NUM_ATTR_SENT_WBALL
			+    // for the wireless ball
			TEAMMEMBERS_BEING_SENT *NUM_ATTR_SENT_TEAMMATE + 

			// 1 +                   // for the size of the remaining world model data
			// NUM_ATTR_SENT_ROB +   // the vision data (x y h varx vary varh varxy varxh varyh)
			// NUM_ATTR_SENT_BALL +

			1
			+                      // for the size of the remaining wireless model data
			NUM_ATTR_SENT_ROB +      // the old world model data (x y h varx vary varh varxy varxh varyh)

			1
			+                      // for size of opponents data
			NUM_OPPONENTS *NUM_ATTR_SENT_OPPONENTS;/* +
            
            3 +
            vision.MAX_EDGE_PIXELS * 2; // Derrick: used for debugging edge detection*/
            


		float data[arraySize];
		int ptr = 0;

		//set player number
		data[ptr] = Behaviours::PLAYER_NUMBER;
		ptr++;

		// set direction team color
		if (gps.targetGoal == vobBlueGoal) {
			data[ptr] = P_RED_DOG;
		}
		else {
			data[ptr] = P_BLUE_DOG;
		}
		ptr++;

		// what we can see
		data[ptr++] = (double) gps.view;

		data[ptr++] = (double) gps.directx;
		data[ptr++] = (double) gps.directy;
		data[ptr++] = (double) gps.directh;

		/////////////////////
		// NEW WORLD MODEL
		/////////////////////
		//cout << "*************** got to world model sending *********" << endl;


		data[ptr] = NUM_ATTR_SENT_ROB
			+ NUM_ATTR_VIS_PULL
			+ NUM_ATTR_SENT_BALL
			+ NUM_ATTR_SENT_WBALL
			+ (TEAMMEMBERS_BEING_SENT * NUM_ATTR_SENT_TEAMMATE);
		ptr++;

		const WMObj &self = gps.self();
		const MMatrix3 scov = gps.selfCov();

		data[ptr] = self.pos.x;
		data[ptr + 1] = self.pos.y;
		data[ptr + 2] = self.h;
		data[ptr + 3] = scov(0, 0); //varx
		data[ptr + 4] = scov(1, 1); //vary
		data[ptr + 5] = scov(2, 2); //varh
		data[ptr + 6] = scov(0, 1); //varxy
		data[ptr + 7] = scov(0, 2); // varxh
		data[ptr + 8] = scov(1, 2); // varyh
		ptr += NUM_ATTR_SENT_ROB;

		const MVec3 &vp = gps.getVisualPull();
		data[ptr] = vp(0, 0);
		data[ptr + 1] = vp(1, 0);
		data[ptr + 2] = vp(2, 0);
		ptr += NUM_ATTR_VIS_PULL;

		//WMObj& ball = gps.ball('g');
		const Vector &ball = gps.getBall(GLOBAL);
		//WMObj& vball = gps.vBall();
		const Vector &vball = gps.getVBall(GLOBAL);
		const MMatrix2 &bcov = gps.getBallCov();
		const MMatrix2 &bvcov = gps.getVelCov();
		//const MMatrix4& bcov = gps.bCov();
		data[ptr] = ball.x; //ball.pos.x;
		data[ptr + 1] = ball.y; //ball.pos.y;
		data[ptr + 2] = vball.x; //vball.pos.x;
		data[ptr + 3] = vball.y; //vball.pos.y;
		data[ptr + 4] = bcov(0, 0); //varx
		data[ptr + 5] = bcov(1, 1); //vary
		data[ptr + 6] = bcov(0, 1); //varxy
		data[ptr + 7] = bvcov(0, 0); //varvx
		data[ptr + 8] = bvcov(1, 1); //varvy
		ptr += NUM_ATTR_SENT_BALL;

		if (gps.shareBallvar < LARGE_VAL) {
			// wireless ball data is valid
			data[ptr] = gps.sbRobotNum;
			data[ptr + 1] = gps.shareBallx;
			data[ptr + 2] = gps.shareBally;
		}
		else {
			// wireless ball data is invalid
			data[ptr] = 0;
			data[ptr + 1] = 0;
			data[ptr + 2] = 0;
		}
		ptr += NUM_ATTR_SENT_WBALL;

		for (int i = 0; i < NUM_TEAM_MEMBER; i++) {
			if (i != (Behaviours::PLAYER_NUMBER)
						- 1) {
				const WMObj &tm = gps.tmObj(i);
				data[ptr] = tm.pos.x;
				data[ptr + 1] = tm.pos.y;
				data[ptr + 2] = tm.h;
				data[ptr + 3] = tm.posVar;
				data[ptr + 4] = tm.hVar;
				ptr += NUM_ATTR_SENT_TEAMMATE;
			}
		}

		/*
		   cout << "**************" << endl;
		   cout << "my player number is " << (Behaviours::PLAYER_NUMBER) << endl;
		   cout << "my location is " << gps.r.pos.x << "," << gps.r.pos.y << endl;
		   for (int i=0; i<NUM_TEAM_MEMBER; i++) {
		   if (i!=((Behaviours::PLAYER_NUMBER)-1)) {
		   cout << "my teammates number " << (i+1) << "'s location is " << gps.wmTeammate[i].pos.x << "," << 
		   gps.wmTeammate[i].pos.y << endl;
		   }
		   }
		   cout << "**************" << endl;
		   */

		/////////////////////
		// VISION MODELS
		/////////////////////

		//data[ptr] = NUM_ATTR_SENT_ROB + NUM_ATTR_SENT_BALL; //number of attributes being sent over
		//ptr++;


		// note - there is no vision state any more as kalman filtering has changed
		// this should be eventually deleted
		/*data[ptr]     = 0;
		  data[ptr + 1] = 0;
		  data[ptr + 2] = 0;
		  data[ptr + 3] = 1; //varx
		  data[ptr + 4] = 1; //vary
		  data[ptr + 5] = 1; //varh
		  data[ptr + 6] = 0; //varxy
		  data[ptr + 7] = 0; // varxh
		  data[ptr + 8] = 0; // varyh
		  ptr += NUM_ATTR_SENT_ROB;*/

		/*data[ptr]     = gps.newBallx;
		  data[ptr + 1] = gps.newBally;
		  data[ptr + 2] = gps.newBallvx;
		  data[ptr + 3] = gps.newBallvy;
		  data[ptr + 4] = (*(gps.ballMeasCov))(0,0); //varx
		  data[ptr + 5] = (*(gps.ballMeasCov))(1,1); //vary
		  data[ptr + 6] = (*(gps.ballMeasCov))(0,1); //varxy
		  data[ptr + 7] = (*(gps.ballMeasCov))(2,2); //varvx
		  data[ptr + 8] = (*(gps.ballMeasCov))(3,3); //varvy
		  ptr += NUM_ATTR_SENT_BALL;*/

		/////////////////////
		// OLD WORLD MODEL
		/////////////////////

		data[ptr] = NUM_ATTR_SENT_ROB; //number of attributes being sent over
		ptr++;

		data[ptr] = gps.oldx;
		data[ptr + 1] = gps.oldy;
		data[ptr + 2] = gps.oldh;
		data[ptr + 3] = gps.oldvx; //varx
		data[ptr + 4] = gps.oldvy; //vary
		data[ptr + 5] = gps.oldvh; //varh
		data[ptr + 6] = gps.oldcxy; //varxy
		data[ptr + 7] = gps.oldcxh; //varxh
		data[ptr + 8] = gps.oldcyh; //varyh
		ptr += NUM_ATTR_SENT_ROB;


		/////////////////////
		// OPPONENTS
		/////////////////////
		data[ptr] = NUM_OPPONENTS * NUM_ATTR_SENT_OPPONENTS;
		ptr++;
		for (int i = 0 ; i < NUM_OPPONENTS ; i++) {
			const Vector v = gps.getOppPos(i, GLOBAL);
			const MMatrix2 &m = gps.getOppCov(i);
			data[ptr] = v.x;
			data[ptr + 1] = v.y;
			data[ptr + 2] = m(0, 0);
			data[ptr + 3] = m(1, 1);
			data[ptr + 4] = m(0, 1);
			ptr += NUM_ATTR_SENT_OPPONENTS;
		}

        // Derrick: for debugging - Sends projected points along with
        // the world model data.
        /*data[ptr++] = (float) vision.numEdgePnts;
        data[ptr++] = (float) vision.MAX_EDGE_PIXELS;
        if (gps.useLineData) data[ptr++] = 1.0;
        else data[ptr++] = 0.0;
        
		if (gps.targetGoal == vobBlueGoal) {
            for (int i = 0; i < (vision.numEdgePnts * 2); i++)
                data[ptr + i] = (float) vision.globalEdgePnts[i];
        } else {
            for (int i = 0; i < (vision.numEdgePnts * 2); i++) {
                data[ptr + i] = (float) vision.globalEdgePnts[i];

                if (i & 1) data[ptr + i] = FIELD_LENGTH - data[ptr + i];
                else data[ptr + i] = FIELD_WIDTH - data[ptr + i];
            }
        }
        for (int i = vision.numEdgePnts * 2; i < vision.MAX_EDGE_PIXELS * 2; i++)
            data[ptr + i] = 0.0;
            
        ptr += (vision.MAX_EDGE_PIXELS * 2);*/

		int dataType = shmdtWorldModel; //i assume this is the number that says "world model stuff"
		int arrayLength = sizeof(data);


		memcpy(sharedDebugMemory->world_model_debug_data, &data, arrayLength);

		int msgType = msgTypeNORMAL;
		objectOut->ClearBuffer();
		objectOut->SetData(&msgType, sizeof(int));
		objectOut->SetData(&dataType, sizeof(int));
		objectOut->SetData(&arrayLength, sizeof(int));
		objectOut->NotifyObservers();
	}        
	displayEnvironmentCounter++;
	displayEnvironmentCounter %= DISPLAY_ENVIRONMENT_INTERVAL;
}


int Vision::storeSanityValues(unsigned char *planeArray, int offset) {
	int i, j, total;
	unsigned char * dataPtr;

	for (i = 0; i < NUM_SANITY_COORDS; i++) {
		total = vision.getSanityArray(i);
		dataPtr = (unsigned char *) &total;
		for (j = 0; j < INT_BYTES; j++) {
			planeArray[offset++] = dataPtr[j];
		}

		if (sanityValuesDebugMsg) {
			cout << "Vision.cc: i = " << i << "; NUM_SANITY_COORDS = "
				<< NUM_SANITY_COORDS << "; total = " << total << "; calculated ";
			cout
				<< ((int) ((planeArray[offset - 4] & 0xff)))
				+ ((int) ((planeArray[offset - 3] & 0xff) << 8))
				+ ((int) ((planeArray[offset - 2] & 0xff) << 16))
				+ ((int) ((planeArray[offset - 1] & 0xff) << 24))
				<< endl;
		}
	}
	return offset;
}

/* Store visual objects / body state values (pan, tilt, horizon) for
** RoboCommander to log. */  
int Vision::storeObjects(unsigned char *planeArray, int offset) {
	int i, j, k, value, objectIndex;
	unsigned char * dataPtr;
	bool objectSend = false;

	/* Store the pan. */
	value = (int) (vision.pan * SEND_SCALE);
	dataPtr = (unsigned char *) &value;
	for (k = 0; k < INT_BYTES; k++) {
		planeArray[offset++] = dataPtr[k];
	}

	/* Store the tilt. bodyTilt is also included so that roll is more accurate. 
       It has the same direction as tilt, and is also in radian. */
	value = (int) ((vision.tilt + vision.bodyTilt) * SEND_SCALE);
	dataPtr = (unsigned char *) &value;
	for (k = 0; k < INT_BYTES; k++) {
		planeArray[offset++] = dataPtr[k];
	}

    /* Crane doesn't need to be sent because only pan and tilt are required to
    ** calcualte the rotation. */


    /* Still not successful but keep for future use. */
#ifdef not_yet_dan_28_Mar_2004

    /* Send the horizon information. */
    /* 4 * INT_BYTES + 1 * INT_BYTES bytes sent altogether. */

    for (int l = 0; l < 3; l++) {

        /* Don't really need double to display horizon. */
        value = (int) vision.hzArray[l];
        dataPtr = (unsigned char *) &value;
        for (k = 0; k < INT_BYTES; k++) {
            planeArray[offset++] = dataPtr[k];
        }
    }
    /* Convert 3 bools into 1 int using binary representation. */
    /* Important: When reading the info in receive side needs to read these
    ** variables in order. */

    /* TODO: Can put in 1 byte instead, but too lazy to try new things, just
    ** use current setup. */
    value = 0;
    value = vision.hzExists  ? (value | 1) : value;
    value = vision.hzUpIsSky ? (value | 2) : value;
    value = vision.hzAllAboveHorizon ? (value | 4) : value;

	dataPtr = (unsigned char *) &value;
	for (k = 0; k < INT_BYTES; k++) {
		planeArray[offset++] = dataPtr[k];
	}
#endif

	// send all stationary objects first (beacons + goals), 
	// followed by all non-stationary objects (dogs + ball).
	// i = 1..8  = stationary beacons + goals
	// i = 9..16 = non-stationary dogs
	// i = 0     = non-stationary ball
	for (i = 0; i < NUM_OBJ; i++) {
		objectIndex = objectOrder[i];
		if (vision.vob[objectIndex].cf > 0
				|| vision.vob[objectIndex].var < LARGE_VAL) {
			if (newOplaneDebugMsg) {
				cout << "objectIndex: " << objectIndex << "; x: "
					<< vision.vob[objectIndex].x << "; y: "
					<< vision.vob[objectIndex].y << "; width: "
					<< vision.vob[objectIndex].width << "; height: "
					<< vision.vob[objectIndex].height << "; d: "
					<< vision.vob[objectIndex].d << endl;
			}
			objectSend = true;
			for (j = 0; j < NUM_INT_COORDS; j++) {
				switch (j) {
					case 0: value = objectIndex; // sends the object number
						break;
					case 1: value = (int) vision.vob[objectIndex].x;
						break;
					case 2: value = (int) vision.vob[objectIndex].y;
						break;
					case 3: value = (int) vision.vob[objectIndex].width;
						break;
					case 4: value = (int) vision.vob[objectIndex].height;
						break;
					case 5: value = (int) vision.vob[objectIndex].d;
						break;
					default: // do nothing
						break;
				}
				dataPtr = (unsigned char *) &value;
				for (k = 0; k < INT_BYTES; k++) {
					planeArray[offset++] = dataPtr[k];
				}
			}
		}
		// don't send empty plane
	}
	if (!objectSend) {
		/* this check is to make sure at least one (filled or empty) plane get sent to RoboCommander, 
		 * in which a blank screen will be showing up in the OPlane
		 * so that it will be realized in the debug tool
		 **/
		for (j = 0; j < NUM_INT_COORDS; j++) {
			for (k = 0; k < INT_BYTES; k++) {
				planeArray[offset++] = zeroBytes[k];
			}
		}
	}
	return offset;
}


void Vision::sendCompressedCPlane() {

    /* Use the cplaneCounter in a simple way usually. */
    if (sendCPlaneMode == SEND_CPLANE_NORMAL) {

        // send CPlane per every 'CPLANE_INTERVAL' frame
        cplaneCounter = (cplaneCounter + 1) % CPLANE_INTERVAL;
    }


    if (sendCPlaneMode == SEND_CPLANE_NORMAL) {
        if (cplaneCounter && Behaviours::burstCounter == 0) return;

    /* If send it in other modes, cplaneCounter is slightly different. Want to
    ** make sure we do not send every frame when condition is true, and do not
    ** want sending only when the interval conincides with the condition. */
    } else {

        if (cplaneCounter < CPLANE_INTERVAL) 
            cplaneCounter ++;

        if (sendCPlaneMode == SEND_CPLANE_ONLY_BAD_GPS) {
            if (!gps.sendCPlane) return;


        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_SEEING_BALL) {
            if (!(vision.vob[vobBall].cf > 0)) return; 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_NOT_SEEING_BALL) {
            if (vision.vob[vobBall].cf > 0) return; 


        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_SEEING_BLUE_GOAL) {
            if (!(vision.vob[vobBlueGoal].cf > 0)) return; 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_NOT_SEEING_BLUE_GOAL) {
            if (vision.vob[vobBlueGoal].cf > 0) return; 


        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_SEEING_YELLOW_GOAL) {
            if (!(vision.vob[vobYellowGoal].cf > 0)) return; 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_NOT_SEEING_YELLOW_GOAL) {
            if (vision.vob[vobYellowGoal].cf > 0) return; 
 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_SEEING_BLUE_LEFT_BEACON) {
            if (!(vision.vob[vobBlueLeftBeacon].cf > 0)) return; 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_NOT_SEEING_BLUE_LEFT_BEACON) {
            if (vision.vob[vobBlueLeftBeacon].cf > 0) return; 
 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_SEEING_BLUE_RIGHT_BEACON) {
            if (!(vision.vob[vobBlueRightBeacon].cf > 0)) return; 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_NOT_SEEING_BLUE_RIGHT_BEACON) {
            if (vision.vob[vobBlueRightBeacon].cf > 0) return; 
 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_SEEING_YELLOW_LEFT_BEACON) {
            if (!(vision.vob[vobYellowLeftBeacon].cf > 0)) return; 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_NOT_SEEING_YELLOW_LEFT_BEACON) {
            if (vision.vob[vobYellowLeftBeacon].cf > 0) return; 
 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_SEEING_YELLOW_RIGHT_BEACON) {
            if (!(vision.vob[vobYellowRightBeacon].cf > 0)) return; 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_NOT_SEEING_YELLOW_RIGHT_BEACON) {
            if (vision.vob[vobYellowRightBeacon].cf > 0) return; 


        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_SEEING_BEACONS) {
            if (!(   (vision.vob[vobYellowLeftBeacon].cf > 0)
                  || (vision.vob[vobYellowRightBeacon].cf > 0)
                  || (vision.vob[vobBlueLeftBeacon].cf > 0)
                  || (vision.vob[vobBlueRightBeacon].cf > 0))
               )
               return; 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_NOT_SEEING_BEACONS) {
            if (   (vision.vob[vobYellowLeftBeacon].cf > 0)
                  || (vision.vob[vobYellowRightBeacon].cf > 0)
                  || (vision.vob[vobBlueLeftBeacon].cf > 0)
                  || (vision.vob[vobBlueRightBeacon].cf > 0)
               )
               return; 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_SEEING_RED_DOG) {
            if (!(vision.vob[vobRedDog].cf > 0)) return; 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_NOT_SEEING_RED_DOG) {
            if (vision.vob[vobRedDog].cf > 0) return; 


        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_SEEING_BLUE_DOG) {
            if (!(vision.vob[vobBlueDog].cf > 0)) return; 

        } else if (sendCPlaneMode == SEND_CPLANE_ONLY_NOT_SEEING_BLUE_DOG) {
            if (vision.vob[vobBlueDog].cf > 0) return; 

        }


        /* Have ball, but has already sent a while ago. */
        if (cplaneCounter < CPLANE_INTERVAL) { 
            return;

        /* Send cplane.*/
        } else {
            cplaneCounter = 0;
        }
    }
    
	//cout << "display cplane " << Behaviours::burstCounter << endl;

	// variable for coding
	uchar* cp = vision.cplane;
	//uchar* cp = vision.edgePlane;
	unsigned char * compressedSizePtr;
	int offset = INT_BYTES; // leave space for the compressed cplane size
	int compressedSize;
	int start;

	//Runlength Encode cplane

	for(int y=0; y < HEIGHT; y++) {
	  start = 0;
	  for(int x=0; x < WIDTH; x++, cp++) {
	    if (*cp != *(cp - 1) || x == WIDTH - 1) {

	      if(x == WIDTH - 1) x++;

	      compressedCPlane[offset++] = x - start;
	      compressedCPlane[offset++] = *(cp-1);
	      start = x;  // reset start for segment just started.
	    }
	  }
	}

	/*
	//When splane has every segment saved
	for(int y = 0, count = 0; y < HEIGHT; y++) {
	  count = 0;
	  for(int x = 0; x < vision.esplane->segmentCount[y]; x++) {
	    count += vision.esplane->segment[y][x].length;
	    if(count <= WIDTH) {
	      compressedCPlane[offset++] = vision.esplane->segment[y][x].length;
	      if(vision.esplane->segment[y][x].isEdge)
		compressedCPlane[offset++] = vision.esplane->segment[y][x].color | 0x10;
	      else
		compressedCPlane[offset++] = vision.esplane->segment[y][x].color;
	    }
	  }
	  if(count < WIDTH) { //shouldn't happen
	    compressedCPlane[offset++] = WIDTH - count;
	    compressedCPlane[offset++] = 13; //maroon - a non-field colour
	  }
	}
	*/
	// store the size of compressed cplane at the begining of the array
	if (compressedCPlaneDebugMsg) {
		cout << "Vision.cc: ";
	}
	compressedSize = offset - INT_BYTES;
	compressedSizePtr = (unsigned char *) &compressedSize;
	for (int j = 0; j < INT_BYTES; j++) {
		compressedCPlane[j] = compressedSizePtr[j];
	}
	if (compressedCPlaneDebugMsg) {
		cout << "compressedCPlane: ";
		cout << (int) (compressedCPlane[0] & 0xff) << " ";
		cout << (int) ((compressedCPlane[1] & 0xff) << 8) << " ";
		cout << (int) ((compressedCPlane[2] & 0xff) << 16) << " ";
		cout << (int) ((compressedCPlane[3] & 0xff) << 24) << " = ";
		cout << ((int) (compressedCPlane[0] & 0xff)
			 + (int) ((compressedCPlane[1] & 0xff) << 8)
			 + (int) ((compressedCPlane[2] & 0xff) << 16)
			 + (int) ((compressedCPlane[3] & 0xff) << 24))
		     << endl;
	}

	if (compressedCPlaneDebugMsg) {
		cout << "cplane: " << offset << " ";
	}
	offset = storeSensors(compressedCPlane, offset);
	if (compressedCPlaneDebugMsg) {
		cout << "sensor: " << offset << " ";
	}
	offset = storeSanityValues(compressedCPlane, offset);
	if (compressedCPlaneDebugMsg) {
		cout << "sanity: " << offset << " ";
	}
	offset = storeObjects(compressedCPlane, offset);
	if (compressedCPlaneDebugMsg) {
		cout << "object: " << offset << " ";
	}
	offset = gps.storeGps(compressedCPlane, offset);
	if (compressedCPlaneDebugMsg) {
		cout << "gps   : " << offset << " ";
	}

	if (isDelayTesting) {
		compressedCPlane[offset++] = (byte) (delayTestingCounter >> 24);
		compressedCPlane[offset++] = (byte) (delayTestingCounter >> 16);
		compressedCPlane[offset++] = (byte) (delayTestingCounter >> 8);
		compressedCPlane[offset++] = (byte) delayTestingCounter;

		cout << "UNSWSkills.cc ---------- delayTestingCounter: "
			<< delayTestingCounter << " ----------" << endl;

		delayTestingCounter++;
	}

	memcpy(sharedDebugMemory->cplane_data, compressedCPlane, offset);

	if (compressedCPlaneDebugMsg) {
		cout << "total = " << offset << endl;
	}

	int msgType = msgTypeNORMAL;
	int dataType = shmdtCPlane ;
	objectOut->ClearBuffer();
	objectOut->SetData(&msgType, sizeof(int));
	objectOut->SetData(&dataType, sizeof(int));
	objectOut->SetData(&offset, sizeof(int));
	objectOut->NotifyObservers();
}

int Vision::storeSensors(unsigned char *planeArray, int offset) {
	int value, i, j;
	unsigned char * dataPtr;

	if (sensorValueDebugMsg) {
		cout << "Sensor: ";
	}

	for (i = 0; i < NUM_SENSOR_COORDS; i++) {
		value = vision.getSensor(i);
		dataPtr = (unsigned char *) &value;
		//        for (j=0; j<INT_BYTES; j++) {
		for (j = 0; j < LONG_BYTES; j++) {
			planeArray[offset++] = dataPtr[j];
		}
		if (sensorValueDebugMsg) {
			cout << value << " ";
		}
	}
	if (sensorValueDebugMsg) {
		cout << value << endl;
	}

	return offset;
	}

	void Vision::sendOPlane() {
		// send OPlane per every 'OPLANE_INTERVAL' frame
		oplaneCounter = (oplaneCounter + 1) % OPLANE_INTERVAL;
		if (oplaneCounter) {
			return;
		}

		int offset = storeObjects(oplane, 0);

		int dataType = shmdtOPlane ;
		memcpy(sharedDebugMemory->oplane_data, oplane, offset);

		int msgType = 1;
		objectOut->ClearBuffer();
		objectOut->SetData(&msgType, sizeof(int));  
		objectOut->SetData(&dataType, sizeof(int));  
		objectOut->SetData(&offset, sizeof(int));                  
		objectOut->NotifyObservers();
	}

	void Vision::sendYUVPlane() {
		int msgType = msgTypeNORMAL;
		int dataType = shmdtYUVPlaneCPlane ;
		int arrayLength = CPLANE_SIZE;

		memcpy(sharedDebugMemory->yuvCplane_data, vision.cplane, arrayLength);

		objectOut->ClearBuffer();
		objectOut->SetData(&msgType, sizeof(int));
		objectOut->SetData(&dataType, sizeof(int));
		objectOut->SetData(&arrayLength, sizeof(int));
		objectOut->NotifyObservers();

		if (yuvDebugMsg) {
			cout << "Vision.cc: sendYUVPlane - CPlane Size : " << arrayLength << endl;
		}

		dataType = shmdtYUVPlane ;
		arrayLength = YUVPLANE_SIZE;

		memcpy(sharedDebugMemory->yuvplane_data, vision.yplane, arrayLength);

		objectOut->ClearBuffer();
		objectOut->SetData(&msgType, sizeof(int));
		objectOut->SetData(&dataType, sizeof(int));
		objectOut->SetData(&arrayLength, sizeof(int));
		objectOut->NotifyObservers();

		if (yuvDebugMsg) {
			cout << "Vision.cc: sendYUVPlane - YPlane Size : " << arrayLength << endl;
		}
	}


	void Vision::sendAA(AtomicAction &nextAA) {
		//sExecuteAA->ClearBuffer(obsExecuteAA); //obsExecuteAA is the observer index of the observer reveiving a msg from this transmission.. number is generated automatically by stub.cfg
		sExecuteAA->ClearBuffer(); //obsExecuteAA is the observer index of the observer reveiving a msg from this transmission.. number is generated automatically by stub.cfg
		sExecuteAA->SetData(&nextAA, sizeof(nextAA));
		sExecuteAA->NotifyObservers();
	}

	void Vision::sendOP(OffsetParams &nextOP) {
		sGaitOptimiser->ClearBuffer();
		sGaitOptimiser->SetData(&nextOP, sizeof(nextOP));
		sGaitOptimiser->NotifyObservers();
	}

	void Vision::sendLearningParams(WalkLearningParams &params) {
		sWalkingLearner->ClearBuffer();
		sWalkingLearner->SetData(&params, sizeof(params));
		sWalkingLearner->NotifyObservers();
	}


	void Vision::sendIndicators(int *inds, size_t sz) {
		//sSetIndication->ClearBuffer(obsSetIndication); //obsSetIndication is the observer index of the observer reveiving a msg from this transmission.. number is generated automatically by stub.cfg
		//probably dont want to clear the buffer for indicators
		//i dont think it will make much of a difference though
		//sSetIndication->ClearBuffer(); //obsSetIndication is the observer index of the observer reveiving a msg from this transmission.. number is generated automatically by stub.cfg
		sSetIndication->SetData(inds, sz);
		sSetIndication->NotifyObservers();
	}


