/*

   Copyright 2003 The University of New South Wales (UNSW) and National  
   ICT Australia (NICTA).

   This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
   redistribute it and/or modify it under the terms of the GNU General  
   Public License as published by the Free Software Foundation; either  
   version 2 of the License, or (at your option) any later version as  
   modified below.  As the original licensors, we add the following  
   conditions to that license:

   In paragraph 2.b), the phrase "distribute or publish" should be  
   interpreted to include entry into a competition, and hence the source  
   of any derived work entered into a competition must be made available  
   to all parties involved in that competition under the terms of this  
   license.

   In addition, if the authors of a derived work publish any conference  
   proceedings, journal articles or other academic papers describing that  
   derived work, then appropriate academic citations to the original work  
   must be included in that publication.

   This rUNSWift source is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of  
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along  
   with this source code; if not, write to the Free Software Foundation,  
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: VisualCortex.h 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 *
 * Object identification from blobs
 * 
 **/


#ifndef VisualCortex_h_DEFINED
#define VisualCortex_h_DEFINED


#include <iostream>
#include <fstream>
#include "../share/Common.h"
#include "../share/minimalMatrix.h"
#include "DistanceCalibrator.h"

#include <cstdio>
using namespace std;

#ifdef OFFLINE
//#include <cstdio> /* Include at all times to enable compilation. */
using namespace std;
#endif
#include "Blob.h"

//class Blobber;
//class HalfBeacon;
class ONotifyEvent;
struct WMObj;


#include "GeneralisedColourDetect.h"
#include "../share/SwitchBoard.h"

#define ALT_BALL

struct PWalkInfoStruct {
	int *leading_leg;
	int *currentWalkType;
	int *currentStep;
	int *currentPG;
    double *frontShoulderHeight;
    double *bodyTilt;
	double *desiredPan;
	double *desiredTilt;
	double *desiredCrane;
    bool *isHighGain;
};

static const int VOB_COUNT = 17;

// for distance methods
static const int PIXEL_COUNT_DISTANCE = 0;
static const int PIXEL_COUNT_PROJECTION = 1;
static const int FIREBALL_PROJECTION = 2;


// active bar challenge code
//#define BAR_CHALL

// denotes infinity
// this is useful in determining when a variance
// is insignificant to be considered in computation
static const int LARGE_VAL = 10000000;

const int NUM_ELEMENT       = 3;
const int X_OFFSET          = 0;
const int Y_OFFSET          = 1;
const int TYPE_OFFSET       = 2;
const int GRID_SPACING      = 10;

static const int LINE_EDGE   = 0;
static const int BORDER_EDGE = 1;
static const int FIELD       = 2;


#ifdef ALT_BALL
static const int VOB_2BALL_COUNT = 18;  // 29/01/02
#endif

/* Use the limits from BlobLimits. */
/* Cannot get it from limits array directly because compiler will complain. */
static const int kMaxPinkBlobs   = 10; //limits[cBEACON_PINK].numBlobs;
static const int kMaxBlueBlobs   = 10; //limits[cBEACON_BLUE].numBlobs;   
static const int kMaxYellowBlobs = 15; //limits[cBEACON_YELLOW].numBlobs;
static const int kMaxGreenBlobs  = 10; //limits[cBEACON_GREEN].numBlobs;
static const int kMaxOrangeBlobs = 10; //limits[cBALL].numBlobs;
static const int numBeacons = 6;  

// Constants for variances, Tim Tam 13/03/02
//static const double BEACON_M = 0.0077388;
//static const double BEACON_M = 0.0077388;

#ifdef SHAPE_CHALL
/******************************/
// defines for Shape Challenge

typedef enum {
	Triangle,
	Rectangle,
	Square,
	TShape,
	LShape
}
ShapeEnum;

static const int SHAPE_COUNT = 5;
static const int BACKGROUND = 8;  //BACKGROUND
static const int FILL_COLOUR = 6; //cROBOT_RED
static const int BORDER_COLOUR = 5; //cROBOT_BLUE

/******************************/
#endif

static const double BEACON_C = 0.0203926;

typedef enum {

#ifdef ALT_BALL
	vobAltBall = 17,    // 29/01/02 new addition
#endif

	vobBall = 0,
	vobBlueGoal = 1,
	vobYellowGoal = 2,
	vobBlueLeftBeacon = 3,
	vobBlueRightBeacon = 4,
	vobGreenLeftBeacon = 5,
	vobGreenRightBeacon = 6,
	vobYellowLeftBeacon = 7,
	vobYellowRightBeacon = 8,
	vobRedDog = 9,
	vobRedDog2 = 10,
	vobRedDog3 = 11,
	vobRedDog4 = 12,
	vobBlueDog = 13,
	vobBlueDog2 = 14,
	vobBlueDog3 = 15,
	vobBlueDog4 = 16
}
VobEnum;

static const char *vobName[VOB_COUNT] = {
	"ball",
	"blue goal",
	"yellow goal",
	"BlueLeft Beacon",
	"BlueRight Beacon",
	"GreenLeft Beacon",
	"GreenRight Beacon",
	"YellowLeft Beacon",
	"YellowRight Beacon",
	"Red Dog",
	"Blue Dog"
};

/*********************************************************
VOBS: Coming out of VisualCortex, the fields have slightly
different meanings to when they come out of Remembrate
e.g. 'head' is offset from current head position, without
reference to the body's position [when Rememebrate is
done with them, they are relative to the body]
 *********************************************************/
class VisualObject {
	public:

        // Ted Wong's robot distance hack.
        int hack;

		int cf;                   // from 0 to 1000 (0 => not seen)
		double var;               // variance of object determined from equation.
		double angleVar;          // variance of the angle of the object
		double cx, cy;            // centroid
		double x, y, height, width;  // actual object bounding box

        // Only used by robots (NOTE: goal cx/cy/x/y/h/w is actually
        // rotated like this as well).
		double xRmin, xRmax, yRmin, yRmax;  // bounding box drawn using rotated blobs. 
		double cRx, cRy;                    // its centroid

		double radius;

		// Added by Ted Wong. These variables are unmodified version
		// from the blob algorithm. "o" stands for original.
		double ox, oy, oheight, owidth, ocx, ocy;

		double projected_x;  // projection of bounding box onto image due to occlusion, or limited view
		double projected_y;
		double projected_h;
		double projected_w;
		HalfBeacon *blob;

		/****
		 * These idea variables are in a bit of a halfway state because of the confusion between visual objects, world model
		 * objects (GPS) and CurObjects (UNSWskills).
		 ****/
		int area;        // area of blob(s)
		double imgHead;  // Image relative tilt and elevation
		double imgElev;
		double tilt;     // Internal rotation of vob (XY) [degrees]
		double head;     // Heading rotation of vob (XZ) [degrees]
		double lhead;    // Heading rotation of vob (XZ) [degrees]
		double rhead;    // Heading rotation of vob (XZ) [degrees]
		double h;
		double elev;     // Elevation rotation of vob (YZ) [degrees]

		double c2c; 
		double pinkXsum;
		double pinkYsum;
		double pinkarea;
		double closestXsum;
		double closestYsum;
		double closestarea;

        // above head and elev are robot relative
		double cam_dist;            // A raw camera distance measurement for ball tracking


		/* Distance from the camera to the centroid of the vob in cm.
		 * It is a manipulated value to give accurate distances. It is unsuitable for ball tracking
		 **/
		double cam_centroid_dist;   

		// Used in calibration process
		double dist;                // distance from the neck to edge of the vob [centimetres]
		double d;                   /* Two Dimensional Distance from to neck to 
                                       vob edge (parallel to the plane of the field) */
		double misc;                // beacon: c2c, ball: raw y co-ordinate of blob
		int pad;                    // may as well have it...

		double objectPosition[3];   //positiion in three dimensional space relative to the neck,
		double f;                   // the absolute direction the visual object is facing in world model.
		double orient;              // orientation of robots seen 0 - facing away chest, 
                                    // 180 - facing towards chest, -90 left, 90 right

		VisualObject() : var(LARGE_VAL), angleVar(LARGE_VAL)
		{
			Init();
		}               // var must be set, else division by 0 occurs
		/*
		 * void setDistance(double myX,double myY,double myZ)
		 * Sets a number of visual object attributes from the 3 dimensional coordinate space relative to the base of the neck.
		 * Attributes are: dist(3d), elev, heading (and h), d(2d)
		 **/
		void setDistance(double myX, double myY, double myZ);
		/*
		 * @return 3 dimensional distance from robot's neck.
		 **/
		inline double getRobotDistance() const {
			return dist;
		}
		/*
		 * @return 2 dimensional distance from robot's neck, on a plane parrallel to the field.
		 **/
		double getFlatDistance() const;

		/*
		 * @return 1 dimensional distance from robot's neck, on a plane parallel to the field and perpendicular to the body's
		 * chest.  (the z axis)
		 **/
		inline double getPerpDistance() const {
			return objectPosition[2];
		}
		/*
		 * @return 1 dimensional distance from robot's neck on a plane parallel to the field and parallel to the body's
		 * chest.  (the x axis)
		 **/
		inline double getLateralOffset() const {
			return objectPosition[0];
		}

		/*
		 * @return 1 dimensional distance from robot's neck on a plane perpendicular to the field plane.  (the y axis)
		 **/
		inline double getVerticalOffset() const {
			return objectPosition[1];
		}
		void Print();

		void printPlayData();

		void Init();

        /* Used in sanity check to forget recognised visual objects. */
		void reset();

};


class VisualCortex {
	public:

		VisualCortex();
		~VisualCortex() {
            shutdownEdgeDetect();
		}

		DogColourDistortion colDistort;

		void SetDogParams(unsigned long mac);

        double VisualCortex::getMinElevation(VisualObject &obj);

		bool isBallTooThinOrTooFlat(double xmax, double xmin,
		                            double ymax, double ymin);

		bool checkIfBallCoverScreen(int, int, bool);

		int ballDistMethod;
        bool lookForBeaconsAndGoals;

        /* Declare at all time so that codes can be compiled at all times. */
		ofstream *outfile;

#ifdef OFFLINE
		// ofstream *outfile; 
		void ProcessImage(uchar *yuv,
				long s0,
				long s1,
				long s2,
				double r2o);
#else // OFFLINE
		void ProcessImage(const ONotifyEvent &event,
				long s0,
				long s1,
				long s2,
				double r2o,
                struct PWalkInfoStruct *pWalkInfo);
#endif // OFFLINE

        /* Empty function for inserting break points in gdb. */
		void VisualCortex::test();

		void VisualCortex::initGoalDistanceLookup();

#ifdef ALT_BALL
		VisualObject vob[VOB_2BALL_COUNT];
#else
		VisualObject vob[VOB_COUNT];
#endif
		ofstream *blobOutput;

		// Derrick: This is for the localisation challenge, stores largest
		// pink blobs as a VisualObject.
		VisualObject localisePink[kMaxPinkBlobs];
		int numLocalisePink;
    
		uchar *yplane, *uplane, *vplane, *cplane;
		SegmentedPlane* splane;
		eSegmentedPlane* esplane;
		uchar* edgePlane;
		long cameraFrame;


        /*------------------------------*/
        /* Used by TrigInit. */
        double shoulderHeight, hipHeight;   // mm's
        double neckBaseHeight;              // mm's

        double crane, degCrane;             // +ive up, -ive down. 
        double pan,   degPan;               // +ive left, -ive right.
        double tilt,  degTilt;              // +ive up, -ive down (original,
                                            // but in TrigInit we reversed it)

	    double bodyTilt, degBodyTilt;       // +ive when back is higher. If offline, 
                                            // set to 0 because already included in
                                            // tilt. If online, get from PWalkInfo.

        double roll,  degRoll;              // The rotation of cplane,
                                            // combined effect by pan+tilt. 

		double effective_tilt, degETilt;    // For ERS7, effective tilt is the
                                            // combine effect of tilt and
                                            // crane, +ive down, -ive up.

		double sin_eroll, cos_eroll;        // values for unrotating the image
		double usin_eroll, ucos_eroll;      // values for rotating image back to normal

		double sin_tilt,  cos_tilt;         
		double sin_pan,   cos_pan;          // All rad's, to save duplicated
		double sin_crane, cos_crane;        // calls. 
        double sin_etilt, cos_etilt;

        double hzArray[4];      // Define the horizon using two points. 
        bool   hzExists;        // If a horizon line exists in cplane.
        bool   hzUpIsSky;       // When this is true, this means cplane region 
                                // "above" horizon line is the sky.

        double hz_m, hz_c;      // The gradient and offset of the horizon.

        double hz_roll;         // Should be similar to degRoll except this
                                // one is calculated from horizon. Should be
                                // more accurate because of adjusted bodyTilt.

        bool   hzAllAboveHorizon;   // When no hz is found, either all
                                    // above or all (whole screen) below horizon. 

        /*------------------------------*/


		GeneralisedColourDetect gcd;
		int goalDistanceLookup[144];                  //Lookup table relating goal height to goal distance.

		// pointer to the current walk type the dog is performing
		const int *currentWalkType;


        /* Check if elevation has been calculated correctly. */
        void debugElevation();


        /* Check if a given point in cplane is below horizon. */
        /* Format of cplane: (0,0) in top-left corner. */
        bool isBelowHorizon(int x, int y, int canGoOveryBy, bool checkEvenIfHzNotExist);

        /* Finds the intersection of a vertical or horizontal line with horizon. */
        bool horizonIntercept(int in, int dir, int *out);

		void RobotRelative(VisualObject &object);  // ANdres make this private!

		void set_mc(int);
		int get_mc();

		// prev_colour is the colour of the previous pixel in run-length encoding
		// and height and width is the height and width location of the pixel.

		// initialise all elements to zero at the begining
		void initialiseSanityArray();

		// fired the given sanity
		void fireSanity(int sanityIndex);

		// get methods for the sanity Array
		int getSanityArray(int i);

		// get methods for the sensors (used in method storeObject in Vision.cc)
		long getSensor(int i);

		int countColouredPixels(int colour, int minX, int minY, int maxX, int maxY);

		int player;


#ifdef VISION_DEBUG
		// Keep these information relative to the camera as
		// visual objects get changed in remembrate to robot relative
		// new variables added in VisualObject eliminates the need for these
		//long ballHead, ballElev, bGoalHead, bGoalElev, yGoalHead, yGoalElev;

		void changeCalibration(int type);
		void switchFile(int suff);

		// Update the calibration dynamically via wireless.
		void updateCalibration(int col, int row, int newcolour);

		// Subroutine to determine whether to process the image or not.
		void freezeImageProcessing(int i);
#endif

#ifdef FIREBALL
		/*************************************
		 * FIREBALL
		 *************************************/

		bool isOutside(int x, int y);
		void findBallEdge(int start,
				int end,
				int fixed,
				bool outside,
				bool reverse,
				bool vertical,
				int *intersect_x,
				int *intersect_y,
				int edgeThreshold);
		void equationEdge(int start,
				int end,
				double m,
				double b,
				bool reverse,
				int *intersect_x,
				int *intersect_y,
				int edgeThreshold);
		void findFireBallCentre(HalfBeacon *preball,
				int *fbX,
				int *fbY,
				double *fbHead,
				double *fbDist,
				double *fbRadius,
				double *fb_obj0,
				double *fb_obj1,
				double *fb_obj2);
		void findPointOnCircumferenceOneAndTwo(int bxmin,
				int bxmax,
				int bymin,
				int bymax,
				bool nl,
				bool nr,
				bool sl,
				bool sr,
				int *p1x,
				int *p1y,
				int *p2x,
				int *p2y);
		void findPointOnCircumferenceThree(int bxmin,
				int bxmax,
				int bymin,
				int bymax,
				bool nl,
				bool nr,
				bool sl,
				bool sr,
				int p1x,
				int p1y,
				int p2x,
				int p2y,
				int *p3x,
				int *p3y);
		void findIntersectionOfTwoChords(int p1x,
				int p1y,
				int p2x,
				int p2y,
				int p3x,
				int p3y,
				int *ix,
				int *iy,
				double *radius);
		void projectPoints(int *inPoints,
				int numPoints,
				double *outPoints,
				double projectHeight = 0, bool debug = false);
		void projectPoints(int *inPoints,
				int numPoints,
				int *outPoints,
				double projectHeight = 0, bool debug = false);

		// test
		void testFunction(int p1x,
				int p1y,
				double p13m,
				double p13b,
				double p23m,
				double p23b,
				int *ix,
				int *iy,
				double *radius);
		// debug
#endif
    public:
        static const int MAX_EDGE_PIXELS = 128;

        // Edge detection output
		int numEdgePnts;
        double calcRobotMatch(MVec3 &, MVec3 &);
        void calcEdgeCentre(MVec2 &);

        // Derrick: I made these two public so that Vision.cc can send data
        // (from either one) to RoboWirelessBase
        double globalEdgePnts[MAX_EDGE_PIXELS * 2];
        double localEdgePnts[MAX_EDGE_PIXELS * 2];

        // Edge detection inputs
        double headMovement;
        bool bodyMoving;
        void invalidateEdgeMap();

    private:
        // Edge detection prototypes
		void initEdgeDetect();
        void shutdownEdgeDetect();
		void initGradField();
		void mirrorGradient(float *); 
        double distanceSq(int, int, int, int);
		void runEdgeDetect();
        double calcPointMatch(double, double, int, double *, double *);
        void calcEdgeWeights();
        void stationaryMap();
        inline bool shouldMerge(double, double, int, int, int *);
        bool tryAddEdge(int, int, int, int *, int);
        int findFieldEdges();
        void localToGlobal(MVec3 &robot, int, double *, double *);
                
        // Edge detection constants
		static const int offsetX = 100;  // How far match function extends
		static const int offsetY = 100;  // beyond beacons
		static const int ARRAY_WIDTH = FIELD_WIDTH / 2 + 2 * offsetX + 1;
		static const int ARRAY_HEIGHT = FIELD_LENGTH / 2 + 2 * offsetY + 1;
		static const double INFINITY_GRADIENT_VALUE = 0;
		static const double groundPlane = 0.0;
        
        // Edge detection variables
		int edgeList[MAX_EDGE_PIXELS * 3];
        int edgePntTypes[MAX_EDGE_PIXELS];
        int cplaneEdgePnts[MAX_EDGE_PIXELS * 2];
        int mapPntTypes[MAX_EDGE_PIXELS];
        int mapPntConfidence[MAX_EDGE_PIXELS];
        double mapPnts[MAX_EDGE_PIXELS * 2];
        double edgeWeights[MAX_EDGE_PIXELS];
		float *matchLine;
		float *matchBorder;
        float *matchField;
        double totalLineWeight, totalFieldWeight;
        int numMapPnts;
        bool mapInvalid;

    public:        
		// This variable is deprecated... doesn't do anything
		bool facingFieldEdge;


		bool cmdAdjustDistance;   // distance calibration testing
		// PLEASE remove this after the march
		double beaconDistanceConstant;
		double beaconDistanceOffset;

		// gets rid of recognizing carpet as beacon
		int beaconSearchPerim;
		double beaconErrorThreshold;
		int filterInnerBeacon;

		double errorScale;   // scale of partially seen obj

		double BEACON_M;

		// the constant to use for variance adjustance of partially seen objects
		double varConst;

		double minGoalDim;

		void wlLoadNNMC(const unsigned char *data, int size);

	private:
		// Struct for a beacon type
		class Beacon {
			public:

				bool valid;
				double c2c;                // c2c distance of the 2 halfbeacons making up the beacon

				double pinkXsum;
				double pinkYsum;
				double pinkarea;

				double closestXsum;
				double closestYsum;
				double closestarea;

				double cx, cy;
				int    xmin, xmax, ymin, ymax;
                int    area;

				double angle;
				int halfmass;              // area of larger halfbeacon

                void Set(
                    const HalfBeacon &pink,
                    const HalfBeacon &closest
                        );
				

		Beacon()
		{
			xmin = -1;
            xmax = -1;
            ymin = -1;
            ymax = -1;
            cx = -1; 
            cy = -1;
		}             

				
		};

		int FindVisualObjects();

        // Derrick: for localisation challenge
        void FindPinkLocalisationBlobs();
        
		void FindBeacons();
		void BeaconsSanities();
        void copyBeaconsToVobs();

		void formBeaconFromPink(HalfBeacon *pink,
				int colorIndex,
				HalfBeacon **hbArray,
				int hbCount,
				HalfBeacon **closest);

        inline bool goalSurroundingCheck(
            int goalX,
            int goalY,
            int goalWidth,
            int goalHeight);


        inline void throwAwayOneBeacon(
                VisualObject *beaconYR,
                VisualObject *beaconBR); 


        inline void recalPinkBlobCentroid(
                HalfBeacon *pink, 
                HalfBeacon *other);

        inline bool beaconHasWhiteBelow(
                int beaconX,
                int beaconY,
                int oldBeaconMinX,
                int oldBeaconMaxX,
                int oldBeaocnMaxY,
                int beaconWidth,
                int beaconHeight
                );

        inline bool isBeaconPartiallySeen(
                HalfBeacon *pink, 
                HalfBeacon *other);

        inline void recalConfusedBeaconBlobCentroid(
                HalfBeacon *pink, 
                HalfBeacon *other,
                int        goodBlobColor,
                int        nonPinkBlobColor);

        inline void calBeaconEdgePoint(
                double &outX,
                double &outY,
                int cx,
                int cy,
                double mBeacon,
                bool goodBlobIsUp,
                bool goodBlobIsLeft,
                int goodBlobColour);

        inline bool isBeaconBlobMergedWithBackground(
                HalfBeacon *pink, 
                HalfBeacon *other,
                double c2c,
                bool relaxThreshold);

		void FindGoal();
		void LookForColourGoal(HalfBeacon **hbArraye,
				int hbCount,
				int beaconLeftOfGoalIndex,
				int beaconRightOfGoalIndex,
				int goalIndex);

		void GoalSanities();

		void LandmarkSanities();

		void FindBall();
		bool isBallBeaconNoise(VisualObject& ball);			
		bool normalFindBall(HalfBeacon *preball, VisualObject &ball);
		bool andresFindBall(HalfBeacon *preball, VisualObject &ball);

		bool BallSanities(VisualObject &,
				int,
				int,
				int,
				int,
				double,
				double,
				double,
				double);

		void FindRobot();
		void FindColorRobot(int colour);
		void RobotNoiseFilter();
		bool isRobotTooFarAwayFromGoal(double goalDist, double
					       robotDist);

		int VobInVob(VisualObject &b1,
				VisualObject &b2,
				int ot,
				int or_,
				int ob,
				int ol);

		void TrigInit(const long &micro_tilt,
				const long &micro_pan,
				const long &micro_crane,
                struct PWalkInfoStruct *pWalkInfo);

		void Reset();

		bool InYellowNoGreen(VisualObject &ball);

		bool NoGreenNearBy(int xmin, int xmax, int ymin, int ymax);
		bool NoColourXNearBy(int colour,
				int xmin,
				int xmax,
				int ymin,
				int ymax,
				int deltaX,
				int deltaY,
				int threshHold);

		int FindHalfBeacons(int colorIndex,
				HalfBeacon **hbArray,
				int maxHB,
				int threshold);

		double NewBallDistance();

		bool DarkBackgroundInBetween(int left, int right, int top, int bottom);
		//void FindColorRobot (int vobNum);

		void PrintBeaconInfo();

		void convertGoalBeacon();

		// used to draw lines between centroids of blobs and then returns
		// the number of 'colour's seen on the line
		inline int calDistBaseOnColor(int x1, int y1, int x2, int y2, 
                HalfBeacon *sourceBlob, uchar colour,
                bool skipBoundary = true);

        inline int unrotatePointReturnX(double x, double y); 
        inline int unrotatePointReturnY(double x, double y); 

	public:
		int numPink, numGreen, numBlue, numYellow;

        /* These are numbered from left to right, top to bottom with the 'top' 
         * of the field being the blue goal, and the 'bottom' being the yellow 
         * goal. ie #0=blue/pink, and #5=pink/yellow. */
		Beacon beacon[numBeacons];


		HalfBeacon *hbPink[kMaxPinkBlobs];
		HalfBeacon *hbBlue[kMaxBlueBlobs];
		HalfBeacon *hbYellow[kMaxYellowBlobs];
		HalfBeacon *hbGreen[kMaxGreenBlobs];
		HalfBeacon *hbBall[kMaxOrangeBlobs];

	private:
		int sanityArray[NUM_SANITY_COORDS];
		bool firedSanity[SANITY_COUNT];

		long sensor0;
		long sensor1;
		long sensor2;
		int infaredDistance;

		DistanceCalibrator *ballCalibrator;

    public:
        //Hack: for accessing red blob info from Python
		//  int CreateRunLengthInfo(int start, RunLengthInfo& info);
		Blobber *blobs;

        /* Functions/variables for ShapeChallenge removed after rev 2701. */
};

/* Other utility function used in VisualCortex.h */
double  getAng(double dx, double dy);
int     imax(int a, int b);
int     pi_abs(int a);
int     imin(int a, int b);

long    SizeToDistance(long physicalSize, long apparentSize);

/* Will be outdated soon. */
double  PointToHeading(double x);
double  PointToElevation(double y);
double  PointToHeading(double x, double y, double sin_eroll, double cos_eroll);
double  PointToElevation(double x, double y, double sin_eroll, double cos_eroll);


/* Calculate the elevation and heading of a cplane point relative to BODY. 
 * Without assuming the distance or height, it's impossible to calculate the
 * angles. */
double xToHeadingFixDist(double cpx, double dist);
double xToHeadingFixHeight(double cpx, double height);
double yToElevationFixDist(double cpy, double dist);
double yToElevationFixHeight(double cpy, double height);


/***
 * Queue class added by Ben
 */
class QueueItem {
	private:
		int item;
		QueueItem *next;

	public:
		QueueItem() {
		}
		~QueueItem() {
		}

		QueueItem(int qitem) {
			item = qitem;
			next = NULL;
		}

		// returns the item in this container
		int getItem() {
			return item;
		}

		// returns the next container
		QueueItem * getNext() {
			return next;
		}

		// set the next container
		void setNext(QueueItem *qitem) {
			next = qitem;
		}
}; // class QueueItem

class Queue {
	private:
		// this queue is implemented as linked
		QueueItem *first, *last;
		int size;

	public:

		Queue() {
			first = NULL;
			last = NULL;
			size = 0;
		}
		~Queue() {
		}

		bool isEmpty() {
			return first == NULL;
		}

		// queue the object to the end of the queue
		void enqueue(int item) {
			// make a new container for this object
			QueueItem *newItem = new QueueItem(item);
			size++;
			// append the object to the end of the queue
			if (first != NULL) {
				last->setNext(newItem);
				last = newItem;
			}
			else {
				last = newItem;
				first = last;
			}
		}

		// dequeue the object from the top of the queue return null if
		// the queue is empty
		int dequeue() {
			// dequeue the first item and reset pointers
			if (first != NULL) {
				int result = first->getItem();
				QueueItem *temp = first->getNext();

				delete first;

				first = temp;
				size--;

				return result;
			}
			else {
				return 0;
			}
		}

		int getSize() {
			return size;
		}
};  // class Queue

#endif // VisualCortex_h_DEFINED



