/*

   Copyright 2003 The University of New South Wales (UNSW) and National  
   ICT Australia (NICTA).

   This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
   redistribute it and/or modify it under the terms of the GNU General  
   Public License as published by the Free Software Foundation; either  
   version 2 of the License, or (at your option) any later version as  
   modified below.  As the original licensors, we add the following  
   conditions to that license:

   In paragraph 2.b), the phrase "distribute or publish" should be  
   interpreted to include entry into a competition, and hence the source  
   of any derived work entered into a competition must be made available  
   to all parties involved in that competition under the terms of this  
   license.

   In addition, if the authors of a derived work publish any conference  
   proceedings, journal articles or other academic papers describing that  
   derived work, then appropriate academic citations to the original work  
   must be included in that publication.

   This rUNSWift source is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of  
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
   General Public License for more details.

   You should have received a copy of the GNU General Public License along  
   with this source code; if not, write to the Free Software Foundation,  
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: VisualCortex.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
 **/

#include "VisualCortex.h"

#ifndef OFFLINE
#include <OPENR/OSubject.h>
#include <OPENR/OObserverVector.h>
#include <OPENR/OPENREvent.h>
#include <OPENR/OPENRAPI.h>
#include <OPENR/OFbkImage.h>
#include "OPENR/ODataFormats.h"
#endif	// OFFLINE
#include <sys/time.h>
#include "VisualCortex-priv.h"
#include "../share/Common.h"
#include "../share/PWalkDef.h"
#include "gps.h"
#include "Blob.h"
#include "PointProjection.h"
#ifndef OFFLINE
#include <MCOOP.h>
#endif
#include <cstdio>
#include <cstring>
#include <cfloat>

#include "../share/SwitchBoard.h"
#include "../share/BallDef.h"

#ifdef VALGRIND
#include <valgrind/memcheck.h>
#endif // VALGRIND

struct DogMacLookupType {
	unsigned long mac;
	DogColourDistortion colMunge;
};

DogMacLookupType DogMacLookup[] = {
	{0, {1.0, 0.0, 1.0, 0.0, 1.0, 0.0}},	// first entry is the default
	{0xc6244f15, {1.0561, 0.5273, 1.0802, -8.9272, 1.0697, -7.0994}},
	{0xc6245473, {0.9696, 1.9431, 1.0218, -2.7382, 0.9651, 5.5388}},
	// {0xc6245457, {1.0, 0.0, 1.0, 0.0, 1.0, 0.0}},
	{0xc624544b, {0.9138, 1.5080, 0.9591, 5.1463, 0.9482, 5.3510}},
	{0xc6245478, {1.0180, 1.1892, 1.0331, -4.2305, 1.0297, -1.5526}},
	{0xc6245481, {0.9443, 2.3447, 0.9502, 5.8307, 0.9766, 3.7230}},
    {0xc62450cd, {0.9731, -1.3547, 0.9887, 1.7451, 0.9777, 1.2692}},
	{0xc6245466, {0.9595, 1.3804, 0.9607, 4.6637, 1.0024, -0.0285}}
};

using namespace std;

static const int ROBOTS_EACH_COLOUR = 4;
static const int MAX_BLOBS_PER_DOG = 100;

/* This is from one of my data: average height when the dog is 1m away, turning 
 * on the spot is 15.17, so this constance = 100*15.17. Need recalibrate before
 * the game.
 * Note that this height is not include the paw. So remove the patch before
 * calibrating it. */

static const double ROBOT_HEIGHT_DISTANCE_CONSTANCE = 1517.0;

static const bool ignoreGreenBeaconsLate  = true;  // this filters out green beacons after we've found them.

/* This never searches for green beacons, and hence might mismatch the extra 
   pink blobs if green beacons do actually exist. Set this to true mean no 
   green beacons will be processed in LocateLandmarkPatches. .*/
static const bool ignoreGreenBeaconsEarly = false; 

static const bool sensorValueDebugMsg   = false;
static const bool beaconSanityDebugMsg  = false;
static const bool sanityBlobDebugMsg    = false;

/***************************************************************************/
/* This is the switches that turn on and off offline debugging messages in
** different functions. */

static const bool debugProcessTime          = false;
static const bool debugFindVisualObjects    = false;
static const bool debugTrigInit             = false;

static const bool debugFindBeacons          = true;
static const bool debugCopyBeaconsToVobs    = true;
static const bool debugBeaconsSanities      = true;
static const bool debugFormBeaconFromPink   = true;
static const bool debugThrowAwayOneBeacon   = true;

static const bool debugBeaconHasWhiteBelow              = true;
static const bool debugRecalPinkBlobCentroid            = false;
static const bool debugRecalConfusedBeaconBlobCentroid  = false;
static const bool debugIsBeaconBlobMergedWithBackground = false;
static const bool debugCalBeaconEdgePoint               = false;
static const bool debugIsBeaconPartiallySeen            = false;

static const bool debugFindGoal             = true;
static const bool debugLookForColourGoal    = true;
static const bool debugGoalSanities         = true;
static const bool debugLandmarkSanities     = true;
static const bool debugGoalSurroundingCheck = true;

static const bool debugFindRobot            = true;
static const bool debugFindColorRobot       = true;
static const bool debugRobotNoiseFilter     = true;

static const bool debugNormalFindBall       = true;
static const bool debugFindBall             = true;
static const bool debugBallSanities         = true;
static const bool debugFindFireBallCentre   = true;

static const bool debugIsBelowHorizon       = false;
static const bool debugRobotRelative        = false;
static const bool debugRunEdgeDetect        = false;

/***************************************************************************/

static const bool calibrateProjectPoints    = false;
static const bool calibrateBeaconDist       = false;

static const double zoidalVarFactor     = 40.0;
static const double zoidalAngVarFactor  = 1.5;

static double lastd = 0;

static const int xOrigin = WIDTH / 2;
static const int yOrigin = HEIGHT / 2;

/*****************************************************************************/
/***********                                                        **********/
/******                VISUAL OBJECT FUNCTIONS                          ******/


/* Used by different sanity checks to forget some possible visual objects. */
void VisualObject::reset() {
    cf          = 0;
    var         = LARGE_VAL;
    angleVar    = LARGE_VAL;
    hack        = LARGE_VAL;
}

void VisualObject::setDistance(double myX, double myY, double myZ) {

    objectPosition[0] = myX;
    objectPosition[1] = myY;
    objectPosition[2] = myZ;

    dist = sqrt(SQUARE(myX) + SQUARE(myY) + SQUARE(myZ));
    
    d = sqrt(SQUARE(myX) + SQUARE(myZ));
    h = head = RAD2DEG(atan2(myX, myZ));

    double temp1 = sqrt(SQUARE(myX) + SQUARE(myZ));
    double temp2 = dist;
    double temp3 = acos(temp1 / temp2);

    if (myY >= 0) {
        elev = RAD2DEG(temp3);
    }
    else {
        elev = -RAD2DEG(temp3);
    }
}



double VisualObject::getFlatDistance() const {
    return sqrt(SQUARE(objectPosition[2]) + SQUARE(objectPosition[0]));
}

void VisualObject::Print() {
    if (!cf) {
        cout << "\tVOB is invalid" << endl;
    }
    else {
        this->printPlayData();
        cout << "\tBBox:       (" << x << "," << y << ")-(" << x + width << ","
            << y + height << ")" << endl;
        cout << "\tBBox H:     " << height << endl;
        cout << "\tBBox W:     " << width << endl;

        cout << "\tProjBox:    (" << projected_x << "," << projected_y << ")-("
            << projected_x + projected_w << "," << projected_y + projected_h
            << ")" << endl;
        cout << "\tProjBox H:  " << projected_h << endl;
        cout << "\tProjBox W:  " << projected_w << endl;
    }
}

void VisualObject::printPlayData() {
    if (cf) {
        cout << "\tConfidence: " << cf << endl;
        cout << "\tCentroid:   (" << cx << "," << cy << ")" << endl;
        cout << "\tImage Head: " << imgHead << endl;
        cout << "\tImage Elev: " << imgElev << endl;
        cout << "\tArea:       " << area << endl;
        cout << "\tHeading:    " << head << endl;
        cout << "\tElevation:  " << elev << endl;
        cout << "\tDistance:   " << dist << endl;
    }
}

void VisualObject::Init() {

    cf = 0;
    var = LARGE_VAL;

    cx = 0;
    cy = 0;
    x = y = height = width = 0;

    cRx = cRy = 0;
    xRmin = xRmax = yRmin = yRmax = 0;

    projected_h = 0;
    projected_w = 0;
    area = 0;
    imgHead = 0;
    imgElev = 0;
    tilt = 0;
    head = 0;
    elev = 0;
    dist = 0;
    misc = 0;
    pad = 0;
    lhead = 0;
    rhead = 0;
    hack = 0;
}
/******                VISUAL OBJECT FUNCTIONS                          ******/
/***********                                                        **********/
/*****************************************************************************/



VisualCortex::VisualCortex() {
	ballCalibrator = new DistanceCalibrator();
	blobs = new Blobber(gcd);
	cplane = gcd.cplane;
	splane = &blobs->splane;
	esplane = &blobs->esplane;
	edgePlane = &blobs->edgePlane[0][0];
	cmdAdjustDistance = 1;

	beaconDistanceConstant = 2187.23; // Calibrated using one beacon. 
	beaconDistanceOffset   = 4.887;	  // Calibrated using one beacon.
    /*    
	beaconDistanceConstant = 2221.02  // Calibrated using one beacon. 
	beaconDistanceOffset   = 2.862;	  // Calibrated using one beacon. 
    */

	beaconSearchPerim = 2;
	beaconErrorThreshold = 0.6;    // % of green field seen before discarding a blob.
	filterInnerBeacon = 3;
	BEACON_M = 0.01; //0.0077388;
	errorScale = 2;
	varConst = 80;   //50;
	minGoalDim = 10;

	currentWalkType = NULL;
    lookForBeaconsAndGoals = true;

	initEdgeDetect();
	initGoalDistanceLookup();

	OFFLINE_CODE(cout << "VisualCortex.cc: OFFLINE turned on" << endl;);
	ONLINE_CODE(cout << "VisualCortex.cc: OFFLINE turned off" << endl;);

	ballDistMethod = PIXEL_COUNT_DISTANCE;

#ifdef PRE_SANITY_BLOB
	cout << "VisualCortex.cc: PRE_SANITY_BLOB turned on" << endl;
#else
	cout << "VisualCortex.cc: PRE_SANITY_BLOB turned off" << endl;
#endif
}

void VisualCortex::SetDogParams(unsigned long mac) {
	int numFields = sizeof(DogMacLookup)/sizeof(DogMacLookupType);
	bool foundDog = false;
	for (int i=0; i<numFields; i++) {
		if (DogMacLookup[i].mac == mac) {
			colDistort = DogMacLookup[i].colMunge;
			cout << "Found Dog Colour Distortion: " << colDistort << endl;
			foundDog = true;
			break;
		}
	}
	if (!foundDog) {
		colDistort = DogMacLookup[0].colMunge;
		cout << "No color distortion table found for dog.  Using first in table: " << colDistort << endl;
	}

	blobs->initDCD(colDistort);
}

/* Empty function for inserting break points in gdb. */
void VisualCortex::test() {}

void VisualCortex::wlLoadNNMC(const unsigned char *data, int size) {
	gcd.wlLoadNNMC(data, size);
}

// initialise all elements to zero at the begining
void VisualCortex::initialiseSanityArray() {
	for (int i = 0; i < NUM_SANITY_COORDS; i++) {
		sanityArray[i] = 0;
	}
	for (int i = 0; i < SANITY_COUNT; i++) {
		firedSanity[i] = false;
	}
}

// fired the given sanity
void VisualCortex::fireSanity(int sanityIndex) {
	if (!firedSanity[sanityIndex]) {
		sanityArray[sanityIndex / INT_BITS] += 1 << (sanityIndex % INT_BITS);
		firedSanity[sanityIndex] = true;
	}
}

// get methods for the sanity Array
int VisualCortex::getSanityArray(int i) {
	return sanityArray[i];
}

// Get methods for the sensors (used in method storeObject in Vision.cc)
// pan, tilt, crane all in micro radians.
long VisualCortex::getSensor(int i) {
	switch (i) {
		case 0: return sensor0 - RAD2MICRO(bodyTilt); // tilt
		case 1: return sensor1; // pan
		case 2: return sensor2; // crane
	}
	return -1;
}

/*
 *  Resets the variables for a new image feed.
 **/
void VisualCortex::Reset() {
	// reset all visual objects
#ifdef ALT_BALL
	for (int x = 0; x < VOB_2BALL_COUNT; x++) {
		vob[x].Init();
	}
#else
	for (int x = 0; x < VOB_COUNT; x++) {
		vob[x].Init();
	}
#endif
	// set all the beacons as invalid
	for (int k = 0; k < numBeacons; k++) {
		beacon[k].valid = 0;
	}
	//reset all blobs and halfBeacons
	blobs->reset();

	// reset the sanity that fired previous
	initialiseSanityArray();
}



/* Init variables relating to pan, tilt and crane, fundamental to
** projectPoints and sanity checks. */
void VisualCortex::TrigInit(const long &micro_tilt, const long &micro_pan, const long &micro_crane, struct PWalkInfoStruct *pWalkInfo) {

    static const string debugHead = "TrigInit| ";
	static const bool useSensorsForPanTiltCrane = true;

	if (useSensorsForPanTiltCrane) {
		pan      = MICRO2RAD(micro_pan);
		crane    = MICRO2RAD(micro_crane);
		tilt     = -MICRO2RAD(micro_tilt);  // Negative because we want to take
											// tilting downward as positive. 
	} else {
		// used for testing - makes sure that we don't have errors caused by the physical robot
		pan     = DEG2RAD(*pWalkInfo->desiredPan);
		crane   = DEG2RAD(*pWalkInfo->desiredCrane);
		tilt    = -DEG2RAD(*pWalkInfo->desiredTilt);
	}

    /* If it is offline, set bodyTilt to 0 because it should have been included in
    ** tilt. */
    if (pWalkInfo == NULL) {

        // Some logs contain bodyTilt, some don't, depending on codes of
        // robot in times of logging.
        static const bool LOG_CONTAIN_REAL_BODY_TILT = true;

        shoulderHeight  = 91;   // Fine-tuned from horizon when standing still.
        hipHeight       = 123;

        if (LOG_CONTAIN_REAL_BODY_TILT) {

            /* bodyTilt should now be included in tilt in cplane logs. */
	        bodyTilt        = 0;
	        degBodyTilt     = 0;

        } else {

            bodyTilt        = asin((hipHeight - shoulderHeight) / BODY_LENGTH);
            degBodyTilt     = RAD2DEG(bodyTilt);

        }

    /* Otherwise get it from pWalkInfo. */
    } else {
        bodyTilt       = *(pWalkInfo->bodyTilt); // originally radian
        shoulderHeight = *(pWalkInfo->frontShoulderHeight);
        hipHeight      = shoulderHeight + sin(bodyTilt) * BODY_LENGTH;
        degBodyTilt    = RAD2DEG(bodyTilt); // then change to degree
    }
   

    effective_tilt = tilt - crane;      // tilt +ive val is up, but crane +ive
                                        // val is down.

	sin_etilt = sin(effective_tilt);
	cos_etilt = cos(effective_tilt);

	sin_tilt = sin(tilt);
	cos_tilt = cos(tilt);

	sin_pan = sin(pan);
	cos_pan = cos(pan);

    /* Hack: RobotRelative function crane to be in same direction as tilt, so,
    ** here this is. Right now I know sin_crane is only used by RobotRelative.*/
	sin_crane = sin(-crane);
	cos_crane = cos(-crane);

    /* Roll is a combine effect of pan and tilt. Crane doesn't affect much. */
	roll = asin(sin_pan * sin(tilt + bodyTilt));

	if (tilt > M_PI / 2) {
		if (roll < 0) {
			roll = -M_PI - roll;
		}
		else {
			roll = M_PI - roll;
		}
	}

	sin_eroll   = sin(roll);
	cos_eroll   = cos(roll);

	usin_eroll  = sin(-roll); // To unrotate things.
	ucos_eroll  = cos(-roll);

    degPan      = RAD2DEG(pan);
    degTilt     = RAD2DEG(tilt);
    degCrane    = RAD2DEG(crane);
    degRoll     = RAD2DEG(roll);
    degETilt    = RAD2DEG(effective_tilt);

    neckBaseHeight = getNeckBaseHeightFromStance(shoulderHeight, hipHeight);

    if (debugTrigInit && flag_OFFLINE) {
        (*outfile) << debugHead
            << "degRoll: " << degRoll << endl;
    }

    /* Get the horizon definitions for current pan, tilt and crane. */
    /* It return true if a horizon line can be found, and hzArray will store
    ** the xy coordinates of the 2 points, hzUpIsSky tell be true if pixels
    ** above the hz line is the real up.
    ** If no hz is found, hzAllAboveHorizon means that the whole screen is
    ** above the horizon (everything should be filtered).
     */

    /* Initialise the horizon array to be empty. */
    for (int i = 0; i < 3; i++) {
        hzArray[i] = 0;
    }
    hzExists = getHorizon(hzArray, &hzUpIsSky, &hzAllAboveHorizon,

                            // TODO: Put getHorizon in ViusalCortex. 

                            degPan,
                            degTilt + degBodyTilt,
                            degCrane,

                            0,

                            76.9,
                            80.0,
                            
                            // neckBaseHeight - BALL_RADIUS * 10.0,
                            0, // Calculate with ref to base of neck now. 
                            
                            APERTURE_DISTANCE,
                            CAMERA_OFFSET,
                            false);
    if (hzExists) { 
        /* By experiment, it is shown that the observed horizon is not
        ** accurate - pan of 90 deg is 10 pixels higher than pan of 0 deg.
        ** And when pan is 0, the real horizon is about 10 pix higher. */
        double hzAdjustment = (90 - ABS(degPan)) / 9.00;

        hzArray[1] = hzArray[1] + hzAdjustment; // Change the y-values of
        hzArray[3] = hzArray[3] + hzAdjustment; // horizon.
    }

    /* hzArray: x1, y1, x2, y2. */
    /* m = (y2-y1)/(x2-x1) */
    /* c = m * x1 + y1 (2 point form line-eq) */
    double  x1 = hzArray[0],
            y1 = hzArray[1],
            x2 = hzArray[2],
            y2 = hzArray[3];

    hz_m = (y2-y1)/(x2-x1);
    hz_c = -(hz_m * x1) + y1; 

    hz_roll = RAD2DEG(atan(hz_m));

    if (debugTrigInit && flag_OFFLINE) {
        (*outfile) 
            << debugHead << "hz_roll: " << hz_roll << endl
            << debugHead << "degPan  : " << degPan << endl
            << debugHead << "degTilt : " << degTilt << endl
            << debugHead << "degCrane: " << degCrane << endl
            << endl;
    }

    /* Printing the horizon in offvision to be displayed in OfflineVision. */
    if (flag_OFFLINE) {
        if (hzExists) {
            (*outfile) 
                << "h " 
                << hzArray[0] << " "
                << hzArray[1] << " "
                << hzArray[2] << " "
                << hzArray[3] << " "
                << hzUpIsSky << " "
                << endl;
        }
    }
}



/*----------------------------------------------------*/
/* Check if a given point in cplane is below horizon. */
/* Format of x,y: (0,0) in top-left corner.

 * The forth input, checkEvenIfHzNotExist, distinguish beacon checking case
   from others. For beacon checking, even the dog is looking up so high and
   there is no horizon, it doesn't mean the beacon is not there. So beacon
   horizon checks only get called if horizon exists in cplane.

   For other objects, goal/ball/robots, if dog is looking up so high and no
   horizon is on cplane, then we should return true if horizon is below the
   cplane. */

bool VisualCortex::isBelowHorizon(
        int x, int y, 
        int  canGoOverBy, 
        bool checkEvenIfHzNotExist = true) {

    static const string debugHead = "isBelowHorizon| ";
   
    /* Sometimes no horizon is observable in cplane. */
    if (checkEvenIfHzNotExist && !(hzExists)) {

        /* Then it depends whether dog is looking up or down. */
        return !hzAllAboveHorizon;
    }

    /* Now there is horizon, have to construct the line params to check.*/
    double inX = TOPLEFT2CENTER_X(x);
    double inY = TOPLEFT2CENTER_Y(y);


    bool isBelowLine;

    if (ABS(hz_m) <= 1) {

        double hzY_from_inX = (hz_m * inX) + hz_c;

        /* Sometimes the horizon is a bit low, so we want some threshold for
        ** go-over. */
        if ((hzY_from_inX - inY) > -canGoOverBy) {
            isBelowLine = true;
        } else {
            isBelowLine = false;
        }

        /* If up is not sky, then invert the result. */
        return (hzUpIsSky ? isBelowLine : !isBelowLine);

    } else {

        /* When gradient is more than 1, we would like to compare the horizontal
        ** offset instead of vertical offset. */
        double hzX_from_inY = (inY - hz_c) / hz_m;

        /* There are four cases in total. Sky could be up or down, m could be
        ** > or < 0. Better understand it with pictures.*/
        if ((hzUpIsSky && hz_m < 0)
            || (!hzUpIsSky && hz_m >= 0)) {

            if ((hzX_from_inY - inX) > -canGoOverBy) {
                return true;
            } else {
                return false;
            }

        } else {

            if ((inX - hzX_from_inY) > -canGoOverBy) {
                return true;
            } else {
                return false;
            }
        }
    }
    return false;
}

/*----------------------------------------------------*/
/* Given an x or y c-plane coordinate this calculates the other coordinate, */
/* where it intersects the horizon.                                         */
/* Inputs: in is the x or y coordinate of the intercept                     */
/*         dir = 0 specifies 'in' is a x value, dir = 1 means 'in' is y     */
/* Outputs: out is the other coordinate of the intercept with the horizon   */
/*          returns true if result is a valid c-plane coordinate, false if  */
/*          it is offscreen.                                                */

bool VisualCortex::horizonIntercept(int in, int dir, int *out) {
    int max[2] = {CPLANE_WIDTH, CPLANE_HEIGHT};
    double p;
    
    if (hzExists) {
        if (dir == 0) {
            p = TOPLEFT2CENTER_X(in);
            *out = (int)(CENTER2TOPLEFT_Y(hz_m * p + hz_c) + 0.5);
        } else {
            if (hz_m == 0.0) return false;
        
            p = TOPLEFT2CENTER_Y(in);
            *out = (int)(CENTER2TOPLEFT_X((p - hz_c) / hz_m) + 0.5);
        }

        return ((*out) >= 0 && (*out) < max[1 - dir]);
    }
    
    return false;
}


void VisualCortex::initGoalDistanceLookup() {

	//initializes the int goalDistanceLookup[144] table. 
	//Between goal heights of zero to 80 pixels (infinite to 50 cm)
	//the expression for distance d relative to height h is:
	//
	//d = 6,878.598/(h^1.125)
	//
	//Between goal heights of 80 to 144 pixels (50 to 0 cm) 
	//the expression for distance d relative to height h is:
	//
	//d = 144-1.28*h
	//
	//This is precomputed to reduce processing time for each frame. 
	goalDistanceLookup[0] = (int) LARGE_VAL;
	for (int h = 1; h < 144; h++)
		goalDistanceLookup[h] = (int)(5818.6 / (h - 0.2054));

/*	for (int h = 85; h < 144; h ++)
		goalDistanceLookup[h] = (int) (144.0 - 1.28 * (double) h * 0.9);*/
}

#ifdef OFFLINE
void VisualCortex::ProcessImage(uchar *yuv, long s0, long s1, long s2, double r2o) {
    bodyMoving = false;
    headMovement = 0.0;

#ifdef VALGRIND
	// invalidates the memory holding all the vobs for valgrind
	VALGRIND_MAKE_WRITABLE(vob, VOB_2BALL_COUNT * sizeof(VisualObject));
#endif	// VALGRIND

#else // OFFLINE

void VisualCortex::ProcessImage(const ONotifyEvent &event, long s0, long s1, long s2, double r2o, struct PWalkInfoStruct *pWalkInfo) {

#endif // OFFLINE

    sensor0 = s0;
    sensor1 = s1;
    sensor2 = s2;
    infaredDistance = (int) r2o;    // stands for range to obstacle

    Reset();

#ifdef OFFLINE
    TrigInit(s0, s1, s2, NULL);
#else //OFFLINE
    TrigInit(s0, s1, s2, pWalkInfo);
#endif //OFFLINE

    if (flag_OFFLINE) {
        if (!flag_PRE_SANITY_BLOB) {
            //(*outfile) << "ProcessImage_o| effective crane " << RAD2DEG(roll) << endl;
        } 

        #ifdef OFFLINE // Must use ifdef. 
            yplane = yuv;
        #endif // OFFLINE

    } else {

        #ifndef OFFLINE // Must use ifdef.
                    
            // Save only the latest image.
            OFbkImageVectorData *imageVec = reinterpret_cast <OFbkImageVectorData*>
                (event.RCData(event.NumOfData() - 1)->Base());

	    OFbkImageInfo *imageInfo = imageVec->GetInfo(0);

            cameraFrame = imageInfo->frameNumber;

	    if ((((int)imageInfo->width) != CPLANE_WIDTH) ||
			    (((int)imageInfo->height) != CPLANE_HEIGHT)) {
		    cout << "Bad Image Size!!" << endl;
	    }

	    if (imageInfo->type != odataFBK_YCrCb_HPF) {
		    cout << "Bad Image Data Type!!" << endl;
	    }
	    
            // pointer to start of actual image data
            // medium resolution index=1
            // high resolution index=0
            yplane = imageVec->GetData(RES);
        
        #endif // OFFLINE

        /*static bool firstTime = true;
        if (firstTime) {
            BMP bmp;
            cout << "Saving BMP " << endl;
            bmp.SaveYCrCb2RGB((const char *)"/MS/OPEN-R/MW/DATA/P/TEST.BMP", imageVec, ofbkimageLAYER_H);
        }
        
        firstTime = false;*/
    }

    uplane = yplane + WIDTH;
    vplane = uplane + WIDTH;

    double t1, t2;
    timeval start, end;

    if (flag_OFFLINE && debugProcessTime) {
        gettimeofday(&start, NULL);
        t1 = start.tv_sec + (start.tv_usec);	
    }

    blobs->replaceCplane(yplane, uplane, vplane, SKIP);
    // blobs->replaceCplane(imageVec->GetInfo(0), yplane);

    if (flag_OFFLINE && debugProcessTime) {
        gettimeofday(&end, NULL);
        t2 = end.tv_sec + (end.tv_usec);
        cout << "Time took by colour detection improvement: " << t2 - t1 << endl;

        gettimeofday(&start, NULL);
        t1 = start.tv_sec + (start.tv_usec);
    }

    blobs->formBlobs();

    if (flag_OFFLINE && debugProcessTime) {
        gettimeofday(&end, NULL);
        t2 = end.tv_sec + (end.tv_usec);
        cout << "Time took by 2003 blob algorithm: " << t2 - t1 << endl;
    }

    if (flag_OFFLINE && debugProcessTime) {
        gettimeofday(&start, NULL);
        t1 = start.tv_sec + (start.tv_usec);
    }

    FindVisualObjects();

    if (flag_OFFLINE && debugProcessTime) {
        gettimeofday(&end, NULL);
        t2 = end.tv_sec + (end.tv_usec);
        cout << "Time took by the entire object recognition and sanity: " << t2 - t1 << endl;
    }

    if (flag_EDGE_DETECT) {

        if (flag_OFFLINE && debugProcessTime) {
            gettimeofday(&start, NULL);
            t1 = start.tv_sec + (start.tv_usec);
        }

        bool lookForEdges = true;

#ifndef OFFLINE
        // Only run edge detection when not doing a kick (point projection
        // can be bad due to large roll values.)
        if (*(pWalkInfo->currentWalkType) > AlanWalkWT)
            lookForEdges = false;       
#endif //OFFLINE

        if (lookForEdges)
            runEdgeDetect();

        if (flag_OFFLINE && debugProcessTime) {
            gettimeofday(&end, NULL);
            t2 = end.tv_sec + (end.tv_usec);
            cout << "Time took by the edge detection algorithm: " << t2 - t1 << endl;
        }
	}

    //debugElevation();
}


/*-----------------------------------------*/
/* Evaluating the RobotRelative function. */

void VisualCortex::debugElevation() {

    //if (flag_OFFLINE) {
    if (((cameraFrame % 60) == 0) && vob[vobBlueRightBeacon].cf > 0) {

        /* HACK: ELEV TESTING. */
        shoulderHeight  = 91;   // Fine-tuned from horizon when standing still.
        hipHeight       = 123;
        degBodyTilt       = RAD2DEG(asin((hipHeight - shoulderHeight) / BODY_LENGTH));
 
        static const string debugHead = "CalibratingElevation| ";
        VisualObject obj, obj2;
        double x    = vob[vobBlueRightBeacon].x + vob[vobBlueRightBeacon].width/2;
        double y    = vob[vobBlueRightBeacon].y + vob[vobBlueRightBeacon].height;
        double dist = vob[vobBlueRightBeacon].dist;// 112.9; // cm, just for testing. 

        obj2.imgElev = obj.imgElev = PointToElevation(y);
        obj2.imgHead = obj.imgHead = PointToHeading(x);
        obj2.dist    = obj.dist    = dist;

        //RobotRelativeERS7(obj);
        RobotRelative(obj2);

        //(*outfile) 
        cout  
            << debugHead << "pan     : " << degPan << endl
            << debugHead << "tilt    : " << degTilt << endl
            << debugHead << "crane   : " << degCrane << endl
            << debugHead << "cplane x: " << x << endl
            << debugHead << "cplane y: " << y << endl
            << debugHead << "est dist: " << dist << endl
            << debugHead << "imgHead:  " << obj.imgHead << endl
            << debugHead << "imgElev:  " << obj.imgElev << endl
            << debugHead << "------- NEW FOR ERS7" << endl
            << debugHead << "obj.h  :  " << obj.h << endl
            << debugHead << "obj.elev: " << obj.elev << endl
            << debugHead << "obj.d:    " << obj.d << endl
            << debugHead << "------- OLD FOR ERS210" << endl
            << debugHead << "obj.h  :  " << obj2.h << endl
            << debugHead << "obj.elev: " << obj2.elev << endl
            << debugHead << "obj.d:    " << obj2.d << endl
            << endl;
    }
}



/******************************************************************************
 * Find beacons: first objects to recognise.
 */
void VisualCortex::FindBeacons() {

    static const string debugHead = "FindBeacons| ";
	int checkedBeacons[6] = {
		0, 0, 0, 0, 0, 0
	};

    numGreen  = ignoreGreenBeaconsEarly ? 0 : blobs->findHalfBeacons(cBEACON_GREEN, hbGreen, kMaxGreenBlobs);
    numPink   = blobs->findHalfBeacons(cBEACON_PINK, hbPink, kMaxPinkBlobs);
	numBlue   = blobs->findHalfBeacons(cBEACON_BLUE, hbBlue, kMaxBlueBlobs);
	numYellow = blobs->findHalfBeacons(cBEACON_YELLOW, hbYellow, kMaxYellowBlobs);


    if (flag_OFFLINE && debugFindBeacons) {
        (*outfile) 
            << debugHead << "numPink    : " << numPink << endl
            << debugHead << "numBlue    : " << numBlue << endl
            << debugHead << "numYellow  : " << numYellow << endl
            << endl;
    }
    
 

    //------------------------------------
	/* Out of the largest 6 pink blobs found, check them one by one. */
	for (int p = 0; p < numPink; p++) {

        int         closestType = -1;
        HalfBeacon  *closest    = NULL;

        HalfBeacon  pinkOriginal = *hbPink[p];
        HalfBeacon  pinkAfterMatchBlue;

        if (flag_OFFLINE && debugFindBeacons) {
            (*outfile) << debugHead
                << "..................Pink at ("
                << hbPink[p]->cx
                << ", "
                << hbPink[p]->cy
                << "), corners ("
                << hbPink[p]->getXMin() << ", "
                << hbPink[p]->getYMin() << ", "
                << hbPink[p]->getXMax() << ", "
                << hbPink[p]->getYMax() << ")......."
                << endl;
        }
		
        //------------------------------------
        /* Find the other (yellow/blue) half beacon that matches the pink half beacon. */ 
		HalfBeacon *blueClosest     = NULL;
		HalfBeacon *yellowClosest   = NULL;
		HalfBeacon *greenClosest    = NULL;

        formBeaconFromPink(hbPink[p], cBEACON_BLUE, hbBlue, numBlue, &blueClosest);

        /* Back up pink blob after any changes from matching blue beacons. */
        /* Use original pink blob for matching yellow beacons. */
        pinkAfterMatchBlue  = *hbPink[p];
        *hbPink[p]          = pinkOriginal;

        formBeaconFromPink(hbPink[p], cBEACON_YELLOW, hbYellow, numYellow, &yellowClosest);

        /* In the real game play (with green beacons physically removed), we
         * would set ignoreGreenBeaconsEarly to true). Right now still check
         * green beacon early because we don't want the pink on green beacon
         * to be messed up with other "real" beacons.  */
        if (!ignoreGreenBeaconsEarly) {
            formBeaconFromPink(hbPink[p],cBEACON_GREEN, hbGreen, numGreen, &greenClosest);
        }


        /* CHECK ------------------------------------------------------------*/

        /* One pink blob can connect to at most 1 other colour to form a blob.
         * When there are more than 1 colour connected to pink, one must be 
         * false and we need to decide. */

        if ((blueClosest != NULL) && (yellowClosest != NULL)) {

            /* It seems that area ratio is not reliable... compare elevation
            ** instead. */

            //#define pick_using_area_ratio
            #ifdef pick_using_area_ratio
            double blueBeaconRatio = (hbPink[p]->area < blueClosest->area) 
                                   ? ((double) hbPink[p]->area   / (double) blueClosest->area) 
                                   : ((double) blueClosest->area / (double) hbPink[p]->area); 

            double yellowBeaconRatio = (hbPink[p]->area < yellowClosest->area) 
                                     ? ((double) hbPink[p]->area     / (double) yellowClosest->area) 
                                     : ((double) yellowClosest->area / (double) hbPink[p]->area); 

            /* Whichever ratio is bigger, meaning pink-other ratio closer
            ** to 1, is the better one and the other should be filtered.
            ** */
            if (blueBeaconRatio > yellowBeaconRatio) {
                closest     = blueClosest;
                closestType = cBEACON_BLUE;

            } else {
                closest     = yellowClosest;
                closestType = cBEACON_YELLOW;
            }

            if (flag_OFFLINE && debugFindBeacons) {
                (*outfile) << debugHead
                    << "...2 beacons recognised from 1 pink blob, choosing "
                    << ColorName[closestType]
                    << endl;
                (*outfile) << debugHead
                    << "   Areas: pink- " << hbPink[p]->area << ", "
                    << " blue- " << blueClosest->area << ", "
                    << " yellow- " << yellowClosest->area 
                    << endl;
            }
            #else 

            double yellowElev = PointToElevation(yellowClosest->cx, yellowClosest->cy,
                                                    sin_eroll, cos_eroll); 
            double blueElev = PointToElevation(blueClosest->cx, blueClosest->cy,
                                                    sin_eroll, cos_eroll); 

            /* Pick the one with lower elevation. */
            if (blueElev < yellowElev) {
                closest     = blueClosest;
                closestType = cBEACON_BLUE;

            } else {
                closest     = yellowClosest;
                closestType = cBEACON_YELLOW;
            }

            if (flag_OFFLINE && debugFindBeacons) {
                (*outfile) << debugHead
                    << "...2 beacons recognised from 1 pink blob, choosing "
                    << ColorName[closestType]
                    << endl;
                (*outfile) << debugHead
                    << debugHead << "    blueElev   - " << blueElev << endl
                    << debugHead << "    yellowElev - " << yellowElev << endl
                    << endl;
            }

            #endif // pick_using_area_ratio
            

        } else if ((blueClosest == NULL) && (yellowClosest != NULL)) {
            closest     = yellowClosest;
            closestType = cBEACON_YELLOW;

        } else if ((blueClosest != NULL) && (yellowClosest == NULL)) {
            closest     = blueClosest;
            closestType = cBEACON_BLUE;

            /* Re-assign the pink blob if choose blue after all. */
            *hbPink[p]  = pinkAfterMatchBlue;

        } else if (greenClosest != NULL) {
            closest     = greenClosest;
            closestType = cBEACON_GREEN;

        } else { 
			fireSanity(CANNOT_MATCH_PINK_BLOB);
			continue;
        }
            
        //------------------------------------
        /* Figure out which beacon we are dealing with, defaulting to
         * having pink on the bottom (ie the left side of the field). */
        closest->matched = true;
        int beaconID = -1;
        switch (closestType) {
            case cBEACON_BLUE:   beaconID = 0;
                         break;
            case cBEACON_GREEN:  beaconID = 2;
                         break;
            case cBEACON_YELLOW: beaconID = 4;
                         break;
        }

        /* Compare elevation after rotating images with amount of roll due to 
         * combined effect of pan and tilt and bodyTilt. */
        double closest_elev = PointToElevation(closest->cx, closest->cy, sin_eroll, cos_eroll);
        double pink_elev    = PointToElevation(hbPink[p]->cx, hbPink[p]->cy, sin_eroll, cos_eroll);

		// Pink is on top, the beacon is on the right hand side of the field.
		if (closest_elev < pink_elev) {
			beaconID++;
		}

        #ifdef want_left_right_beacon_debug
        if (debugFindBeacons && flag_OFFLINE) {
            (*outfile) 
                << debugHead << "Is it left or right beacon? " << endl
                << debugHead << "   Pink  cx, cy: " << hbPink[p]->cx << ", "
                                                    << hbPink[p]->cy << endl
                << debugHead << "   Other cx, cy: " << closest->cx << ", "
                                                    << closest->cy << endl
                << debugHead << "   Pink  elev: " << pink_elev << endl
                << debugHead << "   Other elev: " << closest_elev << endl
                << endl;
        }
        #endif // want_left_right_beacon_debug


        //------------------------------------
        /* Ignore the recognised 2 middle beacons if (any). */
        if (ignoreGreenBeaconsLate && (beaconID == 2 || beaconID == 3)) {

            if (debugFindBeacons && flag_OFFLINE) {
                (*outfile) << debugHead
                    << "...filter out middle beacons!"
                    << endl;
            }
            continue;
        }

        //------------------------------------
		/* Set this beacon as 'seen' so that we can tell if any beacon appeared
        ** twice - one must be false. */
		if (!checkedBeacons[beaconID]) {

			checkedBeacons[beaconID] = 1;
			beacon[beaconID].Set(*hbPink[p], *closest);

            if (flag_OFFLINE && debugFindBeacons) {
                (*outfile) << debugHead
                    << "...beacon #" << beaconID  
                    << " set!" << endl;
                (*outfile) << debugHead << "   Beacon c2c   : " << beacon[beaconID].c2c << endl;
                (*outfile) << debugHead << "   Beacon cx, cy: " << beacon[beaconID].cx << ", " 
                           << beacon[beaconID].cy << endl;

                //#define detail_beacon_info_after_each_pink_match
                #ifdef detail_beacon_info_after_each_pink_match
                (*outfile) << debugHead << "   Beacon xmin/ax, ymin/ax: " 
                    << beacon[beaconID].xmin << ", "
                    << beacon[beaconID].xmax << ", "
                    << beacon[beaconID].ymin << ", "
                    << beacon[beaconID].ymax << endl;
                (*outfile) << debugHead << "   Pink   xmin/ax, ymin/ax: " 
                    << hbPink[p]->getXMin() << ", "
                    << hbPink[p]->getXMax() << ", "
                    << hbPink[p]->getYMin() << ", "
                    << hbPink[p]->getYMax() << endl;
                (*outfile) << debugHead << "   Other  xmin/ax, ymin/ax: " 
                    << closest->getXMin() << ", "
                    << closest->getXMax() << ", "
                    << closest->getYMin() << ", "
                    << closest->getYMax() << endl;

                #endif // detail_beacon_info_after_each_pink_match
            }

        /* This beacon has been set before, which one is correct? */
		} else {

            if (flag_OFFLINE && debugFindBeacons) {
                (*outfile) << debugHead
                    << "...SAME BEACON (id:" << beaconID  
                    << ") FOUND!" << endl;
            }

            /* TODO: THIS MEANS WE'RE SEEING ANOTHER BEACON WE HAVE SEEN, SHOULD 
             * DECIDE WHICH BEACON TO THROW AWAY. */
			continue;
		}
	}

   	copyBeaconsToVobs(); // Copy beacon information from beacon[] into vob[].
}



/******************************************************************************
 * Given ONE pink blob and array of yellow or blue blobs, check if a beacon
 * can be formed by that pink blob. 
 * Outputs:
 *  - HalfBeacon **closest: Pointer pointing to other blob if a beacon is
 *                          found. NULL if no beacon is found.
 *  - int *closestType: What colour index the other blob is.
 */
void VisualCortex::formBeaconFromPink(
        HalfBeacon  *pink,
        int         nonPinkColor, 
        HalfBeacon  **hbArray,
        int         hbCount, 
        HalfBeacon  **closest) {

    static const string debugHead = "formBeaconFromPink| ";

    if (flag_OFFLINE && debugFormBeaconFromPink) {
        (*outfile) << debugHead << endl;
        (*outfile) 
            << debugHead << "Checking against " << ColorName[nonPinkColor] 
                << ", numBlob: " << hbCount << endl 
            << endl;
    }

    /*-----------------------------------*/
    /* Sanity checks on pink blobs only. */ 
    /*-----------------------------------*/
     

    // SANITY CHECK 1 ----------------------------------------------------------

    /* If the dog is looking really down, there should not be any beacon. */
    if (!hzExists && !hzAllAboveHorizon) {

        if (debugFormBeaconFromPink && flag_OFFLINE) {
            (*outfile) << debugHead 
                << "... looking straight down, no beacon will be observed!"
                << endl;
        }
        return;
    }
                            // --------------------------- END OF SANITY CHECK 


    // SANITY CHECK 2 ----------------------------------------------------------

    // Check if the pink blob is too high above horizon (80 pixels buffer).
    if (hzExists && !isBelowHorizon((int)pink->cx, (int)pink->cy, 80, false)) {

        fireSanity(BEACON_PINK_TOO_HIGH_TO_HORIZON);

        if (debugFormBeaconFromPink && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead
                << "...beacon pink blob "
                << "is too high above the horizon." 
                << endl;
        }
        return;
    }
                            // --------------------------- END OF SANITY CHECK 


    // SANITY CHECK 3 ----------------------------------------------------------

    // Check if the pink blob is too low below horizon (80 pixels buffer).
    if (hzExists && isBelowHorizon((int)pink->cx, (int)pink->cy, -80, false)) {

        fireSanity(BEACON_PINK_TOO_LOW_TO_HORIZON);

        if (debugFormBeaconFromPink && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead
                << "...beacon pink blob "
                << "is too low below the horizon." 
                << endl;
        }
        return;
    }
                            // --------------------------- END OF SANITY CHECK 


    // SANITY CHECK 4 ----------------------------------------------------------

    // Check the bounding box size of the pink blob.
    static const int BOUNDING_BOX_MIN_PINK = 3;
    int     pinkWidth   = pink->getXMax() - pink->getXMin() +1;
    int     pinkHeight  = pink->getYMax() - pink->getYMin() +1;

    if (   (pinkWidth < BOUNDING_BOX_MIN_PINK)  
        || (pinkHeight < BOUNDING_BOX_MIN_PINK)) {

        if (debugFormBeaconFromPink && flag_OFFLINE) {
            (*outfile) << debugHead 
                << "...pink blob is too small (height or width < "
                << BOUNDING_BOX_MIN_PINK << " pixels."
                << endl;
        }
        return;
    }
                            // --------------------------- END OF SANITY CHECK 


    /* Robot noise checks - checks on ratios of pink blob. */

    double minPinkSolidRatioThreshold  = 0.4;
    double maxPinkAspectRatioThreshold = 6;

    // More strict for small blobs not in edge.  
    bool pinkInEdge = pink->getXMax() > CPLANE_WIDTH - 5
                    || pink->getXMin() < 5
                    || pink->getYMax() > CPLANE_HEIGHT - 5
                    || pink->getYMin() < 5;

    if (!pinkInEdge) { 

        maxPinkAspectRatioThreshold = 3;

        // Even more strict if the area is small. 
        if (pink->area < 20) { 
            minPinkSolidRatioThreshold  = 0.5;
            maxPinkAspectRatioThreshold = 2.5;
        }
    }

    double pinkSolidRatio = (double) pink->area 
                            / (double) (pinkWidth * pinkHeight);

    double pinkAspectRatio = (pinkHeight > pinkWidth)
                        ? (double) pinkHeight / (double) pinkWidth
                        : (double) pinkWidth  / (double) pinkHeight;

    if (debugFormBeaconFromPink && flag_OFFLINE) {
        (*outfile) 
            << debugHead << "    --pink->area : " << pink->area << endl
            << debugHead << "    --pinkbox    : " << pinkWidth * pinkHeight << endl
            << endl;
    }


    // SANITY CHECK 5 ------------------------------------------------------

    // Check if pink solid ratio is ok.
    if (pinkSolidRatio < minPinkSolidRatioThreshold) {

        if (debugFormBeaconFromPink && flag_OFFLINE) {
            (*outfile) << debugHead 
                << "...pink blob's solid ratio is too low!"
                << pinkSolidRatio << ")"
                << endl;
            (*outfile)
                << debugHead << "    --pinkSolidRatio  : " << pinkSolidRatio  << endl
                << debugHead << "    --minPinkSolidRatioThreshold  : " << minPinkSolidRatioThreshold  << endl
                << endl;
        }
        return;
    }
                            // --------------------------- END OF SANITY CHECK 



    // SANITY CHECK 6 ------------------------------------------------------

    // Check if pink bounding box aspect ratio is ok.
    if (pinkAspectRatio > maxPinkAspectRatioThreshold) {

        if (debugFormBeaconFromPink && flag_OFFLINE) {
            (*outfile) << debugHead 
                << "...pink blob's aspect ratio is too large ("
                << pinkAspectRatio << ")" 
                << endl;
            (*outfile)
                << debugHead << "    --pinkAspectRatio : " << pinkAspectRatio << endl
                << debugHead << "    --maxPinkAspectRatioThreshold : " << maxPinkAspectRatioThreshold << endl
                << endl;
        }
        return;
    }
                            // --------------------------- END OF SANITY CHECK 



    /*---------------------------------------------------------------------*/
    /* Sanity checks on matching of the pink blob and the array of blobs   */ 
    /* with the other colour.                                              */
    /*---------------------------------------------------------------------*/

    /* Go through each of the other colour blob (yellow or blue). */
    for (int i = 0; i < hbCount; i++) {

        HalfBeacon *other = hbArray[i];

        if (debugFormBeaconFromPink && flag_OFFLINE) {
            (*outfile) << debugHead
                << "-----------  BLOB #" << other->blob_number << ": "
                << "(" << other->getXMin() << "," << other->getXMax() 
                << "," << other->getYMin() << "," << other->getYMax() 
                << ") [xmin/max, ymin/max]  (i = " << i 
                << endl;
        }

        // SANITY CHECK 7 ------------------------------------------------------

        // This blob is already formed as another beacon. 
        if (other->matched) {
            if (debugFormBeaconFromPink && flag_OFFLINE) {
                (*outfile) << debugHead
                    << "...blob already matched!" << endl << endl;
            }
            continue;
        }
                            // --------------------------- END OF SANITY CHECK 


        // SANITY CHECK 8 ------------------------------------------------------

        // If the area is zero. 
        if (other->area == 0 || pink->area == 0) {
            if (debugFormBeaconFromPink && flag_OFFLINE) {
                (*outfile) << debugHead
                    << "...Area is zero!" << endl << endl;
            }
            fireSanity(ZERO_HALF_BEACON_AREA_OR_PINK_AREA);
            continue;
        }
                            // --------------------------- END OF SANITY CHECK 



        // SANITY CHECK 9 ------------------------------------------------------

        // Check the bounding box size of the other blob.
        static const int BOUNDING_BOX_MIN_OTHER = 3;
        int     otherWidth   = other->getXMax() - other->getXMin() +1;
        int     otherHeight  = other->getYMax() - other->getYMin() +1;

        if (   otherWidth < BOUNDING_BOX_MIN_OTHER 
            || otherHeight < BOUNDING_BOX_MIN_OTHER) {

            fireSanity(BEACON_NON_PINK_BLOB_TOO_SMALL);
            if (debugFormBeaconFromPink && flag_OFFLINE) {
                (*outfile) << debugHead  
                    << "...the other blob is too small (height or width < "
                    << BOUNDING_BOX_MIN_OTHER << " pixels."
                    << endl << endl;
            }
            continue;
        }
                            // --------------------------- END OF SANITY CHECK 



        /* Robot noise checks - checks on ratios of the other blob. */

        double minOtherSolidRatioThreshold  = 0.4;
        double maxOtherAspectRatioThreshold = 6;

        bool otherInEdge = other->getXMax() > CPLANE_WIDTH - 5
                        || other->getXMin() < 5
                        || other->getYMax() > CPLANE_HEIGHT - 5
                        || other->getYMin() < 5;

        // More strict if other blob is not in edge (so it won't be partial beacons).
        if (!otherInEdge) {
            maxOtherAspectRatioThreshold = 3;

            // Even more strict if the area is small. 
            if (other->area < 20) { 
                minOtherSolidRatioThreshold  = 0.5;
                maxOtherAspectRatioThreshold = 2.5;
            }
        }

        double otherSolidRatio = (double) other->area 
                                / (double) (otherWidth * otherHeight);

        double otherAspectRatio = (otherHeight > otherWidth)
                            ? (double) otherHeight / (double) otherWidth
                            : (double) otherWidth  / (double) otherHeight;

        if (debugFormBeaconFromPink && flag_OFFLINE) {
            (*outfile) 
                << debugHead << "    -- other->area : " << other->area << endl
                << debugHead << "    -- otherBox    : " << otherWidth * otherHeight << endl 
                << endl;
        }


        // SANITY CHECK 10 ------------------------------------------------------

        // Check the solid ratio of the other blob.
        if (otherSolidRatio < minOtherSolidRatioThreshold) {

            if (debugFormBeaconFromPink && flag_OFFLINE) {
                (*outfile) 
                    << debugHead << "    -- otherSolidRatio              : " << otherSolidRatio << endl
                    << debugHead << "    -- minOtherSolidRatioThreshold  : " << minOtherSolidRatioThreshold << endl
                    << endl;
                (*outfile) 
                    << debugHead << "...other blob's solid ratio not right!"
                    << endl << endl;
            }
            continue;
        }
                            // --------------------------- END OF SANITY CHECK 



        // SANITY CHECK 11 ------------------------------------------------------

        // Check the aspect ratio of the other blob.
        if (otherAspectRatio > maxOtherAspectRatioThreshold) {

            fireSanity(BEACON_NON_PINK_BLOB_SMALL_N_LARGE_ASPECT_RATIO);
            if (debugFormBeaconFromPink && flag_OFFLINE) {
                (*outfile) 
                    << debugHead << "    -- otherAspectRatio             : " << otherAspectRatio << endl
                    << debugHead << "    -- maxOtherAspectRatioThreshold : " << maxOtherAspectRatioThreshold << endl
                    << endl;
                (*outfile) << debugHead  
                    << "...other blob is small and not in edge, and also VERY LARGE ASPECT RATIO!"
                    << endl << endl;
            }
            continue;
        }
                            // --------------------------- END OF SANITY CHECK 



        // SANITY CHECK 12 ------------------------------------------------------

        // Check if two blobs are too far apart to be beacon.
        // Calculate distance between centers of pink and the other color. 
        double c2cDist = sqrt(SQUARE(pink->cx - other->cx) + SQUARE(pink->cy - other->cy));

        double pinkRadius  = calDistBaseOnColor(
                            (int) (pink->cx), (int) (pink->cy), 
                            (int) (other->cx), (int) (other->cy), 
                            pink, cBEACON_PINK, false);

        double otherRadius = calDistBaseOnColor(
                            (int) (other->cx),(int) (other->cy), 
                            (int) (pink->cx),  (int) (pink->cy), 
                            other, nonPinkColor, false);

        double factor = 0.7;

        /* Small blob tends to have tiny radius from calDistBaseOnColor, so
        ** relax the threshold a bit. */
        if (pink->area < 20 || other->area < 20) {
            factor = 0.6;
        }

        #ifdef enable_only_for_detail_c2c_debugging
        if (debugFormBeaconFromPink && flag_OFFLINE) { 

            (*outfile) << debugHead  
                << "Details of C2C check..." << endl;
            (*outfile) << debugHead 
                << " pinkRadius: " << pinkRadius << endl 
                << debugHead << " otherRadius: " << otherRadius << endl
                << debugHead << " sum: " << pinkRadius + otherRadius << endl
                << debugHead << " c2cDist: " << c2cDist << endl
                << endl;
            (*outfile) << debugHead 
                << "   pink (xmin/max ,ymin/max): " 
                << pink->getXMin() << " " 
                << pink->getXMax() << " " 
                << pink->getYMin() << " " 
                << pink->getYMax() 
                << endl;
            (*outfile) << debugHead 
                << "   other(xmin/max ,ymin/max): " 
                << other->getXMin() << " " 
                << other->getXMax() << " " 
                << other->getYMin() << " " 
                << other->getYMax() 
                << endl << endl;
        }
        #endif // enable_only_for_detail_c2c_debugging

        if ((c2cDist * factor) > (pinkRadius + otherRadius)) { 

            fireSanity(PINK_TOO_FAR_APART_FROM_COLOUR_BLOB);

            if (debugFormBeaconFromPink && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead  
                    << "...blob too far apart (pink, color)."
                    << endl << endl;
            }
            continue;
        }
                            // --------------------------- END OF SANITY CHECK 


        // SANITY CHECK 13 ------------------------------------------------------

        // Check if the center-center distance exceeds a maximum. 
        if (c2cDist > 100) { 

            fireSanity(BEACON_TOO_LONG);

            if (debugFormBeaconFromPink && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile)  << debugHead 
                    << "...beacon is too long."
                    << endl << endl;
            }
            continue;
        }
                            // --------------------------- END OF SANITY CHECK 


        // SANITY CHECK 14 ------------------------------------------------------

        // Extra harsh on blue solid ratio. This is tailored for having 
        // middle beacons, where cBEACON_GREEN often appears as cBEACON_BLUE.

        // #define see_if_other_checks_can_handle_middle_beacon
        #ifdef see_if_other_checks_can_handle_middle_beacon
        /*
        if (nonPinkColor == cBEACON_BLUE) {

            int horiDist = other->getXMax() - other->getXMin() + 1;
            int vertDist = other->getYMax() - other->getYMin() + 1;
            int boxArea  = (horiDist == 0 || vertDist == 0) ? 1 : horiDist * vertDist;

            if ((double) ((double) other->area / (double) boxArea) < 0.6) { 

                if (debugFormBeaconFromPink && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                    (*outfile) << debugHead  
                        << "...blue is too scattered!"
                        << endl << endl;
                }
                continue;
            }
        }
        */
        #endif // see_if_other_checks_can_handle_middle_beacon
                            // --------------------------- END OF SANITY CHECK 


        // SANITY CHECK 15 ------------------------------------------------------

        // Check if the beacon is vertical to the horizon.
        HalfBeacon *higherBlob = (pink->cy < other->cy) ? pink : other;
        HalfBeacon *lowerBlob  = (pink->cy < other->cy) ? other : pink;

        double dx = lowerBlob->cx - higherBlob->cx;
        double dy = lowerBlob->cy - higherBlob->cy;

        // Invert sign so that / is +ive and \ -ive.
        double beaconToHorizontal = -RAD2DEG(atan2(dy, dx)); 

        // Make things reference to 0 at EAST. 
        if (beaconToHorizontal < 0)
            beaconToHorizontal += 180; 

        double rollToHorizontal = degRoll;

        if (rollToHorizontal < 0)
            rollToHorizontal += 180; 

        double angleToHorizon = abs(rollToHorizontal - beaconToHorizontal);

        if (angleToHorizon < 60.0 || angleToHorizon > 120.0) { 

            fireSanity(BEACON_NOT_VERTICAL_TO_HORIZON);

            if (debugFormBeaconFromPink && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead 
                    << "...beacon is not vertical to the horizon. "
                    << endl;
                (*outfile) 
                    << debugHead << "   pink  cx, cy: " 
                        << pink->cx << ", " << pink->cy << endl
                    << debugHead << "   other cx, cy: " 
                        << other->cx << ", " << other->cy << endl
                    << debugHead << "   dy, dx: " << dy << ", " << dx << endl
                    << debugHead << "   beaconAngle: " << beaconToHorizontal << endl
                    << debugHead << "   rollAngle: " << rollToHorizontal << endl
                    << debugHead << "   angleToHorizon: " << angleToHorizon 
                    << endl << endl;
            }
            continue;
        }
                            // --------------------------- END OF SANITY CHECK 


        // SANITY CHECK 16 ------------------------------------------------------

        // Can't have negative elevation.
        double beaconCentroidX = (pink->cx + other->cx) / 2;
        double beaconCentroidY = (pink->cy + other->cy) / 2;

        VisualObject dummyObjForElevCheck;
        dummyObjForElevCheck.imgHead = PointToHeading(beaconCentroidX);
        dummyObjForElevCheck.imgElev = PointToElevation(beaconCentroidY);
        dummyObjForElevCheck.dist    = 100; // Just want to get a rough estimation.

        RobotRelative(dummyObjForElevCheck);

        if (dummyObjForElevCheck.elev < -5.0) {

            fireSanity(BEACON_HAS_NEGATIVE_ELEVATION);
            if (debugFormBeaconFromPink && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead  
                    << "...beacon elev is too low ("
                    << dummyObjForElevCheck.elev << ")"
                    << endl << endl;
            }
            continue;
        }
                            // --------------------------- END OF SANITY CHECK 


        // SANITY CHECK 17 ------------------------------------------------------

        // Check if mid-base of beacon is too high above horizon.
        int midBaseX, midBaseY;

        if (pink->getYMax() > other->getYMax()) {
            midBaseX = (pink->getXMin() + pink->getXMax()) /2;
            midBaseY = pink->getYMax();

        } else {
            midBaseX = (other->getXMin() + other->getXMax()) /2;
            midBaseY = other->getYMax();
        }
        
        if (hzExists && !isBelowHorizon(midBaseX, midBaseY, 60, false)) {

            fireSanity(BEACON_TOO_HIGH_TO_HORIZON);

            if (debugFormBeaconFromPink && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead  
                    << "...beacon is too high above the horizon." 
                    << endl << endl;
            }
            continue;
        }
                            // --------------------------- END OF SANITY CHECK 


        // SANITY CHECK 18 ------------------------------------------------------

        // Check if mid-base of beacon is too low below horizon.
        if (hzExists && isBelowHorizon(midBaseX, midBaseY, -60, false)) {

            fireSanity(BEACON_TOO_LOW_TO_HORIZON);

            if (debugFormBeaconFromPink && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile)  << debugHead 
                    << "...beacon is too low below the horizon." 
                    << endl << endl;
            }
            continue;
        }
                            // --------------------------- END OF SANITY CHECK 


        /* Values of xmin, xmax, ymax before adjusting centroid for
        ** beaconHasWhiteBelow to use. */
        int oldBeaconMaxX = MAX(pink->getXMax(), other->getXMax());
        int oldBeaconMinX = MIN(pink->getXMin(), other->getXMin());
        int oldBeaconMaxY = MAX(pink->getYMax(), other->getYMax());
        

        /* Found that red/orange noise move the pink centroid too much, so
        ** let disable it for now. */
        #ifdef found_centroid_readjustment_moving_centroid_too_much

        //---------------------------------------
        /* Since pink blob often appears as orange and red in edges, try to 
           recalculate its centroid to accept some number of those noises as
           part of the pink blob. CAN BE DANGEROUS. */
        if (debugFormBeaconFromPink && flag_OFFLINE) {
            (*outfile) << debugHead  
                << "Doing centroid readjustment for pink..."
                << endl;
        }

        recalPinkBlobCentroid(pink, other);
        #endif // found_centroid_readjustment_moving_centroid_too_much



        /********************************************************************/
        /* Now, before checking the distance and elevation makes sense
        ** together, check if centroid of blobs need re-adjustment. */
        /********************************************************************/

        bool isPartialBeacon = false;

        /* In case this is not a beacon at the end, we would like to
        ** reverse the change. */ 
        HalfBeacon backupBlob;
        HalfBeacon *whichBlobWasBackedUp = NULL;

        //---------------------------------------
		/* Check if blob's centroid is wrong because it is only partially
        ** seen. */ 
        if (isBeaconPartiallySeen(pink, other)) {

            isPartialBeacon = true;

            if (debugFormBeaconFromPink && flag_OFFLINE) {
                (*outfile) << debugHead
                    << "...PARTIAL BEACON! Reset one of the blob centroid."
                    << endl << endl;
            }

            HalfBeacon *smallerBlob;
            HalfBeacon *biggerBlob;

            if (pink->area < other->area) {
                smallerBlob = pink;
                biggerBlob  = other;
            } else {
                smallerBlob = other;
                biggerBlob  = pink;
            }
            
            double smallBoxHeight   = smallerBlob->getYMax() - smallerBlob->getYMin();
            double bigBoxHeight     = biggerBlob->getYMax() - biggerBlob->getYMin();
            double heightChange     = (bigBoxHeight - smallBoxHeight) / 2.0;

            /* Compare elevation after rotation. */
            double smallerElev = PointToElevation(smallerBlob->cx, smallerBlob->cy,
                                                    sin_eroll, cos_eroll); 

            double biggerElev  = PointToElevation(biggerBlob->cx, biggerBlob->cy,
                                                    sin_eroll, cos_eroll); 

            backupBlob           = *smallerBlob;
            whichBlobWasBackedUp = smallerBlob;

            if (biggerElev < smallerElev) {
                smallerBlob->setYMin((int) (smallerBlob->getYMax() - bigBoxHeight));
                smallerBlob->cy = smallerBlob->cy - heightChange;

            } else {
                smallerBlob->setYMax((int) (smallerBlob->getYMin() + bigBoxHeight));
                smallerBlob->cy = smallerBlob->cy - heightChange;
            }


        } else { 

            // Only do the check if it is not partial beacon - partial beacon
            // will have in-proportional blobs.


            // SANITY CHECK 19 ------------------------------------------------------

            /* Check if the relationships between the pink blob and blob of other
            ** colour makes sense. */ 
            if (isBeaconBlobMergedWithBackground(pink, other, c2cDist, true)) {

                fireSanity(LAZY_SANITY_MESSAGE1);

                if (debugFormBeaconFromPink && flag_OFFLINE) {
                    (*outfile)  << debugHead 
                        << "...relationship between blobs in beacon is too insane!"
                        << endl << endl;
                }
                continue;
            }
                                // --------------------------- END OF SANITY CHECK 
        }

        /*
        #ifdef enable_only_if_dist_estimate_is_accurate
        else if (isBeaconBlobMergedWithBackground(pink, other, c2cDist, false)) {

            //---------------------------------------
            // Check if blob's centroid is wrong because it is merged with
            // background. 

            // Assume that the smaller blob has a correct shape. 
            HalfBeacon *goodBlob = (pink->area < other->area) ? pink: other;
            HalfBeacon *badBlob  = (pink->area < other->area) ? other: pink;
            int goodBlobColor    = (goodBlob == pink) ? cBEACON_PINK : nonPinkColor;

            backupBlob           = *badBlob;
            whichBlobWasBackedUp = badBlob;

            recalConfusedBeaconBlobCentroid(goodBlob, badBlob, goodBlobColor, nonPinkColor);

            if (debugFormBeaconFromPink && flag_OFFLINE) {
                (*outfile) << debugHead
                    << "... blob merged with background. Centroid re-calculated."
                    << endl;
                (*outfile) << debugHead
                    << "    pink centroid : ("
                    << pink->cx << "," << pink->cy << ")"
                    << endl;
                (*outfile) << debugHead
                    << "    pink xmin/max ymin/max : "
                    << pink->getXMin() << "," << pink->getXMax() << ","
                    << pink->getYMin() << "," << pink->getYMax() << ""
                    << endl;
                (*outfile) << debugHead
                    << "    other centroid: ("
                    << other->cx << "," << other->cy << ")"
                    << endl;
                (*outfile) << debugHead
                    << "    other xmin/max ymin/max : "
                    << other->getXMin() << "," << other->getXMax() << ","
                    << other->getYMin() << "," << other->getYMax() << ""
                    << endl << endl;
            }
        }
        #endif // enable_only_if_dist_estimate_is_accurate
        */


        // SANITY CHECK 20 ------------------------------------------------------

        // Check aspect ratio of bounding box of beacon after centroid is
        // recalculated. ONLY CHECK IF centroid has not been adjusted. 
        // Because if it has, that means the background is disturbing the blob
        // and the bounding box probably won't be in good ratio.
        if (!isPartialBeacon
            && (   ABS(degRoll) < 30  // only if beacon is upright
                || ABS(degRoll) > 60) // or lying horizontally.
           ) {

            int horiDist = max(pink->getXMax(), other->getXMax()) 
                         - min(pink->getXMin(), other->getXMin()) + 1;

            int vertDist = max(pink->getYMax(), other->getYMax()) 
                         - min(pink->getYMin(), other->getYMin()) + 1;

            /* Take roll into account - want the larger value on top. */
            double standingBeaconAspectRatio = (abs(degRoll) < 45) 
                        ? (double) vertDist / (double) horiDist
                        : (double) horiDist / (double) vertDist;

            if (standingBeaconAspectRatio < 0.8) {

                /* Reverse the change if blob has been changed. */
                if (whichBlobWasBackedUp != NULL) {
                    *whichBlobWasBackedUp = backupBlob;
                }

                fireSanity(BEACON_BAD_BOUNDING_BOX_ASPECT_RATIO);

                if (debugFormBeaconFromPink && flag_OFFLINE) {
                    (*outfile) << debugHead 
                        << "...formed beacon bounding box aspect ratio is insane!" << endl;
                    (*outfile) << debugHead << "   horiDist: " << horiDist << endl;
                    (*outfile) << debugHead << "   vertDist: " << vertDist << endl;
                    (*outfile) << debugHead << "   ratio   : " 
                        << standingBeaconAspectRatio << endl << endl;
                }
                continue;
            }
        }
                            // --------------------------- END OF SANITY CHECK 


        // SANITY CHECK 21 ------------------------------------------------------

        // Check if there are enough white pixels below the beacon.
        int beaconBoxX      = MIN(pink->getXMin(), other->getXMin());
        int beaconBoxY      = MIN(pink->getYMin(), other->getYMin());
        int beaconBoxWidth  = MAX(pink->getXMax(), other->getXMax()) - beaconBoxX + 1;
        int beaconBoxHeight = MAX(pink->getYMax(), other->getYMax()) - beaconBoxY + 1;

        if (!beaconHasWhiteBelow(beaconBoxX, beaconBoxY,
                                 oldBeaconMinX, oldBeaconMaxX, oldBeaconMaxY,
                                 beaconBoxWidth, beaconBoxHeight)) {

            /* Reverse the change if blob has been changed. */
            if (whichBlobWasBackedUp != NULL) {
                *whichBlobWasBackedUp = backupBlob;
            }

            fireSanity(BEACON_NOT_ENOUGH_WHITE_BELOW);
            if (debugFormBeaconFromPink && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead 
                    << "...beacon does not have enough white below. "
                    << endl << endl;
            }
            continue;
        }
                            // --------------------------- END OF SANITY CHECK 


        // SANITY CHECK 22 ------------------------------------------------------

        /* Check if the elevation and distance make sense together. */

        // New height check using elevation angle from camera. 
        // Very relaxed thershold because of bodyTilt error.


        static const double MIN_BEACON_HEIGHT_THRESHOLD = 50;  // mm
        static const double MAX_BEACON_HEIGHT_THRESHOLD = 500; // mm

        double c2c        = sqrt(SQUARE(pink->cx - other->cx) + SQUARE(pink->cy - other->cy));

        // Because error in bodyTilt can contribute error in height a lot, 
        // we don't want big distance.
		double beaconDist = MIN(250, beaconDistanceConstant / c2c); // cm 

        beaconCentroidX = (pink->cx + other->cx) / 2;
        beaconCentroidY = (pink->cy + other->cy) / 2;

        double beaconHeight = neckBaseHeight
                                    + getHeightKnowingDistance(degPan, degTilt + degBodyTilt, degCrane,
                                    TOPLEFT2CENTER_X(beaconCentroidX),
                                    TOPLEFT2CENTER_Y(beaconCentroidY),
                                    beaconDist * 10,    // mm 
                                    FACE_LENGTH * 10 ,  // mm
                                    NECK_LENGTH * 10,   // mm
                                    APERTURE_DISTANCE,  // mm
                                    CAMERA_OFFSET);
        double beaconElev = getElevation(degPan, degTilt + degBodyTilt, degCrane,
                                        TOPLEFT2CENTER_X(beaconCentroidX),
                                        TOPLEFT2CENTER_Y(beaconCentroidY), 
                                        FACE_LENGTH * 10, 
                                        NECK_LENGTH * 10, 
                                        APERTURE_DISTANCE, 
                                        CAMERA_OFFSET);

        if (beaconHeight > MAX_BEACON_HEIGHT_THRESHOLD) {

            /* Reverse the change if blob has been changed, since we do not to
            ** leave the effect to later checks. */
            if (whichBlobWasBackedUp != NULL) {
                *whichBlobWasBackedUp = backupBlob;
            }
            fireSanity(BEACON_BAD_HEIGHT);

            if (debugFormBeaconFromPink && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead 
                    << "...beacon's estimated height is too big." << endl
                    << debugHead << "    --degPan, degTilt, degCrane : "
                        << degPan << ", " << degTilt << ", " << degCrane << endl << endl
                    << debugHead << "    --degBodyTilt               : "
                        << degBodyTilt << endl
                    << debugHead << "    --beaconCentroidX/Y         : "
                        << beaconCentroidX << ", " << beaconCentroidY <<
                        endl << endl
                    << debugHead << "    --beaconDist       : " << beaconDist << endl
                    << debugHead << "    --neckBaseHeight   : " << neckBaseHeight << endl
                    << debugHead << "    --beaconHeight     : " << beaconHeight << endl
                    << endl;
            }
            continue;

        } else if (beaconHeight < MIN_BEACON_HEIGHT_THRESHOLD) {

            /* Reverse the change if blob has been changed, since we do not to
            ** leave the effect to later checks. */
            if (whichBlobWasBackedUp != NULL) {
                *whichBlobWasBackedUp = backupBlob;
            }
            fireSanity(BEACON_BAD_HEIGHT);

            if (debugFormBeaconFromPink && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead 
                    << "...beacon's estimated height is too small." << endl
                    << debugHead << "    --degPan, degTilt, degCrane : "
                        << degPan << ", " << degTilt << ", " << degCrane << endl << endl
                    << debugHead << "    --degBodyTilt               : "
                        << degBodyTilt << endl
                    << debugHead << "    --beaconCentroidX/Y         : "
                        << beaconCentroidX << ", " << beaconCentroidY <<
                        endl << endl
                    << debugHead << "    --beaconDist       : " << beaconDist << endl
                    << debugHead << "    --neckBaseHeight   : " << neckBaseHeight << endl
                    << debugHead << "    --beaconHeight: "    << beaconHeight << endl
                    << endl;
            }
            continue;
        }
                            // --------------------------- END OF SANITY CHECK 


        /**************************************
         * Done! This beacon should be ok!
         **************************************
         */
        pink->matched   = true;
        other->matched  = true;

        /* Check if a beacon has actually been recognised already. */
        if (*closest == NULL) {
            *closest = other;

            if (debugFormBeaconFromPink && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead 
                    << "...FOUND BEACON WITH COLOUR " 
                    << ColorName[nonPinkColor] << endl << endl;
            }

        } else {

            /* Compare the rotated elevation of two matched blobs to see which
            ** one has a smaller elevation. */
            double oldElev = PointToElevation((*closest)->cx, (*closest)->cy, sin_eroll, cos_eroll);
            double newElev = PointToElevation(other->cx, other->cy, sin_eroll, cos_eroll);

            if (newElev < oldElev) {
                *closest = other;
            }
        }


        /* Removed a big combination check in r3417 because it has been
        ** replaced by individual checks. */

        /* Removed checking of Green nearby Beacons in r3417 because we will
        ** check the white surrounding it. NoGreenNearyBy is still in file. */
	}
}


//---------------------------------------
/* Since pink blob often appears as orange and red in edges, try to 
   recalculate its centroid to accept some number of those noises as
   part of the pink blob. CAN BE DANGEROUS. */

inline void VisualCortex::recalPinkBlobCentroid(
        HalfBeacon *pink, 
        HalfBeacon *other
        ) {

    static const string debugHead = "recalPinkBlobCentroid| ";

    double edge1x, edge1y, edge2x, edge2y;

    /* m for gradient. 1 instead of -1 because hz_m has different
    ** coord system (y going up). */
    double mBeacon = 1 / hz_m;
    bool   goingDown   = (pink->cy < other->cy) ? true : false;
    bool   goingRight  = (pink->cx < other->cx) ? true : false;

    int    startX = (int) (pink->cx + 0.5); // Avoid round-ground error. 
    int    startY = (int) (pink->cy + 0.5); 

    /* Get the edge points along mBeacon for both directions. */
    calBeaconEdgePoint(edge1x, edge1y,
                       startX, startY,
                       mBeacon, goingDown, goingRight,
                       cBEACON_PINK
                      ); 

    calBeaconEdgePoint(edge2x, edge2y,
                       startX, startY,
                       mBeacon, !goingDown, !goingRight,
                       cBEACON_PINK
                      ); 

    /*
    pink->setXMin(MIN(pink->getXMin(), (int) MIN(edge1x, edge2x)));
    pink->setYMin(MIN(pink->getYMin(), (int) MIN(edge1y, edge2y)));
    pink->setXMax(MIN(pink->getXMax(), (int) MAX(edge1x, edge2x)));
    pink->setYMax(MIN(pink->getYMax(), (int) MAX(edge1y, edge2y)));
    */

    pink->cx    = (double) (edge1x + edge2x) / 2.0;
    pink->cy    = (double) (edge1y + edge2y) / 2.0;

    if (debugRecalPinkBlobCentroid && flag_OFFLINE) {
        (*outfile) << endl;
        (*outfile) << debugHead << "edge1x: " << edge1x << endl;
        (*outfile) << debugHead << "edge1y: " << edge1y << endl;
        (*outfile) << debugHead << "edge2x: " << edge2x << endl;
        (*outfile) << debugHead << "edge2y: " << edge2y << endl;

        (*outfile) << debugHead << "new centroid x: " << pink->cx << endl;
        (*outfile) << debugHead << "new centroid y: " << pink->cy << endl;

        (*outfile) << debugHead 
            << "Centroid re-calculation finished." 
            << endl;
    }
}





/*********************************************************/
/* Check if the goal has enough white colour below it. */
/* Similar to beacon has white below but simpler. */

inline bool VisualCortex::goalSurroundingCheck(
        int goalX,
        int goalY,
        int goalWidth,
        int goalHeight
    ) {

    static const string debugHead = "goalSurroundingCheck| ";

    /* The space where we think should be the white below the goal. */
    int checkSpaceMinX,
        checkSpaceMinY,
        checkSpaceMaxX,
        checkSpaceMaxY;

    /* Roll is not big so we assume goal is vertical - check space below it.
    ** */
    if (abs(degRoll) < 45) { 

        checkSpaceMinX      = goalX;
        checkSpaceMaxX      = checkSpaceMinX + goalWidth;
        checkSpaceMinY      = MIN(goalY + goalHeight, CPLANE_HEIGHT-1);
        checkSpaceMaxY      = MIN(checkSpaceMinY + MIN(goalHeight/4, 10), CPLANE_HEIGHT-1); // max 10

    /* Roll is > 45 we check the right of goal. */
    } else if (degRoll > 45) {

        checkSpaceMinY      = goalY;
        checkSpaceMaxY      = checkSpaceMinY + goalHeight;

        // Check space base on the un-modified edge if that's more right.
        checkSpaceMinX      = MIN(goalX + goalWidth, CPLANE_WIDTH-1);
        checkSpaceMaxX      = MIN(checkSpaceMinX + MIN(goalWidth/4, 10), CPLANE_WIDTH -1);

    /* Roll is < -45 so we check the left of goal. */
    } else {

        checkSpaceMinY      = goalY;
        checkSpaceMaxY      = checkSpaceMinY + goalHeight;

        checkSpaceMaxX      = MAX(goalX, 0);
        checkSpaceMinX      = MAX(checkSpaceMaxX - MIN(goalWidth/4, 10), 0);
    }
    
    if (debugGoalSurroundingCheck && flag_OFFLINE) {
        (*outfile) << endl;
        (*outfile) << debugHead << "Checking white below..." << endl;
        (*outfile) << debugHead << "    vertical? : " 
            << ((abs(degRoll) < 45) ? "true" : "false") 
            << endl;
        (*outfile) << debugHead << "    checkSpace (minX, maxX, minY, maxY): " 
            << checkSpaceMinX << ", "
            << checkSpaceMaxX << ", "
            << checkSpaceMinY << ", "
            << checkSpaceMaxY
            << endl;
    }


    int checkSpaceWidth  = checkSpaceMaxX - checkSpaceMinX;
    int checkSpaceHeight = checkSpaceMaxY - checkSpaceMinY;

    if (checkSpaceWidth <= 3 || checkSpaceHeight <= 3) {
     
        if (debugGoalSurroundingCheck && flag_OFFLINE) {
            (*outfile) << endl;
            (*outfile) << debugHead << "...no need to check - check space too small."
                << endl;
        }
        return true;
    }



    int greenCounter = 0;
    int whiteCounter = 0;
    int curColour;

    /* Go through each pixel in the checking region. */
    for (int i = checkSpaceMinX; i <= checkSpaceMaxX; i++) {
        for (int j = checkSpaceMinY; j <= checkSpaceMaxY; j++) {

            curColour = cplane[i + j * WIDTH] & ~MAYBE_BIT;

            // #define detail_colour_analysis_report
            #ifdef detail_colour_analysis_report
            if (debugGoalSurroundingCheck && flag_OFFLINE) {
                (*outfile) << debugHead
                    << "    (" << i << ", " << j << "): "
                    << curColour << " ie. " << ColorName[curColour] 
                    << endl;
            }
            #endif // detail_colour_analysis_report

            switch (curColour) {

                case cFIELD_GREEN:
                    greenCounter++;
                    break;

                case cWHITE:
                    whiteCounter++;
                    break;

                default:
                    break;
            }
        }
    }

    static const double MIN_GREEN_RATIO = 0.3; 
    static const double MAX_WHITE_RATIO = 0.8; 
   
    double area = (double) (checkSpaceMaxX - checkSpaceMinX) 
                * (double) (checkSpaceMaxY - checkSpaceMinY);
 
    double greenRatio = (double) greenCounter / area;
    double whiteRatio = (double) whiteCounter / area;

    if (debugGoalSurroundingCheck && flag_OFFLINE) {
        (*outfile) 
            << debugHead << "    GREEN ratio: " << greenRatio << endl
            << debugHead << "    WHITE ratio: " << whiteRatio << endl;
    }

    /*
    if (greenRatio < MIN_GREEN_RATIO) {

        if (debugGoalSurroundingCheck && flag_OFFLINE) {
            (*outfile) << debugHead
                << "... green ratio is too small! ( < "
                << MIN_GREEN_RATIO << ")"
                << endl;
        }
        return false;
    } 
    */

    if (whiteRatio > MAX_WHITE_RATIO) {

        if (debugGoalSurroundingCheck && flag_OFFLINE) {
            (*outfile) << debugHead
                << "... white ratio is too large! ( > "
                << MAX_WHITE_RATIO << ")"
                << endl;
        }
        return false;
    } 


    return true;
}



/*********************************************************/
/* Check if the beacon has enough white colour below it. */

inline bool VisualCortex::beaconHasWhiteBelow(
        int beaconX,
        int beaconY,

        int oldBeaconMinX,
        int oldBeaconMaxX,
        int oldBeaconMaxY,

        int beaconWidth,
        int beaconHeight
        ) {

    static const string debugHead = "beaconHasWhiteBelow| ";

    /* The space where we think should be the white below the beacon. */
    int checkSpaceMinX,
        checkSpaceMinY,
        checkSpaceMaxX,
        checkSpaceMaxY;

    int someGap; // between the beacon and the check space.

    /* Roll is not big so we assume beacon is vertical - check space below it.
    ** */
    if (abs(degRoll) < 45) { 

        someGap             = beaconHeight / 10;

        checkSpaceMinX      = beaconX;
        checkSpaceMaxX      = checkSpaceMinX + beaconWidth;

        // Check space base on the un-modified edge if that's lower.
        checkSpaceMinY      = MAX(oldBeaconMaxY, beaconY + beaconHeight) + someGap;
        checkSpaceMaxY      = checkSpaceMinY + MIN(beaconHeight/4, 10); // max 10

        // No point to check this edge.
        if (checkSpaceMaxY >= CPLANE_HEIGHT) {
            return true;
        }
 
    /* Roll is > 45 we check the right of beacon. */
    } else if (degRoll > 45) {

        someGap             = beaconWidth / 10;

        checkSpaceMinY      = beaconY;
        checkSpaceMaxY      = checkSpaceMinY + beaconHeight;

        // Check space base on the un-modified edge if that's more right.
        checkSpaceMinX      = MAX(oldBeaconMaxX, beaconX + beaconWidth) + someGap;
        checkSpaceMaxX      = checkSpaceMinX + MIN(beaconWidth/4, 10);

        // No point to check this edge.
        if (checkSpaceMaxX >= CPLANE_WIDTH) {
            return true;
        }


    /* Roll is < -45 so we check the left of beacon. */
    } else {

        someGap             = beaconWidth / 10;

        checkSpaceMinY      = beaconY;
        checkSpaceMaxY      = checkSpaceMinY + beaconHeight;

        checkSpaceMaxX      = MIN(oldBeaconMinX, beaconX) - someGap;
        checkSpaceMinX      = checkSpaceMaxX - MIN(beaconWidth/4, 10);

        // No point to check this edge.
        if (checkSpaceMinX <= 0) {
            return true;
        }
    }
    
    if (debugBeaconHasWhiteBelow && flag_OFFLINE) {
        (*outfile) << endl;
        (*outfile) << debugHead << "Checking white below..." << endl;
        (*outfile) << debugHead << "    vertical? : " 
            << ((abs(degRoll) < 45) ? "true" : "false") 
            << endl;
        (*outfile) << debugHead << "    checkSpace (minX, maxX, minY, maxY): " 
            << checkSpaceMinX << ", "
            << checkSpaceMaxX << ", "
            << checkSpaceMinY << ", "
            << checkSpaceMaxY
            << endl;
    }

    int whiteCounter = 0,
        blackCounter = 0,
        blueCounter  = 0,
        greyCounter  = 0,
        redCounter   = 0,
        greenCounter = 0;

    int curColour;

    /* Go through each pixel in the checking region. */
    for (int i = checkSpaceMinX; i <= checkSpaceMaxX; i++) {
        for (int j = checkSpaceMinY; j <= checkSpaceMaxY; j++) {

            curColour = cplane[i + j * WIDTH] & ~MAYBE_BIT;

            // #define detail_colour_analysis_report
            #ifdef detail_colour_analysis_report
            if (debugBeaconHasWhiteBelow && flag_OFFLINE) {
                (*outfile) << debugHead
                    << "    (" << i << ", " << j << "): "
                    << curColour << " ie. " << ColorName[curColour] 
                    << endl;
            }
            #endif // detail_colour_analysis_report

            switch (curColour) {

                case cROBOT_GREY:
                    greyCounter++;
                    break;

                case cWHITE:
                    whiteCounter++;
                    break;

                case cBLACK:
                    blackCounter++;
                    break;

                case cROBOT_BLUE:
                    blueCounter++;
                    break;

                case cROBOT_RED:
                    redCounter++;
                    break;

                case cFIELD_GREEN:
                    greenCounter++;
                    break;

                default:
                    break;
            }
        }
    }

    int acceptableColourCounter = whiteCounter + blackCounter + redCounter 
                                    + greyCounter + blueCounter + greenCounter;

    static const double MIN_ACCEPTABLE_COLOUR_RATIO = 0.8; 
    static const double MAX_GREEN_RATIO = 0.6; 
    static const double MAX_BLACK_RATIO = 0.6; 
   
    double area = (double) (checkSpaceMaxX - checkSpaceMinX) 
                * (double) (checkSpaceMaxY - checkSpaceMinY);
 
    double greenRatio = (double) greenCounter / area;
    double blackRatio = (double) blackCounter / area;
    double acceptableColourRatio = (double) acceptableColourCounter / area;

    if (debugBeaconHasWhiteBelow && flag_OFFLINE) {
        (*outfile) 
            << debugHead << "    GREEN ratio: " << greenRatio << endl
            << debugHead << "    BLACK ratio: " << blackRatio << endl
            << debugHead << "    ACCEPTABLE COLOUR ratio: " << acceptableColourRatio << endl;
    }

    if (greenRatio > MAX_GREEN_RATIO) {

        if (debugBeaconHasWhiteBelow && flag_OFFLINE) {
            (*outfile) << debugHead
                << "... green ratio is too large! ( > "
                << MAX_GREEN_RATIO << ")"
                << endl;
        }
        return false;

    } else if (blackRatio > MAX_BLACK_RATIO) {

        if (debugBeaconHasWhiteBelow && flag_OFFLINE) {
            (*outfile) << debugHead
                << "... black ratio is too large! ( > "
                << MAX_BLACK_RATIO << ")"
                << endl;
        }
        return false;

    } else if (acceptableColourRatio < MIN_ACCEPTABLE_COLOUR_RATIO) {

        if (debugBeaconHasWhiteBelow && flag_OFFLINE) {
            (*outfile) << debugHead
                << "... acceptable colour ratio too small! ( < "
                << MIN_ACCEPTABLE_COLOUR_RATIO << ")"
                << endl;
        }
        return false;
    } 

    return true;
}




/******************************************/
/* Check if the beacon is partially seen. */

inline bool VisualCortex::isBeaconPartiallySeen(
        HalfBeacon *pink, 
        HalfBeacon *other) {

    static const string debugHead = "isBeaconPartiallySeen| ";
    
    HalfBeacon *smallerBlob = (pink->area < other->area) ? pink : other;
    HalfBeacon *biggerBlob  = (pink->area < other->area) ? other: pink;

    double blobAreaRatio = smallerBlob->area / biggerBlob->area;

    static const int LEFT_LIMIT   = 2;
    static const int TOP_LIMIT    = 2;
    static const int RIGHT_LIMIT  = CPLANE_WIDTH  - 2;
    static const int BOTTOM_LIMIT = CPLANE_HEIGHT - 2;

    if (debugIsBeaconPartiallySeen && flag_OFFLINE) {
        (*outfile) 
            << debugHead << "Is beacon partially seen? " << endl
            << debugHead << "    degRoll             : " << degRoll << endl
            << debugHead << "    blobAreaRatio       : " << blobAreaRatio << endl
            << debugHead << "    smallerBlob ymin/max: " 
                << smallerBlob->getYMin()
                << ", " << smallerBlob->getYMax()
                << endl;
    }

    /* Condition of "seeing partial beacon". */
    if (abs(degRoll) < 45  // Ok upright
        &&  (blobAreaRatio < 0.5)                       // Area ratio not right. 
        &&  (smallerBlob->getYMin() < TOP_LIMIT 
             || smallerBlob->getYMax() > BOTTOM_LIMIT)) {
        //!(RAD2DEG(tilt) > 45 && ABS(RAD2DEG(pan)) > 60) // Ok upright. 
        return true;
    }

    return false;

}


/**********************************************************/
/* In case a beacon is merged with the background, or somewhat
** in-proportional, this function will correct the centroid of the "bad" blob. */

inline void VisualCortex::recalConfusedBeaconBlobCentroid(
        HalfBeacon *goodBlob, 
        HalfBeacon *badBlob,
        int goodBlobColor,
        int nonPinkBlobColor) {

    static const string debugHead = "recalConfusedBeaconBlobCentroid| ";

    double goodBlobEdgeX, goodBlobEdgeY;

    /* m for gradient. 1 instead of -1 because hz_m has different
    ** coord system (y going up). */
    double mBeacon = 1 / hz_m;
    bool   goodBlobIsUp   = (goodBlob->cy < badBlob->cy) ? true : false;
    bool   goodBlobIsLeft = (goodBlob->cx < badBlob->cx) ? true : false;

    /* Ensure bad blob didn't point the good blob into another direction. */

    int    goodBlobCX = (int) (goodBlob->cx + 0.5); // Avoid round-ground error. 
    int    goodBlobCY = (int) (goodBlob->cy + 0.5); // Avoid round-ground error. 

    calBeaconEdgePoint(goodBlobEdgeX, goodBlobEdgeY,
                       goodBlobCX, goodBlobCY,
                       mBeacon, goodBlobIsUp, goodBlobIsLeft,
                       goodBlobColor
                       //goodBlob->color // ha, color isn't set properly!
                       //(goodBlob == pink ? cBEACON_PINK : nonPinkBlobColor)
                      ); 

    double xDiff = abs(goodBlobEdgeX - goodBlob->cx);
    double yDiff = abs(goodBlobEdgeY - goodBlob->cy);

    /* Reset badBlob's centroid and bounding box. */
    badBlob->cx = goodBlobEdgeX + ((goodBlobEdgeX > goodBlob->cx) ? xDiff : -xDiff);
    badBlob->cy = goodBlobEdgeY + ((goodBlobEdgeY > goodBlob->cy) ? yDiff : -yDiff);

    badBlob->setXMin((int) (badBlob->cx - xDiff));
    badBlob->setXMax((int) (badBlob->cx + xDiff));
    badBlob->setYMin((int) (badBlob->cy - yDiff));
    badBlob->setYMax((int) (badBlob->cy + yDiff));

    /*

    badBlob->cx = (goodBlobEdgeX > goodBlob->cx)
                    ? (goodBlobEdgeX + (goodBlobEdgeX - goodBlob->cx))
                    : (goodBlobEdgeX - (goodBlob->cx - goodBlobEdgeX));
    badBlob->cy = (goodBlobEdgeY > goodBlob->cy)
                    ? (goodBlobEdgeY + (goodBlobEdgeY - goodBlob->cy))
                    : (goodBlobEdgeY - (goodBlob->cy - goodBlobEdgeY));

    badBlob->setXMin((int) MIN(badBlob->getXMin(), goodBlobEdgeX));
    badBlob->setYMin((int) MIN(badBlob->getYMin(), goodBlobEdgeY));
    badBlob->setXMax((int) MIN(badBlob->getXMax(), goodBlobEdgeX));
    badBlob->setYMax((int) MIN(badBlob->getYMax(), goodBlobEdgeY));
    */

    if (debugRecalConfusedBeaconBlobCentroid && flag_OFFLINE) {
        (*outfile) << endl;
        (*outfile) << debugHead << "goodBlobEdgeX: " << goodBlobEdgeX << endl;
        (*outfile) << debugHead << "goodBlobEdgeY: " << goodBlobEdgeY << endl;
        (*outfile) << debugHead << "badBlob->cx: " << badBlob->cx << endl;
        (*outfile) << debugHead << "badBlob->cy: " << badBlob->cy << endl;
        (*outfile) << debugHead << "badBlob xmin,ymin, xmax,ymax: " 
            << badBlob->getXMin() << ","
            << badBlob->getYMin() << ", "
            << badBlob->getXMax() << ","
            << badBlob->getYMax() << endl;
    }
}



/*****************************************************************************/
/* Given the centroid of a "good" beacon blob, find out the coordinates of the 
   edge of the blob, in the direction along gradient towards the "bad blob". */
inline void VisualCortex::calBeaconEdgePoint(
        double &outX,
        double &outY,
        int cx,
        int cy,
        double mBeacon,
        bool goodBlobIsUp,   // == searchDown
        bool goodBlobIsLeft, // == searchRight
        int goodBlobColour) {

    static const string debugHead = "calBeaconEdgePoint| ";

    int traceX = cx;
    int traceY = cy;

    int pixColour;
    int pinkNoiseCounter = 0;

    /* When > 0, orange/red pixels will be accepted as part of the pink blob
    ** for PINK_NOISE_LIMIT number of times. Notet that this is used by
    ** recalConfusedBeaconCentroid AND recalPinkBlobCentroid. And BOTH of the
    ** functions should use the same constant (so set the constant here is
    ** good. */
    static const int PINK_NOISE_LIMIT = 0; 

    int changeInY = (goodBlobIsUp   ?  1: -1);
    int changeInX = (goodBlobIsLeft ?  1: -1);

    double checkedX  = (double) cx;
    double checkedY  = (double) cy;

    if (debugCalBeaconEdgePoint && flag_OFFLINE) {
        (*outfile) << endl << debugHead 
            << "Getting edge point... "
            << endl;
    }


    if (debugCalBeaconEdgePoint && flag_OFFLINE) {
        (*outfile) << debugHead << "    (" << traceX << "," << traceY << ")" << endl;
    }

    /* Go through pixels along mBeacon. */
    while (traceX <= CPLANE_WIDTH &&
           traceX >= 0 &&
           traceY <= CPLANE_HEIGHT &&
           traceY >= 0) {

        // Using (y2-y1)/(x2-x1) = m, change dependent variable first, then
        // calcualte the independent variable.
        if (abs(mBeacon) > 1) {
            traceY += changeInY;
            traceX  = (int) ((double) (traceY - cy) / mBeacon + (double) cx);

        } else {
            traceX += changeInX;
            traceY = (int) ((mBeacon * (double) (traceX - cx)) + (double) cy);
        }

        if (debugCalBeaconEdgePoint && flag_OFFLINE) {
            (*outfile) << debugHead << "    (" << traceX << "," << traceY << ")" ;
        }

        pixColour = cplane[traceX + traceY * WIDTH] & ~MAYBE_BIT;

        if (goodBlobColour == cBEACON_PINK) {

            /* If it is pink, check next pixel immediately. */
            if (pixColour == cBEACON_PINK) {

                if (debugCalBeaconEdgePoint && flag_OFFLINE) {
                    (*outfile) << "... pink. " << endl;
                }

            /* If it is orange/red, still ok, but have a limit. */
            } else if (((pixColour == cROBOT_RED) ||
                       (pixColour == cBALL))
                      && pinkNoiseCounter < PINK_NOISE_LIMIT) {

                if (debugCalBeaconEdgePoint && flag_OFFLINE) {
                    (*outfile) << "... red/orange. " << endl;
                }
                pinkNoiseCounter++; 

            /* Otherwise, it's unbearable. */
            } else {

                if (debugCalBeaconEdgePoint && flag_OFFLINE) {
                    (*outfile) << "... not pink/red/orange! " << endl;
                }
                outX = checkedX;
                outY = checkedY;
                return;
            }

        } else {

            if (pixColour != goodBlobColour) {

                if (debugCalBeaconEdgePoint && flag_OFFLINE) {
                    (*outfile) << "... NOT " << ColorName[goodBlobColour] << endl;
                }
                outX = checkedX;
                outY = checkedY;
                return;

            } else {
                if (debugCalBeaconEdgePoint && flag_OFFLINE) {
                    (*outfile) << "... " << ColorName[goodBlobColour] << endl;
                }
            }
        }

        /* Remember the good pixel. */
        if (abs(mBeacon) > 1) {
            checkedY = (double) traceY + ((double) changeInY/2);
            checkedX = ((double) (checkedY - cy) / mBeacon + (double) cx);

        } else {
            checkedX = (double) traceX + ((double) changeInX/2);
            checkedY = ((mBeacon * (double) (checkedX - cx)) + (double) cy);
        }
    }
}


/**********************************************************/
/* Check if the beacon blobs are merged with background. */

inline bool VisualCortex::isBeaconBlobMergedWithBackground(
        HalfBeacon *pink,
        HalfBeacon *other,
        double c2cDist,
        bool relaxThreshold) { // use relax thresholds for sanity checks.


    const string debugHead = "isBeaconBlobMergedWithBackground| "; 

    int horiDist = max(pink->getXMax(), other->getXMax()) 
                 - min(pink->getXMin(), other->getXMin()) + 1;

    int vertDist = max(pink->getYMax(), other->getYMax()) 
                 - min(pink->getYMin(), other->getYMin()) + 1;

    double pinkHeight = pink->getYMax() - pink->getYMin() + 1;
    double pinkWidth  = pink->getXMax() - pink->getXMin() + 1;

    double otherHeight = other->getYMax() - other->getYMin() + 1;
    double otherWidth  = other->getXMax() - other->getXMin() + 1;


    // BLOB CENTROID CHECK ------------------------------------------------
    /* Check if the two blobs have similar height and width. */
    double heightDiffRatio = abs(pinkHeight - otherHeight) / MAX(pinkHeight, otherHeight);
    double widthDiffRatio  = abs(pinkWidth - otherWidth) / MAX(pinkWidth, otherWidth);

    double maxHeightDiffRatio = relaxThreshold ? 0.6 : 0.4;
    double maxWidthDiffRatio  = relaxThreshold ? 0.6 : 0.4;

    if (    heightDiffRatio > maxHeightDiffRatio 
        ||  widthDiffRatio > maxWidthDiffRatio) {

        if (debugIsBeaconBlobMergedWithBackground && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead 
                << "...height/width too different! Rework of centroid needed!"
                << endl;
        }
        return true;
    }


    // BLOB CENTROID CHECK ------------------------------------------------

    /* Check if the two blobs have ok similar bounding box sizes. */ 
    double pinkBoxArea  = pinkHeight * pinkWidth;
    double otherBoxArea = otherHeight * otherWidth;
    double blobBoxRatio = otherBoxArea < pinkBoxArea   // Want smaller over larger 
                        ? (double) otherBoxArea / pinkBoxArea
                        : (double) pinkBoxArea / otherBoxArea;

    double minBlobBoxRatio = relaxThreshold ? 0.4 : 0.6;

    if (sqrt(blobBoxRatio) < minBlobBoxRatio) {
        if (debugIsBeaconBlobMergedWithBackground && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead 
                << "...blobBoxRatio too small, rework of centroid needed!"
                << endl;
            (*outfile)
                << debugHead << "   -- pinkBoxArea      : " << pinkBoxArea  << endl
                << debugHead << "   -- otherBoxArea     : " << otherBoxArea << endl
                << debugHead << "   -- blobBoxRatio     : " << blobBoxRatio << endl
                << debugHead << "   -- minBlobBoxRatio  : " << minBlobBoxRatio << endl
                << endl;
        }
        return true;
    }


    // BLOB CENTROID CHECK ------------------------------------------------

    /* Check if the two blobs have ok similar sizes. */ 
    double blobRatio = other->area < pink->area   // Want smaller over larger 
                        ? (double) other->area / pink->area
                        : (double) pink->area / other->area;

    double minBlobRatio = relaxThreshold ? 0.4 : 0.6;

    if (blobRatio < minBlobRatio) {
        if (debugIsBeaconBlobMergedWithBackground && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << endl << debugHead 
                << "...blobRatio too small, rework of centroid needed!"
                << endl;
        }
        return true;
    }

    // BLOB CENTROID CHECK ------------------------------------------------
    /* Check ratio of (desired colors pixels / bounding box pixels). */ 

    double boundingBoxArea = (horiDist * vertDist);

    double solidRatio = (pink->area + other->area) * 1.0 // int to double. 
                        / boundingBoxArea; 

    double minSolidRatio = relaxThreshold ? 0.3 : 0.5;

    if (solidRatio < minSolidRatio) {
        if (debugIsBeaconBlobMergedWithBackground && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << endl << debugHead 
                << "...beacon's density is too low, rework of centroid needed!"
                << endl;
            (*outfile) 
                << debugHead << "    blobsArea      : " << pink->area + other->area << endl
                << debugHead << "    boundingBoxArea: " << boundingBoxArea << endl
                << debugHead << "    solidRatio     : " << solidRatio << endl;

        }
        return true;
    }

    // BLOB CENTROID CHECK ------------------------------------------------
    /* Check aspect ratio of the beacon bounding box. */ 

    double aspectRatio = (vertDist > horiDist) 
                        ?(double) vertDist / (double) horiDist
                        :(double) horiDist / (double) vertDist;

    double maxAspectRatio = relaxThreshold ? 4 : 3;

    if (aspectRatio > maxAspectRatio) {

        if (debugIsBeaconBlobMergedWithBackground && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << endl << debugHead 
                << "...beacon's aspect ratio is not right, rework of centroid needed!"
                << endl;
        }
        return true;
    }
    
    // BLOB CENTROID CHECK ------------------------------------------------
    // Check that x and y dimensions are comparable to c2c distance. 
   
    double maxC2CRatio   = (double)max(vertDist, horiDist) / c2cDist;
    double minC2CRatio   = (double)c2cDist / min(vertDist, horiDist);
    double worstC2CRatio = max(maxC2CRatio, minC2CRatio); 

    double maxWorstC2CRatio = relaxThreshold ? 4 : 3;
 
    if (worstC2CRatio > maxWorstC2CRatio) {

        if (debugIsBeaconBlobMergedWithBackground && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << endl << debugHead 
                << "...beacon dimensions are very different from c2c, "
                << "rework of centroid needed!"
                << endl;
        }
        return true;
    }


    if (debugIsBeaconBlobMergedWithBackground && flag_OFFLINE) {
        (*outfile) << endl << debugHead 
            << "Just checked if any beacon blob is merged to background..." << endl;
        (*outfile) 
            << debugHead << "Pink area: " << pink->area << endl
            << debugHead << "Other area: " << other->area << endl
            << debugHead << "== blobRatio: " << blobRatio << endl
            << debugHead << "Beacon bounding box: "
                << horiDist << " x " << vertDist << " = " << boundingBoxArea << endl
            << debugHead << "== solidRatio: " << solidRatio << endl
            << debugHead << "== aspectRatio: " << aspectRatio << endl
            << debugHead << "== maxC2CRatio: " << maxC2CRatio << endl
            << debugHead << "== minC2CRatio: " << minC2CRatio << endl
            << debugHead << "pinkBoxArea: " << pinkBoxArea << endl
            << debugHead << "otherBoxArea: " << otherBoxArea << endl
            << debugHead << "== blobBoxRatio: " << blobBoxRatio << endl;
    }

    if (debugIsBeaconBlobMergedWithBackground && flag_OFFLINE) {
        (*outfile) << endl << debugHead 
            << "... finished checking blob merging into background, OK!"
            << endl;
    }
    return false;
}


/*
 * Copies all the beacon information from beacon[] into vob[].
 * Filter beacons that have been marked invalid. 
 */
void VisualCortex::copyBeaconsToVobs() {

    const string debugHead = "copyBeaconsToVobs| "; 
    VisualObject *cv = &vob[vobBlueLeftBeacon];

    /* Go through all the 6 beacons. */
    for (int i = 0; 
         i < 6;
         i++, cv++) {

		if (!beacon[i].valid) {

			if (debugCopyBeaconsToVobs && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead
                    << "...classified invalid, beacon[" 
                    << i << "]" << endl;
            }
            cv->reset();
			fireSanity(INVALID_BEACON + i);
			continue;

		} 
        
        cv->x      = beacon[i].xmin;
        cv->y      = beacon[i].ymin;
        cv->width  = beacon[i].xmax - beacon[i].xmin;
        cv->height = beacon[i].ymax - beacon[i].ymin;

		cv->cx = beacon[i].cx;
		cv->cy = beacon[i].cy;

		cv->projected_x = cv->x;
		cv->projected_y = cv->y;
		cv->projected_w = cv->width;
		cv->projected_h = cv->height;

		cv->area = beacon[i].area;
		cv->tilt = beacon[i].angle;
		cv->imgHead = PointToHeading(beacon[i].cx);
		cv->imgElev = PointToElevation(beacon[i].cy);

 	    /* There used to be formula that calculate distance base on
        ** pixels in the blob, and was used for "medium" resolution (when
        ** there used to be 2 possible resolutions. Removed since rev
	    ** 2567. */

        cv->cam_dist = cv->dist = beaconDistanceConstant / beacon[i].c2c + beaconDistanceOffset;

		/* Write only our interested attribute for calibration. */
        if (calibrateBeaconDist && flag_OFFLINE) {
            (*outfile) << beacon[i].c2c << endl;
        }

		//ERIC these printouts are all for centroid testing
		cv->c2c = beacon[i].c2c;
		cv->pinkXsum = beacon[i].pinkXsum;
		cv->pinkYsum = beacon[i].pinkYsum;
		cv->pinkarea = beacon[i].pinkarea;
		cv->closestXsum = beacon[i].closestXsum;
		cv->closestYsum = beacon[i].closestYsum;
		cv->closestarea = beacon[i].closestarea;

		int halfExpectedArea = (int) SQUARE(beacon[i].c2c);

        /* Avoid divide by 0 below. */
        if (halfExpectedArea == 0)
            halfExpectedArea = 1;
        
        cv->misc = beacon[i].c2c;
        cv->cf = MIN((1000 * beacon[i].area) / (halfExpectedArea * 2), 1000);
        cv->var = exp(2 * BEACON_C + 2 * BEACON_M * cv->dist)
                  + SQUARE(varConst * (1 - beacon[i].area / (double) (halfExpectedArea * 2)));

        // This is the variance of the beacon angle.
        cv->angleVar = 25 / (beacon[i].area / (double) (halfExpectedArea * 2));


        /* Find the elev/heading/distance relative to dog's neck base. */
        int startBeacon = vobBlueLeftBeacon;
        RobotRelative(vob[startBeacon+i]);

        /* The beacon can't be further than the field diagonal length. */
		cv->cam_dist = cv->d = cv->dist = MIN(cv->dist, 530);

	    if (debugCopyBeaconsToVobs && flag_OFFLINE) {

            (*outfile) << debugHead
                << "...Beacon (" << vobName[startBeacon+i] << ") "
                << "identified, with elev " << vob[startBeacon+i].elev
                << ", distance: " << vob[startBeacon+i].d
                << endl;
		}
    }
}



/******************************************************************************
 * After finding all possible beacons (4 at most), try to figure out if they
 * make sense or not. If any beacon does not make sense (or if there is
 * conflict between 2 beacons), try to throw one out.
 */
void VisualCortex::BeaconsSanities() {

    int numBeacons = 0;
    static const string debugHead = "BeaconsSanities| ";

    /* See how many beacons have been recognised. */
    for (int i = vobBlueLeftBeacon; i <= vobYellowRightBeacon; i++) {
        if (vob[i].cf > 0) {
            numBeacons++;
        }
    }

    /*---------------------------------------------------*/
    /* If there is not more than one beacon, return. */ 
    if (numBeacons <= 1) {
        return;


    /*---------------------------------------------------*/
    /* If there're 3 beacons, one definitely is insane. */
    } else if (numBeacons == 3) {

        if (vob[vobBlueLeftBeacon].cf > 0 && vob[vobBlueRightBeacon].cf > 0) {
            vob[vobYellowLeftBeacon].reset();
            vob[vobYellowRightBeacon].reset();

            if (debugBeaconsSanities && flag_OFFLINE) {
                (*outfile) << debugHead
                    << "3 beacons observed! Throw out yellow beacons."
                    << endl;
            }
            return;
        }

        if (vob[vobYellowLeftBeacon].cf > 0 && vob[vobYellowRightBeacon].cf > 0) {
            vob[vobBlueLeftBeacon].reset();
            vob[vobBlueRightBeacon].reset();

            if (debugBeaconsSanities && flag_OFFLINE) {
                (*outfile) << debugHead
                    << "3 beacons observed! Throw out blue beacons."
                    << endl;
            }
            return;
        }


    /*---------------------------------------------------*/
    /* If 4 beacons can all be seen, one pair must be unreal. */
    } else if (numBeacons == 4) {

        double yellowPairDist = abs(vob[vobYellowLeftBeacon].x -
                                 vob[vobYellowRightBeacon].x);

        double bluePairDist   = abs(vob[vobBlueLeftBeacon].x -
                                 vob[vobBlueRightBeacon].x);

        /* Roughly, when 2 beacons can be observed, they're about 140 pixels
        ** apart. */
        static const int TWO_BEACONS_IDEAL_DIST = 140;
        
        if (abs(yellowPairDist - TWO_BEACONS_IDEAL_DIST)  
            < abs(bluePairDist - TWO_BEACONS_IDEAL_DIST)) {

            vob[vobBlueLeftBeacon].reset();
            vob[vobBlueRightBeacon].reset();

            if (debugBeaconsSanities && flag_OFFLINE) {
                (*outfile) << debugHead
                    << "4 beacons observed! Throw out blue beacons!"
                    << endl;
            }

        } else {
            vob[vobYellowLeftBeacon].reset();
            vob[vobYellowRightBeacon].reset();

            if (debugBeaconsSanities && flag_OFFLINE) {
                (*outfile) << debugHead
                    << "4 beacons observed! Throw out yellow beacons!"
                    << endl;
            }
        }
        return;


    /*---------------------------------------------------*/
    /* Otherwise there're 2 beacons, there are more things to check... */
    } else {

        VisualObject *beaconBL = &vob[vobBlueLeftBeacon];
        VisualObject *beaconBR = &vob[vobBlueRightBeacon];
        VisualObject *beaconYL = &vob[vobYellowLeftBeacon];
        VisualObject *beaconYR = &vob[vobYellowRightBeacon];

        /*---------------------------------------------------*/
        /* The right beacon appears in the left of left beacon? */
        if (beaconBL->cf > 0 && beaconBR->cf > 0
                && beaconBR->h > beaconBL->h) {

            throwAwayOneBeacon(beaconBL, beaconBR);

            if (debugBeaconsSanities && flag_OFFLINE) {
                (*outfile) 
                    << debugHead << "Contradicting Left/right beacons observed!" << endl
                    << debugHead << "Throw out either BlueLeft OR BlueRight!" << endl;
            }
            return; // No need to do more check, only 2 beacons.
        }

        if (beaconYL->cf > 0 && beaconYR->cf > 0
                && beaconYR->h < beaconYL->h) {

            throwAwayOneBeacon(beaconYL, beaconYR);

            if (debugBeaconsSanities && flag_OFFLINE) {
                (*outfile) 
                    << debugHead << "Contradicting Left/right beacons observed!" << endl
                    << debugHead << "Throw out either YellowLeft OR YellowRight!" << endl;
            }
            return; 
        }

        /*---------------------------------------------------*/
        /* One beacon from each opposite end of field? */
        if (beaconYL->cf > 0 && beaconBL->cf > 0) {
            throwAwayOneBeacon(beaconYL, beaconBL);

            if (debugBeaconsSanities && flag_OFFLINE) {
                (*outfile) 
                    << debugHead << "2 beacons from opposite sides observed!" << endl
                    << debugHead << "Throw out either YellowLeft OR BlueLeft!" << endl;
            }
            return;
        }

        if (beaconYL->cf > 0 && beaconBR->cf > 0) {
            throwAwayOneBeacon(beaconYL, beaconBR);

            if (debugBeaconsSanities && flag_OFFLINE) {
                (*outfile) 
                    << debugHead << "2 beacons from opposite sides observed!" << endl
                    << debugHead << "Throw out either YellowLeft OR BlueRight!" << endl;
            }
            return;
        }

        if (beaconYR->cf > 0 && beaconBL->cf > 0) {
            throwAwayOneBeacon(beaconYR, beaconBL);

            if (debugBeaconsSanities && flag_OFFLINE) {
                (*outfile) 
                    << debugHead << "2 beacons from opposite sides observed!" << endl
                    << debugHead << "Throw out either YellowRight OR BlueLeft!" << endl;
            }
            return;
        }

        if (beaconYR->cf > 0 && beaconBR->cf > 0) {
            throwAwayOneBeacon(beaconYR, beaconBR);

            if (debugBeaconsSanities && flag_OFFLINE) {
                (*outfile) 
                    << debugHead << "2 beacons from opposite sides observed!" << endl
                    << debugHead << "Throw out either YellowRight OR BlueRight!" << endl;
            }
            return;
        }


        /*---------------------------------------------------*/
        /* One beacon's elevation is much higher than the other? */

        /* ONLY 2 logical pairs from here on. */
        if (beaconBL->cf > 0 && beaconBR->cf > 0
                && abs(beaconBL->elev - beaconBR->elev) > 10) {
            throwAwayOneBeacon(beaconBL, beaconBR);

            if (debugBeaconsSanities && flag_OFFLINE) {
                (*outfile) 
                    << debugHead << "2 beacons elevation too different!" << endl
                    << debugHead << "Throw out either BlueLeft OR BlueRight!" << endl;
            }

            return;
        }

        if (beaconYL->cf > 0 && beaconYR->cf > 0
                && abs(beaconYL->elev - beaconYR->elev) > 10) {
            throwAwayOneBeacon(beaconYL, beaconYR);

            if (debugBeaconsSanities && flag_OFFLINE) {
                (*outfile) 
                    << debugHead << "2 beacons elevation too different!" << endl
                    << debugHead << "Throw out either YellowLeft OR YellowRight!" << endl;
            }

            return;
        }


        /*---------------------------------------------------*/
        /* Two beacons are too close? */
        if (beaconBL->cf > 0 && beaconBR->cf > 0
                && abs(beaconBL->h - beaconBR->h) < 35) {
            throwAwayOneBeacon(beaconBL, beaconBR);

            if (debugBeaconsSanities && flag_OFFLINE) {
                (*outfile) 
                    << debugHead << "2 beacons headings too close!" << endl
                    << debugHead << "Throw out either BlueLeft OR BlueRight!" << endl
                    << debugHead << "   beaconBL->h: " << beaconBL->h << endl
                    << debugHead << "   beaconBR->h: " << beaconBR->h << endl;
            }
            return;
        }

        if (beaconYL->cf > 0 && beaconYR->cf > 0
                && abs(beaconYL->h - beaconYR->h) < 35) {
            throwAwayOneBeacon(beaconYL, beaconYR);

            if (debugBeaconsSanities && flag_OFFLINE) {
                (*outfile) 
                    << debugHead << "2 beacons headings too close!" << endl
                    << debugHead << "Throw out either YellowLeft OR YellowRight!" << endl
                    << debugHead << "   beaconYL->h: " << beaconYL->h << endl
                    << debugHead << "   beaconYR->h: " << beaconYR->h << endl;
            }
            return;
        }

        /*---------------------------------------------------*/
        /* Their size differ a lot? */
        if (beaconBL->cf > 0 && beaconBR->cf > 0) {

            double areaBL = beaconBL->width * beaconBL->height; 
            double areaBR = beaconBR->width * beaconBR->height; 

            if (abs(areaBL-areaBR) > 500) {
                throwAwayOneBeacon(beaconBL, beaconBR);

                if (debugBeaconsSanities && flag_OFFLINE) {
                    (*outfile) 
                        << debugHead << "2 beacons size differ too much!" << endl
                        << debugHead << "Throw out either BlueLeft OR BlueRight!" << endl;
                }
            }
            return;
        }

        if (beaconYL->cf > 0 && beaconYR->cf > 0) {

            double areaYL = beaconYL->width * beaconYL->height; 
            double areaYR = beaconYR->width * beaconYR->height; 

            if (abs(areaYL-areaYR) > 500) {
                throwAwayOneBeacon(beaconYL, beaconYR);

                if (debugBeaconsSanities && flag_OFFLINE) {
                    (*outfile) 
                        << debugHead << "2 beacons from opposite sides observed!" << endl
                        << debugHead << "Throw out either YellowRight OR YellowRight!" << endl;
                }
            }
        }

    }
}

/* Pick one beacon from the given 2 and reset it. */
inline void VisualCortex::throwAwayOneBeacon(
        VisualObject *beaconOne,
        VisualObject *beaconTwo) { 

    static const string debugHead = "throwAwayOneBeacon| ";

    if (beaconOne->elev < beaconTwo->elev) {

        beaconTwo->reset();

        if (debugThrowAwayOneBeacon && flag_OFFLINE) {
            (*outfile) << debugHead 
                << "Throw away second beacon."
                << endl;
        }
        
    } else {

        beaconOne->reset();

        if (debugThrowAwayOneBeacon && flag_OFFLINE) {
            (*outfile) << debugHead 
                << "Throw away first beacon."
                << endl;
        }
    }

    return;
}

/******************************************************************************
 * Find the blue and yellow goals.  As each goal can be composed of 
 * separate patches, there is sanity checking on each patch to determine 
 * whether it is a valid "goal" patch.
**/
void VisualCortex::FindGoal() {

    /* Rotate blue and yellow blobs according to amount of roll. */
	blobs->rotate(cBEACON_BLUE, sin_eroll, cos_eroll);
	blobs->rotate(cBEACON_YELLOW, sin_eroll, cos_eroll);

    if (debugFindGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
        (*outfile) << "FindGoal| look for blue goal" << endl;
    }
	LookForColourGoal(hbBlue, numBlue, 
                      vobBlueLeftBeacon, vobBlueRightBeacon, 
                      vobBlueGoal);


    if (debugFindGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
        (*outfile) << "FindGoal| look for yellow goal" << endl;
    }

    /* The beacons order is different because YellowRight beacon actually
    ** appears in left of yellow goal in cplane. */
	LookForColourGoal(hbYellow, numYellow, 
                      vobYellowRightBeacon, vobYellowLeftBeacon, 
                      vobYellowGoal);
}


/* Called after finding goals and beacons, to ensure we don't see things that
** are impossible to fit in one frame. */
inline void VisualCortex::LandmarkSanities() {

    static const string debugHead = "LandmarkSanities| ";

    // TODO:  Should compare variance to see which one should be removed.
    //        Probable things to compare - how solid is the patch?

    /* CHECK ----------------------------------------------------------------*/
    /* Cannot see yellow goal and blue beacons together. */
	if (vob[vobYellowGoal].cf > 0 && vob[vobBlueRightBeacon].cf > 0) {
		vob[vobBlueRightBeacon].reset();
		fireSanity(YELLOW_GOAL_NOT_BLUE_RIGHT_BEACON);

        if (debugLandmarkSanities && flag_OFFLINE) {
            (*outfile) << debugHead
                << "...cannot see both yellow goal and blue right beacon!"
                << endl;
        }
	}

	if (vob[vobYellowGoal].cf > 0 && vob[vobBlueLeftBeacon].cf > 0) {
		vob[vobBlueLeftBeacon].reset();
		fireSanity(YELLOW_GOAL_NOT_BLUE_LEFT_BEACON);

        if (debugLandmarkSanities && flag_OFFLINE) {
            (*outfile) << debugHead
                << "...cannot see both yellow goal and blue left beacon!"
                << endl;
        }
	}

    /* CHECK ----------------------------------------------------------------*/
    /* Cannot see blue goal and yellow beacons together. */
	if (vob[vobBlueGoal].cf > 0 && vob[vobYellowRightBeacon].cf > 0) {
		vob[vobYellowRightBeacon].reset();
		fireSanity(BLUE_GOAL_NOT_YELLOW_RIGHT_BEACON);

        if (debugLandmarkSanities && flag_OFFLINE) {
            (*outfile) << debugHead
                << "...cannot see both blue goal and yellow right beacon!"
                << endl;
        }
    }

	if (vob[vobBlueGoal].cf > 0 && vob[vobYellowLeftBeacon].cf > 0) {
		vob[vobYellowLeftBeacon].reset();
		fireSanity(BLUE_GOAL_NOT_YELLOW_LEFT_BEACON);

        if (debugLandmarkSanities && flag_OFFLINE) {
            (*outfile) << debugHead
                << "...cannot see both blue goal and yellow left beacon!"
                << endl;
        }
    }
}

inline void VisualCortex::GoalSanities() {

    static const string debugHead = "GoalSanities| ";

    /* Removed checking using elevation since rev 2726. Use horizon check
    ** in FindGoal instead. */

    /* CHECK ----------------------------------------------------------------*/
    /* Check yellow goal elevation. */
    if (vob[vobYellowGoal].cf && vob[vobYellowGoal].width < WIDTH - 5 
            && vob[vobYellowGoal].elev > 25.0) {
        vob[vobYellowGoal].reset();

        if (debugGoalSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead
                << "...Yellow goal too high, sanitied out." << endl;
        }
        fireSanity(YELLOW_GOAL_TOO_HIGH);
    }

    /* CHECK ----------------------------------------------------------------*/
    /* Check blue goal elevation. */
    if (vob[vobBlueGoal].cf && vob[vobBlueGoal].width < WIDTH - 5 && vob[vobBlueGoal].elev> 25.0) {
        vob[vobBlueGoal].reset();

        if (debugGoalSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead
                << "...blue goal too high, sanitied out" << endl;
        }
        fireSanity(BLUE_GOAL_TOO_HIGH);
    }

	/*
	   if (vob[vobYellowGoal].cf && vob[vobYellowGoal].elev < 0.0)
	   {
	   vob[vobYellowGoal].cf = 0;
	   vob[vobYellowGoal].var = LARGE_VAL;
	   vob[vobYellowGoal].angleVar = LARGE_VAL;
	   fireSanity(YELLOW_GOAL_TOO_LOW);
	   }
	   if (vob[vobBlueGoal].cf && vob[vobBlueGoal].elev < 0.0)
	   {
	   vob[vobBlueGoal].cf = 0;
	   vob[vobBlueGoal].var = LARGE_VAL;
	   vob[vobBlueGoal].angleVar = LARGE_VAL;
	   fireSanity(BLUE_GOAL_TOO_LOW);
	   } 	   
	*/

    /* CHECK ----------------------------------------------------------------*/
    /* We can't see any yellow beacons and the blue goal at the same time. */
    if ((vob[vobYellowLeftBeacon].cf > 200 || vob[vobYellowRightBeacon].cf > 200) 
        && vob[vobBlueGoal].cf) {

		vob[vobBlueGoal].reset();
		fireSanity(YELLOW_BEACON_NOT_BLUE_GOAL);

        if (debugGoalSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << "yellow beacon and blue goal same time, sanitied out" << endl;
        }
	}

    /* CHECK ----------------------------------------------------------------*/
	// we can't see any blue beacons and the yellow goal at the same time
	if ((vob[vobBlueLeftBeacon].cf > 200 || vob[vobBlueRightBeacon].cf > 200) && vob[vobYellowGoal].cf) {

		// if ((vob[vobBlueLeftBeacon].var < get95CF(40) || 
        	// vob[vobBlueRightBeacon].var < get95CF(40)) && vob[vobYellowGoal].cf < LARGE_VAL) {
		// wout << "sanity 2\n";

		vob[vobYellowGoal].reset();
		fireSanity(BLUE_BEACON_NOT_YELLOW_GOAL);

        if (debugGoalSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << "blue beacon and yellow goal same time, sanitied out" << endl;
        }
	}

	/*
		// if the goal is too far away, forget it
		if (vob[vobBlueGoal].cf && vob[vobBlueGoal].dist > (FIELD_LENGTH * 1.2)) {
			// if (vob[vobBlueGoal].var < LARGE_VAL && vob[vobBlueGoal].dist > (FIELD_LENGTH * 1.2)) {
			// wout << "sanity 3\n";
			vob[vobBlueGoal].cf = 0;
			vob[vobBlueGoal].var = LARGE_VAL;
			fireSanity(BLUE_GOAL_TOO_FAR);
		}
		// if the goal is too far away, forget it
		if (vob[vobYellowGoal].cf && vob[vobYellowGoal].dist > (FIELD_LENGTH * 1.2)) {
			// if (vob[vobYellowGoal].var < LARGE_VAL && vob[vobYellowGoal].dist > (FIELD_LENGTH * 1.2)) {
	 		// wout << "sanity 4\n";
			vob[vobYellowGoal].cf = 0;
			vob[vobYellowGoal].var = LARGE_VAL;
			fireSanity(YELLOW_GOAL_TOO_FAR);
		}
	*/
	
    /* CHECK ----------------------------------------------------------------*/
	// We can't see both goals at the same time.
	if (vob[vobBlueGoal].cf && vob[vobYellowGoal].cf) {
		// if (vob[vobBlueGoal].var < LARGE_VAL && vob[vobYellowGoal].var < LARGE_VAL) {
		// wout << "sanity 5\n";
		if (vob[vobBlueGoal].cf >= vob[vobYellowGoal].cf) {
			vob[vobYellowGoal].cf = 0;
			fireSanity(BLUE_GOAL_CONFIDENT_THAN_YELLOW_GOAL);
		} else {
			vob[vobBlueGoal].cf = 0;
			fireSanity(YELLOW_GOAL_CONFIDENT_THAN_BLUE_GOAL);
		}

		if (vob[vobBlueGoal].var <= vob[vobYellowGoal].var) {
			vob[vobYellowGoal].var = LARGE_VAL;
			vob[vobYellowGoal].angleVar = LARGE_VAL;
			fireSanity(BLUE_GOAL_VARIANCE_LESS_THAN_YELLOW_GOAL);

		} else {
			vob[vobBlueGoal].var = LARGE_VAL;
			vob[vobBlueGoal].angleVar = LARGE_VAL;
			fireSanity(YELLOW_GOAL_VAIRANCE_LESS_THAN_BLUE_GOAL);
		}
		
        if (debugGoalSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead
                << "...both goal seen, sanitied out" << endl;
        }
	}
}


/*****************************************************************************
 *  Using the blobs generated by Process(), identify visual objects.
 *
 *  @return     number of visual objects identified.
 */
int VisualCortex::FindVisualObjects() {

    if (lookForBeaconsAndGoals) {
        FindBeacons();      // Find all the beacons first.
        BeaconsSanities();

        FindGoal();         // Find the blue and yellow goals.
        GoalSanities();

        LandmarkSanities(); // Do beacons/goals locations make sense?
    }
    
    if (flag_LOC_CHALLENGE)
        FindPinkLocalisationBlobs();        

    FindBall();         // Find the ball.
    FindRobot();        // Find red and blue robots.


    if (debugFindVisualObjects && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
        (*outfile) << "FindVisualObjects| Done all finding." << endl;
    }

	/* Now return the number of vobs we found. */
    int n = 0;
    for (int i = 0; i < VOB_COUNT; i++) {
        if (vob[i].cf || vob[i].var != LARGE_VAL) {
                n++;
        }
    }

	return n;
}


/*****************************************************************************/
/* Converts large pink blobs into visual objects, for use in the             */
/* localisation challenge.                                                   */

void VisualCortex::FindPinkLocalisationBlobs() {
    double density;
    bool skip;
    int n = 0;

    for (int i = 0; i < numPink; i++) {
        skip = false;
    
        // Calculate robot relative heading and elevation
        localisePink[n].reset();
		localisePink[n].imgHead = PointToHeading(hbPink[i]->cx);
		localisePink[n].imgElev = PointToElevation(hbPink[i]->cy);
		localisePink[n].area = hbPink[i]->area;
		localisePink[n].width = hbPink[i]->getXMax() - hbPink[i]->getXMin();
		localisePink[n].height = hbPink[i]->getYMax() - hbPink[i]->getYMin();
		localisePink[n].x = hbPink[i]->cx - localisePink[n].width / 2.0;
		localisePink[n].y = hbPink[i]->cy - localisePink[n].height / 2.0;

        // We don't know the distance to a simple blob, so just use a large
        // distance so that interference due to head position / stance is
        // small.
		localisePink[n].dist = 500;
        RobotRelative(localisePink[n]);
        
        // Some sanity checks for the pink blobs
        
        // Don't use halfbeacons that are very small
        if (localisePink[n].area < 30) skip = true;
        if (localisePink[n].width < 4) skip = true;
        if (localisePink[n].height < 4) skip = true;

        // We want solid pink blobs, so remove blobs with low density
        density = localisePink[n].area;
        density /= (localisePink[n].width * localisePink[n].height);
        if (density < 0.85) skip = true;

        // Don't use halfbeacons that are inside the main beacons
	    for (int j = vobBlueLeftBeacon; j <= vobYellowRightBeacon; j++) {
            if (vob[j].cf || vob[j].var != LARGE_VAL) {
                if (VobInVob(localisePink[n], vob[j], 5, 5, 5, 5))
                    skip = true;
            }
        }
        
        if (!skip) n++;
    }
    
    numLocalisePink = n;
}

/*****************************************************************************/
// return if b1 is completely inside b2
// return if b1 is completely inside b2
// offset grows b2
int VisualCortex::VobInVob(VisualObject &b1, VisualObject &b2, int ot, int or_,
        		   int ob, int ol) {
    return (b1.x >= b2.x - ol && b1.y >= b2.y - ot
            && (b1.x + b1.width) <= (b2.x + b2.width + or_)
            && (b1.y + b1.height) <= (b2.y + b2.height + ob));
}


/* These values should be put into .h evantually. Right now this is faster to
 * compile. */
// static const int NECK_HEIGHT    = 7;
static const int FRAME = 5;
static const int MAXPOINTS = 8;
static const int TILT_TRIGGER = 25;
static const int INFINITE = 10000;
static const int COVERSCREEN = 22000;

/* 250 is for ERS200, ERS7 value = 250 * (208/160)/(176*144) = 328 */ 
// If less than this not really big enough to be a ball within the frame box 
// thingie anyways.. well it is but not really.
static const int AREA_TRIGGER = 328; 


/*****************************************************************************/
/* Change the cplane co-ordinates into local co-ordinates base on dog's neck. */
void VisualCortex::projectPoints(int *inPoints, int numPoints, double *outPoints,
    				 double projectHeight, bool debug) {

	for (int i = 0; i < numPoints; i++) {

        	double ipx = inPoints[i*2];
        	double ipy = inPoints[i*2 + 1];

	        // cout << "Image " << ipx << " " << ipy << endl;

 	        double projectedX, projectedY;

            /* Call getPointProjection in PointProjection.h. */

            /*
            getPointProjection(&projectedX, &projectedY, degPan, deg-ilt,
                                degCrane, ipx, ipy);
            */

            getPointProjection(&projectedX, &projectedY, degPan, degTilt + degBodyTilt, degCrane, ipx, ipy,
                		   FACE_LENGTH * 10.0, NECK_LENGTH * 10.0,
                           neckBaseHeight - projectHeight * 10.0,
                		   APERTURE_DISTANCE, CAMERA_OFFSET, debug);

        // cout << " Projected  " << projectedX << "  " << projectedY << endl;

        outPoints[i * 2] = projectedX / 10.0;
        outPoints[i * 2 + 1] = projectedY / 10.0;
    }
}

/*****************************************************************************/
/* TODO: this take an extra unneccesary step of copying the whole array.
 * Should delete this sometime. */
/* Similar version, output as ints , NOT SURE WHY some functions needs it?? */
void VisualCortex::projectPoints(int *inPoints, int numPoints, int *outPoints,
        			 double projectHeight, bool debug) {

    double outPointsDouble[numPoints*2];

    projectPoints(inPoints, numPoints, outPointsDouble, projectHeight, debug);

    for (int i = 0; i < 2 * numPoints; i++) {
        outPoints[i] = (int) outPointsDouble[i];
    }

}

void VisualCortex::equationEdge(int start, int end, double m, double b, bool reverse,
				int *intersect_x, int *intersect_y, int edgeThreshold) {

	const int INFINITE = 10000;
	if (reverse) {
		for (int i = start; i >= end; i--) {
			if (m >= INFINITE) {
				int c = cplane[(int) b + i *WIDTH] & ~MAYBE_BIT;
				if (c == cBALL || c == cROBOT_RED) {
					int j = i - 1;
					for (; j >= 0; j--) {
						int nc = cplane[(int) b + j *WIDTH] & ~MAYBE_BIT;
						if (nc != cBALL && nc != cROBOT_RED)
							break;
					}
					if (j <= 0 || j <= i - edgeThreshold) {
						*intersect_x = (int) b;
						*intersect_y = i;
						return;
					}
				}
			}
			else if (abs(m) <= 1) {
				int c = cplane[i + (int) (i *m + b) * WIDTH] & ~MAYBE_BIT;
				if (c == cBALL || c == cROBOT_RED) {
					int j = i - 1;
					for (; j >= 0; j--) {
						int nc = cplane[j + (int) (j *m + b) * WIDTH] & ~MAYBE_BIT;
						if (nc != cBALL && nc != cROBOT_RED)
							break;
					}
					if (j <= 0 || j <= i - edgeThreshold) {
					    *intersect_x = i;
					    *intersect_y = (int) (i * m + b);
					    return;
					}
				}
			}
			else {
				int c = cplane[(int) ((i - b) / m) + i *WIDTH] & ~MAYBE_BIT;
				if (c == cBALL || c == cROBOT_RED) {
					int j = i - 1;
					for (; j >= 0; j--) {
						int nc = cplane[(int) ((j - b) / m) + j *WIDTH] & ~MAYBE_BIT;
						if (nc != cBALL && nc != cROBOT_RED)
							break;
					}
					if (j <= 0 || j <= i - edgeThreshold) {
						*intersect_x = (int) ((i - b) / m);
						*intersect_y = i;
						return;
					}
				}
			}
		}
	}
	else {
		for (int i = start; i <= end; i++) {
			if (m >= INFINITE) {
				int c = cplane[(int) b + i *WIDTH] & ~MAYBE_BIT;
				if (c == cBALL || c == cROBOT_RED) {
					int j = i + 1;
					for (; j < HEIGHT; j++) {
						int nc = cplane[(int) b + j *WIDTH] & ~MAYBE_BIT;
						if (nc != cBALL && nc != cROBOT_RED)
							break;
					}
					if (j >= HEIGHT || j >= i + edgeThreshold) {
						*intersect_x = (int) b;
						*intersect_y = i;
						return;
					}
				}
			}
			else if (abs(m) <= 1) {
				int c = cplane[i + (int) (i *m + b) * WIDTH] & ~MAYBE_BIT;
				if (c == cBALL || c == cROBOT_RED) {
					int j = i + 1;
					for (; j < WIDTH; j++) {
						int nc = cplane[j + (int) (j *m + b) * WIDTH] & ~MAYBE_BIT;
						if (nc != cBALL && nc != cROBOT_RED)
							break;
					}
					if (j >= WIDTH || j >= i + edgeThreshold) {
						*intersect_x = i;
						*intersect_y = (int) (i * m + b);
						return;
					}
				}
			}
			else {
				int c = cplane[(int) ((i - b) / m) + i *WIDTH] & ~MAYBE_BIT;
				if (c == cBALL || c == cROBOT_RED) {
					int j = i + 1;
					for (; j < HEIGHT; j++) {
						int nc = cplane[(int) ((j - b) / m) + j *WIDTH] & ~MAYBE_BIT;
						if (nc != cBALL && nc != cROBOT_RED)
							break;
					}
					if (j >= HEIGHT || j >= i + edgeThreshold) {
						*intersect_x = (int) ((i - b) / m);
					        *intersect_y = i;
						return;
					}
				}
			}
		}
	}    
	*intersect_x = *intersect_y = -1;
}

// TPC (third party comment) : 
// This code is dodgy, Should be rewritten
// 
void VisualCortex::findBallEdge(int start, int end, int fixed, bool outside, bool reverse,
				bool vertical, int *resultx, int *resulty, int edgeThreshold) {
	/*
		#ifdef OFFLINE
		#ifndef PRE_SANITY_BLOB
			(*outfile) << "start " << start << " end " << end << " fixed " << fixed << " reverse " << reverse << " vertical " << vertical << endl;
		#endif
		#endif
	*/

	/*
		// Kim's testing on the number of pixels lying on the edge. 
		int countBall = 0;
		int u = start;
		while (u != end) {
			int c = (vertical ? cplane[fixed + u * WIDTH] : cplane[u + fixed * WIDTH]);
            c = c & ~MAYBE_BIT; 
			if (c == cBALL || c == cROBOT_RED) countBall ++;
			if (reverse) u --;
			else u++;
		}
		{
		int c = (vertical ? cplane[fixed + u * WIDTH] : cplane[u + fixed * WIDTH]);
        c = c & ~MAYBE_BIT; 
		if (c == cBALL || c == cROBOT_RED) countBall ++;
		}
		cout << " There are " << countBall << " ball pixels on this edges " << start << "," << end << "," << fixed<< " " << vertical << endl;
	*/

	if (!reverse) {
		for (int i = start; i <= end; i++) {
			int c = (vertical 
                        ? (cplane[fixed + i *WIDTH] & ~MAYBE_BIT) 
                        : (cplane[i + fixed *WIDTH] & ~MAYBE_BIT)
                    );
			if (outside && (c == cBALL || c == cROBOT_RED)) {
				// see if the subsequent pixels are ball
				int j = i + 1;
				for (; j < (vertical ? HEIGHT : WIDTH); j++) {
					int nc = (vertical ? cplane[fixed + j *WIDTH] : cplane[j + fixed *WIDTH]);
                    nc = nc & ~MAYBE_BIT; 
					if (nc != cBALL && nc != cROBOT_RED)
						break;
				}
				if (j == (vertical ? HEIGHT : WIDTH) || j >= i + edgeThreshold) {
					*resultx = (vertical ? fixed : i);
					*resulty = (vertical ? i : fixed);
					/*
						#ifdef OFFLINE
						#ifndef PRE_SANITY_BLOB
							(*outfile) << "found (" << *resultx << "," << *resulty << ")" << endl;
						#endif
						#endif
					*/
					return;
				}
				i = j;
			}
			else if (!outside && (c != cBALL && c != cROBOT_RED)) {
				// see if the subsequent pixels are bg
				int j = i + 1;
				for (; j < (vertical ? HEIGHT : WIDTH); j++) {
					int nc = (vertical ? cplane[fixed + j *WIDTH] : cplane[j + fixed *WIDTH]);
                    nc = nc & ~MAYBE_BIT; 
					if (nc == cBALL || nc == cROBOT_RED)
						break;
				}
				if (j == (vertical ? HEIGHT : WIDTH) || j >= i + edgeThreshold) {
					*resultx = (vertical ? fixed : i - 1);
					*resulty = (vertical ? i - 1 : fixed);
					/*
						#ifdef OFFLINE
						#ifndef PRE_SANITY_BLOB
							(*outfile) << "found (" << *resultx << "," << *resulty << ")" << endl;
						#endif
						#endif
					*/
					return;
				}
				i = j;
			}
		}
	}
	else {
		for (int i = start; i >= end; i--) {
			int c = (vertical ? cplane[fixed + i *WIDTH] : cplane[i + fixed *WIDTH]);
            c = c & ~MAYBE_BIT; 
			if (outside && (c == cBALL || c == cROBOT_RED)) {
				// see if the subsequent pixels are ball
				int j = i - 1;
				for (; j >= 0; j--) {
					int nc = (vertical ? cplane[fixed + j *WIDTH]
						  : cplane[j + fixed *WIDTH]);
                    nc = nc & ~MAYBE_BIT; 
					if (nc != cBALL && nc != cROBOT_RED)
						break;
				}
				if (j <= 0 || j <= i - edgeThreshold) {
					*resultx = (vertical ? fixed : i);
					*resulty = (vertical ? i : fixed);
					/*
						#ifdef OFFLINE
						#ifndef PRE_SANITY_BLOB
							(*outfile) << "found (" << *resultx << "," << *resulty << ")" << endl;
						#endif
						#endif
					*/
					return;
				}
				i = j;
			}
			else if (!outside && (c != cBALL && c != cROBOT_RED)) {
				// see if the subsequent pixels are bg
				int j = i - 1;
				for (; j >= 0; j--) {
					int nc = (vertical ? cplane[fixed + j *WIDTH]
						  : cplane[j + fixed *WIDTH]);
                    nc = nc & ~MAYBE_BIT; 
					if (nc == cBALL || nc == cROBOT_RED)
						break;
				}
				if (j <= 0 || j <= i - edgeThreshold) {
					*resultx = (vertical ? fixed : i + 1);
					*resulty = (vertical ? i + 1 : fixed);
					/*
						#ifdef OFFLINE
						#ifndef PRE_SANITY_BLOB
							(*outfile) << "found (" << *resultx << "," << *resulty << ")" << endl;
						#endif
						#endif
					*/
					return;
				}
				i = j;
			}
		}
	}
	*resultx = *resulty = -1;
	/*
		#ifdef OFFLINE
		#ifndef PRE_SANITY_BLOB
			(*outfile) << "found (" << *resultx << "," << *resulty << ")" << endl;
		#endif
		#endif
	*/
}        

// Whether a point is outside or inside the circle
// this is a very strict check because it is only applied to 
// pixels outside of the frame anyway, so it basically only
// needs to tell whether we are directly inside a ball or out
// and if directly in, then this should be predominantly orange
// i expect something like.. 20 / 25 pixels orange
bool VisualCortex::isOutside(int px, int py) {
	int count = 0;
	int total = 0;
	for (int i = py - 2; i <= py + 2; i++) {
		for (int j = px - 2; j <= px + 2; j++) {
			int c = cplane[i *WIDTH + j] & ~MAYBE_BIT;
			if (c == cBALL || c == cROBOT_RED)
				count++;
			total++;
		}
	}
	return ((count * 1.0) / total) <= 0.8;
}

// Counts the number of specified colour points in the rectangle specified
// by minX, minY, maxX and maxY
int VisualCortex::countColouredPixels(int colour, int minX, int minY, int maxX, int maxY) {
	int count = 0;        
	for (int i = minY; i <= maxY; i++) {
		for (int j = minX; j <= maxX; j++) {
			int c = cplane[i *WIDTH + j] & ~MAYBE_BIT;
			if (c == colour)
				count++;
		}
	}    
	return count;
}    

// TPC (third party comments) : 
// find two points on the bounding box that intersects circumference of ball
// nl = north left , nr = north right , sl = south left , sr = south right 
// these are for checking if it's fireball or not, and how the fireball looks like
// ONLY WORKS with FIREBALL !!! 
void VisualCortex::findPointOnCircumferenceOneAndTwo(int bxmin, int bxmax, int bymin, int bymax,
						     bool nl, bool nr, bool sl, bool sr, int *p1x,
						     int *p1y, int *p2x, int *p2y) {

		int candidate_x[MAXPOINTS], candidate_y[MAXPOINTS];

		// clear the candidates
		for (int i = 0; i < MAXPOINTS; i++) {
			candidate_x[i] = candidate_y[i] = -1;
		}
		
		int start, end, fixed;
		int edgeThresholdX = max(FRAME - 1, (bxmax - bxmin) / 10);
		int edgeThresholdY = max(FRAME - 1, (bymax - bymin) / 10);
		//int edgeThresholdX = 4;
		//int edgeThresholdY = 4;

		// south
		start = max(bxmin, FRAME);
		end = min(bxmax, WIDTH - FRAME);
		fixed = HEIGHT - FRAME;
		findBallEdge(start, end, fixed, sl, false, false, &(candidate_x[0]), &(candidate_y[0]), edgeThresholdX);
		findBallEdge(end, start, fixed, sr, true, false, &(candidate_x[1]), &(candidate_y[1]), edgeThresholdX);

		// north
		fixed = FRAME;
		findBallEdge(start, end, fixed, nl, false, false, &(candidate_x[2]), &(candidate_y[2]), edgeThresholdX);
		findBallEdge(end, start, fixed, nr, true, false, &(candidate_x[3]), &(candidate_y[3]), edgeThresholdX);

		// left
		start = max(bymin, FRAME);
		end = min(bymax, HEIGHT - FRAME);
		fixed = FRAME;
		findBallEdge(start, end, fixed, nl, false, true, &(candidate_x[4]), &(candidate_y[4]), edgeThresholdY);
		findBallEdge(end, start, fixed, sl, true, true, &(candidate_x[5]), &(candidate_y[5]), edgeThresholdY);

		// right
		fixed = WIDTH - FRAME;
		findBallEdge(start, end, fixed, nr, false, true, &(candidate_x[6]), &(candidate_y[6]), edgeThresholdY);
		findBallEdge(end, start, fixed, sr, true, true, &(candidate_x[7]), &(candidate_y[7]), edgeThresholdY);

		/*
		//DEBUG
			for (int i = 0; i < MAXPOINTS; i++) {
				cout << "Edge point " << i << " " << candidate_x[i] << " , " << candidate_y[i] << endl;
			}
		*/

		// try to pick the points intelligently
		// can be done more cleverly if i actually order the candidates in clockwise and
		// anti clockwise direction, REMIND ME to come back
		if (nl && nr && sl && sr) {
			for (int i = 0; i < MAXPOINTS; i++) {
				if (candidate_x[i] > 0 && i % 2 == 0) {
					*p1x = candidate_x[i];
					*p1y = candidate_y[i];
					*p2x = candidate_x[i + 1];
					*p2y = candidate_y[i + 1];
					break;
				}
			}
		}
		else if (nl && nr && sl) {
			*p1x = candidate_x[0]; *p1y = candidate_y[0];
			*p2x = candidate_x[6]; *p2y = candidate_y[6];
		}
		else if (nl && nr && sr) {
			*p1x = candidate_x[1]; *p1y = candidate_y[1];
			*p2x = candidate_x[4]; *p2y = candidate_y[4];
		}
		else if (sl && sr && nr) {
			*p1x = candidate_x[3]; *p1y = candidate_y[3];
			*p2x = candidate_x[5]; *p2y = candidate_y[5];
		}
		else if (sl && sr && nl) {
			*p1x = candidate_x[2]; *p1y = candidate_y[2];
			*p2x = candidate_x[7]; *p2y = candidate_y[7];
		}
		else if (nl && nr) {
			*p1x = candidate_x[6]; *p1y = candidate_y[6];
			*p2x = candidate_x[4]; *p2y = candidate_y[4];
		}
		else if (sl && sr) {
			*p1x = candidate_x[5]; *p1y = candidate_y[5];
			*p2x = candidate_x[7]; *p2y = candidate_y[7];
		}
		else if (nr && sr) {
			*p1x = candidate_x[3]; *p1y = candidate_y[3];
			*p2x = candidate_x[1]; *p2y = candidate_y[1];
		}
		else if (nl && sl) {
			*p1x = candidate_x[2]; *p1y = candidate_y[2];
			*p2x = candidate_x[0]; *p2y = candidate_y[0];
		}
		else if (nr) {
			*p1x = candidate_x[3]; *p1y = candidate_y[3];
			*p2x = candidate_x[6]; *p2y = candidate_y[6];
		}
		else if (nl) {
			*p1x = candidate_x[2]; *p1y = candidate_y[2];
			*p2x = candidate_x[4]; *p2y = candidate_y[4];
		}
		else if (sr) {
			*p1x = candidate_x[1]; *p1y = candidate_y[1];
			*p2x = candidate_x[7]; *p2y = candidate_y[7];
		}
		else if (sl) {
			*p1x = candidate_x[5]; *p1y = candidate_y[5];
			*p2x = candidate_x[0]; *p2y = candidate_y[0];
		}
	}

	void VisualCortex::findPointOnCircumferenceThree(int bxmin,
			int bxmax,
			int bymin,
			int bymax,
			bool nl,
			bool nr,
			bool sl,
			bool sr,
			int p1x,
			int p1y,
			int p2x,
			int p2y,
			int *p3x,
			int *p3y) {
		// now find definition of perpendicular bisector of p1 and p2
		int p12x, p12y;
		double normal, tangent, b;
		int edgeThresholdX = max(FRAME - 1, (bxmax - bxmin) / 10);
		int edgeThresholdY = max(FRAME - 1, (bymax - bymin) / 10);
		//int edgeThresholdX = 4;
		//int edgeThresholdY = 4;
		p12x = (p1x + p2x) / 2;
		p12y = (p1y + p2y) / 2;
		if (p2x != p1x) {
			tangent = ((p2y - p1y) * 1.0) / (p2x - p1x);
			normal = (tangent == 0) ? INFINITE : (-1 / tangent);
		}
		else {
			tangent = INFINITE;
			normal = 0;
		}            
		if (normal < INFINITE) {
			b = p12y - normal * p12x;
		}
		else {
			b = p12x;
		}

		// now id the 3rd edge
		int leftCount, rightCount, topCount, bottomCount;
		leftCount = rightCount = topCount = bottomCount = 0;
		if (sl) {
			leftCount++; bottomCount++;
		}
		if (nl) {
			leftCount++; topCount++;
		}
		if (sr) {
			rightCount++; bottomCount++;
		}
		if (nr) {
			rightCount++; topCount++;
		}

		if (abs(normal) <= 1) {
			// the case where more x than y
			if (leftCount > rightCount) {
				//cout << "start from left" << endl;
				equationEdge(bxmin, bxmax, normal, b, false, p3x, p3y, edgeThresholdX);
			}
			else if (rightCount > leftCount) {
				//cout << "start from right" << endl;
				equationEdge(bxmax, bxmin, normal, b, true, p3x, p3y, edgeThresholdX);
			}
			else {
				// depends which edge closer
				if ((bxmax - (WIDTH - FRAME)) > (FRAME - bxmin)) {
					//cout << "start from left" << endl;
					equationEdge(bxmin, bxmax, normal, b, false, p3x, p3y, edgeThresholdX);
				}
				else {
					//cout << "start from right" << endl;
					equationEdge(bxmax, bxmin, normal, b, true, p3x, p3y, edgeThresholdX);
				}
			}
		}
		else {
			// the case where more y than x
			if (topCount > bottomCount) {
				//cout << "start from up" << endl;
				equationEdge(bymin, bymax, normal, b, false, p3x, p3y, edgeThresholdY);
			}
			else if (bottomCount > topCount) {
				//cout << "start from down" << endl;
				equationEdge(bymax, bymin, normal, b, true, p3x, p3y, edgeThresholdY);
			}
			else {
				//depends which edge closer
				if ((bymax - (HEIGHT - FRAME)) > (FRAME - bymin)) {
					//cout << "start from up" << endl;
					equationEdge(bymin, bymax, normal, b, false, p3x, p3y, edgeThresholdY);
				}
				else {
					//cout << "start from bottom" << endl;
					equationEdge(bymax, bymin, normal, b, true, p3x, p3y, edgeThresholdY);
				}
			}
		}
	}

	// debugging, basically i put stuff in here so if they crash then i know it's them.. 
	// dumb way of finding out for sure what crashed.. when i have absolutely no idea
	void VisualCortex::testFunction(int p1x,
			int p1y,
			double p13m,
			double p13b,
			double p23m,
			double p23b,
			int *ix,
			int *iy,
			double *radius) {
	}

	// find intersection of the perpendicular bisector between 3 points
	void VisualCortex::findIntersectionOfTwoChords(int p1x,
			int p1y,
			int p2x,
			int p2y,
			int p3x,
			int p3y,
			int *ix,
			int *iy,
			double *radius) {
		// cout << "found points : (" << p1x << "," << p1y << ") (" << p3x << "," << p3y << ") (" << p2x << "," << p2y << ")" << endl;
		// now that we have the three points, try to find the centre by finding the intersection of 
		// perpendicular bisector of the chords p1 p3 and p1 p2
		int p13x, p13y, p23x, p23y;
		double p13m, p13b, p23m, p23b;

		p13x = (p1x + p3x) / 2;
		p13y = (p1y + p3y) / 2;
		if (p1y == p3y) {
			p13m = INFINITE;
		}
		else {
			p13m = (p1x == p3x) ? 0 : -1 / (((p3y - p1y) * 1.0) / (p3x - p1x));
		}
		p13b = (p13m == INFINITE) ? p13x : p13y - p13m * p13x;

		p23x = (p2x + p3x) / 2;
		p23y = (p2y + p3y) / 2;
		if (p2y == p3y) {
			p23m = INFINITE;
		}
		else {
			p23m = (p2x == p3x) ? 0 : -1 / (((p2y - p3y) * 1.0) / (p2x - p3x));
		}
		p23b = (p23m == INFINITE) ? p23x : p23y - p23m * p23x;

		if (p23m == p13m) {
			//cout << "ERROR !!!!!!!!!!!!!!!!!!! parallel lines found !!!!!!!!!!!!!!!!!!!!!! ERROR" << endl;
			return;
		}

		if (p13m == INFINITE) {
			//cout << "p1 and p3 horizontal" << endl;
			*ix = (int) p13b;
			*iy = (int) (*ix * p23m + p23b);
		}
		else if (p23m == INFINITE) {
			//cout << "p2 and p3 horizontal" << endl;
			*ix = (int) p23b;
			*iy = (int) (*ix * p13m + p13b);
		}
		else {
			//cout << "normal simultaneous equation" << endl;
			*ix = (int) ((p13b - p23b) / (p23m - p13m));
			*iy = (int) (*ix * p13m + p13b);
		}
		// when the ball is close and distorted, it is possible to detect 3 points that ALMOST
		// lie in a straight line. this would then create a VERY large circle
		// and throw off heading.
		// restrict the image centre of ball to within one screen width from the centre
		*ix = (*ix < WIDTH / 2) ? MAX(-WIDTH / 2, *ix) : MIN(3 * WIDTH / 2, *ix);
		*iy = (*iy < HEIGHT / 2) ? MAX(-HEIGHT / 2, *iy) : MIN(3 * HEIGHT / 2, *iy);
		double r1 = sqrt((double)
				((*ix - p1x) * (*ix - p1x) + (*iy - p1y) * (*iy - p1y)));
		double r2 = sqrt((double)
				((*ix - p2x) * (*ix - p2x) + (*iy - p2y) * (*iy - p2y)));
		double r3 = sqrt((double)
				((*ix - p3x) * (*ix - p3x) + (*iy - p3y) * (*iy - p3y)));
		*radius = (r1 + r2 + r3) / 3.0;
	}


/* Find center of fireball. */
void VisualCortex::findFireBallCentre(HalfBeacon *preball,
        int *fbX,
        int *fbY,
        double *fbHead,
        double *fbDist,
        double *fbRadius,
        double *fb_obj0,
        double *fb_obj1,
        double *fb_obj2) {

    static const string debugHead = "findFireBallCentre| ";
    int p1x, p1y, p2x, p2y, p3x, p3y;
    int bxmin, bxmax, bymin, bymax;
    bool nl, nr, sl, sr;

    double xmin = preball->getXMin();
    double xmax = preball->getXMax();
    double ymin = preball->getYMin();
    double ymax = preball->getYMax();

    // initialise
    bxmin = (int) xmin;
    bymin = (int) ymin;
    bxmax = (int) xmax;
    bymax = (int) ymax;
    p1x = p2x = p3x = p1y = p2y = p3y = -1;
    nl = isOutside(FRAME, FRAME);
    nr = isOutside(WIDTH - FRAME, FRAME);
    sl = isOutside(FRAME, HEIGHT - FRAME);
    sr = isOutside(WIDTH - FRAME, HEIGHT - FRAME);

    if (flag_OFFLINE && debugFindFireBallCentre) {
        (*outfile) << debugHead
            << "Try to find 1st & 2nd points on circumference.. " << endl;
    }

    findPointOnCircumferenceOneAndTwo(bxmin,
            bxmax,
            bymin,
            bymax,
            nl,
            nr,
            sl,
            sr,
            &p1x,
            &p1y,
            &p2x,
            &p2y);

    if (flag_OFFLINE && debugFindFireBallCentre) {
        (*outfile) << debugHead 
            << "found points .. (" << p1x << "," << p1y << ") (" << p2x
            << "," << p2y << ")" << endl;
    }

    if (p1x < 0 || p2x < 0) {
        *fbRadius = -1; return;
    }
    // else cout << "pt1 : " << p1x << "," << p1y << "  pt2 : " << p2x << "," << p2y << endl;


    if (flag_OFFLINE && debugFindFireBallCentre) {
        (*outfile) << debugHead
            << "try to find 3rd points on circumference.. " << endl;
    }

    findPointOnCircumferenceThree(bxmin,
            bxmax,
            bymin,
            bymax,
            nl,
            nr,
            sl,
            sr,
            p1x,
            p1y,
            p2x,
            p2y,
            &p3x,
            &p3y);

    if (flag_OFFLINE && debugFindFireBallCentre) {
        (*outfile) << debugHead
            << "found points .. (" << p3x << "," << p3y << ")" << endl;
    }

    if (p3x < 0) {
        *fbRadius = -1; return;
    }
    // else cout << "pt3 : " << p3x << "," << p3y << endl;

    if (p1x < 0
            || p2x < 0
            || p3x < 0
            || ((p1x == p3x) && (p2x == p3x))
            || ((p1y == p3y) && (p2y == p3y))) {
        *fbRadius = -1; return;
    }

    int ix, iy;
    double radius = -1;

    if (flag_OFFLINE && debugFindFireBallCentre) {
        (*outfile) << debugHead
            << "try to find intersection of chords " << endl;
    }

    findIntersectionOfTwoChords(p1x, p1y, p2x, p2y, p3x, p3y, &ix, &iy, &radius);

    if (flag_OFFLINE && debugFindFireBallCentre) {
        (*outfile) << debugHead
            << "done intersection of chords " << endl;
    }

    if (radius < 0) {
        *fbRadius = -1; return;
    }

    // if the calculated ball fails to cover the orange blob, then something is wrong
    // it is possible that the orange blob is not covered because part of the orange is noise
    // ie when next to white boundary,
    // but if obviously missing (contain < half of orange blob) then there is error in 
    // detecting edge of ball
    double xdiff = (ix < (bxmax + bxmin) / 2)
        ? ix
        + radius
        - bxmin
        : bxmax
        - (ix - radius);
    double ydiff = (iy < (bymax + bymin) / 2)
        ? iy
        + radius
        - bymin
        : bymax
        - (iy - radius);
    xdiff = MIN(xdiff, bxmax - bxmin);
    ydiff = MIN(ydiff, bymax - bymin);
    double intersectArea = xdiff *ydiff;
    double blobArea = (bymax - bymin) * (bxmax - bxmin);

    if ((intersectArea / blobArea) < 0.9) {

        if (flag_OFFLINE && debugFindFireBallCentre) {
            (*outfile) << debugHead
                << "fireball sanity, blob intersect.. " << intersectArea << "/"
                << blobArea << " = " << (intersectArea / blobArea) << endl;;
        }

        *fbRadius = -1; 
        return;

    } else {

        if (flag_OFFLINE && debugFindFireBallCentre) {
            (*outfile) << debugHead
                << "fireball sanity passed blob intersect.. " << intersectArea
                << "/" << blobArea << endl;;
        }
    }

    // if fireball calculates the ball to be entirely within frame, then it is either unneccesary, or wrong
    if (ix - radius > 0
            && ix + radius < WIDTH
            && iy + radius <HEIGHT
            && iy - radius> 0) {
        *fbRadius = -1;

        if (flag_OFFLINE && debugFindFireBallCentre) {
            (*outfile) << debugHead
                << "fireball sanity, ball not occluded" << endl;
        }
        return;
    }

    // something is wrong if the calculated centre is higher than the top edge of blob
    double centreElev = PointToElevation(ix, iy, sin_eroll, cos_eroll);

    int ox = (int) xmin;
    int oy = (int) ymin;
    int ow = (int) (xmax - xmin);
    int oh = (int) (ymax - ymin);

    int nX, nY, eX, eY, sX, sY, wX, wY;
    int rnX, rnY, reX, reY, rsX, rsY, rwX, rwY;

    nX = xOrigin - (int) (ox + ow / 2);
    nY = yOrigin - (int) (oy);

    rnX = (int) (cos_eroll * nX - sin_eroll * nY);
    rnY = (int) (sin_eroll * nX + cos_eroll * nY);

    sX = xOrigin - (int) (ox + ow / 2);
    sY = yOrigin - (int) (oy + oh);

    rsX = (int) (cos_eroll * sX - sin_eroll * sY);
    rsY = (int) (sin_eroll * sX + cos_eroll * sY);

    eX = xOrigin - (int) (ox);
    eY = yOrigin - (int) (oy + oh / 2);

    reX = (int) (cos_eroll * eX - sin_eroll * eY);
    reY = (int) (sin_eroll * eX + cos_eroll * eY);

    wX = xOrigin - (int) (ox + ow);
    wY = yOrigin - (int) (oy + oh / 2);

    rwX = (int) (cos_eroll * wX - sin_eroll * wY);
    rwY = (int) (sin_eroll * wX + cos_eroll * wY);

    int yRmin = yOrigin - MAX(MAX(MAX(rnY, rsY), reY), rwY);
    double topElev = PointToElevation(yRmin);

    if (centreElev > topElev && bymin > 0) {

        if (flag_OFFLINE && debugFindFireBallCentre) {
            (*outfile) << debugHead
                << "fireball sanity, centre elev " << centreElev
                << " > top Elev " << topElev << endl;;
        }
        *fbRadius = -1; 
        return;
    }

    int inpoints[2] = {
        ((ix - WIDTH / 2)), (-(iy - HEIGHT / 2))
    };
    double outpoints[2] = {
        -1, -1
    };
    projectPoints(inpoints, 1, outpoints, BALL_RADIUS);
    *fbDist = sqrt((double)
            (outpoints[0] * outpoints[0] + outpoints[1] * outpoints[1]));
    //cout << "Distance ball " << *fbDist << endl;

    if (*fbDist == 0) {
        *fbRadius = -1; return;
    }

    if (flag_OFFLINE && debugFindFireBallCentre) {
        (*outfile) << debugHead
            << "fireball found, image centre.. (" << ix << "," << iy << ")"
            << endl;;
    }

    *fbHead = -1.0 * RAD2DEG(asin((1.0 * outpoints[0]) / *fbDist));
    *fbRadius = radius;
    *fbX = ix;
    *fbY = iy;
    *fb_obj0 = outpoints[0];
    *fb_obj2 = outpoints[1];
    *fb_obj1 = NECK_LENGTH;
}    



// This method tries to find ball distance using blob size
// blob size is extended to the largest dimension since ball is round
// ie width = height and extend the smaller to be equal to the greater
// to cater for occlusions.
// This method fails when the ball is in the corner, or in other words
// when both and width and height != diametre of ball.
// at close distances, this error is magnified.
bool VisualCortex::normalFindBall(HalfBeacon *preball, VisualObject &ball) {

	double xmin = preball->getXMin();
	double xmax = preball->getXMax();
	double ymin = preball->getYMin();
	double ymax = preball->getYMax();


	// Added by Ted Wong. Save values before it is change.
	ball.ox = xmin;
	ball.oy = ymin;
	ball.owidth = xmax - xmin;
	ball.oheight = ymax - ymin;
    ball.ocx = (xmax+xmin) /2.0;
    ball.ocy = (ymax+ymin) /2.0;

	double raw_y = ymin;

	// First of all, we expand the rectangle to full size
	// assuming the ball is partially visible.
	// This assumes we can at least see two corners
	// Find the centre of the sides.
	int counter1 = 0;
	int counter2 = 0;
	double cleft = 0;
	double cright = 0;
	double ctop = 0;
	double cbottom = 0;

	for (int y = (int) ymin; y <= (int) ymax; y++) {
		if ((cplane[y * WIDTH + (int) xmin] & ~MAYBE_BIT) == cBALL) {
			cleft += y;
			counter1++;
		}
		if ((cplane[y * WIDTH + (int) xmax] & ~MAYBE_BIT) == cBALL) {
			cright += y;
			counter2++;
		}
	}

	// added by Martin 21/06/2001
	// try get rid of the posibilities of divide by zero
	// but not sure what to sure, so set counter to 1.
	// any other suggestion?
	if (!counter1) {
		counter1 = 1;
	}

	if (!counter2) {
		counter2 = 1;
	}

	cleft /= counter1;
	cright /= counter2;

	counter1 = counter2 = 0;
	for (int x = (int) xmin; x <= (int) xmax; x++) {
		if ((cplane[(int) ymin * WIDTH + x] & ~MAYBE_BIT) == cBALL) {
			ctop += x;
			counter1++;
		}
		if ((cplane[(int) ymax * WIDTH + x] & ~MAYBE_BIT) == cBALL) {
			cbottom += x;
			counter2++;
		}
	}

	if (!counter1) {
		counter1 = 1;
	}

	if (!counter2) {
		counter2 = 1;
	}

	ctop /= counter1;
	cbottom /= counter2;

	// We've now got the centres of the sides.
	// So we expand the box so the centre is in the centre.
	// The basic idea is to make it as big as we can justify

	// Adjust the top or bottom
	// double diff1 = MAX(ymax-cleft, cleft-ymin);
	// double diff2 = MAX(ymax-cright, cright-ymin);
	// double centre_y = (MAX(ymax-cleft, cleft-ymin) > MAX(ymax-cright, cright-ymin)) ? cleft : cright;

	// Adjust the left or right
	// diff1 = MAX(xmax-ctop, ctop-xmin);
	// diff2 = MAX(xmax-cbottom, cbottom-ymin);
	// double centre_x = (MAX(xmax-ctop, ctop-xmin) > MAX(xmax-cbottom, cbottom-ymin)) ? ctop : cbottom;

	// Now finally make it a square
	// This should give us an even better centreing - 'tho not perfect.
	// We should get max(sin(th),cos(th))*rad, where rad is the real
	// radius and th is the angle from the nearest axis of the chord to
	// the edgemost point available. (That sort of idea anyway)
	// Hmm.. so maybe the expansion stuff above is unnecessary?
	// But not, it's good if it's obscured anyway.
	// The justification for some of this is slightly dodgy... :)

	double width = xmax - xmin;
	double height = ymax - ymin;
	double offset;
	int delta = 3;

	if (width > height) {
		if (ymax >= USABLE_HEIGHT - delta)        // object is on the bottom edge of the FOV
			ymax = ymin + width;
		else if (ymin <= delta)        // object is on the top edge of the FOV
			ymin = ymax - width;
		else {
			// object's bbox is not rectangular?
			offset = (width - height) / 2.0;
			ymin -= offset;
			ymax += offset;
		}
		height = width;
	}
	else if (width < height) {
		if (xmax >= WIDTH - delta)        // object is on the right edge of the FOV
			xmax = xmin + height;
		else if (xmin <= delta)        // object is on the left edge of the FOV
			xmin = xmax - height;
		else {
			// object's bbox is not rectangular?
			offset = (height - width) / 2.0;
			xmin -= offset;
			xmax += offset;
		}
		width = height;
	}

	double centre_x = (xmin + xmax) / 2.0;
	double centre_y = (ymax + ymin) / 2.0;

	ball.blob = preball;
	ball.area = preball->area;
	ball.x = xmin;
	ball.y = ymin;
	ball.width = width;
	ball.height = height;
	ball.cx = centre_x;
	ball.cy = centre_y;
	ball.radius = MAX(width, height) / 2;

	ball.cf = ball.cf;
	ball.tilt = 0;
	ball.imgHead = PointToHeading(centre_x);
	ball.imgElev = PointToElevation(centre_y);
	ball.misc = raw_y;

	// Ted: I don't understand why it is 1558.0? Kim please explain?
	// ball.dist = 1200.0 / ball.height;
	ball.dist = 1558.0 / ball.height;
		
	if (ball.dist < 25 && ABS(pan) < 0.5)        //0.5rad = 30degree
	{
		ball.dist = MIN(ball.dist, NewBallDistance());
	}
		
	ball.cam_dist = ball.dist;

	// Ted: This has a serious problem.	
	RobotRelative(ball);

	ball.tilt = 0;

	ball.projected_x = MAX(ball.x, 0);
	ball.projected_y = MAX(ball.y, 0);
	ball.projected_w = ball.width;
	ball.projected_h = ball.height;

	double expected = MIN(pi *SQUARE(ball.radius), HEIGHT *WIDTH);        //25/1/02 used to be width/2
	ball.cf = (int) MIN(ball.area * 1000.0 / expected, 1000);
	// WARNING: not sure how these magic number works
	ball.var = exp(-0.864452 + 0.0215735 * ball.dist) / pow(ball.area
			/ expected, errorScale);
	return true;
}








///////////////////////////////////////////////////////////////////////////////////////////////////////////
// above 40cm, use number of pixels of ball to determine distance
// under 40cm, use projection method
// when using projection, if the ball is occluded
// use fireball, otherwise just project centre of ball.
//////////////////////////////////////////////////////////////////////////////////////////////////////////
void VisualCortex::FindBall() {

    static const string debugHead = "FindBall| ";

	double fbHead, fbDist, fbRadius;
    double fb_obj0, fb_obj1, fb_obj2;
    bool fbFound = false;
    bool screenCovered = false;
    bool nbFound = true;
    int fbX, fbY;

    fbHead = fbDist = fbRadius = fb_obj0 = fb_obj1 = fb_obj2 = fbX = fbY = -1;

	/* Selection sort the base blobs in order, return the largest 6 in order.
    ** */
	int numBalls        = blobs->findHalfBeacons(cBALL, hbBall, kMaxOrangeBlobs);

    /* TODO: When there're a lot of oranges in the background, this may force
    ** the real ball out of our consideration, thus findHalfBeacons might want
    ** to make sure all orange blobs are returned when it is looking for ball. */

    int totalRedArea    = 0;
	int totalOrangeArea = 0;


    /* Find the total ball blobs and red robot blobs area. */
	for (int n = 0; n < numBalls; n++) {
        HalfBeacon *pball = hbBall[n];
        totalOrangeArea += pball->area;
    }
    int nb = blobs->count(cROBOT_RED);
    for (int n = 0; n < nb; n++) {
        totalRedArea += blobs->getBlob(cROBOT_RED, n).area;
    }

	// Look for all the orange blobs.
	for (int ballNum = 0; ballNum < numBalls; ballNum++) {
	
		/* preball is a candidate ball. */
        /* Use it to get low level blob information. */
        HalfBeacon *preball = hbBall[ballNum];
        preball->initBallBlob(&blobs->splane); // Use to init the splane. 

        double xmin = preball->getXMin();
        double xmax = preball->getXMax();
        double ymin = preball->getYMin();
        double ymax = preball->getYMax();
        double blobArea = preball->area;
        double raw_y = ymin;

        if (debugFindBall && flag_OFFLINE) {
            (*outfile) << debugHead 
                << "========= Checking ball #" << ballNum 
                << ", (xmin/ax, ymin/ax, area) " 
                << xmin << " " << xmax << " " << ymin << " " 
                << ymax << ", " << blobArea << endl; 
        }

        /* Check ----------------------------------------------------*/

        // It's not the ball if it's too thin or too flat.
        if (xmax + 1 - xmin <= 4 || ymax + 1 - ymin <= 4) {
            if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead 
                    << "... blob filtered because too thin or too flat." << endl;
            }
            fireSanity(BALL_TOO_THIN_OR_TOO_FLAT);
            continue;
        }

        /* Check ----------------------------------------------------*/

        /* Quick check - can't be a ball if not occupying enough orange. */
        double boundingBoxArea = (xmax-xmin + 1) * (ymax-ymin + 1);
        double BALL_ORANGE_RATIO_THRESHOLD = 0.4;

        if (blobArea > 4000) {
            BALL_ORANGE_RATIO_THRESHOLD = 0.3;
        }

        if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead 
                << "    Checking ball blobArea/boundingBoxArea..." << endl
                << debugHead << "    - ballBlob   : " << blobArea << endl 
                << debugHead << "    - boundingBox: " << boundingBoxArea << endl 
                << debugHead << "    - density    : "
                    << blobArea/boundingBoxArea << endl; 
        }
 
        if ((blobArea / boundingBoxArea) < BALL_ORANGE_RATIO_THRESHOLD) {
            if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead 
                    << "... blob filtered because of too little orange." << endl;
            }
            fireSanity(TOO_LITTLE_ORANGE_IN_BALL);
            continue;
        }


        /* Check ----------------------------------------------------*/

        /* Check if any corner of the ball is higher than the horizon. */
        int BALL_OVER_HORIZON_THRESHOLD = 30; 

        /* For very large ball, the horizon may be just touching the ball
        ** edge, so relax the threshold a bit. */
        if (blobArea > 8000) {
            BALL_OVER_HORIZON_THRESHOLD = 50; 
        }


        if (!isBelowHorizon   ((int) xmin, (int) ymin, BALL_OVER_HORIZON_THRESHOLD) 
            || !isBelowHorizon((int) xmax, (int) ymin, BALL_OVER_HORIZON_THRESHOLD)
            || !isBelowHorizon((int) xmin, (int) ymax, BALL_OVER_HORIZON_THRESHOLD)
            || !isBelowHorizon((int) xmax, (int) ymax, BALL_OVER_HORIZON_THRESHOLD)
            ) {

            if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead 
                    << "... blob filtered because above horizon." 
                    << endl;
            }
            fireSanity(BALL_ABOVE_HORIZON);
            continue;
        }
         
        /* Check if the cplane has a fireball. */ 
        if (preball->area > AREA_TRIGGER * 8) { // Big ball.
            /*
        if ((RAD2DEG(effective_tilt) > TILT_TRIGGER && 
                    preball->area > AREA_TRIGGER)        // Ball is close. 
                || (preball->area > AREA_TRIGGER * 8)) { // Big ball.
                */

            if (flag_OFFLINE && debugFindBall) {
                (*outfile) << debugHead 
                    << "Total orange area " << totalOrangeArea 
                    << "+ total red area " << totalRedArea 
                    << " = " << totalOrangeArea + totalRedArea
                    << " (out of shrinked cplane area " 
                    << (WIDTH - 10) * (HEIGHT - 10) << ")"<< endl;
            }

            findFireBallCentre(preball, &fbX, &fbY, &fbHead, &fbDist, &fbRadius, &fb_obj0, &fb_obj1, &fb_obj2);

            fbFound = (fbRadius > 0); // whether a fireball is found

            /* Check ------------------------------------------------------- */
            double maxDimension = MAX(preball->getXMax() - preball->getXMin(),
                                    preball->getYMax() - preball->getYMin());

            /* If the radius of FireBall found is shorter than the real 
             * dimension of orange blob, then something is wrong, most likely 
             * the edges of ball were detected falsely due to noise around ball. 
             * */
            if (2 * fbRadius < 0.9 * maxDimension) {
                fbFound = false;

                if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                    (*outfile) << debugHead  
                        << "... diametre not right."
                        << endl;
                }
            }


            // Check whether the ball covers the entire screen.
            // however cant check for full orange because when ball close 
            // enough there'll be alot of red as well on the other hand, it's 
            // possible that you are looking straight into red dog, 
            // so if ball is reasonable size, but rest of screen is covered 
            // by red dog, then fire.

            // WARNING: 21000 is the constant for old dogs.
            // Using the same ratio (21000/176*144), new camera should 
            // use 27575.

            /* Check ------------------------------------------------------- */

            /* Now already said no fireball, but check again if the ball 
             * actually covered the entire screen, foundBall=true again if yes. */
            static const int BALL_ON_ENTIRE_SCREEN_CONST = 27575;

            if (totalOrangeArea + totalRedArea >= BALL_ON_ENTIRE_SCREEN_CONST && !fbFound) {

                int inpoints[2] = {0, 0};
                double outpoints[2] = {-1, -1};
                
                projectPoints(inpoints, 1, outpoints, BALL_RADIUS);
                fbDist = sqrt((double) (outpoints[0] * outpoints[0] + outpoints[1] * outpoints[1]));
                fbHead = -1.0 * RAD2DEG(asin((1.0 * outpoints[0]) / fbDist));
                fbRadius = max(WIDTH / 2, HEIGHT / 2);
                fbX = WIDTH / 2;
                fbY = HEIGHT / 2;
                fb_obj0 = outpoints[0];
                fb_obj2 = outpoints[1];
                fb_obj1 = NECK_LENGTH; 
                fbFound = true;
                screenCovered = true;

                if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                    (*outfile) << debugHead 
                        << ".. fireball screen covered with ball, fbFound = true"
                        << endl;
                }

            } else {

                if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                    (*outfile) << debugHead
                        << "... fireball not cover screen, area "
                        << totalOrangeArea + totalRedArea 
                        << ", fbFound = " << fbFound
                        << endl;
                }
            }
                
            /* Check ----------------------------------------------------*/

            // If fireball returns big distance, then dont trust it.
            if (fbDist > 30) {
                fbFound = false;
                if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                    (*outfile) << debugHead
                        << "... fireball dist > 30, fbFound = false" << endl;
                }
            }

        }

        nbFound = normalFindBall(preball, vob[vobBall]);

        const int PROJECTION_TRIGGER = 40;
        //const int NECK_HEIGHT = 7;
        int pxmin = (int) MAX(xmin, 0);
        int pxmax = (int) MIN(xmax, WIDTH);
        int pymin = (int) MAX(ymin, 0);
        int pymax = (int) MIN(ymax, HEIGHT);
        VisualObject &ball = vob[vobBall];

        if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {

            if (fbFound || nbFound) {
                (*outfile) << debugHead
                    << "Info: Before BallSanities," 
                    << " fbFound? " << (fbFound ? "True" : "False")
                    << " nbFound? " << (nbFound ? "True" : "False")
                    << endl;
            }
        }

        // Perform sanity checks on ball, if false is returned then it can't be a ball.
        // cover is true if the ball covers the entire screen.
        if (screenCovered || BallSanities(ball, pxmin, pxmax, pymin, pymax, xmin, xmax, ymin, ymax)) {

            int    inpoints[2] = {((int) (ball.cx) - WIDTH / 2), -1 * ((int) (ball.cy) - HEIGHT / 2)};
            double outpoints[2] = {-1, -1};

            projectPoints(inpoints, 1, outpoints, BALL_RADIUS);

            double currentRDist = sqrt((double) (outpoints[0] * outpoints[0] + outpoints[1] * outpoints[1]));
            // cout << "Distance ball 3 " << currentRDist << endl;
            double currentRHead = -1.0 * RAD2DEG(asin((1.0 * outpoints[0]) / currentRDist));

            /* Check ----------------------------------------------------*/

            // If ball is small but vision says it's very close then likely it 
            // is reflection from looking down at other robot
            if (currentRDist < 50 && ball.area < 50 && ball.radius <= 5) {

                if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {

                    (*outfile) << debugHead 
                        << "... small ball with close distance, impossible." 
                        << endl;
                }
                ball.reset();
                continue;
            }

            // If estimate distance to be < 40 use fireBall or projection.
            if (currentRDist < PROJECTION_TRIGGER) {

                /* Use fireBall method to calculate ball details. */
                if (fbFound && 
                        (   ball.x <= 0 || (ball.x + ball.radius * 2) >= WIDTH - 1
                         || ball.y <= 0 || (ball.y + ball.radius * 2) >= HEIGHT - 1)
                   ) {
                    ball.radius = fbRadius;
                    ball.cx = fbX;
                    ball.cy = fbY;
                    ball.width = fbRadius * 2;
                    ball.height = fbRadius * 2;
                    ball.x = fbX - fbRadius;
                    ball.y = fbY - fbRadius;
                    ball.area = preball->area;

                    xmin = ball.x;
                    xmax = ball.x + ball.width;
                    ymin = ball.y;
                    ymax = ball.y + ball.height;

                    // calculating confidence
                    double expected = MIN(pi *SQUARE(ball.radius), HEIGHT *WIDTH);        //25/1/02 used to be width/2
                    // ball.cf = (int) MIN (ball.area * 1000.0 / expected, 1000);
                    ball.cf = 500;

                    // apply clippings
                    int pxmin = (int) MAX(xmin, 0);
                    int pxmax = (int) MIN(xmax, WIDTH);
                    int pymin = (int) MAX(ymin, 0);
                    int pymax = (int) MIN(ymax, HEIGHT);
                    ball.projected_x = pxmin;
                    ball.projected_y = pymin;
                    ball.projected_w = pxmax - pxmin;
                    ball.projected_h = pymax - pymin;

                    ball.tilt = 0;
                    ball.imgHead = PointToHeading(fbX);
                    ball.imgElev = PointToElevation(fbY);

                    ball.cam_dist = sqrt(fbDist * fbDist + NECK_LENGTH * NECK_LENGTH)
                            - FACE_LENGTH;
                    ball.dist = sqrt(fbDist * fbDist + NECK_LENGTH * NECK_LENGTH);
                    ball.d = fbDist;
                    ball.h = ball.head = fbHead;
                    ball.objectPosition[0] = fb_obj0;
                    ball.objectPosition[1] = fb_obj1;
                    ball.objectPosition[2] = fb_obj2;
                    ball.var = exp(-0.864452 + 0.0215735 * ball.dist) / pow(ball.area
                            / expected, errorScale);

                    if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                        if (screenCovered) {
                            (*outfile) << debugHead
                                << "Info: ballDistMethod = FIREBALL_PROJECTION."
                                << endl;
                        }
                    }

                    ballDistMethod = FIREBALL_PROJECTION;

                /* Use pixel count projection method to calculate ball details. */
                } else {

                    ball.cam_dist = sqrt(currentRDist * currentRDist
                            + NECK_LENGTH * NECK_LENGTH) - FACE_LENGTH;
                    ball.dist = sqrt(currentRDist * currentRDist
                            + NECK_LENGTH * NECK_LENGTH);
                    ball.d = currentRDist;
                    ball.h = ball.head = currentRHead;

                    if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                        (*outfile) << debugHead
                            << "Info: ballDistMethod = PIXEL_COUNT_PROJECTION."
                            << endl;
                    }
            
                    ballDistMethod = PIXEL_COUNT_PROJECTION;
                }

            /* Use pixel count distance method. */
            } else {

                if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                    (*outfile) << debugHead
                        << "Info: ballDistMethod = PIXEL_COUNT_DISTANCE."
                        << endl;
                }
        
                ballDistMethod = PIXEL_COUNT_DISTANCE;
            }



            // SANITY CHECK --------------------------------------------------

            /* Final desparate distance check - don't care about balls 3.5m
            ** away, this helps throwing out lots of small balls other sanity
            ** cannot throw out. */

            // Kim: man, this is the worse sanity check ever seen. Instead of 
            // sanity out the ball on red dog, you throws out ball that's looks 
            // like ball on red dog.
            
            // Dan: Ok ok - I change to 350. After long discussion,
            // Kim is not convinced that there're nice orange blobs in robot
            // patches that can only be sanity out by this check, and I'm not
            // convinced that colour can be gauranteed to be good. Since I
            // happened to be the vision guy here, let me change this again.
            // ;]

#ifdef ONLY_ONE_DOG
             if (ball.dist > 400) 
#else             
             if (ball.dist > 350) 
#endif //ONLY_ONE_DOG                 
             {

                fireSanity(BALL_TOO_FAR);

                if (debugFindBall && flag_OFFLINE) {

                    (*outfile) << debugHead
                        << "...ball's dist (" << ball.dist << ") is over 3m,"
                        << " assume noise from red dog!"
                        << endl;
                }
                ball.reset();
                continue;
            }


            // SANITY CHECK ----------------------------------------------------

            // New height check using elevation angle from camera. 
            // 200 - it will miss ball sometimes, due to bad body tilt,
            // but I don't want to increase it to risk seeing flying ball.
            static const double BALL_HEIGHT_THRESHOLD = 250;

            // ballHeight is relative to the neckBase.
            double ballHeight = getHeightKnowingDistance(degPan, degTilt + degBodyTilt, degCrane,
                                    TOPLEFT2CENTER_X(ball.cx),
                                    TOPLEFT2CENTER_Y(ball.cy),
                                    MIN(ball.dist, 150) * 10, 
                                    FACE_LENGTH * 10 , 
                                    NECK_LENGTH * 10, 
                                    APERTURE_DISTANCE, 
                                    CAMERA_OFFSET);

            #define detail_formed_ball_height_check
            #ifdef detail_formed_ball_height_check
            if (debugFindBall && flag_OFFLINE) { 

                double camElev = getElevation(degPan, degTilt + degBodyTilt, degCrane,
                                        TOPLEFT2CENTER_X(ball.cx),
                                        TOPLEFT2CENTER_Y(ball.cy), 
                                        FACE_LENGTH * 10, 
                                        NECK_LENGTH * 10, 
                                        APERTURE_DISTANCE, 
                                        CAMERA_OFFSET);

                (*outfile) << debugHead
                    << "---Detail ball height check info: " << endl
                    << debugHead << "    cx, cy        : " 
                        << ball.cx << ", "
                        << ball.cy << endl
                    << debugHead << "    dist          : " << ball.dist * 10 << endl
                    << debugHead << "    camElev       : " << camElev << endl
                    << debugHead << "    ball's height : " 
                        << neckBaseHeight + ballHeight
                        << " (mm) " << endl
                        << endl;
            }
            #endif // detail_formed_ball_height_check

            if (ballHeight + neckBaseHeight > BALL_HEIGHT_THRESHOLD){

                if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                    (*outfile) << debugHead
                        << "... ball height > 25cm"
                        << endl << endl;
                }
                ball.reset();
                fireSanity(BALL_HEIGHT_TOO_LARGE);
                continue;
            }


            // SANITY CHECK --------------------------------------------------

            /* Final elevation check. */
            /* There is an error in elev produced by RobotRelative - used
            ** height check from elev of camera instead. */
            /*
            RobotRelative(ball);
            double ballHeight = tan(DEG2RAD(ball.elev)) * ball.dist;

             if (ballHeight > 5) {

                if (debugFindBall && flag_OFFLINE) {

                    (*outfile) << debugHead
                        << "...ball's height (" << ballHeight << ") is too large!"
                        << endl;
                    (*outfile) 
                        << debugHead << "   imgHead: " << ball.imgHead << endl
                        << debugHead << "   imgElev: " << ball.imgElev << endl
                        << debugHead << "   dist   : " << ball.dist << endl
                        << debugHead << "   elev   : " << ball.elev << endl;
                }
                ball.reset();
                continue;
            }
            */


            if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {

                if (ball.cf > 0) {
                    (*outfile) << debugHead
                        << "... ball found! " << endl;
                }
            }


            // SANITY CHECK --------------------------------------------------

            /* Now ball confirmed, remove false beacons near the corners of 
               cplane and edge of ball.*/
            static const int BIG_BALL_AREA_FOR_FALSE_BEACONS =
                                           CPLANE_WIDTH*CPLANE_HEIGHT*2/5;

            if (ball.area > BIG_BALL_AREA_FOR_FALSE_BEACONS) {
                for (int beaconIndex = vobBlueLeftBeacon; 
                        beaconIndex < vobYellowRightBeacon;
                        beaconIndex++) {

                    static const int FALSE_BEACON_BORDER = 40; 
                    if ((vob[beaconIndex].x < FALSE_BEACON_BORDER &&
                         vob[beaconIndex].y < FALSE_BEACON_BORDER) 
                     || (vob[beaconIndex].x < FALSE_BEACON_BORDER &&
                         vob[beaconIndex].y > CPLANE_HEIGHT - FALSE_BEACON_BORDER) 
                     || (vob[beaconIndex].x > (CPLANE_WIDTH - FALSE_BEACON_BORDER) &&
                         vob[beaconIndex].y < FALSE_BEACON_BORDER)
                     || (vob[beaconIndex].x > CPLANE_WIDTH - FALSE_BEACON_BORDER &&
                         vob[beaconIndex].y > CPLANE_HEIGHT - FALSE_BEACON_BORDER)
                        ) {

                        vob[beaconIndex].reset();
                        fireSanity(BEACON_ACTUALLY_PART_OF_BALL);

                        if (debugFindBall && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                            (*outfile) << debugHead
                                << "...beacon is actually part of the ball."
                                << endl;
                        }
                    }
                }
            }
           
            /* Extra details independent of calculation method. */
            ball.misc = raw_y;

            // temporary fix
            ball.cf = (ball.var >= LARGE_VAL) ? 3 : ball.cf;

            // smallest dist from base of neck to ball is ball_radius
            ball.d = (ball.d <= 0) ? BALL_RADIUS : ball.d;

            // Ball variances are affected by the head movement speed
            ball.var += get95CF(ball.d * headMovement);
            ball.angleVar = get95CF(PAN_95CF +  RAD2DEG(headMovement));
            ball.var = (ball.var >= LARGE_VAL) ? LARGE_VAL - 1 : ball.var;

            /* TODO: zoidalVarFactor and zoidalAngVarFactor needs to be
            ** updated. */

            /* Adjust ball's variace base on the current walk type. */
            if (currentWalkType && (((*currentWalkType) == ZoidalWalkWT) || ((*currentWalkType) == OffsetWalkWT))) {
                    vob[cBALL].var *= zoidalVarFactor;
                    vob[cBALL].angleVar *= zoidalAngVarFactor;
            }
            
            return;
            
        } else {
            continue;
        }
    }
        
   	lastd = 0;
   	vob[cBALL].reset();
}


// Added by Ted Wong. This function check if a orange blob is a ball by comparing
// it against any visable beacon. Basically check if the orange blob is a beacon noise.
//What? sanity check by "Blobber" guy? 
//This is very funny, do you know how many pixel that 4 degree elevation is?
//is disabled by Kim.
bool VisualCortex::isBallBeaconNoise(VisualObject &ball) {
	for (int id = vobBlueLeftBeacon; id <= vobYellowRightBeacon; id++) {
		// Ted: This cf may need to change. Not sure a correct value yet.
		if (vob[id].cf > 100) {
			// If the ball and beacon's heading are very close.
			if (ABS(ball.head - vob[id].head) <= 4) {
				// If the ball and beacons's elev is also very close.
				if (ABS(ball.elev - vob[id].elev) <= 5) {
					return true;
				}
			}
		}
	}
	return false;		
}

/* pxmin, pymin, pxmax, pymax are the bounded values inside cplanes, ie. no
** negatives, no > CPLANE_HEIGHT/WIDTH of the ball bounding box.
*/
bool VisualCortex::BallSanities(VisualObject &ball, int pxmin, int pxmax, 
        int pymin, int pymax, double xmin, double xmax, double ymin,
        double ymax) {

    static const string debugHead = "BallSanities| ";

    /* Some horizon check by Raymond in 2003 is removed here since rev 2726 
       because I already had some earlier in FindBall. */


    VisualObject dummyBall;
    dummyBall.x      = pxmin;
    dummyBall.y      = pymin;
    dummyBall.width  = pxmax - pxmin; 
    dummyBall.height = pymax - pymin; 

	double  ballElev            = getMinElevation(dummyBall);

    // SANITY CHECK ----------------------------------------------------------

	// Check elev of ball against that of goals.
    for (int elevPassedCounter = vobBlueGoal; elevPassedCounter <= vobYellowGoal; elevPassedCounter++)  {

        if (vob[elevPassedCounter].cf > 100) 
        {
            // goal cy is already rotated, just compare the imgElev.
            double goalElev = PointToElevation (vob[elevPassedCounter].cy);

            static const int BALL_OVER_GOAL_ELEV_THRESHOLD = 2;
            if (ballElev > goalElev + BALL_OVER_GOAL_ELEV_THRESHOLD ) {
                if (debugBallSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                    (*outfile) 
                        << debugHead << "...BALL is higher than goal!" << endl
                        << debugHead << "   -- ballElev = " << ballElev << endl 
                        << debugHead << "   -- goalElev = " << goalElev << endl;
                }
                fireSanity(BALL_ELEVATION_GREATER_THAN_GOAL_ELEVATION 
                        + elevPassedCounter - vobBlueGoal); 
                return false;
            }
        }
    }


    // SANITY CHECK ----------------------------------------------------------

	// Check elev of ball against that of beacons.
    // beacon vob IDs starts from left top to right bottom.
	for (int elevPassedCounter = vobBlueLeftBeacon; elevPassedCounter <= vobYellowRightBeacon; elevPassedCounter++) {
        
		if (vob[elevPassedCounter].cf > 100) { 

            double beaconElev = getMinElevation(vob[elevPassedCounter]);
            // if ( ball.elev > vob[k].elev )
            static const int BALL_OVER_BEACON_ELEV_THRESHOLD = 1;
            if (ballElev > beaconElev + BALL_OVER_BEACON_ELEV_THRESHOLD) {
                if (debugBallSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                    (*outfile) << debugHead
                        << "...BALL is higher than beacon!"<< endl;
                }
                fireSanity(BALL_ELEVATION_GREATER_THAN_BEACON_ELEVATION 
                        + elevPassedCounter - vobBlueLeftBeacon);
                return false;
            }
        }
	}


    // SANITY CHECK ----------------------------------------------------------

    // Check if the orange blob is really a beacon noise. May be very useful 
    // for 2004 color classification. Of couse this function only work if the 
    // beacon could be identified.
    /*
	if (isBallBeaconNoise(ball)) {

        if (flag_OFFLINE && debugBallSanities) {
            (*outfile) << debugHead
                << "...BALL is a beacon noise!"<< endl;
        }
		return false;
    }
    */


    // SANITY CHECK ----------------------------------------------------------

    // Check if the ball is scattered and in edges, but don't check for large
    // orange blob.
    int     boundingBoxArea = (pxmax- pxmin) * (pymax - pymin); 
    double  density         = (double) ball.area / (double) boundingBoxArea;

    if (ball.area < 600
            && density < 0.5 
            && 
               (   (   ball.ocx > WIDTH - 20
                    && ball.ox  > WIDTH - 35)
                ||  (  ball.ocx < 20
                    && ball.ox  < 35)
               )
            ) {

        fireSanity(BALL_DENSITY_SMALL_AND_IN_EDGE);
        if (flag_OFFLINE && debugBallSanities) {
            (*outfile) << debugHead
                << "...BALL's density < 0.6 and in edge - robot noise!"
                << endl;
        }
		return false;
    }


    // SANITY CHECK ----------------------------------------------------------

    // Number of red pixels near the projected bounding box should be low.
    if (ball.area <= 150) {

        int numRed      = 0;
        int numYellow   = 0;
        int numField    = 0;
        int starty  = pymin - ((int) (ball.projected_h / 2.0) + 1);
        int endy    = pymax + ((int) (ball.projected_h / 2.0) + 1);
        int startx  = pxmin - ((int) (ball.projected_w / 2.0) + 1);
        int endx    = pxmax + ((int) (ball.projected_w / 2.0) + 1);

        if (starty < 0)
            starty = 0;
        if (endy >= HEIGHT)
            endy = HEIGHT - 1;
        if (startx < 0)
                startx = 0;
        if (endx >= WIDTH)
                endx = WIDTH - 1;

        // Count how many red, yellow and green pixels.
        for (int i = starty; i <= endy; i++) {
            for (int j = startx; j <= endx; j++) {
                if ((cplane[i * WIDTH + j] & ~MAYBE_BIT) == cROBOT_RED)
                    numRed++;
                if ((cplane[i * WIDTH + j] & ~MAYBE_BIT) == cBEACON_YELLOW)
                    numYellow++;
                if ((cplane[i * WIDTH + j] & ~MAYBE_BIT) == cFIELD_GREEN)
                        numField++;
            }
        }

		// Too many red pixels.
        if (numRed > ball.area + numField) {

            if (debugBallSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                if (flag_OFFLINE && debugBallSanities) {
                    (*outfile) << debugHead
                        << "...too much red in ball. " << endl;
                }
            }
            fireSanity(TOO_MUCH_RED_IN_BALL);
            return false;

        } 

        // To filter out ball formed btwn red goalie and yellow goal, toned 
        // down slightly to check green field around perimetre. If green field
        // exists, then probably a ball.
        if (numRed + numYellow > numField + ball.area) {
            if (debugBallSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead
                    << "...too much yellow in ball. " << endl;
            }
           fireSanity(TOO_MUCH_YELLOW_RED_IN_BALL);
           return false;
        } 
    }


    // SANITY CHECK ----------------------------------------------------------

    // Check if the bounding box of ball is completely inside the largest red
    // blob.
    if (ball.area <= 1000) {

        int redxmax = 0, redymax = 0;
        int redxmin = 511, redymin = 511;

        // Get the largest red blob.
        for (int i = 0; i < blobs->count(cROBOT_RED); i++) {
            BlobInfo &info = blobs->getBlob(cROBOT_RED, i);
            if (info.area > 20) {
                redxmax = MAX(redxmax, info.getXMax());
                redymax = MAX(redymax, info.getYMax());
                redxmin = MIN(redxmin, info.getXMin());
                redymin = MIN(redymin, info.getYMin());
            }
        }
        int redArea = (redxmax-redxmin)*(redymax-redymin);
    
        int bxmin = (int) MAX(0, ball.x);
        int bxmax = (int) MIN(WIDTH - 1, ball.x + ball.width);
        int bymin = (int) MAX(0, ball.y);
        int bymax = (int) MIN(HEIGHT - 1, ball.y + ball.height);

        static const double BALL_IN_ROBOT_THRESHOLD = 0.2;
       
        if (xmax <= redxmax 
                && ymax <= redymax 
                && xmin >= redxmin 
                && ymin >= redymin 
                && NoGreenNearBy(bxmin, bxmax, bymin, bymax)
                && (ball.area / redArea) < BALL_IN_ROBOT_THRESHOLD ) {

            if (debugBallSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                 (*outfile) << debugHead
                     << "...ball completely inside a large red blob bounding box." 
                     << endl;
            }

            fireSanity(BALL_BOUNDING_BOX_INSIDE_RED_ROBOT_BOUNDING_BOX);
            return false;
        }
    }


    // SANITY CHECK ----------------------------------------------------------

    // If ball is small and it touches the red blob -> unlikely it's a ball.
    if (ball.area <= 40) {

        int redBlobPassed = 0;

        for (; redBlobPassed < blobs->count(cROBOT_RED); redBlobPassed++) {
            BlobInfo &info = blobs->getBlob(cROBOT_RED, redBlobPassed);
            if (info.area > 20) {
                if (info.getXMax() >= xmax
                        && info.getXMin() <= xmin
                        && info.getYMin() >= ymin - ball.radius) {

                    break;
                }
            }
        }

        if (redBlobPassed < blobs->count(cROBOT_RED)) {
            fireSanity(BALL_TOUCHES_RED_BLOB);
            if (debugBallSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead
                     << "...ball is small and touches red blob." << endl;
            }
            return false;
        }
    }


    // SANITY CHECK ----------------------------------------------------------

    if (ball.area <= 50 && InYellowNoGreen(ball)) {
        // eout << "Sanity G" << endl;
        // wout << "Sanity G\n";
        fireSanity(BALL_INSIDE_YELLOW_GOAL);
        if (debugBallSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead
                     << "...ball is small and inside yellow, no green." << endl;
        }
        return false;
    }


    // SANITY CHECK ----------------------------------------------------------

    // too far away outside the field, cant be the ball
    if (ball.dist > FIELD_DIAGONAL) {
        // eout << "Sanity H" << endl;
        // wout << "Sanity H\n";
        fireSanity(BALL_OUTSIDE_FIELD);

        if (debugBallSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead
                 << "...ball is too far - outside field distance." << endl;
        }
        return false;
    }


    // SANITY CHECK ----------------------------------------------------------

	// This will get rid of the problem of mis-recognise the
    // edge of a red robot (inside the yellow goal) as a ball.
    // The number of orange pixels must occupy
    // more than a certain percentage of the area of
    // projected bounding box.
    double pi = 3.14159;
    double occupied_area = (long) (ball.radius *ball.radius *pi);

	if (occupied_area == 0) {
        if (debugBallSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead
                 << "...ball's area calculated using radius is 0." << endl;
        }
        fireSanity(BALL_OCCUPIED_AREA_0);
        return false;

    } else if ((ball.area < 100) && (((double) ball.area / occupied_area) < 0.5)) {

        if (debugBallSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead
                 << "...ball (area < 100) has too little orange pixels in it."
                 << endl;
        }
        return false;

	} else if ((ball.area < 150) && (((double) ball.area / occupied_area) < 0.3)) {

        if (debugBallSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead
                 << "...ball (area < 150) has too little orange pixels in it."
                 << endl;
        }
    
        fireSanity(BALL_AREA_LT150_NOT_OCCUPY_MT25_PROJECTED_BOUNDING_BOX);
        return false;

	} 


    // SANITY CHECK ----------------------------------------------------------

	// If yellow goal is inside ball, reject the goal.
    VisualObject &yGoal = vob[vobYellowGoal];

    int buffer = (int) ((ball.width + ball.height)/ 10);

    /* Unrotate goal x/y about center of cplane. */
	int uYGoalX = xOrigin - (int) (  ucos_eroll*(xOrigin - yGoal.x) 
                                   - usin_eroll*(yOrigin - yGoal.y));

 	int uYGoalY = yOrigin - (int) (  usin_eroll*(xOrigin - yGoal.x) 
                                   + ucos_eroll*(yOrigin - yGoal.y));

  	if ((ball.x <= uYGoalX + buffer) 
            && (ball.x + ball.width >= uYGoalX + yGoal.width - buffer)
            && (ball.y <= uYGoalY + buffer) 
            && (ball.y + ball.height >= uYGoalY + yGoal.height - buffer)) {

        yGoal.cf = 0;
        yGoal.var = LARGE_VAL;
        fireSanity(YELLOW_GOAL_INSIDE_BALL);

        if (debugBallSanities && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead
                 << "...goal in ball, doesn't make sense." << endl;
        }
    }

	return true;
}


/* Return the minimum image elevation of this visual object. */
double VisualCortex::getMinElevation(VisualObject &obj) {

    double elev1 = PointToElevation(obj.x,            obj.y,            sin_eroll, cos_eroll);
    double elev2 = PointToElevation(obj.x+ obj.width, obj.y,            sin_eroll, cos_eroll);
    double elev3 = PointToElevation(obj.x,            obj.y+obj.height, sin_eroll, cos_eroll);
    double elev4 = PointToElevation(obj.x+ obj.width, obj.y+obj.height, sin_eroll, cos_eroll);

    return MIN(elev1,
                MIN(elev2,
                    MIN(elev3, elev4)));
}


// If orange blob index is inside yellow goal and there is no green
// of the field nearby, it is not the ball.
bool VisualCortex::InYellowNoGreen(VisualObject &ball) {
	if (ball.x < vob[vobYellowGoal].x
	    || ball.y <vob[vobYellowGoal].y
	    || ball.x + ball.width  > vob[vobYellowGoal].x + vob[vobYellowGoal].width
	    || ball.y + ball.height > vob[vobYellowGoal].y + vob[vobYellowGoal].height) {
	    return false;
	}
	
	// Make bounding box bigger to check the green color.
	return NoGreenNearBy((int) ball.x, (int) (ball.x + ball.width),
			     (int) ball.y, (int) (ball.y + ball.height));
}

bool VisualCortex::NoColourXNearBy(int colour, int xmin, int xmax, int ymin, int ymax,
				   int delta_x, int delta_y, int threshHold) {
	// check if there is any green near this box
	int x, y, num;
	//int min_delta = 7;
	//int delta_x = MAX(min_delta, (xmax-xmin)/3);
	//int delta_y = MAX(min_delta, (ymax-ymin)/3);

	xmin -= delta_x;
	if (xmin < 0) {
		xmin = 0;
	}

	ymin -= delta_y;
	if (ymin < 0) {
		ymin = 0;
	}

	xmax += delta_x;
	if (xmax > WIDTH) {
		xmax = WIDTH;
	}

	ymax += delta_y;
	if (ymax > USABLE_HEIGHT) {
		ymax = USABLE_HEIGHT;
	}

	// so what if its at the bottom of the image ?
	/*
	   if (ymax==USABLE_HEIGHT)
	   return false; // at the bottom of the image
	   */

	//threshold = ((xmax-xmin)*(ymax-ymin))/(3*3*5);
	//if (threshold>30) threshold = 30;

	num = 0;
	for (y = ymax; y >= ymin; y--) {
		for (x = xmin; x <= xmax; x++) {
			if ((cplane[y * WIDTH + x] & ~MAYBE_BIT) == colour) {
				num++;

				if (num >= threshHold)
					return false;
			}
		}
	}

	return true;
}


/* Look for goal using the given blobs array. Also given
   beacons in left/right of it for preliminary sanity check. */
void VisualCortex::LookForColourGoal(
        HalfBeacon **hbArray, 
        int hbCount, 
        int beaconLeftOfGoalIndex,
        int beaconRightOfGoalIndex,
        int goalIndex
    ) {

	const string debugHead = "LookForColourGoal| "; 

    VisualObject *goal               = &vob[goalIndex];
    VisualObject *beaconLeftOfGoal   = &vob[beaconLeftOfGoalIndex]; 
    VisualObject *beaconRightOfGoal  = &vob[beaconRightOfGoalIndex];


#ifndef ONLY_ONE_DOG
    // SANITY CHECK ------------------------------------------------------

    /* If the dog is looking really down or really up, there should not be 
       any goal. */
    if (!hzExists && goalIndex == vobYellowGoal) {

        if (debugLookForColourGoal && flag_OFFLINE) {
            (*outfile) << debugHead 
                << "... looking straight down, no yellow goal could be observed!"
                << endl;
        }
        return;
    }
#endif //ONLY_ONE_DOG


    /*----------------------------------------------------------------------*/
	/* Combine all the unmatched half_beacons into 1 big one. */
    /*----------------------------------------------------------------------*/

    /* Init a dummy final goal blob, its xmin/max, ymin/max is supposed to be
    ** MAX(original max, rotated max). */
	HalfBeacon goal_blob;
	goal_blob.setXMin(100000);
	goal_blob.setYMin(100000);
	goal_blob.setXMax(0);
	goal_blob.setYMax(0);
	goal_blob.area = 0;

	/* Also keep track of unrotated and rotated info. */
	int uxmin = 100000;
	int uymin = 100000;
	int uxmax = 0;
	int uymax = 0;    

	int xRmin = 100000;
	int yRmin = 100000;
	int xRmax = 0;
	int yRmax = 0;    


    bool multipleBlobJoined = false;

    /* Go through all the blue/yellow blobs to form a goal. */
	for (int b = 0; b < hbCount; b++) {

        HalfBeacon& goalBlob = *hbArray[b];


#ifdef ONLY_ONE_DOG
//hacked, if goal blob is occupying the whole screen, must be inside the goal.
        if ( goalBlob.getXMax() - goalBlob.getXMin() >= 207 &&
             goalBlob.getYMax() - goalBlob.getYMin() >= 158 ){
            goal->x = 0;
            goal->y = 0;
            goal->width = WIDTH;
            goal->height = HEIGHT;

            goal->cx = WIDTH / 2.0;
            goal->cy = HEIGHT / 2.0;

            goal->projected_x = 0;
            goal->projected_y = 0;
            goal->projected_w = WIDTH;
            goal->projected_h = HEIGHT;

            goal->area = goalBlob.area;
            goal->tilt = 0;
            goal->misc = 0;
            goal->dist = 5;
            goal->d = 5;
            goal->cf = 1000;
            goal->var = 10;
            goal->angleVar = 10;
            return;
        }

#endif //ONLY_ONE_DOG

        // SANITY CHECK ------------------------------------------------------

		if (goalBlob.matched) {

			//wout << "Sanity C\n";
			fireSanity(BLOB_ALREADY_MATCHED);
			continue;
		}

		if (debugLookForColourGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << endl << debugHead 
                << "Checking goal blob #" << goalBlob.blob_number
                << " with cx, cy, area: " 
                << goalBlob.cx << ", "
                << goalBlob.cy << ", "
                << goalBlob.area
                << endl;
        }


        // SANITY CHECK ------------------------------------------------------

        /* If the blob is too small, forget it. */
		if (goalBlob.area < 30) {

            if (debugLookForColourGoal && flag_OFFLINE) {
                (*outfile) << debugHead << "... area under 30 -> filtered" << endl;
            }
			
			fireSanity(BLOB_AREA_LT30);
			continue;
		}


        // SANITY CHECK ------------------------------------------------------

        /* Check if goal is too much above horizon. Compare using the lower 
         * edge so that it won't be affected by whether it's far/near. */
        int goalBaseX, goalBaseY;
            
        #define use_centroid_for_horizon_check
        #ifdef use_centroid_for_horizon_check
        goalBaseX = (int) goalBlob.cx;
        goalBaseY = (int) goalBlob.cy;

        #else
        /* Vertical. */
        if (ABS(degRoll) < 30) {
            goalBaseX = (int) goalBlob.cx;
            goalBaseY = (int) goalBlob.getYMax();

        /* Around 45 deg anit-clockwise. */
        } else if (degRoll > 30 && degRoll < 60) {
            goalBaseX = (int) goalBlob.getXMax();
            goalBaseY = (int) goalBlob.getYMax();
            
        /* Around 45 deg clockwise. */
        } else if (degRoll < -30 && degRoll > -60) {
            goalBaseX = (int) goalBlob.getXMin();
            goalBaseY = (int) goalBlob.getYMax();

        /* Much roll anti-clockwise. */
        } else if (degRoll > 60) {
            goalBaseX = (int) goalBlob.getXMax();
            goalBaseY = (int) goalBlob.cy;

        /* Much roll clockwise. */
        } else if (degRoll < -60) {
            goalBaseX = (int) goalBlob.getXMin();
            goalBaseY = (int) goalBlob.cy;
        }
        #endif //use_centroid_for_horizon_check

        int goalHeight= (int) (goalBlob.getYMax() - goalBlob.getYMin());
        int goalWidth = (int) (goalBlob.getXMax() - goalBlob.getXMin());
    
        /* Only check the horizon if it exists. */
        if (hzExists 
                && goalHeight < 100
                && goalWidth  < 100) {

            /* Check if goal is too much above horizon. */
            if (!isBelowHorizon(goalBaseX, goalBaseY, 20)) {

                if (flag_OFFLINE && debugLookForColourGoal) {
                    (*outfile) << debugHead << "...blob too high -> filtered" << endl;
                }
                fireSanity(BLOB_TOO_HIGH);
                continue;
            }

            /* Check if goal is too much below horizon. */
            if (isBelowHorizon(goalBaseX, goalBaseY, -40)) {

                if (flag_OFFLINE && debugLookForColourGoal) {
                    (*outfile) << debugHead 
                        << "...blob is too low below horizon. "
                        << endl;
                }
                continue;
            }
        }


        // SANITY CHECK ------------------------------------------------------
        /* Check elevation and heading comparing to beacons. */

		bool filteredUsingBeacon = false;

		VisualObject goalPatch;

        /* cx/cy is unrotated, only cRx/cRy are. */
		goalPatch.imgHead = PointToHeading(goalBlob.cx, goalBlob.cy, sin_eroll, cos_eroll);
		goalPatch.imgElev = PointToElevation(goalBlob.cx, goalBlob.cy, sin_eroll, cos_eroll);

        #define detail_goal_against_beacon_elev_check
        #ifdef detail_goal_against_beacon_elev_check
        if (debugLookForColourGoal && flag_OFFLINE) {
            (*outfile) 
                << debugHead << "Checking elev/head against beacons... " << endl
                << debugHead << "   goal imgHead  : " << goalPatch.imgHead << endl
                << debugHead << "   goal imgElev  : " << goalPatch.imgElev << endl
                << debugHead << "   goal cx, cy   : " 
                    << goalBlob.cx << ", " << goalBlob.cy << endl;
        }
        #endif // detail_goal_against_beacon_elev_check

		for (int k = vobBlueLeftBeacon; k <= vobYellowRightBeacon; k++) {

			if (vob[k].cf > 0) {

                /* Check against rotated imgHead/Elev of beacons. */
				VisualObject beacon;
				beacon.imgHead = PointToHeading(vob[k].cx, vob[k].cy, sin_eroll, cos_eroll);
				beacon.imgElev = PointToElevation(vob[k].cx, vob[k].cy, sin_eroll, cos_eroll);

                #define detail_goal_against_beacon_elev_check
                #ifdef detail_goal_against_beacon_elev_check
                if (debugLookForColourGoal && flag_OFFLINE) {
                    (*outfile) 
                        << endl
                        << debugHead << "-- against " << vobName[k] << endl

                        << debugHead << "   beacon imgHead: " << beacon.imgHead << endl
                        << debugHead << "   beacon imgElev: " << beacon.imgElev << endl
                        << debugHead << "   beacon cx, cy : " 
                            << vob[k].cx << ", " << vob[k].cy << endl;
                }
                #endif // detail_goal_against_beacon_elev_check

                /* CHECK -------------------------------------------------- */
                /* Goal cannot be too high above beacon. */
				if (goalPatch.imgElev > beacon.imgElev + 4) { // 4 for error from roll.

                    if (flag_OFFLINE && !flag_PRE_SANITY_BLOB) {

                        if (debugLookForColourGoal) {
                            (*outfile) << debugHead 
                                << "... blob is too high above beacon -> "
                                << "filtered." 
                                << endl;
                        }
                    }
                    
					filteredUsingBeacon = true;
					fireSanity(BLOB_ELEVATION_GT_BEACON_ELEVATION + k - vobBlueLeftBeacon);
					break;
				}


                /* CHECK -------------------------------------------------- */
                /* Goal cannot be too low below beacon. */
				if (beacon.imgElev - goalPatch.imgElev > 17) {

                    if (flag_OFFLINE && !flag_PRE_SANITY_BLOB) {

                        if (debugLookForColourGoal) {
                            (*outfile) << debugHead 
                                << "... blob is too low above beacon -> "
                                << "filtered." 
                                << endl;
                        }
                    }
                    
					filteredUsingBeacon = true;
					fireSanity(BLOB_ELEVATION_GT_BEACON_ELEVATION + k - vobBlueLeftBeacon);
					break;
				}


                /* CHECK -------------------------------------------------- */
                /* Goal cannot be too close to the beacon. */
                /* Exception: In some angle, beacon in left of goal is very
                ** close to goal. */ 
				if (
                       !(   (k == beaconLeftOfGoalIndex
                             && beaconLeftOfGoal->d < 300
                             && beacon.imgHead      > goalPatch.imgHead)

                         || (k == beaconRightOfGoalIndex
                             && beaconRightOfGoal->d < 300
                             && beacon.imgHead       < goalPatch.imgHead)
                        )
                     &&
                        ABS(goalPatch.imgHead - beacon.imgHead) < 17
                   ) {

                    if (flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                        if (debugLookForColourGoal) {
                            (*outfile) << debugHead 
                                << "... blob is too close to a beacon. "
                                << endl;
                        }
                    }

					filteredUsingBeacon = true;
					break;
				}
			}
		}

		if (filteredUsingBeacon) {
			continue;
		}
	

        // SANITY CHECK ------------------------------------------------------

        // Check if blue goal is probably due to ring effect. 
        int     boxArea  = goalWidth * goalHeight;
        double  density  = (double) goalBlob.area / (double) boxArea;

        if (density < 0.6) {

            if ((goalBlob.cx < 20 && goalBlob.cy < 20) ||
                (goalBlob.cx < 20 && goalBlob.cy > 140) ||
                (goalBlob.cx > 188 && goalBlob.cy < 20) ||
                (goalBlob.cx > 188 && goalBlob.cy > 140))
            {
                goalBlob.ResetBlob();

                if (debugLookForColourGoal && flag_OFFLINE) {
                    (*outfile) << debugHead
                        << "...blue goal in corner, likely to be a ring effect." 
                        << endl;
                }
                fireSanity(BLUE_GOAL_IN_CORNER);
            }
        }



        // SANITY CHECK ------------------------------------------------------

		// Its dimension is too small, so forget about it altogether
		if (goalWidth < 12 || goalHeight < 6) {

            if (flag_OFFLINE && !flag_PRE_SANITY_BLOB) {

                if (debugLookForColourGoal) {
                    (*outfile) << debugHead 
                        << "... dimension is too small to be a goal -> "
                        << "filtered."
                        << endl;
                    (*outfile) << debugHead
                        << "    -- goal xMin/ax, width : "
                            << goalBlob.getXMin() << ", "
                            << goalBlob.getXMax() << ", " 
                            << goalWidth << endl
                        << debugHead << "    -- goal yMin/ax, height: "
                            << goalBlob.getYMin() << ", "
                            << goalBlob.getYMax() << ", "
                            << goalHeight << endl
                        << endl;
                }
            }

			fireSanity(BLOB_TOO_SMALL);
			continue;	
		}


        // SANITY CHECK ------------------------------------------------------

		// A goal blob cannot be to the left of the left goal beacon
		// so we conclude it cannot be the goal

        int rotatedBeaconLeftOfGoalx = WIDTH / 2 
                            - (int) (  cos_eroll * (WIDTH  / 2 - beaconLeftOfGoal->cx)
                                     - sin_eroll * (HEIGHT / 2 - beaconLeftOfGoal->cy));

		if (beaconLeftOfGoal->cf > 0 
            && rotatedBeaconLeftOfGoalx > goalBlob.cRx) {

            if (debugLookForColourGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead 
                    << "... goal in left of left goal beacon -> filtered."
                    << endl;
            }
	    
			fireSanity(BLOB_LEFT_OF_LEFT_BEACON);
			continue;
		}
		

        // SANITY CHECK ------------------------------------------------------

		// A goal blob cannot be to the right of the right beacon
		// so we conclude it cannot be the goal

        int rotatedBeaconRightOfGoalx = WIDTH / 2 
                                - (int) (  cos_eroll * (WIDTH  / 2 - beaconRightOfGoal->cx)
                                         - sin_eroll * (HEIGHT / 2 - beaconRightOfGoal->cy));

		if (beaconRightOfGoal->cf > 0 
            && rotatedBeaconRightOfGoalx < goalBlob.cRx) {

            if (debugLookForColourGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead 
                    << "... goal in right of right goal beacon -> filtered."
                    << endl;
            }

			fireSanity(BLOB_RIGHT_OF_RIGHT_BEACON);
			continue;
		}


        // SANITY CHECK ------------------------------------------------------

		// See if this blob should be included as part of the
        // already-classified goal blobs.

        if (goal_blob.area != 0) {

            int blob_cy = (goalBlob.getYMin() + goalBlob.getYMax()) / 2;
            int blob_cx = (goalBlob.getXMin() + goalBlob.getXMax()) / 2;

            int oldWidth  = MAX(uxmax-uxmin, uymax-uymin);
            int blobWidth = MAX(goalBlob.getXMax()-goalBlob.getXMin(),
                                goalBlob.getYMax()-goalBlob.getYMin());

            int halfSmallerWidth = MIN(oldWidth, blobWidth) / 2;

            int bgPix   = calDistBaseOnColor((uxmin + uxmax) / 2, (uymin + uymax) / 2,
                                              blob_cx, blob_cy, NULL, BACKGROUND);
            int gnPix   = calDistBaseOnColor((uxmin + uxmax) / 2, (uymin + uymax) / 2,
                                              blob_cx, blob_cy, NULL, cFIELD_GREEN);

            // too many background colours between the blobs.
            // shudnt be field in between either
            // also should use unrotated coordinates since accessing cplane
            if ((bgPix > halfSmallerWidth || gnPix > halfSmallerWidth/2)) {

                if (debugLookForColourGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                    (*outfile) << debugHead 
                        << "... too much bg color btwn blob -> filtered."
                        << endl;
                    (*outfile) 
                        << debugHead << "    bgPix    : " << bgPix << endl
                        << debugHead << "    gnPix    : " << gnPix << endl
                        << debugHead << "    oldWidth : " << oldWidth << endl
                        << debugHead << "    blobWidth: " << blobWidth << endl
                        << debugHead << "    halfSmallerWidth: " 
                            << halfSmallerWidth << endl
                        << endl;
                }

                fireSanity(TOO_MUCH_BACKGROUND_COLOUR_BETWEEN_BLOBS);
                continue;
            }
        }


        /*--------------------------------*/

        if (debugLookForColourGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead
                << "...TAKEN!" 
                << endl;
        }

        /* goal_blob x/y min/max can always contain the original goal. */
		goal_blob.setXMax(MAX(goal_blob.getXMax(), goalBlob.xRmax));
		goal_blob.setYMax(MAX(goal_blob.getYMax(), goalBlob.yRmax));
		goal_blob.setXMin(MIN(goal_blob.getXMin(), goalBlob.xRmin));
		goal_blob.setYMin(MIN(goal_blob.getYMin(), goalBlob.yRmin));

        /* However original ones are also keep tracked. */
		uxmin = MIN(uxmin, goalBlob.getXMin());
		uymin = MIN(uymin, goalBlob.getYMin());
		uxmax = MAX(uxmax, goalBlob.getXMax());
		uymax = MAX(uymax, goalBlob.getYMax());

        xRmin = MIN(xRmin, goalBlob.xRmin);
        yRmin = MIN(yRmin, goalBlob.yRmin);
        xRmax = MAX(xRmax, goalBlob.xRmax);
        yRmax = MAX(yRmax, goalBlob.yRmax);


        if (goal_blob.area != 0) 
            multipleBlobJoined = true;

		goal_blob.area += goalBlob.area;
	}


    /* Basic check on patch has finished, now form a goal and check again. */

	goal_blob.cx = (goal_blob.getXMax() + goal_blob.getXMin()) / 2.0;
	goal_blob.cy = (goal_blob.getYMax() + goal_blob.getYMin()) / 2.0;
	long xmin = goal_blob.getXMin();
	long xmax = goal_blob.getXMax();
	long ymin = goal_blob.getYMin();
	long ymax = goal_blob.getYMax();;
	double width = xmax + 1 - xmin;
	double height = ymax + 1 - ymin;

    double ucx = (uxmin + uxmax) / 2.0;
    double ucy = (uymin + uymax) / 2.0;

    if (debugLookForColourGoal && flag_OFFLINE) {
        (*outfile) 
            << debugHead << "Checking FORMED goal (rotated):" << endl
            << debugHead << "    xmin/max: " << xmin << ", " << xmax << endl
            << debugHead << "    ymin/max: " << ymin << ", " << ymax << endl
            << debugHead << "    area    : " << goal_blob.area << endl
            << endl;
    }


    // SANITY CHECK ------------------------------------------------------
    int minGoalBlobSize = 200;

    if (uxmin < 5 || uxmax > 203
            || uymin < 5 || uymax > 155) {
        minGoalBlobSize = 100;
    }

	if (goal_blob.area < minGoalBlobSize) {

        if (debugLookForColourGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead 
                << "... FORMED goal area less than minimum goal size."
                << endl;
        }

		fireSanity(BLOB_AREA_LESS_THAN_MINIMUM_GOAL_SIZE);
		return;
	}


    // SANITY CHECK ------------------------------------------------------
    /* Take out goals of low density. Remember that area is actually number 
     * of pixels. */

    int goalRotatedBoundingBoxArea = (xRmax - xRmin +1) * (yRmax - yRmin +1);
	double goal_rotatedBox_density     = (double) goal_blob.area / (double) goalRotatedBoundingBoxArea;
	double goal_oriBoundingBox_density = (goal_blob.area * 1.0) / ((uymax - uymin + 1) * (uxmax - uxmin + 1));

    double goal_density = MAX(goal_rotatedBox_density, goal_oriBoundingBox_density);

    if (debugLookForColourGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
        (*outfile) << debugHead 
            << "Checking goal density... " << endl
            << "    -- xRmin/ax, yRmin/ax: "
                << xRmin << ", " << xRmax << ", "
                << yRmin << ", " << yRmax << endl
            << "    -- blob_area: " << goal_blob.area << endl
            << "    -- goalRotatedBoundingBoxArea: " << goalRotatedBoundingBoxArea << endl
            << "    -- goal_density: " << goal_density << endl
            << endl;
    }


	if (!multipleBlobJoined && goal_density < 0.4) {

		fireSanity(BLOB_DENSITY_TOO_LOW);

        if (debugLookForColourGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {

            (*outfile) << debugHead 
                << "... FORMED goal density too low (" << goal_density
                << endl; 
        }
		return;
	} 
           
    // SANITY CHECK ------------------------------------------------------
    /* Take out goals of extreme bounding box ratio. */

	double goal_proportion = ((goal_blob.getXMax() - goal_blob.getXMin() + 1) * 1.0)
		                 / (goal_blob.getYMax() - goal_blob.getYMin() + 1);

	if (goal_proportion <0.2 || goal_proportion> 5.0) {
		fireSanity(BLOB_DENSITY_TOO_LOW);

        if (debugLookForColourGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {

            (*outfile) << debugHead 
                << "... FORMED goal aspect ratio too extreme! (" 
                << goal_proportion << ")" << endl; 
        }
		return;
	} 
     

    // SANITY CHECK ------------------------------------------------------

	/* A goal with small blob area not on the field? can't be! */

	//if (NoGreenNearBy(goal_blob.getXMin (), goal_blob.getXMax (), goal_blob.getYMin (),
	//                  goal_blob.getYMax ()) && goal_blob.area < 7000) {

    /*
    bool noGreenNearBy = NoGreenNearBy(uxmin, uxmax, uymin, uymax);

    // Constant is about the size when goal is at a distance where green field
    // can still be seen.
	if (noGreenNearBy && goal_blob.area < 7000) {
		if (debugLookForColourGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead 
                <<  "... FORMED goal - no green field around."
                << endl;
        }
		fireSanity(NO_GREEN_NEAR_BLOB);
		return;
	}
    */

    
    // SANITY CHECK ------------------------------------------------------

	/* A goal with not enough green below - throw away. */

    if (goal_blob.area < 7000) {

        bool goalSurroundingOk = goalSurroundingCheck(uxmin, uymin, uxmax - uxmin, uymax - uymin);

        // Constant is about the size when goal is at a distance where green field
        // can still be seen.
        if (!goalSurroundingOk) {

            if (debugLookForColourGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead 
                    <<  "... FORMED goal - surrounding is not right (too much white/too little green)."
                    << endl;
            }
            fireSanity(NO_GREEN_NEAR_BLOB);
            return;
        }
    }




    // SANITY CHECK ------------------------------------------------------

	/* Get rid of goals that has tiny width or height. */

	if (width < minGoalDim || height < minGoalDim) {
		goal->reset();
		fireSanity(TINY_BLOB_WIDTH_OR_HEIGHT);

        if (debugLookForColourGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead 
                << "... FORMED goal aspect ratio does not make sense. "
                << endl;
        }
		return;
	}

	int delta = 4;

	if (height <= 0) {
		height = 1;
	}

	if ((width / height) > kGoalDimensionRatio) {
		// need to expand the height
		double new_height = width / kGoalDimensionRatio;

		// what is delta??
		// give it some tolerance so that the edge of the
		// object doesnt have to be exactly on the edge of the screen
		if (ymax >= USABLE_HEIGHT - 1 - delta)        // object is on the bottom edge of the FOV
			ymax = ymin + (int) new_height;
		else if (ymin <= delta)        // object is on the top edge of the FOV
			ymin = ymax - (int) new_height;
		else {
			// object's bbox is not in ratio?
			double offset;
			offset = (new_height - height) / 2.0;
			ymin -= (int) offset;
			ymax += (int) offset;
		}

		height = new_height;
	}


    /*
     * Store all the visual object information regarding the goal.
     */
    goal->x = xmin;          // These are rotated ones.
    goal->y = ymin;
    goal->width = width;
    goal->height = height;

    goal->cx = (xmin + xmax) / 2.0;
    goal->cy = (ymin + ymax) / 2.0;

    // apply clippings
    long pxmin = (long) MAX(xmin, 0);
    long pxmax = (long) MIN(xmax, WIDTH);
    long pymin = (long) MAX(ymin, 0);
    long pymax = (long) MIN(ymax, HEIGHT);

    goal->projected_x = pxmin;
    goal->projected_y = pymin;
    goal->projected_w = pxmax - pxmin;
    goal->projected_h = pymax - pymin;

    goal->area = goal_blob.area;
    goal->tilt = 0;
    goal->misc = 0;

    double goalCF = 0;
    int sanities = 0;


    /* Rotating image back to normal about center of cplane, to get the real
    ** imgHead and imgElev. */
    int uGoalX = unrotatePointReturnX(goal->cx, goal->cy);
    int uGoalY = unrotatePointReturnY(goal->cx, goal->cy);

    goal->imgHead = PointToHeading(uGoalX);
    goal->imgElev = PointToElevation(uGoalY);


    /*-----------------------------------------------------------------------*/
    /* All the codes below fine-tuned the variance and confidence.
     * I didn't review them, from the game, they seem to be ok.
     *-----------------------------------------------------------------------*/


    // Add the variance for the goal Tim Tam 13/03/2002
    goal->var = LARGE_VAL;
    goal->angleVar = LARGE_VAL;

    double expectedArea = goal->width *goal->height;

    /* Removed shape challenge stuff since rev 2726. */

    // add 10 since the calibrated measurements go to the mouth of the goal which is
    // 10cm from the start of the goal color itself.
    int goalPixHeight = (int)CLIP(goal->height, 143.0);
    goal->dist     = (double)goalDistanceLookup[goalPixHeight];
    goal->var      = SQUARE(400.0 / goalPixHeight) + exp(goalPixHeight / 16.0)
                    + SQUARE(30.0 * (1 - goal->area / expectedArea));
    goal->angleVar = exp(goalPixHeight / 16.0) + 50;

    // wout << "goal dist, var: " << goal->dist << ", " << goal->var << "\n";
    // wout << "x,y,width,height: " << goal->x << ", " << goal->y << ", " 
    //     << goal->width << ", " << goal->height << "\n";


    // There are no beacons that we can cross-ref against, so we will have to reduce
    // our certainty by a bit.
    if (!sanities) {
        goalCF = 666;
        goal->var *= 2;                // this is an adhoc value
        goal->angleVar *= 2;
    }
    // figure out our new confidence factor.
    if (sanities <= 0) {
        sanities = 1;
    }

    goal->cf = (int) (goalCF / sanities);

    /* Used to have old goal dist calculation using size of goal. Removed
     ** since rev 2570. */

    // modified by Spencer confidence is capped at 1000
    goal->cf = MIN(goal->cf, 1000);

    if (goal->cf || goal->var != LARGE_VAL) {
        RobotRelative(*goal);

        goal->lhead = goal->head + (PointToHeading(xmin) - PointToHeading(goal->cx));
        goal->rhead = goal->head + (PointToHeading(xmin + width) - PointToHeading(goal->cx));

        if (debugLookForColourGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) << debugHead << "goal found, elev " << goal->elev << " dist " << goal->d << endl;
        }

        // wout << "left: " << goal->lhead << " mid: " << goal->head << " right: " << goal->rhead << "\n";
    }

    if (debugLookForColourGoal && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
        (*outfile) << debugHead << "...Goal Sanity passed! " << endl;
    }
}


void VisualCortex::Beacon::Set(
        const HalfBeacon &pink,
        const HalfBeacon &closest) {

    valid = true;
    cx = (pink.cx + closest.cx) / 2.0;
    cy = (pink.cy + closest.cy) / 2.0;
    area = pink.area + closest.area;
    halfmass = pink.area > closest.area ? pink.area : closest.area;

    c2c = sqrt(SQUARE(pink.cx - closest.cx) + SQUARE(pink.cy - closest.cy));

    pinkXsum = pink.xsum;
    pinkYsum = pink.ysum;
    pinkarea = pink.area;
    closestXsum = closest.xsum;
    closestYsum = closest.ysum;
    closestarea = closest.area;


    angle = NormalizeAngle_0_360(RAD2DEG(atan2(pink.cy - closest.cy,
                    pink.cx - closest.cx)));

    xmin = MIN(pink.getXMin(), closest.getXMin());
    ymin = MIN(pink.getYMin(), closest.getYMin());

    xmax = MAX(pink.getXMax(), closest.getXMax());
    ymax = MAX(pink.getYMax(), closest.getYMax());

    /*
    width = (pink.getXMin() < closest.getXMin())
        ? (closest.getXMax() - pink.getXMin() + 1)
        : (pink.getXMax() - closest.getXMin() + 1);

    height = pink.getYMin() < closest.getYMin()
        ? closest.getYMax() - pink.getYMin() + 1 
        : pink.getYMax() - closest.getYMin() + 1;
        */
}


bool VisualCortex::NoGreenNearBy(int xmin, int xmax, int ymin, int ymax) {

    int min_delta = 7;
    int delta_x = MAX(min_delta, (xmax - xmin) / 3);
    int delta_y = MAX(min_delta, (ymax - ymin) / 3);
    return NoColourXNearBy(cFIELD_GREEN,
            xmin,
            xmax,
            ymin,
            ymax,
            delta_x,
            delta_y,
            kFieldGreenThreshold);
}

double getAng(double dx, double dy) {
    double t;

    if (!dy) {
        return (dx >= 0) ? 0 : 180;
    }
    if (!dx) {
        return (dy >= 0) ? 90 : 270;
    }

    t = atan((int) abs((int) dy) / dx) * (double) (180 / 3.14159);
    if (dx < 0) {
        t += 180;
    }
    return (dy >= 0) ? t : 360 - t;
}

int imax(int a, int b) {
    if (a > b) {
        return a;
    }
    else {
        return b;
    }
}

int imin(int a, int b) {
    if (a < b) {
        return a;
    }
    else {
        return b;
    }
}

int pi_abs(int a) {
    if (a < 0) {
        return -a;
    }
    else {
        return a;
    }
}

// Calculates Image Heading using cplane x value.
double PointToHeading(double x) {
    return RAD2DEG(atan(((WIDTH / 2.0) - x) * 2.0 * tan(xField / 2.0) / WIDTH));
}

// Calculates Image Elevation using cplane y value.
double PointToElevation(double y) {
    return RAD2DEG(atan(((HEIGHT / 2.0) - y) * 2.0 * tan(yField / 2.0) / HEIGHT));
}

// Calculates Rotated Image Heading.
double PointToHeading(double x, double y, double sin_eroll, double cos_eroll) {
    return RAD2DEG(
            atan(
                (
                      (WIDTH  / 2.0 - x) * cos_eroll
                    - (HEIGHT / 2.0 - y) * sin_eroll) * 2.0 * tan(xField / 2.0)
                / WIDTH)
            );
}

// Calculates Rotated Image Elevation
double PointToElevation(double x, double y, double sin_eroll, double cos_eroll) {
    return RAD2DEG(
            atan(
                (
                    (WIDTH  / 2.0 - x) * sin_eroll
                  + (HEIGHT / 2.0 - y) * cos_eroll) * 2.0 * tan(yField / 2.0)
                / HEIGHT)
            );
}

// arguments are in centimetres and pixels
long SizeToDistance(long physicalSize, long apparentSize) {
    // OK, when the chessboard was 495mm away, it appeared to be
    // 60 by 54 pixels in size. The chessboard is IRL 325 by 330mm

    // Time for averages here... chessboard is 33cm. Its image
    // at a distance of 50cm appeared over 57 pixels.

    // Say height is 1cm. Then at 50cm away it would take up
    // 57/33 pixels. i.e. apparent = sz/z, or 57/33 = sz/50.
    // So sz ('distance' to screen) is (50*57)/33.
    // In general eqn is apparent = physical*sz/z.
    // So now z = (physical*sz)/apparent

    if (!apparentSize) {
        return 400;
    }                        // a long way away

    return (long) ((physicalSize * DIST_CONST) / apparentSize);
}

//WARNING: not sure what this does
double VisualCortex::NewBallDistance() {
    // assume the image is pretty much horizontally alligned
    double elev_of_top_of_ball = effective_tilt
        + DEG2RAD(-PointToElevation(vob[vobBall].misc));

    // ball above horizon... abnormal ball position
    if (elev_of_top_of_ball < 0) {
        return 999;
    }

    double height_snout = NECKBASE_HEIGHT
        + NECK_LENGTH *cos_tilt
        - FACE_LENGTH *sin_tilt *cos_pan;
    double similar_triangle_height_at_centre_of_ball = BALL_RADIUS
        / cos(elev_of_top_of_ball)
        + BALL_RADIUS;
    double distance_snout_2_ball = (height_snout
            - similar_triangle_height_at_centre_of_ball)
        / tan(elev_of_top_of_ball);
    /*
       cout << "elev_of_top_of_ball: " << elev_of_top_of_ball << endl;
       cout << "height_snout: " << height_snout << endl;
       cout << "similar_triangle_height_at_centre_of_ball: " 
            << similar_triangle_height_at_centre_of_ball << endl;
       cout << "distance_snout_2_ball: " << distance_snout_2_ball << endl;
       cout << "dist_neck2snout: " << dist_neck2snout << endl;
       */
    return distance_snout_2_ball;        //(distance_snout_2_ball + dist_neck2snout);
}



/* Convert VisualObject coordinates from relative-to-camera into relative-to-robot.
 * Using standard Rotation and Translation matrices.
 * Reference for the Math Calucalation obtained from Computer Graphics comp3421 lecture.
 *
 * Inputs are object.dist estimated earlier, imgElev and imgHead.
 */
void VisualCortex::RobotRelative(VisualObject &object) {

	double ihead = DEG2RAD(object.imgHead);
	double ielev = DEG2RAD(object.imgElev);
	double x1 = 0;
	double y1 = 0;
	double z1 = object.dist;

	// Rotation for Image Heading
    double x2 = x1 *cos(ihead) + z1 *sin(ihead);
    double y2 = y1;
   	double z2 = -x1 *sin(ihead) + z1 *cos(ihead);

	// Rotation for Image Elevation
	double x3 = x2;
    double y3 = y2 *cos(ielev) + z2 *sin(ielev);
    double z3 = -y2 *sin(ielev) + z2 *cos(ielev);

    // Translation for Face Length
	double x4 = x3;
    double y4 = y3;
    double z4 = z3 + FACE_LENGTH;

	// Translation for Camera Offset
	double x4b = x4;
    double y4b = y4 - (CAMERA_OFFSET / 10);
    double z4b = z4;

	// Rotation for Head Crane
	double x5 = x4b;
    double y5 = y4b *cos_crane - z4b *sin_crane;
    double z5 = y4b *sin_crane + z4b *cos_crane;

    // Rotation for Head Pan
   	double x5b = x5 *cos_pan + z5 *sin_pan;
    double y5b = y5;
    double z5b = -x5 *sin_pan + z5 *cos_pan;

	// Translation for Neck Length
	double x6 = x5b;
    double y6 = y5b + NECK_LENGTH;
    double z6 = z5b;

	// Rotation for Head Tilt
	double x7 = x6;
    double y7 = y6 *cos_tilt - z6 *sin_tilt;
    double z7 = y6 *sin_tilt + z6 *cos_tilt;

	// Rotation for Body Tilt
	double x8 = x7;
    double y8 = y7 *cos(bodyTilt) - z7 *sin(bodyTilt);
    double z8 = y7 *sin(bodyTilt) + z7 *cos(bodyTilt);

	// Calculate actual distance and robot relative heading and elevation
	object.setDistance(x8, y8, z8);
}




/* Convert VisualObject coordinates from relative-to-camera into relative-to-robot.
 * Using standard Rotation and Translation matrices.
 * Reference for the Math Calucalation obtained from Computer Graphics comp3421 lecture.
 *
 * Inputs are object.dist estimated earlier, imgElev and imgHead.
 */
/*
void VisualCortex::RobotRelative(VisualObject &object, bool dp) {

	double ihead = DEG2RAD(object.imgHead);
	double ielev = DEG2RAD(object.imgElev);
	double x1 = 0;
	double y1 = 0;
	double z1 = object.dist;

	if (debugRobotRelative && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
        	(*outfile) << "1: " << x1 << " " << y1 << " " << z1 << endl;
	}

	// Rotation for Image Heading
    double x2 = x1 *cos(ihead) + z1 *sin(ihead);
    double y2 = y1;
   	double z2 = -x1 *sin(ihead) + z1 *cos(ihead);

	if (debugRobotRelative && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
        (*outfile) << "2: " << x2 << " " << y2 << " " << z2 << endl;
	}

	// Rotation for Image Elevation
	double x3 = x2;
    double y3 = y2 *cos(ielev) + z2 *sin(ielev);
    double z3 = -y2 *sin(ielev) + z2 *cos(ielev);

	if (debugRobotRelative && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
        (*outfile) << "3: " << x3 << " " << y3 << " " << z3 << endl;
	}

	// Translation for Face Length
	double x4 = x3;
    double y4 = y3;
    double z4 = z3 + FACE_LENGTH;

    if (debugRobotRelative && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
(*outfile) << "4: " << x4 << " " << y4 << " " << z4 << endl;
    }

    // Rotation for Head Pan
   	double x5 = x4 *cos_pan + z4 *sin_pan;
    double y5 = y4;
    double z5 = -x4 *sin_pan + z4 *cos_pan;

    if (debugRobotRelative && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
        (*outfile) << "5: " << x5 << " " << y5 << " " << z5 << endl;
	}

	// Translation for Neck Length
	double x6 = x5;
    double y6 = y5 + NECK_LENGTH;
    double z6 = z5;

	if (debugRobotRelative && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
        (*outfile) << "6: " << x6 << " " << y6 << " " << z6 << endl;
	}

	// Rotation for Head Tilt
	double x7 = x6;
    double y7 = y6 *cos_tilt - z6 *sin_tilt;
    double z7 = y6 *sin_tilt + z6 *cos_tilt;

	if (debugRobotRelative && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
        (*outfile) << "7: " << x7 << " " << y7 << " " << z7 << endl;
	}

	// Calculate actual distance and robot relative heading and elevation
	object.setDistance(x7, y7, z7);
}
*/





/* Check if there're dark background in a bounding box, return true if it's
** above a specified density. */
/* Temp disabled 28/3/2004. */
bool VisualCortex::DarkBackgroundInBetween(int left,
        int right,
        int top,
        int bottom) {

    static const int    minRequiredBlack = 144;
    static const double minRequiredDensity = 0.8;

    int xMin = CPLANE_WIDTH, xMax = 0, yMin = CPLANE_HEIGHT, yMax = 0;
    int num = 0;

    /* Go through each pixel in the bounding box. */
    for (int y = top; y >= bottom; y--) {
        for (int x = left; x <= right; x++) {

            // if background i.e. robot grey or black
            if ((cplane[y * WIDTH + x] & ~MAYBE_BIT) == cROBOT_GREY
                    || (cplane[y * WIDTH + x] & ~MAYBE_BIT) == cBLACK) {
                num++;

                /* I doubt this: won't the box shrink to nothing very quickly?
                ** */
                xMin = MIN(xMin, x);
                xMax = MAX(xMax, x);
                yMin = MIN(yMin, y);
                yMax = MAX(yMax, y);

                double density = (double) num / ((xMax - xMin) * (yMax - yMin));

                if (num >= minRequiredBlack && 
                        density > minRequiredDensity)
                    return true;
            }
        }
    }

    return false;
}


/* The given color is either cROBOT_BLUE or cROBOT_RED. */
/* Scan robot blobs from left to right, and form robot bounding box as long as
** their horizontal distance is not too far away. After getting all dogs
** drawn, estimate their distances and headings from the camera. */

/* Daniel 2004: For a real sensible way of merging the blobs, I think we need
** to check the distance between blobs, whether there are white/black/green
** between these blobs, and hori/vert in both direction as a whole, rather
** than just scanning once from left to right. This will be done after Aus
** open. */

//TODO: please log sanity by calling fireSanity(), or we will never know what 
//      happened with the dog
void VisualCortex::FindColorRobot(int color) {
#ifdef ONLY_ONE_DOG    
    static const int ROBOT_AREA_THRESHOLD = 80;
#else
    static const int ROBOT_AREA_THRESHOLD = 120;
#endif //ONLY_ONE_DOG

    VisualObject *robot;
    int vobNum;
    const string debugHead = "FindColorRobot| ";

    if (debugFindColorRobot && flag_OFFLINE) {
        (*outfile) << debugHead 
            << "Start finding robot, color " << ColorName[color] 
            << endl << endl;
    }


    /* Set Color speicific variables. */
    if (color == cROBOT_RED) {
        robot = &vob[vobRedDog];
        vobNum = vobRedDog;

    } else if (color == cROBOT_BLUE) {
        robot = &vob[vobBlueDog];
        vobNum = vobBlueDog;

    } else {
        cout << "Error: invalid visual object" << endl;
        return;
    }

    int bufferX = 0,
        bufferY = 0;

    static const int NUM_BLUE_ROBOT = 5;    
    int blueRobotMixedArea[NUM_BLUE_ROBOT]; // As long as this is more than num of robots.
    int blueRobotMaybeArea[NUM_BLUE_ROBOT];

    for (int i = 0; i < NUM_BLUE_ROBOT; i++) {
        blueRobotMixedArea[i] = 0;
        blueRobotMaybeArea[i] = 0;
    }


    /* Extrems of the dog bounding box, both after/before rotation. */
    int xMin = CPLANE_WIDTH, xMax = 0, yMin = CPLANE_HEIGHT, yMax = 0;
    int uxMin = CPLANE_WIDTH, uxMax = 0, uyMin = CPLANE_HEIGHT, uyMax = 0;

    /* Values set to opposite extrems so that they can be more and more
    ** refined back to what they really mean in the loop (the xMin getting
    ** smaller and smaller). */

    int totalArea       = 0;
    int numberOfBlobs   = 0; // Number of blobs belong to one robot. 

    bool isFirstBlob    = true;

    /* Go through all blobs of a particular color. */
    /* More info in 2002 report page 44. */
    for (int i = 0; i < blobs->count(color); i++) {

        BlobInfo&   robotBlob = blobs->getBlob(color,i);

        int         robotXMin = robotBlob.getXMin();
        int         robotYMin = robotBlob.getYMin();
        int         robotXMax = robotBlob.getXMax();
        int         robotYMax = robotBlob.getYMax();

        /* Some blobs have been filtered (in RobotNoiseFilter and FindRobot),
         * they have area = 0 and are not considered anymore. */
        if (robotBlob.area > 0) {

            /* Heuristic : extend horizontally by 3 times the [width|height]
            ** (take the larger because sometimes it maybe partially seen) of 
            ** [current bounding box | new blob] (take the smaller because we
            ** don't want to join far away small blobs with the large blob, as
            ** it's usually noise). */

            int blobSide =  MAX(robotBlob.xRmax - robotBlob.xRmin + 1 , 
                                robotBlob.yRmax - robotBlob.yRmin + 1);

            int currentBoundingBoxSide = MAX(xMax - xMin + 1, 
                                             yMax - yMin + 1);

            bufferX = MIN(3 * blobSide, 3 * currentBoundingBoxSide);

            /* For Kim's KeepAway Soccer Open Challenge:
             * Make threshold very large. so that there's little chance that it 
             * seperate as 2 dogs. */
            #ifdef ONLY_ONE_DOG
            bufferY = MIN(3 * blobSide, 3 * currentBoundingBoxSide);

            #else // ONLY_ONE_DOG
            bufferY = MIN(2 * blobSide, 2 * currentBoundingBoxSide);

            #endif // ONLY_ONE_DOG


            if (debugFindColorRobot && flag_OFFLINE) {
                (*outfile) 
                    << debugHead << "Go through blob #" << robotBlob.blob_number << endl
                    << debugHead << "    xMin/ax,  yMin/ax, area: "
                        << robotXMin << ", " << robotXMax << ", "
                        << robotYMin << ", " << robotYMax << ", "
                        << robotBlob.area << endl
                    << debugHead << "    xRmin/ax, yRmin/ax     : "
                        << robotBlob.xRmin << ", " << robotBlob.xRmax << ", "
                        << robotBlob.yRmin << ", " << robotBlob.yRmax
                        << endl

                    // 1st buffer will be wrong, but it's ok because 1st blob
                    // must pass the condition below anyway.
                    << debugHead << "    bufferX/Y              : "
                        << bufferX << ", " << bufferY << endl;
            }


            if (isFirstBlob ||         // First blob should always pass this condition. 
                  (
                      (robotBlob.xRmin <= (xMax + bufferX)) // buffer from last red blob.
                   && (   robotBlob.yRmax >= (yMin - bufferY)
                       && robotBlob.yRmin <= (yMax + bufferY)
                      )
                  )
               ) {
                  

                // Forget about infrared now. 
                // ||    (infaredDistance / 10000.0 < 15 && 
                //     DarkBackgroundInBetween(robotBlob.getXMin(), 
                //                             uxMax, utop, ubottom))
 

                if (debugFindColorRobot && flag_OFFLINE) {
                    if (isFirstBlob)
                        (*outfile) << debugHead
                            << "...First blob, new robot" << endl << endl;
                    else
                        (*outfile) << debugHead 
                            << "...Blob get taken into current bounding box."
                            << endl << endl;
                }

                isFirstBlob = false;


                xMin = MIN(xMin, robotBlob.xRmin);
                xMax = MAX(xMax, robotBlob.xRmax);
                yMin = MIN(yMin, robotBlob.yRmin);
                yMax = MAX(yMax, robotBlob.yRmax);

                uxMin = MIN(uxMin, robotXMin);
                uxMax = MAX(uxMax, robotXMax);
                uyMin = MIN(uyMin, robotYMin);
                uyMax = MAX(uyMax, robotYMax);

                robot->cf = 1000;

                if (color == cROBOT_BLUE) {
                    blueRobotMixedArea[vobNum-vobBlueDog] += robotBlob.area;
                    blueRobotMaybeArea[vobNum-vobBlueDog] += robotBlob.maybeArea;
                }

            /* This blob is possibly a new dog (if it passes few checks) */
            } else {

                xMin = robotBlob.xRmin;
                xMax = robotBlob.xRmax;
                yMin = robotBlob.yRmin;
                yMax = robotBlob.yRmax;

                uxMin = robotXMin;
                uxMax = robotXMax;
                uyMin = robotYMin;
                uyMax = robotYMax;

                // SANITY CHECK ----------------------------------------------
                /* Check if 3 bounding box already drawn. */

                if ((color == cROBOT_RED && vobNum >= vobRedDog4)
                    || (color == cROBOT_BLUE && vobNum >= vobBlueDog4)
                    ) {
                    if (debugFindColorRobot && flag_OFFLINE) {
                        (*outfile) << debugHead
                            << "...have 3 dogs already!" 
                            << endl << endl;
                    }
                    break;
                }

                // SANITY CHECK ----------------------------------------------
                /* If last dog's bounding box is very tiny, it should be a
                ** false dog, so override its details using current bounding
                ** box. */
                double boundingBoxArea = robot->width * robot->height;

                if (boundingBoxArea < ROBOT_AREA_THRESHOLD) {

                    if (debugFindColorRobot && flag_OFFLINE) {
                        (*outfile) << debugHead
                            << "...last robot bounding box was far from this blob and too small. " 
                                << "(" << boundingBoxArea << ")"<< endl
                            << debugHead << "   Delete last one and form a new dog using current blob."
                            << endl << endl;
                    }

                } else {

                    if (debugFindColorRobot && flag_OFFLINE) {
                        (*outfile) << debugHead 
                            << "...Starting a new dog! " << endl
                            << debugHead << "    uxMin/ax, uyMin/ax: "
                                << uxMin << ", " << uxMax << ", "
                                << uyMin << ", " << uyMax << endl 
                            << endl;
                    }
                    vobNum++;
                }

                if (color == cROBOT_BLUE) {
                    blueRobotMixedArea[vobNum-vobBlueDog] += robotBlob.area;
                    blueRobotMaybeArea[vobNum-vobBlueDog] += robotBlob.maybeArea;
                }


                robot = &vob[vobNum];
                robot->cf = 1000;

                totalArea = 0;
                numberOfBlobs = 0;

            }

            robot->x = uxMin;
            robot->y = uyMin;
            robot->width = uxMax - uxMin;
            robot->height = uyMax - uyMin;

            robot->xRmin = xMin;
            robot->xRmax = xMax;
            robot->yRmin = yMin;
            robot->yRmax = yMax;


            /* TODO: ANYTHING WRONG HERE? */
            if (numberOfBlobs >= MAX_BLOBS_PER_DOG)
                continue;

            totalArea += robotBlob.area;

            numberOfBlobs++;

            /* Check if robot is really up close. */
            if ((totalArea > 10000)
                || ((numberOfBlobs <= 2) && (totalArea > 5000))){

                robot->dist = 20; // Arbitary value for behaviour to know that
                                  // robot is really up close.
            } 
        }
    }

    /* Same check as the one above, but after the loop, so that no small dog 
       will be accepted. */
    if (robot != NULL){

        double boundingBoxArea = robot->width * robot->height; 

        if (boundingBoxArea < ROBOT_AREA_THRESHOLD) {

            if (debugFindColorRobot && flag_OFFLINE) {
                (*outfile) << debugHead
                    << "...last robot bounding box too small ("
                    << boundingBoxArea << ") Delete it! "
                    << endl << endl;
            }

            robot->cf = 0;
            vobNum--;
        }
    }


    /* Reset vobNum because it's been changed in loop above. */ 
    vobNum = (color == cROBOT_RED) ? vobRedDog : vobBlueDog;


    /* Before anything is done, filter out blue robot that has too much
    ** maybeBlue. */
    for (int i = vobBlueDog; i < vobBlueDog4; i++) {

        if (vob[i].cf > 0) {

            int    dogNum     = i - vobBlueDog;
            double maybeRatio = (double) blueRobotMaybeArea[dogNum]
                               /(double) blueRobotMixedArea[dogNum];

            if (debugFindColorRobot && flag_OFFLINE) {
                (*outfile) << debugHead
                    << "    -- dog (x,y,width,height): "
                        << robot->x << ", "
                        << robot->y << ", "
                        << robot->width << ", "
                        << robot->height << endl 
                    << "    -- maybeRatio: "
                        << maybeRatio << endl
                        << endl;
            }

            if (maybeRatio > 0.7) {
                robot->cf = 0;
            }
        }
    }



    /* Robots have been found, now estimates its distance and heading. */ 
    for (int i = vobNum, robotNum = 0;
            i < vobNum + ROBOTS_EACH_COLOUR;
            i++, robotNum++) {

        robot = &vob[i];

        if (robot->cf != 0) {

            robot->cx = robot->x + robot->width / 2.0;
            robot->cy = robot->y + robot->height / 2.0;
            robot->cf = (int) (robot->height * 1000 / 144.0);

            robot->cRx = (robot->xRmin + robot->xRmax) / 2.0;
            robot->cRy = (robot->yRmin + robot->yRmax) / 2.0;

            robot->projected_x = robot->x;
            robot->projected_y = robot->y;
            robot->projected_w = robot->width;
            robot->projected_h = robot->height;
            robot->tilt = 0;        // not being used

            robot->imgHead  = PointToHeading(robot->cx);
            robot->imgElev  = PointToElevation(robot->cy);
            robot->head = 0;


            /* This means that the distance has been checked and it was really
            ** close, don't overwrite it with projection forumla. */
            if (robot->dist == 20) {

                if (debugFindColorRobot && flag_OFFLINE) {
                    (*outfile) << debugHead 
                        << "DISTANCE: Large area, set robot dist to 20."
                        << endl << endl;
                }

            } else {

                double ipx = TOPLEFT2CENTER_X(robot->x + robot->width/2);
                double ipy = TOPLEFT2CENTER_Y(robot->y + robot->height);
                ipy = ipy - 10;
                double projectedX, projectedY;

                getPointProjection(&projectedX, &projectedY, degPan, degTilt + degBodyTilt, degCrane, ipx, ipy,
                               FACE_LENGTH * 10.0, NECK_LENGTH * 10.0,
                               neckBaseHeight,
                               APERTURE_DISTANCE, CAMERA_OFFSET, false);

                robot->dist = DISTANCE(projectedX/10, projectedY/10, 0, 0);

                /* If projected point gives large distance, get the min of that 
                 * with distance estimation from bouding height 
                 * 200cm is the distance that the dog almost can't see the red
                 * patch in the paw. */

                if (ABS(robot->dist) > 200) {

                    double distanceFromHeight = ROBOT_HEIGHT_DISTANCE_CONSTANCE / robot->height;
                    robot->dist = MIN(robot->dist, distanceFromHeight);

                    if (debugFindColorRobot && flag_OFFLINE) {
                        (*outfile) << debugHead 
                            << " DISTANCE: PointProj gives large dist, use bounding height "
                            << robot->height << endl << endl;
                    }

                } else {

                    if (debugFindColorRobot && flag_OFFLINE) {
                        (*outfile) << debugHead 
                            << " DISTANCE: Use PointProjection method." 
                            << endl << endl;
                    }
                }
            }
            /* Get the heading and elevation relative to neck base. */
            RobotRelative(*robot);


            /* Let say, the closer the robot is, the less variance it has. */
            double varEstimate = robot->dist;
            robot->var = varEstimate;

            /* Use a fixed angular variance for the moment */
            robot->angleVar = 100.0;

        }
        /* An equation was here to calculate robot distance using height of
        ** bounding box, removed since rev 2567.  */
    }

}


/* Filter blobs which have robot colour and are "unusual", by setting their cf
** to be 0. */
void VisualCortex::RobotNoiseFilter() {

    const string debugHead = "RobotNoiseFilter| "; 

    /* Go through both red and blue colour. */
	for (int robotColorIndex = cROBOT_BLUE; robotColorIndex <= cROBOT_RED; robotColorIndex++) {

        if (debugRobotNoiseFilter && flag_OFFLINE) {
            (*outfile) << debugHead 
                << "Filtering " 
                << ColorName[robotColorIndex] << " blobs, total = " 
                << blobs->count(robotColorIndex) << endl 
                << endl;
        }

        /* Go through all blobs of each colour. */
		for (int blobIndex = 0; 
                blobIndex < blobs->count(robotColorIndex); 
                blobIndex++) {

			BlobInfo     &robotBlob  = blobs->getBlob(robotColorIndex, blobIndex);

            int          xMin       = robotBlob.getXMin();
            int          yMin       = robotBlob.getYMin();
            int          xMax       = robotBlob.getXMax();
            int          yMax       = robotBlob.getYMax();
            int          area       = robotBlob.area;
            int          maybeArea  = robotBlob.maybeArea;
            double       maybeRatio = (double) maybeArea / (double) area;


            if (debugRobotNoiseFilter && flag_OFFLINE) {
                (*outfile) << debugHead 
                    << "Checking " << ColorName[robotColorIndex] 
                    << " blob#" << robotBlob.blob_number
                        << endl
                    << debugHead << "    -- xMin/ax, yMin/ax: " 
                        << xMin << ", " << xMax << ", "
                        << yMin << ", " << yMax << endl
                    << debugHead << "    -- area, maybeArea: "
                        << area << ", " << maybeArea << endl
                    << debugHead << "    -- maybeArea/area: "
                        << maybeRatio << endl
                    << endl; 
           }


            // SANITY CHECK --------------------------------------------------

            // Use maybe colour to help filtering out black object that has
            // blue on it. 
            if (robotColorIndex == cROBOT_BLUE
                    && maybeRatio > 0.9) {

                if (debugRobotNoiseFilter && flag_OFFLINE) {
                    (*outfile) << debugHead 
                        << "...blue blob with too much maybeArea! ("
                        << maybeRatio << ")" << endl
                        << endl;
                }
                robotBlob.ResetBlob();
                continue;
            }

			VisualObject patch;
			VisualObject beacon;

			patch.imgElev = PointToElevation(robotBlob.cRy); 

            //#define detail_robot_noise_beacon_elev_check
            #ifdef detail_robot_noise_beacon_elev_check
            if (debugRobotNoiseFilter && flag_OFFLINE) {
                (*outfile) 
                    << debugHead << "    Checking elevation..." << endl
                    << debugHead << "    Robot cRx/y: "
                        << robotBlob.cRx << ", " << robotBlob.cRy << endl
                    << debugHead << "    Robot blob rotated imgElev: "
                        << patch.imgElev << endl;
            }
            #endif // detail_robot_noise_beacon_elev_check


            // SANITY CHECK --------------------------------------------------
            /* Check if there're robot blobs above the beacons. */

			for (int beaconIndex = vobBlueLeftBeacon; 
                    beaconIndex <= vobYellowRightBeacon; 
                    beaconIndex++) {

                if (vob[beaconIndex].cf > 0) {

                    beacon.imgElev = PointToElevation(vob[beaconIndex].cx,
                            vob[beaconIndex].cy,
                            sin_eroll,
                            cos_eroll);

                    #ifdef detail_robot_noise_beacon_elev_check
                    if (debugRobotNoiseFilter && flag_OFFLINE) {
                        (*outfile) 
                            << debugHead << "    " << vobName[beaconIndex]
                                << "'s rotated imgElev: " 
                                << beacon.imgElev << endl;
                    }
                    #endif // detail_robot_noise_beacon_elev_check

                    if (patch.imgElev > beacon.imgElev) {

                        if (robotColorIndex == cROBOT_BLUE) {
                            fireSanity(BLUE_BLOB_ABOVE_BEACON + beaconIndex - vobBlueLeftBeacon);

                        } else if (robotColorIndex == cROBOT_RED) {
                            fireSanity(RED_BLOB_ABOVE_BEACON + beaconIndex - vobBlueLeftBeacon);
                        }

                        if (debugRobotNoiseFilter && flag_OFFLINE) {
                            (*outfile) << debugHead 
                                << "...blob filtered due to above beacons" 
                                << endl << endl;
                        }
                        robotBlob.ResetBlob();
                        break;
                    }
                }
			}

            /* No need to do furthur check if blob already reset. */
            if (robotBlob.area == 0) {
                continue;
            }


            // SANITY CHECK --------------------------------------------------
            /* Check if there're robot blobs above the goal. */

			for (int goalIndex = vobBlueGoal; goalIndex <= vobYellowGoal; goalIndex++) {

				if (vob[goalIndex].cf > 0) {

					if (patch.imgElev > PointToElevation(vob[goalIndex].y)) {

						if (robotColorIndex == cROBOT_BLUE) {
							fireSanity(BLUE_BLOB_ABOVE_GOAL + goalIndex - vobBlueGoal);
						}
						else if (robotColorIndex == cROBOT_RED) {
							fireSanity(RED_BLOB_ABOVE_GOAL + goalIndex - vobBlueGoal);
						}

                        if (debugRobotNoiseFilter && flag_OFFLINE) {
                            (*outfile) << debugHead 
                                << "...blob filtered due to above the goal." 
                                << endl << endl;
                        }
                        robotBlob.ResetBlob();
                        break;
					}
				}
			}

            /* No need to do furthur check if blob already reset. */
            if (robotBlob.area == 0) {
                continue;
            }


            // SANITY CHECK --------------------------------------------------
            /* Check elevation instead of horizon.
               Check once for each blob with large threshold (10) and once for each 
               formed dog later with small threshold (0) 
             */
            static const int ELEVATION_TOO_HIGH = 10;//degree
            double camElev = getElevation(degPan, degTilt+degBodyTilt, degCrane,
                               TOPLEFT2CENTER_X((xMin + xMin)/2) , 
                               TOPLEFT2CENTER_Y(yMax), 
                               FACE_LENGTH * 10, 
                               NECK_LENGTH * 10, 
                               APERTURE_DISTANCE, 
                               CAMERA_OFFSET); 

            if (camElev > ELEVATION_TOO_HIGH) {

                if (debugRobotNoiseFilter && flag_OFFLINE) {
                    (*outfile) << debugHead << "...Blob elevation too high : "
                        << "  cx = " << TOPLEFT2CENTER_X((xMin + xMin)/2)
                        << "  bottom y = " << TOPLEFT2CENTER_Y(yMax) 
                        << "  elev = " << camElev 
                        << endl << endl;
                }

                robotBlob.ResetBlob();
                continue;
            }
        

            // Doing the same thing as checking camElev, except camElev
            // formalise the degree instead of using cplane pixels, so check
            // camElev only should be enough.

            // SANITY CHECK --------------------------------------------------
            /* Check against horizon. */
            /*
            static const int ROBOT_GO_OVER_HORIZON_THRESHOLD = 15;

            if (!isBelowHorizon(
                        (xMin+xMax)/2,
                        (yMin+yMax)/2,
                        ROBOT_GO_OVER_HORIZON_THRESHOLD)) {

                if (robotColorIndex == cROBOT_BLUE) {
                    fireSanity(BLUE_BLOB_ABOVE_HEAD);
                } else if (robotColorIndex == cROBOT_RED) {
                    fireSanity(RED_BLOB_ABOVE_HEAD);
                }

                if (debugRobotNoiseFilter && flag_OFFLINE) {
                    (*outfile) << debugHead 
                        << "...blob filtered due to above horizon. " 
                        << endl << endl;
                }
                robotBlob.ResetBlob();
                continue;
            }
            */

            if (debugRobotNoiseFilter && flag_OFFLINE) {
                (*outfile) << debugHead
                    << "...Blob accepted!" 
                    << endl << endl;
            }
        }
    }
}

/* Find robots of all colour from CPlane. */
void VisualCortex::FindRobot() {

    static const string debugHead = "FindRobot| ";
    VisualObject &ball = vob[vobBall];

    /* get the xRmin/max, yRmin/max, cRx/y values. */ 
    blobs->rotate(cROBOT_RED, sin_eroll, cos_eroll);
    blobs->rotate(cROBOT_BLUE, sin_eroll, cos_eroll);

	RobotNoiseFilter();

    /* ----------------------------------------------------------------*/
    /* Cleans up red noise inside ball and red blobs that are too close. */
	double  bcx = 0.0,
            bcy = 0.0,
            ball_crx = 0.0,
            ball_cry = 0.0,
            ball_r = 0.0,
            leftX = 0.0,
            rightX = 0.0,
            topY = 0.0,
            bottomY = 0.0,
            BUFFER = 0.0;

    /* Get ball information for this sanity check. */
	if (ball.cf > 0) {
		// rotate the ball cx and cy
		bcx = (WIDTH / 2) - ball.cx;
		bcy = (HEIGHT / 2) - ball.cy;
		ball_crx    = (WIDTH / 2) - (cos_eroll * bcx - sin_eroll * bcy);
		ball_cry    = (HEIGHT / 2) - (sin_eroll * bcx + cos_eroll * bcy);
		ball_r      = ball.radius;
		leftX       = ball_crx - ball_r;
		rightX      = ball_crx + ball_r;
		topY        = ball_cry - ball_r;
		bottomY     = ball_cry + ball_r;
		BUFFER      = MAX((rightX - leftX) * 0.20, 6.0);
	}

    /* Sort the order of blobs array according to the cplane x-value, so that
    ** we can check blobs from left to right. */

    /* TODO: Daniel 15/6/2004 - It is noticed that the sorting sorts all blobs
    ** despite of the size, and then only the first BlobLimits.numBlobs blobs
    ** are accessed. So if the numBlobs limit is not large enough, some robot
    ** blobs could be missed. */
    blobs->qsort(cROBOT_RED);
    blobs->qsort(cROBOT_BLUE);


    // SANITY CHECK -------------------------------------------------------- 

    /* Filter red blobs near ball. */
    for (int j = 0; j < blobs->count(cROBOT_RED); j++) {
    	BlobInfo &redBlob = blobs->getBlob(cROBOT_RED, j);

        if (redBlob.area > 0) { 

            if (flag_OFFLINE && debugFindRobot) {
                (*outfile) << debugHead
                    << "Checking Red blob #" << redBlob.blob_number << ": (rotated) " 
                    << redBlob.xRmin << " " << redBlob.yRmin << " "
                    << redBlob.xRmax << " " << redBlob.yRmax << endl;
            }

            if (ball.cf > 0) {
                
                // clean up all red around and below the ball
                if (    redBlob.xRmin >= leftX   - BUFFER 
                     && redBlob.xRmax <= rightX  + BUFFER
                     && redBlob.yRmin >= topY    - BUFFER * 2
                     && redBlob.yRmax <= bottomY + BUFFER * 2
                   ) {

                    if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                        (*outfile) << debugHead 
                            << "... blob #" << redBlob.blob_number 
                            << " filtered because near ball." << endl << endl;

                        #define detail_redrobot_near_ball_check
                        #ifdef detail_redrobot_near_ball_check
                        (*outfile) 
                            << debugHead << "    RED blob (xRmin/ax, yRmin/ax): "
                                << redBlob.xRmin << ", " << redBlob.xRmax << ", "
                                << redBlob.yRmin << ", " << redBlob.yRmax << endl
                            << debugHead << "    Rotated BALL bounding box: " << endl
                            << debugHead << "        leftX,   leftX   -BUFFER  : "
                                << leftX << ", " << leftX-BUFFER << endl
                            << debugHead << "        rightX,  rightX  +BUFFER  : "
                                << rightX << ", " << rightX+BUFFER << endl
                            << debugHead << "        topY,    topY    -BUFFER*2: "
                                << topY << ", " << topY-BUFFER*2 << endl
                            << debugHead << "        bottomY, bottomY +BUFFER*2: "
                                << bottomY << ", " << bottomY+BUFFER*2 << endl
                                << endl;
                        #endif // detail_redrobot_near_ball_check
                    }

                    fireSanity(RED_BLOB_NEAR_BALL);
                    redBlob.ResetBlob();
                    continue;
                }
            }

            if (flag_OFFLINE && debugFindRobot) {
                (*outfile) << debugHead
                    << "...Ok! " << endl << endl;
            }
        }
    }


    /* Find the red robot first, then blue robot. */
    FindColorRobot(cROBOT_RED);


    /* Do some checks on blue blobs before identifying blue dogs. */
    for (int j = 0; j < blobs->count(cROBOT_BLUE); j++) {

        BlobInfo &blueBlob = blobs->getBlob(cROBOT_BLUE, j);

        if (blueBlob.area == 0) {
            continue;
        }

        double blueArea = (blueBlob.xRmax - blueBlob.xRmin) *
                          (blueBlob.yRmax - blueBlob.yRmin);

        if (flag_OFFLINE && debugFindRobot) {
            (*outfile) << debugHead
                << "Checking Blue blob #" << blueBlob.blob_number 
                    << ": (rotated xmin/ax ymin/ax, area) " 
                    << blueBlob.xRmin << " " << blueBlob.xRmax << " "
                    << blueBlob.yRmin << " " << blueBlob.yRmax << " "
                    << blueBlob.area << endl;
        }


        // SANITY CHECK -------------------------------------------------------- 

        /* Kill blue blobs near ball. */
        if (ball.cf > 0) {

            if (   blueBlob.xRmin >= leftX - BUFFER
                && blueBlob.xRmax <= rightX + BUFFER
                && blueBlob.yRmin >= topY - BUFFER * 2
                && blueBlob.yRmax <= bottomY + BUFFER * 2
                ) {

                fireSanity(BLUE_BLOB_NEAR_BALL);
                blueBlob.ResetBlob();

                if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                    (*outfile) << debugHead
                        << "...blue blob near ball filtered! Rotated coords: " 
                            << endl << endl;

                    // #define detail_bluerobot_near_ball_check
                    #ifdef detail_bluerobot_near_ball_check
                    (*outfile) << debugHead << "    BLUE blob (xRmin/ax, yRmin/ax): "
                        << blueBlob.xRmin << ", " << blueBlob.xRmax << ", "
                        << blueBlob.yRmin << ", " << blueBlob.yRmax << endl
                        << debugHead << "    Rotated BALL bounding box: " << endl
                        << debugHead << "        leftX,   leftX   -BUFFER  : "
                            << leftX << ", " << leftX-BUFFER << endl
                        << debugHead << "        rightX,  rightX  +BUFFER  : "
                            << rightX << ", " << rightX+BUFFER << endl
                        << debugHead << "        topY,    topY    -BUFFER*2: "
                            << topY << ", " << topY-BUFFER*2 << endl
                        << debugHead << "        bottomY, bottomY +BUFFER*2: "
                            << bottomY << ", " << bottomY+BUFFER*2 << endl
                            << endl;
                    #endif // detail_bluerobot_near_ball_check
                }
                continue;
            }
        }

        /* Remove blue dog blob around red dog as well, because
         * head/shadows/paw of red dogs can turn to blue dogs. */
        for (int redRobotIndex = vobRedDog; redRobotIndex <= vobRedDog4; redRobotIndex++) {

            double redLeftX     = vob[redRobotIndex].xRmin;
            double redRightX    = vob[redRobotIndex].xRmax;
            double redTopY      = vob[redRobotIndex].yRmin;
            double redBottomY   = vob[redRobotIndex].yRmax;
            double redWidth     = (redRightX  - redLeftX);
            double redHeight    = (redBottomY - redTopY);
                                  // MAX((redRightX - redLeftX) * 0.20, 5.0);

            double redArea  = redWidth * redHeight;

            /* Red dog needs to be bigger than blue blob to assume blue dog might
               be a noise. */
            if (vob[redRobotIndex].cf > 0 && redArea > blueArea) {

                // #define detail_blue_around_red_dog_check
                #ifdef detail_blue_around_red_dog_check
                (*outfile) << debugHead << "----Check aginst red dog ("
                    << redRobotIndex << "):" << endl

                    << debugHead << "    Red dog:" << endl
                    << debugHead << "        xMin/ax, yMin/ax: "
                        << redLeftX << ", " << redRightX << ", "
                        << redTopY  << ", " << redBottomY << endl
                    << debugHead << "        width, height: "
                        << redWidth << ", " << redHeight << endl
                        << endl;
                #endif // detail_blue_around_red_dog_check


                // SANITY CHECK -------------------------------------------------------- 

                /* Blue blob within red dog horizontally. */
                if (   blueBlob.cRx >= redLeftX
                    && blueBlob.cRx <= redRightX
                    && blueBlob.cRy >= redTopY      - redHeight * 0.2
                    && blueBlob.cRy <= redBottomY   + redHeight * 0.5
                   ) {

                    if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                        (*outfile) << debugHead
                            << "...blue blob within red dog horizontally." 
                            << endl << endl;
                    }

                    fireSanity(BLUE_BLOB_AROUND_RED_ROBOT + redRobotIndex - vobRedDog);
                    blueBlob.ResetBlob();
                    break;


                // SANITY CHECK -------------------------------------------------------- 

                /* Blue blob within red dog vertically (+some below) */
                } else if (blueBlob.cRy >= redTopY
                        && blueBlob.cRy <= redBottomY + redHeight * 0.2
                        && blueBlob.cRx >= redLeftX   - redWidth * 0.5
                        && blueBlob.cRx <= redRightX  + redWidth * 0.5
                   ) {

                    if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                        (*outfile) << debugHead
                            << "...blue blob within red dog vertically (or slightly below)." 
                            << endl << endl;
                    }

                    fireSanity(BLUE_BLOB_AROUND_RED_ROBOT + redRobotIndex - vobRedDog);
                    blueBlob.ResetBlob();
                    break;
                }

            }
        }
        if (blueBlob.blob_number == -1) {
            continue;
        }

        if (flag_OFFLINE && debugFindRobot) {
            (*outfile) << debugHead
                << "...ok!" << endl << endl;
        }
    }

    FindColorRobot(cROBOT_BLUE);


    // SANITY CHECK -------------------------------------------------------- 

    /* Filter out red robots in blue, and vice versa. */
    for (int redDogIndex = vobRedDog; redDogIndex <= vobRedDog4; redDogIndex++) {
        if (vob[redDogIndex].cf > 0) {

            for (int blueDogIndex = vobBlueDog; blueDogIndex <= vobBlueDog4;
                    blueDogIndex++) {

                if (vob[blueDogIndex].cf > 0) {

                    int rxmin = (int) (vob[redDogIndex].x);
                    int rxmax = (int) (vob[redDogIndex].x + vob[redDogIndex].width);
                    int rymin = (int) (vob[redDogIndex].y);
                    int rymax = (int) (vob[redDogIndex].y + vob[redDogIndex].height);

                    int bxmin = (int) (vob[blueDogIndex].x);
                    int bxmax = (int) (vob[blueDogIndex].x + vob[blueDogIndex].width);
                    int bymin = (int) (vob[blueDogIndex].y);
                    int bymax = (int) (vob[blueDogIndex].y + vob[blueDogIndex].height);

                    if (rxmin > bxmin && rxmax <bxmax && rymin> bymin && rymax < bymax) {
                        // check red in blue
                        vob[redDogIndex].reset();

                        if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                            (*outfile) << debugHead
                                << "...red robot within blue robot." << endl;
                        }

                    } else if (bxmin > rxmin
                            && bxmax <rxmax
                            && bymin> rymin
                            && bymax < rymax) {
                        // check blue in red
                        vob[blueDogIndex].reset();

                        if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                            (*outfile) << debugHead
                                << "...blue robot within red robot." << endl;
                        }
                    }
                }
            }
        }
    }

    /* CHECK --------------------------------------------------------------- */
    /* Other logical sanity check after probable red and blue dogs. */
    for (int dogIndex = vobRedDog; dogIndex <= vobBlueDog4; dogIndex++) {

        VisualObject& dog = vob[dogIndex];

        if (dog.cf <= 0) {
            dog.cf = 0;
            continue;
        }

        dog.lhead = dog.head
                            + (PointToHeading(dog.x) 
                                - PointToHeading(dog.cx));

        dog.rhead = dog.head
                            + (PointToHeading(dog.x + dog.width)
                                - PointToHeading(dog.cx));

        double dogMinX = dog.x;
        double dogMaxX = dog.x + dog.width; 
        double dogMinY = dog.y;
        double dogMaxY = dog.y + dog.height; 

        double dogCX   = dog.x + dog.width/2; 
        double dogCY   = dog.y + dog.height/2.; 

        if (debugFindRobot && flag_OFFLINE) {
            (*outfile)
                << debugHead << "Checking FORMED dog:" << endl
                << debugHead << "    minx/y, maxx/y: "
                    << dogMinX << "," << dogMinY << "," 
                    << dogMaxX << "," << dogMaxY << endl
                << debugHead << "    width/height  : "
                    << dog.width << ", " << dog.height 
                    << endl << endl;
        }


        // SANITY CHECK -------------------------------------------------------- 

        // Second elevation check for dog, which is stricter, any dog that has height > 10cm is bad
        // elevation is of the bottom edge of the robot bounding box.
        // note that elevation is relative to the camera.

        static const double DOG_HEIGHT_THRESHOLD = 250;//mm, this is pretty strict, coz we don't want false robot

        /* Unrotate the rotated robot cx/topY about center of cplane. */
        /* This can reduce error caused by rotated bounding box. */
        int uDogCX     = unrotatePointReturnX(dog.cRx, dog.yRmin); 
        int uDogTopY   = unrotatePointReturnY(dog.cRx, dog.yRmin); 

        // dogHeight is relative to the neckBase.
        double dogHeight = getHeightKnowingDistance(degPan, degTilt + degBodyTilt, degCrane,
                                TOPLEFT2CENTER_X(uDogCX),
                                TOPLEFT2CENTER_Y(uDogTopY),
                                dog.dist * 10, 
                                FACE_LENGTH * 10 , 
                                NECK_LENGTH * 10, 
                                APERTURE_DISTANCE, 
                                CAMERA_OFFSET);

        #define detail_formed_dog_height_check
        #ifdef detail_formed_dog_height_check
        if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {

            double camElev = getElevation(degPan, degTilt + degBodyTilt, degCrane,
                                    TOPLEFT2CENTER_X(uDogCX),
                                    TOPLEFT2CENTER_Y(uDogTopY), 
                                    FACE_LENGTH * 10, 
                                    NECK_LENGTH * 10, 
                                    APERTURE_DISTANCE, 
                                    CAMERA_OFFSET);

            (*outfile) << debugHead
                << "---Detail height check info: " << endl
                << debugHead << "    uDogCX, uDogTopY : " 
                    << uDogCX << ", "
                    << uDogTopY << endl
                << debugHead << "    dist             : " << dog.dist * 10 << endl
                << debugHead << "    camElev          : " << camElev << endl
                << debugHead << "    dog's height     : " 
                    << neckBaseHeight + dogHeight
                    << " (mm) " << endl
                    << endl;
        }
        #endif // detail_formed_dog_height_check

        if (dogHeight + neckBaseHeight > DOG_HEIGHT_THRESHOLD){
                /*
                cout << "  cx = " << TOPLEFT2CENTER_X((dogMinX + dogMaxX)/2 )
                << "  bottom y = " << TOPLEFT2CENTER_Y(dogMaxY) 
                << "  robot dist = " << dog.dist * 10
                << "  pan = " << degPan 
                << "  tilt = " << degTilt + degBodyTilt 
                << "  crane = " << degCrane 
                << "  height = " << dogHeight 
                << endl;
                */
                
                if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                    (*outfile) << debugHead
                        << "...Robot bottom edge height > 25cm"
                        << endl << endl;
                }
                dog.reset();
                fireSanity(LAZY_SANITY_MESSAGE1);
                continue;
        }


        // SANITY CHECK -------------------------------------------------------- 

        // A blue robot in corner is probably just a ring effect, not real dog.
        if (dogIndex >= vobBlueDog && dogIndex <= vobBlueDog4) {

            if ((dog.cx < 20 && dog.cy < 20) ||
                (dog.cx < 20 && dog.cy > 140) ||
                (dog.cx > 188 && dog.cy < 20) ||
                (dog.cx > 188 && dog.cy > 140))
            {
                dog.reset();

                if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                    (*outfile) << debugHead
                        << "...blue robot in corner, likely to be a ring effect." 
                        << endl;
                }
                fireSanity(BLUE_ROBOT_IN_CORNER);
                continue;
            }
        }


        // SANITY CHECK -------------------------------------------------------- 

        // A robot cannot be too high up the ball. 
        static const int DOG_ABOVE_BALL_ELEV_THRESHOLD = 10;

        if (ball.cf > 0 && ball.elev > dog.elev +
                DOG_ABOVE_BALL_ELEV_THRESHOLD) {

            dog.reset();

            if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead
                    << "...dog lower than ball, remove dog!"
                    << endl << endl;
            }
            fireSanity(LAZY_SANITY_MESSAGE1);
            continue;
        }        


        // SANITY CHECK -------------------------------------------------------- 

        /* Do not want dogs over 3m away. */
        if (dog.dist > 300) {

            dog.reset();

            if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead
                    << "...Distance is too large = " << dog.dist
                    << endl << endl;
            }
            fireSanity(LAZY_SANITY_MESSAGE1);
            continue;
        }        


        // SANITY CHECK -------------------------------------------------------- 

        /* Check the aspect ratio of the dog's bounding box, in attempt to 
         * filter out blue dog steaks from edges of white and background. */
        if (   (dog.width * 1.0) / dog.height >= 7.5
            || (dog.height * 1.0) / dog.width >= 7.5
           ) {

            dog.reset();

            if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                (*outfile) << debugHead
                    << "...Aspect Ratio screwed." 
                    << endl << endl;
            }
            fireSanity(LAZY_SANITY_MESSAGE1);
            continue;
        }        


        /* Check against beacons ------------------------------------------ */
        bool resetByBeacon = false;

        for (int beaconIndex = vobBlueLeftBeacon; beaconIndex <= vobYellowRightBeacon; 
                beaconIndex++) {

            VisualObject &beacon = vob[beaconIndex];

            if (beacon.cf > 0) {

                if (debugFindRobot && flag_OFFLINE) {
                    (*outfile) << debugHead
                        << "    Checking " << vobName[beaconIndex] << " against dogs..."
                        << endl;
                    (*outfile) 
                        << debugHead << "    beacon (xmin/ax, ymin/max: "
                            << beacon.x << ", " << beacon.x + beacon.width << ", "
                            << beacon.y << ", " << beacon.y + beacon.height << endl
                            << endl; 
                }



                // SANITY CHECK -----------------------------------------------

                /* Beacons should be further away than robot, added 20 cm 
                 * leniency due to crappy robot distance. */
                static const int ROBOT_AWAY_FROM_BEACON_THRESHOLD = 20;
                if (beacon.dist + ROBOT_AWAY_FROM_BEACON_THRESHOLD 
                        < dog.dist) {

                    // I don't think we need to check heading... just check
                    // anytime nomatter what heading is. 
                    // ABS(beacon.head - dog.head) < 10 && 

                    if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                        (*outfile) << debugHead
                            << "... dog " << dogIndex 
                            << " is furthur than beacon. Set var to LARGE_VAL"
                            << endl;
                    }
                    dog.var = LARGE_VAL;

                    // No breaking - who knows if this robot will break
                    // another check down the road?
                }
		

                // SANITY CHECK -----------------------------------------------

                /* Check if dog is flying above beacon. */
                static const int DOG_HIGHER_THAN_BEACON_THRESHOLD = 0;

                // Check base rather than cy because there can be noise just
                // around the beacon. 
                double dogBaseElev = PointToElevation(dog.x + dog.width/2,
                                                      dog.y + dog.height,
                                                      sin_eroll, cos_eroll);

                double beaconElev = PointToElevation(beacon.x + beacon.width/2,
                                                     beacon.y + beacon.height,
                                                     sin_eroll, cos_eroll);

                if (dogBaseElev - beaconElev > DOG_HIGHER_THAN_BEACON_THRESHOLD) {

                    if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                        (*outfile) << debugHead
                            << "...dog is higher than beacon." << endl;
                    }
                    dog.reset();
                    resetByBeacon = true;
                    fireSanity(ROBOT_BLOB_TOO_HIGH_TO_BEACON + beaconIndex - vobBlueLeftBeacon);
                    break;
                } 


                /* Reset the beacon if it is totally inside a red dog's
                 * bounding box. */

                if (dogIndex >= vobRedDog && dogIndex <= vobRedDog4
                        && beacon.x > dogMinX
                        && beacon.x + beacon.width < dogMaxX
                        && beacon.y > dogMinY
                        && beacon.y + beacon.height < dogMaxY
                    ) {

                    beacon.reset();

                    if (debugFindRobot && flag_OFFLINE) {
                        (*outfile) << debugHead
                            << "...beacon is totally inside red dog, sanity out!"
                            << endl;
                    }
                    fireSanity(LAZY_SANITY_MESSAGE1);
                }
            }
        }
        if (resetByBeacon) { 
            continue;
        }


        /* Check against Goals ------------------------------------------ */
        bool resetByGoal = false;

        for (int goalIndex = vobBlueGoal; goalIndex <= vobYellowGoal;
                goalIndex++) {

            VisualObject& goal = vob[goalIndex];

            if (goal.cf > 0) {

                // SANITY CHECK -----------------------------------------------

                /* Goal should be further away than robot, added 20 cm 
                 * leniency due to crappy robot distance. */
                static const int ROBOT_AFTER_GOAL_THRESHOLD = 20;

                if (goal.dist + ROBOT_AFTER_GOAL_THRESHOLD 
                        < dog.dist) {

                    if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                        (*outfile) << debugHead
                            << "...Dog is too furthur than the "
                            << vobName[goalIndex] << endl;
                    }

                    // No breaking - who knows if this robot will break
                    // another check down the road?
                    dog.var = LARGE_VAL;
                }
	

                // SANITY CHECK -----------------------------------------------

                /* Robot cannot be too high above goal. */
                static const int ROBOT_ABOVE_GOAL_THRESHOLD = 5;
                if (dog.elev > 
                        goal.elev + ROBOT_ABOVE_GOAL_THRESHOLD) {

                    if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                        (*outfile) << debugHead
                            << "...dog is higher than "
                            << vobName[goalIndex] << endl;
                    }

                    dog.reset();
                    resetByGoal = true;
                    fireSanity(ROBOT_BLOB_TOO_HIGH_TO_GOAL + goalIndex - vobBlueGoal);
                    break;
                }


                // SANITY CHECK (FOR GOAL) ------------------------------------

                /* A goal cannot be completely inside a dog. */

                /* Unrotate goal corners about center of cplane. */
                int uGoal_x1 = unrotatePointReturnX(goal.x, goal.y);
                int uGoal_y1 = unrotatePointReturnY(goal.x, goal.y);

                int uGoal_x2 = unrotatePointReturnX(goal.x + goal.width, goal.y);
                int uGoal_y2 = unrotatePointReturnY(goal.x + goal.width, goal.y);

                int uGoal_x3 = unrotatePointReturnX(goal.x + goal.width, goal.y + goal.height);
                int uGoal_y3 = unrotatePointReturnY(goal.x + goal.width, goal.y + goal.height);

                int uGoal_x4 = unrotatePointReturnX(goal.x, goal.y + goal.height);
                int uGoal_y4 = unrotatePointReturnY(goal.x, goal.y + goal.height);

                double goalMinX = MIN(MIN(MIN(uGoal_x1, uGoal_x2), uGoal_x3), uGoal_x4);
                double goalMaxX = MAX(MAX(MAX(uGoal_x1, uGoal_x2), uGoal_x3), uGoal_x4);
                double goalMinY = MIN(MIN(MIN(uGoal_y1, uGoal_y2), uGoal_y3), uGoal_y4);
                double goalMaxY = MAX(MAX(MAX(uGoal_y1, uGoal_y2), uGoal_y3), uGoal_y4);

                if (goalMinX > dogMinX
                        && goalMinY > dogMinY
                        && goalMaxX < dogMaxX
                        && goalMaxY < dogMaxY) {

                    goal.reset();
                    fireSanity(GOAL_INSIDE_ROBOT);

                    if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
                        (*outfile) << debugHead
                            << "..."
                            << vobName[goalIndex]
                            << " cannot be inside a robot!" << endl;
                        (*outfile) << debugHead 
                            << "    goal coords: " 
                            << goalMinX << " " << goalMinY << " " 
                            << goalMaxX << " " << goalMaxY << " "  
                            << endl;
                    }
                }
            }
        }
        if (resetByGoal) {
            continue;
        }


        /* Checks about balls ---------------------------------------------- */

        if (ball.cf > 0 
                && dogIndex >= vobRedDog && dogIndex <= vobRedDog4) {

            /* Do not want to trust modified bounding box. */
            double ballArea = ball.owidth * ball.oheight;
            double dogArea  = dog.width * dog.height;


            double ball2dogRatio = ballArea / dogArea;

            static const double BALL2DOG_RATIO_THRESHOLD = 400.0 / 8000.0; 
            static const int    BALL_NEAR_DOG_THRESHOLD  = 10;

            if (debugFindRobot && flag_OFFLINE) {
                (*outfile) << debugHead 
                    << "    Checking ball against robot..."
                    << endl;
                (*outfile) 
                    << debugHead << "    ball original (x,y,w,h): "
                        << ball.ox << ", " << ball.oy << ", " 
                        << ball.owidth << ", " << ball.oheight << endl 
                        << endl;
            }

            // SANITY CHECK (FOR BALL) ----------------------------------------

            // A ball beside a red robot, if ball is very small compare to it, 
            // ball is probably a noise.
            if (ball2dogRatio < BALL2DOG_RATIO_THRESHOLD
                    && ball.ocx > dog.x - BALL_NEAR_DOG_THRESHOLD
                    && ball.ocx < ((dog.x+dog.width)  + BALL_NEAR_DOG_THRESHOLD)
                    && ball.ocy > dog.y - BALL_NEAR_DOG_THRESHOLD
                    && ball.ocy < ((dog.y+dog.height) + BALL_NEAR_DOG_THRESHOLD)
                ) {

                ball.reset();
                fireSanity(LAZY_SANITY_MESSAGE2);


            // SANITY CHECK (FOR BALL) ----------------------------------------

            // If the ball is inside a red dog, and not in the lower part of
            // it, it's probably some phantom orange of its body.
            } else if (ball.ox >= dogMinX
                    && ball.ox + ball.owidth <= dogMaxX
                    && ball.oy >= dogMinY
                    && ball.oy + ball.oheight <= dogMaxY) {

                // Check if ball is not around the lower part of the dog. 
                double ballElev        = PointToElevation(ball.ocx, ball.ocy, sin_eroll, cos_eroll);
                double dogCentroidElev = PointToElevation(dog.cx, dog.cy, sin_eroll, cos_eroll);
                double dogMinElev      = getMinElevation(dog);

                if (debugFindRobot && flag_OFFLINE) {
                    (*outfile) << debugHead 
                        << "    Checking ball inside robot..."
                        << endl;
                    (*outfile) 
                        << debugHead << "    ballElev: "   << ballElev << endl 
                        << debugHead << "    dogCenElev: " << dogCentroidElev << endl 
                        << debugHead << "    dogMinElev: " << dogMinElev << endl 
                        << endl;
                }


                if (ballElev > (dogCentroidElev + dogMinElev)/2) {
                    ball.reset();
                    fireSanity(LAZY_SANITY_MESSAGE2);

                    if (debugFindRobot && flag_OFFLINE) {
                        (*outfile) 
                            << debugHead << "...ball reset because it is totally inside the bounding box" 
                                << " and not lower than the dog."
                                << endl;
                    }
                }
            }
        }

        
        if (debugFindRobot && flag_OFFLINE && !flag_PRE_SANITY_BLOB) {
            (*outfile) 
                << debugHead << "...Finally dog (" << dogIndex  
                    << ") found!" << endl
                << debugHead << "    neckElev: " << dog.elev << endl
                << debugHead << "    camElev : " << 
                        getElevation(degPan,degTilt + degBodyTilt,degCrane,
                                TOPLEFT2CENTER_X(dog.cx) , 
                                TOPLEFT2CENTER_Y(dog.y + dog.height),
                                FACE_LENGTH * 10 , NECK_LENGTH * 10, 
                                APERTURE_DISTANCE, CAMERA_OFFSET) << endl
                << debugHead << "    head    : " << dog.head << endl
                << debugHead << "    dist    : " << dog.d 
                << endl << endl;
        }
    }
    

    // A hack added by Ted Wong, otherwise no robot distance will ever pass to behaviour and no backoff will
    // ever occured.
    if (vob[vobRedDog].cf > 0)
        vob[vobRedDog].hack = 9231;
    else
        vob[vobRedDog].hack = 0;

    if (vob[vobRedDog2].cf > 0)
        vob[vobRedDog2].hack = 9231;
    else
        vob[vobRedDog2].hack = 0;

    if (vob[vobRedDog3].cf > 0)
        vob[vobRedDog3].hack = 9231;
    else
        vob[vobRedDog3].hack = 0;
        
    if (vob[vobRedDog4].cf > 0)
        vob[vobRedDog4].hack = 9231;
    else
        vob[vobRedDog4].hack = 0;        

}


/* Given a rotated point, unrotate it back to what it originally is in the
 * cplane. */
inline int VisualCortex::unrotatePointReturnX(double x, double y) {
    return (xOrigin - (int) (  ucos_eroll*(xOrigin - x) 
                             - usin_eroll*(yOrigin - y))
           );
}

inline int VisualCortex::unrotatePointReturnY(double x, double y) {
    return (yOrigin - (int) (  usin_eroll*(xOrigin - x) 
                             + ucos_eroll*(yOrigin - y))
           );
}





/* Draw a line from 1 point to the other, measure the distance between the
** starting point and wherever the color changes to be different from the 
** specified color. */
inline int VisualCortex::calDistBaseOnColor(
        int Ax, int Ay, 
        int Bx, int By, 
        HalfBeacon *sourceBlob,
        uchar color,
        bool skipBoundary) {
    int numOfPixels = 0;

    //------------------------------------------------------------------------
    // INITIALIZE THE COMPONENTS OF THE ALGORITHM THAT ARE NOT AFFECTED BY THE
    // SLOPE OR DIRECTION OF THE LINE
    //------------------------------------------------------------------------
    int dX = abs(Bx - Ax);        // store the change in X and Y of the line endpoints
    int dY = abs(By - Ay);

    int CurrentX = Ax;             // store the starting point (just point A)
    int CurrentY = Ay;

    //------------------------------------------------------------------------
    // DETERMINE "DIRECTIONS" TO INCREMENT X AND Y (REGARDLESS OF DECISION)
    //------------------------------------------------------------------------
    int Xincr, Yincr;
    if (Ax > Bx) {
        Xincr = -1;
    } else {
        Xincr = 1;
    }                             // which direction in X?
    if (Ay > By) {
        Yincr = -1;
    }
    else {
        Yincr = 1;
    }                              // which direction in Y?

    /* This pinkNoise stuff is an experiment to see if it is better to count
    ** some red/orange as noise of pinkBlob, so accept them as part of
    ** beacon blob when doing c2c calculation.
    ** Setting to 0 means that we don't want to use this hack. */
    int pinkNoiseCounter = 0;
    static const int PINK_NOISE_LIMIT = 0;

    //------------------------------------------------------------------------
    // DETERMINE INDEPENDENT VARIABLE (ONE THAT ALWAYS INCREMENTS BY 1 (OR -1) )
    // AND INITIATE APPROPRIATE LINE DRAWING ROUTINE (BASED ON FIRST OCTANT
    // ALWAYS). THE X AND Y'S MAY BE FLIPPED IF Y IS THE INDEPENDENT VARIABLE.
    //------------------------------------------------------------------------
    if (dX >= dY) {
        // if X is the independent variable
        int dPr = dY << 1;              // amount to increment decision if right is chosen (always)
        int dPru = dPr - (dX << 1);     // amount to increment decision if up is chosen
        int P = dPr - dX;               // decision variable start value
        int thisColour;

        for (; dX >= 0; dX--) {

            thisColour = cplane[CurrentX + CurrentY * WIDTH] & ~MAYBE_BIT;

            /* Accept both cROBOT_GREY and cWHITE as background. */ 
            if (color == BACKGROUND
                && (thisColour == BACKGROUND || thisColour == cWHITE)
                ) {
                numOfPixels++;

            } else if (thisColour == color) {
                    numOfPixels++;

            /* Exception for pink - take red / ball orange as ok. */
            } else if (color == cBEACON_PINK 
                    && (thisColour == cROBOT_RED || thisColour == cBALL)
                    && (pinkNoiseCounter < PINK_NOISE_LIMIT)
                ) {
                pinkNoiseCounter++;
                numOfPixels++;

            }
            
            /* Return if the counting has exeeded the boundary box of the
            ** starting blob. */
            if (!skipBoundary) {

                if (CurrentX < sourceBlob->getXMin()
                        || CurrentX > sourceBlob->getXMax()
                        || CurrentY < sourceBlob->getYMin()
                        || CurrentY > sourceBlob->getYMax()) {

                    return numOfPixels;
                }
            }

            // is the pixel going right AND up?
            if (P > 0) {
                CurrentX += Xincr;  // increment independent variable
                CurrentY += Yincr;  // increment dependent variable
                P += dPru;          // increment decision (for up)

            } else {

                // is the pixel just going right?
                CurrentX += Xincr;  // increment independent variable
                P += dPr;           // increment decision (for right)
            }
        }

    } else {

        // if Y is the independent variable 
        int dPr = dX << 1;        // amount to increment decision if right is chosen (always)
        int dPru = dPr - (dY << 1);        // amount to increment decision if up is chosen
        int P = dPr - dY;                // decision variable start value
        int thisColour;

        // process each point in the line one at a time (just use dY)      
        for (; dY >= 0; dY--) {
            //SetPixel(CurrentX, CurrentY, Color);  // plot the pixel

            thisColour = cplane[CurrentX + CurrentY * WIDTH] & ~MAYBE_BIT;

            if (color == BACKGROUND
                && (thisColour == BACKGROUND || thisColour == 9)) {
                numOfPixels++;

            } else if (thisColour == color) {
                numOfPixels++;

            /* Exception for pink - take red / ball orange as ok. */
            } else if (color == cBEACON_PINK 
                    && (thisColour == cROBOT_RED || thisColour == cBALL)
                    && (pinkNoiseCounter < PINK_NOISE_LIMIT)
                ) {
                    pinkNoiseCounter++;
                    numOfPixels++;

            /* This pixel does not match our interest, return immediately. */ 
            }

            /* Return if the counting has exeeded the boundary box of the
            ** starting blob. */
            if (!skipBoundary) {

                if (CurrentX < sourceBlob->getXMin()
                        || CurrentX > sourceBlob->getXMax()
                        || CurrentY < sourceBlob->getYMin()
                        || CurrentY > sourceBlob->getYMax()) {

                    return numOfPixels;
                }
            }

            // is the pixel going up AND right?
            if (P > 0) {
                CurrentX += Xincr;        // increment dependent variable
                CurrentY += Yincr;        // increment independent variable
                P += dPru;        // increment decision (for up)
            }
            else {
                // is the pixel just going up?
                CurrentY += Yincr;        // increment independent variable
                P += dPr;                 // increment decision (for right)
            }
        }
    }
    return numOfPixels;
}

#ifdef VISION_DEBUG
		void VisualCortex::changeCalibration(int type) {
			gcd.changeCalibration(type);
		}

		void VisualCortex::switchFile(int suff) {
			gcd.switchFile(suff);
		}

		void VisualCortex::updateCalibration(int col, int row, int newcolour) {
			gcd.updateCalibration(col, row, newcolour, yplane);
		}

		void VisualCortex::freezeImageProcessing(int i) {
		}
#endif

		void VisualCortex::set_mc(int x) {
			blobs->multi_colour = x;
		}

		int VisualCortex::get_mc() {
			return blobs->multi_colour;
		}


/* Functions for "Shape Challenge" removed after rev 2701. */



/*********************************************************************/
/* This is called by the constructor of VisualCortex                 */

void VisualCortex::initEdgeDetect() {
    matchBorder = new float[ARRAY_WIDTH * ARRAY_HEIGHT];
    matchLine = new float[ARRAY_WIDTH * ARRAY_HEIGHT];
    matchField = new float[ARRAY_WIDTH * ARRAY_HEIGHT];
    bodyMoving = true;
    headMovement = 0.0;
    mapInvalid = false;
    numEdgePnts = 0;

    initGradField();
}
// This is called by the destructor of VisualCortex

void VisualCortex::shutdownEdgeDetect() {
    delete matchBorder;
    delete matchLine;
    delete matchField;
}


/*********************************************************************/
/* Initialise the match functions (line, border & field), for        */
/* gradient ascent                                                   */

void VisualCortex::initGradField() {

    double CENTER_CIRCLE_RADIUS = 15.0;
    double SPREAD = 0.004; // A bit contradictory, small values => very spread out
    double dist, cur_dist;
    int index, x, y;

    /*
     * Bottom left of the field
     *         
     *        *2*
     *   _____________
     *  |          \__
     *  |           *7* 
     *  |
     *  |*1*
     *  |         *4* 
     *  |        _____
     *  |    *5*|     
     *   \      |
     * *8*\_____|_____
     *      *3*    *6*
     *          
     */
     
    for (int yInd = -offsetY; yInd <= FIELD_LENGTH / 4; yInd++) {
        for (int xInd = -offsetX; xInd <= FIELD_WIDTH / 4; xInd++) {
            index = (yInd + offsetY) * ARRAY_WIDTH + (xInd + offsetX);
            x = 2 * xInd; y = 2 * yInd;
        
            /*************
             *           *
             * LINE EDGE *
             *           *
             *************/
         
            // line 2 - mid-line
            if (x < WALL_THICKNESS)
                dist = distanceSq(x, y, WALL_THICKNESS, FIELD_LENGTH/2);
            else
                dist = SQUARE(y - FIELD_LENGTH/2);


            // line 4 - goalbox line horizontal
            if (x < FIELD_WIDTH/2 - GOALBOX_WIDTH/2)
                cur_dist = distanceSq(x, y, FIELD_WIDTH/2 - GOALBOX_WIDTH/2,
                                            GOALBOX_DEPTH + WALL_THICKNESS);
            else
                cur_dist = SQUARE(y - GOALBOX_DEPTH - WALL_THICKNESS);       

            dist = min(cur_dist, dist);

 
            // line 5 - goalbox line vertical
            if (y < WALL_THICKNESS)            
                cur_dist = distanceSq(x, y, FIELD_WIDTH/2 - GOALBOX_WIDTH/2,
                                            WALL_THICKNESS);
            else if (y > GOALBOX_DEPTH + WALL_THICKNESS)
                cur_dist = distanceSq(x, y, FIELD_WIDTH/2 - GOALBOX_WIDTH/2,
                                            GOALBOX_DEPTH + WALL_THICKNESS);
            else 
                cur_dist = SQUARE(FIELD_WIDTH/2 - GOALBOX_WIDTH/2 - x);

            dist = min(cur_dist, dist);


            // line 6 - goal line horizontal INNER         
            if (x < FIELD_WIDTH/2 - GOAL_WIDTH/2)
                cur_dist = distanceSq(x, y, FIELD_WIDTH/2 - GOAL_WIDTH/2, WALL_THICKNESS);
            else
                cur_dist = SQUARE(y - WALL_THICKNESS);
          
            cur_dist *= 4;
            dist = min(cur_dist, dist);


            // line 7 - quarter circle
            cur_dist = sqrt(distanceSq(x, y, FIELD_WIDTH/2, FIELD_LENGTH/2));
            cur_dist = SQUARE(cur_dist - CENTER_CIRCLE_RADIUS);
            cur_dist *= 4;
          
            dist = min(cur_dist, dist);
            
            matchLine[index] = (float)(1.0 / (1.0 + SPREAD * dist));
          

            /***************
             *             *
             * BORDER EDGE *
             *             *
             ***************/
          
            // line 1 - left border
            if (y < CORNER_WIDTH)            
                dist = distanceSq(x, y, WALL_THICKNESS, CORNER_WIDTH);
            else 
                dist = (x - WALL_THICKNESS) * (x - WALL_THICKNESS);
 
          
            // line 3 - bottom border 
            if (x < CORNER_WIDTH)
                cur_dist = distanceSq(x, y, CORNER_WIDTH, WALL_THICKNESS);
            else if (x > FIELD_WIDTH/2 - GOAL_WIDTH/2)
                cur_dist = distanceSq(x, y, FIELD_WIDTH/2 - GOAL_WIDTH/2, WALL_THICKNESS);
            else
                cur_dist = SQUARE(y - WALL_THICKNESS);

            dist = min(cur_dist, dist);


            // line 8 - corner border
            if (y - x - CORNER_WIDTH + WALL_THICKNESS > 0)
                cur_dist = distanceSq(x, y, WALL_THICKNESS, CORNER_WIDTH);
            else if (y - x + CORNER_WIDTH - WALL_THICKNESS < 0)
                cur_dist = distanceSq(x, y, CORNER_WIDTH, WALL_THICKNESS);
            else {
                double x0 = WALL_THICKNESS, y0 = CORNER_WIDTH;
                double x1 = CORNER_WIDTH,   y1 = WALL_THICKNESS;
                double dx = x1 - x0,        dy = y1 - y0;
                double numerator, denominator;

                numerator = dy * (x - x0) - dx * (y - y0);
                numerator *= numerator;
                denominator = dx * dx + dy * dy;
                    
                cur_dist = numerator / denominator;
            }

            dist = min(cur_dist, dist);

            // little border in the goal
            if (y < 0)
                cur_dist = distanceSq(x, y, FIELD_WIDTH/2 - GOAL_WIDTH/2, 0);
            else if (y > WALL_THICKNESS)
                cur_dist = distanceSq(x, y, FIELD_WIDTH/2 - GOAL_WIDTH/2, WALL_THICKNESS);
            else
                cur_dist = SQUARE(x - (FIELD_WIDTH/2 - GOAL_WIDTH/2));

            dist = min(cur_dist, dist);
    
            // Long outer border (for Portugal 2004 fields)
            // How far outside game field the outer border is (cm)
            static const int outerBorderDist = 110;
            static const int borderLimit = -WALL_THICKNESS - 40;
            
            if (y > -outerBorderDist)
                cur_dist = SQUARE(x + outerBorderDist);
            else
                cur_dist = distanceSq(x, y, -outerBorderDist, -outerBorderDist);
                
            dist = min(cur_dist, dist);
            
            if (x > -outerBorderDist)
                cur_dist = SQUARE(y + outerBorderDist);
            else
                cur_dist = distanceSq(x, y, -outerBorderDist, -outerBorderDist);
                
            dist = min(cur_dist, dist);

            matchBorder[index] = (float)(1.0 / (1.0 + SPREAD * dist));
            
            /*********
             *       *
             * FIELD *
             *       *
             *********/

            double LARGE_DIST = 10000000.0;           
            dist = LARGE_DIST; 

            // line 1 - left border
            if (x < WALL_THICKNESS && y > CORNER_WIDTH) {
                cur_dist = SQUARE(x - WALL_THICKNESS);
                dist = min(cur_dist, dist);
            }
                  
            if (y < WALL_THICKNESS) {
                if (x > CORNER_WIDTH && x < FIELD_WIDTH/2 - GOAL_WIDTH/2) {
                    // line 3 - bottom border
                    cur_dist = (y - WALL_THICKNESS) * (y - WALL_THICKNESS);
                    dist = min(cur_dist, dist);
                    
                    // Goal side
                    if (y <= -GOAL_LENGTH)
                        cur_dist = distanceSq(x, y, FIELD_WIDTH/2 - GOAL_WIDTH/2,
                            -GOAL_LENGTH);
                    else
                        cur_dist = SQUARE(x - (FIELD_WIDTH/2 - GOAL_WIDTH/2));

                    dist = min(cur_dist, dist);

                } else if (x >= FIELD_WIDTH/2 - GOAL_WIDTH/2) {

                    // Back of goal
                    if (y < -GOAL_LENGTH) {
                        cur_dist = SQUARE(y + GOAL_LENGTH);
                        dist = min(cur_dist, dist);
                    }
                }
            }

            if (x + y - (CORNER_WIDTH + WALL_THICKNESS) <= 0) {
                // line 8 - corner border
                if (y - x - CORNER_WIDTH + WALL_THICKNESS > 0)
                    cur_dist = distanceSq(x, y, WALL_THICKNESS, CORNER_WIDTH);
                else if (y - x + CORNER_WIDTH - WALL_THICKNESS < 0)
                    cur_dist = distanceSq(x, y, CORNER_WIDTH, WALL_THICKNESS);
                else {
                    double x0 = WALL_THICKNESS, y0 = CORNER_WIDTH;
                    double x1 = CORNER_WIDTH,   y1 = WALL_THICKNESS;
                    double dx = x1 - x0,        dy = y1 - y0;
                    double numerator, denominator;
    
                    numerator = dy * (x - x0) - dx * (y - y0);
                    numerator *= numerator;
                    denominator = dx * dx + dy * dy;
                    
                    cur_dist = numerator / denominator;
                }
                
                dist = min(cur_dist, dist);
            }

            if (dist >= LARGE_DIST) dist = 0.0;
            
            if (x < -outerBorderDist || y < -outerBorderDist) {
                if (x < -outerBorderDist && y < -outerBorderDist)
                    cur_dist = distanceSq(x, y, -outerBorderDist, -outerBorderDist);

                else if (x < -outerBorderDist)
                    cur_dist = min(SQUARE((double)x + outerBorderDist), cur_dist);
                    
                else
                    cur_dist = min(SQUARE((double)y + outerBorderDist), cur_dist);

            } else if (x > borderLimit && y > borderLimit){
                cur_dist = min(SQUARE(x - borderLimit), SQUARE(y - borderLimit));

            } else cur_dist = 0.0;
            
            dist = min(cur_dist, dist);
                    
            matchField[index] = (float)(1.0 / (1.0 + SPREAD * 0.5 * dist));
        }
    }
    
    mirrorGradient(matchLine);
    mirrorGradient(matchField);
    mirrorGradient(matchBorder);
    
    // Derrick: for debugging - this outputs match tables to file (use with
    // offvision). Produces .raw files in offvision directory after process is
    // run. Image sizes are width 346 height 421 (as of 30/6/2004). Use rawtopgm
    // and then ppmtotga / ppmtobmp / ppmtojpeg on each file to view in an
    // image viewer. eg. At shell, type:
    // > rawtopgm 346 421 LineMap.Raw > LineMap.pgm
    // > ppmtotga LineMap.pgm > LineMap.tga
    
    /*if (flag_OFFLINE) {
        uchar t;    
        ofstream fout;
    
        fout.open("LineMap.Raw", ios::out);        
        for (int i = 0; i < ARRAY_WIDTH * ARRAY_HEIGHT; i++) {
            t = (uchar)(255 * matchLine[i]);
            fout << t;
        }    
        fout.close();
    
        fout.open("BordMap.Raw", ios::out);
        for (int i = 0; i < ARRAY_WIDTH * ARRAY_HEIGHT; i++) {
            t = (uchar)(255 * matchBorder[i]);
            fout << t;
        }
        fout.close();

        fout.open("FieldMap.Raw", ios::out);
        for (int i = 0; i < ARRAY_WIDTH * ARRAY_HEIGHT; i++) {
            t = (uchar)(255 * matchField[i]);
            fout << t;
        }
        fout.close();
    }*/
}

void VisualCortex::mirrorGradient(float *gradient)
{
    int x, y;
    float val;
    
    for (y = 0; y <= ARRAY_HEIGHT / 2; y++) {
        for (x = 0; x <= ARRAY_WIDTH / 2; x++) {

            val = gradient[y * ARRAY_WIDTH + x];

            // top left corner
            gradient[(ARRAY_HEIGHT - 1 - y) * ARRAY_WIDTH + x] = val;
            
            // top right corner
            gradient[(ARRAY_HEIGHT - 1 - y) * ARRAY_WIDTH + (ARRAY_WIDTH - 1 - x)] = val;

            // bottom left corner
            gradient[y * ARRAY_WIDTH + (ARRAY_WIDTH - 1 - x)] = val;
        }
    }
}

// Returns the square of the distance between two (integer) points
 
inline double VisualCortex::distanceSq(int x1, int y1, int x2, int y2)
{
    double dx = (double)x2 - x1;
    double dy = (double)y2 - y1;

    return dx * dx + dy * dy;
}

/*****************************************************************************/
/* Called by ProcessImage.                                                   */

void VisualCortex::runEdgeDetect() {
    numEdgePnts = 0;

    // Don't look for edge pixels if the head is moving too fast, since fast
    // head rotations can cause image distortion, such as straight lines
    // looking curved.
    if (headMovement < 0.09) {
        // Edge points pixels will be stored in the edgeList and number of
        //  points will be returned.
        numEdgePnts = findFieldEdges();

        for (int i = 0; i < numEdgePnts; i ++) {
            edgePntTypes[i] = edgeList[i * 3 + 2];
            cplaneEdgePnts[i * 2] = edgeList[i * 3] - WIDTH / 2;
            cplaneEdgePnts[i * 2 + 1] = -(edgeList[i * 3 + 1] - HEIGHT / 2);
            edgeWeights[i] = 0.0;
        }

        // Project the edge pixels into the ground plane.
        projectPoints(cplaneEdgePnts, numEdgePnts, localEdgePnts, groundPlane);
        calcEdgeWeights();
    }
    
    // If we're not moving, keep track of old points
    stationaryMap();    
}

/*****************************************************************************/
/* Calculates the match value and gradient, given a robot position & heading.*/
/* Note: Outputs gradient to 'gradient', returns match value.                */

double VisualCortex::calcRobotMatch(MVec3 &robot, MVec3 &gradient)
{
    double totalLineMatch, totalFieldMatch;
    double localGX, localGY, dX, dY, match;
    MVec3 lineGrad, fieldGrad, grad;

    localToGlobal(robot, numEdgePnts, localEdgePnts, globalEdgePnts);
    
    lineGrad(0, 0) = 0.0; fieldGrad(0, 0) = 0.0;
    lineGrad(1, 0) = 0.0; fieldGrad(1, 0) = 0.0;
    lineGrad(2, 0) = 0.0; fieldGrad(2, 0) = 0.0;
    totalLineMatch = 0.0; totalFieldMatch = 0.0;
            
    int i = 0;
    for (int c = 0; c < numEdgePnts; c++) {
        match = calcPointMatch(globalEdgePnts[i], globalEdgePnts[i + 1],
                        edgePntTypes[c], &localGX, &localGY);
                        
        match *= edgeWeights[c];
        localGX *= edgeWeights[c];
        localGY *= edgeWeights[c];
        
        dX = globalEdgePnts[i] - robot(0, 0);
        dY = globalEdgePnts[i + 1] - robot(1, 0);

        grad(0, 0) = localGX;
        grad(1, 0) = localGY;
        grad(2, 0) = dX * localGY - dY * localGX;
        
        if (edgePntTypes[c] == FIELD) {
            totalFieldMatch += match;
            fieldGrad += grad;
        } else {
            totalLineMatch += match;
            lineGrad += grad;
        }
                                   
        i += 2;
    }

    if (totalLineWeight > 0.0) {
        totalLineMatch /= totalLineWeight;
        lineGrad /= totalLineWeight;
    } else totalLineMatch = 1.0;

    if (totalFieldWeight > 0.0) {
        totalFieldMatch /= totalFieldWeight;
        fieldGrad /= totalFieldWeight;
    } else totalFieldMatch = 1.0;

    // The actual match value is the product of the field and line match values
    match = totalLineMatch * totalFieldMatch;
    fieldGrad *= totalLineMatch;
    lineGrad *= totalFieldMatch;
    gradient = fieldGrad;
    gradient += lineGrad;
    
    return match;
}

/*****************************************************************************/
/* Calculates the match value and local gradient of a single point           */
/*  (specified by x, y, type) on the field, outputs gradient to gX, gY,      */ 
/*  returns match value.                                                     */

inline double VisualCortex::calcPointMatch(double inX, double inY, int type,
                                           double *gX, double *gY)
{
    double mX[3], mY[3], dMdX[2], dMdY[2];
    double x, y, biasX, biasY, match;
    int iX, iY, c, index;

    x = inX * 0.5; y = inY * 0.5;
    iX = (int)(x + 0.5); iY = (int)(y + 0.5);
    biasX = x - iX;      biasY = y - iY;
    iX += offsetX;       iY += offsetY;

    (*gX) = 0.0; (*gY) = 0.0;

    float *table;
    table = NULL;
    if (type == FIELD) table = matchField;
    if (type == LINE_EDGE) table = matchLine;
    if (type == BORDER_EDGE) table = matchBorder;

    if (table == NULL)
        return INFINITY_GRADIENT_VALUE;
    
    if (iX < -1 || iX > (ARRAY_WIDTH + 1) || iY < -1 || iY > (ARRAY_HEIGHT + 1))
        return INFINITY_GRADIENT_VALUE;

    // Retrieve match values along horizontal line
    index = iY * ARRAY_WIDTH + (iX - 1);
    for (c = -1; c <= 1; c++) {
    
        if ((iX + c) >= 0 && (iX + c) < ARRAY_WIDTH && iY >= 0 && iY < ARRAY_HEIGHT)
             mX[c + 1] = (double)table[index];
        else mX[c + 1] = INFINITY_GRADIENT_VALUE;

        index++;
    }

    // Retrieve match values along vertical line
    index = (iY - 1) * ARRAY_WIDTH + iX;
    for (c = -1; c <= 1; c++) {
    
        if (iX >= 0 && iX < ARRAY_WIDTH && (iY + c) >= 0 && (iY + c) < ARRAY_HEIGHT)
             mY[c + 1] = (double)table[index];
        else mY[c + 1] = INFINITY_GRADIENT_VALUE;

        index += ARRAY_WIDTH;
    }

    // Calculate match gradients
    for (c = 0; c < 2; c++) {
        dMdX[c] = mX[c + 1] - mX[c];
        dMdY[c] = mY[c + 1] - mY[c];
    }
    
    (*gX) = (0.5 - biasX) * dMdX[0] + (0.5 + biasX) * dMdX[1];
    (*gY) = (0.5 - biasY) * dMdY[0] + (0.5 + biasY) * dMdY[1];
    (*gX) *= 0.5; (*gY) *= 0.5;
    
    // Calculate match value
    match = mX[1];

    if (biasX < 0.0) match += biasX * dMdX[0];
    else match += biasX * dMdX[1];

    if (biasY < 0.0) match += biasY * dMdY[0];
    else match += biasY * dMdY[1];

    return match;
}

/*****************************************************************************/
/* Weights points according to their distance and culls points that are too  */
/*  far away (usually these are noise)                                       */

void VisualCortex::calcEdgeWeights()
{
    static const double reduceDist = 70.0;
    static const double cullDist = 350.0;
    static const int minPoints = 4;

    double dx, dy, d;
    int numPnts, typeCount[3];
    bool useType[2];
    
    // First count the number of points of each type
    typeCount[FIELD] = 0;
    typeCount[LINE_EDGE] = 0;
    typeCount[BORDER_EDGE] = 0;
    for (int i = 0; i < numEdgePnts; i++)
        typeCount[edgePntTypes[i]]++;

    // If there are too few points of one type, don't use that type
    for (int i = 0; i < 3; i++)
        useType[i] = (typeCount[i] >= minPoints);
    
    numPnts = 0;
    totalLineWeight = 0.0; totalFieldWeight = 0.0;
    for (int i = 0; i < numEdgePnts; i++) {
        // Calculate distance d from robot to point
        dx = localEdgePnts[i * 2];
        dy = localEdgePnts[i * 2 + 1];
        d = sqrt(dx * dx + dy * dy);

        if (edgePntTypes[i] != FIELD) {
            // For field lines/borders, the weight for each point decreases as
            // distance increases, since projection accuracy decreases with
            // distance.
            if (d < reduceDist)
                edgeWeights[numPnts] = 1.0;
            else if (d < cullDist)
                edgeWeights[numPnts] = (cullDist - d) / (cullDist - reduceDist);

        } else {
            // Field points are constant weight because they don't need to be
            // particalarly accurate.
            edgeWeights[numPnts] = 1.0;
        }
            
        // Only use point if it is within the culling distance
        if (d < cullDist && useType[edgePntTypes[i]]) {
            
            if (numPnts < i) {
                localEdgePnts[numPnts * 2] = localEdgePnts[i * 2];
                localEdgePnts[numPnts * 2 + 1] = localEdgePnts[i * 2 + 1];
                edgePntTypes[numPnts] = edgePntTypes[i];
            }

            if (edgePntTypes[i] != FIELD)
                totalLineWeight += edgeWeights[numPnts];
            else
                totalFieldWeight += edgeWeights[numPnts];

            numPnts++;
        }
    }

    numEdgePnts = numPnts;
}

/*****************************************************************************/
/* Calculates the centre of the matched projected points                     */

void VisualCortex::calcEdgeCentre(MVec2 &projEdgeCentre)
{
    double x, y, weight, total, gX, gY, closeDist;
    int t, closest;

    total = 0.0;
    projEdgeCentre(0, 0) = 0.0;
    projEdgeCentre(1, 0) = 0.0;
    
    // Calculate average centre point
    for (int i = 0; i < numEdgePnts; i++) {
        t = edgePntTypes[i];
        x = globalEdgePnts[i * 2];
        y = globalEdgePnts[i * 2 + 1];
        
        if (t != FIELD) {
            weight = calcPointMatch(x, y, t, &gX, &gY) * edgeWeights[i];

            projEdgeCentre(0, 0) += weight * x;
            projEdgeCentre(1, 0) += weight * y;
            total += weight;
        }
    }
    
    if (total > 0) projEdgeCentre /= total;
    
    // Find projected point that is closest to the average centre
    closest = -1;
    closeDist = 0.0;
    for (int i = 0; i < numEdgePnts; i++) {
        t = edgePntTypes[i];
        x = globalEdgePnts[i * 2] - projEdgeCentre(0, 0);
        y = globalEdgePnts[i * 2 + 1] - projEdgeCentre(1, 0);
        weight = x * x + y * y;
        
        if ((weight < closeDist || closest == -1) && t != FIELD) {
            closest = i;
            closeDist = weight;
        }
    }

    // Set the centre to closest point
    projEdgeCentre(0, 0) = globalEdgePnts[closest * 2];
    projEdgeCentre(1, 0) = globalEdgePnts[closest * 2 + 1];
}

/*****************************************************************************/
/* If the dog is standing still, instead of throwing away points from        */
/*  previous frames, we build up a map from them                             */

static const int maxMapConf = 100;

void VisualCortex::stationaryMap()
{
    int index, localInd, oldInd, pT, pC, failedAdds;
    double mergeRad, scaler, pX, pY, pDistSq;
    bool pointUsed;

    // We have to restart map building if the robot moves position
    if (bodyMoving || mapInvalid) {
        numMapPnts = 0;
        mapInvalid = false;
        return;
    }

    failedAdds = 0;    
    // Incorporate edge pixel data into map
    for (localInd = 0; localInd < numEdgePnts; localInd++) {
        pointUsed = false;

        pC = 1;
        pT = edgePntTypes[localInd];
        pX = localEdgePnts[localInd * 2];
        pY = localEdgePnts[localInd * 2 + 1];
        oldInd = -1;
        
        // Check if current point is close to a map point
        while (shouldMerge(pX, pY, pT, oldInd, &index)) {
            // If current point matches a map point, merge it
            pX = pX * pC + mapPnts[index * 2] * mapPntConfidence[index];
            pY = pY * pC + mapPnts[index * 2 + 1] * mapPntConfidence[index];
            pC += mapPntConfidence[index];
            pX /= pC; pY /= pC;
            pC = min(pC, maxMapConf);
            
            mapPnts[index * 2] = pX;
            mapPnts[index * 2 + 1] = pY;
            mapPntConfidence[index] = pC;
                
            if (pointUsed) mapPntConfidence[oldInd] = 0;
            oldInd = index;
            pointUsed = true;
        }
        
        // If point could not be merged, insert it into map array
        if (!pointUsed) {
            // Otherwise add a new map point
            if (index < 0 && numMapPnts < MAX_EDGE_PIXELS) {
                index = numMapPnts;
                numMapPnts++;
            }
            
            if (index >= 0) {
                mapPntConfidence[index] = pC;
                mapPntTypes[index] = pT;
                mapPnts[index * 2] = pX;
                mapPnts[index * 2 + 1] = pY;
            } else failedAdds++;
        }
    }
    
    // If we couldn't add points, then decrease confidence of all map points
    if (failedAdds > 0) {
        for (int i = 0; i < numMapPnts; i++) {
            if (mapPntConfidence[i] <= failedAdds)
                mapPntConfidence[i] = 0;
            else mapPntConfidence[i] -= failedAdds;
        }
    }

    // Copy map points to local points array
    totalLineWeight = 0.0; totalFieldWeight = 0.0;
    numEdgePnts = 0;
    for (int i = 0; i < numMapPnts; i++) {
        if (mapPntConfidence[i] > 0) {
            pX = mapPnts[i * 2];
            pY = mapPnts[i * 2 + 1];
            localEdgePnts[numEdgePnts * 2] = pX;
            localEdgePnts[numEdgePnts * 2 + 1] = pY;
            edgePntTypes[numEdgePnts] = mapPntTypes[i];
            edgeWeights[numEdgePnts] = mapPntConfidence[i];

            if (edgePntTypes[numEdgePnts] == FIELD)
                totalFieldWeight += edgeWeights[numEdgePnts];
            else
                totalLineWeight += edgeWeights[numEdgePnts];

            numEdgePnts++;
        }
    }
}

inline bool VisualCortex::shouldMerge(double x, double y, int t, int dontMergeWith, int *mergeInd)
{
    double mX, mY, mDSq, dX, dY, dSq, dotProd, cosSq, mergeRad;
    int mapInd;

    dSq = x * x + y * y;
    if (t == FIELD) mergeRad = 20.0 + 0.2 * sqrt(dSq);
    else mergeRad = 7.0 + 0.15 * sqrt(dSq);
    mergeRad *= mergeRad;

    (*mergeInd) = -1;
    for (mapInd = 0; mapInd < numMapPnts; mapInd++) {
        if (mapPntConfidence[mapInd] == 0) {
            (*mergeInd) = mapInd;
            
        } else if (t == mapPntTypes[mapInd]) {
            // Calculate square distance between local and map points
            mX = mapPnts[mapInd * 2];
            mY = mapPnts[mapInd * 2 + 1];
            dX = mX - x; dY = mY - y;
                
            if ((dX * dX + dY * dY) <= mergeRad) {
                // Calculate cosine of angle between local and map points
                dotProd = x * mX + y * mY;
                mDSq = mX * mX + mY * mY;
                cosSq = dotProd * dotProd / (dSq * mDSq);

                // Don't want to do angle check for field points
                if ((t == FIELD || cosSq > 0.997) && mapInd != dontMergeWith) {
                    (*mergeInd) = mapInd;
                    return true;
                }
            }
        }
    }
    
    return false;
}

void VisualCortex::invalidateEdgeMap()
{
    mapInvalid = true;
}

/*****************************************************************************/
/* Adds in edge pixel to the edgeList, but only if it passes some checks     */

bool VisualCortex::tryAddEdge(int x, int y, int t, int *numEdges, int dir)
{
    int odir = 1 - dir;
    uchar cPixel;

    // Make sure point is not inside a dog or ball vob
    for (int i = vobBall; i <= vobBlueDog4; i++) {
        if (vob[i].cf > 0) {
            if (x >= vob[i].x && x <= (vob[i].x + vob[i].width) &&
                y >= vob[i].y && y <= (vob[i].y + vob[i].height))
                
                return true;
        }
        
        if (i == vobBall) i = vobRedDog - 1;
    }

    // Only add point if its not in corner of c-plane (lots of noise there)
    int lineConst = 10; // How much of the corner to chop off (in pixels)
    
    for (int cornerY = 0; cornerY <= CPLANE_HEIGHT; cornerY += CPLANE_HEIGHT) {
        for (int cornerX = 0; cornerX <= CPLANE_WIDTH; cornerX += CPLANE_WIDTH) {
            int xVal = abs(x - cornerX);
            int yVal = abs(y - cornerY);

            if ((xVal + yVal - lineConst) < 0) return true;
        }
    }

    int pos[2], greenCount = 0;

    if (t == LINE_EDGE) {
        // If this pixel is part of a line, check for white (appoximately)
        // along the line, ie. perpendicular to the scan direction
        
        for (int i = -1; i <= 1; i += 2) {
            pos[0] = x; pos[1] = y;
            pos[odir] += i;
 
            if (pos[0] >= 0 && pos[0] < CPLANE_WIDTH &&
                pos[1] >= 0 && pos[1] < CPLANE_HEIGHT) {
    
                cPixel = cplane[pos[0] + pos[1] * CPLANE_WIDTH] & COLOR_MASK;
                if (cPixel == cFIELD_GREEN) greenCount++;
            }
        }

        if (greenCount > 0) return true;
    
    } else if (t == BORDER_EDGE) {
        // If the pixel is part of a border, then it must contain alot of
        // white in one of the perpendicular directions as well as the 
        // current scan direction

        static const int checkLength = 15;
        int edge[2], intercept;
        bool onScreen;
        
        edge[0] = -1;
        if (odir == 0) edge[1] = CPLANE_WIDTH;
        else edge[1] = CPLANE_HEIGHT;

        pos[0] = x; pos[1] = y;
        onScreen = horizonIntercept(pos[dir], dir, &intercept);
        if (onScreen) {
            if (intercept < pos[odir]) edge[0] = intercept;
            else edge[1] = intercept;
        }
        
        if (edge[0] < (pos[odir] - checkLength))
            edge[0] = pos[odir] - checkLength;
        
        if (edge[1] > (pos[odir] + checkLength))
            edge[1] = pos[odir] + checkLength;
        
        for (; pos[odir] > edge[0]; pos[odir]--) {
            cPixel = cplane[pos[0] + pos[1] * CPLANE_WIDTH] & COLOR_MASK;

            if (cPixel == cFIELD_GREEN) {
                greenCount++;
                pos[odir] = edge[0];
            }
        }

        if (dir == 1) {
            pos[0] = x; pos[1] = y;
            for (; pos[odir] < edge[1]; pos[odir]++) {
                cPixel = cplane[pos[0] + pos[1] * CPLANE_WIDTH] & COLOR_MASK;
            
                if (cPixel == cFIELD_GREEN) {
                    greenCount++;
                    pos[odir] = edge[1];
                }
            }

            if (greenCount > 1) return true;

        // If we ran into the border while scanning horizontally, we should
        // check upwards (only) for white
        } else if (greenCount > 0) return true;
    
    } else if (t == FIELD) {
        // If the pixel is field, it must contain a lot of green in both
        // perpendicular scan directions
        
        static const int checkLength = 15;
        int screenMax, max;
        
        if (odir == 0) screenMax = CPLANE_WIDTH;
        else screenMax = CPLANE_HEIGHT;
                
        pos[0] = x; pos[1] = y;
        max = pos[odir] + checkLength + 1;
        if (max > screenMax) return true;
        
        for (pos[odir]++; pos[odir] < max; pos[odir]++) {
            cPixel = cplane[pos[0] + pos[1] * CPLANE_WIDTH] & COLOR_MASK;
            if (cPixel == cFIELD_GREEN) greenCount++;
        }
            
        pos[0] = x; pos[1] = y;
        max = pos[odir] - checkLength - 1;
        if (max < 0) return true;
        
        for (pos[odir]--; pos[odir] > max; pos[odir]--) {
            cPixel = cplane[pos[0] + pos[1] * CPLANE_WIDTH] & COLOR_MASK;
            if (cPixel == cFIELD_GREEN) greenCount++;
        }
        
        if (greenCount < (2 * checkLength - 2)) return true;
    }

    // If point passed the tests, add it
    edgeList[(*numEdges) * 3 + 0] = x;
    edgeList[(*numEdges) * 3 + 1] = y;
    edgeList[(*numEdges) * 3 + 2] = t;
    (*numEdges)++;

    return ((*numEdges) < MAX_EDGE_PIXELS);
}

/*****************************************************************************/
/* This scans the c-plane in a series of horizontal and vertical lines,      */
/* searching for green to white, or white to green transitions that mark the */
/* field lines or edges. All transition pixels will be stored in the edgeList*/
/* and number in the array  will be returned.                                */

int VisualCortex::findFieldEdges()
{
    static const int SOLID_GREEN_THRESHOLD = 2;
    static const int NOISE_THRESHOLD = 2;
    static const int MIN_BORDER_WIDTH = 25;
    static const int MAX_LINE_WIDTH   = 18;

    bool green, white, edgeStarted, fieldStarted, intOffScreen, useFieldPoints;
    int pos[2], max[2], st[2], field[2], dir, odir, numEdges;
    int greenSeen, greenSinceField, othersSeen, eType;
    int stop, intercept, fieldInterval, edgeInterval;
    double in;
    uchar cPixel, previous;

    numEdges = 0;
    max[0] = CPLANE_WIDTH;
    max[1] = CPLANE_HEIGHT;

    // Scan both horizontal and vertical lines
    for (dir = 0; dir < 2; dir++) {
        odir = 1 - dir;

        for (pos[odir] = 0; pos[odir] < max[odir]; pos[odir] += GRID_SPACING) {
            // Only scan as far as horizon permits
            intercept = 0;
            intOffScreen = !horizonIntercept(pos[odir], odir, &intercept);

            pos[dir] = max[dir] - 1;
            stop = 0;

            if (!intOffScreen) {
                if (isBelowHorizon(pos[0], pos[1], 0)) stop = intercept;
                else pos[dir] = intercept;
            } else if (!isBelowHorizon(pos[0], pos[1], 0)) stop = max[dir];
            
            // Record field green points only occasionally
            useFieldPoints = (dir == 1 && pos[odir] % (5 * GRID_SPACING) == 0);
            
            // Scan along line, checking for white/green transitions
            greenSeen = 0;
            othersSeen = 0;
            greenSinceField = 0;
            previous = cNONE;
		    edgeStarted = false;
            edgeInterval = 0;

            for (; pos[dir] >= stop; pos[dir]--) {
                cPixel = cplane[pos[1] * CPLANE_WIDTH + pos[0]] & COLOR_MASK;
                
                // Count number of consecutive green pixels
                if (cPixel == cFIELD_GREEN) greenSeen++;
                else greenSeen = 0;
                
                // Count number of non-green & non-white pixels
                if (cPixel != cFIELD_GREEN
                    && (cPixel & COLOR_MASK) != cWHITE)
                    othersSeen++;
                else othersSeen = 0;
                
                green = (greenSeen >= SOLID_GREEN_THRESHOLD);
                white = (cPixel == cWHITE);
                
                if (othersSeen > NOISE_THRESHOLD) previous = cNONE;
                
                // Check for green -> white transitions
                if (previous == cFIELD_GREEN && white) {
                    st[0] = pos[0]; st[1] = pos[1];
                    edgeStarted = true;
                    edgeInterval = 0;
                }
                
                // Check for white -> green transitions
                if (previous == cWHITE && green) {
                    bool addEdge = true;
                
                    if (edgeStarted) {
                        if (edgeInterval <= MAX_LINE_WIDTH) {
                            // For white lines thin enough to be classified as a
                            // field line, put the edge pixel in the centre of
                            // the line.
                            st[dir] += pos[dir];
                            st[dir] += SOLID_GREEN_THRESHOLD;
                            st[dir] /= 2;
                            eType = LINE_EDGE;
                         
                        } else if (edgeInterval > MIN_BORDER_WIDTH) {
                            // If the white line was fat enough to be classified
                            // as a field border, we have to decide which side
                            // of the line to put the edge pixel on. When
                            // scanning up the cplane, always take the lower
                            // one. When scanning across take the closer of the
                            // two points to the centre.
                            if (dir == 0) {
                                int endLinePos, startLineDist, endLineDist;
                                endLinePos = pos[dir] + SOLID_GREEN_THRESHOLD;
                                endLineDist = abs(endLinePos - max[dir] / 2);
                                startLineDist = abs(st[dir] - max[dir] / 2); 
                            
                                if (endLineDist < startLineDist) {
                                    st[0] = pos[0];
                                    st[1] = pos[1];
                                    st[dir] += SOLID_GREEN_THRESHOLD;
                                }
                            }

                            eType = BORDER_EDGE;
                        
                        // Don't add points if unsure if it is a line or border.    
                        } else addEdge = false;

                    // If we are scanning from bottom of cplane up, we will not
                    // be able to tell if the first white->green transition
                    // (without a preceeding green->white transition) is a field
                    // border or line, so don't use it.
                    } else if (dir != 1) {
                        st[0] = pos[0];
                        st[1] = pos[1];
                        st[dir] += SOLID_GREEN_THRESHOLD;
                        eType = BORDER_EDGE;
                    
                    } else addEdge = false;

                    if (addEdge) {
                        if(!tryAddEdge(st[0], st[1], eType, &numEdges, dir))
                            return numEdges;
                    }
                
                    edgeStarted = false;
                }

                // Check for substantial amounts of solid green, and mark them
                // as field green (only do this in one direction, so we don't
                // get too many points)
                if (useFieldPoints) {

                    // Record more field points as we get closer to the horizon
                    // or top of the screen.
                    fieldInterval = min((pos[dir] - stop) >> 2, 25);
                
                    if (greenSeen > fieldInterval) {
                        if (greenSinceField == 0) {
                            field[0] = pos[0]; field[1] = pos[1];

                        } else if (greenSinceField > fieldInterval) {
                            if(!tryAddEdge(field[0], field[1], FIELD, &numEdges, dir))
                                return numEdges;
    
                            field[0] = pos[0]; field[1] = pos[1];
                            greenSinceField = 0;
                        }
                    
                        greenSinceField++;
                    
                    } else greenSinceField = 0;
                }
        
                if (green) previous = cFIELD_GREEN;
                if (white) previous = cWHITE;
                edgeInterval++;
            }
            
            if (edgeStarted) {
                if(!tryAddEdge(st[0], st[1], BORDER_EDGE, &numEdges, dir))
                    return numEdges;            
            }
        }
    }

    // For each edge pixel found, mark it on the c-plane
    for (int i = 0; i < numEdges * 3; i++) {
        int x = edgeList[i++]; int y = edgeList[i++];
		int type = edgeList[i];

        if (type == BORDER_EDGE) cplane[y * CPLANE_WIDTH + x] = cFIELD_BORDER;
        else if (type == LINE_EDGE) cplane[y * CPLANE_WIDTH + x] = cFIELD_LINE;
        else if (type == FIELD) cplane[y * CPLANE_WIDTH + x] = cROBOT_GREY;
    }
    
    return numEdges;
}

//Takes inNumPoints points as local co-ordinate pairs in inLocal and transforms them into
//global co-ordinate pairs in outGlobal based on inX, inY and inT. NOTE: outGlobal must be 
//twice inNumPoints in size (as must inLocal) or we'll have problems. This routine 
//works even if inLocal and outGlobal are the same. 

void VisualCortex::localToGlobal(MVec3 &robot, int inNumPoints, double *inLocal, double *outGlobal) {

    //First, we rotate all the points about the origin by an amount equal to inT-90 where
    //positive angle is COUNTERCLOCKWISE. For instance, if we're facing left (inT = 180) we 
    //need to rotate all the points counterclockwise by 90 degrees.
    
    double newAng = robot(2, 0) - pi * 0.5;
    double sinT = sin(newAng);
    double cosT = cos(newAng);
    int i, index = 0;
    
    for (int i = 0; i < inNumPoints; i ++) {
        double in0 = inLocal[index + 0];
        double in1 = inLocal[index + 1];

        outGlobal[index + 0] = (in0 * cosT - in1 * sinT) + robot(0, 0);
        outGlobal[index + 1] = (in0 * sinT + in1 * cosT) + robot(1, 0);
        
        index += 2;
    }
}

/* Calculate the elevation and heading of a cplane point relative to BODY. 
 * Without assuming the distance or height, it's impossible to calculate the
 * angles. */
double xToHeadingFixDist(double cpx, double dist) {
    /* Use inline function in PointProjection.h to get relative to CAMERA
    ** readings. */

    return cpx; // dummy return.
}

double xToHeadingFixHeight(double cpx, double height) {

    return cpx; // dummy return.
}

double yToElevationFixDist(double cpy, double dist){
    return cpy; // dummy return.
}

double yToElevationFixHeight(double cpy, double height){
    return cpy; // dummy return.
}

